//
// cont_sw.cpp
//
// implementation of the synchronization controller (pure simulation)
//
// this version is for BigEndian
// *) includes evalconv.hpp
// *) ExecuteCommand(): byte order of messages is converted
// *) Set/GetParameter(): byte order of messages is converted
// *) ConvertParameter() converts parameter structures
// *) ConvertCommand() converts command structures
//
// created: 25.01.97
//
// last change: 04.09.98
//
// RCS version control information:
//     $Revision: 1.6 $$State: Exp $$Date: 1999/02/05 15:51:42 $
//


#include "ci.hpp"
#include "cont_sw.hpp"
#include "error.hpp"
#include "evalconv.hpp"
#include "globals.hpp"
#include "network.hpp"
#include "sim_glob.hpp"
#include "supervis.hpp"
#include "template.hpp"


#ifdef LINUX_SIM

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>  // also needed for usleep()
#include <stdio.h>

extern EBool g_finished;
// right now the 'tester' process is notified of the end of the simulation
// (as requested from the EvalSys with the command KillSimulation) through
// this variable

extern double g_delayTime;
// the time that the controller sleeps (thus delaying the whole simulation);
// it is set by the main() function (user definable via an option), is given
// in seconds (granularity is microseconds), and is needed to delay the
// simulation, because on a fast processor the events are so close together
// that they clog up the evaluation (XWindow-Server does not update screen
// anymore)

#endif

#ifndef NO_RESOURCE
extern void ClearISRs (SINT32 nodeID, int area);
// the function is defined in interupt.cpp and only available in pure
// simulation if NO_RESOURCE is not defined; in that case, ISRs from
// deleted CSAs have to be removed from the vector table by hand
#endif

// dummy functions for ci_init(); should never be called
int dummy_announce (renvx *, bufx *, INT16, int)
{
    Assert (False);
    return 0;
}
void dummy_error_announce (tenvx *, bufx *, INT16, int, int)
{
    Assert (False);
}

extern int ci_init (int if_num, ci_fn ci_ann_addr, ci_efn ci_ann_err_addr);
// ci_init() is called directly, since the module is only needed in
// pure simulation, where direct calls are possible


extern char *ftoa (double f, int width, char *string);  // from ftoa.cpp
// returns 'f' as a string with at least 'width' characters in %e format;
// stores the string in 'string', which must be big enough to hold the
// converted string; if 'string' is NULL, NULL is returned


//**************************************************************************
// class CSoftwareController: this class controls all clock synchronization
//     applications and the network; for the network it is the link to
//     the simulation
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

CSoftwareController::CSoftwareController (char qname[4], ACSocketAddress *addr)
    : ACController (qname, addr)
{
    networkID_ = 1;  // network IDs start with 1
}

CSoftwareController::~CSoftwareController ()
{
    DeleteAllSimulationObjects();
    // note: the function must be called in every derived class, because
    // it calls StopApplication() if there are any supervisors in the
    // listOfSupervisors, and this function would not be available in the
    // dtor of the base class
}


//--------------------------------------------------------------------------
//--- reset of the simulation
//--------------------------------------------------------------------------

void CSoftwareController::Reset()
{
    networkID_ = 1;  // network IDs again start with 1
    ACController::Reset();  // reset the base class
}


// reset the simulation (set all parameters in all applications and
// networks to their default values); returns True if at least one
// object exists, False if no simulation objects exist
EBool CSoftwareController::ResetTheSimulation()
{
    EBool result;

    for (ACNetwork *network = listOfNetworks.GetFirst();
         network != NULL;
         network = listOfNetworks.GetNext())
    {
        network->Reset();
    } // for

    result = ACController::ResetTheSimulation();  // reset supervisors etc.

    if (!result && !listOfNetworks.Empty()) result = True;
    return result;
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// byte order conversion for controller/network parameters
static void ConvertParameter (SParameter& param)
{
    static char data[512];  // buffer for the output
    static char val1[30], val2[30], val3[30], val4[30], val5[30];
    // buffers for converting the double values

    switch (gi_Convert((INT32)param.type))  // convert byte order
    {
        // SSingleNetworkParameter
        case CrashStatus:
        case PCrash:
        case PRecover:
            Convert ((SSingleNetworkParameter*)&param, True);
            sprintf (data, "module = %04d, type = %04d, netID = %04d, "
                           "value = %s",
                     param.module, param.type,
                     ((SSingleNetworkParameter*)&param)->networkID,
                     ftoa(((SSingleNetworkParameter*)&param)->value, 14, val1));
            break;

        // SNetDistribution
        case DistributionLambda:
        case DistributionOmega:
            Convert ((SNetDistribution*)&param, True);
            sprintf (data, "module = %04d, type = %04d, netID = %04d,\n"
                           "    dist = %04d, low = %s, high = %s, mean = %s, "
                           "constant = %s, freeparam = %s",
                     param.module, param.type,
                     ((SNetDistribution*)&param)->networkID,
                     ((SNetDistribution*)&param)->data.dist,
                     ftoa(((SNetDistribution*)&param)->data.low, 7, val1),
                     ftoa(((SNetDistribution*)&param)->data.high, 7, val2),
                     ftoa(((SNetDistribution*)&param)->data.mean, 7, val3),
                     ftoa(((SNetDistribution*)&param)->data.constant, 7, val4),
                     ftoa(((SNetDistribution*)&param)->data.freeparam,7,val5));
            break;

        // SNetDistributionP2P
        case DistributionEarly:
        case DistributionTimely:
        case DistributionLate:
            Convert ((SNetDistributionP2P*)&param, True);
            sprintf (data, "module = %04d, type = %04d, netID = %04d, "
                           "sender = %04d, receiver = %04d,\n"
                           "    dist = %04d, low = %s, high = %s, mean = %s, "
                           "constant = %s, freeparam = %s",
                     param.module, param.type,
                     ((SNetDistributionP2P*)&param)->networkID,
                     ((SNetDistributionP2P*)&param)->sender,
                     ((SNetDistributionP2P*)&param)->receiver,
                     ((SNetDistributionP2P*)&param)->data.dist,
                     ftoa(((SNetDistributionP2P*)&param)->data.low,7,val1),
                     ftoa(((SNetDistributionP2P*)&param)->data.high,7,val2),
                     ftoa(((SNetDistributionP2P*)&param)->data.mean,7,val3),
                     ftoa(((SNetDistributionP2P*)&param)->data.constant,7,val4),
                  ftoa(((SNetDistributionP2P*)&param)->data.freeparam,7,val5));
            break;

        // SNetworkData
        case PEarly:
        case PLate:
            Convert ((SNetworkData*)&param, True);
            sprintf (data, "module = %04d, type = %04d, netID = %04d, "
                           "value = %s, sender = %04d, receiver = %04d",
                     param.module, param.type,
                     ((SNetworkData*)&param)->networkID,
                     ftoa(((SNetworkData*)&param)->value,14,val1),
                     ((SNetworkData*)&param)->sender,
                     ((SNetworkData*)&param)->receiver);
            break;

        // SSingleParameter
        case InterfacePeriod:
        case TimerGranularity:
            Convert ((SSingleParameter*)&param, True);
            sprintf (data, "module = %04d, type = %04d, value = %s",
                     param.module, param.type,
                     ftoa(((SSingleParameter*)&param)->value, 14, val1));
            break;

        // SEventReport
        case EventClassReport:
        case EventReport:
            Convert ((SEventReport*)&param, True);
            sprintf (data, "module = %04d, type = %04d, value = %s, "
                           "set = %04d, masksize = %04d",
                     param.module, param.type,
                     ftoa(((SEventReport*)&param)->value, 14, val1),
                     ((SEventReport*)&param)->set,
                     ((SEventReport*)&param)->masksize);
            break;

        default:
            cout << "unknown parameter type to controller socket: ";
            cout << gi_Convert((INT32)param.type);
            cout << " (converted: " << param.type << ")" << endl;
            return;
    } // switch

    cout << "parameter to " << data << endl;
}


// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
EBool CSoftwareController::SetParameter (const SParameter& param)
{
    ACNetwork *network;

    ConvertParameter (*(SParameter*)&param);
    // convert to local order

    switch (param.module)
    {
        case Controller:
            return ACController::SetParameter (param);
            break;
        case Network:
            // find out the network
            network = listOfNetworks.Find (((SNetworkParameter*)&param)->networkID);
            if (network == NULL) return False;
            else return network->SetParameter (param);
            break;

        default:
            return False;  // unknown parameter
    } // switch
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned (the
// parameter-fields may nevertheless be changed!)
EBool CSoftwareController::GetParameter (SParameter& param) const
{
    ACNetwork *network;

    ConvertParameter (*(SParameter*)&param);
    // convert to local order
    EBool result;

    switch (param.module)
    {
        case Controller:
            result = ACController::GetParameter (param);
            break;
        case Network:
            // again, find out the network
            network = listOfNetworks.Find (((SNetworkParameter*)&param)->networkID);
            if (network == NULL) result = False;
            else result = network->GetParameter (param);
            break;

        default:
            result = False;  // unknown parameter
    } // switch

    ConvertParameter (*(SParameter*)&param);
    // convert back to remote order

    return result;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// byte order conversion for controller/network commands
static void ConvertCommand (SCommand& command)
{
    static char data[256];

    switch (gi_Convert((INT32)command.type))  // convert byte order
    {
        // SCommand
        case KillSimulation:
        case ResetSimulation:
        case SuspendSimulation:
        case ResumeSimulation:
            Convert ((SCommand*)&command, True);
            sprintf (data, "module = %04d, type = %04d",
                     command.module, command.type);
            break;

        // SApplCommand
        case CreateApplication:
        case DeleteApplication:
        case SuspendApplication:
        case ResumeApplication:
            Convert ((SApplCommand*)&command, True);
            sprintf (data, "module = %04d, type = %04d, nodeID = %04d, "
                           "area = %04d",
                     command.module, command.type,
                     ((SApplCommand*)&command)->nodeID,
                     ((SApplCommand*)&command)->area);
            break;

        // SNetworkCreateCommand
        case CreateNetwork:
            Convert ((SNetworkCreateCommand*)&command, True);
            sprintf (data, "module = %04d, type = %04d, netType = %04d",
                     command.module, command.type,
                     ((SNetworkCreateCommand*)&command)->networkType);
            break;

        // SNetworkCommand
        case CrashModule:
        case RecoverModule:
        case DeleteNetwork:
            Convert ((SNetworkCommand*)&command, True);
            sprintf (data, "module = %04d, type = %04d, netID = %04d",
                     command.module, command.type,
                     ((SNetworkCommand*)&command)->networkID);
            break;

        // SNetworkConnectCommand
        case ConnectNode:
        case DisconnectNode:
            Convert ((SNetworkConnectCommand*)&command, True);
            sprintf (data, "module = %04d, type = %04d, netID = %04d, "
                           "nodeID = %04d",
                     command.module, command.type,
                     ((SNetworkConnectCommand*)&command)->networkID,
                     ((SNetworkConnectCommand*)&command)->nodeID);
            break;

        case ResetModule:
            if (gi_Convert((INT32)command.module)==Controller)
            {
                Convert ((SCommand*)&command, True);
                sprintf (data, "module = %04d, type = %04d",
                         command.module, command.type);
            }
            else
            {
                Convert ((SNetworkCommand*)&command, True);
                sprintf (data, "module = %04d, type = %04d, netID = %04d",
                         command.module, command.type,
                         ((SNetworkCommand*)&command)->networkID);
            }
            break;

        default:
            cout << "unknown command type to controller socket: ";
            cout << gi_Convert((INT32)command.type);
            cout << " (converted: " << command.type << ")" << endl;
            return;
    } // switch

    cout << "command to " << data << endl;
}


// executes the command; returns True if successfull, otherwise False
// existing commands: all Controller-Commands of ECommandType in params.hpp
EBool CSoftwareController::ExecuteCommand (const SCommand& command)
{
    ConvertCommand (*(SCommand*)&command);
    // convert to local byte order

    switch (command.module)  // check the module (Controller or Network)
    {
        case Controller:  // is handled below
            break;
        case Network:  // deliver to the network
            return ExecuteNetworkCommand (*(SNetworkCommand*)&command);
            break;

        default:
            return False;
    } // switch

    switch (command.type)  // check the type of the Controller command
    {
        case CreateNetwork:
            return CreateNewNetwork (*(SNetworkCreateCommand*)&command);
            break;
        case DeleteNetwork:
            return DeleteANetwork (*(SNetworkCommand*)&command);
            break;

        default:  // deliver to the base class
            return ACController::ExecuteCommand (command);
    } // switch
}


//--------------------------------------------------------------------------
//--- starting/stopping applications
//--------------------------------------------------------------------------

// activates the supervisor process
// assert: the supervisor must exist
EBool CSoftwareController::StartApplication (SINT32 nodeID, int area)
{
    CSupervisor *sup = listOfSupervisors.Find (nodeID, area);
    Assert (sup != NULL);  // should never happen

    listOfClocks.Insert (sup->GetClock(), nodeID, area);
    // insert the clock into the list of clocks

    listOfLances.Insert (sup->GetLance(), nodeID, area);
    // insert the lance into the list of lances

#ifndef NO_RESOURCE
    Resource::ref (sup);  // ref the process in C++SIM
#endif
    sup->Activate();  // activate the process

    return True;
}

// cancels the supervisor process, terminates it
// assert: the supervisor must exist
EBool CSoftwareController::StopApplication (SINT32 nodeID, int area)
{
    CSupervisor *sup = listOfSupervisors.Find (nodeID, area);
    Assert (sup != NULL);  // should never happen

    sup->SuspendOperation();  // stop working (no new interrupts)
    sup->Reset();  // reset everything (that also kills the clock processes)

    sup->Cancel();
    sup->terminate();  // terminate the process

    listOfClocks.Remove (nodeID, area);  // remove objects from the lists
    listOfLances.Remove (nodeID, area);
    // note: it is assumed that the supervisor is deleted (or unref'd) in
    // the base class (by calling listOfSupervisors.Delete())

#ifndef NO_RESOURCE
    ClearISRs (nodeID, area);
    // remove the ISRs of that CSA from the vector table (since the
    // CSA is not deleted at once and therefore cannot do it, and we
    // cannot know if EvalSys does not create another CSA with the same
    // nodeID and area)
#endif

    return True;
}


//--------------------------------------------------------------------------
//--- stop program execution
//--------------------------------------------------------------------------

// empty clock list, delete network list, call base class function
void CSoftwareController::DeleteAllSimulationObjects ()
{
    listOfClocks.RemoveAll();  // remove all clock entries
    // note: the clocks will be deleted during deletion of the supervisors

    listOfLances.RemoveAll();  // remove all lance entries
    // note: lances will be deleted during deletion of the supervisors

    ACController::DeleteAllSimulationObjects();  // delete all supervisors

    listOfNetworks.DeleteAll();  // delete all networks
    // has to come after the deletion of the supervisor, because otherwise
    // the application may still try to send messages to a network that is
    // not there anymore (it's no problem, but it induces an error report)

    // note: these statements assume that there are never two controllers
    // in the same context (using the same lists)!
}


// deschedule the process (the function will not return)
void CSoftwareController::StopExecution()
{
    g_finished = True;  // notify the tester
    Cancel();  // stop execution
}


//--------------------------------------------------------------------------
//--- reaction to network commands
//--------------------------------------------------------------------------

// find out the network and call its ExecuteCommand() function
EBool CSoftwareController::ExecuteNetworkCommand (SNetworkCommand& command)
{
    ACNetwork *network;
    network = listOfNetworks.Find (command.networkID);
    if (network == NULL) return False;
    return network->ExecuteCommand (command);
}

// create a new network, return the networkID to EvalSys
// if the network type is not installed, False is returned
EBool CSoftwareController::CreateNewNetwork (SNetworkCreateCommand& command)
{
    NetworkFunction func = networkManager.GetFunction (command.networkType);
    // get the create function

    if (func==NULL) return False;  // check if the network type is installed

    func (this, networkID_);  // create an object of the network

    Assert (ci_init (networkID_, dummy_announce, dummy_error_announce) == 0);
    // initialize the IFC

    ReplyToMessage ((double)networkID_);  // send the networkID to EvalSys
    networkID_++;  // increment the first free ID for networks
    return True;
}

// if the network exists, it is deleted and True is returned; else False
// is returned
EBool CSoftwareController::DeleteANetwork (SNetworkCommand& command)
{
    return listOfNetworks.Delete (command.networkID);
}



