//
// supervis.cpp
//
// implementation of the synchronization supervisor
//
// this version is for BigEndian
// *) includes evalconv.hpp and stdio.h
// *) ExecuteCommand(): byte order of messages is converted
// *) Set/GetParameter(): byte order of messages is converted
// *) ConvertParameter() converts parameter structures
// *) ConvertCommand() converts command structures
//
// created: 06.12.96
//
// last change: 26.01.99
//
// RCS version control information:
//     $Revision: 1.9 $$State: Exp $$Date: 1999/02/05 15:51:42 $
//


#include <stdio.h>  // for sprintf()

#include "clockifc.hpp"
#include "clocksyn.hpp"
#include "error.hpp"
#include "evalconv.hpp"
#include "event.hpp"
#include "general.hpp"
#include "globals.hpp"
#include "lance.hpp"
#include "supervis.hpp"

#define DATA_BUFFER_LENGTH 8000

extern char *ftoa (double f, int width, char *string);  // from ftoa.cpp
// returns 'f' as a string with at least 'width' characters in %e format;
// stores the string in 'string', which must be big enough to hold the
// converted string; if 'string' is NULL, NULL is returned


//**************************************************************************
// class CSupervisor: this class controls the initialization and execution
//      of the clock synchronization; it creates all necessary moduls and
//      is their link to the evaluation
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CSupervisor::CSupervisor (char qname[4], ACSocketAddress *addr, SINT32 nodeID,
    int area, double clockFrequency) :
    CControllingObject (qname, addr, nodeID)
{
    nodeID_ = nodeID;
    area_ = area;
    snapshot_ = False;

    clock_ = new CClockInterface (this, clockFrequency, area_);
    cs_ = new CClockSync (this, area_);
    lance_ = new CLance (this, area_);

    Assert (clock_ != NULL);
    Assert (cs_ != NULL);
    Assert (lance_ != NULL);

    lance_->SetClockSync (cs_);  // connect the objects
    lance_->SetClock (clock_);
    cs_->SetClock (clock_);
    cs_->SetLance (lance_);

    listOfSupervisors.Insert (this, nodeID_, area_);
    // store the supervisor itself in the list of supervisors

    SendModule (Supervisor);  // send the module type to EvalSys
}


CSupervisor::~CSupervisor ()
{
    listOfSupervisors.Remove (nodeID_, area_);
    // remove the supervisor itself from the list of supervisors

    delete cs_;  // ClockSync must be deleted before Clock
    delete lance_;  // Lance must be deleted before Clock
    delete clock_;
}


//--------------------------------------------------------------------------
//--- reset of the application
//--------------------------------------------------------------------------

// set all parameters in all modules to their default values
void CSupervisor::Reset()
{
    snapshot_ = False;

    INT32 oldmode = DisableInterrupts();  // disable application interrupts
/**/// note: this should become obsolete; the Reset() functions should disable
    // interrupts if necessary

    clock_->Reset ();  // reset all modules
    cs_->Reset ();
    lance_->Reset ();

    EnableInterrupts (oldmode);  // enable application interrupts

    CControllingObject::Reset();  // reset the event queue as well
}


//--------------------------------------------------------------------------
//--- get the nodeID/area identifiers
//--------------------------------------------------------------------------

// copy the internal nodeID/area into the parameters
void CSupervisor::GetIdentification (SINT32& nodeID, int& area) const
{
    nodeID = nodeID_;
    area = area_;
}


//--------------------------------------------------------------------------
//--- suspend/resume application
//--------------------------------------------------------------------------

// suspend ClockSync and Lance
void CSupervisor::SuspendOperation ()
{
    cs_->Suspend ();  // suspend the clock sync. (disable duty interrupts)
    lance_->Suspend ();  // suspend the lance
    cs_->Suspend ();  // suspend the clock sync. (now really suspend)
    // note: the order of the three statements is important!
/**/// alternative: put lance_->Suspend() into CS::Suspend(), between
    // the disabling of the interrupt and the rest of the suspend
    // action; disadvantage: CS has to know when and how to suspend the
    // Lance
}

// resume operation
void CSupervisor::ResumeOperation ()
{
    cs_->Resume();  // let CS generate the next interrupt
}


//--------------------------------------------------------------------------
//--- return associated objects
//--------------------------------------------------------------------------

CClockInterface *CSupervisor::GetClock ()
{
    return clock_;
}

CClockSync *CSupervisor::GetClockSync ()
{
    return cs_;
}

CLance *CSupervisor::GetLance ()
{
    return lance_;
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

static char data[DATA_BUFFER_LENGTH];  // for output of the structures

// byte order conversion for application parameters
static void ConvertParameter (SParameter& param)
{
    //static char data[256];  // for output of the structures
    char val[8][30];  // buffers for double values

    switch (gi_Convert((INT32)param.type))  // convert byte order
    {
        // SParameter
        case UserParameter:
            Convert ((SParameter*)&param, True);
            sprintf (data, "module = %04d, type = %04d",
                     param.module, param.type);
            break;

        // SSingleParameter
        case Accuracy:
        case AccuracyTransmissionLoss:
        case AmortizationDuration:
        case AssociatedNetwork:
        case ClockDrift:
        case ClockGranularity:
        case ComputationDelayCompensation:
        case CrashStatus:
        case DelayCompensation:
        case DelayMeasurementCycles:
        case InterfacePeriod:
        case NumberOfFaultyNodes:
        case NumberOfNodes:
        case PCrash:
        case PRecover:
        case PDriftChange:
        case PDuplicateSend:
        case PDuplicateRecv:
        case PFast:
        case PSlow:
        case PReadFault:
        case PWriteFault:
        case PScrambleSend:
        case PScrambleRecv:
        case RateSynchronizationCycles:
        case RhoPlus:
        case RhoMinus:
        case RoundPeriod:
        case SynchronizationAlgorithm:
        case ThetaPlus:
        case ThetaMinus:
        case TimerGranularity:
        case U:
            Convert ((SSingleParameter*)&param, True);
            sprintf (data, "module = %04d, type = %04d, value = %s",
                     param.module, param.type,
                     ftoa (((SSingleParameter*)&param)->value, 14, val[0]));
            break;

        // SApplDistribution
        case DistributionFast:
        case DistributionRegular:
        case DistributionSlow:
        case DistributionSend:
        case DistributionRecv:
            Convert ((SApplDistribution*)&param, True);
            sprintf (data, "module = %04d, type = %04d,\n"
                           "    dist = %04d, low = %s, high = %s, mean = %s, "
                           "constant = %s, freeparam = %s",
                     param.module, param.type,
                     ((SApplDistribution*)&param)->data.dist,
                     ftoa(((SApplDistribution*)&param)->data.low, 7, val[0]),
                     ftoa(((SApplDistribution*)&param)->data.high, 7, val[1]),
                     ftoa(((SApplDistribution*)&param)->data.mean, 7, val[2]),
                     ftoa(((SApplDistribution*)&param)->data.constant,7,val[3]),
                   ftoa(((SApplDistribution*)&param)->data.freeparam,7,val[4]));
            break;

        // SApplWobbleDrift
        case WobbleDrift:
            Convert ((SApplWobbleDrift*)&param, True);
            sprintf (data, "module = %04d, type = %04d,\n"
                           "    dist = %04d, low = %s, high = %s, mean = %s, "
                           "constant = %s, freeparam = %s\n"
                           "    period = %s, rhoMax = %s, rhoMin = %s",
                     param.module, param.type,
                     ((SApplWobbleDrift*)&param)->data.dist,
                     ftoa(((SApplWobbleDrift*)&param)->data.low, 7, val[0]),
                     ftoa(((SApplWobbleDrift*)&param)->data.high, 7, val[1]),
                     ftoa(((SApplWobbleDrift*)&param)->data.mean, 7, val[2]),
                     ftoa(((SApplWobbleDrift*)&param)->data.constant,7,val[3]),
                     ftoa(((SApplWobbleDrift*)&param)->data.freeparam,7,val[4]),
                     ftoa(((SApplWobbleDrift*)&param)->period, 7, val[5]),
                     ftoa(((SApplWobbleDrift*)&param)->rhoMax, 7, val[6]),
                     ftoa(((SApplWobbleDrift*)&param)->rhoMin, 7, val[7]));
            break;

        // STimeParameter
        case ClockValue:
            Convert ((STimeParameter*)&param, True);
            sprintf (data, "module = %04d, type = %04d, ms = %04d, ts = %04d",
                     param.module, param.type,
                     ((STimeParameter*)&param)->macrostamp,
                     ((STimeParameter*)&param)->timestamp);
            break;

        // STransmissionData
        case Delta:
        case EpsilonPlus:
        case EpsilonMinus:
            Convert ((STransmissionData*)&param, True);
            sprintf (data, "module = %04d, type = %04d, netID = %04d, "
                           "nodeID = %04d, value = %s",
                     param.module, param.type,
                     ((STransmissionData*)&param)->networkID,
                     ((STransmissionData*)&param)->nodeID,
                     ftoa(((STransmissionData*)&param)->value, 14, val[0]));
            break;

        // SEventReport
        case EventClassReport:
        case EventReport:
            Convert ((SEventReport*)&param, True);
            sprintf (data, "module = %04d, type = %04d, value = %s, "
                           "set = %04d, masksize = %04d",
                     param.module, param.type,
                     ftoa(((SEventReport*)&param)->value, 14, val[0]),
                     ((SEventReport*)&param)->set,
                     ((SEventReport*)&param)->masksize);
            break;

        default:
            cout << "unknown parameter type to supervisor socket: ";
            cout << gi_Convert((INT32)param.type);
            cout << " (converted: " << param.type << ")" << endl;
            return;
    } // switch

    cout << "parameter to " << data << endl;
}

// sets the given parameter to its new value (or passes it on to one
// of the submodules); if the parameter does not exist or if the value
// is not valid, False is returned
EBool CSupervisor::SetParameter (const SParameter& param)
{
    ConvertParameter (*(SParameter*)&param);
    // convert to local order

    // see if one of the objects can use the parameter
    switch (param.module)
    {
        case Supervisor:
            return CControllingObject::SetParameter (param);
            break;
        case ClockSync:
/**/        // check the type, maybe disable interrupts
            return cs_->SetParameter (param);
            break;
        case Clock:
            return clock_->SetParameter (param);
            break;
        case Lance:
            return lance_->SetParameter (param);
            break;

        default:
            return False;  // unknown parameter
    } // switch
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned (the
// parameter-fields may nevertheless be changed!)
EBool CSupervisor::GetParameter (SParameter& param) const
{
    EBool result;

    ConvertParameter (*(SParameter*)&param);
    // convert to local order

    // again, see if one of the objects can use the parameter
    switch (param.module)
    {
        case Supervisor:
            result = CControllingObject::GetParameter (param);
            break;
        case ClockSync:
/**/        // check the type, maybe disable interrupts
            result = cs_->GetParameter (param);
            break;
        case Clock:
            result = clock_->GetParameter (param);
            break;
        case Lance:
            result = lance_->GetParameter (param);
            break;

        default:
            result = False;  // unknown parameter
    } // switch

    ConvertParameter (*(SParameter*)&param);
    // convert back to remote order

    return result;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// byte order conversion for application commands
static void ConvertCommand (SCommand& command)
{
    //static char data[256];

    switch (gi_Convert((INT32)command.type))  // convert byte order
    {
        // SCommand
        case CrashModule:
        case EnableSnapshot:
        case DisableSnapshot:
        case RecoverModule:
        case ResetModule:
        case UserCommand:
            Convert ((SCommand*)&command, True);
            sprintf (data, "module = %04d, type = %04d",
                     command.module, command.type);
            break;

        // SGpsConnectCommand
        case ConnectGps:
        case DisconnectGps:
            Convert ((SGpsConnectCommand*)&command, True);
            sprintf (data, "module = %04d, type = %04d, gpu = %04d",
                     command.module, command.type,
                     ((SGpsConnectCommand*)&command)->gpu);
            break;

        // SNetworkCommand
        case ConnectNetwork:
        case DisconnectNetwork:
            Convert ((SNetworkCommand*)&command, True);
            sprintf (data, "module = %04d, type = %04d, netID = %04d",
                     command.module, command.type,
                     ((SNetworkCommand*)&command)->networkID);
            break;

        default:
            cout << "unknown command type to supervisor socket: ";
            cout << gi_Convert((INT32)command.type);
            cout << " (converted: " << command.type << ")" << endl;
    } // switch

    cout << "command to " << data << endl;
}


// executes the command (or passes it on to one of the submodules);
// returns True if successfull, otherwise False
// own commands: EnableSnapshot, DisableSnapshot
EBool CSupervisor::ExecuteCommand (const SCommand& command)
{
    ConvertCommand (*(SCommand*)&command);
    // convert to local byte order

    // see if one of the objects can use the command
    switch (command.module)
    {
        case Supervisor:
            if (command.type == EnableSnapshot)
            {
                snapshot_ = True;
                return True;
            }
            else if (command.type == DisableSnapshot)
            {
                snapshot_ = False;
                return True;
            }
            else return CControllingObject::ExecuteCommand (command);
            break;
        case ClockSync:
/**/        // check the type, maybe disable interrupts
            return cs_->ExecuteCommand (command);
            break;
        case Clock:
            return clock_->ExecuteCommand (command);
            break;
        case Lance:
            return lance_->ExecuteCommand (command);
            break;

        default:
            return False;  // unknown module
    } // switch
}


//--------------------------------------------------------------------------
//--- event logging
//--------------------------------------------------------------------------

// stores the event until it can be processed
void CSupervisor::NoteEvent (SEvent *event)
{
    Assert (event != NULL);
    Assert (event->size >= sizeof(SApplEvent));
    // events from the application modules must have that type
/**/// that's not very secure - think of something better!

    SApplEvent *appevent = (SApplEvent *)event;

    clock_->GetTimestamp (appevent->clockTime);
    // get the clock timestamp
    appevent->nodeID = nodeID_;
    appevent->area = area_;
    CControllingObject::NoteEvent (event);
}


//--------------------------------------------------------------------------
//--- periodic activities of derived classes
//--------------------------------------------------------------------------

// is called once every period; initiates a snapshot if the triggering of
// periodic snapshots is enabled
void CSupervisor::DoPeriodicActivities()
{
    if (snapshot_) MakeSystemSnapshot();
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int CSupervisor::ReportError (EReportType type, char *file, int line,
                              char *expression)
{
    return ::ReportError (type, file, line, expression, Supervisor,
                          nodeID_, area_);
    // call the global error report with module and nodeID/area
}

