//
// evalcfg.cpp
//
// contains values of all globals needed for configuring the evaluation
//
// notes:
// *) to change the number of nodes or number of networks, change the
//    constants in evalcfg.hpp
//
// created: 16.09.98
//
// last change: 05.02.99
//
// RCS version control information:
//     $Revision: 1.12 $$State: Exp $$Date: 1999/02/05 15:33:51 $
//


#include "clockcfg.hpp"
#include "constant.hpp"
#include "evalcfg.hpp"


//**************************************************************************
// the variables used for accessing the data
//**************************************************************************

const SEvalsysConfigData evalsysConfigData;
const SSimutcConfigData simutcConfigData;


//**************************************************************************
// constants for the evaluation
//**************************************************************************

const double TwoPowMinus16 = 1.525878e-05;
// resolution of duty timer, approx. 15 us


SEvalsysConfigData::SEvalsysConfigData()
{
    //----------------------------------------------------------------------
    // configure how long evalsys should run
    //----------------------------------------------------------------------

    NumberOfRounds = 100; //40;
    // maximum number of rounds that are printed on screen; a round is defined
    // as the reception of an event of type RoundEvent from node RoundNodeID
    // (see below); if the constant is exceeded, the simulation is stopped
    // (killed)

    NumberOfMessages = 300;
    // maximum number of messages that are received from the simulation; if
    // the constant is exceeded, the simulation is stopped (killed)
    // note: multiply this constant with the SelectTimeout to get the number
    // of seconds evalsys will delay before killing a simulation that does
    // not send any messages


    //----------------------------------------------------------------------
    // configure from when on to collect precision data
    //----------------------------------------------------------------------

    FirstSyncPrecisionRound = 3;  // >= 0
    // evalsys traces the synchronization rounds of the simulation; it
    // starts to collect precision data from the snapshots if the simulation
    // is at least in synchronization round FirstSyncPrecisionRound
    // note: collecting precision data only makes sense after the sync.
    // algorithm has been executed; so the round should be at least the
    // first round in which the sync. algorithm has been executed


    //----------------------------------------------------------------------
    // the area that is occupied on the nodes
    //----------------------------------------------------------------------

    NodeArea = 0;
    // the area of the UTCSUs; currently, you can only have one area per
    // nodeID


    //----------------------------------------------------------------------
    // nodes with special meaning (round count)
    //----------------------------------------------------------------------

    RoundNodeID = 1;
    // the ID of the node used for counting the simulation rounds; you have
    // to make sure that the node you select periodically sends the event
    // you are using to count the rounds (p.e., if RoundEvent is set to
    // ClockUpdate, then you should not use a P-node)

    RoundArea = NodeArea;
    // the area of the node used for counting the simulation rounds

    RoundEvent = ClockUpdate;
    // the event that causes the round counter to be incremented; only
    // use events that are sent periodically by the node you have selected
    // in RoundNodeID

    //----------------------------------------------------------------------
    // snapshot node
    //----------------------------------------------------------------------

    SnapshotNodeID = 2;
    // the ID of the node used for triggering snapshots
    // in the future (currently *not* recommended!!!):
    // set it to a value >= SCommonNodeConfigData::NumberOfNodes if you do
    // not want snapshots

    SnapshotArea = NodeArea;
    // the area of the node used for triggering snapshots

    SnapshotPeriod = 3.0;//100;  // should be > 0
    // the period (in seconds) of the supervisor that triggers periodic
    // snapshots; this value overrides SNodeConfigData::SupervisorPeriod


    //----------------------------------------------------------------------
    // timeouts when waiting for simulation messages
    //----------------------------------------------------------------------

    BlockTime = 20;
    // time (in seconds) the ReceiveMessage() function blocks when waiting
    // for the header

    SelectTimeout = 20;
    // time (in seconds) SelectSocket() waits for one of the supervisor
    // sockets to become readable


    //----------------------------------------------------------------------
    // general evaluation specific things
    //----------------------------------------------------------------------

    MaskSize = EventCount/8 + 1;
    // size of the event mask used in SEventReport; the value should not
    // exceed the size of an INT32
}



//**************************************************************************
// constants for the simulation
//**************************************************************************

//--------------------------------------------------------------------------
// Controller configuration
//--------------------------------------------------------------------------

SControllerConfigData::SControllerConfigData()
{
    //----------------------------------------------------------------------
    // event reports (from the network)
    //----------------------------------------------------------------------

    // enabled event classes; uncomment the classes you need, comment
    // out those that you do not need
    EventClasses =
        //(1 << SystemReport) +
        //(1 << InterruptEvent) +
        //(1 << FaultInjection) +
        0;

    // enabled events; again, uncomment the events you need and comment
    // out the rest
    Events =
        (1 << CrashStatusChange) +
        (1 << DeliverMessage) +
        (1 << EarlyMessage) +
        (1 << LateMessage) +
        (1 << MessageLost) +
        (1 << SendMessage) +
        0;


    //----------------------------------------------------------------------
    // Controller period
    //----------------------------------------------------------------------

    ControllerPeriod = DefaultIfcPeriod;  // should be > 0
    // the period of the controller in seconds; this value affects how
    // fast the controller reacts to user commands and how often it sends
    // network events to the evaluation
    // in pure simulation, the value is not very important, so it is set
    // on its default value
}


//--------------------------------------------------------------------------
// node configuration
//--------------------------------------------------------------------------

SCommonNodeConfigData::SCommonNodeConfigData()
{
    //----------------------------------------------------------------------
    // number of nodes
    //----------------------------------------------------------------------

    NumberOfNodes = numberOfNodes;  // do not change this assignment!
    // number of nodes in the system
    // note: if you have to change the number of nodes, change the constant
    // 'numberOfNodes' in evalcfg.hpp

    NumberOfFaultyNodes = 0;
    // maximum number of faulty nodes


    //----------------------------------------------------------------------
    // round specific things (period, delay cycles)
    //----------------------------------------------------------------------

    RoundPeriod = 12;//100;  // should be > 0, multiple of 2^16
    // period of a synchronization round in seconds
    // the value should have duty timer granularity (2^-16 seconds)

    DelayCycles = 2;//4;  // set it to 0 to turn off delay measurement
    // every 'delayCycles'-th message does delay measurement

    RateCycles = 4;//0;  // set it to 0 to turn off rate sync.
    // every 'rateCycles'-th message sends rate synchronization data


    //----------------------------------------------------------------------
    // general settings (algorithm)
    //----------------------------------------------------------------------

    // the synchronization algorithm that is used; uncomment the algorithm
    // you want to use and leave the rest commented out
    SynchronizationAlgorithm =
        //FaultTolerantAverage;
        //FreeRun;
        //Midpoint;
        //Fetzer;
        //OP;
        //Marzullo;
        Rate;
        //FaultTolerantIntersection;
}


SSpecificNodeConfigData::SSpecificNodeConfigData()
{
    //----------------------------------------------------------------------
    // configurations concerning the clock
    //----------------------------------------------------------------------

    ClockDrift = 0;  // should be >= -1
    // clock drift
    // SimUTC does only accept the parameter in pure simulation
    // note: the ctor of SSimutcConfigData overrides the drift

    ConnectedGpus = 0;  // default: S-Node
    // the GPUs of the UTCSU which have GPS receivers connected to them
    // this is a binary representation, so if a receiver is connected to
    // GPU1, bit 1<<GPU1 is set in connectedGpus
    // this value is not directly sent to the Clock module, but for each
    // GPU set, command ConnectGps is called
    // note: the ctor of SSimutcConfigData overrides this value

    Macrostamp = 0;
    Timestamp = 0;
    // initial clock value (each is a 32-bit value)

    Accuracy = 0; //0x20000000;
    // initial clock accuracy
    // the value contains the UTCSU bits S:[-8,-38] of both accuracies
    // take care when comparing this input with the accuracies you get
    // with the clock timestamps: those are unsigned and contain
    // bits [-8,-23] in the upper and lower two bytes of the 'acc' field

    AmortizationDuration = DefaultAmortizationDuration; // should be >= 0
    // the duration of a clock amortization phase (in seconds)
    // the value should have duty timer granularity (2^-16 seconds)

    RhoPlus = 1e-4;//WcRhoPlus;  // should be in [0, WcRhoPlus]
    // upper bound on the regular clock drift (the worst case bound for
    // faulty drift is taken from WcRhoPlus)

    RhoMinus = 1e-4;//WcRhoMinus;  // should be in [0, WcRhoMinus]
    // lower bound on the regular clock drift (the worst case bound for
    // faulty drift is taken from WcRhoMinus)


    //----------------------------------------------------------------------
    // configurations concerning the clock (fault injection)
    //----------------------------------------------------------------------

    ProbClockCrash = 0;  // should be in [0, 1]
    // probability of "crashing" the clock (i.e., the clock stops counting)

    ProbClockRecover = 0;  // should be in [0, 1]
    // probability of recovering a crashed clock

    ProbDriftChange = 0;  // should be in [0, 1]
    // probability that the clock drift changes

    ProbFastDrift = 0;  // should be in [0, 1]
    // probability that the clock drift changes to a fast value

    ProbSlowDrift = 0;  // should be in [0, 1]
    // probability that the clock drift changes to a slow value

    ProbReadFault = 0;  // should be in [0, 1]
    // probability that a read fault occurs

    ProbWriteFault = 0;  // should be in [0, 1]
    // probability that a write fault occurs


    //----------------------------------------------------------------------
    // clock drift wobbling
    //----------------------------------------------------------------------

    WobblePeriod = 0;  // should be >= 0, multiple of 2^16
    // the period of drift wobbling (in seconds)
    // set it to zero to disable drift wobbling
    // the value should have duty timer granularity (2^-16 seconds)

    WobbleRhoPlus = RhoPlus;  // should be in [0, WcRhoPlus] (constant.hpp)
    // upper bound on the drift during wobbling

    WobbleRhoMinus = RhoMinus;  // in [0, WcRhoMinus] (constant.hpp)
    // lower bound on the drift during wobbling

    WobbleAddRangeMin = 0;  // should be >= 0
    WobbleAddRangeMax = 0;  // should be >= 0
    // bounds of the interval from which the additional drift should be
    // chosen (it is taken from this interval with Uniform distribution)


    //----------------------------------------------------------------------
    // configurations concerning the clock synchronization module
    //----------------------------------------------------------------------

    DelayCompensation = TwoPowMinus16 * 120;  // >= 0, multiple of 2^16
    // delay compensation (of network transmission times) in seconds
    // the value should have duty timer granularity (2^-16 seconds)

    CompDelayCompensation = 0;  // should be >= 0, multiple of 2^16
    // computation delay compensation (algorithm execution time) in seconds
    // can be set to zero in pure simulation, because the algorithm
    // execution does not take any (simulated) time
    // the value should have duty timer granularity (2^-16 seconds)


    //----------------------------------------------------------------------
    // configurations concerning the network access (fault injection)
    //----------------------------------------------------------------------

    ProbLanceCrash = 0;  // should be in [0, 1]
    // probability of a node crash

    ProbLanceRecover = 0;  // should be in [0, 1]
    // probability of node recovery

    DmaxSendLance = 0;  // should be >= 0
    // maximum delay of outgoing messages in seconds

    DmaxRecvLance = 0;  // should be >= 0
    // maximum delay of incoming messages in seconds

    ProbLanceSendScramble = 0;  // should be in [0, 1]
    // probability of scrambling send message

    ProbLanceRecvScramble = 0;  // should be in [0, 1]
    // probability of scrambling receive message

    ProbLanceSendDuplicate = 0;  // should be in [0, 1]
    // probability of duplicating send message

    ProbLanceRecvDuplicate = 0;  // should be in [0, 1]
    // probability of duplicating receive message


    //----------------------------------------------------------------------
    // event reports (from the nodes)
    //----------------------------------------------------------------------

    // enabled event classes; uncomment the classes you need, comment
    // out those that you do not need
    EventClasses =
        (1 << SystemReport) +
        (1 << InterruptEvent) +
        //(1 << FaultInjection) +
        0;

    // enabled events; again, uncomment the events you need and comment
    // out the rest
    Events =
        //(1 << ChangeAssociation) +
        (1 << ClockPreload) +
        //(1 << ClockReadFault) +
        (1 << ClockUpdate) +
        //(1 << ClockWriteFault) +
        //(1 << ConvergenceFailed) +
        //(1 << CrashStatusChange) +
        //(1 << DelayEstimation) +
        //(1 << DiscardRecvMessage) +
        //(1 << DiscardSendMessage) +
        //(1 << GpsEvent) +
        //(1 << IgnoreSetting) +
        //(1 << IntervalEvent) +
        (1 << RateEvent) +
        //(1 << ReceivedMessage) +
        //(1 << ScrambleRecvMessage) +
        //(1 << ScrambleSendMessage) +
        //(1 << DuplicateRecvMessage) +
        //(1 << DuplicateSendMessage) +
        //(1 << SendMessage) +
        (1 << Snapshot) +
        //(1 << TimerEvent) +
        //(1 << UserEvent) +
        0;
    // note: GPS events are only sent if the clock needs resynchronization to
    // GPS; so the event may be turned on without harm (it will only occur
    // once or twice at the beginning of the reports, until the clock is
    // synchronized to GPS


    //----------------------------------------------------------------------
    // Supervisor period
    //----------------------------------------------------------------------

    SupervisorPeriod = DefaultIfcPeriod;  // should be > 0
    // the period of the supervisor in seconds; like the controller period,
    // it affects how the supervisor reacts to user commands and how often
    // it sends CSA events to the evaluation
    // if the node triggers snapshot, SEvalsysConfigData::SnapshotPeriod
    // overrides this value
    // in pure simulation, the value is not very important

}


//--------------------------------------------------------------------------
// network configuration
//--------------------------------------------------------------------------

SNetworkConfigData::SNetworkConfigData()
{
    //----------------------------------------------------------------------
    // number of networks
    //----------------------------------------------------------------------

    NumberOfNetworks = numberOfNetworks;  // do not change this assignment!
    // the number of simulated networks
    // note: if you have to change the number of networks, change the
    // constant 'numberOfNetworks' in evalcfg.hpp
    // (always 1 in pure simulation, 0 in hardware-based simulation)
    // note: if there is more than one network, the first one is used as
    // associated network


    DefaultCINetwork = 2;
    // the associated network in hardware (is always used if numberOfNetworks
    // is zero); this should be the CI network interface number


    //----------------------------------------------------------------------
    // message transmission characteristics
    //----------------------------------------------------------------------

    Delta = 60e-6;//.6;    // should be >= EpsilonMinus (see below)
    // deterministic part of the transmission delay (in seconds)
    // first measures: d =^= 70 us

    EpsilonPlus = 20e-6;//.1;  // should be >= 0
    // indeterministic upper part of the transmission delay (in seconds)
    // eps =^= 2-3 us

    EpsilonMinus = 10e-6;//.0;  // should be <= Delta
    // indeterministic lower part of the transmission delay (in seconds)

    Lambda = 1e-3;//.2;  // should be >= 0
    // broadcast delay (in seconds)

    Omega = 0;  // should be >= 0
    // broadcast operation delay (in seconds)
    // ignored by the currently used broadcast network


    //----------------------------------------------------------------------
    // message transmission characteristics (fault injection)
    //----------------------------------------------------------------------

    Dmin = 0;  // should be in [0, Delta-EpsilonMinus]
    // minimum transmission delay of early messages (in seconds)

    Dmax = 5;  // should be >= Delta+EpsilonPlus
    // maximum transmission delay of late messages (in seconds)

    ProbNetCrash = 0;  // should be in [0, 1]
    // probability of a network crash

    ProbNetRecover = 0;  // should be in [0, 1]
    // probability of network recovery

    ProbNetEarly = 0;  // should be in [0, 1]
    // probability of early messages

    ProbNetLate = 0;  // should be in [0, 1]
    // probability of late messages
}


//--------------------------------------------------------------------------
// complete simulation configuration
//--------------------------------------------------------------------------

// the ctor modifies the default values of the nodes
SSimutcConfigData::SSimutcConfigData()
{
    int nodeNumber;  // for iterating over the nodes

    nodesDiffer = False;
    // this is the default if all nodes are configured alike
    // in this case, PrintConfiguration() in evalfunc.cpp only prints
    // the configuration of one node, not that of all nodes

    // controllerData is taken over unchanged

    // commonNodeData is taken over unchanged

    // nodeData is adapted; add your own node-specific modifications here

    nodeData = new SSpecificNodeConfigData[numberOfNodes];
    // create the default configurations

    nodesDiffer = True;  // we make changes to some nodes

    for (nodeNumber = 0;
         nodeNumber < simutcConfigData.commonNodeData.NumberOfNodes;
         nodeNumber++)
    {

        // last node is a P-Node, but only if there are at least two nodes
        if ((nodeNumber == simutcConfigData.commonNodeData.NumberOfNodes-1) &&
            (nodeNumber > 0))
        {
            //nodeData[nodeNumber].ConnectedGpus = 1<<GPU3;
        } // if

        double stop = 2.0;

        switch (nodeNumber+1)  // set clock drifts
        {
            case 0: break;  // do not use!!!
                // we use the real node number (1-numberOfNodes)

            case 1:
                nodeData[nodeNumber].ClockDrift = 0; //1e-6;

                // drift wobbling
                nodeData[nodeNumber].WobblePeriod = stop*TwoPowMinus16*0.0;

                break;

            case 2:
                // average drift
                nodeData[nodeNumber].ClockDrift = 0.5e-6; //0;

                // drift wobbling
                nodeData[nodeNumber].WobblePeriod = stop * TwoPowMinus16 * 5e5;
                nodeData[nodeNumber].WobbleRhoPlus = 0.3e-6 *3;
                nodeData[nodeNumber].WobbleRhoMinus = 0.1e-6 *3;
                nodeData[nodeNumber].WobbleAddRangeMin = 0.0e-6 *3;
                nodeData[nodeNumber].WobbleAddRangeMax = 0.05e-6 *3;

                break;

            case 3:
                // average drift
                nodeData[nodeNumber].ClockDrift = -0.5e-6;//-1e-6;

                // drift wobbling
                nodeData[nodeNumber].WobblePeriod = stop * TwoPowMinus16 * 3e5;
                nodeData[nodeNumber].WobbleRhoPlus = 0.1e-6 *3;
                nodeData[nodeNumber].WobbleRhoMinus = 0.1e-6 *3;
                nodeData[nodeNumber].WobbleAddRangeMin = 0.01e-6 *3;
                nodeData[nodeNumber].WobbleAddRangeMax = 0.05e-6 *3;

                break;

            case 4:  // drift wobbling
                // average drift
                nodeData[nodeNumber].ClockDrift = -1e-6;

                // drift wobbling
                nodeData[nodeNumber].WobblePeriod = stop * TwoPowMinus16 * 5e5;
                nodeData[nodeNumber].WobbleRhoPlus = 0.2e-6 *3;
                nodeData[nodeNumber].WobbleRhoMinus = 0.3e-6 *3;
                nodeData[nodeNumber].WobbleAddRangeMin = 0.01e-6 *3;
                nodeData[nodeNumber].WobbleAddRangeMax = 0.01e-6 *3;

                break;

            case 5:
                // average drift
                nodeData[nodeNumber].ClockDrift = 2e-6;

                // drift wobbling
                nodeData[nodeNumber].WobblePeriod = stop * TwoPowMinus16 * 1e6;
                nodeData[nodeNumber].WobbleRhoPlus = 0.5e-6 *3;
                nodeData[nodeNumber].WobbleRhoMinus = 0.5e-6 *3;
                nodeData[nodeNumber].WobbleAddRangeMin = 0.03e-6 *3;
                nodeData[nodeNumber].WobbleAddRangeMax = 0.03e-6 *3;

                break;
        } // switch
    } // for

    // networkData is taken over unchanged
}

// the dtor deletes the array
SSimutcConfigData::~SSimutcConfigData()
{
    delete [] nodeData;
}


