//
// evalcfg.hpp
//
// contains all globals needed for configuring the evaluation and simulation
//
// notes:
// *) to change the number of nodes or number of networks, change the
//    constants below
// *) to change all other values, edit evalcfg.cpp
// *) if you want to change the values of only one node, adapt the ctor
//    of SSimutcConfigData so that it sets the values of the node you
//    want to change
//
// created: 01.02.97
//
// last change: 05.02.99
//
// RCS version control information:
//     $Revision: 1.8 $$State: Exp $$Date: 1999/02/05 15:31:47 $
//


#ifndef _EVALCFG_HPP_
#define _EVALCFG_HPP_

#include "psos.h"

#include "event.hpp"
#include "general.hpp"


//#define NO_PRINTEVENT
// uncomment this line if you do not wish to print out events, but
// do not want to redirect cout (because you are interested in the
// precision summary)

#define PLOT_DATA
// uncomment it to make plots of precision, consonance, ...


//**************************************************************************
// number of nodes and networks
//**************************************************************************

const INT32 numberOfNodes = 5;
// the number of nodes in the system; in our case, the number of nodes
// equals the number of CSAs, because we only use one area on each node
// (to simplify the configuration in evalsys.cpp)

const INT32 numberOfNetworks = 1;
// the number of networks in the system; should be 1 for pure simulation,
// 0 in hardware-based simulation
// note: if there is more than one network, the first one is used as
// associated network



//**************************************************************************
// constants for the evaluation
//**************************************************************************

// this structure contains all constants the evaluation needs to set up
// the simulation and work with it; it does not contain any simulation
// specific constants
struct SEvalsysConfigData
{
    //----------------------------------------------------------------------
    // configure how long evalsys should run
    //----------------------------------------------------------------------

    int NumberOfRounds;
    // maximum number of rounds that are printed on screen; a round is defined
    // as the reception of an event of type RoundEvent from node RoundNodeID
    // (see below); if the constant is exceeded, the simulation is stopped
    // (killed)

    int NumberOfMessages;
    // maximum number of messages that are received from the simulation; if
    // the constant is exceeded, the simulation is stopped (killed)
    // note: multiply this constant with the SelectTimeout to get the number
    // of seconds evalsys will delay before killing a simulation that does
    // not send any messages


    //----------------------------------------------------------------------
    // configure from when on to collect precision data
    //----------------------------------------------------------------------

    INT32 FirstSyncPrecisionRound;
    // evalsys traces the synchronization rounds of the simulation; it
    // starts to collect precision data from the snapshots if the simulation
    // is at least in synchronization round FirstSyncPrecisionRound


    //----------------------------------------------------------------------
    // set the area that is occupied on the nodes
    //----------------------------------------------------------------------

    int NodeArea;
    // the area of the nodes; currently, you can only have one area per
    // node


    //----------------------------------------------------------------------
    // nodes with special meaning
    //----------------------------------------------------------------------

    SINT32 RoundNodeID;
    // the ID of the node used for counting the simulation rounds; you have
    // to make sure that the node you select periodically sends the event
    // you are using to count the rounds (p.e., if RoundEvent is set to
    // ClockUpdate, then you should not use a P-node)

    int RoundArea;
    // the area of the node used for counting the simulation rounds

    EEvent RoundEvent;
    // the event that causes the round counter to be incremented; only
    // use events that are sent periodically by the node you have selected
    // in RoundNodeID


    //----------------------------------------------------------------------
    // snapshot node
    //----------------------------------------------------------------------

    SINT32 SnapshotNodeID;
    // the ID of the node used for triggering snapshots

    int SnapshotArea;
    // the area of the node used for triggering snapshots

    double SnapshotPeriod;
    // period (in seconds) of the supervisor that triggers periodic
    // snapshots; this value overrides SNodeConfigData::SupervisorPeriod


    //----------------------------------------------------------------------
    // timeouts when waiting for messages
    //----------------------------------------------------------------------

    int BlockTime;
    // time (in seconds) the ReceiveMessage() function blocks when waiting
    // for the header

    int SelectTimeout;
    // time (in seconds) SelectSocket() waits for one of the supervisor
    // sockets to become readable


    //----------------------------------------------------------------------
    // general evaluation specific things
    //----------------------------------------------------------------------

    int MaskSize;
    // size of the event mask used in SEventReport


    //----------------------------------------------------------------------
    // the constructor sets all values
    //----------------------------------------------------------------------

    SEvalsysConfigData();
};


//**************************************************************************
// constants for the simulation
//**************************************************************************

//--------------------------------------------------------------------------
// Controller configuration
//--------------------------------------------------------------------------

// this structure contains the configuration data for the simulation
// controller
struct SControllerConfigData
{
    //----------------------------------------------------------------------
    // event reports
    //----------------------------------------------------------------------

    INT32 EventClasses;
    // enabled event classes

    INT32 Events;
    // enabled events


    //----------------------------------------------------------------------
    // Controller period
    //----------------------------------------------------------------------

    double ControllerPeriod;
    // the period of the controller in seconds; this value affects how
    // fast the controller reacts to user commands and how often it sends
    // network events to the evaluation
    // in pure simulation, the value is not very important


    //----------------------------------------------------------------------
    // the constructor sets all values
    //----------------------------------------------------------------------

    SControllerConfigData();
};


//--------------------------------------------------------------------------
// node configuration
//--------------------------------------------------------------------------

// this structure contains all node configuration data that is the same
// for all nodes
struct SCommonNodeConfigData
{
    //----------------------------------------------------------------------
    // number of nodes
    //----------------------------------------------------------------------

    int NumberOfNodes;  // -> ClockSync
    // number of nodes in the system

    int NumberOfFaultyNodes;  // -> ClockSync
    // maximum number of faulty nodes


    //----------------------------------------------------------------------
    // round specific things (period, delay cycles)
    //----------------------------------------------------------------------

    double RoundPeriod;  // -> ClockSync
    // period of a synchronization round in seconds

    int DelayCycles;  // -> ClockSync
    // every 'delayCycles'-th message does delay measurement

    int RateCycles;  // -> ClockSync
    // every 'rateCycles'-th message sends rate synchronization data


    //----------------------------------------------------------------------
    // general settings (algorithm)
    //----------------------------------------------------------------------

    EAlgorithmType SynchronizationAlgorithm;  // -> ClockSync
    // synchronization algorithm used


    //----------------------------------------------------------------------
    // the constructor sets all values
    //----------------------------------------------------------------------

    SCommonNodeConfigData();
};


// this structure contains node specific data; the default values are that
// of an S-node, any P-Node specific changes are done in the ctor of
// SSimutcConfigData
struct SSpecificNodeConfigData
{
    //----------------------------------------------------------------------
    // configurations concerning the clock
    //----------------------------------------------------------------------

    double ClockDrift;  // -> Clock
    // clock drift
    // SimUTC does only accept the parameter in pure simulation

    INT8 ConnectedGpus;
    // the GPUs of the UTCSU which have GPS receivers connected to them
    // this is a binary representation, so if a receiver is connected to
    // GPU1, bit 1<<GPU1 is set in connectedGpus
    // this value is not directly sent to the Clock module, but for each
    // GPU set, command ConnectGps is called

    INT32 Macrostamp, Timestamp;  // -> Clock
    // initial clock value

    INT32 Accuracy;  // -> Clock
    // initial clock accuracy (one value for both accuracies)
    // the value contains the UTCSU bits S:[-8,-38] of both accuracies

    double AmortizationDuration;  // -> Clock
    // the duration of a clock amortization phase (in seconds)

    double RhoPlus;  // -> Clock
    // upper bound on the regular clock drift (the worst case bound for
    // faulty drift is taken from WcRhoPlus)

    double RhoMinus;  // -> Clock
    // lower bound on the regular clock drift (the worst case bound for
    // faulty drift is taken from WcRhoMinus)


    //----------------------------------------------------------------------
    // configurations concerning the clock (fault injection)
    //----------------------------------------------------------------------

    double ProbClockCrash;  // -> Clock
    // probability of "crashing" the clock (i.e., the clock stops counting)

    double ProbClockRecover;  // -> Clock
    // probability of recovering a crashed clock

    double ProbDriftChange;  // -> Clock
    // probability that the clock drift changes

    double ProbFastDrift;  // -> Clock
    // probability that the clock drift changes to a fast value

    double ProbSlowDrift;  // -> Clock
    // probability that the clock drift changes to a slow value

    double ProbReadFault;  // -> Clock
    // probability that a read fault occurs

    double ProbWriteFault;  // -> Clock
    // probability that a write fault occurs


    //----------------------------------------------------------------------
    // clock drift wobbling
    //----------------------------------------------------------------------

    double WobblePeriod;  // -> Clock
    // the period of drift wobbling (in seconds)

    double WobbleRhoPlus;  // -> Clock
    // upper bound on the drift during wobbling

    double WobbleRhoMinus;  // -> Clock
    // lower bound on the drift during wobbling

    double WobbleAddRangeMin;  // -> Clock
    double WobbleAddRangeMax;  // -> Clock
    // bounds of the interval from which the additional drift should be
    // chosen (it is taken from this interval with Uniform distribution)


    //----------------------------------------------------------------------
    // configurations concerning the clock synchronization module
    //----------------------------------------------------------------------

    double DelayCompensation;  // -> ClockSync
    // delay compensation (of network transmission times) in seconds

    double CompDelayCompensation;  // -> ClockSync
    // computation delay compensation (algorithm execution time) in seconds


    //----------------------------------------------------------------------
    // configurations concerning the network access (fault injection)
    //----------------------------------------------------------------------

    double ProbLanceCrash;  // -> Lance
    // probability of a node crash

    double ProbLanceRecover;  // -> Lance
    // probability of node recovery

    double DmaxSendLance;  // -> Lance
    // maximum delay of outgoing messages in seconds

    double DmaxRecvLance;  // -> Lance
    // maximum delay of incoming messages in seconds

    double ProbLanceSendScramble;  // -> Lance
    // probability of scrambling send message

    double ProbLanceRecvScramble;  // -> Lance
    // probability of scrambling receive message

    double ProbLanceSendDuplicate;  // -> Lance
    // probability of duplicating send message

    double ProbLanceRecvDuplicate;  // -> Lance
    // probability of duplicating receive message


    //----------------------------------------------------------------------
    // event reports
    //----------------------------------------------------------------------

    INT32 EventClasses;  // -> Supervisor
    // enabled event classes

    INT32 Events;  // -> Supervisor
    // enabled events


    //----------------------------------------------------------------------
    // Supervisor period
    //----------------------------------------------------------------------

    double SupervisorPeriod;  // -> Supervisor
    // the period of the supervisor in seconds; like the controller period,
    // it affects how the supervisor reacts to user commands and how often
    // it sends CSA events to the evaluation
    // if the node triggers snapshot, SEvalsysConfigData::SnapshotPeriod
    // overrides this value
    // in pure simulation, the value is not very important


    //----------------------------------------------------------------------
    // the constructor sets all values
    //----------------------------------------------------------------------

    SSpecificNodeConfigData();
};


//--------------------------------------------------------------------------
// network configuration
//--------------------------------------------------------------------------

struct SNetworkConfigData
{
    //----------------------------------------------------------------------
    // number of networks
    //----------------------------------------------------------------------

    int NumberOfNetworks;
    // the number of simulated networks
    // (always 1 in pure simulation, 0 in hardware-based simulation)
    // note: if there is more than one network, the first one is used as
    // associated network

    SINT32 DefaultCINetwork;
    // the associated network in hardware (is always used if numberOfNetworks
    // is zero); this should be the CI network interface number


    //----------------------------------------------------------------------
    // message transmission characteristics
    // currently the same for all nodes!
    //----------------------------------------------------------------------

    double Delta;
    // deterministic part of the transmission delay (in seconds)

    double EpsilonPlus;
    // indeterministic upper part of the transmission delay (in seconds)

    double EpsilonMinus;
    // indeterministic lower part of the transmission delay (in seconds)

    double Lambda;
    // broadcast delay (in seconds)

    double Omega;
    // broadcast operation delay (in seconds)


    //----------------------------------------------------------------------
    // message transmission characteristics (fault injection)
    //----------------------------------------------------------------------

    double Dmin;
    // minimum transmission delay of early messages (in seconds)

    double Dmax;
    // maximum transmission delay of late messages (in seconds)

    double ProbNetCrash;
    // probability of a network crash

    double ProbNetRecover;
    // probability of network recovery

    double ProbNetEarly;
    // probability of early messages

    double ProbNetLate;
    // probability of late messages


    //----------------------------------------------------------------------
    // the constructor sets all values
    //----------------------------------------------------------------------

    SNetworkConfigData();
};


//--------------------------------------------------------------------------
// complete simulation configuration
//--------------------------------------------------------------------------

// this structure contains all data of the simulation
struct SSimutcConfigData
{
    SControllerConfigData controllerData;
    // controller data

    SCommonNodeConfigData commonNodeData;
    // data common to all nodes

    SSpecificNodeConfigData *nodeData;
    // node specific data

    SNetworkConfigData networkData;
    // network data


    //----------------------------------------------------------------------
    // the ctor creates the array objects and modifies their default values;
    // the dtor deletes the array objects
    //----------------------------------------------------------------------

    SSimutcConfigData();  // overrides some of the node specific defaults
    ~SSimutcConfigData();

    EBool nodesDiffer;
    // set it to True if some nodes are configured differently from others
};



//**************************************************************************
// the variables used for accessing the data
//**************************************************************************

extern const SEvalsysConfigData evalsysConfigData;
extern const SSimutcConfigData simutcConfigData;

#endif


