//
// evalconv.cpp
//
// all structures and classes that are needed to manage system parameters
// and events
//
// created: 09.03.98
//
// last change: 08.10.98
//
// RCS version control information:
//     $Revision: 1.11 $$State: Exp $$Date: 1998/10/08 09:46:04 $
//


#include "evalconv.hpp"


//**************************************************************************
// unconditional big/little endian conversion of simple types
//**************************************************************************

// convert the INT32 (just read 'src' backwards)
static void g_Convert (INT32& src)
{
    INT32 tmp;
    tmp  = (src << 24) & 0xFF000000;
    tmp += (src <<  8) & 0x00FF0000;
    tmp += (src >>  8) & 0x0000FF00;
    tmp += (src >> 24) & 0x000000FF;
    src = tmp;
}

// macro for declaring a conversion function g_Convert() for a certain
// enumeration; this assumes that all enums have the same data size (4 bytes)
// (usually, you can enforce this with a compiler option)
#define DeclareConversion(type)		\
static void g_Convert (type &src)	\
{					\
    INT32 tmp = src;			\
    g_Convert (tmp);			\
    src = (type)tmp;			\
}


// conversion functions for diverse enumerations
DeclareConversion (int);
DeclareConversion (SINT32);

DeclareConversion (EBool);
DeclareConversion (EDistribution);
DeclareConversion (EEvent);
DeclareConversion (EGpsUnit);
DeclareConversion (EModule);
DeclareConversion (ENetworkType);
DeclareConversion (EReplyType);
DeclareConversion (ESystemCommand);
DeclareConversion (ESystemParameter);


// conversion functions for other data types
static void g_Convert (STimestamp& src)
{
    g_Convert (src.ms);
    g_Convert (src.ts);
    g_Convert (src.acc);
}

static void g_Convert (CNtpTime& src)
{
    g_Convert (src.ms);
    g_Convert (src.ts);
    g_Convert (src.us);
}

// convert the double, return the new value (just read 'src' backwards)
static void g_Convert (double& src)
{
    double res;
    INT8 *srcptr = ((INT8*)&src)+7;  // last byte of source
    INT8 *resptr = (INT8*) &res;  // first byte of destination

    for (int i=0; i<8; i++)
    {
        *resptr = *srcptr;
        resptr++;  // go to next byte in result
        srcptr--;  // go to previous byte in source
    } // for

    src = res;
}


//**************************************************************************
// macro for defining a conversion function for a certain type
//**************************************************************************

// note: as you can see, the macro only defines half of the function;
// therefore, using the macro results in non-C style code; however, it
// makes the user code shorter and emphasizes the important things
// (who is the base struct, which are the own members that need conversion)
// on the other hand, debugging this code is almost certainly impossible
// please keep this in mind when stepping into a Convert() function!
#define DefineConvert(type, basetype)                                   \
void Convert (type *s, EBool convert)                                    \
{                                                                       \
    if (!convert) return;  \
    Convert ((basetype *)s, convert);

// convert a single member
#define DoConvert(name)							\
    g_Convert (s->name);


//**************************************************************************
// these are the Convert() declarations for all structures that are needed
// by the current dummy evaluation
//**************************************************************************

// SModuleBase does not need a call to the Convert() function of the base,
// so the macro cannot be used
void Convert (SModuleBase *s, EBool convert)
{
    if (!convert) return;
    g_Convert (s->module);
}


//--------------------------------------------------------------------------
// event structures
//--------------------------------------------------------------------------

DefineConvert (SEvent, SModuleBase)
    DoConvert (type);
    DoConvert (time);
}

DefineConvert (SNetEvent, SEvent)
    DoConvert (networkID);
}

DefineConvert (SNetCrashEvent, SNetEvent)
    DoConvert (crashed);
}

DefineConvert (SNetMessageEvent, SNetEvent)
    DoConvert (sender);
    DoConvert (area);
    DoConvert (receiver);
    DoConvert (msgID);
}

DefineConvert (SApplEvent, SEvent)
    DoConvert (clockTime);
    DoConvert (nodeID);
    DoConvert (area);
}

DefineConvert (SApplNetEvent, SApplEvent)
    DoConvert (networkID);
}

DefineConvert (SApplGpsEvent, SApplEvent)
    DoConvert (diff);
}

DefineConvert (SApplClockEvent, SApplEvent)
    DoConvert (newTime);
}

DefineConvert (SApplPreloadEvent, SApplClockEvent)
    DoConvert (loadAt);
}

DefineConvert (SApplTimeEvent, SApplEvent)
    DoConvert (input);
    DoConvert (interval.low);
    DoConvert (interval.high);
    DoConvert (interval.refpoint);
}

DefineConvert (SApplRateEvent, SApplEvent)
    DoConvert (input);
    DoConvert (rate.thetaPlus);
    DoConvert (rate.thetaMinus);
    DoConvert (rate.u);
}

DefineConvert (SApplCrashEvent, SApplEvent)
    DoConvert (crashed);
}

DefineConvert (SApplSendMsgEvent, SApplEvent)
    DoConvert (msgID);
    DoConvert (receiver);
}

DefineConvert (SApplRecvMsgEvent, SApplSendMsgEvent)
    DoConvert (sender);
    DoConvert (senderArea);
}

DefineConvert (SApplDelayEvent, SApplRecvMsgEvent)
    DoConvert (delaytime);
}

DefineConvert (SApplDelayEstimationEvent, SApplEvent)
    DoConvert (networkID);
    DoConvert (sender);
    DoConvert (delay.delta);
    DoConvert (delay.epsilonMinus);
    DoConvert (delay.epsilonPlus);
}

//--------------------------------------------------------------------------
// system parameter structures
//--------------------------------------------------------------------------

void Convert (SParameter *s, EBool convert)
{
    if (!convert) return;     // only do conversion if requested
    Convert ((SModuleBase *)s, convert);  // convert the inherited members
    g_Convert (s->type);
}


DefineConvert (SSingleNetworkParameter, SNetworkParameter)
    DoConvert (value);
}

DefineConvert (SNetworkParameter, SParameter)
    DoConvert (networkID);
}

DefineConvert (SNetDistribution, SNetworkParameter)
    DoConvert (data.dist);
    DoConvert (data.low);
    DoConvert (data.high);
    DoConvert (data.mean);
    DoConvert (data.constant);
    DoConvert (data.freeparam);
}

DefineConvert (SNetDistributionP2P, SNetDistribution)
    DoConvert (sender);
    DoConvert (receiver);
}

DefineConvert (SApplDistribution, SParameter)
    DoConvert (data.dist);
    DoConvert (data.low);
    DoConvert (data.high);
    DoConvert (data.mean);
    DoConvert (data.constant);
    DoConvert (data.freeparam);
}

DefineConvert (SApplWobbleDrift, SApplDistribution)
    DoConvert (period);
    DoConvert (rhoMin);
    DoConvert (rhoMax);
}

DefineConvert (SNetworkData, SSingleNetworkParameter)
    DoConvert (sender);
    DoConvert (receiver);
}

DefineConvert (SSingleParameter, SParameter)
    DoConvert (value);
}

DefineConvert (STimeParameter, SParameter)
    DoConvert (macrostamp);
    DoConvert (timestamp);
}

DefineConvert (STransmissionData, SParameter)
    DoConvert (networkID);
    DoConvert (nodeID);
    DoConvert (value);
}

DefineConvert (SRateData, STransmissionData)
    DoConvert (area);
}

DefineConvert (SEventReport, SSingleParameter)
    DoConvert (set);
    DoConvert (masksize);
}


//--------------------------------------------------------------------------
// system command structures
//--------------------------------------------------------------------------

DefineConvert (SCommand, SModuleBase)
    DoConvert (type);
}

DefineConvert (SGpsConnectCommand, SCommand)
    DoConvert (gpu);
}

DefineConvert (SApplCommand, SCommand)
    DoConvert (nodeID);
    DoConvert (area);
}

DefineConvert (SNetworkCreateCommand, SCommand)
    DoConvert (networkType);
}

DefineConvert (SNetworkCommand, SCommand)
    DoConvert (networkID);
}

DefineConvert (SNetworkConnectCommand, SNetworkCommand)
    DoConvert (nodeID);
}


//--------------------------------------------------------------------------
// reply structures
//--------------------------------------------------------------------------

DefineConvert (SReply, SModuleBase)
    DoConvert (type);
}

DefineConvert (SBooleanReply, SReply)
    DoConvert (result);
}

DefineConvert (SDoubleReply, SReply)
    DoConvert (value);
}


