//
// evalfunc.cpp
//
// provides functions for the evaluation
//
// created: 01.02.97
//
// last change: 05.02.99
//
// RCS version control information:
//     $Revision: 1.25 $$State: Exp $$Date: 1999/02/05 15:34:47 $
//


#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <stream.h>    // must stand *before* iostream, otherwise 'output'...
#include <iostream.h>  // ...is not defined!
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "bool.hpp"
#include "constant.hpp"
#include "error.hpp"
#include "event.hpp"
#include "sockstre.hpp"
#include "structs.hpp"
#include "time.hpp"

#include "evalcfg.hpp"   // configuration (all necessary globals)
#include "evalconv.hpp"  // Convert() functions for all events
#include "evalfunc.hpp"  // prototypes of useful functions
#include "evalglob.hpp"  // structure for receiving replies

#ifndef DOS_SIM
  #include <sys/time.h>
#endif

#ifdef PLOT_DATA
  #define PLOT_DIR "gnuplot/"
  // this define is put before the names of the files needed to produce
  // the gnuplot data
  // note: do not forget the slash after the name! if you want the files
  // in the working directory, then you can leave the string empty ("")
#endif


#ifndef NDEBUG
  extern CStreamServer *serverSocket;
  // necessary to remove the server socket before an Assert() fails

  #ifdef Assert
    #undef Assert
  #endif
  #define Assert(expr) \
    if (!(expr)) \
    { \
        (void)delete serverSocket; \
        (void)ReportError (Asserting, __FILE__, __LINE__, #expr); \
    } // if
#endif


//**************************************************************************
// necessary globals
//**************************************************************************

static double piMin = 1;  // minimum precision of all snapshot rounds
static double piMax = 0;  // maximum precision of all snapshot rounds
static double piSum = 0;  // sum of all precisions (from all snapshot rounds)
static INT32 piRoundNumber = 0;  // number of encountered snapshot rounds
// best, worst and average precision (average = piSum/piRoundNumber)
// note: in fact, the piMin/Max/Sum values are not from all rounds,
// but their computation is started from a user-definable round on
// (the user sets the synchronization round in SEvalsysConfigData, and
// from this value the first valid snapshot round is computed)
// if there are not enough rounds to compute precision values, function
// PrintPrecision() states that it has not collected any data


static INT32 g_Convert (INT32 src);
// if g_byteOrderReversed is True, the reverted value of 'src' is returned;
// if it is False, 'src' is returned unchanged

#ifdef PLOT_DATA
  // used for the output of simulation data
  static ostream outPik[numberOfNodes-1];  // k-th precision
  static ostream outState[numberOfNodes];  // state
  static ostream outAdj[numberOfNodes];  // adjustments
  static ostream outAcc[numberOfNodes];  // accuracy
  static ostream outRate[numberOfNodes];  // rates
  static ostream outGammak[numberOfNodes-1];
  static ostream outDrift[numberOfNodes];  // drifts
  static ostream outMult[numberOfNodes];

  const double TwoPowMinus23 = 1.192092e-07;
  // factor for accuracy

  extern char *ftoa (double f, int width, char *string);  // from ftoa.cpp
  // returns 'f' as a string with at least 'width' characters in %e format;
  // stores the string in 'string', which must be big enough to hold the
  // converted string; if 'string' is NULL, NULL is returned
  // don't forget the link and entry in the Makefile!

  static void SortArray (double* unsortedArray,double* sortedArray,INT32 size);
  // sort an array of doubles

  static double PrecisionKth (double* pi, INT32 upper, INT32 order);
  // computes the kth precision among pi[0],...,pi[upper-1] which
  // are supposed to be sorded ascendingly
  // if order=0 -> maximum distance
  // if order=1 -> 2nd maximum distance
  // etc.
#endif


//**************************************************************************
// function implementations
//**************************************************************************

//--------------------------------------------------------------------------
// message transmission
//--------------------------------------------------------------------------

// send the message (message header and data) over the socket
static void SendData (CStreamInterface *sock, EMessageType msgheader,
                      SBase *data)
{
    Assert (sock != NULL);
    sock->SendMessage (msgheader, data);
}


//--------------------------------------------------------------------------
// parameter/command transmission
//--------------------------------------------------------------------------

// send the command over the socket
void ExecuteCommand (CStreamInterface *sock, SCommand *command)
{
    SendData (sock, MsgCommand, command);
}


// send the parameter over the socket
void SetParameter (CStreamInterface *sock, SParameter *param)
{
    SendData (sock, MsgSetParameter, param);
}


// send the parameter over the socket, get the reply, modify param->value
EBool GetParameter (CStreamInterface *sock, SSingleParameter *param)
{
    SendData (sock, MsgGetParameter, param);
    SStructureReply *reply;
    if (ReceiveReply (sock, (SReply**)&reply))
    {
        param->value = ((SSingleParameter*)(reply->data))->value;
        delete reply;
        return True;
    }
    else return False;
}


//--------------------------------------------------------------------------
// file output in gnuplot format
//--------------------------------------------------------------------------

// prints the line "<x> <y>" on stream 'out'
static void PlotPoint (ostream& out, double x, double y)
{
    //out << x << " " << y << endl;
    // note: this fast version does not work; the plot is not correct

    char buf[40];

    ftoa (x, 14, buf);
    out << buf << " ";

    ftoa (y, 14, buf);
    out << buf << endl;
}

// prints the line "<x> <y> <l> <u>" on stream 'out'
static void PlotInterval (ostream& out, double x, double y, double l, double u)
{
    //out << x << " " << y << " " << l << " " << u << endl;
    // note: this fast version does not work; the plot is not correct

    char buf[40];

    ftoa (x, 14, buf);
    out << buf << " ";

    ftoa (y, 14, buf);
    out << buf << " ";

    ftoa (l, 14, buf);
    out << buf << " ";

    ftoa (u, 14, buf);
    out << buf << endl;
}


//--------------------------------------------------------------------------
// precision management
//--------------------------------------------------------------------------

static INT32 FirstSnapshotPrecisionRound =
    (INT32)((evalsysConfigData.FirstSyncPrecisionRound*
             simutcConfigData.commonNodeData.RoundPeriod)/
             evalsysConfigData.SnapshotPeriod);
// this is the first snapshot round that is taken into consideration when
// computing the precision
// do not confuse it with the synchronization round: the number of snapshot
// rounds depends on the SnapshotPeriod of the Supervisor triggering the
// snapshots, whereas the number of sync. rounds depends on the RoundPeriod
// of the CSAs

// compute precision per round, store it in global precision data
// piMin, piMax, piSum and piRoundNumber
static void AddPrecision (double accuracy, double simTime)
{
    static double currTime = -1;
    // the time of the current snapshot round; it is set to -1 at the
    // beginning to indicate that the initial values of currPiMax/Min are
    // not valid; from then on, it is always set to the 'simTime' value
    // of the last call

#ifdef PLOT_DATA
    static double TradAcc[numberOfNodes];
    static double TradAccSorted[numberOfNodes];
    static int fill = 0;

    char buf[40];
#else
    static double currAlphaMin;  // minimum accuracy of current round
    static double currAlphaMax;  // maximum accuracy of current round
#endif

    // the following code is parted into two blocks: one if we start
    // a new snapshot round, and one if the accuracy belongs to the
    // same round as that of the previous call

    if (simTime != currTime)  // we start a new snapshot round
    {
        piRoundNumber++;  // increment the round counter

        // we start a new round, so we have to check if the old round
        // was valid; if it was, we have to compute the precision from
        // the minimum and maximum accuracies that we have computed
        // for this round; the precision only affects the global statistics
        // if the snapshot round should be counted

        if ((currTime != -1) &&  // the old round has left valid data
            (piRoundNumber > FirstSnapshotPrecisionRound))  // round is counted
        {
            // if the last round was valid and the round should be counted,
            // we have to compute the precision of this previous round and
            // this value is taken over into the global variables
            // piMin/Max/Sum

            double piLastRound;   // precision of the previous round

#ifdef PLOT_DATA
            SortArray (TradAcc, TradAccSorted, fill);

            piLastRound = PrecisionKth (TradAccSorted, fill, 0);

            // output simulation data
            for (int i=0; i<numberOfNodes-1; i++)
            {
                PlotPoint (outPik[i], currTime,
                    PrecisionKth (TradAccSorted, fill, i));
                // plot the time (from the former round) and the i-th
                // precision
            } // for
#else
            piLastRound = currAlphaMax - currAlphaMin;
            // compute the precision of the last round
#endif

            if (piMax < piLastRound) piMax = piLastRound;
            if (piMin > piLastRound) piMin = piLastRound;
            // adapt the minimum and maximum precision if necessary

            piSum += piLastRound;
            // add the precision to the sum value
        } // if

        // now that the old round has been handled, consider the new data
        // we have got
        // remember that we are in a new round; so the 'accuracy' we
        // got is the best and worst accuracy we currently have and
        // we have to store it as minimum and as maximum

#ifdef PLOT_DATA
        fill = 0;
        TradAcc[fill] = accuracy;
        fill++;
        // first entry for the new snapshot
#else
        currAlphaMax = accuracy;  // store the minimum accuracy
        currAlphaMin = accuracy;  // store the maximum accuracy
#endif

        currTime = simTime;  // store the time of the new snapshot round
    }
    else  // the data belongs to the current snapshot round
    {
#ifdef PLOT_DATA
        Assert( fill < numberOfNodes );
        TradAcc[fill] = accuracy;
        fill++;
#else
        // if the accuracy belongs to the same round as that of the
        // previous call, we just have to update the minimum/maximum
        // accuracy values if necessary

        if (currAlphaMax < accuracy) currAlphaMax = accuracy;
        // if the 'accuracy' is greater than the current maximum -> store it

        if (currAlphaMin > accuracy) currAlphaMin = accuracy;
        // if the 'accuracy' is less than the current minimum -> store it
#endif
    } // else
}


//--------------------------------------------------------------------------
// consonance management
//--------------------------------------------------------------------------

#ifdef PLOT_DATA
// compute consonance per round, store it in global consonance data
static void AddConsonance (double rate, double simTime)
{
    static double currTime = -1;
    // the time of the current snapshot round; it is set to -1 at the
    // beginning to indicate that the initial values of currPiMax/Min are
    // not valid; from then on, it is always set to the 'simTime' value
    // of the last call

    static double CurrentRate[numberOfNodes];
    static double CurrentRateSorted[numberOfNodes];
    static int fill = 0;

    char buf[40];

    if (simTime != currTime)  // new snapshot round
    {
        if (currTime != -1)   // compute consonance of previous round
        {
            double consLastRound;

            SortArray (CurrentRate, CurrentRateSorted, fill);

            consLastRound = PrecisionKth (CurrentRateSorted, fill, 0);

            for (int i=0; i<numberOfNodes-1; i++)
            {
                PlotPoint (outGammak[i], currTime,
                    PrecisionKth (CurrentRateSorted, fill, i)*1e6);
                // plot the time (from the former round) and the i-th
                // consonance
            } // for
        } // if

        fill = 0;
        CurrentRate[fill] = rate;
        fill++;
        // first entry for the new snapshot

        currTime = simTime;
        // store the time of the new snapshot round
    }
    else  // the data belongs to the current snapshot round
    {
        Assert( fill < numberOfNodes );
        CurrentRate[fill] = rate;
        fill++;
    }
}
#endif


//--------------------------------------------------------------------------
// event printing
//--------------------------------------------------------------------------

// print events from the network
// note: you must convert the event, even if its data is not needed;
// the calling function depends on it
static void PrintNetworkEvent (SEvent *event, _IO_ostream_withassign& out)
{
    SNetEvent *netev = (SNetEvent *)event;  // basic network event
    SNetMessageEvent *msgev = (SNetMessageEvent *)event;
    SNetCrashEvent *netcrash = (SNetCrashEvent *)event;

    out << "network (" << g_Convert (netev->networkID) << "): ";
    // every network event output starts with the network ID

    switch (g_Convert (event->type))
    {
        case CrashStatusChange:
            Convert (netcrash, g_byteOrderReversed);
            if (netcrash->crashed) out << "crashed";
            else out << "recovered";
            break;
        case DeliverMessage:
            Convert (msgev, g_byteOrderReversed);
            out << "delivered message (" << msgev->sender;
            out << ", " << msgev->area << ", " << msgev->receiver;
            out << ", " << msgev->msgID << ")";
            break;
        case EarlyMessage:
            Convert (msgev, g_byteOrderReversed);
            out << "sent early message (" << msgev->sender;
            out << ", " << msgev->area << ", " << msgev->receiver;
            out << ", " << msgev->msgID << ")";
            break;
        case LateMessage:
            Convert (msgev, g_byteOrderReversed);
            out << "sent late message (" << msgev->sender;
            out << ", " << msgev->area << ", " << msgev->receiver;
            out << ", " << msgev->msgID << ")";
            break;
        case MessageLost:
            Convert (msgev, g_byteOrderReversed);
            out << "discarded message (" << msgev->sender;
            out << ", " << msgev->area << ", " << msgev->receiver;
            out << ", " << msgev->msgID << ")";
            break;
        case SendMessage:
            Convert (msgev, g_byteOrderReversed);
            out << "transmitted message (" << msgev->sender;
            out << ", " << msgev->area << ", " << msgev->receiver;
            out << ", " << msgev->msgID << ")";
            break;

        default:
            Convert (netev, g_byteOrderReversed);
            out << "unknown event: " << event->type;
            break;
    } // switch
}


// converts a module number into a string constant
// (same function as in error.cpp)
static const char *ModuleToString (EModule module)
{
    switch (module)
    {
        case Controller: return "Controller"; break;
        case Network: return "Network"; break;
        case Supervisor: return "Supervisor"; break;
        case ClockSync: return "ClockSync"; break;
        case Clock: return "Clock"; break;
        case Lance: return "Lance"; break;

        default: return "Unknown"; break;
    } // switch
}

// print the data of one message
// format: "message <msgID> (<sender>,<area> => <receiver>) "
static void PrintMsg (SINT32 msgID, SINT32 sender, int area, SINT32 receiver,
                      _IO_ostream_withassign& out)
{
    out << "message " << msgID << " (" << sender << "," << area;
    out << " => " << receiver << ") ";
}

// print events from the CSA
// note: you must convert the event, even if its data is not needed;
// the calling function depends on it
static void PrintCSAEvent (SEvent *event, _IO_ostream_withassign& out)
{
    SApplEvent *appev = (SApplEvent *)event;
    SApplClockEvent *ev = (SApplClockEvent *)event;
    SApplPreloadEvent *pre = (SApplPreloadEvent *)event;
    SApplGpsEvent *gpsev = (SApplGpsEvent *)event;
    SApplCrashEvent *crashev = (SApplCrashEvent *)event;
    SApplSendMsgEvent *appsmsg = (SApplSendMsgEvent *)event;
    SApplRecvMsgEvent *apprmsg = (SApplRecvMsgEvent *)event;
    SApplDelayEvent *appdelay = (SApplDelayEvent *)event;
    SApplDelayEstimationEvent *appest = (SApplDelayEstimationEvent *)event;
    SApplNetEvent *netev = (SApplNetEvent *)event;
    SApplTimeEvent *intev = (SApplTimeEvent *)event;
    SApplRateEvent *rateev = (SApplRateEvent *)event;

#ifdef PLOT_DATA
    int nodeIndex;
    double tradacc;
    double edge_lower, edge_upper;
    double rate;
    static double state_old[numberOfNodes];
    static double time_old[numberOfNodes];
    static EBool inhibit[numberOfNodes];
    static EBool firstCall = True;

    if (firstCall)  // initialize the inhibit[] array
    {
        for (int i=0; i<numberOfNodes; i++) inhibit[i] = True;
        firstCall = False;
    } // if
#endif

    out << "node (" << g_Convert (appev->nodeID);
    out << "," << g_Convert (appev->area) << "): ";
    // every CSA event output starts with the node ID


    // check if we should update the round counter
    if ((g_Convert(appev->type) == evalsysConfigData.RoundEvent) &&
        (g_Convert(appev->nodeID) == evalsysConfigData.RoundNodeID) &&
        (g_Convert(appev->area) == evalsysConfigData.RoundArea))
    {
        round++; // update round counter
    } // if


    switch (g_Convert(event->type))
    {
        case ChangeAssociation:  // associated network was changed
            Convert (netev, g_byteOrderReversed);
            out << "associated network set to " << netev->networkID;
            break;
        case ClockPreload:  // this is a clock preload event
            Convert (pre, g_byteOrderReversed);
            out << "clock preloaded from ";
            out << pre->loadAt << " to ";
            out << NtpToSeconds (pre->newTime);
            if ((ev->nodeID == evalsysConfigData.RoundNodeID) &&
                (ev->area == evalsysConfigData.RoundArea))
                round++; // update round counter
            break;
        case ClockReadFault:  // clock returned wrong value on read access
            Convert (appev, g_byteOrderReversed);
            out << "clock return arbitrary value on read";
            break;
        case ClockUpdate:  // this is a clock event
            Convert (ev, g_byteOrderReversed);
            out << "clock set from ";
            out << NtpToSeconds (ev->clockTime);
            out << " to ";
            out << NtpToSeconds (ev->newTime);

#ifdef PLOT_DATA
            // --- output simulation data

            nodeIndex = ev->nodeID-1;
            Assert (nodeIndex >= 0);
            // index is nodeID-1, so it starts from 0

            PlotPoint (outAdj[nodeIndex], event->time,
                NtpToSeconds (ev->newTime) - NtpToSeconds (ev->clockTime));
            // print the pair <time, adjustment>

            inhibit[nodeIndex] = True;
            // skip the next snapshot
#endif

            break;
        case ClockWriteFault:  // clock was set to wrong value on write access
            Convert (appev, g_byteOrderReversed);
            out << "clock is set to wrong value on write";
            break;
        case ConvergenceFailed:  // convergence function failed
            Convert (appev, g_byteOrderReversed);
            out << "convergence function failed";
            break;
        case CrashStatusChange:  // crash status has changed
            Convert (crashev, g_byteOrderReversed);
            if (crashev->crashed) out << "crashed";
            else out << "recovered";
            break;
        case DelayEstimation:  // new estimated transmission delay
            Convert (appest, g_byteOrderReversed);
            out << "estimated delay from " << appest->sender;
            out << " on network " << appest->networkID << ": ";
            out << "(" << appest->delay.delta << "; ";
            out << appest->delay.delta - appest->delay.epsilonMinus << ", ";
            out << appest->delay.delta + appest->delay.epsilonPlus << ")";
            break;
        case DiscardRecvMessage:  // received message was discarded
            Convert (apprmsg, g_byteOrderReversed);
            PrintMsg (apprmsg->msgID, apprmsg->sender, apprmsg->senderArea,
                      apprmsg->nodeID, out);
            out << " was discarded by receiver";
            break;
        case DiscardSendMessage:  // sent message was discarded
            Convert (appsmsg, g_byteOrderReversed);
            PrintMsg (appsmsg->msgID, appsmsg->nodeID, appsmsg->area,
                      appsmsg->receiver, out);
            out << " was discarded by sender";
            break;
        case GpsEvent:  // this is a GPS update event
            Convert (gpsev, g_byteOrderReversed);
            out << "clock differs from GPS time by ";
            out << gpsev->diff << " seconds";
            break;
        case IgnoreSetting:  // clock has ignored register access
            Convert (appev, g_byteOrderReversed);
            out << "clock has ignored register access";
            break;
        case IntervalEvent:  // time interval from convergence function
            Convert (intev, g_byteOrderReversed);
            if (intev->input) out << "CV got input interval (";
            else out << "CV computed interval (";
            out << NtpToSeconds (intev->interval.low) << ", ";
            out << NtpToSeconds (intev->interval.high) << "; ";
            out << NtpToSeconds (intev->interval.refpoint) << ")";
            break;
        case RateEvent:  // rate interval from rate algorithm
            Convert (rateev, g_byteOrderReversed);
            if (rateev->input) out << "rate alg. got input rate (";
            else out << "rate alg. computed rate (";
            out << rateev->rate.thetaMinus << ", ";
            out << rateev->rate.thetaPlus << "; ";
            out << rateev->refPoint << "; ";
            out << rateev->rate.u << ")";

#ifdef PLOT_DATA
            // --- output simulation data

            edge_lower = 1.0 - rateev->rate.thetaMinus;
            edge_upper = 1.0 + rateev->rate.thetaPlus;
            // edges of the local rate interval

            nodeIndex = rateev->nodeID-1;
            Assert (nodeIndex >= 0);
            // index is nodeID-1, so it starts from 0

            if (!rateev->input)   // only use the computed rate
            {
                PlotInterval (outDrift[nodeIndex], event->time + nodeIndex*1.0,
                    1.0, edge_lower, edge_upper);
                // put some space between the nodes, otherwise their
                // data is plotted at the same x-position

                PlotPoint (outMult[nodeIndex], event->time,
                    (rateev->refPoint - 1) * 1e6);
            } // if
#endif

            break;
        case ReceivedMessage:  // message was received
            Convert (apprmsg, g_byteOrderReversed);
            PrintMsg (apprmsg->msgID, apprmsg->sender, apprmsg->senderArea,
                      apprmsg->nodeID, out);
            out << " was received";
            break;
        case ScrambleRecvMessage:  // incoming message was scrambled
            Convert (apprmsg, g_byteOrderReversed);
            PrintMsg (apprmsg->msgID, apprmsg->sender, apprmsg->senderArea,
                      apprmsg->nodeID, out);
            out << " was scrambled by receiver";
            break;
        case ScrambleSendMessage:  // outgoing message was scrambled
            Convert (appsmsg, g_byteOrderReversed);
            PrintMsg (appsmsg->msgID, appsmsg->nodeID, appsmsg->area,
                      appsmsg->receiver, out);
            out << " was scrambled by sender";
            break;
        case DuplicateRecvMessage:  // incoming message was duplicated
            Convert (apprmsg, g_byteOrderReversed);
            PrintMsg (apprmsg->msgID, apprmsg->sender, apprmsg->senderArea,
                      apprmsg->nodeID, out);
            out << " was duplicated by receiver";
            break;
        case DuplicateSendMessage:  // outgoing message was duplicated
            Convert (appsmsg, g_byteOrderReversed);
            PrintMsg (appsmsg->msgID, appsmsg->nodeID, appsmsg->area,
                      appsmsg->receiver, out);
            out << " is a duplication at sender";
            // note: the msg ID is that of the duplicated message, not
            // that of the original message (however, most of the time,
            // the msg ID of the original msg will be msgID-1)
            break;
        case SendMessage:  // message was sent
            Convert (appsmsg, g_byteOrderReversed);
            PrintMsg (appsmsg->msgID, appsmsg->nodeID, appsmsg->area,
                      appsmsg->receiver, out);
            out << " was sent";
            break;
        case Snapshot:  // system snapshot was taken
            Convert (ev, g_byteOrderReversed);
            out << "clock made snapshot ";
            out << NtpToSeconds (ev->newTime);
            out << " (acc = [" << ((ev->newTime.acc) & 0x0FFFF) << ", ";
            out << ((ev->newTime.acc>>16) & 0x0FFFF) << "])";
            AddPrecision (ev->time - NtpToSeconds (ev->newTime), ev->time);

#ifdef PLOT_DATA
            // --- output simulation data

            nodeIndex = ev->nodeID-1;
            Assert (nodeIndex >= 0);
            // index is nodeID-1, so it starts from 0

            tradacc = NtpToSeconds (ev->newTime) - event->time;
            // current tradition accuracy

            edge_lower = tradacc -
                (double)((ev->newTime.acc >> 16) & 0x0FFFF) * TwoPowMinus23;
            edge_upper = tradacc +
                (double)((ev->newTime.acc) & 0x0FFFF) * TwoPowMinus23;
            // lower/upper accuracies

            if (!inhibit[nodeIndex])  // rate can be computed
                rate = (NtpToSeconds (ev->newTime) - state_old[nodeIndex]) /
                    (event->time - time_old[nodeIndex]);
            else
                rate = 1.0;
            // current rate

            state_old[nodeIndex] = NtpToSeconds (ev->newTime);
            time_old[nodeIndex]  = event->time;
            // remember for the next snapshot

            PlotPoint (outState[nodeIndex], event->time, tradacc);

            PlotInterval (outAcc[nodeIndex], event->time + nodeIndex*0.4,
                tradacc, edge_lower, edge_upper);


            if (!inhibit[nodeIndex])  // print the rate
            {
                PlotPoint (outRate[nodeIndex], event->time + nodeIndex*0.4,
                    (rate-1) * 1e6);
                AddConsonance (rate, ev->time);
            }
            else inhibit[nodeIndex] = False;
            // from the second call on, we can print the rate
#endif

            break;
        case TimerEvent:  // a clock timer has produced interrupt
            Convert (appev, g_byteOrderReversed);
            out << "duty timer interrupt for module ";
            out << ModuleToString (appev->module);
            break;
        case TransmissionDelay:  // measured transmission delay
            Convert (appdelay, g_byteOrderReversed);
            PrintMsg (appdelay->msgID, appdelay->sender, appdelay->senderArea,
                      appdelay->nodeID, out);
            out << " was transmitted in " << NtpToSeconds(appdelay->delaytime);
            out << " seconds";
            break;
        case UserEvent:  // contains user-specific data
            Convert (appev, g_byteOrderReversed);
            out << "user defined event occured";

        default:
            Convert (appev, g_byteOrderReversed);
            out << "unknown event: " << event->type;
            break;
    } // switch
}


// print the event data to cout
static void PrintEvent (SEvent *event)
{
    SApplEvent *appev;  // for events from a CSA

    _IO_ostream_withassign out;  // output stream

#ifdef NO_PRINTEVENT  // redirect event output to /dev/null
    filebuf fbuf;
    if (fbuf.open ("/dev/null", "w") == 0)
    {
        cerr << "cannot open /dev/null" << endl;
        out = cout;  // use cout instead
        out.setf (ios::fixed, ios::floatfield);
        // always print 6 (default) digits
    }
    else
    {
        ostream tmpout (&fbuf);  // output stream
        out = tmpout;
    }
#else
    out = cout;
    out.setf (ios::fixed, ios::floatfield);
    // always print 6 (default) digits
#endif

    switch (g_Convert(event->module))
    {
        case Network:
            PrintNetworkEvent (event, out);
            // the call also converts the event if necessary
            break;
        case Lance:
        case ClockSync:
        case Clock:
            PrintCSAEvent (event, out);
            // the call also converts the event if necessary
            break;

        default:
            Convert (event, g_byteOrderReversed);  // convert base event
            out << "module not supported: " << event->module;
    } // switch

    out << " at simulation time " << event->time;

    switch (g_Convert(event->module))
    {
        case Lance:      // convert it to an application event, extract
        case ClockSync:  // the clock time
        case Clock:
            appev = (SApplEvent *)event;
            out << " (";
            out << NtpToSeconds (appev->clockTime);
            out << ")";
            break;
    } // switch

    out << endl;
}


//--------------------------------------------------------------------------
// message reception
//--------------------------------------------------------------------------

// returns True if the message with the appropriate header was received,
// otherwise False; if the message was received sucessfully, it is returned
// in 'buf'
EBool ReceiveMessage (CStreamInterface *sock, EMessageType header,
                      SBase **buf)
{
    Assert (sock != NULL);
    Assert (buf != NULL);

    EMessageType type;
    int failureCount = 0;  // number of different messages encountered

    do {
        // wait for the reply (block for BlockTime seconds)
        if (!sock->ReceiveMessageHeader (&type, evalsysConfigData.BlockTime))
        {
            // that can only be EOF (socket killed???)
            // (when using timeouts, it can also be a timeout)
            Report (evalsys: encountered timeout or EOF reading header);
            return False;
        }

        if (type != header)
        {
            if (failureCount == 0)
            {
                cerr << "evalsys: encountering unexpected message(s)." << endl;
            }
            failureCount++;

            if (sock->ReceiveMessage (buf))
            // read a message if it's there, process it
            {
                switch (type)
                {
                    case MsgEvent:
                        PrintEvent ((SEvent *)(*buf));
                        break;
                    case MsgReply:
                        cerr << "evalsys: was waiting for header " << header;
                        cerr << ", have received a reply" << endl;
                        break;

                    default:
                        cerr << "evalsys: was waiting for header " << header;
                        cerr << ", have received " << type << endl;
                        break;
                } // switch

                sock->ReturnMessageBuffer (*buf);  // return the buffer
            } // if
        } // if
        else if (!sock->ReceiveMessage (buf))
        {
            return False;
        }
    } while (type != header);

    if (failureCount > 0)
    {
        cerr << "evalsys: encountered " << failureCount << " different";
        cerr << " messages before reading the expected message" << endl;
    }
    return True;
}


//--------------------------------------------------------------------------
// special message reception
//--------------------------------------------------------------------------

// receives an event and prints it to cout; returns True if the event
// was received sucessfully, else False
EBool ReceiveEvent (CStreamInterface *sock)
{
    SBase *buf;  // msg buffer

    if (!ReceiveMessage (sock, MsgEvent, &buf))
    {
        Report (evalsys: msg was no event);
        return False;
    }

    PrintEvent ((SEvent *)buf);
    // assume that the buffer contains an event

    sock->ReturnMessageBuffer (buf);
    // return the buffer

    return True;
}


// receives a reply, allocates the appropriate reply structure and returns
// it in 'reply'; returns True if the reply was received successfully,
// else False
EBool ReceiveReply (CStreamInterface *sock, SReply **reply)
{
    Assert (sock != NULL);
    Assert (reply != NULL);

    SBase *buf;  // msg buffer

    if (!ReceiveMessage (sock, MsgReply, &buf))
    {
        Report (evalsys: msg was no reply);
        return False;
    }

    if (buf->size < sizeof(SReply))
    {
        Report (evalsys: reply was not complete);
        sock->ReturnMessageBuffer (buf);
        return False;
    }

    char *param;  // for structure replies
    INT32 size;  // size of the param structure

    switch (g_Convert(((SReply*)buf)->type))
    {
        case ReplyBoolean:
            *reply = new SBooleanReply;
            memcpy (*reply, buf, buf->size);  // copy type and data
            Convert ((SBooleanReply *)*reply, g_byteOrderReversed);
            break;
        case ReplyDouble:
            *reply = new SDoubleReply;
            memcpy (*reply, buf, buf->size);  // copy type and data
            Convert ((SDoubleReply *)*reply, g_byteOrderReversed);
            break;
        case ReplyStructure:
            size = buf->size - sizeof(SReply);
            // size of the param structure
            param = new char[size];
            memcpy (param, (char*)buf+sizeof(SReply), size);
            // copy the parameter structure
            *reply = new SStructureReply (param, size);
            Convert ((SStructureReply *)*reply, g_byteOrderReversed);
            break;

        default:  // unknown type
            Report (evalsys: reply type was unknown);
            sock->ReturnMessageBuffer (buf);
            return False;
    } // switch

    sock->ReturnMessageBuffer (buf);  // not needed anymore
    return True;
}


// waits for a reply, prints an error message containing 'command' to cerr
// if (a) there was no reply, (b) it was no boolean reply, or (c) the reply
// contained the boolean value False
void CheckBooleanReply (CStreamInterface *sock, const char *command)
{
    SBooleanReply *boolReply;
    if (!ReceiveReply (sock, (SReply **)&boolReply))
    {
        cerr << "error receiving a reply" << endl;
        return;
    }
    if (boolReply->type != ReplyBoolean) Report (evalsys: wrong reply!);
    if (boolReply->result != True)
    {
        cerr << command << ": have received the reply ";
        cerr << boolReply->result << endl;
    }
    delete boolReply;
}


//--------------------------------------------------------------------------
// socket management
//--------------------------------------------------------------------------

#ifndef DOS_SIM
// select() is only known in UNIX, not in DOS

// waits for the supervisor sockets to become readable (using select);
// returns the index of the first socket that is readable (if more than one
// socket is readable, the smallest index is returned)
int SelectSocket ()
{
    fd_set readSet;  // read mask for select()
    timeval tv;  // for the timeout of select()
    int err;  // return value of select()
    static INT32 maxsid=0;  // the highest supervisor socket ID

    FD_ZERO (&readSet);  // initialize the set

    for (int i=0; i<numberOfNodes+1; i++)
    {
        FD_SET (supsid[i], &readSet);
        // set the read mask for the socket IDs

        if (supsid[i] > maxsid) maxsid = supsid[i];
        // find out the highest socket ID
    }

    tv.tv_sec  = (long)evalsysConfigData.SelectTimeout;  // extract the seconds
    tv.tv_usec = (long)(evalsysConfigData.SelectTimeout*1e6) % (long)1e6;
    // extract the microseconds

    // now wait for the socket to become readable
    err = select (maxsid+1, &readSet, 0, 0, &tv);

    if (err == -1)
    {
        Report (error using select);
        return NoAccount;
    }
    if (err == 0) return NoAccount;  // timeout

    // now check the read mask for the first socket that is set
    for (int i=0; i<numberOfNodes+1; i++)
    {
        if (FD_ISSET (supsid[i], &readSet)) return i;
    }

    return NoAccount;  // should never happen
}

#else
int SelectSocket ()  // just some dummy function
{
    return NoAccount;
}
#endif


// checks if the socket with the given address already exists and returns
// True if the check succeeds, False if the socket does not exist
// the function performs its check by creating a socket with this address
// and type and by checking if this creation (call to s_create()) fails;
// if it fails, the function returns True, otherwise it deletes the socket
// again and returns False
EBool SocketExists (const ACSocketAddress *addr, ESocketType type)
{
    Assert (addr != NULL);

    INT32 sid;

    if (s_create (addr, type, &sid) != SOkay) return True;
    // note: this also returns True if the socket could not be created
    // due to some other reason!

    // if we come here, then the socket was created successfully
    s_delete (sid, addr);  // close the socket again
    return False;
}


//--------------------------------------------------------------------------
// 'normalize' a double time value so that it is > 1
//--------------------------------------------------------------------------

// converts 'input' (in seconds) to a double >= 1, stores the unit
// needed in 'unit'; unit should be at least 8 characters wide
static void Normalize (double input, double& output, char *unit)
{
    int i;
    short sign = 0;  // sign of 'input'; if input < 0 -> sign = 1
    char *unitname[5] = { "s", "ms", "us", "ns", "ps" };
    // the unit of the output value

    if (input == 0)  // just copy the zero
    {
        output = 0;
        strcpy (unit, unitname[0]);
        return;
    }

    if (input < 0)  // make input positive (makes computation easier)
    {
        input = -input;
        sign = 1;
    }

    for (i=0; (i < 4) && (input < 1); i++) input = input * 1000;
    // now 'input' is >= 1, and i is the power (base 10)

    if (sign) output = -input;  // copy the result to 'output'
    else output = input;

    // check the power, get the appropriate unit name
    if ((i >= 0) && (i < 5)) strcpy (unit, unitname[i]);
    else strcpy (unit, "unknown");
}


//--------------------------------------------------------------------------
// printing the configuration
//--------------------------------------------------------------------------

// prints the value to 'o' in the form "<param> = <norm> <unit>", where
// 'norm' and 'unit' are obtained from Normalize()
static void PrintNormal (ostream& o, const char* param, double value)
{
    double norm;   // "normalized" number for output
    char unit[8];  // unit name

    Normalize (value, norm, unit);
    o << param << " = " << norm << " " << unit << endl;
}

// prints the string "probability of <prob> = <value>" to 'o'
static void PrintProbability (ostream& o, const char *prob, double value)
{
    o << "probability of " << prob << " = " << value << endl;
}

// prints the configuration data
void PrintConfiguration ()
{
    double norm;   // "normalized" number for output
    char unit[8];  // unit name

    cout << endl << "configuration of the simulation:" << endl;

    if (numberOfNetworks > 0)  // pure simulation configured a network
    {
        cout << endl << "  Networks: " << numberOfNetworks << endl;
        cout << "    All nodes are connected to the first network." << endl;
        cout << "    "; PrintProbability (cout, "network crashes",
                            simutcConfigData.networkData.ProbNetCrash);
        cout << "    "; PrintProbability (cout,"network recovery",
                            simutcConfigData.networkData.ProbNetRecover);
        cout << "    "; PrintProbability (cout, "early messages",
                            simutcConfigData.networkData.ProbNetEarly);
        cout << "    "; PrintProbability (cout, "late messages",
                            simutcConfigData.networkData.ProbNetLate);
        cout << "    "; PrintNormal (cout, "lambda",
                            simutcConfigData.networkData.Lambda);
        cout << "    "; PrintNormal (cout, "dmin",
                            simutcConfigData.networkData.Dmin);
        cout << "    "; PrintNormal (cout, "dmax",
                            simutcConfigData.networkData.Dmax);
        cout << "    for all nodes i,j with i!=j:" << endl;
        cout << "       "; PrintNormal (cout, "delta",
                               simutcConfigData.networkData.Delta);
        cout << "       "; PrintNormal (cout, "epsMinus",
                               simutcConfigData.networkData.EpsilonMinus);
        cout << "       "; PrintNormal (cout, "epsPlus",
                               simutcConfigData.networkData.EpsilonPlus);
    } // if

    cout << endl << "  Nodes: " << numberOfNodes << endl;
    cout << "    "; PrintNormal (cout, "round period",
                        simutcConfigData.commonNodeData.RoundPeriod);

    cout << "    delay measurement every ";
    cout << simutcConfigData.commonNodeData.DelayCycles << " cycles" << endl;

    cout << "    rate synchronization every ";
    cout << simutcConfigData.commonNodeData.RateCycles << " cycles" << endl;

    cout << "    synchronization algorithm is ";
    switch (simutcConfigData.commonNodeData.SynchronizationAlgorithm)
    {
        case FreeRun: cout << "free-run";
            break;
        case Midpoint: cout << "FTA (using midpoint)";
            break;
        case FaultTolerantAverage: cout << "FTA (using reference point)";
            break;
        case Fetzer: cout << "FTA (reference point) with Fetzer for P-Nodes";
            break;
        case OP: cout << "OP (Optimal Precision)";
            break;
        case Marzullo: cout << "Marzullo";
            break;
        case Rate: cout << "Rate Synchronization";
            break;
        case FaultTolerantIntersection: cout << "FTI (Fault Tolerant"
                                                " Intersection)";
            break;
        default:
            cout << simutcConfigData.commonNodeData.SynchronizationAlgorithm;
            cout << " (unknown)";
    } // switch
    cout << endl;

    cout << endl;
    cout << "    node specific data:" << endl;
    for (int i=0; i<numberOfNodes; i++)
    {
        // if the node data is different or if there is only one node,
        // we print the text "node 1"; if the node data is the same for
        // all nodes and there are at least two, we print "nodes 1-n"
        // where 'n' is the number of nodes
        if (simutcConfigData.nodesDiffer || (numberOfNodes==1))
            cout << "    node " << i+1 << ": " << endl;
        else cout << "    nodes 1-" << numberOfNodes << ": " << endl;

        // now print the node data
        cout << "      "; PrintNormal (cout, "clock drift",
                            simutcConfigData.nodeData[i].ClockDrift);

        cout << "      node type = ";
        if (simutcConfigData.nodeData[i].ConnectedGpus) cout << "P-Node";
        else cout << "S-Node";
        cout << endl;

        cout << "      "; PrintNormal (cout, "delay compensation",
                            simutcConfigData.nodeData[i].DelayCompensation);
        cout << "      "; PrintNormal (cout, "computation delay compensation",
                            simutcConfigData.nodeData[i].CompDelayCompensation);
        cout << "      "; PrintProbability (cout, "node crashes",
                            simutcConfigData.nodeData[i].ProbLanceCrash);
        cout << "      "; PrintProbability (cout, "node recovery",
                            simutcConfigData.nodeData[i].ProbLanceRecover);
        cout << "      ";PrintNormal(cout, "maximum delay of outgoing messages",
                            simutcConfigData.nodeData[i].DmaxSendLance);
        cout << "      ";PrintNormal(cout, "maximum delay of incoming messages",
                            simutcConfigData.nodeData[i].DmaxRecvLance);
        cout << "      "; PrintProbability (cout, "scrambling send messages",
                            simutcConfigData.nodeData[i].ProbLanceSendScramble);
        cout << "      "; PrintProbability (cout, "scrambling recv messages",
                            simutcConfigData.nodeData[i].ProbLanceRecvScramble);
        cout << "      "; PrintProbability (cout, "duplicating send messages",
                           simutcConfigData.nodeData[i].ProbLanceSendDuplicate);
        cout << "      "; PrintProbability (cout, "duplicating recv messages",
                           simutcConfigData.nodeData[i].ProbLanceRecvDuplicate);


        if (!simutcConfigData.nodesDiffer) break;
        // jump out of the loop if the other nodes have the same data
    } // for

    cout << endl;
}


//--------------------------------------------------------------------------
// printing the measured precision
//--------------------------------------------------------------------------

// prints the precision data collected during evaluation
void PrintPrecision ()
{
    double norm;   // "normalized" number for output
    char unit[8];  // unit name

    cout << endl << "precision data";

    if (piRoundNumber > FirstSnapshotPrecisionRound)
    // the round number always includes the current (possibly unfinished)
    // round; this round is not included in the summary, so there must be
    // at least FirstSnapshotPrecisionRound+1 rounds for valid precision
    // data
    {
        double average;  // the average value

        average = piSum/(piRoundNumber-FirstSnapshotPrecisionRound);

        cout << " (from round ";
        cout << evalsysConfigData.FirstSyncPrecisionRound << " on):" << endl;
        cout << "    snapshots: " << piRoundNumber << endl;

        Normalize (piMin, norm, unit);
        cout << "    minimum: " << norm << " " << unit << endl;
        Normalize (piMax, norm, unit);
        cout << "    maximum: " << norm << " " << unit << endl;
        Normalize (average, norm, unit);
        cout << "    average: " << norm << " " << unit << endl;
    } // if
    else cout << ": none collected" << endl;
}


//--------------------------------------------------------------------------
// conversion function
//--------------------------------------------------------------------------

// if g_byteOrderReversed is True, the reverted value of 'src' is returned;
// if it is False, 'src' is returned unchanged
static INT32 g_Convert (INT32 src)
{
    if (!g_byteOrderReversed) return src;  // nothing to convert

    INT32 tmp;
    tmp  = (src << 24) & 0xFF000000;
    tmp += (src <<  8) & 0x00FF0000;
    tmp += (src >>  8) & 0x0000FF00;
    tmp += (src >> 24) & 0x000000FF;
    return tmp;
}


#ifdef PLOT_DATA

//--------------------------------------------------------------------------
//--- open files
//--------------------------------------------------------------------------

// used for the output of simulation data
static filebuf Frun;  // LaTeX wrapper file
static ostream outRun;  // stream for Frun
static filebuf Fpik[numberOfNodes-1];  // k-th precision
static filebuf Fstate[numberOfNodes];  // state
static filebuf Fadj[numberOfNodes];  // adjustments
static filebuf Facc[numberOfNodes];  // accuracy
static filebuf Frate[numberOfNodes];  // rates
static filebuf Fgammak[numberOfNodes-1];  // k-th consonance
static filebuf Fdrift[numberOfNodes];  // drifts
static filebuf Fmult[numberOfNodes];


// generates the filename from 'name' followed by 'number'+1, with the
// extension ".dat" in the subdirectory PLOT_DIR, opens the filebuffer
// file[number], associates the buffer with the ostream stream[number]
static void OpenFile (filebuf *file, ostream* stream, int number,
                      const char *name)
{
    static char filename[256];  // for generating the file names
    static char buf[20];  // buffer for itoa()

    strcpy (filename, PLOT_DIR);
    strcat (filename, name);
    strcat (filename, itoa (number+1, buf, 10));
    strcat (filename, ".dat");
    // filenames are generated from the name and the number

    Assert (file[number].open (filename, output) != 0);
    // open the file

    stream[number].rdbuf (&file[number]);
    // associate the filebuffer with its stream
}

// same as OpenFile(), but the filename is generated from 'name' followed
// by 'number' (i.e., starting from 0, as required by k-th order plots)
static void OpenOrderFile (filebuf *file, ostream* stream, int number,
                           const char *name)
{
    static char filename[256];  // for generating the file names
    static char buf[20];  // buffer for itoa()

    strcpy (filename, PLOT_DIR);
    strcat (filename, name);
    strcat (filename, itoa (number, buf, 10));
    strcat (filename, ".dat");
    // filenames are generated from the name and the number

    Assert (file[number].open (filename, output) != 0);
    // open the file

    stream[number].rdbuf (&file[number]);
    // associate the filebuffer with its stream
}

static void GenerateTexFile();
// prints the TeX-Output into run.tex

// opens all file buffers
void OpenPlotFiles()
{
    int i;
    char filename[256];

    strcpy (filename, PLOT_DIR);
    strcat (filename, "run.tex");

    // open and generate the TeX-File
    Assert (Frun.open (filename, output) != 0);
    GenerateTexFile();

    // for a simulation run output precision by each snapshot
    for (i=0; i<numberOfNodes-1; i++)
    {
        OpenOrderFile (Fpik, outPik, i, "pik");  // plots a k-th order
    } // for

    // for a simulation run output state by each snapshot
    for (i=0; i<numberOfNodes; i++)
    {
        OpenFile (Fstate, outState, i, "state");
    } // for

    // for a simulation run output adjustment by each clock update
    for (i=0; i<numberOfNodes; i++)
    {
        OpenFile (Fadj, outAdj, i, "adj");
    } // for

    // for a simulation run output accuracy interval by each snapshot
    for (i=0; i<numberOfNodes; i++)
    {
        OpenFile (Facc, outAcc, i, "acc");
    } // for

    // for a simulation run output current rate by each snapshot
    for (i=0; i<numberOfNodes; i++)
    {
        OpenFile (Frate, outRate, i, "rate");
    } // for

    // for a simulation run output consonance by each snapshot
    for (i=0; i<numberOfNodes-1; i++)
    {
        OpenOrderFile (Fgammak, outGammak, i, "gammak");  // plots a k-th order
    } // for

    // for a simulation run output rate interval by each clock update
    for (i=0; i<numberOfNodes; i++)
    {
        OpenFile (Fdrift, outDrift, i, "drift");
    } // for

    // for a simulation run output step multiplicator by each clock update
    for (i=0; i<numberOfNodes; i++)
    {
        OpenFile (Fmult, outMult, i, "mult");
    } // for
}

// closes all plot files
void ClosePlotFiles()
{
    int i;

    Frun.close();  // close the TeX-File

    // precision
    for (i=0; i<numberOfNodes-1; i++)
    {
        Fpik[i].close();
    } // for

    // state
    for (i=0; i<numberOfNodes; i++)
    {
        Fstate[i].close();
    } // for

    // adjustment
    for (i=0; i<numberOfNodes; i++)
    {
        Fadj[i].close();
    } // for

    // accuracy interval
    for (i=0; i<numberOfNodes; i++)
    {
        Facc[i].close();
    } // for

    // current rate
    for (i=0; i<numberOfNodes; i++)
    {
        Frate[i].close();
    } // for

    // consonance
    for (i=0; i<numberOfNodes-1; i++)
    {
        Fgammak[i].close();
    } // for

    // rate interval
    for (i=0; i<numberOfNodes; i++)
    {
        Fdrift[i].close();
    } // for

    // step multiplicator
    for (i=0; i<numberOfNodes; i++)
    {
        Fmult[i].close();
    }
}

// prints the text for displaying the postscript picture 'picture'
static ostream& TexPrintPicture (ostream& out, const char *picture)
{
    out << "\\begin{figure}[h]\n";
    out << "\\epsfxsize=0.9\\textwidth\n";
    out << "\\hfill\n";
    out << "\\epsfbox{";
    out << picture;
    out << "}\n";
    out << "\\hfill\\hbox{}\n";
    out << "\\end{figure}\n\n";

    return out;
}

// print the time and date, followed by a newline
static ostream& TexPrintTime (ostream& out)
{
    out << "\\medskip\n";
    out << "\\begin{center}\n";
    out << "{\\em \\today}\n";
    out << "\\end{center}\n";
    out << "\\newpage\n\n";

    return out;
}

// prints the TeX-Output into run.tex
static void GenerateTexFile()
{
    double norm;
    char unit[8];

    ostream outRun(&Frun);
    // create the latex file to show the simulation results

    static char *buf_top =
    "
    \\documentclass[10pt,a4paper]{article}
    \\setlength{\\textwidth}{15truecm}
    \\setlength{\\textheight}{23truecm}
    \\setlength{\\oddsidemargin}{0truecm}
    \\setlength{\\evensidemargin}{0truecm}
    \\setlength{\\topmargin}{-0.06truecm}
    \\setlength{\\topskip}{0.0truecm}
    \\setlength{\\headheight}{0.0truecm}
    \\setlength{\\headsep}{0.0truecm}
    \\input{epsf}
    \\renewcommand{\\textfraction}{0.0}
    \\renewcommand{\\topfraction}{1.0}

    \\begin{document}
    ";

    outRun << buf_top << endl;

    outRun << "\\noindent" << endl;
    outRun << "{\\bf " << endl;
    switch (simutcConfigData.commonNodeData.SynchronizationAlgorithm)
    {
        case FreeRun: outRun << "Free run";
            break;
        case Midpoint: outRun << "FTA applied on midpoints";
            break;
        case FaultTolerantAverage: outRun << "FTA applied on reference points";
            break;
        case Fetzer: outRun << "FTA applied on reference points "
                            "with Fetzer for P-Nodes";
            break;
        case OP: outRun << "Optimal Precision";
            break;
        case Marzullo: outRun << "Marzullo";
            break;
        case Rate: outRun << "Rate";
            break;
        case FaultTolerantIntersection: outRun << "FTI";
            break;
        default:
            outRun << simutcConfigData.commonNodeData.SynchronizationAlgorithm;
            outRun << " (unknown)";
    } // switch
    outRun << "}: " << endl;
    outRun << "$n = " << simutcConfigData.commonNodeData.NumberOfNodes << "$, ";
    outRun<<"$P_S = " << simutcConfigData.commonNodeData.RoundPeriod << "$~s, ";
    outRun << "$P_R = " << simutcConfigData.commonNodeData.RoundPeriod *
        simutcConfigData.commonNodeData.RateCycles << "$~s, ";
    outRun<<"$f = "<<simutcConfigData.commonNodeData.NumberOfFaultyNodes<<"$, ";
    Normalize (simutcConfigData.nodeData[0].DelayCompensation, norm, unit);
    outRun << "$\\Delta = " << norm << "~{\\rm " << unit << "}$";
    outRun << " \\newline" << endl;

    Normalize (simutcConfigData.nodeData[0].ClockDrift, norm, unit);
    outRun << "clocks:~$\\rho = (";
    outRun << norm << "~{\\rm " << unit << "/s}";
    for (int i=1; i<simutcConfigData.commonNodeData.NumberOfNodes; i++)
    {
        Normalize (simutcConfigData.nodeData[i].ClockDrift, norm, unit);
        outRun << ", " << norm << "~{\\rm " << unit << "/s}";
    }
    outRun << ")$ \\newline" << endl;

    outRun << "network:~$\\delta = ( ";
    Normalize (simutcConfigData.networkData.Delta, norm, unit);
    outRun <<  norm << "~{\\rm " << unit << "}, ";
    Normalize (simutcConfigData.networkData.EpsilonMinus, norm, unit);
    outRun <<  norm << "~{\\rm " << unit << "}, ";
    Normalize (simutcConfigData.networkData.EpsilonPlus, norm, unit);
    outRun <<  norm << "~{\\rm " << unit << "} )$, ";
    Normalize (simutcConfigData.networkData.Lambda, norm, unit);
    outRun << "$\\lambda = " << norm << "~{\\rm " << unit << "}$";
    outRun << endl;


    TexPrintPicture (outRun, "state.ps");
    TexPrintPicture (outRun, "pi.ps");
    TexPrintTime (outRun);

    TexPrintPicture (outRun, "adj.ps");
    TexPrintPicture (outRun, "acc.ps");
    TexPrintTime (outRun);

    TexPrintPicture (outRun, "rate.ps");
    TexPrintPicture (outRun, "gamma.ps");
    TexPrintTime (outRun);

    TexPrintPicture (outRun, "mult.ps");
    TexPrintPicture (outRun, "drift.ps");
    outRun << "\\medskip" << endl << endl;


    outRun << "\\noindent" << endl;
    outRun << "evaluation:~rounds$_{\\max} = ";
    outRun <<     evalsysConfigData.NumberOfRounds << "$, ";
    outRun << "messages$_{\\max} = ";
    outRun <<     evalsysConfigData.NumberOfMessages << "$, ";
    outRun << "$P_{\\rm snap} = ";
    outRun <<     evalsysConfigData.SnapshotPeriod << "$" << endl;
    outRun << "\\vspace*{0.5in}" << endl;

    static char *buf_tail =
    "
    \\begin{center}
    {\\em \\today}
    \\end{center}

    \\end{document}
    ";

    outRun << buf_tail << endl;
}


//--------------------------------------------------------------------------
//--- sort an array of doubles, compute precision
//--------------------------------------------------------------------------

static void SortArray (double* unsortedArray, double* sortedArray, INT32 size)
{
    Assert (unsortedArray != NULL);
    Assert (sortedArray != NULL);
    Assert (size > 0);

    int i;
    int free;
    int up, low, curr;
    double item;

    free = 0;
    low = up = curr = 0;

    for (i=0;i<size;i++)
    {
        item = unsortedArray[i];

        if (free != 0)  // find the entry
        {
            up = (free>0) ? free-1 : 0;  // prepare for binary search
            low = 0;
            curr = (up+low)/2;

            while ((item != sortedArray[curr]) && (low < up))
            {
                if (item < sortedArray[curr]) up = (curr>0) ? curr-1 : 0;
                else if (item > sortedArray[curr]) low = curr+1;

                curr = (low+up)/2;
            }

            if (item > sortedArray[curr]) curr++;
            // didn't find the entry, need to insert behind the current one
        } // if

        Assert (curr < (SINT32)(size));
        Assert (curr >= 0);

        memmove (sortedArray+curr+1, sortedArray+curr,  // move everything
                 (size-curr-1)*sizeof(double));

        sortedArray[curr] = item;  // enter the timestamp

        free++;  // increment the free index
    } // for
}


static double PrecisionKth (double* pi, INT32 upper, INT32 order)
{
    INT32 lower;
    int i;

    if (upper <= 1 + order) return 0.0;
    // not enough elements

    lower = 0;
    upper--;
    // now we have pi[lower],...,pi[upper]


    for (i=0; i<order; i++)
    {
        // in each round eliminate one peripheral element
        // which one? lower or upper one that has the larger
        // distance to the next inner neigbor

        if ( (pi[lower+1] - pi[lower]) >= (pi[upper] - pi[upper-1]) )
        {
            lower++; // take away the lower
        }
        else
        {
            upper--; // take away the upper
        }

    }

    Assert( lower < upper );
    return (pi[upper] - pi[lower]);
    // compute the distance
}
#endif


