//
// evalsys.cpp
//
// main file of the dummy evaluation; contains main(), which initializes
// the simulation, starts one run, prints the events, computes the precision
// from the snapshot events, then kills the simulation, prints precision
// statistics and the simulation configuration used
//
// to change the configuration, edit evalcfg.hpp and recompile
//
// created: 01.02.97
//
// last change: 05.02.99
//
// RCS version control information:
//     $Revision: 1.20 $$State: Exp $$Date: 1999/02/05 15:30:26 $
//


#include <sys/socket.h>
#include <sys/types.h>
#include <errno.h>
#include <iostream.h>
#include <stdlib.h>
#include <stream.h>
#include <unistd.h>

#include "bool.hpp"
#include "constant.hpp"
#include "errbase.hpp"
#include "error.hpp"
#include "general.hpp"
#include "sockinet.hpp"
#include "sockunix.hpp"
#include "sockstre.hpp"
#include "structs.hpp"
#include "time.hpp"

#include "evalcfg.hpp"   // configuration (all necessary globals)
#include "evalconv.hpp"  // Convert() functions for all events
#include "evalfunc.hpp"  // prototypes of useful functions
#include "evalglob.hpp"  // structure for receiving replies

#ifndef DOS_SIM
  #include <sys/time.h>
#endif


//**************************************************************************
// necessary globals (are used in evalfunc.cpp)
//**************************************************************************

const INT32 ControllerIndex = numberOfNodes;
// the controller has the highest ID in all arrays; the supervisor IDs
// are the same as their nodeIDs and start with 0 and go up to
// NumberOfNodes-1

CStreamServer *serverSocket = NULL;
// server socket; it is deleted by evalfunc.cpp before an Assert() fails

INT32 supsid[ControllerIndex+1];
// supervisor socket IDs (last: controller)
// needed for SelectSocket()

INT32 round = 0;
// current round (set by PrintEvent())

EBool g_byteOrderReversed = False;
// byte order of communication peer is different from own (for g_Convert())


//**************************************************************************
// local functions and their globals
//**************************************************************************

#define SOCKNAME "/tmp/SimUTC"
// name of the UNIX socket


// all possible return values of the program
enum EErrorNumbers {
    ErrNoError = 0,          // everything was okay
    ErrAcceptHasFailed,      // accept() has returned an error
    ErrNoNodesAvailable,     // numberOfNodes is zero
    ErrSocketExists          // the UNIX socket already exists
};


static SINT32 networkID[numberOfNetworks];
// IDs of the networks
// set by ConfigureNetworks(), read by ConfigureNodes()

static CStreamInterface *ctrl;  // socket for the controller
static CStreamInterface *supsock[numberOfNodes];
// supervisor sockets

static CUnixAddress addr (SOCKNAME);  // socket address
//static CInternetAddress addr (EvalsysPort);


static void ConfigureNetworks();
// create and configure numberOfNetworks networks

static int ConfigureNodes();
// create and configure numberOfNodes nodes


//**************************************************************************
// main task: create the socket, wait for the connection from the Controller,
//      create a network and a few applications, set their parameters, and
//      start the whole thing; then wait for the event reports (clock
//      updates), print them to cout, and finally kill the simulation and
//      terminate the program
//**************************************************************************

int main ()
{
    SSingleParameter param;  // for setting parameters
    SEventReport report;  // for setting event report parameters
    INT8 mask[evalsysConfigData.MaskSize];  // mask for the event reports
    EModule module;  // the module type of the communication partner
    SApplCommand applc;

    int i, j, k;


    //----------------------------------------------------------------------
    // check if the server socket already exists

    // to avoid program crashes if the socket already exists (in this
    // case, creating the socket causes an assert to fail), we check for it

    if (SocketExists (&addr, SStream))  // check if the socket exists
    {
        cerr << "error: the server socket could not be created" << endl;
        cerr << "       Check if another instance of evalsys is currently ";
        cerr << "being executed." << endl;
        cerr << "       If a previous execution of evalsys has been ";
        cerr << "interrupted and if you use" << endl;
        cerr << "       UNIX domain sockets, then the socket file may still ";
        cerr << "exist. In this " << endl;
        cerr << "       case, remove the file "<< SOCKNAME <<" and try again.";
        cerr << endl;
        cerr << endl;
        return ErrSocketExists;
    } // if


    //----------------------------------------------------------------------
    // install the error handler before creating objects that report errors

    CErrorHandler errorHandler;  // for error outputs
    InstallErrorHandler (&errorHandler);  // install the handler

    cout.setf (ios::fixed, ios::floatfield);
    // always print 6 (default) digits

    cerr << endl;
    // always start output on a new line


    //----------------------------------------------------------------------
    // create the server socket

    serverSocket = new CStreamServer (&addr);  // create main socket


    //----------------------------------------------------------------------
    // generate the gnuplot files if desired

#ifdef PLOT_DATA
    OpenPlotFiles();
#endif


    //----------------------------------------------------------------------
    // wait for the controller

    if ((ctrl = serverSocket->Accept()) == NULL)  // controller can connect?
    {
        cerr << "accept() has failed" << endl;
        cerr << "errno = " << errno << endl;
        delete serverSocket;  // remove the server socket
        return ErrAcceptHasFailed;
    }
    ctrl->SetTimeout (20);  // set socket timeout to 20 seconds

    g_byteOrderReversed = ctrl->ByteOrderReversed();
    // the byte order on the target may be different from the own order

    if (g_byteOrderReversed) Report (client uses different byte order);
    else Report (client uses same byte order);


    supsid[ControllerIndex] = ctrl->GetSocketID();  // store the socket ID

    if (!ctrl->ReceiveModule (module, 20)) Report (no module message);
    // get the module type (wait for 20 seconds)
    else if (module != Controller) Report (unexpected module);

    EMessageType type;  // for receiving the message header
    if (ctrl->ReceiveMessageHeader (&type, 10))  // get the enum-size msg
    {
        if (type != MsgEnumSizes) Report (no enumeration size message);
        SEnumSizes *msg;
        if (!ctrl->ReceiveMessage ((SBase**)&msg))
            Report (could not receive enumMsg);
        else ctrl->ReturnMessageBuffer (msg);  // not needed in dummy-ES
    }
    else Report (no enumSize message);


    //----------------------------------------------------------------------
    // configure the controller

    // set event and class mask of the controller
    report.module = Controller;
    memcpy (mask, &simutcConfigData.controllerData.Events,
            evalsysConfigData.MaskSize);
    report.Set (EventReport, mask, evalsysConfigData.MaskSize);
    Convert(&report, g_byteOrderReversed);
    SetParameter (ctrl, &report);
    CheckBooleanReply (ctrl, "set event mask");

    report.module = Controller;
    memcpy (mask, &simutcConfigData.controllerData.EventClasses,
            evalsysConfigData.MaskSize);
    report.Set (EventClassReport, mask, evalsysConfigData.MaskSize);
    Convert(&report, g_byteOrderReversed);
    SetParameter (ctrl, &report);
    CheckBooleanReply (ctrl, "set event class mask");


    //----------------------------------------------------------------------
    // create and configure the networks

    if (simutcConfigData.networkData.NumberOfNetworks > 0) ConfigureNetworks();
    // only configure them if they are simulated (in hardware-based
    // simulation, the network is not simulated)


    //----------------------------------------------------------------------
    // create and configure the nodes

    int error;  // for the result of the node configuration

    if (simutcConfigData.commonNodeData.NumberOfNodes > 0)
    // only configure nodes if any exist
    {
        error = ConfigureNodes();
        if (error != 0)
        {
            delete serverSocket;  // remove the server socket
            return error;
        }
    } // if
    else  // we cannot work without any nodes
    {
        Report (no nodes available);
        delete serverSocket;  // remove the server socket
        return ErrNoNodesAvailable;
    }


    //----------------------------------------------------------------------
    // one supervisor triggers periodic snapshots

    // give that supervisor a different (snapshot) period
    param.module = Supervisor;
    param.type = InterfacePeriod;
    param.value = evalsysConfigData.SnapshotPeriod;
    Convert(&param, g_byteOrderReversed);
    SetParameter (supsock[evalsysConfigData.SnapshotNodeID-1], &param);
    CheckBooleanReply (supsock[evalsysConfigData.SnapshotNodeID-1],
                       "set interface period");

    SCommand command;
    command.module = Supervisor;
    command.type = EnableSnapshot;
    Convert(&command, g_byteOrderReversed);
    ExecuteCommand (supsock[evalsysConfigData.SnapshotNodeID-1], &command);
    CheckBooleanReply (supsock[evalsysConfigData.SnapshotNodeID-1],
                       "enable snapshot");


    //----------------------------------------------------------------------
    // resume the supervisors --> start synchronization

    command.module = Controller;
    command.type = ResumeSimulation;
    Convert(&command, g_byteOrderReversed);
    ExecuteCommand (ctrl, &command);
    CheckBooleanReply (ctrl, "resume simulation");


    //----------------------------------------------------------------------
    // now poll the sockets, print out the data

    // wait for 'NumberOfRounds' reports per socket, then stop everything

    int msgcounter = 0;
    INT32 index;  // socket index of the next readable socket

    while ((round < evalsysConfigData.NumberOfRounds) &&
           (msgcounter < evalsysConfigData.NumberOfMessages))
    // note: I have two bounds on this loop: either the number of rounds
    // is exceeded (so one node has updated his clock for a certain number
    // of times), or the maximum number of messages is exceeded (a timeout
    // counts as a "message" as well)
    // it is vital that evalsys stops at some point and kills the
    // simulation, so there must be some condition to *ensure* that
    // this loop is exited
    {
        index = SelectSocket();  // check the sockets

        if (index == NoAccount) Report (evalsys: sockets were empty);
        else if (index==ControllerIndex) ReceiveEvent (ctrl);
        else ReceiveEvent (supsock[index]);

        msgcounter++;  // even timeouts (or read failures) are counted
    } // while


    //----------------------------------------------------------------------
    // stop the simulation (irreversibly)

    command.module = Controller;
    command.type = KillSimulation;
    Convert(&command, g_byteOrderReversed);
    ExecuteCommand (ctrl, &command);
    cout << endl << "--- simulation is being shut down ---" << endl;
    CheckBooleanReply (ctrl, "delete simulation");


    //----------------------------------------------------------------------
    // delete the supervisor and controller sockets

    for (i=0; i < simutcConfigData.commonNodeData.NumberOfNodes; i++)
        delete supsock[i];
    delete ctrl;

    delete serverSocket;  // delete the server socket


    //----------------------------------------------------------------------
    // that's all, folks

    PrintConfiguration ();
    PrintPrecision ();

#ifdef PLOT_DATA
    ClosePlotFiles();
#endif

    return ErrNoError;
}


//**************************************************************************
// local functions
//**************************************************************************

// create numberOfNetworks networks, configure them
static void ConfigureNetworks()
{
    int i, j, k;
    SDoubleReply *doubleReply;  // that will contain the network ID
    SNetworkCreateCommand netc;
    SNetworkConnectCommand netconn;
    SNetworkData netdata;
    SSingleNetworkParameter sparam;
    SNetDistribution netdistp;  // for lambda, omega
    SNetDistributionP2P netdist;  // for transmission delays


    for (k=0; k<simutcConfigData.networkData.NumberOfNetworks; k++)
    {
        netc.module = Controller;
        netc.type = CreateNetwork;
        netc.networkType = Broadcast;
        Convert(&netc, g_byteOrderReversed);
        ExecuteCommand (ctrl, &netc);
        if (!ReceiveReply (ctrl, (SReply **)&doubleReply))
        {
            Report (evalsys: error receiving the reply);
        }
        else
        {
            if (doubleReply->type!=ReplyDouble) Report (evalsys: wrong reply!);
            else
            {
                networkID[k] = (SINT32)doubleReply->value;
                Check (networkID[k] == k+1); // networkID should be equal to k+1
            }
            delete doubleReply;
        }
        CheckBooleanReply (ctrl, "create network");


        // first add the nodes to the network
        for (i=0; i<simutcConfigData.commonNodeData.NumberOfNodes; i++)
        {
            netconn.module = Network;
            netconn.type = ConnectNode;
            netconn.networkID = networkID[k];
            netconn.nodeID = i+1;  // node IDs start with 1
            Convert(&netconn, g_byteOrderReversed);
            ExecuteCommand (ctrl, &netconn);
            CheckBooleanReply (ctrl, "connect node");
        }

        // set the transmission delays and transmission error probabilities
        for (i=1; i <= simutcConfigData.commonNodeData.NumberOfNodes; i++)
        for (j=1; j <= simutcConfigData.commonNodeData.NumberOfNodes; j++)
        {
            if (i != j)
            {
                // set late transmissions (dmax)
                netdist.module = Network;
                netdist.type = DistributionLate;
                netdist.networkID = networkID[k];
                netdist.sender = i;
                netdist.receiver = j;
                netdist.data.dist = Uniform;
                netdist.data.constant = 0;
/*
                netdist.data.low = simutcConfigData.networkData.Delta +
                                   simutcConfigData.networkData.EpsilonPlus;
*/
                netdist.data.low = 0;  // DistributionTimely will set it
                netdist.data.high = simutcConfigData.networkData.Dmax;
                netdist.data.mean = (netdist.data.low+netdist.data.high)/2;
                netdist.data.freeparam = 0;
                Convert(&netdist, g_byteOrderReversed);
                SetParameter (ctrl, &netdist);
                CheckBooleanReply (ctrl, "set late transmission");

                // set timely transmission
                netdist.module = Network;
                netdist.type = DistributionTimely;
                netdist.networkID = networkID[k];
                netdist.sender = i;
                netdist.receiver = j;
                netdist.data.dist = Normal; //Uniform;
                netdist.data.constant = 0;
                netdist.data.low = simutcConfigData.networkData.Delta -
                                   simutcConfigData.networkData.EpsilonMinus;
                netdist.data.high = simutcConfigData.networkData.Delta +
                                    simutcConfigData.networkData.EpsilonPlus;
                netdist.data.mean = simutcConfigData.networkData.Delta;
                netdist.data.freeparam =
                    (simutcConfigData.networkData.EpsilonMinus +
                     simutcConfigData.networkData.EpsilonPlus)/4; //0;
                Convert(&netdist, g_byteOrderReversed);
                SetParameter (ctrl, &netdist);
                CheckBooleanReply (ctrl, "set timely transmission");

                // set early transmissions (dmin)
                netdist.module = Network;
                netdist.type = DistributionEarly;
                netdist.networkID = networkID[k];
                netdist.sender = i;
                netdist.receiver = j;
                netdist.data.dist = Uniform;
                netdist.data.constant = 0;
                netdist.data.low = simutcConfigData.networkData.Dmin;
                netdist.data.high = simutcConfigData.networkData.Delta -
                                    simutcConfigData.networkData.EpsilonMinus;
                netdist.data.mean = (netdist.data.low+netdist.data.high)/2;
                netdist.data.freeparam = 0;
                Convert(&netdist, g_byteOrderReversed);
                SetParameter (ctrl, &netdist);
                CheckBooleanReply (ctrl, "set early transmission");


                // set the probabilities for early/late messages
                netdata.module = Network;
                netdata.networkID = networkID[k];
                netdata.type = PEarly;
                netdata.value = simutcConfigData.networkData.ProbNetEarly;
                netdata.sender = i;
                netdata.receiver = j;
                Convert(&netdata, g_byteOrderReversed);
                SetParameter (ctrl, &netdata);
                CheckBooleanReply (ctrl, "set probEarly");

                netdata.module = Network;
                netdata.networkID = networkID[k];
                netdata.type = PLate;
                netdata.value = simutcConfigData.networkData.ProbNetLate;
                netdata.sender = i;
                netdata.receiver = j;
                Convert(&netdata, g_byteOrderReversed);
                SetParameter (ctrl, &netdata);
                CheckBooleanReply (ctrl, "set probLate");
            } // if
        } // for

        // set lambda
        netdistp.module = Network;
        netdistp.type = DistributionLambda;
        netdistp.networkID = networkID[k];
        netdistp.data.dist = Uniform;
        netdistp.data.constant = 0;
        netdistp.data.low = 0;
        netdistp.data.high = simutcConfigData.networkData.Lambda;
        netdistp.data.mean = (netdistp.data.low+netdistp.data.high)/2;
        netdistp.data.freeparam = 0;
        Convert(&netdistp, g_byteOrderReversed);
        SetParameter (ctrl, &netdistp);
        CheckBooleanReply (ctrl, "set Lambda");

        // set omega
        netdistp.module = Network;
        netdistp.type = DistributionOmega;
        netdistp.networkID = networkID[k];
        netdistp.data.dist = Uniform;
        netdistp.data.constant = 0;
        netdistp.data.low = 0;
        netdistp.data.high = simutcConfigData.networkData.Omega;
        netdistp.data.mean = (netdistp.data.low+netdistp.data.high)/2;
        netdistp.data.freeparam = 0;
        Convert(&netdistp, g_byteOrderReversed);
        SetParameter (ctrl, &netdistp);
        CheckBooleanReply (ctrl, "set Omega");


        // set the probabilities for crash/recovery
        sparam.module = Network;
        sparam.networkID = networkID[k];
        sparam.type = PCrash;
        sparam.value = simutcConfigData.networkData.ProbNetCrash;
        Convert(&sparam, g_byteOrderReversed);
        SetParameter (ctrl, &sparam);
        CheckBooleanReply (ctrl, "set network crash probability");

        sparam.module = Network;
        sparam.networkID = networkID[k];
        sparam.type = PRecover;
        sparam.value = simutcConfigData.networkData.ProbNetRecover;
        Convert(&sparam, g_byteOrderReversed);
        SetParameter (ctrl, &sparam);
        CheckBooleanReply (ctrl, "set network recover probability");
    } // for (iteration over networks)
}


// create and configure numberOfNodes nodes
static int ConfigureNodes()
{
    int i;
    SSingleParameter param;  // for setting parameters
    SEventReport report;  // for setting event report parameters
    INT8 mask[evalsysConfigData.MaskSize];  // mask for the event reports
    EModule module;  // the module type of the communication partner
    SApplCommand applc;
    SGpsConnectCommand gpscmd;
    SNetworkCommand netcommand;
    SApplDistribution distp;
    STimeParameter timep;
    SApplWobbleDrift wobblep;

    for (i=0; i < numberOfNodes; i++)
    {
        applc.module = Controller;
        applc.type = CreateApplication;
        applc.area = evalsysConfigData.NodeArea;  // same for all nodes
        applc.nodeID = i+1;
        Convert(&applc, g_byteOrderReversed);
        ExecuteCommand (ctrl, &applc);

        // wait for the supervisor to connect
        if ((supsock[i] = serverSocket->Accept()) == NULL)
        {
            cerr << "accept() has failed" << endl;
            return ErrAcceptHasFailed;
        }

        if (!supsock[i]->ReceiveModule (module, 20)) Report (no module message);
        // get the module type (wait for 20 seconds)
        else if (module != Supervisor) Report (unexpected module);

        supsid[i] = supsock[i]->GetSocketID();  // store the socket ID
        supsock[i]->SetTimeout (20);
        // set the timeout for message reception to 20 seconds (quite long)

        CheckBooleanReply (ctrl, "create application");
    } // for


    //----------------------------------------------------------------------
    // set the parameters

    // note: I set all params of one supervisor before doing the
    // same with the next, because this is much faster (allows the
    // supervisor to process all params in the same round)
    // for the same reason, I collect all replies at the end of the
    // for-loop

    int gpus;  // number of GPUs connected; type must be signed!

    for (i=0; i<numberOfNodes; i++)
    {
        //--- send commands to Clock

        gpus = 0;
        // if it is a P-node, enable GPS on the given GPUs
        if (simutcConfigData.nodeData[i].ConnectedGpus != 0)
        {
            if (simutcConfigData.nodeData[i].ConnectedGpus & (1<<GPU1))
            {
                gpscmd.module = Clock;
                gpscmd.type = ConnectGps;
                gpscmd.gpu = GPU1;
                Convert(&gpscmd, g_byteOrderReversed);
                ExecuteCommand (supsock[i], &gpscmd);
                gpus++;
            }
            if (simutcConfigData.nodeData[i].ConnectedGpus & (1<<GPU2))
            {
                gpscmd.module = Clock;
                gpscmd.type = ConnectGps;
                gpscmd.gpu = GPU2;
                Convert(&gpscmd, g_byteOrderReversed);
                ExecuteCommand (supsock[i], &gpscmd);
                gpus++;
            }
            if (simutcConfigData.nodeData[i].ConnectedGpus & (1<<GPU3))
            {
                gpscmd.module = Clock;
                gpscmd.type = ConnectGps;
                gpscmd.gpu = GPU3;
                Convert(&gpscmd, g_byteOrderReversed);
                ExecuteCommand (supsock[i], &gpscmd);
                gpus++;
            }
        } // if P-Node


        //--- send commands to ClockSync

        // add the networkID as associated network (first network)
        netcommand.module = ClockSync;
        netcommand.type = ConnectNetwork;
        if (numberOfNetworks > 0) netcommand.networkID = networkID[0];
        else netcommand.networkID =
                 simutcConfigData.networkData.DefaultCINetwork;
        Convert(&netcommand, g_byteOrderReversed);
        ExecuteCommand (supsock[i], &netcommand);


        //--- set the params for ClockSync

        param.module = ClockSync;
        param.type = DelayMeasurementCycles;
        param.value = simutcConfigData.commonNodeData.DelayCycles;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = ClockSync;
        param.type = RateSynchronizationCycles;
        param.value = simutcConfigData.commonNodeData.RateCycles;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = ClockSync;
        param.type = NumberOfNodes;
        param.value = numberOfNodes;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = ClockSync;
        param.type = NumberOfFaultyNodes;
        param.value = simutcConfigData.commonNodeData.NumberOfFaultyNodes;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = ClockSync;
        param.type = RoundPeriod;
        param.value = simutcConfigData.commonNodeData.RoundPeriod;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = ClockSync;
        param.type = DelayCompensation;
        param.value = simutcConfigData.nodeData[i].DelayCompensation;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = ClockSync;
        param.type = ComputationDelayCompensation;
        param.value = simutcConfigData.nodeData[i].CompDelayCompensation;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);


        //--- set the params for Clock

        timep.module = Clock;
        timep.type = Accuracy;
        timep.macrostamp = simutcConfigData.nodeData[i].Macrostamp;
        timep.timestamp = simutcConfigData.nodeData[i].Timestamp;
        Convert(&timep, g_byteOrderReversed);
        SetParameter (supsock[i], &timep);

        param.module = Clock;
        param.type = Accuracy;
        param.value = simutcConfigData.nodeData[i].Accuracy;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Clock;
        param.type = ClockDrift;
        param.value = simutcConfigData.nodeData[i].ClockDrift;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Clock;
        param.type = AmortizationDuration;
        param.value = simutcConfigData.nodeData[i].AmortizationDuration;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        distp.module = Clock;
        distp.type = DistributionRegular;
        distp.data.dist = Uniform;
        distp.data.low = 0;
        distp.data.high = 0;
        distp.data.mean = (distp.data.low + distp.data.high)/2;
        distp.data.constant = 0;
        distp.data.freeparam = 0;
        Convert(&distp, g_byteOrderReversed);
        SetParameter (supsock[i], &distp);

        distp.module = Clock;
        distp.type = DistributionFast;
        distp.data.dist = Uniform;
        distp.data.low = simutcConfigData.nodeData[i].RhoPlus;
        distp.data.high = WcRhoPlus;
        distp.data.mean = (distp.data.low + distp.data.high)/2;
        distp.data.constant = 0;
        distp.data.freeparam = 0;
        Convert(&distp, g_byteOrderReversed);
        SetParameter (supsock[i], &distp);

        distp.module = Clock;
        distp.type = DistributionSlow;
        distp.data.dist = Uniform;
        distp.data.low = simutcConfigData.nodeData[i].RhoMinus;
        distp.data.high = WcRhoMinus;
        distp.data.mean = (distp.data.low + distp.data.high)/2;
        distp.data.constant = 0;
        distp.data.freeparam = 0;
        Convert(&distp, g_byteOrderReversed);
        SetParameter (supsock[i], &distp);

        wobblep.module = Clock;
        wobblep.type = WobbleDrift;
        wobblep.period = simutcConfigData.nodeData[i].WobblePeriod;
        wobblep.rhoMin = simutcConfigData.nodeData[i].WobbleRhoMinus;
        wobblep.rhoMax = simutcConfigData.nodeData[i].WobbleRhoPlus;
        wobblep.data.dist = Uniform;
        wobblep.data.constant = 0;
        wobblep.data.low = simutcConfigData.nodeData[i].WobbleAddRangeMin;
        wobblep.data.high = simutcConfigData.nodeData[i].WobbleAddRangeMax;
        wobblep.data.mean = (wobblep.data.low+wobblep.data.high)/2;
        wobblep.data.freeparam = 0;
        Convert(&wobblep, g_byteOrderReversed);
        SetParameter (supsock[i], &wobblep);

        param.module = Clock;
        param.type = PCrash;
        param.value = simutcConfigData.nodeData[i].ProbClockCrash;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Clock;
        param.type = PRecover;
        param.value = simutcConfigData.nodeData[i].ProbClockRecover;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Clock;
        param.type = PDriftChange;
        param.value = simutcConfigData.nodeData[i].ProbDriftChange;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Clock;
        param.type = PFast;
        param.value = simutcConfigData.nodeData[i].ProbFastDrift;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Clock;
        param.type = PSlow;
        param.value = simutcConfigData.nodeData[i].ProbSlowDrift;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Clock;
        param.type = PReadFault;
        param.value = simutcConfigData.nodeData[i].ProbReadFault;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Clock;
        param.type = PWriteFault;
        param.value = simutcConfigData.nodeData[i].ProbWriteFault;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);


        //--- set the params for ClockSync

        param.module = ClockSync;
        param.type = SynchronizationAlgorithm;
        param.value = simutcConfigData.commonNodeData.SynchronizationAlgorithm;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);


        //--- set the params for Supervisor

        report.module = Supervisor;
        memcpy (mask, &simutcConfigData.nodeData[i].EventClasses,
                evalsysConfigData.MaskSize);
        report.Set (EventClassReport, mask, evalsysConfigData.MaskSize);
        Convert(&report, g_byteOrderReversed);
        SetParameter (supsock[i], &report);

        report.module = Supervisor;
        memcpy (mask, &simutcConfigData.nodeData[i].Events,
                evalsysConfigData.MaskSize);
        report.Set (EventReport, mask, evalsysConfigData.MaskSize);
        Convert(&report, g_byteOrderReversed);
        SetParameter (supsock[i], &report);

        param.module = Supervisor;
        param.type = InterfacePeriod;
        param.value = simutcConfigData.nodeData[i].SupervisorPeriod;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);


        //--- set the params for Lance

        param.module = Lance;
        param.type = PCrash;
        param.value = simutcConfigData.nodeData[i].ProbLanceCrash;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Lance;
        param.type = PRecover;
        param.value = simutcConfigData.nodeData[i].ProbLanceRecover;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        distp.module = Lance;
        distp.type = DistributionSend;
        distp.data.dist = Uniform;
        distp.data.low = 0;
        distp.data.high = simutcConfigData.nodeData[i].DmaxSendLance;
        distp.data.mean = (distp.data.low + distp.data.high)/2;
        distp.data.constant = 0;
        distp.data.freeparam = 0;
        Convert(&distp, g_byteOrderReversed);
        SetParameter (supsock[i], &distp);

        distp.module = Lance;
        distp.type = DistributionRecv;
        distp.data.dist = Uniform;
        distp.data.low = 0;
        distp.data.high = simutcConfigData.nodeData[i].DmaxRecvLance;
        distp.data.mean = (distp.data.low + distp.data.high)/2;
        distp.data.constant = 0;
        distp.data.freeparam = 0;
        Convert(&distp, g_byteOrderReversed);
        SetParameter (supsock[i], &distp);

        param.module = Lance;
        param.type = PScrambleSend;
        param.value = simutcConfigData.nodeData[i].ProbLanceSendScramble;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Lance;
        param.type = PScrambleRecv;
        param.value = simutcConfigData.nodeData[i].ProbLanceRecvScramble;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Lance;
        param.type = PDuplicateSend;
        param.value = simutcConfigData.nodeData[i].ProbLanceSendDuplicate;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);

        param.module = Lance;
        param.type = PDuplicateRecv;
        param.value = simutcConfigData.nodeData[i].ProbLanceRecvDuplicate;
        Convert(&param, g_byteOrderReversed);
        SetParameter (supsock[i], &param);


        // now collect the replies (take care not to mess up the order!)
        while (gpus > 0)
        {
            CheckBooleanReply (supsock[i], "connect GPS");
            gpus--;
        }
        CheckBooleanReply (supsock[i], "connect network");

        CheckBooleanReply (supsock[i], "set rate synchronization cycle");
        CheckBooleanReply (supsock[i], "set delay measurement cycle");
        CheckBooleanReply (supsock[i], "set number of nodes");
        CheckBooleanReply (supsock[i], "set number of faulty nodes");
        CheckBooleanReply (supsock[i], "set round period");
        CheckBooleanReply (supsock[i], "set delay compensation");
        CheckBooleanReply (supsock[i], "set computation delay compensation");

        CheckBooleanReply (supsock[i], "set clock value");
        CheckBooleanReply (supsock[i], "set clock accuracy");
        CheckBooleanReply (supsock[i], "set clock drift");
        CheckBooleanReply (supsock[i], "set clock amortization duration");
        CheckBooleanReply (supsock[i], "set distribution of regular drift");
        CheckBooleanReply (supsock[i], "set distribution of fast drift");
        CheckBooleanReply (supsock[i], "set distribution of slow drift");
        CheckBooleanReply (supsock[i], "set drift wobbling");
        CheckBooleanReply (supsock[i], "set clock crash probability");
        CheckBooleanReply (supsock[i], "set clock recover probability");
        CheckBooleanReply (supsock[i], "set clock drift change probability");
        CheckBooleanReply (supsock[i], "set clock fast drift probability");
        CheckBooleanReply (supsock[i], "set clock slow drift probability");
        CheckBooleanReply (supsock[i], "set clock read fault probability");
        CheckBooleanReply (supsock[i], "set clock write fault probability");

        // must be last, after settings of clock
        CheckBooleanReply (supsock[i], "set algorithm type");

        CheckBooleanReply (supsock[i], "set event class report");
        CheckBooleanReply (supsock[i], "set event report");
        CheckBooleanReply (supsock[i], "set supervisor period");

        CheckBooleanReply (supsock[i], "set node crash probability");
        CheckBooleanReply (supsock[i], "set node recover probability");
        CheckBooleanReply (supsock[i], "set delay of outgoing messages");
        CheckBooleanReply (supsock[i], "set delay of incoming messages");
        CheckBooleanReply(supsock[i], "set probability of scrambling send msg");
        CheckBooleanReply(supsock[i], "set probability of scrambling recv msg");
        CheckBooleanReply(supsock[i],"set probability of duplicating send msg");
        CheckBooleanReply(supsock[i],"set probability of duplicating recv msg");
    } // for

    return ErrNoError;
}

