//
// cont_hw.cpp
//
// implementation of the synchronization controller (hardware)
//
// this version is an example
// *) ApplTask(): task self-deletion is done in two steps
// *) ApplTask(): sets socket ID to 0, deletes the supervisor
// *) StartApplication(): stores task ID in an array; accepts up to
//	MAXASICS areas
// *) StopApplication(): removes the supervisor from the list
// *) StopApplication(): removes the task ID from the array
// *) taskname[] is an array of task names (at most MAXASIC tasks)
//
// note: this version does not install the master clock (because we only
// have one UTCSU and want to test the CSA);
// note: if you define the macro INSTALL_MASTER, the master clock is
// installed on area 4
//
// *) StopExecution() sends an event to the ROOT task
// *) there are no private functions concerning the task list
// *) there is no member listOfTasks
// *) StartApplication() gets the supervisor from the listOfSupervisors and
//	starts a task if none exists, else it returns False
// *) StopApplication() sends the EvKillTask event to the task (if it exists)
//
// created: 30.03.97
//
// last change: 20.05.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Exp $$Date: 1998/07/03 13:11:34 $
//


#ifdef INSTALL_MASTER
#include <math.h>
#endif

#include <prepc.h>
#include <pna.h>

#include "ci.hpp"
#include "clock_hw.hpp"  // for g_masterClock
#include "cont_hw.hpp"
#include "event.hpp"
#include "error.hpp"
#include "general.hpp"
#include "globals.hpp"
#include "supervis.hpp"
#include "template.hpp"

#define NoTaskID 0
// the return value of the task list, if an entry was not found

static INT32 tid_app[MAXASICS];
// the IDs of the application tasks

static char taskname[MAXASICS][5] = {"TAP1", "TAP2", "TAP3", "TAP4"};
// the names of the application tasks

extern const INT32 EvKillSim;
// the event is sent to the ROOT task to signal that the controller
// has deleted its own task (simulation is stopped)

extern const INT32 EvKillTask;
// the event is sent to the TAPP task to signal that the application
// task should be deleted


#ifdef INSTALL_MASTER
extern INT32 g_masterArea;
// the area of the master clock, defined in contbase.cpp
// it has to be set to its correct value in the ctor of the hardware
// controller (default is NoAccount)

extern CClock *g_masterClock;
// the master clock itself; use the clock template 'clockTemplate' in
// template.hpp to obtain the clock


static const double TwoPow51 = 4294967296.*524288.;
// needed for computing the value of the STEP and STEPLOW registers
#endif


//**************************************************************************
// class CHardwareController: this class implements all task specific
//      things needed to control the clock synchronization applications
//      (which are tasks)
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CHardwareController::CHardwareController (char qname[4], ACSocketAddress *addr)
    : ACController (qname, addr)
{
    for (int i=0; i<MAXASICS; i++) tid_app[i] = NoTaskID;
    // initialize the task array

#ifdef INSTALL_MASTER
    g_masterArea = 4;
    // set it to some value in [0, NumberOfAreas-1] if you want to use the
    // master clock (and function GetMasterTime())

    double masterFrequency = GetClockFrequency (GetNodeID(), g_masterArea);
    // the frequency of the clock will depend on the oscillator used to
    // feed it (and its peers on the other nodes); all master clocks use
    // the same frequency

    // create the master clock from the clock template
    g_masterClock = clockTemplate->NewClock (masterFrequency,
                                             GetNodeID(), g_masterArea);
    Assert (g_masterClock != NULL);


    // initialize the clock (code taken from CClockInterface and adapted)
    double dummy;
    INT32 steplow = (INT8)(modf (TwoPow51/masterFrequency, &dummy) * 256);
    steplow  = steplow<<24 & 0xFF000000;

    g_masterClock->SetSTEPPURE ((INT32)dummy);  // preload step (pure phase)
    g_masterClock->SetAMORTTIMESET (1);  // short amortization
    g_masterClock->SetSTARTAMORT (1);  // to take over the step pure value
    g_masterClock->SetMSSET (steplow);  // load steplow, set macrostamp to 0
    g_masterClock->SetTSSET (0);
    g_masterClock->SetUSSET (0);
    // the STEPPURE and STEPLOW registers are set according to the frequency,
    // and the NTPTIME registers are set to 0
    g_masterClock->SetNTPSET (0);  // load the clock

    Report (master clock is set up);
#endif
}

CHardwareController::~CHardwareController()
{
    DeleteAllSimulationObjects();
    // note: the function must be called in every derived class, because
    // it calls StopApplication() if there are any supervisors in the
    // listOfSupervisors, and this function would not be available in the
    // dtor of the base class

#ifdef INSTALL_MASTER
    delete g_masterClock;
#endif
}


//--------------------------------------------------------------------------
//--- starting/stopping applications
//--------------------------------------------------------------------------

typedef void (*TaskPrototype)();
// prototype of a task function, as required by t_start()

// create and start the application task
EBool CHardwareController::StartApplication (SINT32 nodeID, int area)
{
    const INT32 Mode_app = T_PREEMPT | T_NOTSLICE | T_SUPV;
    const INT32 SUPST = 1000;
    const INT32 USRST = 1000;
    INT32 args[4] = {nodeID, area, 0, 0};
    INT32 tid;

    void ApplTask (INT32 nodeID, INT32 area);  // application task

    if (tid_app[area] != NoTaskID)
    {
        Report (request for more than one CSA denied);
        return False;
    }

    // get the supervisor, change its socket ID
    INT32 oldmode = DisableInterrupts();
    CSupervisor *sup = listOfSupervisors.Find (nodeID, area);
    EnableInterrupts (oldmode);
    Assert (sup != NULL);  // the supervisor should exist

    // create and start the application task
    t_create (taskname[area], 128, SUPST, USRST, 0, &tid);
    sup->ChangeSocketID (tid);  // change socket ID
    t_start (tid, Mode_app, (TaskPrototype)ApplTask, args);
    tid_app[area] = tid;  // store the task ID

    return True;
}

// delete the application task
EBool CHardwareController::StopApplication (SINT32 nodeID, int area)
{
    if (tid_app[area] == NoTaskID) return False;

    INT32 oldmode = DisableInterrupts();
    listOfSupervisors.Remove (nodeID, area);
    EnableInterrupts (oldmode);

    ev_send (tid_app[area], EvKillTask);  // stop the application task
    tid_app[area] = NoTaskID;  // remove the task ID
    return True;
}


//--------------------------------------------------------------------------
//--- stop program execution
//--------------------------------------------------------------------------

// delete all tasks, call base class function
void CHardwareController::DeleteAllSimulationObjects ()
{
    ACController::DeleteAllSimulationObjects ();
}

// delete the task
void CHardwareController::StopExecution()
{
    INT32 tid_root;  // ROOT task
    t_ident ("ROOT", 0, &tid_root);  // get the ROOT task ID
    ev_send (tid_root, EvKillSim);  // notify ROOT task of the shutdown
}



//**************************************************************************
// application task
//**************************************************************************

void ApplTask(INT32 nodeID, INT32 area)
{
    CSupervisor *sup;

    // get the supervisor
    INT32 oldmode = DisableInterrupts();
    sup = listOfSupervisors.Find (nodeID, area);
    EnableInterrupts (oldmode);
    Assert (sup != NULL);  // the supervisor should exist

    // start the supervisor
    sup->Body();  // will only return when the task is done

    // delete the supervisor
    sup->ClearSocketID();  // set socket ID to 0
    delete sup;  // delete the supervisor

    // delete the task
    fclose (0);  // return some pREPC resources
    close (0);  // return pNA resources
    t_delete (0);  // delete this task

    Report (deletion of supervisor task has failed);

    free ((void *)-1);  // return pREPC memory that is still unreturned
    t_delete (0);  // delete this task

    Assert (0);  // t_delete() should succeed
}


