//
// contbase.cpp
//
// implementation of the synchronization controller (abstract base class)
//
// pSOS changes:
// *) ExecuteCommand(), KillSimulation: if PSOS_SIM, the StopExecution()
//	command is expected to return
// *) DeleteAllSimulationObjects(): if PSOS_SIM, then GetFirst() is used
//	instead of GetNext() when iterating through the listOfSupervisors
//	note:	this assumes that CHardwareController::StopApplication()
//		removes the Supervisor from its list; if it does not, then
//		an endless loop will occur!
//		the Controller will try to report this, but the report may
//		not come through
//
// created: 25.01.97
//
// last change: 18.04.98
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/17 10:32:42 $
//


#include "contbase.hpp"
#include "error.hpp"
#include "event.hpp"
#include "general.hpp"
#include "ci.hpp"
#include "clockbas.hpp"
#include "globals.hpp"
#include "supervis.hpp"


CClock *g_masterClock = NULL;
// set it to a value different from NULL if you want to use the global
// function GetMasterTime()
INT32 g_masterArea = NoAccount;
// the area of the master clock; set it to a value within [0, MAXASICS-1]
// if you want to create a master clock (that is, if the variable above
// is set to some clock object)


//**************************************************************************
// class ACController: provides an abstract base class for controlling all
//      simulation specific things
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

ACController::ACController (char qname[4], ACSocketAddress *addr) :
    CControllingObject (qname, addr, 9)
{
    SendModule (Controller);  // send the module type to EvalSys
    SendEnumSizes();  // send the enum-size message to EvalSys
}


ACController::~ACController ()
{
    DeleteAllSimulationObjects ();
    // that empties all lists and deletes the simulation objects

    DeleteTemplates();
    // delete the template objects and set the variables to NULL
}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

// reset the controller (deletes all simulation objects)
void ACController::Reset()
{
    DeleteAllSimulationObjects();  // delete all objects
    CControllingObject::Reset();  // reset the event queue as well
}

// reset the simulation (set all parameters in all applications to
// their default values); returns True if at least one object exists,
// False if no simulation objects exist
BOOL ACController::ResetTheSimulation()
{
    INT32 oldmode = DisableInterrupts();

    for (CSupervisor *supervisor = listOfSupervisors.GetFirst();
         supervisor != NULL;
         supervisor = listOfSupervisors.GetNext())
    {
        supervisor->Reset();
    } // for

    EnableInterrupts (oldmode);

    CControllingObject::Reset();  // reset the event queue as well

    if (listOfSupervisors.Empty()) return False;
    else return True;
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
// existing params: all of CControllingObject
BOOL ACController::SetParameter (const SParameter& param)
{
    if (param.module != Controller) return False;
    // the base class does not deal with networks

    return CControllingObject::SetParameter (param);
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned
BOOL ACController::GetParameter (SParameter& param) const
{
    if (param.module != Controller) return False;
    // the base class does not deal with networks

    return CControllingObject::GetParameter (param);
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False
// existing commands: all Controller commands of ECommandType concerning
// the application (not the Network!) in params.hpp
BOOL ACController::ExecuteCommand (const SCommand& command)
{
    SApplCommand *appl;  // all normal appl. commands
    CSupervisor *sup;  // for Suspend/ResumeOperation()
    INT32 oldmode;  // for disabling interrupts

    if (command.module != Controller) return False;
    // the base class does not deal with networks

    switch (command.type)
    {
        case CreateApplication:  // make a new application
            return CreateNewApplication (*(SApplCommand*)&command);
            break;
        case DeleteApplication:
        case SuspendApplication:
        case ResumeApplication:
            appl = (SApplCommand *)&command;
            oldmode = DisableInterrupts();
            sup = listOfSupervisors.Find (appl->nodeID, appl->area);
            EnableInterrupts (oldmode);
            if (sup == NULL) return False;

            switch (command.type)
            {
                case DeleteApplication:  // delete the supervisor
                    return DeleteSupervisor (appl->nodeID, appl->area);
                    break;
                case SuspendApplication:  // suspend the supervisor
                    sup->SuspendOperation ();
                    break;
                case ResumeApplication:  // resume the supervisor
                    sup->ResumeOperation ();
                    break;
                default: Assert (False);  // should never happen
            } // switch
            break;
        case SuspendSimulation:
        case ResumeSimulation:
            if (listOfSupervisors.Empty()) return False;
            // nothing to suspend/resume
            ExecuteCommandForAll (command.type);
            break;
        case ResetSimulation:
            return ResetTheSimulation();  // reset the simulation objects
            break;
        case KillSimulation:  // delete all objects, stop operation
            DeleteAllSimulationObjects();  // empty the lists

            ReplyToMessage (True);
            // reply at once, because the function will not return

            StopExecution();  // this call will not return
#ifndef PSOS_SIM
            Report (StopExecution() has returned);  // should never happen
            Assert (False);  // kill the program per force
#endif
            break;

        default:
            return CControllingObject::ExecuteCommand (command);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int ACController::ReportError (EReportType type, char *file, int line,
                              char *expression)
{
    return ::ReportError (type, file, line, expression, Controller);
    // call the global error report with module
}



//--------------------------------------------------------------------------
//--- stop program execution
//--------------------------------------------------------------------------

// delete all entries in the supervisor list
void ACController::DeleteAllSimulationObjects ()
{
    SINT32 nodeID;
    int area;
    CSupervisor *oldsup = NULL;  // for checking if we are in endless loop

    INT32 oldmode = DisableInterrupts();

    for (CSupervisor *sup = listOfSupervisors.GetFirst();
         sup != NULL;
#ifdef PSOS_SIM
         sup = listOfSupervisors.GetFirst())
#else
         sup = listOfSupervisors.GetNext())
#endif
    {
        if (oldsup == sup)
        {
            Report (endless loop in ACController::DeleteAllSimulationObjects!);
            tm_wkafter (1000);  // let other tasks (the error task) to CPU
        }
        sup->GetIdentification (nodeID, area);
        StopApplication (nodeID, area);  // stop the application
        oldsup = sup;
    } // for

    listOfSupervisors.DeleteAll();  // delete or unref all supervisors
    // that's just a formality, it is not to be expected that another
    // controller will be started

    // note: this statement assumes that there are never two controllers
    // in the same context (using the same lists)!

    EnableInterrupts (oldmode);
}


//--------------------------------------------------------------------------
//--- internal command execution
//--------------------------------------------------------------------------

// create a new supervisor (start the application); returns False if
// area not in [0, min {MAXASICS-1, 15}], if area is equal to the area
// of the master clock, if nodeID is not in [0, 16^3-1], or if an
// application with that (nodeID, area) exists; else the application
// is created and True is returned
BOOL ACController::CreateNewApplication (SApplCommand& applc)
{
    CSupervisor *sup;
    static char qname[2*sizeof(int)+1];  // name of the supervisor queue
    // note: the name is only unique for #nodes <= 16^3 (only the first
    // three places are used for the nodeID) and #areas <= 16 (the first
    // place is used for the area)!

    if ((applc.nodeID < 0) || (applc.nodeID >= 4096)) return False;
    // nodeID may only be in the range [0, 16^3-1]

    if ((applc.area < 0) || (applc.area >= MAXASICS)) return False;
    // area may only be in the range [0, MAXASICS-1]

    if (applc.area == g_masterArea) return False;
    // the area of the master clock (used in hardware-based simulation)
    // may not be used by the application

    if (applc.area > 15) return False;
    // apart from that, area must not exceed 15

    INT32 oldmode = DisableInterrupts();
    if (listOfSupervisors.Find (applc.nodeID, applc.area)) return False;
    // nodeID/area already in use
    EnableInterrupts (oldmode);

    itoa (applc.nodeID*16 + applc.area%16, qname, 16);
    // convert the nodeID and area to a queue name

    oldmode = DisableInterrupts();
    sup = new CSupervisor (qname, GetSocketAddress(), applc.nodeID, applc.area,
                           GetClockFrequency (applc.nodeID, applc.area));
    EnableInterrupts (oldmode);

    Assert (sup != NULL);  // new supervisor is created

    return StartApplication (applc.nodeID, applc.area);
}


// delete the supervisor (stop the application); returns True if the
// application was deleted successfully, else False
BOOL ACController::DeleteSupervisor (SINT32 nodeID, int area)
{
    BOOL result = StopApplication (nodeID, area);  // stop the application
    INT32 oldmode = DisableInterrupts();
    listOfSupervisors.Delete (nodeID, area); // delete or unref the process
    EnableInterrupts (oldmode);
    return result;
}


// executes the command (SuspendSimulation, ResumeSimulation) by suspending
// or resuming all supervisors
void ACController::ExecuteCommandForAll (ESystemCommand type)
{
    INT32 oldmode = DisableInterrupts();
    for (CSupervisor *sup = listOfSupervisors.GetFirst();
         sup != NULL;
         sup = listOfSupervisors.GetNext())
    {
        switch (type)
        {
            case SuspendSimulation:
                sup->SuspendOperation();
                break;
            case ResumeSimulation:
                sup->ResumeOperation();
                break;
            default:
                Assert (False);
        } // switch
    } // for
    EnableInterrupts (oldmode);
}


