//
// ctrling.cpp
//
// implementation of the base class used for all modules that are
// controlling some other module
//
// hardware changes:
// *) added public function ChangeSocketID()
// *) added public function ClearSocketID()
// *) variable length queues are used
// *) Body() is left if EvKillTask is received
// *) tm_wkafter() in Body() is replaced with ev_received(EvKillTask) with
//	the same timeout as the tm_wkafter()
//
// created: 13.01.97
//
// last change: 22.05.98
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/17 10:32:42 $
//


#ifdef PSOS_SIM
  #include <pna.h>  // shr_socket()
  #include "socket.hpp"
#endif
#include <string.h>  // for memset()

#include "ctrling.hpp"
#include "error.hpp"
#include "sockstre.hpp"

#ifdef PSOS_SIM
  extern const INT32 EvKillTask;
  // when this event is received, the Body() function should be left
#endif


/**/
#ifdef PSOS_SIM
  const INT32 MaxEventLength = 256;
  const INT32 MaxEventNumber = 100;
#endif


//**************************************************************************
// module global functions
//**************************************************************************

static INT32 SecondsToTicks (double period);
// converts 'period' (in seconds) into pSOS ticks



//**************************************************************************
// class CControllingObject: the class is a basic interface between modules
//    and the evaluation; it provides a function for reporting events, and
//    executes the base loop that services the sockets to the evaluation
//**************************************************************************


BOOL CControllingObject::initialized_ = False;
EEventClass CControllingObject::table_[EventCount];


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// creates the event queue and the stream socket to which it should connect
CControllingObject::CControllingObject (char qname[4], ACSocketAddress *addr,
    SINT32 nodeID)
    : ACProcess (nodeID)
{
    socket_ = new CStreamClient (addr);
    Assert (socket_ != NULL);

#ifdef PSOS_SIM
    // use a variable length queue
    q_vcreate (qname, Q_LOCAL, MaxEventNumber, MaxErrorLength, &eventQueueID_);
#else
    q_create (qname, 0, Q_LOCAL | Q_NOLIMIT, &eventQueueID_);
#endif

    // initialize the event classification table
    if (!initialized_ && !InitEventTable ())
    {
        Report (event table has leak(s));
    }

    Assert (sizeof (void*) <= sizeof (INT32));
    // this condition is needed because the queue stores event pointers
    // using INT32 variables

    resync_ = False;

    waitPeriod_ = DefaultIfcPeriod;  // set the default wait period
}

CControllingObject::~CControllingObject ()
{
    classMask_.Reset();  // reset the class mask (no more events)
    EmptyQueue ();  // empty the event queue
#ifdef PSOS_SIM
    q_vdelete (eventQueueID_);
#else
    q_delete (eventQueueID_);
#endif
    delete socket_;
}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

// empty the event queue, reset event report masks
void CControllingObject::Reset()
{
    eventMask_.Reset();  // reset event mask
    classMask_.Reset();  // reset the class mask
    EmptyQueue ();  // discard all events in the queue
    waitPeriod_ = DefaultIfcPeriod;  // set the default wait period
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
// existing parameters: EventReport, EventClassReport, InterfacePeriod
BOOL CControllingObject::SetParameter (const SParameter& param)
{
    switch (param.type)
    {
        case EventReport:
            UpdateMask (eventMask_, (SEventReport *) &param);
            break;
        case EventClassReport:
            UpdateMask (classMask_, (SEventReport *) &param);
            break;
        case InterfacePeriod:
            if (((SSingleParameter *)&param)->value <= 0) return False;
            // periods have to be positive
            waitPeriod_ = ((SSingleParameter *)&param)->value;
            break;

        default:  // unknown parameter
            return False;
    }  // switch

    return True;
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned
// existing parameters: EventReport, EventClassReport, InterfacePeriod,
// TimerGranularity
BOOL CControllingObject::GetParameter (SParameter& param) const
{
    switch (param.type)
    {
        case EventReport:
            GetBit (eventMask_, (SEventReport *) &param);
            break;
        case EventClassReport:
            GetBit (classMask_, (SEventReport *) &param);
            break;
        case InterfacePeriod:
            ((SSingleParameter *)&param)->value = waitPeriod_;
            break;
        case TimerGranularity:
            ((SSingleParameter *)&param)->value = 1/TICKS2SEC;
            break;

        default:  // unknown parameter
            return False;
    }  // switch

    return True;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False
// existing command: ResetModule
BOOL CControllingObject::ExecuteCommand (const SCommand& command)
{
    switch (command.type)
    {
        case ResetModule:
            Reset ();
            break;

        default:
            return False;  // unknown command
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- event logging
//--------------------------------------------------------------------------

// stores the event until it can be processed
void CControllingObject::NoteEvent (SEvent *event)
{
    Assert (event != NULL);

    event->time = GetGlobalTime ();  // get the global time, store it

    // now check if the event should be reported
    if (!EventReportAllowed (event->type))  // don't report
    {
        delete event;  // discard the event
        return;
    }

    if (event->type == Snapshot) event->time = GetSnapshotTime();
    // use the snapshot time instead of the global time

#ifdef PSOS_SIM
    q_vsend (eventQueueID_, event, event->size);  // store the event
    delete event;  // not needed anymore
#else
    INT32 msg[4];

    msg[0] = (INT32)event;  // store the event pointer in the message!

    q_send (eventQueueID_, msg);  // store the event in the queue
#endif
}


//--------------------------------------------------------------------------
//--- the main function
//--------------------------------------------------------------------------

void CControllingObject::Body()
{
    SEvent *event;  // for the handling of events
    EMessageType type;  // first byte in a socket msg should be the type
    INT32 ev;

    for (;;)
    {
        // check the input socket for new messages from the evaluation
        while (socket_->ReceiveMessageHeader (&type))
        // something came from the socket
        {
#ifdef PSOS_SIM
            if (ev_receive (EvKillTask, EV_NOWAIT, 0, &ev)==0) return;
            // got the signal to kill the task -> return from the main loop
#endif

            if (resync_ && (type != MsgSynchronizationByte)) continue;
            // ignore all data until the sync byte is received

            switch (type)
            {
                case MsgCommand:
                case MsgSetParameter:
                case MsgGetParameter:  // all three expect further data
                    ProcessData (type);
                    break;
                case MsgResynchronize:  // send resync byte
                    socket_->SendSynchronizationByte();
                    break;
                case MsgSynchronizationByte:  // turn off resync mode
                    resync_ = False;
                    break;

                default:  // oops, something went wrong --> resynchronize
                    Resynchronize();
                    break;
            } // switch
        } // while

        // send all events to the evaluation
        while ((event = GetEvent()) != NULL)
        {
            TransmitEvent (event);  // send the event
#ifdef PSOS_SIM
            // in pSOS, vqueues are used, and the memory returned by
            // GetEvent() is of type char[], not SEvent*; so we have to
            // return it via delete[]
            delete[] (char*)event;
#else
            delete event;  // delete the event
#endif
        }  // while

        DoPeriodicActivities();
        // give the derived class a chance to do own periodic work

#ifdef PSOS_SIM
        if (ev_receive (EvKillTask, EV_WAIT, SecondsToTicks (waitPeriod_),
                        &ev)==0) return;
        // got the signal to kill the task -> return from the main loop
#else
        tm_wkafter (SecondsToTicks (waitPeriod_));  // wait some time
#endif
    } // for
}


#ifdef PSOS_SIM
//--------------------------------------------------------------------------
//--- changing socket ID
//--------------------------------------------------------------------------

// the function calls the pSOS system call shr_socket() to get a socket
// descriptor that is valid for the specified task, then calls
// s_delete() to close the current socket and sets the socket ID to
// the new ID
// note: the function must be called by the task that has created the
// object, and only if the Body() of the object is executed by another
// task (the one with ID 'tid')
// note: from the call to ChangeSocketID() on, all actions that result on
// an access to the socket in the context of the currently calling task
// will not succeed
void CControllingObject::ChangeSocketID (INT32 tid)
{
    INT32 oldsid = socket_->GetSocketID();  // get current ID
    INT32 newsid = shr_socket(oldsid, tid);  // get ID for specified task
    Assert (newsid != -1);  // the call must succeed
    if (oldsid == newsid) return;  // what should we do in this case?
    socket_->SetSocketID (newsid);  // set the new ID
    s_delete (oldsid);  // close the old socket
}

// sets the socket ID to zero
// this is needed to allow the supervisors to delete only their own
// descriptors (close(id) always deletes the socket, even if it is
// still used by another task)
// note: do not use the socket after this function has been called!
void CControllingObject::ClearSocketID ()
{
    socket_->SetSocketID (0);  // set socket ID to zero
}

#endif


//--------------------------------------------------------------------------
//--- internal socket handling
//--------------------------------------------------------------------------

// returns the address of the socket
ACSocketAddress *CControllingObject::GetSocketAddress ()
{
    return socket_->GetSocketAddress();
}


// send a MsgReply (containing the value) over the socket
void CControllingObject::ReplyToMessage (double value)
{
    SDoubleReply reply (value);
    socket_->ReplyToMessage (&reply);
}

// send a MsgReply over the socket, containing the result of a message
void CControllingObject::ReplyToMessage (BOOL result)
{
    SBooleanReply reply (result);
    socket_->ReplyToMessage (&reply);
}

// send a MsgReply over the socket, followed by the data structure
void CControllingObject::ReplyToMessage (SBase *data)
{
    Assert (data != NULL);

    SReply reply;
    INT32 replysize = reply.size;
    char *buf;

    reply.size = replysize+data->size;  // that's the whole message size

    buf = (char*)socket_->GetMessageBuffer (reply.size);
    Assert (buf != NULL);

    memcpy (buf, &reply, replysize);  // copy the reply into the buffer
    memcpy (buf+replysize, data, data->size); // followed by the data itself
    // note: the first item in the buffer is the reply, so the buffer can
    // be cast to SBase

    socket_->ReplyToMessage ((SBase*)buf);
    socket_->ReturnMessageBuffer (buf);
}

// send the module type over the socket
// assert: module is one of {Controller, Supervisor}
void CControllingObject::SendModule (EModule module)
{
    Assert ((module == Controller) || (module == Supervisor));
    if (!socket_->SendModule (module)) Report (could not send module type);
}

// send the byte-order message over the socket
void CControllingObject::SendByteOrder()
{
    if (!socket_->SendByteOrder()) Report (could not send byte-order);
}

// send the enum-size message over the socket
void CControllingObject::SendEnumSizes()
{
    SEnumSizes msg;  // the sizes are set automatically in the ctor
    if (socket_->SendMessage (MsgEnumSizes, &msg)==(INT32)NoAccount)
        Report (could not send enum-sizes);
}


//--------------------------------------------------------------------------
//--- internal event retrieval
//--------------------------------------------------------------------------

// returns the next event from the queue, or NULL if the queue is empty
SEvent *CControllingObject::GetEvent ()
{
#ifdef PSOS_SIM
    char msg[MaxEventLength];  // local receive buffer for the event
    INT32 len;  // size of the message received
    char *event;  // the destination buffer that is returned to the caller

    if (q_vreceive (eventQueueID_, Q_NOWAIT, 0, msg, MaxEventLength, &len)==0)
    {
        Assert (len > 0);  // there must be some data

        event = new char[len];     // alloc memory for returning the event
        memcpy (event, msg, len);  // copy it from 'msg'
        return (SEvent*)event;     // cast to SEvent and return
/**/    // the event gets copied way too often --> think of something better!)
    }
    else return NULL;
#else
    INT32 msg[4];

    if (q_receive (eventQueueID_, Q_NOWAIT, 0, msg) == 0)
    {
        // msg[0] is the pointer to the event
        return (SEvent*) msg[0];
    }
    else return NULL;  // no more events in the queue
#endif
}


//--------------------------------------------------------------------------
//--- periodic activities of derived classes
//--------------------------------------------------------------------------

// is called once every period; derived classes should overload this
// function if they have own periodic work to do
void CControllingObject::DoPeriodicActivities()
{
    // all activities of the base class are carried out in Body()
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int CControllingObject::ReportError (EReportType type, char *file, int line,
                                     char *expression)
{
    return ::ReportError (type, file, line, expression);
    // just call the global function
    // note: cannot make this a pure virtual function because in that case
    // the compiler (GNU) complains about its use in the ctor
}


//--------------------------------------------------------------------------
//--- internal queue management
//--------------------------------------------------------------------------

// deletes all events that are stored in the queue
// you should set the event class mask to zero before calling this function
void CControllingObject::EmptyQueue ()
{
#ifdef PSOS_SIM
    char msg[MaxEventLength];
    INT32 len;

    while (q_vreceive (eventQueueID_, Q_NOWAIT, 0, msg, MaxEventLength, &len)
           == 0)
    {
        // no need to do anything
    }
#else
    SEvent *event;
    INT32 msg[4];

    while (q_receive (eventQueueID_, Q_NOWAIT, 0, msg) == 0)
    {
        // msg[0] is the pointer to the event
        event = (SEvent*) msg[0];
        delete event;  // delete the event
    }
    // let's hope this will never result in an endless loop
#endif
}


//--------------------------------------------------------------------------
//--- internal event management
//--------------------------------------------------------------------------

// returns True if the event may be reported, else False
BOOL CControllingObject::EventReportAllowed (EEvent event)
{
    if (!(eventMask_ & event) ||  // event report turned off
        !(classMask_ & table_[event]))  // whole class turned off
    {
        return False;
    }
    else return True;  // both event and class are turned on
}


//--------------------------------------------------------------------------
//--- internal management of the bitmasks
//--------------------------------------------------------------------------

void CControllingObject::UpdateMask (CBitMask& bitmask, const SEventReport *report)
{
    if (report->masksize <= 0)  // have to set a single event
    {
        SINT32 bit;

        report->Get (bit);
        if (report->set) bitmask |= bit;  // set the bit
        else if (bitmask & bit) bitmask ^= bit;
        // bit is set -> clear the bit
    }
    else  // have to set the whole mask
    {
        bitmask.Set (report->mask, report->masksize);
    }
}

// store the value of a bit in 'report'
void CControllingObject::GetBit (const CBitMask& bitmask,
                                 SEventReport *report) const
{
    SINT32 bit;

    report->Get (bit);  // obtain the bit that is of interest
    if (bitmask & bit) report->set = True;  // the bit is set
    else report->set = False;  // the bit is not set
}


//--------------------------------------------------------------------------
//--- internal management of event classifications
//--------------------------------------------------------------------------

BOOL CControllingObject::InitEventTable ()
{
    if (initialized_) return True;
    // table has already been initialized

    int i;

    initialized_ = True;

    for (i=0; i<EventCount; i++) table_[i] = NoClassification;
    // initialize the table to allow checking for leaks (if events are
    // added later on)


    // below follows a simple table that may be put into an own file event.cpp
    // in future versions

    // Clock
    table_[ClockUpdate] = SystemReport;
    table_[ClockPreload] = SystemReport;
    table_[Snapshot] = InterruptEvent;
    table_[GpsEvent] = InterruptEvent;
    table_[ClockReadFault] = InterruptEvent;
    table_[ClockWriteFault] = InterruptEvent;
    table_[IgnoreSetting] = InterruptEvent;

    // Network
    table_[MessageLost] = FaultInjection;
    table_[DeliverMessage] = SystemReport;
    table_[EarlyMessage] = FaultInjection;
    table_[LateMessage] = FaultInjection;
    table_[CrashStatusChange] = FaultInjection;

    // Lance
    table_[DiscardSendMessage] = FaultInjection;
    table_[DiscardRecvMessage] = FaultInjection;
    table_[ScrambleSendMessage] = FaultInjection;
    table_[ScrambleRecvMessage] = FaultInjection;

    // ClockSync
    table_[ChangeAssociation] = SystemReport;
    table_[ConvergenceFailed] = SystemReport;
    table_[IntervalEvent] = SystemReport;
    table_[RateEvent] = SystemReport;
    table_[ReceivedMessage] = InterruptEvent;
    table_[SendMessage] = InterruptEvent;
    table_[TimerEvent] = InterruptEvent;
    table_[UserEvent] = SystemReport;


    // now go through the table once more, search for leaks;
    // if a leak is found: check the above table against EEvent, make sure
    // that all events of EEvent are listed above
    // if the leak occurs because the values of EEvent are not continuous,
    // and if this is intended, skip the hole in the code below
    for (i=0; i<EventCount; i++)
    {
        if (table_[i] == NoClassification) return False;  // found a leak
    } // for

    return True;
}


//--------------------------------------------------------------------------
//--- internal resynchronization
//--------------------------------------------------------------------------

// ask the evaluation to resynchronize, go into resync mode
void CControllingObject::Resynchronize()
{
    socket_->Resynchronize();
    resync_ = True;
}


//--------------------------------------------------------------------------
//--- internal socket data handling
//--------------------------------------------------------------------------

// process data from the socket
void CControllingObject::ProcessData (EMessageType msgtype)
{
    SBase *msgbuf;  // rest of the data, structure SBase is assumed

    // get the message size and body
    if (!socket_->ReceiveMessage (&msgbuf))
    {
        Resynchronize();
        return;
    }

    switch (msgtype)
    {
        case MsgCommand:
            ReplyToMessage (ExecuteCommand (*(SCommand*)msgbuf));
            break;
        case MsgSetParameter:
            ReplyToMessage (SetParameter (*(SParameter*)msgbuf));
            break;
        case MsgGetParameter:
            if (!GetParameter (*(SParameter*)msgbuf))
                ReplyToMessage (False);
            else
                ReplyToMessage (msgbuf);
            break;

        default:
            Assert (False);  // should never happen
    } // switch

    socket_->ReturnMessageBuffer (msgbuf);  // not needed anymore
}

// send the event to the evaluation
// assert: event != NULL
void CControllingObject::TransmitEvent (SEvent *event)
{
    Assert (event != NULL);
    socket_->SendMessage (MsgEvent, event);
}



//**************************************************************************
// module global functions
//**************************************************************************

// converts 'period' (in seconds) into pSOS ticks
static INT32 SecondsToTicks (double period)
{
    if (period*TICKS2SEC < 1) return 1;  // wait at least one tick!
    else return (INT32)(period*TICKS2SEC);
}


