//
// ctrling.hpp
//
// declaration of the base class used for all modules that are controlling
// some other module
//
// hardware changes:
// *) added public function ChangeSocketID()
// *) added public function ClearSocketID()
//
// created: 13.01.97
//
// last change: 26.03.98
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/17 10:37:06 $
//


#ifndef _CTRLING_HPP_
#define _CTRLING_HPP_

#include "psos.h"

#include "bitmask.hpp"
#include "bool.hpp"
#include "process.hpp"
#include "structs.hpp"


class ACSocketAddress;
class CStreamInterface;


//**************************************************************************
// class CControllingObject: the class is a basic interface between modules
//    and the evaluation; it provides a function for reporting events, and
//    executes the base loop that services the socket to the evaluation
//**************************************************************************

class CControllingObject : public ACProcess
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CControllingObject (char qname[4], ACSocketAddress* addr,
                        SINT32 nodeID = DefaultProcessID);
    // creates the event queue and the stream socket to which it should
    // connect

    virtual ~CControllingObject ();
    // deletes the queue, closes the socket


    //----------------------------------------------------------------------
    //--- reset
    //----------------------------------------------------------------------

    virtual void Reset();
    // empty the event queue, reset event report masks and waitPeriod


    //----------------------------------------------------------------------
    //--- configuration of system parameters
    //----------------------------------------------------------------------

    virtual BOOL SetParameter (const SParameter& param);
    // sets the given parameter to its new value; if the parameter does
    // not exist or if the value is not valid, False is returned
    // existing parameters: EventReport, EventClassReport, InterfacePeriod

    virtual BOOL GetParameter (SParameter& param) const;
    // returns the value of the given parameter in the appropriate field of
    // SParameter; if the parameter does not exist, False is returned
    // existing parameters: EventReport, EventClassReport, InterfacePeriod,
    // TimerGranularity


    //----------------------------------------------------------------------
    //--- execution of system commands
    //----------------------------------------------------------------------

    virtual BOOL ExecuteCommand (const SCommand& command);
    // executes the command; returns True if successfull, otherwise False
    // existing command: ResetModule


    //----------------------------------------------------------------------
    //--- event logging
    //----------------------------------------------------------------------

    virtual void NoteEvent (SEvent *event);
    // stores the event until it can be processed


    //----------------------------------------------------------------------
    //--- the main function
    //----------------------------------------------------------------------

    virtual void Body ();
    // this function is executed in pure simulation and should be used as
    // main function in the HW;
    // it reads the messages from the socket and calls Set/GetParameter()
    // or ExecuteCommand() accordingly, then sends all events to the
    // socket and finally calls DoPeriodicActivities()


#ifdef PSOS_SIM
    //----------------------------------------------------------------------
    //--- changing socket ID
    //----------------------------------------------------------------------

    void ChangeSocketID (INT32 tid);
    // the function calls the pSOS system call shr_socket() to get a socket
    // descriptor that is valid for the specified task, then calls
    // s_delete() to close the current socket and sets the socket ID to
    // the new ID
    // note: the function must be called by the task that has created the
    // object, and only if the Body() of the object is executed by another
    // task (the one with ID 'tid')
    // note: from the call to ChangeSocketID() on, all actions that result
    // on an access to the socket in the context of the currently calling
    // task will not succeed

    void ClearSocketID ();
    // sets the socket ID to zero
    // this is needed to allow the supervisors to delete only their own
    // descriptors (close(id) always deletes the socket, even if it is
    // still used by another task)
    // note: do not use the socket after this function has been called!
#endif


protected:
//public: /**/
    //----------------------------------------------------------------------
    //--- internal socket handling
    //----------------------------------------------------------------------

    ACSocketAddress *GetSocketAddress ();
    // returns the address of the socket

    void ReplyToMessage (double value);
    // send a MsgReply (containing the value) over the socket

    void ReplyToMessage (BOOL value);
    // send a MsgReply over the socket, containing the result of a message

    void ReplyToMessage (SBase *data);
    // send a MsgReply over the socket, containing the data

    void SendModule (EModule module);
    // send the module type over the socket
    // assert: module is one of {Controller, Supervisor}

    void SendByteOrder();
    // send the byte-order message over the socket

    void SendEnumSizes();
    // send the enum-size message over the socket


    //----------------------------------------------------------------------
    //--- internal event retrieval
    //----------------------------------------------------------------------

    SEvent *GetEvent ();
    // returns the next event from the queue, or NULL if the queue is empty


    //----------------------------------------------------------------------
    //--- periodic activities of derived classes
    //----------------------------------------------------------------------

    virtual void DoPeriodicActivities();
    // is called once every period; derived classes should overload this
    // function if they have own periodic work to do


    //----------------------------------------------------------------------
    //--- error reports
    //----------------------------------------------------------------------

    virtual int ReportError (EReportType type, char *file, int line,
                             char *expression);


private:

    INT32 eventQueueID_;  // ID of the queue used to store the events
    CStreamInterface *socket_;  // IFC to the evaluation
    BOOL resync_;  // True, if the communication is out of sync

    CBitMask eventMask_;  // bitmask for turning on/off events
    CBitMask classMask_;  // bitmask for turning on/off event classes

    static EEventClass table_[EventCount];
    // table for association of an event class with the event type
    static BOOL initialized_;
    // the first object that is created should initialize it in its ctor

    double waitPeriod_;
    // period of the ctrling object in seconds (default: 1 second)


    //----------------------------------------------------------------------
    //--- internal queue management
    //----------------------------------------------------------------------

    void EmptyQueue ();
    // deletes all events that are stored in the queue
    // set the event class mask to zero before calling this function


    //----------------------------------------------------------------------
    //--- internal event management
    //----------------------------------------------------------------------

    BOOL EventReportAllowed (EEvent event);
    // returns True if the event may be reported, else False


    //----------------------------------------------------------------------
    //--- internal management of the bitmasks
    //----------------------------------------------------------------------

    void UpdateMask (CBitMask& bitmask, const SEventReport *report);
    // update the bitmask according to the data in 'report'

    void GetBit (const CBitMask& bitmask, SEventReport *report) const;
    // store the value of a bit in 'report'


    //----------------------------------------------------------------------
    //--- internal management of event classifications
    //----------------------------------------------------------------------

    static BOOL InitEventTable ();
    // initializes the event table, returns False if there are events which
    // do not have a classification; in that case, add the classification
    // to the code of InitEventTable()

//public: /**/
    //----------------------------------------------------------------------
    //--- internal resynchronization
    //----------------------------------------------------------------------

    void Resynchronize();
    // ask the evaluation to resynchronize


    //----------------------------------------------------------------------
    //--- internal socket data handling
    //----------------------------------------------------------------------

    void ProcessData (EMessageType msgtype);
    // process data from the socket

//public:/**/
    void TransmitEvent (SEvent *event);
    // send the event to the evaluation
    // assert: event != NULL
};


#endif


