//
// impl_hw.cpp
//
// *) double GetSnapshotTime() returns the value in the MS/TSSNU registers
//	of the master clock
// 	if INSTALL_MASTER is True, the snapshot time is returned, otherwise
//	GetGlobalTime() is returned
//
// function implementations for the hardware
//
// *) SetGpsReceiver() stores the type of a GPS receiver
// *) GetGpsReceiver() returns the type of a GPS receiver
//
// GetClockIntvec() returns value of nti_irq_base[] from initnti.cpp
// GetGpsIntvec() ignores all parameters, returns one fictive vector
//	that is used by the GPS task
// GetClockIntLevel() returns NTI_IRQ_LEVEL (must be same as in initnti.cpp!)
// GetGlobalTime() returns GetMasterTime()
// 	if INSTALL_MASTER is True, GetMasterTime() is returned, otherwise a
//      counter is used
// GetPosition() returns the default position of the receive timestamp
// GetClockFrequency() returns 10 MHz
// MakeSystemSnapshot() accesses LA-Trigger on NTI of g_masterArea (=5,6)
//	the address is hard-coded (vme_base_addr + 0x00600000 + 0x0E for
//	slot 0, vme_base_addr + 0x00600100 + 0x0E for slot 1)
// SampleReceiveTimestamps() samples the time of all NTIs on slots 2-5
//	(area 0-3) into the MS/TSRCV registers
// 	this is done via the NTI on area 0 (slot 2), where a write access to
// 	offset 0x2000 + 0x1C on the NTI memory space causes such behaviour
//
// created: 13.02.97
//
// last change: 29.06.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Exp $$Date: 1998/07/03 13:11:52 $
//


#include <stdlib.h>  // for NULL
#include <psos.h>

#include "error.hpp"
#include "general.hpp"

#include "ci.h"
#include "clockcfg.hpp"
#include "msgform.hpp"

#include "clockbas.hpp"
#include "time.hpp"

#define NTI_IRQ_LEVEL 6          /* VME IRQ-Vector from Carrier Board */
// the define is copied from initnti.cpp!


extern CClock *g_masterClock;
extern int g_masterArea;
// set in cont_hw.cpp
// pointer to and area of the master clock; needed in MakeSystemSnapshot()


extern INT32 g_ntiAddress[6];
// defined in simutc.cpp, set in initnti.cpp

extern INT32 vme_base_addr;
extern INT32 nti_io[4];
extern INT32 nti_irq_base[4];
// defined and set in initnti.cpp


//**************************************************************************
// obtaining interrupt vectors and levels
//**************************************************************************

// returns the clock interrupt vector for the area on the given node;
// assert: area < MAXASICS
// assert: area < 16
int GetClockIntvec (SINT32, int area)
{
    Assert (area < MAXASICS);
    Assert (area < 16);
    return nti_irq_base[area];
}

// returns the GPS-receiver interrupt vector for the area on the given node
// assert: area < MAXASICS, area < 16, gpu < EGpsUnitCount
// note: when porting to hardware, return the appropriate vector for the
// area and gpu and ignore the nodeID
int GetGpsIntvec (SINT32, int, EGpsUnit)
{
    return 0x0FFFF;
    // note: this assumes that there is no pSOS vector with this number
    // (but since pSOS vectors are only 1 byte long, we can be quite
    // certain that 0x0FFFF is an invalid vector)
}

// returns the interrupt level of the UTCSU and LANCE interrupts; the levels
// should be equal (if they are not, correct behaviour of the simulation
// cannot be guaranteed).
INT8 GetClockIntLevel()
{
    return NTI_IRQ_LEVEL;
}


//**************************************************************************
// get global time
//**************************************************************************

// returns the master time (in seconds)
double GetMasterTime ()
{
    Assert (g_masterClock != NULL);
    // the master clock must have been installed by the Controller

    INT32 ms, ts, us;  // the clock registers that are important

    ts = g_masterClock->GetTSGETL();  // latch the clock value
    ms = g_masterClock->GetMSGET() & 0x0FFFFFF;  // get the other registers
    us = g_masterClock->GetUSGET();

    return NtpToSeconds (ms, ts, us);
}

// returns the global time (in seconds)
double GetGlobalTime ()
{
#ifndef INSTALL_MASTER
    static int i=1000;
    return i++;
#else
    return GetMasterTime();
#endif
}

// returns the snapshot time (in seconds)
double GetSnapshotTime ()
{
#ifndef INSTALL_MASTER
    return GetGlobalTime();
#else
    Assert (g_masterClock != NULL);

    INT32 ms, ts;  // the clock registers that are important

    ts = g_masterClock->GetTSSNU();  // get timestamp
    ms = g_masterClock->GetMSSNU() & 0x0FFFFFF;  // mask out checksum

    return NtpToSeconds (ms, ts);
#endif
}


//**************************************************************************
// get the position of network time value in message
//**************************************************************************

// just default position
void GetPosition (SMsgPosition& data, SINT32 nodeID, SINT32 networkID)
{
    ti_acx netTime;  // this structure is put in front of the message
    INT16 offset = 0;  // the offset of this structure

    data.position[SMsgPosition::Area] = 0;  // the area is the first item
    data.length[SMsgPosition::Area] = sizeof(INT32);
    offset = data.length[SMsgPosition::Area];

    data.position[SMsgPosition::Macrostamp] = 
        offset + (INT32)&(netTime.ms) - (INT32)&netTime;
    data.position[SMsgPosition::Timestamp] =
        offset + (INT32)&(netTime.ts) - (INT32)&netTime;
    data.position[SMsgPosition::Accuracy] =
        offset + (INT32)&(netTime.acc) - (INT32)&netTime;

    data.length[SMsgPosition::Macrostamp] = sizeof(netTime.ms);
    data.length[SMsgPosition::Timestamp] = sizeof(netTime.ts);
    data.length[SMsgPosition::Accuracy] = sizeof(netTime.acc);
}


//**************************************************************************
// get clock frequency
//**************************************************************************

// returns the frequency of a clock; the function is called by the Controller
// whenever a new CSA is created
// if no frequency is stored, a default frequency is returned
// note: all areas have a 10MHz quarz
double GetClockFrequency (SINT32, int)
{
    static const double FREQUENCY = 10e6;

    return FREQUENCY;  // return the default frequency
}


//**************************************************************************
// trigger hardware snapshots on all UTCSUs
//**************************************************************************

// initiates a hardware snapshot on all UTCSUs in the system (that assumes
// that all UTCSUs are hardwired together, regardless of their subnets)
// note: triggering the snapshot occurs via reading (or writing) to the
// LA-Trigger on the NTI IO space of the master clock
// computation of the address is hard-coded!
void MakeSystemSnapshot()
{
    int i;
    Report (triggering system snapshot);

    switch (g_masterArea)
    {
        case 4:  // slot 0
            i = *(INT16*)(vme_base_addr + 0x00600000 + 0x0E);
            // VME base + IO Space Offset + LA-Trigger Offset
            break;
        case 5:  // slot 1
            i = *(INT16*)(vme_base_addr + 0x00600100 + 0x0E);
            // VME base + IO Space Offset + LA-Trigger Offset
            break;
        default:
            Assert (0);  // other masters are not supported
    } // switch
}


//**************************************************************************
// simulate packet reception
//**************************************************************************

// this is done via the NTI on area 0 (slot 2), where a write access to
// offset 0x2000 + 0x1C on the NTI memory space samples the time of all
// NTIs on slots 2-5 (area 0-3) into the MS/TSRCV registers
void SampleReceiveTimestamps()
{
    Assert (g_ntiAddress[2] != 0);  // the NTI slot must be occupied
    *(INT32*)(g_ntiAddress[2] + 0x2000 + 0x1C) = 1;
    // this should trigger sampling on all areas
}


//**************************************************************************
// receiver handling
//**************************************************************************

static EGpsReceiver s_gpsReceiverType[MAXASICS][EGpsUnitCount];

// stores the type of a GPS receiver
void SetGpsReceiver (int area, EGpsUnit gpu, EGpsReceiver gpsrec)
{
    Assert ((area > 0) && (area < MAXASICS));
    s_gpsReceiverType[area][gpu] = gpsrec;
}

// returns the type of a GPS receiver
EGpsReceiver GetGpsReceiver (int area, EGpsUnit gpu)
{
    Assert ((area > 0) && (area < MAXASICS));
    return s_gpsReceiverType[area][gpu];
}


