//
// opnew.cpp
//
// provides an operator new which uses pt_getbuf() to allocate a partition
// buffer; operator delete uses pt_retbuf() to return the buffer to the
// pool
// there are several memory partitions with different buffer sizes
//
// contents:
//	NumberOfPartitions	number of different partitions
//	SPartitionData		data of one partition
//	g_PartitionTable	an array containing data of all partitions
//	InitMemory()		creates partitions, initializes data table
//	DisposeMemory()		deletes partitions (returns memory to region)
//	CheckMemory()		reports statistical data to the error stream
//	operator new()		for allocating memory with new
//	operator delete()	for freeing memory with delete
//
// created: 27.04.98
//
// last change: 20.05.98
//
// RCS version control information:
//     $Revision: 1.2 $$State: Exp $$Date: 1998/07/03 13:11:52 $
//


#include <stdio.h>  // for sprintf()

#include "psos.h"

#include "error.hpp"
#include "opnew.hpp"


static const INT32 NumberOfPartitions = 4;
// the number of partitions that are used; if you change this constant,
// you have to adapt InitMemory() as well


//**************************************************************************
// data of a partition
//**************************************************************************

// contains the data of one partition
struct SPartitionData
{
    INT32 pid;           // partition ID
    INT32 plen;          // partition length
    INT32 bsize;         // buffer size in the partition
    INT32 nbuf;          // number of buffers in the partition

    // the following entries are statistics for debugging; access to them
    // is not atomic, so their values may be slightly wrong, but are
    // accurate enough for a general view
    // note: if correct values are important, you should disable interrupts
    // during the update of the values; do not forget that operator new
    // is called from tasks and ISRs (so you can't use t_mode())

    INT32 requestCount;  // current number of buffers in use
    INT32 requestMax;    // the maximum number of used buffers
    INT32 requestTotal;  // the total number of requests
    INT32 requestError;  // number of requests that could not be fulfilled
};


//**************************************************************************
// partition table
//**************************************************************************

SPartitionData g_PartitionTable[NumberOfPartitions];
// the partition table; it is initialized in InitMemory()


//**************************************************************************
// partition table management
//**************************************************************************

static char *memoryPointer = NULL;
// this pointer contains the start of the memory block allocated in
// InitMemory(); the memory is freed in DisposeMemory()
// note: the pointer should only be used in these two functions

extern "C" {
// creates memory partition 'p' with name 'name'
// note: the function assumes that 'bsize' and 'plen' of the p-th partition
// table entry are correct, and that 'plen' of all entries from 0 to p-1
// are correct
// the function sets 'pid' and 'nbuf', and initializes the statistics entries
// assert: p < NumberOfPartitions
static void CreatePartition (INT8 p, char name[4])
{
    Assert (p < NumberOfPartitions);  // partition entry must exist

    INT32 offset = 0;  // offset of the partition from the base pointer

    // compute the offset of the partition
    for (int i=0; i<p; i++) offset += g_PartitionTable[i].plen;
    // the offset is the sum of the lengths of all previous partitions

    // create the partition
    INT32 res = pt_create (name,            // partition name
            memoryPointer+offset, memoryPointer+offset,  // address
            g_PartitionTable[p].plen,       // partition length
            g_PartitionTable[p].bsize,      // buffer size
            PT_LOCAL | PT_NODEL,            // flags
            &g_PartitionTable[p].pid,       // resulting partition ID
            &g_PartitionTable[p].nbuf);     // # buffers in the partition

    Assert (res == 0);  // partition creation must succeed

    // initialize the statistical data
    g_PartitionTable[p].requestCount = 0;
    g_PartitionTable[p].requestMax = 0;
    g_PartitionTable[p].requestTotal = 0;
    g_PartitionTable[p].requestError = 0;
}

// allocates a 1 MB memory block via malloc(), creates the partitions and
// initializes the data table
// must be called in the context of a task at the beginning of the program
// (before the first dynamic memory is allocated)
void InitMemory()
{
    memoryPointer = (char *)malloc (1024*1024);
    Assert (memoryPointer != NULL);
    // allocate 1 MB of memory

    // initialize the partition length and buffer size entries
    // name | length [bytes] | buffer size [bytes]
    // -------------------------------------------
    // PT#0 |     368128     |     64
    // PT#1 |     314880     |    256
    // PT#2 |     209920     |   1024
    // PT#3 |     155648     |   8192

    g_PartitionTable[0].plen  = 368128;
    g_PartitionTable[0].bsize = 64;

    g_PartitionTable[1].plen  = 314880;
    g_PartitionTable[1].bsize = 256;

    g_PartitionTable[2].plen  = 209920;
    g_PartitionTable[2].bsize = 1024;

    g_PartitionTable[3].plen  = 155648;
    g_PartitionTable[3].bsize = 8192;


    // create the partitions
    CreatePartition (0, "PT#0");
    CreatePartition (1, "PT#1");
    CreatePartition (2, "PT#2");
    CreatePartition (3, "PT#3");
}

// deletes the partitions and frees the memory
// note: the function assumes (but does not check!) that InitMemory() has
// been called
void DisposeMemory ()
{
    INT32 res;

    // delete the partitions
    for (int i=0; i<NumberOfPartitions; i++)
    {
        res = pt_delete (g_PartitionTable[i].pid);
        if (res != 0) Report (partition cannot be deleted);
        // this could point to a memory leak (check if res==0x2B, ERR_BUFINUSE)
    } // for

    // free the memory
    free (memoryPointer);
}


// report the statistical data of the partition table
void CheckMemory ()
{
    char report[256];  // the output string for ReportError()

    for (int p=0; p<NumberOfPartitions; p++)
    {
        sprintf (report, "partition %d:\n"
                         "   buffers in use: %ld (%ld bytes each)\n"
                         "   max. buffers in use: %ld (at most %ld)\n"
                         "   total #buffers requested: %ld\n"
                         "   unsuccessful requests: %ld\n",
                         p,
                         g_PartitionTable[p].requestCount,
                         g_PartitionTable[p].bsize,
                         g_PartitionTable[p].requestMax,
                         g_PartitionTable[p].nbuf,
                         g_PartitionTable[p].requestTotal,
                         g_PartitionTable[p].requestError);
        ReportMessage (report);
    } // for
}
} // extern

//**************************************************************************
// operators new and delete
//**************************************************************************

// requests a buffer from a partition
void *operator new (size_t s)
{
    int p;  // the partition from which the buffer is taken
    void *addr;  // the pointer to the buffer

    // find the partition from which to allocate the buffer
    for (p=0; p<NumberOfPartitions; p++)
    {
        if (g_PartitionTable[p].bsize > s)  // take buffer from here
        {
            g_PartitionTable[p].requestTotal++;  // count this request

            // allocate a buffer
            if (pt_getbuf (g_PartitionTable[p].pid, &addr) != 0)
            {
                // could not allocate -> note the error
                g_PartitionTable[p].requestError++;
                return NULL;
            }
            else
            {
                g_PartitionTable[p].requestCount++;
                if (g_PartitionTable[p].requestCount >
                    g_PartitionTable[p].requestMax)
                    g_PartitionTable[p].requestMax =
                        g_PartitionTable[p].requestCount;
                return addr;  // return the buffer pointer
            } // else
        } // if
    } // for

    Assert (False);
    // if we come here, we need a partition with bigger buffers
    return NULL;  // never reached, but compiler needs it
}

void operator delete (void *ptr)
{
    int p;  // the partition from which the buffer was taken
    INT32 offset=0;  // the offset of the partition

    // find the partition to which the buffer belongs
    for (p=0; p<NumberOfPartitions; p++)
    {
        if ((ptr > memoryPointer+offset) &&
            (ptr < memoryPointer+offset+g_PartitionTable[p].plen))
        {
             // we have found the partition to which the pointer belongs
             Assert (pt_retbuf (g_PartitionTable[p].pid, ptr) == 0);
             // returning the buffer should always succeed

             g_PartitionTable[p].requestCount--;
             return;  // we are finished
        } // if
        else offset += g_PartitionTable[p].plen;  // check next partition
    } // for

    Assert (False);
    // if we come here, the memory was not taken from a partition
}


