//
// account.hpp
//
// declaration of the accountant class that is used as a base class for
// the transmission data accountants in the system (delay and network
// accountant); it stores the transmission delay data
//
// created: 15.11.96
//
// last change: 02.09.98
//
// RCS version control information:
//     $Revision: 1.6 $$State: Exp $$Date: 1998/09/02 16:04:33 $
//


#ifndef _ACCOUNT_HPP_
#define _ACCOUNT_HPP_


#include "ci.h"
#include "psos.h"

#include "bool.hpp"
#include "distrib.hpp"


class CTransmissionList;


struct SNetworkDelay  // contains the transmission delay data
{
    SRandomData slow;  // high is not used (regular.low is used instead)
    SRandomData regular;  // this is the main interval, all data is used
    SRandomData fast;  // low is not used (regular.high is used instead)
    double probEarly;  // probability of early messages
    double probLate;  // probability of late messages

    SNetworkDelay() {
        probEarly=probLate = 0;
        slow.dist = Uniform; slow.low=slow.high=slow.mean=slow.constant = 0;
        regular = slow; fast=slow; }
};


//**************************************************************************
// class CAccountant: this class is a database for the storage of network
//      transmission delays; the delays are associated with a pair of
//      keys <SINT32, CNetworkAddress>
//      key1 denotes the primary key, key2 the secondary key
//**************************************************************************


class CAccountant
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CAccountant ();
    virtual ~CAccountant ();


    //----------------------------------------------------------------------
    //--- reset all members to their initial values
    //----------------------------------------------------------------------

    void Reset();
    // deletes all entries, restores the initial state of the class


protected:

    CTransmissionList *listOfDelays;
    // stores the transmission delays

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- access to the transmission delay data
    //----------------------------------------------------------------------

    void SetDelay (SNetworkDelay delay, SINT32 key1, SINT32 key2);
    // the delay data is stored; if there was a former entry, its data is
    // overwritten

    EBool GetDelay (SNetworkDelay *delay, SINT32 key1, SINT32 key2);
    // if the entry exists, the delay data is written into delay and True
    // is returned, else False is returned
    // assert: delay != NULL


    //----------------------------------------------------------------------
    //--- deletion of entries
    //----------------------------------------------------------------------

    EBool DeleteKey1 (SINT32 key1);
    // deletes all entries with key1; if at least one entry exists, True is
    // returned, else False is returned

    EBool DeleteKey2 (SINT32 key2);
    // deletes all entries with key2; if at least one entry exists, True is
    // returned, else False is returned


    //----------------------------------------------------------------------
    //--- initialization and destruction of the data members
    //----------------------------------------------------------------------

    virtual void Init();
    // initializes all class members to default values; a second call
    // (without Dispose() in between) has no effect

    virtual void Dispose();
    // deletes all dynamic members; a second call (without Init() in
    // between) has no effect


    //----------------------------------------------------------------------
    //--- internal management of the data entry (key1, key2, data)
    //----------------------------------------------------------------------

    void ResetDataEntry ();
    // resets the data entries to invalid values


    //----------------------------------------------------------------------
    //--- internal management of transmission data
    //----------------------------------------------------------------------

    EBool GetInformation (SINT32 key1, SINT32 key2);
    // tries to get the data for the given pair; if the data is
    // available, key1_, key2_ and data_ will be updated accordingly
    // and the function will return True;
    // if no entry was found, the method has no effect and False is returned
    
    void SetInformation ();
    // take over the information stored in key1_, key2_ and data_
    // and save it in the list
    // assert: data_ != NULL
    // assert: entry does not exist

    void SetupDataEntry (SINT32 key1, SINT32 key2);
    // search for the data of the pair; if no data exist, create
    // a new data entry for the pair and insert it into the list;
    // in any case store the new data in key1_, key2_ and data_

    
private:

    SINT32 key1_;
    SINT32 key2_;      // keys of the currently accessed entry
    SNetworkDelay *data_;     // the transmission data for the above pair

    EBool initCalled_;  // set to True by Init(), and to False by Dispose()
};


#endif

