//
// addrman.hpp
//
// address manager to store triples <networkID,nodeID,nodeAddress>;
// searches can be conducted for <networkID,nodeID> (returning the
// nodeAddress) or for <networkID,nodeAddress> (returning the nodeID)
//
// created: 03.01.97
//
// last change: 21.06.97
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/07/23 08:55:28 $
//


#ifndef _ADDRMAN_HPP_
#define _ADDRMAN_HPP_


#include "bool.hpp"


class CAddr;
class CAddressList;
class CNodeList;


enum EResult {NoAssociation = -1, AssocStored = 0, NodeIDInUse,
              NodeAddressInUse, BothInUse};
// error codes for Insert() [from AssocStored on] and for
// CheckAddress [all return values are possible]


//**************************************************************************
// class CAddressManager: this class is used as a database by the delay
//      accountant to store pairs of nodeIDs (used for identification of
//      a node by the user) and node addresses (used by the network);
//      the pairs have to be unique in a given (sub-)network
//**************************************************************************


class CAddressManager
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CAddressManager ();
    virtual ~CAddressManager ();


    //----------------------------------------------------------------------
    //--- reset
    //----------------------------------------------------------------------

    void Reset ();
    // discards all data


    //----------------------------------------------------------------------
    //--- access to the database
    //----------------------------------------------------------------------

    EResult Insert (SINT32 networkID, SINT32 nodeID, const CAddr& nodeAddress);
    // stores the <networkID,nodeID,nodeAddress> triple if possible;
    // return values:
    // NodeIDInUse:      the nodeID is already associated with another
    //                   nodeAddress (in the given network)
    // NodeAddressInUse: the nodeAddress is already associated with another
    //                   nodeID (in the given network)
    // BothInUse:        both nodeID and nodeAddress are associated with
    //                   other entries
    // AssocStored:      nodeID and nodeAddress were not associated (in
    //                   that case, they are now associated with each
    //                   other), or they already were (with each other)

    EBool FindAddress (SINT32 networkID, SINT32 nodeID, CAddr& nodeAddress);
    // returns True, if an associated network address exists; in that case,
    // the address is copied into nodeAddress; if the assoc. does not exist,
    // False is returned and nodeAddress is not modified
    // assert: nodeAddress != NULL

    EBool FindNodeID (SINT32 networkID,const CAddr& nodeAddress,SINT32 *nodeID);
    // returns True, if an associated nodeID exists; in that case, the
    // value is copied into nodeID; if the assoc. does not exist, False is
    // returned and nodeID is not modified
    // assert: nodeID != NULL


    //----------------------------------------------------------------------
    //--- management of the database
    //----------------------------------------------------------------------

    EBool DeleteAddress (SINT32 networkID, const CAddr& nodeAddress);
    // if the nodeAddress is associated with a nodeID, the association is
    // deleted and True is returned; else False is returned

    EBool DeleteNodeID (SINT32 networkID, SINT32 nodeID);
    // if the nodeID is associated with a node address, the association is
    // deleted and True is returned; else False is returned


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- initialization and destruction of the data members
    //----------------------------------------------------------------------

    virtual void Init();
    // initializes all class members to default values; if it is called
    // twice in a row (without Dispose() in between), nothing happens

    virtual void Dispose();
    // deletes all dynamic members; if it is called twice in a row
    // (without Init() in between), nothing happens


private:

    CAddressList *listOfAddresses_;  // for mapping nodeID --> nodeAddress
    CNodeList *listOfNodes_;  // for mapping nodeAddress --> nodeID
    EBool initCalled_;  // so Init/Dispose() can be called twice in a row


    //----------------------------------------------------------------------
    //--- management of the address lists
    //----------------------------------------------------------------------

    EResult CheckAddress (SINT32 networkID, SINT32 nodeID, const CAddr& nodeAddress);
    // searches for <networkID,nodeID> in the address list and for
    // <networkID,nodeAddress> in the nodeID list;
    // return values:
    // NoAssociation: both entries do not exist
    // AssocStored:   both entries exist and correlate (ie. the stored
    //                values are equal to the appropriate parameter values)
    // NodeIDInUse:   the first entry (key nodeID) exists, the 2nd does not
    // AddressInUse:  the second entry exists, the first does not
};

#endif

