//
// algbase.hpp
//
// declaration of the (abstract) base class of the clock synchronization
// algorithm
//
// created: 12.02.97
//
// last change: 22.01.99
//
// RCS version control information:
//     $Revision: 1.10 $$State: Exp $$Date: 1999/01/22 09:52:53 $
//


#ifndef _ALGBASE_HPP_
#define _ALGBASE_HPP_


#include "psos.h"

#include "bool.hpp"
#include "error.hpp"


class CApplicationDatabase;
class CAddr;
class CClockInterface;
class CClockSync;
class CNtpTime;

struct SCommand;
struct SEvent;
struct SExtendedTimestamp;
struct SParameter;
struct SRate;
struct STimestamp;
struct STimeInterval;



//**************************************************************************
// class ACClockSyncAlgorithm: abstract base class for implementing the
//      currently used clock synchronization algorithm; it can store data
//      of other nodes (time intervals, transmission delays, rate info)
//**************************************************************************

class ACClockSyncAlgorithm
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    ACClockSyncAlgorithm (CClockSync *cs);
    // assert: cs != NULL

    virtual ~ACClockSyncAlgorithm ();


    //----------------------------------------------------------------------
    //--- operators
    //----------------------------------------------------------------------

    ACClockSyncAlgorithm& operator= (const ACClockSyncAlgorithm& alg);
    // copies all interesting things (#nodes, #faulty, assocNetwork, cs)


    //----------------------------------------------------------------------
    //--- reset
    //----------------------------------------------------------------------

    virtual void Reset ();
    // discard all data


    //----------------------------------------------------------------------
    //--- suspend the synchronization
    //----------------------------------------------------------------------

    virtual void Suspend ();
    // discard the collected interval data


    //----------------------------------------------------------------------
    //--- configuration of system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the given parameter to its new value; if the parameter does
    // not exist or if the value is not valid, False is returned
    // existing parameters: NumberOfNodes, NumberOfFaultyNodes, Delta,
    // EpsilonPlus, EpsilonMinus, ThetaPlus, ThetaMinus, U, AssociatedNetwork

    virtual EBool GetParameter (SParameter& param) const;
    // returns the value of the given parameter in the appropriate field of
    // SParameter; if the parameter does not exist, False is returned


    //----------------------------------------------------------------------
    //--- execution of system commands
    //----------------------------------------------------------------------

    virtual EBool ExecuteCommand (const SCommand& command);
    // executes the command; returns True if successfull, otherwise False
    // existing commands: ConnectNetwork, DisconnectNetwork


    //----------------------------------------------------------------------
    //--- add time intervals
    //----------------------------------------------------------------------

    virtual EBool StoreTime (const STimestamp& time, double resync,
        SINT32 networkID, const CAddr& netAddress, int area, EBool gps);
    // computes the current time interval from the timestamp and puts it
    // into the database; returns True if successful, False if an entry was
    // already there; 'resync' is the (local) time until the next
    // resynchronization
    // 'gps' should be set to True if the timestamp came from a clock with
    // GPS; the parameter is ignored in this implementation


    //----------------------------------------------------------------------
    //--- add transmission delay intervals
    //----------------------------------------------------------------------

    virtual EBool StoreDelay (double localDelay, double remoteDelay,
        const SRate& remoteRate, SINT32 networkID, const CAddr& netAddress);
    // process the delays (do drift compensation for both), compute new
    // transmission delay data and store it in the database;
    // the function returns False if the upper bound for the transmission
    // delay is negative; otherwise it calls ComputeDelay() and returns its
    // value


    //----------------------------------------------------------------------
    //--- manage rate synchronization data
    //----------------------------------------------------------------------

    virtual EBool StoreRate (const SRate& remoteRate, const CNtpTime& localTime,
        const CNtpTime& remoteTime, SINT32 networkID,
        const CAddr& netAddress, int area);
    // Since no rate algorithm is implemented, the rate information of the
    // remote node is not needed and thus ignored. The function always
    // returns True.
    // overload the function in the derived class and implement the rate
    // algorithm!


    EBool GetLocalRate (SRate& data) const;
    // if ClockSync::clock != NULL, the current rate information of the
    // local clock (stored in member rate) is copied into 'data' (the data
    // is broadcast to the other nodes) and True is returned; if clock=NULL,
    // False is returned


    //----------------------------------------------------------------------
    //--- manage clock changes
    //----------------------------------------------------------------------

    virtual void ClockChanged ();
    // sets clock dependent data (i.e., the local rate) to default values
    // the local rate is set to default values (wc. bounds of the clock,
    // or global wc. bounds if GetClock() returns NULL)
    // derived classes should overload the function to implement their
    // own clock dependent initializations
    // call this function in ClockSync after the clock pointer is set to a
    // new clock


    //----------------------------------------------------------------------
    //--- manage the association nodeID <--> network address
    //----------------------------------------------------------------------

    virtual EBool SetAssociation (SINT32 networkID, const CAddr& address,
                                 SINT32 nodeID);
    // stores the association <networkID, address, nodeID>; if an
    // association already exists, the new data is ignored; overload the
    // function in the derived class to obtain different behaviour; the
    // function returns True if the triple could be stored, else False

    EBool DeleteAssociation (SINT32 networkID, SINT32 nodeID);
    // deletes the nodeAddress associated with the <networkID, nodeID> key
    // from the list of addresses; if there is no entry, False is returned,
    // else True is returned


    //----------------------------------------------------------------------
    //--- compute and set the new clock interval
    //----------------------------------------------------------------------

    virtual EBool ConvergenceFunction () =0;
    // returns True if a new interval could be computed (depends on 'nodes'
    // (= number of nodes in the system) and 'faulty' (the maximum number
    // of faulty nodes); in that case the new clock value should be
    // stored in 'newClockValue' for later use in SetNewTimestamp() and
    // 'timeValid' should be set to True; if no interval could be computed,
    // 'timeValid' should be set to False

    virtual void SetNewClockValue ();
    // if 'timeValid' is True, the function sets the clock to the new value
    // stored in 'newClockValue'; if it is False, the clock value is not
    // changed; in any case, 'timeValid' is set to False
    // assert: ClockSync::clock != NULL

    virtual void PreloadNewClockValue (double loadAt);
    // same as SetNewClockValue(), but only calls clock::PreloadTimestamp()
// DEBUG

protected:

    SRate *rate;  // the rate data of the associated clock
    CApplicationDatabase *database;  // stores the data of the other appl's

    INT32 numberOfNodes;  // number of nodes in the system
    INT32 numberOfFaultyNodes;  // the maximum(!) number of faulty nodes

    EBool timeValid;
    // True if the member below is valid and the clock should be set to
    // its value, False if the clock should not be set to it (in the next
    // call to SetNewClockValue())

    SExtendedTimestamp *newClockValue;
    // the convergence function should store its result in this member;
    // SetNewClockValue() sets the clock to the value


#ifdef PUBLIC_IN_TEST
public:
#endif

    //----------------------------------------------------------------------
    //--- compute the new transmission delay data (interval)
    //----------------------------------------------------------------------

    virtual EBool ComputeNewDelay (double dmin, double dmax,
        SINT32 networkID, const CAddr& netAddress);
    // the function is called by StoreDelay() to add new transmission delays
    // (given by the bounds (dmin, dmax)) to the database;
    // the new delta is the mean of any stored delta and the mean of dmin,
    // dmax; epsilonPlus, epsilonMinus are chosen so that the resulting
    // interval encompasses both any old interval stored and (dmin, dmax);
    // the function always returns True
    // overload the function in the derived class to implement a different
    // strategy for the computation of the delay data


    //----------------------------------------------------------------------
    //--- internal handling of the CSA identification
    //----------------------------------------------------------------------

    void GetOwnID (SINT32& nodeID, int& area) const;
    // returns the nodeID and area of ClockSync


    //----------------------------------------------------------------------
    //--- internal handling of the associated network
    //----------------------------------------------------------------------

    SINT32 GetAssociatedNetwork () const;
    // returns the ID of the associated network of ClockSync, or
    // NoAccount if no associated network exists

    EBool SetAssociatedNetwork (SINT32 networkID);
    // if 'networkID' is in the list of connected networks of ClockSync,
    // the function sets the associated network to 'networkID' and returns
    // True; otherwise, False is returned and the function has no effect
    // the change is reported to EvalSys with the event 'ChangeAssociation'


    //----------------------------------------------------------------------
    //--- internal handling of the resynchronization mode of CS
    //----------------------------------------------------------------------

    EBool GetResynchronizationStatus() const;
    // returns the value of ClockSync::resynchronizing

    void SetResynchronizationStatus (EBool status) const;
    // sets the value of ClockSync::resynchronizing to 'status'


    //----------------------------------------------------------------------
    //--- internal handling of the clock
    //----------------------------------------------------------------------

    CClockInterface *GetClock () const;
    // returns the clock of ClockSync


    //----------------------------------------------------------------------
    //--- internal handling of synchronization round duration data
    //----------------------------------------------------------------------

    double GetRoundPeriod () const;
    // returns the round period duration of ClockSync

    double GetRatePeriod () const;
    // returns the rate period of ClockSync
    // returns 0 if there is no rate synchronization

    double GetDelayCompensation() const;
    // returns the delay compensation duration of ClockSync

    double GetComputationDelayCompensation() const;
    // returns the computation delay compensation duration of ClockSync


    //----------------------------------------------------------------------
    //--- internal event reports
    //----------------------------------------------------------------------

    virtual void ReportEvent (SEvent *event);
    // pass the event to ClockSync (set module=ClockSync, type=UserEvent)
    // assert: event != NULL

    virtual void ReportInterval (EBool input, const STimeInterval& interval);
    // creates an event of type SIntervalEvent, copies 'interval' into it,
    // sets the event module to 'ClockSync', the type to 'IntervalEvent' and
    // its member 'input' to the parameter 'input' (which should be True if
    // the interval is used as input to the CV, False if it is its result)

    virtual void ReportRate (EBool input, const SRate& rate, double refPoint);
    // creates an event of type SRateEvent, copies 'rate' and 'refPoint'
    // into it, sets the event module to 'ClockSync', the type to
    // 'RateEvent' and its member 'input' to the parameter 'input'
    // (which should be True if the rate is used as input to the rate
    // algorithm, False if it is its result)


    //----------------------------------------------------------------------
    //--- error reports
    //----------------------------------------------------------------------

    virtual int ReportError (EReportType type, char *file, int line,
                             char *expression);
    // calls the error report function of ClockSync


    //----------------------------------------------------------------------
    //--- create/destroy the members that are modified during reset
    //----------------------------------------------------------------------

    virtual void Init ();
    // creates the members that are modified during a reset; a second call
    // (without Dispose() in between) has no effect; sets #nodes=1,
    // #faulty=0

    virtual void Dispose ();
    // deletes the dynamic members created by Init(); a second call
    // (without Init() in between), has no effect


private:

    CClockSync *cs_;  // the associated clocksync application
    EBool initCalled_;  // set to True by Init(), to False by Dispose()
    SINT32 assocNetwork_;  // the associated network
};


#endif


