//
// algrate.hpp
//
// declaration of the "clock rate synchronization algorithm"
//
// created: 07.10.98, Klaus Schossmaier
//
// last change: 20.11.98, Klaus Schossmaier
//
// last revision: 29.01.99
//
// RCS version control information:
//     $Revision: 1.6 $$State: Exp $$Date: 1999/01/29 10:49:01 $
//


#ifndef _ALGRATE_HPP_
#define _ALGRATE_HPP_


#include "algbase.hpp"
#include "alggen.hpp"
#include "database.hpp"
#include "time.hpp"


// additions for relative rate measurement
struct SRateTime : SRate    
{
    CNtpTime localTime;  
    CNtpTime remoteTime;    
    double ref;
};


//**************************************************************************
// class CRate: provides the Rate algorithm
//**************************************************************************

class CRate : public ACClockSyncAlgorithm
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CRate (CClockSync *cs);
    // assert: cs != NULL


    //----------------------------------------------------------------------
    //--- compute and set the new clock interval
    //----------------------------------------------------------------------

    virtual EBool ConvergenceFunction ();
    // returns True if a new interval could be computed (depends on the
    // number of nodes in the system and the maximum number of faulty
    // nodes); in that case 'newTime' is set to the new clock value; if
    // no valid interval could be computed, False is returned and 'newTime'
    // is not changed


    //----------------------------------------------------------------------
    //--- add time intervals
    //----------------------------------------------------------------------

    virtual EBool StoreTime (const STimestamp& time, double resync,
        SINT32 networkID, const CAddr& netAddress, int area, EBool gps);
    // if CS:resync is True, the time is not stored; so the algorithm
    // only starts working after resync is False; then the time is
    // passed on to StoreTime() of the base class


    //----------------------------------------------------------------------
    //--- add transmission delay intervals
    //----------------------------------------------------------------------

    virtual EBool StoreDelay (double localDelay, double remoteDelay,
        const SRate& remoteRate, SINT32 networkID, const CAddr& netAddress);
    // counts to N before setting CS::resync to False, calls StoreDelay
    // of base class to process delays


    //----------------------------------------------------------------------
    //--- add rate intervals
    //----------------------------------------------------------------------

    virtual EBool StoreRate (const SRate& remoteRate, 
        const CNtpTime& localTime, const CNtpTime& remoteTime, 
        SINT32 networkID, const CAddr& netAddress, int area);
    // called when rate intervals are arriving my node
    // do the preprocessing and fill the database with remote rate intervals


    //----------------------------------------------------------------------
    //--- manage clock changes
    //----------------------------------------------------------------------

    virtual void ClockChanged();
    // computes initialization data (clock deterioration, accuracy) for
    // the clock, and some clock dependent parameters needed by the algorithm


    //----------------------------------------------------------------------
    //--- creation function of class CRate
    //----------------------------------------------------------------------

    static ACClockSyncAlgorithm *CreateRate (CClockSync *cs);
    // creates and returns a new object of type CRate


protected:

    //----------------------------------------------------------------------
    //--- set members that are modified during reset
    //----------------------------------------------------------------------

    virtual void Init ();
    // sets cntDelay=0, resync. status to True, calls Init() of base class


private:

    int cntDelay;  
    // number of delay measurements

    SINT32 nodeID;
    int area;
    // for identification

    double piHMinus;
    double piHPlus;
    // maximal precision interval

    double piOwnMinus; 
    double piOwnPlus;
    // own precision interval

    double ups;
    // maximal state correction

    double sigma;
    // oscillator stability

    double gammaHMinus;
    double gammaHPlus;
    // maximal consonance interval

    double gammaOwnMinus; 
    double gammaOwnPlus;
    // own consonance interval

    double thetaMax;
    // maximal rate correction
   
    EBool rateRound;
    // indicates if this round is also for rate synchronization
    // assumes that all nodes are in synch! 
  
#ifdef PUBLIC_IN_TEST
public:
#endif

    //----------------------------------------------------------------------
    //--- state algorithm
    //----------------------------------------------------------------------

    EBool State (CNtpTime& point, double& alpha_minus, double& alpha_plus,
        SINT32 networkID);
    // computes the new clock interval for the network; returns True if the
    // computation was successful, False otherwise

    //----------------------------------------------------------------------
    //--- rate algorithm
    //----------------------------------------------------------------------

    EBool Rate (double& multStep, double& theta_minus, double& theta_plus, 
        SINT32 networkID);
    // compute the new coupling factor (i.e. register STEP)
    // returns True if the computation was successful, False otherwise
};

#endif






