//
// bitmask.hpp
//
// declaration of a class that provides an array of bits that can be set
// using the place of the bit (instead of the value, which is a power of
// two); useful for enumerations
//
// created: 23.01.97
//
// last change: 02.02.97
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 09:57:06 $
//


#ifndef _BITMASK_HPP_
#define _BITMASK_HPP_

#include "constant.hpp"


//**************************************************************************
// class CBitMask: allows access to the bits specifying the place instead
//    of the value (places start with 0)
//**************************************************************************

class CBitMask
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CBitMask (INT32 aSize = ArrayAllocationSize);
    // allocate enough memory to store aSize bits;
    // assert: aSize > 0

    CBitMask (void *mask, INT32 size);
    // allocate the memory needed to store 'size' bytes of mask;
    // assert: size > 0, mask != NULL

    CBitMask (const CBitMask& bitmask);
    // copy the information in bitmask

    ~CBitMask ();


    //----------------------------------------------------------------------
    //--- set (store a mask) and reset (set all bits to zero)
    //----------------------------------------------------------------------

    void Set (const void *mask, INT32 size);
    // discard all old data, store 'size' bytes of the mask;
    // assert: size > 0, mask != NULL

    void Reset (INT32 aSize = ArrayAllocationSize);
    // set all bits to zero, keep enough memory to store aSize bits;
    // assert: aSize > 0


    //----------------------------------------------------------------------
    //--- operators
    //----------------------------------------------------------------------

    int operator& (const INT32 place) const;
    CBitMask& operator|= (const INT32 place);
    CBitMask& operator^= (const INT32 place);


private:

    INT8 *bitmask_;
    INT32 size_;


    //----------------------------------------------------------------------
    //--- compute the index and offset of a place
    //----------------------------------------------------------------------

    INT32 GetIndex (INT32 place) const;
    // returns the array index of the INT8 in which place is located

    INT32 GetOffset (INT32 place) const;
    // returns the offset in the INT8 in which place is located


    //----------------------------------------------------------------------
    //--- set/get a bit
    //----------------------------------------------------------------------

    void SetBit (INT32 place);
    // adds the bit to the bitmask

    int GetBit (INT32 place) const;
    // returns 0 if the bit is not set, otherwise != 0


    //----------------------------------------------------------------------
    //--- internal array management
    //----------------------------------------------------------------------

    void Allocate (INT32 size);
    // assumes that the array is deleted, allocates a new one with size
    // 'size', fills the array with 0

    void ReAllocate (INT32 size);
    // deletes the old array, allocates a new with size 'size'; stores the
    // new size in size_, and fills the array with 0
};


#endif


