//
// classman.hpp
//
// provides facilities for dynamic installation of network and algorithm
// classes
//
// created: 02.07.97
//
// last change: 18.02.98
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 09:57:06 $
//


#ifndef _CLASSMAN_HPP_
#define _CLASSMAN_HPP_


#include <stdlib.h>

#include "psos.h"

#include "general.hpp"


//**************************************************************************
// class CClassManager: provides dynamic installation facilities for later
//      additions of (network and algorithm) classes
//      you need to specify an identifier and a creation function which
//      should create and return a new object of the class you want to
//      install
//**************************************************************************

class CClassManager
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CClassManager (INT32 arraySize);
    // create an array with the given size
    // assert: arraySize > 0

    virtual ~CClassManager ();
    // delete the array


    //----------------------------------------------------------------------
    //--- insert/return function pointers
    //----------------------------------------------------------------------

    void SetFunction (INT32 id, INT32 *functionPointer);
    // insert a function with the identifier 'id'
    // assert: id < size

    INT32 *GetFunction (INT32 id);
    // return the function that was stored for the identifier 'id', or
    // NULL if no function was stored


private:

    INT32 **createFunction_;  // the array of function pointers
    INT32 size_;  // the size of the array
};



//**************************************************************************
// class CAlgorithmManager: provides dynamic installation facilities for
//      clock synchronization algorithm classes
//      the creation function must have the prototype
//          ACClockSyncAlgorithm *YourCreateFunction (CClockSync *)
//**************************************************************************


class ACClockSyncAlgorithm;
class CClockSync;


typedef ACClockSyncAlgorithm *(*AlgFunction)(CClockSync *);
// the prototype of an algorithm creation function


class CAlgorithmManager : public CClassManager
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CAlgorithmManager () : CClassManager (EAlgorithmTypeCount) {}
    // create an array which can hold all algorithm types


    //----------------------------------------------------------------------
    //--- insert/return function pointers
    //----------------------------------------------------------------------

    // insert a function with the type 'algorithm'
    void SetFunction (EAlgorithmType algorithm, AlgFunction functionPointer)
      { CClassManager::SetFunction (algorithm, (INT32*)functionPointer); }

    // return the function that was stored for the type 'algorithm', or
    // NULL if no function was stored
    AlgFunction GetFunction (EAlgorithmType algorithm)
      { return (AlgFunction) CClassManager::GetFunction (algorithm); }
};



//**************************************************************************
// class CNetworkManager: provides dynamic installation facilities for
//      network classes
//      the creation function must have the prototype
//          ACNetwork *YourCreateFunction (CControllingObject *, SINT32)
//**************************************************************************
/**/// note: the class is only needed in pure simulation and may be put
// into some other file in the future


class ACNetwork;
class CControllingObject;


typedef ACNetwork *(*NetworkFunction)(CControllingObject *, SINT32);
// the prototype of a network creation function


class CNetworkManager : public CClassManager
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CNetworkManager () : CClassManager (ENetworkTypeCount) {}
    // create an array which can hold all network types


    //----------------------------------------------------------------------
    //--- insert/return function pointers
    //----------------------------------------------------------------------

    // insert a function with the type 'network'
    void SetFunction (ENetworkType network, NetworkFunction functionPointer)
      { CClassManager::SetFunction (network, (INT32*)functionPointer); }

    // return the function that was stored for the type 'algorithm', or
    // NULL if no function was stored
    NetworkFunction GetFunction (ENetworkType network)
      { return (NetworkFunction) CClassManager::GetFunction (network); }
};


#endif


