//
// clock_sw.hpp
//
// declaration of the base class used for accessing the software clock
//
// created: 19.02.97
//
// last change: 24.09.98
//
// RCS version control information:
//     $Revision: 1.5 $$State: Exp $$Date: 1998/09/24 12:33:57 $
//


#ifndef _CLOCK_SW_HPP_
#define _CLOCK_SW_HPP_


#include "clockbas.hpp"
#include "clockcfg.hpp"
#include "error.hpp"


class CClockInterrupt;
class CGpsInterrupt;
class CInterruptList;
class CNtpTime;

struct SInternalRegisters;


//**************************************************************************
// class CSoftwareClock: interface to the software clock
//**************************************************************************

class CSoftwareClock : public CClock
{
    friend void MakeSystemSnapshot();
    // the software implementation calls HwSnapshot()

public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CSoftwareClock (double frequency, SINT32 nodeID, int area);
    // assert: frequency > 0 (through the ctor of CClock)

    virtual ~CSoftwareClock ();
    // assert: UTCINTEN1,2 == 0


    //----------------------------------------------------------------------
    //--- configuration of system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the given parameter to its new value; if the parameter does
    // not exist or if the value is not valid, False is returned
    // known params: ClockDrift

    virtual EBool GetParameter (SParameter& param) const;
    // returns the value of the given parameter in the appropriate field of
    // SParameter; if the parameter does not exist, False is returned


    //----------------------------------------------------------------------
    //--- read access functions that have to be simulated
    //----------------------------------------------------------------------

    // LTU (Local Time Unit)
    virtual INT32 GetTSGETL();
    virtual INT32 GetAMORTTIMEGETL();

    // ACU (Accuracy Unit)
    virtual INT32 GetSTATEPGETL();
    virtual INT32 GetALPHAPGETL();
    virtual INT32 GetALPHANGETL();

    // ITU (Interrupt Unit)
    virtual INT32 GetUTCSTAT1();
    virtual INT32 GetUTCCONF1();
    virtual INT32 GetUTCCONF2();


    //----------------------------------------------------------------------
    //--- write access functions that have to be simulated
    //----------------------------------------------------------------------

    // LTU (Local Time Unit)
    virtual void SetMSSET(INT32 val);
    virtual void SetNTPSET(INT32 val);
    virtual void SetSTEPAMORT(INT32 val);
    virtual void SetSTARTAMORT(INT32 val);

    // ACU (Accuracy Unit)
    virtual void SetALPHAPNSET(INT32 val);

    // CLU (Clock Unit)
    virtual void SetSWRESET(INT32 val);

    // ITU (Interrupt Unit)
    virtual void SetUTCINTEN1(INT32 val);
    virtual void SetUTCINTEN2(INT32 val);
    virtual void SetUTCINTCLEAR1(INT32 val);
    virtual void SetUTCINTCLEAR2(INT32 val);

    // SNU (Snapshot Unit)
    virtual void SetSWSNAP(INT32 val);

    // APU (Application Unit)
    virtual void SetAPPDUTYH(INT32 val);
    virtual void SetAPPDUTYL(INT32 val);

    // SSU1-6 (Synchronization Subnet Unit 1-6)
    virtual void SetDUTYAH (ESSUnit ssu, INT32 val);
    virtual void SetDUTYAL (ESSUnit ssu, INT32 val);
    virtual void SetDUTYBH (ESSUnit ssu, INT32 val);
    virtual void SetDUTYBL (ESSUnit ssu, INT32 val);


    //----------------------------------------------------------------------
    //--- GPS 1PPS handling; FOR CGpsInterrupt ONLY!
    //----------------------------------------------------------------------

    void CaptureTime (EGpsUnit gpu);
    // copies the new clock value into MSGPS and TSGPS; sets 'gpu_' to 'gpu'
    // assert: 0 <= gpu < 3


    //----------------------------------------------------------------------
    //--- GPS interrupt handling; FOR CSoftwareGps ONLY!
    //----------------------------------------------------------------------

    CGpsInterrupt *GetGpsInterrupt (EGpsUnit gpu);
    // returns the GPS interrupt process of the given GPU; if the process
    // does not exist, it is created
    // assert: gpu < EGpsUnitCount

    EGpsUnit GetGpsUnit();
    // returns the GPS unit that has last called CaptureTime()


protected:

    //----------------------------------------------------------------------
    //--- initialization/destruction
    //----------------------------------------------------------------------

    virtual void Init();
    virtual void Dispose();


    //----------------------------------------------------------------------
    //--- error reports
    //----------------------------------------------------------------------

    virtual int ReportError (EReportType type, char *file, int line,
                             char *expression);


private:

    SInternalRegisters *internals_;  // internal clock registers
    double assocSimTime_;  // last time of clock update
    double fOsc_;
    // real oscillator frequency (as opposed to the ideal frequency stored
    // in CClock; fOsc can be manipulated through the ClockDrift param)

    CInterruptList *listOfInterrupts_;
    SINT32 nodeID_;  // ID of the node (needed for the interrupt processes)
    int area_;  // area of the clock (for setting the interrupt vector)

    EGpsUnit gpu_;  // the GPU that has last captured the clock value

#ifdef PUBLIC_IN_TEST
public:
#endif
    enum EDutyTimer {
        DutyTimerAPP   = 29,  // APPDUTY in APU
        DutyTimerASSU1 = 52,  // duty timer A in SSU1
        DutyTimerASSU2 = 53,
        DutyTimerASSU3 = 54,
        DutyTimerASSU4 = 55,
        DutyTimerASSU5 = 56,
        DutyTimerASSU6 = 57,
        DutyTimerBSSU1 = 58,  // duty timer B in SSU1
        DutyTimerBSSU2 = 59,
        DutyTimerBSSU3 = 60,
        DutyTimerBSSU4 = 61,
        DutyTimerBSSU5 = 62,
        DutyTimerBSSU6 = 63
    };  // all implemented duty timers


    //----------------------------------------------------------------------
    //--- snapshots
    //----------------------------------------------------------------------

    void HwSnapshot();
    // make a hardware snapshot, generate an interrupt (if enabled)

    void SwSnapshot();
    // make a software snapshot (captures all dynamically changing
    // registers)


    //----------------------------------------------------------------------
    //--- internal access functions for status/configuration bits
    //----------------------------------------------------------------------

    void SetSignal (INT8 signalBit);
    // sets the bit in UTCSTAT1 (0..31) or UTCSTAT2 (32..63)
    // assert: signalBit < 64

    EBool GetSignal (INT8 signalBit);
    // returns True if the bit in UTCSTAT1 (0..31) or UTCSTAT2 (32..63) is
    // set, otherwise False
    // assert: signalBit < 64

    void ClearSignal (INT8 signalBit);
    // clears the bit in UTCSTAT1 (0..31) or UTCSTAT2 (32..63)
    // assert: signalBit < 64


    EBool GetPolarity (INT8 bitnumber);
    // returns True if the bit in UTCCONF1 (0..31) or UTCCONF2 (32..63) is
    // set, otherwise False
    // assert: bitnumber < 64


    //----------------------------------------------------------------------
    //--- internal clock update functions
    //----------------------------------------------------------------------

    void Update();
    // update the clock registers

    void UpdateNtp (INT32 step, const CNtpTime& cycles);
    // add the step-value for 'cycles' cycles to the NTPTIME, take STEPLOW
    // as the [-52,-59] portion of step

    void UpdateEnvelopes (EBool pure, const CNtpTime& cycles);
    // if pure=True, LAMBDAP(N)PUREACT is added to ALPHAP(N) for 'cycles'
    // cycles; if pure=False, LAMBDAP(N)AMORT is used, and if AMORTTIMER is
    // equal to AMORTTIMESET, the envelopes are realigned using STATEP(N)SET

    void SetUpperLimit (EBool positive);
    // sets ALPHAP(N) and NALPHAP(N) to the upper limit (which is 0:1...1);
    // if 'positive' is True, the P-envelope is used, else the N-envelope
    // is set
/**/// right now, no interrupt INTPLALPHAP(N) is raised

    void SetLowerLimit (EBool positive);
    // sets ALPHAP(N) and NALPHAP(N) to the lower limit (which is 1:0...0);
    // if 'positive' is True, the P-envelope is used, else the N-envelope
    // is set
/**/// right now, no interrupt INTNLALPHAP(N) is raised

    void GetEnvelope (EBool positive, INT32& alpha, INT16& nalpha);
    // reads internals_->ALPHAP(N) into 'alpha' and internals_->NALPHAP(N)
    // into 'nalpha'; if 'positive' is True, the P-envelope is used, else
    // the N-envelope is read

    void SetEnvelope (EBool positive, INT32 alpha, INT16 nalpha);
    // stores 'alpha' into internals_->ALPHAP(N) and 'nalpha' into
    // internals_->NALPHAP(N); if 'positive' is True, the P-envelope is
    // used, else the N-envelope is modified

    double GetRealDelay (double delta);
    // returns the global time (in seconds) that will pass until the
    // clock time "delta" (in seconds) has passed

    void GetClockCycles (double delta, CNtpTime& cycles);
    // computes the number of clock cycles that occur during a duration of
    // 'delta' real seconds and stores it in 'cycles'

    INT32 GetAmortCycles (const CNtpTime& cycles);
    // returns the number of amortization cycles in the next 'cycles' clock
    // cycles

    double GetRate (INT32 step);
    // returns the real clock rate (clock seconds per second) if step and
    // STEPLOW is used


    //----------------------------------------------------------------------
    //--- internal interrupt handling
    //----------------------------------------------------------------------

    virtual void EnableInterrupt (INT8 interruptBit);
    // enable the interrupt; interruptBit is in the range from 0..63
    // (0..31: UTCINTEN1; 32..63: UTCINTEN2)

    virtual void DisableInterrupt (INT8 interruptBit);
    // if the interrupt of the source was enabled, it is turned off; else
    // nothing happens; interruptBit is in the range from 0..63
    // (0..31: UTCINTEN1; 32..63: UTCINTEN2)


    CClockInterrupt *CreateInterrupt (INT8 interruptBit);
    // create and return an interrupt process; if an interrupt with the ID
    // already exists, that interrupt is returned instead
    // a new interrupt is put into the interrupt list, enabled, and its
    // vector is set according to its ID
    // assert: 0 <= interruptBit <= 63

    void RescheduleInterrupts ();
    // reschedules all interrupt processes after the clock state has been
    // changed; processes that were not scheduled will be ignored

    void ScheduleInterrupt (INT8 interruptBit);
    void ScheduleInterrupt (CClockInterrupt *clockint);
    // (re-)schedule the process so that the interrupt occurs at its stored
    // scheduleTime; the function assumes that the clock is up to date
    // assert: clockint != NULL

    void DescheduleInterrupt (INT8 interruptBit);
    // removes the interrupt process from the scheduler queue; if the
    // process does not exist or is not scheduled, nothing happens


    //----------------------------------------------------------------------
    //--- internal timestamp/accuracy access
    //----------------------------------------------------------------------

    double GetCurrentClockValue();
    // returns the current clock value in seconds

    INT32 GetAccuracy();
    // returns the current accuracy (from INTPBUS)

    void SetOverflow (int overflow);
    // if overflow != 0, internals_->overflow is set to True (and the over-
    // flow bits are removed from the macrostamp); else it is set to False


    //----------------------------------------------------------------------
    //--- internal duty timer handling
    //----------------------------------------------------------------------

    void ModifiedDutyTimer (EDutyTimer dutyBit, const CNtpTime& duty);
    // if the duty timer (duty.ms = DutyHigh, duty.ts = DutyLow) is enabled
    // the clock interrupt process is scheduled accordingly; else the
    // process is disabled
    // if duty timer is enabled, assert: 0 <= dutyBit <= 63

#ifndef NDEBUG
public:
    //----------------------------------------------------------------------
    //--- debugging
    //----------------------------------------------------------------------

    virtual void DumpRegisters();
    // prints the contents of all registers of the UTCSU to cout; in the
    // header, it prints the current simulation time (split into an integer
    // and a fractional part to avoid rounding errors because of limited
    // output precision), the difference between the simulation time and
    // the current associated time, and the current clock time (also split
    // into an integral and a fractional part)
#endif
};

#endif


