//
// clockbas.hpp
//
// declaration of the base class used for accessing the clock
//
// created: 19.02.97
//
// last change: 24.10.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _CLOCKBAS_HPP_
#define _CLOCKBAS_HPP_


#include "bool.hpp"
#include "clockcfg.hpp"
#include "utcsu.hpp"


struct SParameter;


//**************************************************************************
// class CClock: interface to the clock (both hardware and software);
//      provides functions for accessing all registers of the clock; could
//      be used as a hardware clock, but a derived class is used for that
//**************************************************************************

class CClock
{
public:

virtual void DumpRegisters() {}
// DEBUG


    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CClock (double f, SUtcsuRead *read, SUtcsuWrite *write);
    // assert: f > 0; read,write != NULL

    virtual ~CClock() {}


    //----------------------------------------------------------------------
    //--- configuration of system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the given parameter to its new value; if the parameter does
    // not exist or if the value is not valid, False is returned
    // known params: ClockGranularity, AccuracyTransmissionLoss

    virtual EBool GetParameter (SParameter& param) const;
    // returns the value of the given parameter in the appropriate field of
    // SParameter; if the parameter does not exist, False is returned


    //----------------------------------------------------------------------
    //--- access to the oscillator frequency
    //----------------------------------------------------------------------

    double GetFrequency () const { return oscillatorFrequency_; }


    //----------------------------------------------------------------------
    //--- access to the granularities
    //----------------------------------------------------------------------

    double GetGranularity () const { return granularity_; }
    double GetGranularityAcc () const { return granularityAcc_; }
    double GetGranularityDT () const { return granularityDT_; }
    double GetGranularityOsc () const { return granularityOsc_; }
    double GetGranularityS () const { return granularityS_; }


    //----------------------------------------------------------------------
    //--- read access functions
    //----------------------------------------------------------------------

    // note: don't make the functions 'const', since they are not constant
    // in the derived class (software clock) and if the function is const
    // here, and not const in CSoftwareClock, they are seen as different
    // functions (the 'virtual' keyword will be "ignored")

    // LTU (Local Time Unit)
    virtual INT32 GetMSGET() { return utcsuRead_->MSGET; }
    virtual INT32 GetTSGET() { return utcsuRead_->TSGET; }
    virtual INT32 GetUSGET() { return utcsuRead_->USGET; }
    virtual INT32 GetTSGETL() { return utcsuRead_->TSGETL; }
    virtual INT32 GetNSGET() { return utcsuRead_->NSGET; }
    virtual INT32 GetAMORTTIMEGETL() { return utcsuRead_->AMORTTIMEGETL; }
    virtual INT32 GetAMORTTIMEGET() { return utcsuRead_->AMORTTIMEGET; }

    // ACU (Accuracy Unit)
    virtual INT32 GetSTATEPGETL() { return utcsuRead_->STATEPGETL; }
    virtual INT32 GetSTATEPGET() { return utcsuRead_->STATEPGET; }
    virtual INT32 GetALPHAPGETL() { return utcsuRead_->ALPHAPGETL; }
    virtual INT32 GetALPHANGETL() { return utcsuRead_->ALPHANGETL; }
    virtual INT32 GetALPHAPGET() { return utcsuRead_->ALPHAPGET; }
    virtual INT32 GetALPHANGET() { return utcsuRead_->ALPHANGET; }
    virtual INT32 GetNALPHAPGET() { return utcsuRead_->NALPHAPGET; }

    // ITU (Interrupt Unit)
    virtual INT32 GetIDENTIFICATION() { return utcsuRead_->IDENTIFICATION; }
    virtual INT32 GetUTCSTAT1() { return utcsuRead_->UTCSTAT1; }
    virtual INT32 GetUTCSTAT2() { return utcsuRead_->UTCSTAT2; }
    virtual INT32 GetUTCCONF1() { return utcsuRead_->UTCCONF1; }
    virtual INT32 GetUTCCONF2() { return utcsuRead_->UTCCONF2; }
    virtual INT32 GetUTCINTSTAT1() { return utcsuRead_->UTCINTSTAT1; }
    virtual INT32 GetUTCINTSTAT2() { return utcsuRead_->UTCINTSTAT2; }

    // SNU (Snapshot Unit)
    virtual INT32 GetMSSNU() { return utcsuRead_->MSSNU; }
    virtual INT32 GetTSSNU() { return utcsuRead_->TSSNU; }
    virtual INT32 GetACCPSNU() { return utcsuRead_->ACCPSNU; }

    // APU (Application Unit)
    virtual INT32 GetMSAPP() { return utcsuRead_->MSAPP; }
    virtual INT32 GetTSAPP() { return utcsuRead_->TSAPP; }
    virtual INT32 GetACCPAPP() { return utcsuRead_->ACCPAPP; }
    virtual INT32 GetMSAPPX() { return utcsuRead_->MSAPPX; }
    virtual INT32 GetTSAPPX() { return utcsuRead_->TSAPPX; }
    virtual INT32 GetACCPAPPX() { return utcsuRead_->ACCPAPPX; }

    // GPU1-3 (GPS Unit 1-3)
    virtual INT32 GetMSGPS(EGpsUnit gpu) { return utcsuRead_->gpu[gpu].MSGPS; }
    virtual INT32 GetTSGPS(EGpsUnit gpu) { return utcsuRead_->gpu[gpu].TSGPS; }

    // SSU1-6 (Synchronization Subnet Unit 1-6)
    virtual INT32 GetMSRCV(ESSUnit ssu) { return utcsuRead_->ssu[ssu].MSRCV; }
    virtual INT32 GetTSRCV(ESSUnit ssu) { return utcsuRead_->ssu[ssu].TSRCV; }
    virtual INT32 GetMSXMT(ESSUnit ssu) { return utcsuRead_->ssu[ssu].MSXMT; }
    virtual INT32 GetTSXMT(ESSUnit ssu) { return utcsuRead_->ssu[ssu].TSXMT; }
    virtual INT32 GetACCPXMT(ESSUnit ssu) { return utcsuRead_->ssu[ssu].ACCPXMT; }
    virtual INT32 GetACCNPLXMT(ESSUnit ssu) { return utcsuRead_->ssu[ssu].ACCNPLXMT; }
    virtual INT32 GetACCNPHXMT(ESSUnit ssu) { return utcsuRead_->ssu[ssu].ACCNPHXMT; }


    //----------------------------------------------------------------------
    //--- write access functions
    //----------------------------------------------------------------------

    // LTU (Local Time Unit)
    virtual void SetMSSET(INT32 val) { utcsuWrite_->MSSET = val; }
    virtual void SetTSSET(INT32 val) { utcsuWrite_->TSSET = val; }
    virtual void SetUSSET(INT32 val) { utcsuWrite_->USSET = val; }
    virtual void SetNTPSET(INT32 val) { utcsuWrite_->NTPSET = val; }
    virtual void SetSTEPAMORT(INT32 val) { utcsuWrite_->STEPAMORT = val; }
    virtual void SetSTEPPURE(INT32 val) { utcsuWrite_->STEPPURE = val; }
    virtual void SetAMORTTIMESET(INT32 val) { utcsuWrite_->AMORTTIMESET = val; }
    virtual void SetSTARTAMORT(INT32 val) { utcsuWrite_->STARTAMORT = val; }

    // ACU (Accuracy Unit)
    virtual void SetLAMBDAPPURE(INT32 val) { utcsuWrite_->LAMBDAPPURE = val; }
    virtual void SetLAMBDANPURE(INT32 val) { utcsuWrite_->LAMBDANPURE = val; }
    virtual void SetSTATEPSET(INT32 val) { utcsuWrite_->STATEPSET = val; }
    virtual void SetSTATENSET(INT32 val) { utcsuWrite_->STATENSET = val; }
    virtual void SetBOUNDP(INT32 val) { utcsuWrite_->BOUNDP = val; }
    virtual void SetALPHASET(INT32 val) { utcsuWrite_->ALPHASET = val; }
    virtual void SetALPHAPNSET(INT32 val) { utcsuWrite_->ALPHAPNSET = val; }

    // CLU (Clock Unit)
    virtual void SetSWRESET(INT32 val)
      { utcsuWrite_->SWRESET = val; utcinten1_ = 0; utcinten2_ = 0;}

    // ITU (Interrupt Unit)
    virtual void SetUTCCONF1(INT32 val) { utcsuWrite_->UTCCONF1 = val; }
    virtual void SetUTCCONF2(INT32 val) { utcsuWrite_->UTCCONF2 = val; }
    virtual void SetUTCINTEN1(INT32 val) { utcsuWrite_->UTCINTEN1 = val; }
    virtual void SetUTCINTEN2(INT32 val) { utcsuWrite_->UTCINTEN2 = val; }
    virtual void SetUTCINTCLEAR1(INT32 val) { utcsuWrite_->UTCINTCLEAR1 = val; }
    virtual void SetUTCINTCLEAR2(INT32 val) { utcsuWrite_->UTCINTCLEAR2 = val; }

    // SNU (Snapshot Unit)
    virtual void SetSWSNAP(INT32 val) { utcsuWrite_->SWSNAP = val; }
    virtual void SetDUTYH(INT32 val) { utcsuWrite_->DUTYH = val; }
    virtual void SetDUTYL(INT32 val) { utcsuWrite_->DUTYL = val; }

    // APU (Application Unit)
    virtual void SetAPPDUTYH(INT32 val) { utcsuWrite_->APPDUTYH = val; }
    virtual void SetAPPDUTYL(INT32 val) { utcsuWrite_->APPDUTYL = val; }
    virtual void SetAPPSEL(INT32 val) { utcsuWrite_->APPSEL = val; }
    virtual void SetAPPCLEAR(INT32 val) { utcsuWrite_->APPCLEAR = val; }

    // SSU1-6 (Synchronization Subnet Unit 1-6)
    virtual void SetDUTYAH(ESSUnit ssu, INT32 val) { utcsuWrite_->ssu[ssu].DUTYAH=val; }
    virtual void SetDUTYAL(ESSUnit ssu, INT32 val) { utcsuWrite_->ssu[ssu].DUTYAL=val; }
    virtual void SetDUTYBH(ESSUnit ssu, INT32 val) { utcsuWrite_->ssu[ssu].DUTYBH=val; }
    virtual void SetDUTYBL(ESSUnit ssu, INT32 val) { utcsuWrite_->ssu[ssu].DUTYBL=val; }


    //----------------------------------------------------------------------
    //--- enable/disable single interrupt bits
    //----------------------------------------------------------------------

    virtual EBool EnableInterrupt (int area, INT32 mask1, INT32 mask2);
    // enables all set bits of mask1 in UTCINTEN1, and all bits of mask2
    // in UTCINTEN2; the bits that are not set in the mask keep their old
    // state; the function always returns True
    // note: the area parameter is ignored

    virtual EBool DisableInterrupt (int area, INT32 mask1, INT32 mask2);
    // disables all set bits of mask1 in UTCINTEN1, and all bits of mask2
    // in UTCINTEN2; the bits that are not set in the mask keep their old
    // state; the function always returns True
    // note: the area parameter is ignored


protected:

    SUtcsuRead *utcsuRead_;  // the clock structure itself
    SUtcsuWrite *utcsuWrite_;
    // note: have to make them protected, since the software clock needs
    // to delete them


private:

    double oscillatorFrequency_;  // frequency of the clock oszillator

    double granularity_;  // external clock granularity (configurable)
    double granularityAcc_;  // accuracy transmission loss (configurable)
    double granularityDT_;  // granularity of the duty timers (fixed)
    double granularityOsc_;  // oscillator granularity (fixed)
    double granularityS_;  // internal clock granularity (fixed)

    INT32 utcinten1_;  // the current values of the UTCINTEN registers
    INT32 utcinten2_;
};



//**************************************************************************
// class CClockTemplate: the template class used for obtaining an instance
//      of the currently used clock (hardware or software)
//**************************************************************************

class CClockTemplate
{
public:

    //----------------------------------------------------------------------
    //--- return a new instance of the class
    //----------------------------------------------------------------------

    virtual CClock *NewClock (double frequency, SINT32 nodeID, int area);
    // returns a new object of the class; this function is needed so that
    // the clock interface can obtain the correct clock (HW or SW)
    // implementation of the function is left open to the programmer of
    // the hardware/software clock class
    // note: the nodeID is only needed for the software clock and is
    // ignored in the hardware
};


#endif


