//
// clockifc.hpp
//
// declaration of the clock interface
//
// created: 26.10.96
//
// last change: 03.02.99
//
// RCS version control information:
//     $Revision: 1.17 $$State: Exp $$Date: 1999/02/03 11:43:37 $
//


#ifndef _CLOCKIFC_HPP_
#define _CLOCKIFC_HPP_

#include "psos.h"

#include "bool.hpp"
#include "clockcfg.hpp"
#include "crashing.hpp"
#include "ctrled.hpp"
#include "distrib.hpp"
#include "interupt.hpp"


class CClock;
class CRandom;

struct SParameter;
struct STimestamp;


enum EInterruptSource {
    // UTCINTxx1
    IntSnapshot = 16,

    // UTCINTxx2
    IntGps1PPS1 = 0 + 32,
    IntGps1PPS2 = 1 + 32,
    IntGps1PPS3 = 2 + 32,

    IntDutyA1 = 20 + 32,
    IntDutyA2 = 21 + 32,
    IntDutyA3 = 22 + 32,
    IntDutyA4 = 23 + 32,
    IntDutyA5 = 24 + 32,
    IntDutyA6 = 25 + 32,
    IntDutyB1 = 26 + 32,
    IntDutyB2 = 27 + 32,
    IntDutyB3 = 28 + 32,
    IntDutyB4 = 29 + 32,
    IntDutyB5 = 30 + 32,
    IntDutyB6 = 31 + 32
};
// possible interrupt sources; the value is the bit number in the UTCSU
// interrupt registers


// contains all data necessary for wobbling the drift
struct SWobbleDriftData
{
    SRandomData distrib;  // the distribution data for getting rhoAdd
    double period;        // the period of adding rhoAdd
    double rhoMin;        // the lower bound on the drift during wobbling
    double rhoMax;        // the upper bound on the drift during wobbling
    EBool addDrift;       // True -> rhoAdd is added, False -> subtracted
};


//**************************************************************************
// class CClockInterface: this class is the interface to the clock (both
//      hardware and software), provides an ISR for HW-snapshots (INTA)
//**************************************************************************

class CClockInterface : public CApplControlledObject,
                        public CCrashingModule,
                        public CNeedInterrupt
{
    friend class ACGps;
    friend void CallWobbleISR (CNeedInterrupt *obj);


public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CClockInterface (CControllingObject *ctrl, double clockFrequency,
                     int area);
    virtual ~CClockInterface ();


    //----------------------------------------------------------------------
    //--- reset
    //----------------------------------------------------------------------

    virtual void Reset();
    // set members and the base classes


    //----------------------------------------------------------------------
    //--- configuration of system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the given parameter to its new value; if the parameter does
    // not exist or if the value is not valid, False is returned
    // known params: WobbleDrift, AmortizationDuration, ClockValue,
    // Accuracy, PFast, PSlow, DistributionFast, DistributionRegular,
    // DistributionSlow, PReadFault, PWriteFault, PDriftChange,
    // and all of CClock (HW or SW), CApplControlledObject, CCrashingModule

    virtual EBool GetParameter (SParameter& param) const;
    // returns the value of the given parameter in the appropriate field of
    // SParameter; if the parameter does not exist, False is returned
    // additional parameter: CrashStatus, RhoPlus, RhoMinus


    //----------------------------------------------------------------------
    //--- execution of system commands
    //----------------------------------------------------------------------

    virtual EBool ExecuteCommand (const SCommand& command);
    // returns True if the command was executed, else False
    // known commands: ConnectGps, DisconnectGps, and all commands of
    // CApplControlledObject and CCrashingModule


    //----------------------------------------------------------------------
    //--- crash/recover
    //----------------------------------------------------------------------

    virtual void Crash();  // stop the clock (STEP=STEPLOW=0)
    virtual void Recover();  // set step and steplow to old values


    //----------------------------------------------------------------------
    //--- handle GPS receivers
    //----------------------------------------------------------------------

    EBool GpsConnected ();
    // returns True if there is at least one GPS receiver connected to
    // the clock


    //----------------------------------------------------------------------
    //--- configuration of the amortization duration
    //----------------------------------------------------------------------

    EBool SetAmortizationDuration (double amorttime);
    // stores the maximum amortization duration (only if 'amorttime' >= 0;
    // if the clock has GPS, 'amorttime' must be in (0,1) )
    // returns True if the amortization time was taken over, else False
    // amorttime==0 causes SetTimestamp() to set the clock instantaneously


    //----------------------------------------------------------------------
    //--- access clock data
    //----------------------------------------------------------------------

    double GetRhoPlus () const;
    // returns the rate uncertainty (upper bound)

    double GetRhoMinus () const;
    // returns the rate uncertainty (lower bound)

    INT32 GetAccuracy () const;
    // returns the current accuracy of the clock (value of STATEPNGET)

    double GetFrequency () const;
    // returns the (ideal) clock frequency

    double GetGranularity () const;
    // returns the external granularity of the clock

    double GetGranularityDT () const;
    // returns the duty timer granularity of the clock

    double GetGranularityS () const;
    // returns the internal granularity of the clock

    double GetGranularityAcc () const;
    // returns the accuracy transmission loss of the clock

    double GetGranularityOsc () const;
    // returns the oscillator granularity of the clock


    //----------------------------------------------------------------------
    //--- clock value access functions
    //----------------------------------------------------------------------

    virtual EBool GetTimestamp (STimestamp& time);
    // obtain the clock value (ms, ts, acc);
    // injects faults; if there is a read fault, 'time' is set to arbitrary
    // values and True is returned;
    // if there are no faults, the return value of the function
    // depends on the number of GPS receivers connected to the clock: if
    // there are no receivers, the function always returns True; if there
    // are GPS receivers connected, the function returns the value of
    // 'gpsStatus'
    // context: ISR


    virtual void SetTimestamp (const STimestamp& time);
    virtual void SetTimestamp (const SExtendedTimestamp& time);
    // set the clock value; in the latter function, the microstamp can be
    // set to a value different from 0
    // if the clock has GPS (P-Node), a call to the functions has no effect;
    // if amortTime=0, the clock is directly set to the new values (in that
    // case, the accuracies are both set to the higher value in time.acc);
    // if amortTime > 0, the differences are amortized (and the accuracies
    // can be non-symetrical)
    // may inject faults (crash, drift change); if there is a write fault,
    // ms,ts,us and ns are set to random values; if the clock is crashed,
    // the clock is not set


    virtual void PreloadTimestamp (double loadAt, const STimestamp& time);
    virtual void PreloadTimestamp (double loadAt, const SExtendedTimestamp& time);
    // if 'amorttime' is zero, SetTimestamp() is called; otherwise, the
    // functions call PreloadAmort() to load the amortization data and
    // then call InterruptAt() with 'loadAt' to arm duty timer B1 (which
    // upon activation starts the amortization phase)
    // context: ISR


    //----------------------------------------------------------------------
    //--- access to other clock registers
    //----------------------------------------------------------------------

    void SetSteppure (INT32 step);
    // sets the STEPPURE register; if the clock is crashed, the value is
    // not set

    INT32 GetSteppure ();
    // returns the current value of the STEPPURE register

    void SetSteplow (INT8 steplow);
    // sets the STEPLOW register; if the clock is crashed, the value is
    // not set

    void SetLambdapure (INT16 lambdaP, INT16 lambdaN);
    // sets the LAMBDAPURE registers; if the clock is crashed, the values
    // are not set
    // note: if you want to use PreloadTimestamp() to preload a clock
    // value at a given time in the future, then you should only change
    // the LAMBDA values of the clock through this function; setting the
    // registers directly will result in wrong computation of the
    // correction values for the accuracy!

    void SetBounds (INT32 bounds);
    // sets the BOUNDP(N) register (the upper 16 bits of 'bounds' are the
    // value of BOUNDN, the lower 16 bits the value of BOUNDP);
    // if the clock is crashed, the value is not set


    //----------------------------------------------------------------------
    //--- interrupt handling
    //----------------------------------------------------------------------

    virtual void EnableInterrupt (EInterruptSource source);
    // enable the interrupt of the source; that assumes that the vector
    // has already been set;
    // if source<0 or source>63, an assert fails

    virtual void DisableInterrupt (EInterruptSource source);
    // if the interrupt of the source was enabled, it is turned off; else
    // nothing happens
    // if source<0 or source>63, an assert fails


    virtual EBool InterruptPending (EInterruptSource source);
    // returns True if the IP bit of the source is set, else False
    // if source<0 or source>63, an assert fails
    // context: ISR

    virtual EBool InterruptPending (EClockInterrupt inttype);
    // returns True if an interrupt of the given type is pending, else
    // False


    virtual void ClearInterrupt (EInterruptSource source);
    // if the IP bit of the source was set, it is cleared; else nothing
    // happens
    // if source<0 or source>63, an assert fails
    // context: ISR

    virtual void ClearInterrupt (EClockInterrupt inttype);
    // clears all interrupts of the given interrupt type (INTA, INTN, INTT)


    virtual void InterruptAt (EInterruptSource source, double at);
    // produces a duty timer interrupt at clock time 'at' (in seconds)

    virtual void InterruptAfter (EInterruptSource source, double after);
    // produces a duty timer interrupt in 'after' (clock) seconds

    virtual double InterruptAtMult (EInterruptSource source, double multTime);
    // produces a duty timer interrupt when the clock time is a multiple of
    // 'multTime' (in seconds);
    // example: if multTime is 100 sec, and clock time is 185, the clock
    // will raise an interrupt at clock time 200
    // the function returns the (clock) time at which the interrupt is
    // triggered

    virtual void DisableDutyTimer (EInterruptSource source);
    // if 'source' is a duty timer, DUTYL is set to zero (thus disabling
    // the timer); otherwise, nothing happens
    // context: ISR


    //----------------------------------------------------------------------
    //--- get the clock itself
    //----------------------------------------------------------------------

    CClock *GetClock();
    // returns the clock itself; only to be used if direct access to the
    // clock registers is absolutely necessary!


protected:
#ifdef PUBLIC_IN_TEST
public:
#endif

    //----------------------------------------------------------------------
    //--- initialization and destruction of the data members
    //----------------------------------------------------------------------

    virtual void Init ();
    // initialize all members; enable the snapshot and DTB6 interrupts

    virtual void Dispose ();
    // delete 'gps' -> go back to S-Node
    // all duty timers and interrupts enabled by this class are disabled;
    // other interrupts and timers should be disabled in the classes that
    // have enabled them!

    void InitializeClock ();
    // initialize the clock registers; the function does not change the
    // UTCINTEN registers


    //----------------------------------------------------------------------
    //--- rate changes
    //----------------------------------------------------------------------

    void SetStepRegisters (INT32 steppure, INT32 steplow);
    // sets STEPPURE and STEPLOW to the parameters (bits 0-23 of steplow are
    // masked out before setting STEPLOW in MSSET); if steppure or steplow
    // are not 0, the delta values of 'deltaSteppure' and 'deltaSteplow' are
    // added and the clock registers are set to the result; if both steppure
    // and steplow are 0, then the registers are set to 0 as well (stopping
    // the clock)
    // note: steppure_ and steplow_ are not changed


    //----------------------------------------------------------------------
    //--- fault injection
    //----------------------------------------------------------------------

    virtual void ReportCrashChange();
    // creates an event of type SApplCrashEvent, fills in all parameters
    // (type=CrashStatusChange) and sends it to the evaluation

    void InjectFaults();
    // modify crash status, maybe change state and rate

    void SetFaultyDrift (double rho);
    // sets deltaSteppure and deltaSteplow so that the clock drifts by 'rho'
    // and calls SetStepRegisters() to take over this drift

    double GetFaultyDrift ();
    // returns the current value of the faulty drift


    //----------------------------------------------------------------------
    //--- error reports
    //----------------------------------------------------------------------

    virtual int ReportError (EReportType type, char *file, int line,
                             char *expression);


private:

    CClock *clock_;  // the clock itself (either hardware or simulated)
    INT32 steplow_;  // value of STEPLOW (part of the MSSET register)
    INT32 steppure_;  // value of STEPPURE (needed for amortization)

    INT16 lambdaP_;  // value of LAMBDAPPURE (needed for accuracy preload)
    INT16 lambdaN_;  // value of LAMBDANPURE (needed for accuracy preload)
    // both values have the format S:[-37,-51]

    INT32 deltaSteppure_;  // rate modification for fault injection
    INT16 deltaSteplow_;  // bit 8 is sign-bit, bit 9 that of deltaSteppure!
    double deltaRho_;  // stores the faulty drift in easier format

    double rhoPlus_, rhoMinus_;  // bounds on the rate drift
    double amorttime_;  // maximum amortization duration

    ACGps *gps_;  // the GPS class; if it is != NULL, we have P-Node
    // functionality (SetTimestamp() has no effect); if it is ==NULL, an
    // S-Node is simulated
    EBool gpsStatus_;  // the status of the GPS receiver(s)
    // True if all receivers are healthy, False if at least one is not okay;
    // if gps_ is NULL, the value of the variable is ignored

    int area_;  // the area of the clock (for the GPS class)

    CRandom *random_;  // for obtaining random numbers
    double pReadFault_, pWriteFault_;  // prob. for wrong read/write access
    double pDriftChange_;  // prob. of a drift change
    double pFast_, pSlow_;  // prob. for new rate being out of bounds
    SRandomData distDataFast_, distDataRegular_, distDataSlow_;
    // data for the rate change distributions

    SWobbleDriftData wobbleData_;  // for drift wobbling


#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- snapshot ISR
    //----------------------------------------------------------------------

    virtual void isr();
    // called if a HW snapshot interrupt (INTA) occurs; reads the NTP-time
    // and sends an event to the Supervisor; clears the interrupt
    // context: ISR


    //----------------------------------------------------------------------
    //--- drift wobbling ISR
    //----------------------------------------------------------------------

    virtual void wobble_isr();
    // called if a duty timer B6 interrupt (INTT) occurs; gets a value to
    // add to the drift and sets the deltaSteppure/low values accordingly;
    // clears the interrupt
    // context: ISR


    //----------------------------------------------------------------------
    //--- connect/disconnect GPS receivers
    //----------------------------------------------------------------------

    void AddGps (EGpsUnit gpu);
    // enables the GPS receiver specified by <nodeID, area, gpu> (nodeID
    // and area are internal members) and installs an ISR for it
    // from this point on, the CSA will be a P-Node, so the clock will
    // refuse to take over new clock values set by the clock synchronization
    // algorithm
    // assert: gpu < EGpsUnitCount

    void RemoveGps (EGpsUnit gpu);
    // disables the GPS receiver specified by <nodeID, area, gpu>; if it
    // was not enabled, the function has no effect
    // if there are no further GPS receivers enabled, the node type will
    // switch to S-Node and the clock will accept new clock values set by
    // the clock synchronization algorithm


    //----------------------------------------------------------------------
    //--- clock amortization
    //----------------------------------------------------------------------

    SINT32 DragAccuracy (SINT32 currAlpha, SINT16 lambda, double afterTime);
    // computes the accuracy after 'afterTime' clock seconds if the current
    // accuracy is 'currAlpha' (in S:[-8,-38] format) and the deterioration
    // value is 'lambda' (in S:[-37,-51] format); the value returned is in
    // S:[-8,-38] format
    // assert: afterTime >= 0

    void ComputeAccuracyCorrection (INT32 acc, double afterTime,
        INT32& diffPlus, INT32& diffMinus);
    // computes the accuracy difference that one should amortize in 'afterTime' 
    // (clock) seconds (through a call to Amortize() or PreloadAmort())

    virtual void PreloadAmort (double diff, INT32 diffPlus, INT32 diffMinus);
    // programs the clock to amortize the given time difference (in
    // seconds); if the time cannot be fully amortized, as much as
    // possible is amortized;
    // 'diffPlus', 'diffMinus' are the differences of the accuracies,
    // given in format S:[-8,-38]
    // assert: amorttime_ > 0

    virtual void Amortize (double diff, INT32 diffPlus, INT32 diffMinus);
    // calls PreloadAmort() to load everything for the amortization phase,
    // then writes to STARTAMORT to start the amortization phase
    // the function is also called by the associated GPS class (if the
    // clock has GPS functionality)


    //----------------------------------------------------------------------
    //--- odds and ends
    //----------------------------------------------------------------------

    CNtpTime GetCurrentClockTime ();
    // return the current clock time (timestamp and macrostamp)

    void SetDutyTimer (EInterruptSource source, const CNtpTime& time);
    // program the duty timer to raise an interrupt at NTPTIME time
    // if 'source' is no duty timer, an assert fails
};

#endif


