//
// clockirq.hpp
//
// clock interrupt processes
//
// created: 26.10.96
//
// last change: 25.08.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/23 08:55:28 $
//


#ifndef _CLOCKIRQ_HPP_
#define _CLOCKIRQ_HPP_


#include <stdlib.h>  // for NULL

#include "clockcfg.hpp"
#include "intproc.hpp"


class CSoftwareClock;


//**************************************************************************
// class CClockInterrupt: an interrupt that was produced by the clock
//**************************************************************************

class CClockInterrupt : public CInterruptProcess
{
public:
    //----------------------------------------------------------------------
    //--- constructor
    //----------------------------------------------------------------------

    CClockInterrupt (unsigned long nodeID)
        { SetNodeID (nodeID); scheduleTime_=-1; intstat_=NULL; ipBit_=32; }


    //----------------------------------------------------------------------
    //--- management of the schedule time and status register/bit
    //----------------------------------------------------------------------

    void SetScheduleTime (double time) { scheduleTime_ = time; }
    // store the (UTCSU) clock value (in seconds) for which the process
    // waits

    double GetScheduleTime () { return scheduleTime_; }
    // returns the clock value for which the process waits


    void SetStatusRegister (INT32& statusreg) { intstat_ = &statusreg; }
    // set the status register that will contain the interrupt pending bit;

    void SetPendingBit (INT8 bitnumber)
        { if (bitnumber < 32) ipBit_ = bitnumber; }
    // store the IP bit


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- interrupt specific work
    //----------------------------------------------------------------------

    // is executed prior to calling the ISR; sets the IP bit
    virtual void PreExecution ()
        { if ((intstat_ != NULL) && (ipBit_ < 32)) *intstat_ |= 1<<ipBit_;
          scheduleTime_ = -1; }

/**/// note: PostExecution() should schedule the interrupt again if the
    // interrupt condition has not changed; right now, if the interrupt
    // is a Duty Timer and the INTEN bit and the EA bit in DutyL stay
    // high (are not cleared in the ISR), the interrupt is not called again
    // until INTEN is modified


private:

    double scheduleTime_;
    // the clock value (in seconds), at which the interrupt should occur

    INT32 *intstat_;
    // the status register that the interrupt will modify when it is
    // raised

    INT8 ipBit_;
    // number of the interrupt pending bit
};


//**************************************************************************
// class CDutyB1Interrupt: an interrupt of duty timer B on SSU1; this duty
//    timer is hardwired to AMORTDUTY and starts an amortization phase
//    (i.e., writes to register STARTAMORT)
//**************************************************************************

class CDutyB1Interrupt : public CClockInterrupt
{
public:

    //----------------------------------------------------------------------
    //--- constructor
    //----------------------------------------------------------------------

    CDutyB1Interrupt (unsigned long nodeID, CSoftwareClock *clock);
    // store a pointer to the clock; needed for setting STARTAMORT


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- interrupt specific work
    //----------------------------------------------------------------------

    virtual void PreExecution ();
    // is executed prior to calling the ISR; sets the IP bit, writes to
    // register STARTAMORT


private:

    CSoftwareClock *clock_;  // clock value is captured at the 1PPS pulse
};


//**************************************************************************
// class CGpsInterrupt: an interrupt of the GPS receiver (every full
//      second); captures the clock value, calls the 1PPS ISR (if enabled),
//      then waits for 'GpsDelay' ms and calls the ISR of the GPS-receiver 
//      (time information ready)
//**************************************************************************

class CGpsInterrupt : public CClockInterrupt
{
public:

    //----------------------------------------------------------------------
    //--- constructor
    //----------------------------------------------------------------------

    CGpsInterrupt (unsigned long nodeID, CSoftwareClock *clock, EGpsUnit gpu);
    // store a pointer to the clock that is "connected" to the GPS receiver
    // assert: gpu  < EGpsUnitCount

    ~CGpsInterrupt();
    // unreference or delete the GPS time interrupt process


    //----------------------------------------------------------------------
    //--- handling of the GPS time interrupt
    //----------------------------------------------------------------------

    void SetTimeVector (int area);
    // obtains and sets the interrupt vector of the GPS time ISR

    void EnableTime ();
    // the GPS time ISR is called in PostExecution()

    void DisableTime();
    // the GPS time ISR is not called in PostExecution()


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- interrupt specific work
    //----------------------------------------------------------------------

    virtual void PreExecution ();
    // is executed prior to calling the ISR; sets the IP bit, prompts the
    // clock to capture the 1PPS pulse

    virtual void PostExecution ();
    // is executed after calling the ISR; it schedules the GPS time interrupt
    // for activation in 'GpsDelay' seconds from now, then schedules itself
    // for the next full seconds


private:

    CSoftwareClock *clock_;  // clock value is captured at the 1PPS pulse
    CInterruptProcess *gpstime_;  // the GPS time interrupt
    EGpsUnit gpu_;  // the GPS unit to which the interrupt belongs
};

#endif

