//
// clocksyn.hpp
//
// declaration of the clock synchronization class
//
// created: 09.11.96
//
// last change: 06.01.98
//
// RCS version control information:
//     $Revision: 1.4 $$State: Exp $$Date: 1998/07/28 12:31:25 $
//


#ifndef _CLOCKSYN_HPP_
#define _CLOCKSYN_HPP_

#include "ci.h"

#include "ctrled.hpp"
#include "event.hpp"
#include "general.hpp"
#include "interupt.hpp"


class CClockInterface;
class CLance;
class CNodeIdList;

struct SSendData;
struct SReplyData;
struct STimeData;


//**************************************************************************
// class CClockSync: this class is the main clock synchronization
//**************************************************************************

class CClockSync : public CApplControlledObject, public CNeedInterrupt
{
    friend class ACClockSyncAlgorithm;
    // so it has access to the network ID and can call the ReportEvent()
    // functions (the class is called CSA in the comments below)


public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CClockSync (CControllingObject *ctrl, int area);
    virtual ~CClockSync ();
    // the clock must not have been deleted when the dtor is called


    //----------------------------------------------------------------------
    //--- suspend/resume the synchronization
    //----------------------------------------------------------------------

    void Suspend ();
    // don't accept messages, suspend the CSA

    void Resume ();
    // trigger an interrupt at the next multiple of the round period


    //----------------------------------------------------------------------
    //--- configuration of system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the requested parameter value; returns True if successful,
    // otherwise False
    // existing parameters: RoundPeriod, DelayMeasurementCycles,
    // ComputationDelayCompensation, DelayCompensation,
    // RateSynchronizationCycles, SynchronizationAlgorithm, and all params
    // of CApplControlledObject
    // parameters to CSA: NumberOfNodes, NumberOfFaultyNodes, Delta,
    // EpsilonPlus, EpsilonMinus, ThetaPlus, ThetaMinus, U, UserParameter,
    // AssociatedNetwork

    virtual EBool GetParameter (SParameter& param) const;
    // returns the requested parameter value; returns True if successful,
    // otherwise False
    // existing parameters: TimerGranularity and all of SetParameter()


    //----------------------------------------------------------------------
    //--- execution of system commands
    //----------------------------------------------------------------------

    virtual EBool ExecuteCommand (const SCommand& command);
    // executes the command; returns True if successfull, otherwise False;
    // existing commands: ConnectNetwork, DisconnectNetwork, and all commands
    // of CApplControlledObject
    // commands to CSA: ConnectNetwork (also processed in this class),
    // DisconnectNetwork (also processed in this class), UserCommand


    //----------------------------------------------------------------------
    //--- set associated objects
    //----------------------------------------------------------------------

    void SetLance (CLance *lance);
    // set the lance pointer to the lance

    void SetClock (CClockInterface *clock);
    // set the clock pointer to the clock (and notify CSA)


    //----------------------------------------------------------------------
    //--- clock synchronization functions
    //----------------------------------------------------------------------

    void BroadcastTime ();
    // sends a broadcast message to the Lance; if 'resynchronizing' is
    // True, messages that only consist of the timestamp are not send, and
    // in messages with additional information the timestamp is marked
    // invalid
    // assert: Lance != NULL

    void Announce (renvx *renv, STimeData sendtime, char *message,
                   INT16 msgsize, int if_num, INT32 msgID);
    // take over the data, process it and give it to CSA
    // assert: renv != NULL

    void ApplyConvergenceFunction ();
    // prompt CSA to use the collected data to compute a new clock value

    void Resynchronize ();
    // prompt CSA to set the clock to the new value


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- create/destroy the members that are modified during reset
    //----------------------------------------------------------------------

    virtual void Init ();
    // creates and/or initializes the members that are modified during
    // reset; a second call (without Dispose() in between) has no effect

    virtual void Dispose ();
    // deletes all dynamic members; a second call (without Init() in
    // between), has no effect


    //----------------------------------------------------------------------
    //--- management of interrupts at multiples of the round period
    //----------------------------------------------------------------------

    void InterruptAtMult ();
    // prompts the duty timer of the clock to generate an interrupt at
    // the next multiple of the round period, sets timeOfResync_
    // accordingly
    // assert: clock != NULL


    //----------------------------------------------------------------------
    //--- management of the synchronization algorithm
    //----------------------------------------------------------------------

    EBool ChangeAlgorithm (EAlgorithmType newType);
    // change the synchronization algorithm (change csa and csaType)
    // the algorithm can be changed on the fly (no invalid pointers)
    // if the algorithm type has been installed, True is returned and the
    // algorithm is changed, else False is returned


    //----------------------------------------------------------------------
    //--- error reports
    //----------------------------------------------------------------------

    virtual int ReportError (EReportType type, char *file, int line,
                             char *expression);


private:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- actions taken in Announce()
    //----------------------------------------------------------------------

    void ComputeDelay (renvx *renv, STimeData& time, SReplyData& data,
                       STimeData& myTime, int if_num);
    // compute the transmission delay

    void ReplyToMessage (renvx *renv, STimeData& time, int if_num);
    // reply to a request for transmission delay measurement
    // assert: clock, lance != NULL

    void StoreTime (renvx *renv, STimeData& time, int if_num, EBool gps=False);
    // store the time information (gps is True if the sender had a GPS
    // receiver connected to it)


    //----------------------------------------------------------------------
    //--- interrupt service routine
    //----------------------------------------------------------------------

    virtual void isr ();
    // called if a duty timer interrupt from DTB4 (INTT) occurs; calls one
    // of BroadcastTime(), ApplyConvergenceFunction() or Resynchronize(),
    // and sets the duty timer to a new value (depending on the round period)


private:

    CLance *lance_;
    CClockInterface *clock_;

    ACClockSyncAlgorithm *csa_;  // user-definable part of the sync. alg.
    EAlgorithmType csaType_;  // the currently used algorithm type

    CNodeIdList *listOfConnectedNetworks_;  // all connected networks

    int area_;
    // this is the DMA area of the clock

    double roundPeriod_;  // the round period (in seconds)

    double delayCompensation_;  // time from FME to the call to CV
    double computationDelayCompensation_;// time from call to CV to resync.

    INT32 delayCycles_;
    // the period of delay measurement messages (in number of calls to FME);
    // example: if every n-th message should include delay measurement, set
    // delayCycles to n
    // n=0 prevents delay measurement

    INT32 delayCounter_;
    // needed to count the calls to FME for delay measurement (can't make
    // it static in FME because it has to be unique for all objects of
    // this class)

    INT32 rateCycles_;
    // the period of rate synchronization messages (in number of calls to
    // FME); see the example for 'delayCycles'
    // n=0 prevents rate synchronization

    INT32 rateCounter_;
    // needed to count the calls to FME for rate synchronization

    EBool suspended_;
/**/// might become obsolete, don't know right now

    EBool initCalled_;  // set to True by Init(), to False by Dispose()

    INT8 roundStage_; // what function the ISR should call (FME, CV, Resync)

    EBool resynchronizing_; // True if CS does not send own timestamp in FME

    double timeOfResync_;
    // if not negative, it contains the time of the next resynchronization

    EBool reportDelays_;
    // True if measured transmission delays should be reported to EvalSys
};


#endif


