//
// constant.hpp
//
// declaration of all constants in the system
// the corresponding definitions are in constant.cpp; edit this file
// and link again to change system constants
// note: constants are devided into three groups (changeable by simulation,
// by evaluation, by user), ordering in the groups is alphabetically
//
// created: 28.09.96
//
// last change: 28.09.98
//
// RCS version control information:
//     $Revision: 1.5 $$State: Exp $$Date: 1998/09/28 10:58:18 $
//


#ifndef _CONSTANT_HPP_
#define _CONSTANT_HPP_

#include "psos.h"  // needed for INT32


//**************************************************************************
// the following constants are system constants; the user may change them
// through the evaluation system
//**************************************************************************

//--------------------------------------------------------------------------
// synchronization round parameters
//--------------------------------------------------------------------------

extern const double DefaultRoundPeriod;
// the default round period of the clock synchronization application (in
// seconds)
// used by: CClockSync

extern const double DefaultDelayCompensation;
// the default wait period from FME to application of the convergence
// function (in seconds)
// used by: CClockSync

extern const double DefaultComputationDelayCompensation;
// the default wait period from application of the convergence function to
// resynchronization (in seconds)
// used by: CClockSync


//--------------------------------------------------------------------------
// Supervisor/Controller parameters
//--------------------------------------------------------------------------

extern const double DefaultIfcPeriod;
// the default period (in seconds) of the interface modules responsible for
// communication with the evaluation; in the case of the Supervisor, this
// is also the period of the (periodic) system snapshot
// used by: CControllingObject


//--------------------------------------------------------------------------
// network parameters
//--------------------------------------------------------------------------

extern const double DefaultOwnDelay;
// the default delay of a message that is transmitted to the sender node
// itself (in seconds)
// if you do not specify transmission delay data from a node to itself,
// then this default is used; in consequence, messages to oneself are
// always delivered if the node is connected, even if no delay data is
// stored
// used by: CNetworkAccountant


//--------------------------------------------------------------------------
// clock parameters
//--------------------------------------------------------------------------

extern const double WcRhoPlus;
// the upper bound on the worst case clock drift (global, for all clocks)
// used by: CClockInterface, ACClockSyncAlgorithm

extern const double WcRhoMinus;
// the lower bound on the worst case clock drift (global, for all clocks)
// used by: CClockInterface, ACClockSyncAlgorithm

extern const double DefaultAmortizationDuration;
// the default duration of the amortization phase
// used by: CClockInterface

extern const double DefaultClockGranularity;
// the external clock granularity
// used by: CClock

extern const double DefaultAccGranularity;
// the accuracy transmission loss
// used by: CClock


//**************************************************************************
// the following constants are used internally and in communication with the
// evaluation; they may be changed by the programmer of the simulation or
// by the evaluation, but have no effect on the user
//**************************************************************************

// right now, no constants fall into this group


//**************************************************************************
// the following constants are simulation details; they may be changed by
// the programmer of the simulation, but have no (direct) effect on the
// evaluation or the user
//**************************************************************************

//--------------------------------------------------------------------------
// general constants
//--------------------------------------------------------------------------

extern const SINT32 NoAccount;
// this constant is generally used to indicate that the requested data
// was not found
// used by: a lot of classes

extern const INT32 ArrayAllocationSize;
// number of array elements that are freshly allocated whenever there is
// not enough allocated memory to store a new object
// used by: the list classes (ACObjectList, ...)


//--------------------------------------------------------------------------
// node topology constants
//--------------------------------------------------------------------------

extern const INT32 NumberOfAreas;
// the number of areas on one node; the constant is determined by the
// number of NTI modules that the carrier board can hold
// used by: several classes that check if the 'area' parameter requested
// is within acceptable bounds
// note: the simulation assumes that every area within [0, NumberOfAreas)
// can be used; if a slot on the carrier board is not occupied, the user
// must not create an application on that area, since the simulation does
// not check if a slot is occupied


//--------------------------------------------------------------------------
// socket constants
//--------------------------------------------------------------------------

extern const int MaxTrials;
// default number of s_ident() calls before giving up
// used by: CClientSocket

extern const double SocketTimeout;
// the default time an s_receive() call may block (in seconds)
// used by: CSocket

extern const INT32 MaxMessageSize;
// maximum size of a message (useful for avoiding allocation of too
// much memory if the byte order conversion fails)
// used by: CStreamSocket

extern const int EvalsysPort;
// the default port of the evaluation system (stream socket)
// used by: main()

extern const int ErrorPort;
// the default port of the evaluation system error handler (datagram socket)
// used by: main()


//--------------------------------------------------------------------------
// GPS constants
//--------------------------------------------------------------------------

extern const double GpsDelay;
// the delay (in seconds) from the GPS 1PPS signal to calling the ISR
// (pure simulation only)
// used by: CGpsInterrupt

extern const double GpsEpsilon;
// the clock value of a P-Node must differ from the GPS time by more
// than this value if the difference is to be amortized
// used by: ACGps


//--------------------------------------------------------------------------
// error report constants
//--------------------------------------------------------------------------

extern const double ErrorReportPeriod;
// the period (in seconds) of the error report task/process
// used by: CErrorTask (and derived classes)

extern const unsigned int MaxErrorMessages;
// the number of error messages the error report task/process processes
// in one round
// used by: CErrorTask (and derived classes)

extern const unsigned int MaxErrorLength;
// the maximum length of an error string
// used by: error.cpp, CErrorTask

extern const unsigned int MaxErrorNumber;
// the maximum number of error messages that the error queue can hold
// used by: CErrorTask

#endif

