//
// contbase.hpp
//
// declaration of the synchronization controller (abstract base class)
// note: define StartApplication(), StopApplication() and StopExecution()
//       in the derived class to get a working HW/SW implementation
//
// created: 25.01.97
//
// last change: 26.03.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/07/23 08:55:28 $
//


#ifndef _CONTBASE_HPP_
#define _CONTBASE_HPP_

#include "ctrling.hpp"


//**************************************************************************
// class ACController: provides an abstract base class for controlling all
//      simulation specific things
//**************************************************************************

class ACController : public CControllingObject
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    ACController (char qname[4], ACSocketAddress *addr);
    // call the queue "CTRL" or something like that; *don't* use a string
    // of (hex) numbers!

    virtual ~ACController ();


    //----------------------------------------------------------------------
    //--- reset
    //----------------------------------------------------------------------

    virtual void Reset();
    // reset the controller (delete all simulation objects)

    virtual EBool ResetTheSimulation();
    // reset the simulation (set all parameters in all applications to
    // their default values); returns True if at least one object exists,
    // False if no simulation objects exist


    //----------------------------------------------------------------------
    //--- configuration of system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the given parameter to its new value; if the parameter does
    // not exist or if the value is not valid, False is returned
    // existing params: all of CControllingObject

    virtual EBool GetParameter (SParameter& param) const;
    // returns the value of the given parameter in the appropriate field of
    // SParameter; if the parameter does not exist, False is returned


    //----------------------------------------------------------------------
    //--- execution of system commands
    //----------------------------------------------------------------------

    virtual EBool ExecuteCommand (const SCommand& command);
    // executes the command; returns True if successfull, otherwise False
    // existing commands: all Controller commands of ECommandType concerning
    // the application (not the Network!) in params.hpp


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- error reports
    //----------------------------------------------------------------------

    virtual int ReportError (EReportType type, char *file, int line,
                             char *expression);


    //----------------------------------------------------------------------
    //--- starting/stopping applications
    //----------------------------------------------------------------------

    virtual EBool StartApplication (SINT32 nodeID, int area) =0;
    // start the application; derived classes should define this function
    // to implement hardware/software specific start procedures

    virtual EBool StopApplication (SINT32 nodeID, int area) =0;
    // stop the application; derived classes should define this function
    // to implement hardware/software specific stop procedures


    //----------------------------------------------------------------------
    //--- stop program execution
    //----------------------------------------------------------------------

    virtual void DeleteAllSimulationObjects ();
    // delete all entries in the supervisor list (disables interrupts at
    // the start of the function, enables them again at the end)

    virtual void StopExecution() =0;
    // stop the program; derived classes should define this function
    // to implement hardware/software specific exit procedures


private:

    //----------------------------------------------------------------------
    //--- internal command execution
    //----------------------------------------------------------------------

    EBool CreateNewApplication (SApplCommand& applc);
    // create a new supervisor (start the application); returns False if
    // area not in [0, min {ASICCNT-1, 15}], if area is equal to the area
    // of the master clock, if nodeID is not in [0, 16^3-1], or if an
    // application with that (nodeID, area) exists; else the application
    // is created and True is returned

    EBool DeleteSupervisor (SINT32 nodeID, int area);
    // delete the supervisor (stop the application); returns True if the
    // application was deleted successfully, else False

    void ExecuteCommandForAll (ESystemCommand type);
    // executes the command (SuspendSimulation, ResumeSimulation) by
    // suspending or resuming all supervisors
};


#endif


