//
// database.hpp
//
// declaration of the database class that is used to store the information
// collected about the network transmission times and about other nodes
//
// created: 29.04.97
//
// last change: 08.10.98
//
// RCS version control information:
//     $Revision: 1.6 $$State: Exp $$Date: 1998/10/08 17:07:52 $
//


#ifndef _DATABASE_HPP_
#define _DATABASE_HPP_


#include <stdlib.h>  // for NULL

#include "bool.hpp"
#include "gentypes.hpp"  // for TRateTheta and TRateU
#include "time.hpp"


class CAddr;
class CAddressManager;
class CDelaySetList;
class CRateSetList;
class CTimeSetList;

struct SParameter;


//**************************************************************************
// structures that may be accessed in the database
//**************************************************************************

// contains the transmission delay data
struct SDelay
{
    double delta;         // deterministic part of the transmission delay
    double epsilonPlus;   // upper bound of variable part
    double epsilonMinus;  // lower bound of variable part
};

// contains clock rate information of the other nodes
struct SRate
{
    TRateTheta thetaPlus;  // computed upper bound on the clock rate
    TRateTheta thetaMinus;  // computed lower bound on the clock rate
    TRateU u;  // sum of the amortizations since the last rate sync.
};

// contains the timestamp intervals obtained from the other nodes
struct STimeInterval
{
    CNtpTime low;       // lower envelope
    CNtpTime high;      // upper envelope
    CNtpTime refpoint;  // reference point
};



//**************************************************************************
// class CApplicationDatabase: stores information about other nodes (rates,
//      current time intervals, transmission characteristics); storing
//      additional information is possible
//      data is accessed with the key <if_num, netAddress[, area]>
//**************************************************************************

class CApplicationDatabase
{
public:

    enum EDataType {Delay, Rate, Time};
    // use the enum to specify which data is to be accessed


    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CApplicationDatabase();
    virtual ~CApplicationDatabase();


    //----------------------------------------------------------------------
    //--- reset
    //----------------------------------------------------------------------

    virtual void Reset ();
    // discard all data


    //----------------------------------------------------------------------
    //--- configuration of system parameters - IFC to evalsys
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the given parameter to its new value; if the parameter does
    // not exist or the entry is not found or the value is not valid, False
    // is returned
    // parameters: Delta, EpsilonPlus, EpsilonMinus, ThetaPlus, ThetaMinus, U

    virtual EBool GetParameter (SParameter& param) const;
    // returns the value of the given parameter in the appropriate field of
    // SParameter; if the parameter or entry does not exist, False is
    // returned


    //----------------------------------------------------------------------
    //--- general access to the data of all types
    //----------------------------------------------------------------------

    INT32 Cardinality (EDataType type, SINT32 networkID);
    // returns the number of entries for the given type and network

    void DeleteAll (EDataType type, SINT32 networkID);
    // deletes all entries of the given type and network

    void DeleteAll (EDataType type);
    // deletes all entries of the given type (for all networks)

    void Delete (EDataType type, SINT32 networkID,
                 const CAddr& netAddress, int area);
    // deletes the entry; if type==Delay, the area parameter is ignored
    // note: area is no default param to avoid errors if Rate or Time is
    // accessed; inconvenient use for Delay is better than error prone use
    // for the other types


    //----------------------------------------------------------------------
    //--- access to the transmission characteristics
    //----------------------------------------------------------------------

    void SetDelay (const SDelay& delay, SINT32 networkID,
                   const CAddr& netAddress);
    // store the data; any old data is overwritten

    EBool GetDelay (SDelay& delay, SINT32 networkID,
                   const CAddr& netAddress) const;
    // if the entry exists, it is copied into 'delay' and True is returned;
    // else False is returned

    EBool RemoveDelay (SDelay& delay, SINT32 networkID, const CAddr& netAddress);
    // remove the entry; if it is found, it is copied into 'delay' and True
    // is returned, else False is returned

    EBool GetFirstDelay (SDelay& delay, SINT32 networkID);
    // if the list is not empty, the first entry is copied into 'delay'
    // and True is returned; else False is returned

    EBool GetNextDelay (SDelay& delay, SINT32 networkID);
    // if the list is not empty and the last element is not reached, the
    // next entry is copied into 'delay' and True is returned; else False
    // is returned


    //----------------------------------------------------------------------
    //--- access to the rate information
    //----------------------------------------------------------------------

    void SetRate (SRate *rate, SINT32 networkID,
                  const CAddr& netAddress, int area);
    // store the data; if there is already an entry stored, it is deleted
    // assert: rate != NULL

    SRate *GetRate (SINT32 networkID, const CAddr& netAddress, int area) const;
    // if the entry exists it is returned; otherwise, the function returns
    // NULL

    SRate *RemoveRate (SINT32 networkID, const CAddr& netAddress, int area);
    // remove the entry; if it is found, the pointer to the data is returned;
    // otherwise, NULL is returned

    SRate *GetFirstRate (SINT32 networkID);
    // if the list is not empty, the first entry is returned; else NULL is
    // returned

    SRate *GetNextRate (SINT32 networkID);
    // if the list is not empty and the last element is not reached, the
    // next entry is returned; else NULL is returned


    //----------------------------------------------------------------------
    //--- access to the time intervals
    //----------------------------------------------------------------------

    void SetTime (const STimeInterval& time, SINT32 networkID,
                  const CAddr& netAddress, int area);
    // store the data; any old data is overwritten

    EBool GetTime (STimeInterval& time, SINT32 networkID,
                  const CAddr& netAddress, int area) const;
    // if the entry exists, it is copied into 'time' and True is returned;
    // else False is returned

    EBool RemoveTime (STimeInterval& time, SINT32 networkID,
                     const CAddr& netAddress, int area);
    // remove the entry; if it is found, it is copied into 'time' and True
    // is returned, else False is returned

    EBool GetFirstTime (STimeInterval& time, SINT32 networkID);
    // if the list is not empty, the first entry is copied into 'time'
    // and True is returned; else False is returned

    EBool GetNextTime (STimeInterval& time, SINT32 networkID);
    // if the list is not empty and the last element is not reached, the
    // next entry is copied into 'time' and True is returned; else False
    // is returned


    //----------------------------------------------------------------------
    //--- manage association nodeID <-> network address - for IFC to Evalsys
    //----------------------------------------------------------------------

    virtual EBool SetAssociation (SINT32 networkID, const CAddr& address,
                                 SINT32 nodeID);
    // stores the association <networkID, address, nodeID>; if another
    // association already exists, the new data is ignored and False is
    // returned, else the triple is stored and True is returned; overload
    // the function in the derived class to obtain different behaviour

    EBool GetAssociation (SINT32 networkID, SINT32 nodeID, CAddr& netAddress);
    // returns True, if an associated network address exists; in that case,
    // the address is copied into netAddress; if the assoc. does not exist,
    // False is returned and netAddress is not modified

    EBool GetAssociation (SINT32 networkID, const CAddr& netAddress,
                         SINT32 *nodeID);
    // returns True, if an associated nodeID exists; in that case, the
    // value is copied into nodeID; if the assoc. does not exist, False is
    // returned and nodeID is not modified
    // assert: nodeID != NULL

    EBool DeleteAssociation (SINT32 networkID, SINT32 nodeID);
    // deletes the netAddress associated with the <networkID,nodeID> key
    // from the list of addresses; if there is no entry, False is returned,
    // else True is returned


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- create/destroy the members that are modified during reset
    //----------------------------------------------------------------------

    virtual void Init ();
    // creates the members that are modified during a reset; a second call
    // (without Dispose() in between) has no effect

    virtual void Dispose ();
    // deletes the dynamic members created by Init(); a second call
    // (without Init() in between), has no effect


private:

    CAddressManager *addressManager_;  // assoc. nodeID <-> network address
    CDelaySetList *listOfDelays_;  // list of transmission delays
    CRateSetList *listOfRates_;  // list of clock rates
    CTimeSetList *listOfTimes_;  // list of time intervals
    EBool initCalled_;  // set to True by Init(), to False by Dispose()
};


#endif


