//
// dataset.hpp
//
// three classes store time intervals and clock rates with the key <network
// address, area>, and transmission delays with the key <network address>
//
// created: 12.11.96
//
// last change: 01.05.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _DATASET_HPP_
#define _DATASET_HPP_


#include "address.hpp"
#include "bool.hpp"
#include "database.hpp"
#include "error.hpp"
#include "listbase.hpp"
#include "time.hpp"


//**************************************************************************
// class CTimeSet: this class stores time intervals with the network
//      address and area of the sending node as key
//**************************************************************************

class CTimeSet : private ACObjectList
// note: ACObjectList is private so that the user of the class can't call
// one of the access functions of ACObjectList that are not needed in
// CTimeSet; bad style, I know
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CTimeSet (INT32 aSize=ArrayAllocationSize) : ACObjectList (aSize) {}

    virtual ~CTimeSet () { DeleteAll (); }
    // the destructor deletes all objects in the array


    //----------------------------------------------------------------------
    //--- information about the set
    //----------------------------------------------------------------------

    INT32 Cardinality() { return ACObjectList::Occupied(); }
    // return the number of entries

    EBool Empty() { return ACObjectList::Empty(); }
    // returns True if the set is empty, else False


    //----------------------------------------------------------------------
    //--- move through the list
    //----------------------------------------------------------------------

    STimeInterval *GetFirst()
        { return (STimeInterval *)ACObjectList::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    STimeInterval *GetNext()
        { return (STimeInterval *)ACObjectList::GetNext(); }
    // get the next object in the array or NULL if the end of the array
    // has been reached

    STimeInterval *Find (const CAddr& netAddress, int area)
    {
        SKey key (netAddress, area);
        return (STimeInterval *) ACObjectList::Find (&key);
    }
    // returns the object with this key; if the key does not exist,
    // NULL is returned


    //----------------------------------------------------------------------
    //--- add, remove objects
    //----------------------------------------------------------------------

    // put a new interval into the list; returns True if successful,
    // False if an entry exists
    EBool Insert (STimeInterval *object, const CAddr& netAddress, int area)
    {
        SKey *key = new SKey (netAddress, area);

        if (!ACObjectList::Insert (object, key))  // object not inserted
        {
            delete key;  // the key is not needed
            return False;
        }
        else return True;
    }

    // remove an interval from the list; if the key does not exist, NULL
    // is returned
    STimeInterval *Remove (const CAddr& netAddress, int area)
    {
        SKey key (netAddress, area);
        return (STimeInterval *) ACObjectList::Remove (&key);
    }

    // remove the last interval from the list; if the list is empty, NULL
    // is returned
    STimeInterval *RemoveLast ()
    {
        STimeInterval *object;

        if (Empty()) return NULL;
        else
        {
            object = (STimeInterval *) GetLast();  // get the last item
            RemoveAtCurrent();  // remove the array entry
            return object;
        }
    }


    //----------------------------------------------------------------------
    //--- delete one or all objects
    //----------------------------------------------------------------------

    // delete an object; if the index exists, the object with this index
    // is deleted and True is returned; otherwise False is returned
    // current points to the position where the element was deleted (or
    // should have be inserted), or to the last element if the deleted
    // object was (would have been) the last
    EBool Delete (const CAddr& netAddress, int area)
    {
        STimeInterval *object;
        EBool found = False;

        object = Remove (netAddress, area);  // remove the object

        if (object != NULL)  // index does exist
        {
            delete object;  // delete the object
            found = True;
        }

        if (!CurrentIsValid()) GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        STimeInterval *object;

        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = (STimeInterval *) GetLast();  // get the last item
                RemoveAtCurrent();  // remove the array entry
                delete object;  // delete the object
            }

            return True;
        }
    }


protected:

    //----------------------------------------------------------------------
    //--- management of the sort key
    //----------------------------------------------------------------------

    // this function is used to compare keys
    virtual int Compare (void *key1, void *key2) const
    {
        Assert (key1 != NULL);
        Assert (key2 != NULL);

        return ((SKey*)key1)->CompareTo ((SKey*)key2);
    }

    virtual void DeleteKey (void *key) { delete (SKey*) key; }
    // delete the key


private:

    struct SKey
    {
        CAddr netAddress_;
        int area_;

        SKey (const CAddr& netAddress, int area)
        {
            netAddress_ = netAddress;
            area_ = area;
        }

        int CompareTo (const SKey *key1)
        {
            if (netAddress_ < key1->netAddress_) return -1;
            if (netAddress_ > key1->netAddress_) return +1;

            if (area_ < key1->area_) return -1;
            if (area_ > key1->area_) return +1;

            return 0;
        }
    };
};


//**************************************************************************
// class CRateSet: this class stores clock rate information with the network
//      address and area of the sending node as key
//**************************************************************************

class CRateSet : public ACObjectList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CRateSet (INT32 aSize=ArrayAllocationSize) : ACObjectList (aSize) {}

    virtual ~CRateSet () { DeleteAll (); }
    // the destructor deletes all objects in the array


    //----------------------------------------------------------------------
    //--- information about the set
    //----------------------------------------------------------------------

    INT32 Cardinality() { return ACObjectList::Occupied(); }
    // return the number of entries


    //----------------------------------------------------------------------
    //--- move through the list
    //----------------------------------------------------------------------

    SRate *GetFirst() { return (SRate *)ACObjectList::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    SRate *GetNext() { return (SRate *)ACObjectList::GetNext(); }
    // get the next object in the array or NULL if the end of the array
    // has been reached

    SRate *Find (const CAddr& netAddress, int area)
    {
        SKey key (netAddress, area);
        return (SRate *) ACObjectList::Find (&key);
    }
    // returns the object with this key; if the key does not exist,
    // NULL is returned


    //----------------------------------------------------------------------
    //--- add, remove objects
    //----------------------------------------------------------------------

    // put a new rate into the list; returns True if successful,
    // False if an entry exists
    EBool Insert (SRate *object, const CAddr& netAddress, int area)
    {
        SKey *key = new SKey (netAddress, area);

        if (!ACObjectList::Insert (object, key))  // object not inserted
        {
            delete key;  // the key is not needed
            return False;
        }
        else return True;
    }

    // remove a rate from the list; if the key does not exist, NULL
    // is returned
    SRate *Remove (const CAddr& netAddress, int area)
    {
        SKey key (netAddress, area);
        return (SRate *) ACObjectList::Remove (&key);
    }

    // remove the last rate from the list; if the list is empty, NULL
    // is returned
    SRate *RemoveLast ()
    {
        SRate *object;

        if (Empty()) return NULL;
        else
        {
            object = (SRate *) GetLast();  // get the last item
            RemoveAtCurrent();  // remove the array entry
            return object;
        }
    }


    //----------------------------------------------------------------------
    //--- delete one or all objects
    //----------------------------------------------------------------------

    // delete an object; if the index exists, the object with this index
    // is deleted and True is returned; otherwise False is returned
    // current points to the position where the element was deleted (or
    // should have be inserted), or to the last element if the deleted
    // object was (would have been) the last
    EBool Delete (const CAddr& netAddress, int area)
    {
        SRate *object;
        EBool found = False;

        object = Remove (netAddress, area);  // remove the object

        if (object != NULL)  // index does exist
        {
            delete object;  // delete the object
            found = True;
        }

        if (!CurrentIsValid()) GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        SRate *object;

        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = (SRate *) GetLast();  // get the last item
                RemoveAtCurrent();  // remove the array entry
                delete object;  // delete the object
            }

            return True;
        }
    }


protected:

    //----------------------------------------------------------------------
    //--- management of the sort key
    //----------------------------------------------------------------------

    // this function is used to compare keys
    virtual int Compare (void *key1, void *key2) const
    {
        Assert (key1 != NULL);
        Assert (key2 != NULL);

        return ((SKey*)key1)->CompareTo ((SKey*)key2);
    }

    virtual void DeleteKey (void *key) { delete (SKey*) key; }
    // delete the key


private:

    struct SKey
    {
        CAddr netAddress_;
        int area_;

        SKey (const CAddr& netAddress, int area)
        {
            netAddress_ = netAddress;
            area_ = area;
        }

        int CompareTo (const SKey *key1)
        {
            if (netAddress_ < key1->netAddress_) return -1;
            if (netAddress_ > key1->netAddress_) return +1;

            if (area_ < key1->area_) return -1;
            if (area_ > key1->area_) return +1;

            return 0;
        }
    };
};


//**************************************************************************
// class CDelaySet: this class stores transmission delays with the network
//      address of the sending node as key
//**************************************************************************

class CDelaySet : public ACObjectList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CDelaySet (INT32 aSize=ArrayAllocationSize) : ACObjectList (aSize) {}

    virtual ~CDelaySet () { DeleteAll (); }
    // the destructor deletes all objects in the array


    //----------------------------------------------------------------------
    //--- information about the set
    //----------------------------------------------------------------------

    INT32 Cardinality() { return ACObjectList::Occupied(); }
    // return the number of entries


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    SDelay *GetFirst() { return (SDelay *)ACObjectList::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    SDelay *GetNext() { return (SDelay *)ACObjectList::GetNext(); }
    // get the next object in the array or NULL if the end of the array
    // has been reached

    SDelay *Find (const CAddr& netAddress)
    {
        SKey key (netAddress);
        return (SDelay *) ACObjectList::Find (&key);
    }
    // returns the object with this key; if the key does not exist,
    // NULL is returned


    //----------------------------------------------------------------------
    //--- add, remove objects
    //----------------------------------------------------------------------

    // put a new rate into the list; returns True if successful,
    // False if an entry exists
    EBool Insert (SDelay *object, const CAddr& netAddress)
    {
        SKey *key = new SKey (netAddress);

        if (!ACObjectList::Insert (object, key))  // object not inserted
        {
            delete key;  // the key is not needed
            return False;
        }
        else return True;
    }

    // remove a rate from the list; if the key does not exist, NULL
    // is returned
    SDelay *Remove (const CAddr& netAddress)
    {
        SKey key (netAddress);
        return (SDelay *) ACObjectList::Remove (&key);
    }

    // remove the last rate from the list; if the list is empty, NULL
    // is returned
    SDelay *RemoveLast ()
    {
        SDelay *object;

        if (Empty()) return NULL;
        else
        {
            object = (SDelay *) GetLast();  // get the last item
            RemoveAtCurrent();  // remove the array entry
            return object;
        }
    }


    //----------------------------------------------------------------------
    //--- delete one or all objects
    //----------------------------------------------------------------------

    // delete an object; if the index exists, the object with this index
    // is deleted and True is returned; otherwise False is returned
    // current points to the position where the element was deleted (or
    // should have be inserted), or to the last element if the deleted
    // object was (would have been) the last
    EBool Delete (const CAddr& netAddress)
    {
        SDelay *object;
        EBool found = False;

        object = Remove (netAddress);  // remove the object

        if (object != NULL)  // index does exist
        {
            delete object;  // delete the object
            found = True;
        }

        if (!CurrentIsValid()) GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        SDelay *object;

        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = (SDelay *) GetLast();  // get the last item
                RemoveAtCurrent();  // remove the array entry
                delete object;  // delete the object
            }

            return True;
        }
    }


protected:

    //----------------------------------------------------------------------
    //--- management of the sort key
    //----------------------------------------------------------------------

    // this function is used to compare keys
    virtual int Compare (void *key1, void *key2) const
    {
        Assert (key1 != NULL);
        Assert (key2 != NULL);

        return ((SKey*)key1)->CompareTo ((SKey*)key2);
    }

    virtual void DeleteKey (void *key) { delete (SKey*) key; }
    // delete the key


private:

    struct SKey
    {
        CAddr netAddress_;

        SKey (const CAddr& netAddress)
        {
            netAddress_ = netAddress;
        }

        int CompareTo (const SKey *key1)
        {
            if (netAddress_ < key1->netAddress_) return -1;
            if (netAddress_ > key1->netAddress_) return +1;

            return 0;
        }
    };
};


#endif


