//
// datatemp.hpp
//
// lists of CTimeSet, CRateSet and CDelaySet classes, all stored with the
// key <networkID>
//
// created: 14.11.96
//
// last change: 01.05.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _DATATEMP_HPP_
#define _DATATEMP_HPP_

#include "listbase.hpp"
#include "dataset.hpp"


//**************************************************************************
// class CTimeSetList: is used to store the interval sets of a clock
//      synchronization module; the key is <networkID>
//**************************************************************************

class CTimeSetList : public ACObjectList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CTimeSetList (INT32 aSize = ArrayAllocationSize) :
        ACObjectList (aSize) {};
    // allocate memory

    virtual ~CTimeSetList () { DeleteAll(); }
    // the destructor deletes all objects in the array


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    CTimeSet* GetFirst () { return (CTimeSet*) ACObjectList::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    CTimeSet* GetLast () { return (CTimeSet*) ACObjectList::GetLast(); }
    // get the last object or NULL if the array is empty

    CTimeSet* GetNext () { return (CTimeSet*) ACObjectList::GetNext(); }
    // get the next object in the array or the last object if the end of
    // the array has been reached

    CTimeSet* GetPrev () { return (CTimeSet*) ACObjectList::GetPrev(); }
    // return the previous object in the array or the first if the beginning
    // of the array has been reached

    CTimeSet* Find (SINT32 networkID) // put the keys here
    {
        SKey key (networkID);
        return (CTimeSet*) ACObjectList::Find (&key);
    }
    // returns the object with this key; if the key does not exist,
    // NULL is returned


    //----------------------------------------------------------------------
    //--- get information about the current element
    //----------------------------------------------------------------------

    CTimeSet* GetCurrentObject ()
        { return (CTimeSet*) ACObjectList::GetCurrentObject(); }
    // returns a pointer to the member that is currently being accessed or
    // NULL, if no element is being accessed

    EBool GetCurrentObjectKey (INT32& networkID)
    // references to the key variables
    {
        SKey *key = (SKey*) ACObjectList::GetCurrentObjectKey();

        if (key != NULL)
        {
            networkID = key->networkID_;
            // set the argument variables
            return True;
        }
        else return False;
    }
    // returns True if a member is accessed (in that case the networkID will
    // contain the key), or False if no element is currently accessed


    //----------------------------------------------------------------------
    //--- add/remove objects
    //----------------------------------------------------------------------

    EBool Insert (CTimeSet* object, SINT32 networkID)
    {
        SKey *key = new SKey (networkID);
        EBool inserted;

        inserted = ACObjectList::Insert (object, key);

        if (!inserted)  // the object was not inserted
        {
            delete key;  // the key is not needed
        }

        return inserted;
    }

    CTimeSet* Remove (SINT32 networkID)
    {
        SKey key (networkID);
        return (CTimeSet*) ACObjectList::Remove (&key);
    }
    // remove an object from the array; if the key does not exist, NULL
    // is returned



    //----------------------------------------------------------------------
    //--- delete one or all member(s) of the array
    //----------------------------------------------------------------------

    // delete an object; if the index exists, the object with this index
    // is deleted and True is returned; otherwise False is returned
    // current points to the position where the element was deleted (or
    // should have be inserted), or to the last element if the deleted
    // object was (would have been) the last
    EBool Delete (SINT32 networkID)
    {
        CTimeSet* object;
        EBool found = False;

        object = Remove (networkID);  // remove the object

        if (object != NULL)  // index does exist
        {
            delete object;  // delete the object

            found = True;
        }

        if (!CurrentIsValid()) object = GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // delete all members of the array
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        CTimeSet* object;
        
        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = GetLast();  // get the last item in the array
                RemoveAtCurrent();  // remove the array entry
                delete object;  // delete the object
            }

            return True;
        }
    } // DeleteAll


protected:

    // this function is used to compare keys
    virtual int Compare (void* key1, void *key2) const
    {
        if (((SKey*)key1)->networkID_ < ((SKey*)key2)->networkID_)
            return -1;
        if (((SKey*)key1)->networkID_ > ((SKey*)key2)->networkID_)
            return +1;
        return 0;
    }


    virtual void DeleteKey (void* key) { delete (SKey*) key; }
    // delete the key


private:
    struct SKey  // this structure contains the key(s) for sorting
    {
        // all key elements are put here
        SINT32 networkID_;

        SKey (SINT32 networkID) {networkID_ = networkID;}
        // initialize the members
    };
};

//**************************************************************************
// class CRateSetList: is used to store the rate information of the clocks;
//      the key is <networkID>
//**************************************************************************

class CRateSetList : public ACObjectList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CRateSetList (INT32 aSize = ArrayAllocationSize) :
        ACObjectList (aSize) {};
    // allocate memory

    virtual ~CRateSetList () { DeleteAll(); }
    // the destructor deletes all objects in the array


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    CRateSet* GetFirst () { return (CRateSet*) ACObjectList::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    CRateSet* GetLast () { return (CRateSet*) ACObjectList::GetLast(); }
    // get the last object or NULL if the array is empty

    CRateSet* GetNext () { return (CRateSet*) ACObjectList::GetNext(); }
    // get the next object in the array or the last object if the end of
    // the array has been reached

    CRateSet* GetPrev () { return (CRateSet*) ACObjectList::GetPrev(); }
    // return the previous object in the array or the first if the beginning
    // of the array has been reached

    CRateSet* Find (SINT32 networkID) // put the keys here
    {
        SKey key (networkID);
        return (CRateSet*) ACObjectList::Find (&key);
    }
    // returns the object with this key; if the key does not exist,
    // NULL is returned


    //----------------------------------------------------------------------
    //--- get information about the current element
    //----------------------------------------------------------------------

    CRateSet* GetCurrentObject ()
        { return (CRateSet*) ACObjectList::GetCurrentObject(); }
    // returns a pointer to the member that is currently being accessed or
    // NULL, if no element is being accessed

    EBool GetCurrentObjectKey (INT32& networkID)
    // references to the key variables
    {
        SKey *key = (SKey*) ACObjectList::GetCurrentObjectKey();

        if (key != NULL)
        {
            networkID = key->networkID_;
            // set the argument variables
            return True;
        }
        else return False;
    }
    // returns True if a member is accessed (in that case the networkID will
    // contain the key), or False if no element is currently accessed


    //----------------------------------------------------------------------
    //--- add/remove objects
    //----------------------------------------------------------------------

    EBool Insert (CRateSet* object, SINT32 networkID)
    {
        SKey *key = new SKey (networkID);
        EBool inserted;

        inserted = ACObjectList::Insert (object, key);

        if (!inserted)  // the object was not inserted
        {
            delete key;  // the key is not needed
        }

        return inserted;
    }

    CRateSet* Remove (SINT32 networkID)
    {
        SKey key (networkID);
        return (CRateSet*) ACObjectList::Remove (&key);
    }
    // remove an object from the array; if the key does not exist, NULL
    // is returned



    //----------------------------------------------------------------------
    //--- delete one or all member(s) of the array
    //----------------------------------------------------------------------

    // delete an object; if the index exists, the object with this index
    // is deleted and True is returned; otherwise False is returned
    // current points to the position where the element was deleted (or
    // should have be inserted), or to the last element if the deleted
    // object was (would have been) the last
    EBool Delete (SINT32 networkID)
    {
        CRateSet* object;
        EBool found = False;

        object = Remove (networkID);  // remove the object

        if (object != NULL)  // index does exist
        {
            delete object;  // delete the object

            found = True;
        }

        if (!CurrentIsValid()) object = GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // delete all members of the array
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        CRateSet* object;
        
        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = GetLast();  // get the last item in the array
                RemoveAtCurrent();  // remove the array entry
                delete object;  // delete the object
            }

            return True;
        }
    } // DeleteAll


protected:

    // this function is used to compare keys
    virtual int Compare (void* key1, void *key2) const
    {
        if (((SKey*)key1)->networkID_ < ((SKey*)key2)->networkID_)
            return -1;
        if (((SKey*)key1)->networkID_ > ((SKey*)key2)->networkID_)
            return +1;
        return 0;
    }


    virtual void DeleteKey (void* key) { delete (SKey*) key; }
    // delete the key


private:
    struct SKey  // this structure contains the key(s) for sorting
    {
        // all key elements are put here
        SINT32 networkID_;

        SKey (SINT32 networkID) {networkID_ = networkID;}
        // initialize the members
    };
};

//**************************************************************************
// class CDelaySetList: is used to store the transmission times of the
//      nodes of a particular network; the key is <networkID>
//**************************************************************************

class CDelaySetList : public ACObjectList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CDelaySetList (INT32 aSize = ArrayAllocationSize) :
        ACObjectList (aSize) {};
    // allocate memory

    virtual ~CDelaySetList () { DeleteAll(); }
    // the destructor deletes all objects in the array


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    CDelaySet* GetFirst () { return (CDelaySet*) ACObjectList::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    CDelaySet* GetLast () { return (CDelaySet*) ACObjectList::GetLast(); }
    // get the last object or NULL if the array is empty

    CDelaySet* GetNext () { return (CDelaySet*) ACObjectList::GetNext(); }
    // get the next object in the array or the last object if the end of
    // the array has been reached

    CDelaySet* GetPrev () { return (CDelaySet*) ACObjectList::GetPrev(); }
    // return the previous object in the array or the first if the beginning
    // of the array has been reached

    CDelaySet* Find (SINT32 networkID) // put the keys here
    {
        SKey key (networkID);
        return (CDelaySet*) ACObjectList::Find (&key);
    }
    // returns the object with this key; if the key does not exist,
    // NULL is returned


    //----------------------------------------------------------------------
    //--- get information about the current element
    //----------------------------------------------------------------------

    CDelaySet* GetCurrentObject ()
        { return (CDelaySet*) ACObjectList::GetCurrentObject(); }
    // returns a pointer to the member that is currently being accessed or
    // NULL, if no element is being accessed

    EBool GetCurrentObjectKey (INT32& networkID)
    // references to the key variables
    {
        SKey *key = (SKey*) ACObjectList::GetCurrentObjectKey();

        if (key != NULL)
        {
            networkID = key->networkID_;
            // set the argument variables
            return True;
        }
        else return False;
    }
    // returns True if a member is accessed (in that case the networkID will
    // contain the key), or False if no element is currently accessed


    //----------------------------------------------------------------------
    //--- add/remove objects
    //----------------------------------------------------------------------

    EBool Insert (CDelaySet* object, SINT32 networkID)
    {
        SKey *key = new SKey (networkID);
        EBool inserted;

        inserted = ACObjectList::Insert (object, key);

        if (!inserted)  // the object was not inserted
        {
            delete key;  // the key is not needed
        }

        return inserted;
    }

    CDelaySet* Remove (SINT32 networkID)
    {
        SKey key (networkID);
        return (CDelaySet*) ACObjectList::Remove (&key);
    }
    // remove an object from the array; if the key does not exist, NULL
    // is returned



    //----------------------------------------------------------------------
    //--- delete one or all member(s) of the array
    //----------------------------------------------------------------------

    // delete an object; if the index exists, the object with this index
    // is deleted and True is returned; otherwise False is returned
    // current points to the position where the element was deleted (or
    // should have be inserted), or to the last element if the deleted
    // object was (would have been) the last
    EBool Delete (SINT32 networkID)
    {
        CDelaySet* object;
        EBool found = False;

        object = Remove (networkID);  // remove the object

        if (object != NULL)  // index does exist
        {
            delete object;  // delete the object

            found = True;
        }

        if (!CurrentIsValid()) object = GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // delete all members of the array
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        CDelaySet* object;
        
        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = GetLast();  // get the last item in the array
                RemoveAtCurrent();  // remove the array entry
                delete object;  // delete the object
            }

            return True;
        }
    } // DeleteAll


protected:

    // this function is used to compare keys
    virtual int Compare (void* key1, void *key2) const
    {
        if (((SKey*)key1)->networkID_ < ((SKey*)key2)->networkID_)
            return -1;
        if (((SKey*)key1)->networkID_ > ((SKey*)key2)->networkID_)
            return +1;
        return 0;
    }


    virtual void DeleteKey (void* key) { delete (SKey*) key; }
    // delete the key


private:
    struct SKey  // this structure contains the key(s) for sorting
    {
        // all key elements are put here
        SINT32 networkID_;

        SKey (SINT32 networkID) {networkID_ = networkID;}
        // initialize the members
    };
};

#endif


