//
// DISTRIB.HPP - Distribution Classes
//
// Random Number Generators for 32-bit Machines
//
// created: ??.??.98, Gerald Hummel
//
// last change: 25.05.98, Gerald Hummel
//
// last revision: 02.09.98
//
// RCS version control information:
//     $Revision: 1.16 $$State: Exp $$Date: 1998/09/02 16:04:33 $
//


#ifndef _DISTRIB_HPP_
#define _DISTRIB_HPP_

#include "general.hpp"


const int SizeOfSeries = 128;
// size of the 'series' array
// do not change that value; it was taken from the original generator


//**************************************************************************
// data structure for drawing random numbers
//**************************************************************************

// this structure contains all data needed to draw a random number within
// a certain range according to certain parameters
struct SRandomData
{
    EDistribution dist;  // the distribution function

    double low, high;  // the interval of the random number
    double mean;       // the mean value; it should be within [low, high]
    double freeparam;  // a second parameter which is needed for some dists
    double constant;   // the constant value that is added to the random var

    SRandomData() {dist=Uniform; low=high=mean=0; freeparam=0; constant=0;}
    // sets all members to default values

    EBool DataValid() const;
    // checks if low>=0, low+constant>=0, low<=mean<=high; if all three
    // conditions are fulfilled, then the function returns True, otherwise
    // False
};


//**************************************************************************
// distribution functions
//**************************************************************************

class distributions
{
public:
  distributions (long MGSeed=772531L, long LCGSeed=1878892440L);
  virtual void SetSeeds (long mgSeed, long lcgSeed, long distance);
  virtual void GetSeeds (long& mgSeed, long& lcgSeed,long& distance) const;
 
protected:
  double lo,low,hi,range;	   /* lo,hi of destination interval */	
  double GetUniform (); // returns a uniform random variable in (low, high)

private:
  double MGen ();
  double series[SizeOfSeries];
  long MSeed, LSeed;
  long oldMSeed, oldLSeed;
  long difference;
  long MSeedGen (long mSeed) const;
  long LSeedGen (long lSeed) const;
};

//------------------------- U N I F O R M ---------------------------------

class CUniform : public distributions
{
public:

// Constructor -> set bounds:
  CUniform (double low, double high, int streampos=0);

// get value:
  virtual double operator() ();

// manipulate lower and upper bound
  virtual void SetRange (double c, double low, double high);
  
protected:
  double con;
};

//---------------------- E X P O N E N T I A L -----------------------------

class C_exponential: public distributions
{
public:

// Constructor -> set bounds and mean:
  C_exponential(double c, double low, double high, double E, int streampos=0);

// get value:
  virtual double operator() ();

// change specifications -> same as constructor:
  virtual void set(double c, double low, double high, double E);

protected:		
  double mean,limit,con;	
};

//-------------------------  N O R M A L -----------------------------------

class C_normal: public distributions
{
public:

// Constructor -> set bounds, mean and standard deviation: 
  C_normal(double c, double low, double high, double E, double stdev,
           int streampos=0);

// get value:
  virtual double operator() ();	

// change specifications -> same as constructor:
  virtual void set(double c, double low, double high, double E,
		   double stdev);

protected:
  int buf_idx; /* buf_idx points at the last entry */
  double norm_var_buf[10];     /* buffer for 2nd calculated value v2 */
  double mean,stdev,con;
};	

//--------------------------- G A M M A ------------------------------------

class C_gamma: public distributions
{
public:

// Constructor -> set bounds and mean:
  C_gamma(double c, double low, double high, double mean, int streampos=0);

// get value:
  virtual double operator() ();

// change specifications -> same as constructor:
  virtual void set(double c, double low, double high, double mean);

protected:
  int flag;
  double mean,con;
};


//-------------------------- W E I B U L L ---------------------------------

class C_weibull: public distributions
{
public:

// Constructor -> set bounds and mean, create weibull table:
  C_weibull(double c, double low, double high, double mean, int streampos=0);

// get_value:
  virtual double operator() ();

// change_specifications -> change bounds or mean:
  virtual void set(double c, double low, double high, double mean);

protected:
  double l,a,weibull_limit;/* lambda, alpha, upper bound for computation */
  double mean,con;	
};

//--------------------------- G E N E R I C ---------------------------------

class C_generic: public distributions
{
public:

// Constructor -> set table-pointer to NULL
  C_generic(int streampos=0);

// get_value:
  virtual double operator() ();

// select a special table containing function values:
  virtual void set(double c, double low, double high,
		   double *inv_funct, unsigned long tabsize);

protected:
  double *inv_funct;	/* pointer to start address of (X,Y)-table */
  long tabsize;		/* length of table */
  double con;
};
 
//--------------------------- R A N D O M  ----------------------------------

class CRandom: public distributions
{
public:

// Constructor
  CRandom(int streampos=0);

// Access to the Generator:
  EBool CastDice (double prob);
  // returns 'True' with probability 'prob'
  // assert: 'prob' is in [0, 1]

  double RandomVar(EDistribution dist);

// Access to the Generator, for each new value perform a new set:
  double RandomVar(EDistribution dist, double c, double low,
		   double high, double mean, double free_par=0);
  double RandomVar(SRandomData& data);
  double RandomVar(double low, double high);  // return uniform value

// Change Settings:
  void Set(EDistribution dist, double c, double low, double high,
	   double mean, double free_par=0);	

// Table-selector for generic-function:
  void SetGeneric(double c, double low, double high,
                  double *inv_funct, unsigned long tabsize);

// manipulate the seeds directly:
  virtual void SetSeeds (EDistribution dist, long mgSeed, long lcgSeed,
                         long distance);
// takes over the seeds for the generator;
// PLEASE: only use values that were obtained through GetSeeds(), since
// it may reduce the quality of the generator if arbitrary values are
// loaded
// assert: mgSeed, ldgSeed > 0
// assert: the distribution is supported

  virtual void GetSeeds (EDistribution dist, long& mgSeed,
                         long& lcgSeed, long& distance) const;
// return the seeds of the generators
// assert: the distribution is supported
 
private:
  CUniform	uniform;
  C_exponential exponential;
  C_normal	normal;
  C_gamma	gamma;
  C_weibull	weibull;
  C_generic	generic;
};

#endif

