//
// duplist.hpp
//
// declaration of a list class using arrays; optimized for repeated
// access and sequential access; use binary search for other access;
// duplicates are possible
// the sort key is of type double
//
// created: 25.07.96
//
// last change: 15.11.96
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _DUPLIST_HPP_
#define _DUPLIST_HPP_

#include <stdlib.h>  // for the NULL macro

#include "bool.hpp"
#include "constant.hpp"


//**************************************************************************
// class ACSortedArray: this abstract class manages a sorted array
//     it is assumed that most of the time the access is sequential; if
//     not, binary search is used; duplicates are possible
//**************************************************************************

class ACSortedArray
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    ACSortedArray (EBool duplicates = False, INT32 aSize = ArrayAllocationSize);
    // allocate memory

    virtual ~ACSortedArray () =0;
    // pure virtual, but nevertheless it deletes the array_


    //----------------------------------------------------------------------
    //--- get the array status
    //----------------------------------------------------------------------

    EBool Empty() const;
    // returns True if the array is empty, False otherwise
    
    EBool BeginOfList () const;
    // returns True if current_ is at the beginning of
    // the array, False otherwise (or if the array is empty)

    EBool EndOfList () const;
    // returns True if current_ is at the end of the
    // array, False otherwise (or if the array is empty)


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    void* GetFirst ();
    // get the first object in the array, or NULL if the array is empty

    void* GetLast ();
    // get the last object or NULL if the array is empty

    void* GetNext ();
    // get the next object in the array or NULL if the end of the array
    // has been reached

    void* GetPrev ();
    // return the previous object in the array or NULL if the beginning
    // of the array has been reached

    void* Find (double index);
    // returns the object with this index (the first in the array
    // if duplicates are allowed); if no such object does exist,
    // NULL is returned

    void* FindNext ();
    // returns the next object with the same index as the current_ index;
    // if no more objects with this index exist, NULL is returned and
    // the function has no effect

    //----------------------------------------------------------------------
    //--- get information about the current element
    //----------------------------------------------------------------------

    void* GetCurrentObject () const;
    // returns a pointer to the object that is currently being accessed or
    // NULL, if no element is being accessed

    double GetCurrentObjectID () const;
    // returns the object ID of the currently accessed element or NoAccount,
    // if no element is being accessed


    //----------------------------------------------------------------------
    //--- add/remove objects
    //----------------------------------------------------------------------

    EBool Insert (void *object, double index);
    // insert a new element into the (sorted) array;
    // if an object with this index exists but duplicates are not allowed,
    // False is returned
    // if duplicates are allowed and elements with the index exist, the new
    // element is inserted as the first

    void* Remove (double index);
    // remove an object; if duplicates are allowed, the first object in the
    // array with the appropriate index will be removed
    // if the index does not exist, NULL is returned


protected:

    //----------------------------------------------------------------------
    //--- internal information concerning the current element
    //----------------------------------------------------------------------

    EBool CurrentIsValid () const;
    // returns True if current_ is an occupied slot, False otherwise

    double GetCurrentID () const;
    // returns the object ID of the current element; the current element
    // must exist!

    void* GetCurrent () const;
    // returns the object of the current element; the current element must
    // exist!

    
    //----------------------------------------------------------------------
    //--- internal memory management
    //----------------------------------------------------------------------

    void UpdateArray();
    // allocates a new array of size size_ and copies the data of firstFree_
    // elements from array_ to the new array; array_ is then deleted and set
    // to the new array
    // assert: firstFree_ <= size_

    void AllocateMemory();
    // allocates a new array that is allocSize_ elements bigger than the
    // current array size stored in size_; the data is copied to the new
    // array and array_ is set to it; the old array is discarded

    void FreeMemory();
    // frees the last allocSize_ elements of the current array; note that
    // the function will assume that these elements are not in use!
    // if they are, an assert will fail!


    //----------------------------------------------------------------------
    //--- internal array information
    //----------------------------------------------------------------------

    INT32 Size() const;
    // returns the current size of the array

    INT32 Occupied() const;
    // returns the number of array elements that are currently in use

    INT32 AllocationSize() const;
    // returns the additional size used for expansion (contraction) of
    // the array

    EBool Duplicates() const;
    // returns True if duplicate entries are allowed, False otherwise


    //----------------------------------------------------------------------
    //--- internal array manipulation
    //----------------------------------------------------------------------

    void InsertAtCurrent (void *object, double index);
    // insert the data at the current_ index; shifts the rest of the
    // array one element to the right; there has to be at least one free
    // element in the array
    // if the array is full (after insertion), new memory is allocated
    
    void RemoveAtCurrent ();
    // removes the data at the current_ index; shifts the rest of the
    // array one element to the left
    // if more than allocSize_+1 elements are unused, memory is freed


private:

    //**********************************************************************
    // class CArrayElement: this class represents an element of the sorted array;
    //                      it contains the data that is stored in the array
    //**********************************************************************

    class CArrayElement
    {
    public:

        //------------------------------------------------------------------
        //--- constructor
        //------------------------------------------------------------------

        CArrayElement ();  // constructor to initialize the pointer


        //------------------------------------------------------------------
        //--- set the members of the class
        //------------------------------------------------------------------

        void SetObject (void *object, double objectID);


        //------------------------------------------------------------------
        //--- get the members of the class
        //------------------------------------------------------------------

        void* GetObject () const;
        double GetObjectID () const;


    private:

        double objectID_;  // this is the sorting criteria for the array
        void*  object_;    // pointer to the object that is stored
    };


    CArrayElement *array_;  // pointer to the beginning of the sorted array
    EBool duplicates_;  // True if duplicate entries are allowed
    
    INT32 allocSize_;  // number of new elements if array is expanded
    INT32 size_;  // size of the array

    INT32 firstFree_;  // index of the first free entry
    INT32 current_;   // index of the currently accessed entry
};

#endif


