//
// error.hpp
//
// declaration of the global error report functions that should be used
// in the ReportError() function of a class, plus the macros that should
// be used for error handling: Assert(), Check() and Report()
// if NDEBUG is defined, the macros will expand to (void)0
//
// IMPORTANT: never call Check() or Assert() with expressions that have to
//	be evaluated even if error handling is turned off! Expressions will
//	*not* get evaluated if error handling is turned off (NDEBUG is
//	defined)!
//
// note: if you wish to see the error output, you have to install an error
// report object using InstallErrorHandler(); in that case, please call
// DeinstallErrorHandler() before deleting the error report object
// (BTW: take care in hardware when calling these functions: they should
// never interrupt another task/isr calling ReportError(), since access to
// the internal error reporter pointer is not atomic for the ReportError
// functions!)
//
// note: a function ReportMessage() is provided to print messages without
//	any additional information
//
// created: 06.01.97
//
// last change: 05.02.99
//
// RCS version control information:
//     $Revision: 1.3 $$State: Exp $$Date: 1999/02/05 11:51:11 $
//


#ifndef _ERROR_HPP_
#define _ERROR_HPP_

#include "psos.h"
#include "bool.hpp"
#include "params.hpp"


//--------------------------------------------------------------------------
// use these functions to (de-)install an error report object; if none is
// installed, the macros will still behave as described below (Assert()
// causes program abortion, Check() returns the value of the expression),
// but there won't be any output
//--------------------------------------------------------------------------

class CErrorHandler;  // base class of all error report objects

void InstallErrorHandler (CErrorHandler *handler);
// set the internal error handler pointer to the handler
// context: Task

void DeinstallErrorHandler ();
// set the internal error handler pointer to NULL
// context: Task


//--------------------------------------------------------------------------
// the following functions report the expression plus file and line number
// of the call to the error reporter (if installed); if they return (which
// is only the case if type != Asserting), they always return 0
// please call these functions from the ReportError() function of the class
//--------------------------------------------------------------------------

enum EReportType {Asserting, Checking, Reporting};
// Asserting stops program execution, the others don't


int ReportError (EReportType type, char *file, int line, char *expression);
int ReportError (EReportType type, char *file, int line, char *expression,
                 EModule module);
int ReportError (EReportType type, char *file, int line, char *expression,
                 EModule module, SINT32 moduleID);
int ReportError (EReportType type, char *file, int line, char *expression,
                 EModule module, SINT32 nodeID, int area);
// context: ISR, Task


//--------------------------------------------------------------------------
// the following function just reports a message
//--------------------------------------------------------------------------

void ReportMessage (const char *msg);
// context: ISR, Task


//--------------------------------------------------------------------------
// the following macros are used for error checks and reports of all types
// (asserts, checks or simple reports); don't use anything else to ensure
// consistent handling of error reports
//--------------------------------------------------------------------------

#ifdef NDEBUG

#define Assert(expr) ((void)0)
#define Check(expr) ((void)0)
#define Report(expr) ((void)0)

#else

// the following three macros use the fact that they are expanded and can
// use the context of the caller: so a class using those macros can (but
// need not) overload the error report function:
// int ReportError (EBool abortType, char *file, int line, char *expression);
//
// overloading the global function allows an object to report additional
// information like its class and its identification (use the different
// ReportError() functions to report additional information, *never* use
// any other report functions than those defined in this file)

// calling Assert() stops program execution if expr evaluates to 0, has
// no "return" value
#define Assert(expr) ((expr)?(void)0: \
    (void)ReportError (Asserting, __FILE__, __LINE__, #expr))

// calling Check() never stops program execution; it returns 1 if expr is
// true, and 0 (return value of ReportError) if expr is false
#define Check(expr) ((expr)?1: \
    ReportError (Checking, __FILE__, __LINE__, #expr))

// calling Report() simply reports the condition, has no return value
#define Report(expr) ( \
    (void)ReportError (Reporting, __FILE__, __LINE__, #expr))

// remember: these are macros, macros are considered A Bad Thing, so use
// them with care (but anyway, these are okay, they don't evaluate anything
// more than once); but note that Report() *does not* evaluate!
#endif

#endif

