//
// general.hpp
//
// some general declarations of structures, classes and global functions
//
// created: 29.06.96
//
// last change: 19.01.99
//
// RCS version control information:
//     $Revision: 1.12 $$State: Exp $$Date: 1999/01/19 14:39:06 $
//


#ifndef _GENERAL_HPP_
#define _GENERAL_HPP_

#include "psos.h"

#include "bool.hpp"  // boolean type EBool


class CAddr;
class CClock;

struct SMsgPosition;
struct STimestamp;


//**************************************************************************
// some general declarations
//**************************************************************************

// possible distribution functions
// the comment lists the parameters needed in addition to constant, low,
// high (in struct SRandomData)
enum EDistribution
{
    Uniform=0,    // no params
    Normal,       // mean, freeparam
    Exponential,  // mean
    Gamma,        // mean
    Weibull,      // mean
    Generic,
    EDistributionCount
};
// the last entry EDistributionCount is used by Random() to determine the
// number of distributions; it has to stay last even if new distributions
// are added!


// possible network types
enum ENetworkType
{
    Broadcast = 0,  // broadcast network (may transmit dedicated messages)

    ENetworkTypeCount
};
// the last entry EDistributionCount is used for accounting purposes
// and has to stay last even if new items are added!


// possible algorithm types
enum EAlgorithmType
{
    FreeRun = 0,  // no synchronization at all, lets the clock drift
    FaultTolerantAverage,  // fault tolerant average (takes reference point)
    Midpoint,  // like fault tolerant average (but takes midpoint)
    Fetzer,  // like FTA, but P-Nodes are handled separatly
    OP,  // Optimal Precision
    Marzullo,  // Marzullo-Algorithm
    Rate,  // Optimal Precision with rate synchronization
    FaultTolerantIntersection,  // FTI algorithm

    EAlgorithmTypeCount
};
// the last entry EDistributionCount is used for accounting purposes
// and has to stay last even if new items are added!


//**************************************************************************
// global functions for dis-/enabling interrupts
//**************************************************************************

extern INT32 DisableInterrupts ();
// disables all interrupts, returns old value of the task mode

extern void EnableInterrupts (INT32 tmode);
// set interrupt mask to the 'tmode' value


//**************************************************************************
// global random functions
//**************************************************************************

extern EBool CastDice (double prob);
// returns True with the probability prob, False otherwise
// assert: the probability must be in [0, 1]


//**************************************************************************
// get global time and snapshot time
//**************************************************************************

extern double GetGlobalTime ();
// returns the master time (in seconds)
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp

extern double GetSnapshotTime ();
// returns the master time at the last snapshot (in seconds)
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp


//**************************************************************************
// get the transmit timestamp
//**************************************************************************

extern void GetTransmitTimestamp (const CAddr& address, int area, const
                                  CClock *clock, STimestamp& time);
// for measuring the transmission characteristics of the network, each node
// triggers (upon sending a timestamped message) the sampling of the clock
// value into a register dedicated to this node (and area of the sender);
// this function should simply return the macro- and timestamp of the
// UTCSU registers corresponding to the sender with the given address/area
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp


//**************************************************************************
// get the position of network time value in message
//**************************************************************************

extern void SetPosition (const SMsgPosition& data, SINT32 nodeID, SINT32 networkID);
// store the length and position of all data that is automatically added
// to a message by the network into the appropriate fields of 'data'
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp

extern void GetPosition (SMsgPosition& data, SINT32 nodeID, SINT32 networkID);
// puts the length and position of all data that is automatically added
// to a message by the network into the appropriate fields of 'data'
// if there is no data stored, you should provide default position data
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp


//**************************************************************************
// set/get the clock frequencies
//**************************************************************************

extern void SetClockFrequency (SINT32 nodeID, int area, double frequency);
// stores the frequency of a clock; the frequencies of all clocks should be
// set after the program is started (when calling the other initialization
// functions
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp

extern double GetClockFrequency (SINT32 nodeID, int area);
// returns the frequency of a clock; the function is called by the Controller
// whenever a new CSA is created
// if no frequency is stored, 25 MHz is returned
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp


//**************************************************************************
// trigger hardware snapshots on all UTCSUs
//**************************************************************************

extern void MakeSystemSnapshot();
// initiates a hardware snapshot on all UTCSUs in the system (that assumes
// that all UTCSUs are hardwired together, regardless of their subnets)
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp


//**************************************************************************
// convert integer to string
//**************************************************************************

#ifndef DOS_SIM
// itoa() is not ANSI, so it is only available when using Watcom (DOS)

extern char *itoa (int value, char *buffer, int radix);
// converts the value "value" in base "radix" to a string and writes the
// result into "buffer"; a pointer to that buffer is returned
// if radix < 2 or radix > 36, NULL is returned and nothing happens

#endif


//**************************************************************************
// enumerations needed for fault injection
//**************************************************************************

enum EArrival {Early = 0, InTime, Late, EArrivalCount};
// possibilities for msg arrival; the last entry is for accounting purposes
// only and has to stay last even if new items are added!
// used by: CNetworkAccountant

// DEBUG
//enum EProbabilityType {PCrash = 0, PRecover, PEarly, PLate};
// probability types that can be accessed by the evaluation
// used by: all classes derived from CCrashingModule, CNetworkAccountant

#endif

