//
// gpsbase.hpp
//
// declaration of the (abstract) base class of the GPS handling class
// note: implement the virtual functions in the derived class to obtain
// a working class; implement CGpsTemplate::NewGPS() to get a new instance
// of the currently used GPS class
// cf. to gps_sw.hpp/cpp for an example
//
// created: 30.03.97
//
// last change: 09.01.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/07/23 08:55:28 $
//


#ifndef _GPSBASE_HPP_
#define _GPSBASE_HPP_


#include "bool.hpp"
#include "clockcfg.hpp"
#include "interupt.hpp"


class CClockInterface;
class CNtpTime;

struct SEvent;


//**************************************************************************
// class ACGps: abstract base class providing an ISR for the GPS1PPS
//      interrupt of the UTCSU; it modifies the clock time according to the
//      time obtained from the GPS receiver
//**************************************************************************

class ACGps : public CNeedInterrupt
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    ACGps (CClockInterface *clock, int area);
    // stores the clock that is used, and its area
    // assert: clock != NULL

    virtual ~ACGps ();
    // assert: 'enableMask' == 0


    //----------------------------------------------------------------------
    //--- managing GPS receivers
    //----------------------------------------------------------------------

    EBool GpsEnabled();
    // returns True if at least one GPS receiver is enabled ('enableMask'
    // is different from zero), otherwise False


    //----------------------------------------------------------------------
    //--- enabling/disabling interrupts
    //----------------------------------------------------------------------

    virtual void EnableGps (EGpsUnit gpu) =0;
    // enable the interrupt of the GPS receiver that is triggered by the
    // end of the GPS time information

    virtual void DisableGps (EGpsUnit gpu) =0;
    // disable the interrupt of the GPS receiver that is triggered by the
    // end of the GPS time information


protected:

    CClockInterface *clock_;  // the associated clock
    int area_;  // the area of the clock
    INT8 enableMask_;  // the GPUs that are connected to enabled receivers
    INT8 statusMask_;  // status of connected receivers (bit=1 -> healthy)


#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- handling GPS information
    //----------------------------------------------------------------------

    virtual void GetGpsTime (CNtpTime& time) =0;
    // obtain the NTP time of the GPS receiver that has triggered the
    // interrupt; the function is called from the ISR

    virtual EGpsUnit GetGpsNumber() =0;
    // return the identifier of the GPU to which the receiver that has
    // triggered the interrupt is connected; the function is called from
    // the ISR

    virtual void ClearGpsInterrupt() =0;
    // clear the interrupt that was triggered by the end of the GPS time
    // information; the function is called from the ISR


    //----------------------------------------------------------------------
    //--- GPS status
    //----------------------------------------------------------------------

    virtual void SetGpsStatus (EGpsUnit gpu, EBool status);
    // notifies the clock interface of the GPS status; set 'status' to
    // True if the GPS receiver is healthy, to False if it is not
    // assert: the GPS receiver of unit 'gpu' is enabled


    //----------------------------------------------------------------------
    //--- internal event reports
    //----------------------------------------------------------------------

    virtual void ReportEvent (SEvent *event);
    // pass the event to the clock (also set module and type of the event)
    // assert: event != NULL


private:
#ifdef PUBLIC_IN_TEST
public:
#endif

    //----------------------------------------------------------------------
    //--- get clock time (GPS)
    //----------------------------------------------------------------------

    void GetClockGps (EGpsUnit gpu, CNtpTime& clockTime);
    // reads the MSGPS and TSGPS registers of the clock; uses the GPU
    // specified in variable 'gpu'
    // assert: gpu < EGpsUnitCount


    //----------------------------------------------------------------------
    //--- GPS ISRs
    //----------------------------------------------------------------------

    virtual void isr();
    // called after the time information from GPS is available; reads the
    // GPS time and compares it with the clock value stored; if the value
    // difference is more than 'GpsEpsilon', the clock is amortized
};



//**************************************************************************
// class CGpsTemplate: the template class used for obtaining an instance
//      of the urrently used GPS class (software or hardware)
//**************************************************************************

class CGpsTemplate
{
public:

    //----------------------------------------------------------------------
    //--- return a new instance of the class
    //----------------------------------------------------------------------

    virtual ACGps *NewGPS (CClockInterface *clock, int area);
    // returns a new object of the class; this function is needed so that
    // CClockInterface can obtain the correct GPS class
    // implementation of the function is left open to the programmer of
    // the GPS class
};


#endif


