//
// idlist.hpp
//
// list of nodeIDs
//
// created: 18.10.96
//
// last change: 07.02.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _IDLIST_HPP_
#define _IDLIST_HPP_

#include "psos.h"

#include "constant.hpp"


//**************************************************************************
// class CNodeIdList: this class stores a sorted array of SINT32
//     it is assumed that most of the time the access is sequential; if
//     not, binary search is used
//**************************************************************************

class CNodeIdList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CNodeIdList (INT32 aSize = ArrayAllocationSize);
    // allocate memory
    
    virtual ~CNodeIdList ();
    // delete the memory


    //----------------------------------------------------------------------
    //--- get the array status
    //----------------------------------------------------------------------

    EBool Empty() const;
    // returns True if the array is empty, False otherwise
    
    EBool BeginOfList () const;
    // returns True if _current is at the beginning of
    // the array, False otherwise (or if the array is empty)

    EBool EndOfList () const;
    // returns True if _current is at the end of the
    // array, False otherwise (or if the array is empty)


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    SINT32 GetFirst ();
    // get the first object in the array, or NoAccount if the array is empty

    SINT32 GetLast ();
    // get the last object or NoAccount if the array is empty

    SINT32 GetNext ();
    // get the next object in the array or NoAccount if the end of the
    // array has been reached

    EBool Find (SINT32 index);
    // returns True is the index is stored, False otherwise


    //----------------------------------------------------------------------
    //--- add/remove objects
    //----------------------------------------------------------------------

    EBool Insert (SINT32 index);
    // insert a new element into the (sorted) array; if the index exists,
    // False is returned

    EBool Remove (SINT32 index);
    // remove the index; if the index does not exist, False is returned


protected:

    //----------------------------------------------------------------------
    //--- internal memory management
    //----------------------------------------------------------------------

    void UpdateArray();
    // allocates a new array of size size_ and copies the data of firstFree_
    // elements from array_ to the new array; array_ is then deleted and set
    // to the new array
    // assert: firstFree_ <= size_

    void AllocateMemory();
    // allocates a new array that is allocSize_ elements bigger than the
    // current array size stored in size_; the data is copied to the new
    // array and array_ is set to it; the old array is discarded

    void FreeMemory();
    // frees the last allocSize_ elements of the current array;
    // assert: size_ > allocSize_
    // assert: firstFree_ <= size_-allocSize_


    //----------------------------------------------------------------------
    //--- internal array information
    //----------------------------------------------------------------------

    INT32 Size() const;
    // returns the current size of the array

    INT32 Occupied() const;
    // returns the number of array elements that are currently in use

    INT32 AllocationSize() const;
    // returns the additional size used for expansion (contraction) of
    // the array

    EBool CurrentIsValid () const;
    // returns True if current stands on a valid element

    SINT32 GetCurrentID() const;
    // returns the ID of the current element

    SINT32 GetCurrent() const;
    // returns the value of the current element


    //----------------------------------------------------------------------
    //--- internal array manipulation
    //----------------------------------------------------------------------

    void InsertAtCurrent (SINT32 index);
    // insert the data at the current_ index; shifts the rest of the
    // array one element to the right; there has to be at least one free
    // element in the array
    // if the array is full (after insertion), new memory is allocated
    
    void RemoveAtCurrent ();
    // removes the data at the current_ index; shifts the rest of the
    // array one element to the left
    // if more than allocSize_+1 elements are unused, memory is freed


private:

    SINT32 *array_;  // pointer to the beginning of the sorted array
    
    INT32 allocSize_;  // number of new elements if array is expanded
    INT32 size_;  // size of the array

    INT32 firstFree_;  // index of the first free entry
    INT32 current_;   // index of the currently accessed entry
};


#endif

