//
// idtemp.hpp
//
// stores a nodeID with the pair <networkID, nodeAddress> as key
//
// created: 03.01.97
//
// last change: 06.01.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _IDTEMP_HPP_
#define _IDTEMP_HPP_

#include "address.hpp"
#include "error.hpp"
#include "listbase.hpp"


//**************************************************************************
// class CNodeList: this class stores a pointer to a nodeID with the pair
//      <networkID, nodeAddress> as key
//**************************************************************************

class CNodeList : public ACObjectList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CNodeList (INT32 aSize = ArrayAllocationSize) :
        ACObjectList (aSize) {}
    // allocate memory

    virtual ~CNodeList () { DeleteAll(); }
    // the destructor deletes all objects in the array


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    SINT32* GetFirst ()
        { return (SINT32*) ACObjectList::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    SINT32* GetLast ()
        { return (SINT32*) ACObjectList::GetLast(); }
    // get the last object or NULL if the array is empty

    SINT32* GetNext ()
        { return (SINT32*) ACObjectList::GetNext(); }
    // get the next object in the array or the last object if the end of
    // the array has been reached

    SINT32* GetPrev ()
        { return (SINT32*) ACObjectList::GetPrev(); }
    // return the previous object in the array or the first if the beginning
    // of the array has been reached

    SINT32* Find (SINT32 networkID, const CAddr& netAddress)
    {
        SKey key (networkID, netAddress);
        return (SINT32*) ACObjectList::Find (&key);
    }
    // returns the object with this key; if the key does not exist,
    // NULL is returned


    //----------------------------------------------------------------------
    //--- get information about the current element
    //----------------------------------------------------------------------

    SINT32* GetCurrentObject ()
        { return (SINT32*) ACObjectList::GetCurrentObject();}
    // returns a pointer to the member that has been accessed last, or
    // NULL, if this information is not available anymore


    EBool GetCurrentObjectKey (SINT32& networkID, CAddr& netAddress)
    // references to the key variables
    {
        SKey *key = (SKey*) ACObjectList::GetCurrentObjectKey();

        if (key != NULL)
        {
            networkID = key->networkID_;
            netAddress = key->netAddress_;
        //for (int i=0; i<ADDRCNT; i++) (*netAddress)[i] = key->netAddress_[i];
            // set the argument variables
            return True;
        }
        else return False;
    }
    // returns True if a member is accessed (in that case networkID and
    // netAddress will contain the key), or False if no element is
    // currently accessed


    //----------------------------------------------------------------------
    //--- add/remove objects
    //----------------------------------------------------------------------

    EBool Insert (SINT32* object, SINT32 networkID, const CAddr& netAddress)
    {
        SKey *key = new SKey (networkID, netAddress);
        EBool inserted;

        inserted = ACObjectList::Insert (object, key);

        if (!inserted)  // the object was not inserted
        {
            delete key;  // the key is not needed
        }

        return inserted;
    }

    SINT32* Remove (SINT32 networkID, const CAddr& netAddress)
    {
        SKey key (networkID, netAddress);
        return (SINT32*) ACObjectList::Remove (&key);
    }
    // remove an object from the array; if the key does not exist, NULL
    // is returned


    //----------------------------------------------------------------------
    //--- delete one or all member(s) of the array
    //----------------------------------------------------------------------

    // delete an object; if the index exists, True is returned; otherwise
    // False is returned
    EBool Delete (SINT32 networkID, const CAddr& netAddress)
    {
        SINT32 *object;
        EBool found = False;

        object = Remove (networkID, netAddress);

        if (object != NULL)  // index does exist
        {
            delete object;  // delete the object
            found = True;
        }

        if (!CurrentIsValid()) object = GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // delete all members of the array
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        SINT32 *object;
        
        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = GetLast();  // get the last item in the array
                RemoveAtCurrent();  // remove the array entry
                delete object;  // delete the object
            }

            return True;
        }
    } // DeleteAll


protected:

    // this function is used to compare keys
    virtual int Compare (void* key1, void *key2) const
    {
        Assert (key1 != NULL);
        Assert (key2 != NULL);

        return ((SKey*)key1)->CompareTo ((SKey*)key2);
    }

    virtual void DeleteKey (void* key) { delete (SKey*) key; }
    // delete the key


private:

    struct SKey
    {
        SINT32 networkID_;
        CAddr netAddress_;

        SKey (SINT32 networkID, const CAddr& netAddress)
        {
            networkID_ = networkID;
            netAddress_ = netAddress;
        }

        int CompareTo (const SKey* key1)
        {
            if (networkID_ < key1->networkID_) return -1;
            if (networkID_ > key1->networkID_) return +1;

            if (netAddress_ < key1->netAddress_) return -1;
            if (netAddress_ > key1->netAddress_) return +1;

            return 0;
        }
    }; // struct SKey
};


#endif


