//
// interupt.hpp
//
// for interrupt handling
// use GetClockIntvec() to obtain an NTI base vector of the UTCSU for the
// given node and area
// use GetGpsIntvec() to obtain an NTI base vector of the GPS-receiver for
// the given node and area
// use SetISR() to set an interrupt vector to an ISR
// use RemoveISR() to remove an ISR from an interrupt vector
// use InvokeISR() to call the ISR for an interrupt vector
//
// created: 18.01.97
//
// last change: 24.06.98
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/06/24 10:29:43 $
//


#ifndef _INTERUPT_HPP_
#define _INTERUPT_HPP_


#include "psos.h"

#include "clockcfg.hpp"


//**************************************************************************
// CNeedInterrupt: this is the base class needed for all classes that have
//    an interrupt service routine
//**************************************************************************

class CNeedInterrupt
{
public:
    friend void CallISR (CNeedInterrupt *obj);
    // to invoke the default isr of an object; only to be used by the
    // interrupt handler!

private:
    virtual void isr() =0;
    // implementation of the default isr is a must for all derived classes
};


typedef void (*CallFunc)(CNeedInterrupt*);
// so "CallFunc" is a function that takes a CNeedInterrupt* as argument;
// the idea is to have friend functions for calling the isr's:
// if you have only one ISR, simply overload isr() and call the SetISR()
// functions below with the object pointer; in that case the default ISR
// isr() is used when the interrupt is generated
// if you have *more* than one ISR, you still can use the default for
// calling isr(), but for any other isr's you have to define new friend
// functions and specify them together with the object, and you have to
// cast from base to derived in the friend function
// note: instead of friends, normal functions could be used, but in that
// case the ISR's would have to be public; I thought that private ISR's
// are more secure, so I've used friends (also, you can use the friend as
// an ISR; might come in handy in some situations)


//**************************************************************************
// use only these functions for interrupt handling
//**************************************************************************

extern int GetClockIntvec (SINT32 nodeID, int area);
// returns the clock interrupt vector for the area on the given node
// assert: area < ASICCNT, area < 16
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp

extern int GetGpsIntvec (SINT32 nodeID, int area, EGpsUnit gpu);
// returns the GPS-receiver interrupt vector for the area and GPU on the
// given node
// assert: area < ASICCNT, area < 16, gpu < EGpsUnitCount
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp

extern INT8 GetClockIntLevel();
// returns the interrupt level of the UTCSU and LANCE interrupts; the levels
// should be equal (if they are not, correct behaviour of the simulation
// cannot be guaranteed)
// note: hardware/software implementations differ, the implementation can
// be found in impl_?w.cpp


extern void SetISR (int vector, CNeedInterrupt *obj, CallFunc caller=CallISR);
// stores the object with this vector in the interrupt vector table; if the
// vector is in use, the object is simply added to a list of ISRs that
// are called by InvokeISR()
// the function caller() will be invoked by the interrupt
// solely use this function to set the object (isr) for a vector!
// note: make sure that the interrupt in question is disabled when you call
// this function, even if there is already an objectISR stored


extern void RemoveISR(int vector,CNeedInterrupt *obj,CallFunc caller=CallISR);
// removes this particular object from the interrupt vector table
// solely use this function to remove the object (isr) for a vector!
// note: make sure that the interrupt in question is disabled when you call
// this function


extern void InvokeISR (int vector);
// calls the ISR of the object(s) associated with this vector
// assert: an entry must exist (i.e., at least one ISR must have been
// installed)


#endif

