//
// irqtemp.hpp
//
// stores clock interrupt processes with an integer as the key (which
// should be the bitnumber in the interrupt registers (0..63))
//
// created: 19.02.97
//
// last change: 19.02.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _IRQTEMP_HPP_
#define _IRQTEMP_HPP_

#include "listbase.hpp"
#include "clockirq.hpp"


//**************************************************************************
// class CInterruptList: stores clock interrupt processes; uses the
//      bitnumber in the interrupt registers (0..63) as the key
//**************************************************************************

class CInterruptList : public ACObjectList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CInterruptList (INT32 aSize = ArrayAllocationSize) :
        ACObjectList (aSize) {}
    // allocate memory

    virtual ~CInterruptList () { DeleteAll(); }
    // the destructor deletes all objects in the array


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    CClockInterrupt* GetFirst () { return (CClockInterrupt*) ACObjectList::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    CClockInterrupt* GetLast () { return (CClockInterrupt*) ACObjectList::GetLast(); }
    // get the last object or NULL if the array is empty

    CClockInterrupt* GetNext () { return (CClockInterrupt*) ACObjectList::GetNext(); }
    // get the next object in the array or the last object if the end of
    // the array has been reached

    CClockInterrupt* GetPrev () { return (CClockInterrupt*) ACObjectList::GetPrev(); }
    // return the previous object in the array or the first if the beginning
    // of the array has been reached

    CClockInterrupt* Find (int key1) // put the keys here
    {
        return (CClockInterrupt*) ACObjectList::Find (&key1);
    }
    // returns the object with this key; if the key does not exist,
    // NULL is returned


    //----------------------------------------------------------------------
    //--- get information about the current element
    //----------------------------------------------------------------------

    CClockInterrupt* GetCurrentObject ()
        { return (CClockInterrupt*) ACObjectList::GetCurrentObject(); }
    // returns a pointer to the member that is currently being accessed or
    // NULL, if no element is being accessed

    EBool GetCurrentObjectKey (int& key1)
    // references to the key variables
    {
        int *key = (int*) ACObjectList::GetCurrentObjectKey();

        if (key != NULL)
        {
            key1 = *key;
            // set the argument variables
            return True;
        }
        else return False;
    }
    // returns True if a member is accessed (in that case the key1 will
    // contain the key), or False if no element is currently accessed


    //----------------------------------------------------------------------
    //--- add/remove objects
    //----------------------------------------------------------------------

    EBool Insert (CClockInterrupt* object, int key1)
    {
        int *key = new int (key1);

        if (!ACObjectList::Insert (object, key))  // object not inserted
        {
            delete key;  // the key is not needed
            return False;
        }
        else return True;
    }

    CClockInterrupt* Remove (int key1)
    {
        return (CClockInterrupt*) ACObjectList::Remove (&key1);
    }
    // remove an object from the array; if the key does not exist, NULL
    // is returned



    //----------------------------------------------------------------------
    //--- delete one or all member(s) of the array
    //----------------------------------------------------------------------

    // note: since the object is a process, it may be unref'd instead of
    // deleted; so a special function is needed to delete processes; it
    // is called by Delete() and DeleteAll()
    void DeleteProcess (ACProcess* object)
    {
        object->Cancel();
        object->terminate();  // terminate the process

#ifndef NO_RESOURCE
        Resource::unref(object);  // either unref the process in C++SIM
#else
        delete object;  // or delete it
#endif
    }


    // delete an object; if the index exists, the object with this index
    // is deleted and True is returned; otherwise False is returned
    // current points to the position where the element was deleted (or
    // should have be inserted), or to the last element if the deleted
    // object was (would have been) the last
    EBool Delete (int key1)
    {
        CClockInterrupt* object;
        EBool found = False;

        object = Remove (key1);  // remove the object

        if (object != NULL)  // index does exist
        {
            DeleteProcess (object);  // delete the process
            found = True;
        }

        if (!CurrentIsValid()) object = GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // delete all members of the array
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        CClockInterrupt* object;
        
        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = GetLast();  // get the last item in the array
                DeleteProcess (object);  // delete the process
                RemoveAtCurrent();  // remove the array entry
            }

            return True;
        }
    } // DeleteAll


protected:

    // this function is used to compare keys
    virtual int Compare (void* key1, void *key2) const
    {
        if (*(int*)key1 < *(int*)key2) return -1;
        if (*(int*)key1 > *(int*)key2) return +1;
        return 0;
    }


    virtual void DeleteKey (void* key) { delete (int*) key; }
    // delete the key
};

#endif

