//
// isrtemp.hpp
//
// provides a vector table that can store more than one object ISR per
// interrupt vector
//
// created: 28.08.97
//
// last change: 07.12.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _ISRTEMP_HPP_
#define _ISRTEMP_HPP_

#include "duplist.hpp"
#include "interupt.hpp"

struct SInterrupt
{
    CNeedInterrupt *object;  // object whose ISR should be called
    CallFunc caller;  // call function

    SInterrupt (CNeedInterrupt *obj, CallFunc call)
    {
        object = obj;
        caller = call;
    }
};


//**************************************************************************
// class CVectorTable : stores one or more object ISRs with the interrupt
//    vector as the key
//**************************************************************************

class CVectorTable : public ACSortedArray
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CVectorTable (INT32 aSize=ArrayAllocationSize)
        : ACSortedArray (True, aSize) {}
    // allocate memory

    virtual ~CVectorTable () { DeleteAll(); }
    // the destructor deletes all objects in the array; since the base
    // destructor is called although it is pure virtual, destruction of
    // the _array itself is done there


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    SInterrupt* GetFirst () { return (SInterrupt*) ACSortedArray::GetFirst(); }
    // get the first object in the array, or NULL if the array is empty

    SInterrupt* GetLast () { return (SInterrupt*) ACSortedArray::GetLast(); }
    // get the last object or NULL if the array is empty

    SInterrupt* GetNext () { return (SInterrupt*) ACSortedArray::GetNext(); }
    // get the next object in the array or the last object if the end of
    // the array has been reached

    SInterrupt* GetPrev () { return (SInterrupt*) ACSortedArray::GetPrev(); }
    // return the previous object in the array or the first if the beginning
    // of the array has been reached

    SInterrupt* Find (double index) { return (SInterrupt*) ACSortedArray::Find (index); }
    // returns the object with this index; if no such object exists,
    // NULL is returned and _current is the index where the object should
    // be inserted

    SInterrupt* FindNext () { return (SInterrupt*) ACSortedArray::FindNext (); }
    // returns the next object with the same index as the _current index;
    // if no more objects with this index exist, NULL is returned and
    // the function has no effect

    //----------------------------------------------------------------------
    //--- get information about the current element
    //----------------------------------------------------------------------

    SInterrupt* GetCurrentObject () { return (SInterrupt*) ACSortedArray::GetCurrentObject();}
    // returns a pointer to the member that is currently being accessed or
    // NULL, if no element is being accessed


    //----------------------------------------------------------------------
    //--- remove objects
    //----------------------------------------------------------------------

    SInterrupt* Remove (double index) { return (SInterrupt*) ACSortedArray::Remove (index); }
    // remove an object from the array

    void RemoveAtCurrent () { ACSortedArray::RemoveAtCurrent (); }
    // remove the object that is currently active
    // if there is no currently active object, an assert fails!


    //----------------------------------------------------------------------
    //--- delete one or all member(s) of the array
    //----------------------------------------------------------------------

    // delete an object; if the index exists, all objects with this index
    // are deleted and True is returned; otherwise False is returned
    // current points to the position where the element was deleted (or
    // should have be inserted), or to the last element if the deleted
    // object was (would have been) the last
    EBool Delete (double index)
    {
        SInterrupt* object;
        EBool found = False;

        object = Remove (index);

        if (object != NULL)  // index does exist
        {
            do {
                delete object;  // delete all objects with this index
                object = Remove (index);
            } while (object != NULL);
                
            found = True;
        }

        if (!CurrentIsValid()) object = GetLast();
        // set current to the last entry if it is not valid

        return found;
    } // Delete
    
    // delete all members of the array
    // if the array is empty, False is returned; else all objects are
    // deleted and True is returned
    EBool DeleteAll ()
    {
        SInterrupt* object;
        
        if (Empty()) return False;
        else
        {
            while (!Empty())
            {
                object = GetLast();  // get the last item in the array
                RemoveAtCurrent();  // remove the array entry
                delete object;  // delete the object
            }

            return True;
        }
    } // DeleteAll
    // note: _array now points to an empty array of size _allocSize
};


#endif

