//
// lance.hpp
//
// declaration of the Lance interface
//
// created: 21.10.96
//
// last change: 22.01.99
//
// RCS version control information:
//     $Revision: 1.7 $$State: Exp $$Date: 1999/01/26 14:14:06 $
//


#ifndef _LANCE_HPP_
#define _LANCE_HPP_

#include "ci.h"

#include "bool.hpp"
#include "crashing.hpp"
#include "ctrled.hpp"
#include "interupt.hpp"


class CClockSync;
class CDelayList;
class CRandom;

struct SDelayMcb;
struct STimeData;


//**************************************************************************
// class CLance: this class is the interface between the simulation and
//    the CI functions; it can scramble and/or delay messages
//**************************************************************************

class CLance : public CApplControlledObject, public CCrashingModule,
               public CNeedInterrupt
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CLance (CControllingObject *ctrl, int area);
    ~CLance();


    //----------------------------------------------------------------------
    //--- reset
    //----------------------------------------------------------------------

    virtual void Reset();
    // set members and the base classes


    //----------------------------------------------------------------------
    //--- control the system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the requested parameter value
    // existing params: PCrash, PRecover, DmaxSend, DistributionSend,
    // DmaxRecv, DistributionRecv, and all params of CApplControlledObject

    virtual EBool GetParameter (SParameter& param) const;
    // returns the requested parameter value
    // additional parameter: CrashStatus


    //----------------------------------------------------------------------
    //--- execution of system commands
    //----------------------------------------------------------------------

    virtual EBool ExecuteCommand (const SCommand& command);
    // executes the command; returns True if successfull, otherwise False;
    // existing commands: all of CApplControlledObject and CCrashingModule


    //----------------------------------------------------------------------
    //--- suspend the Lance
    //----------------------------------------------------------------------

    void Suspend();
    // go into suspension mode; in that mode, all messages that are
    // currently hold back will be discarded, and message announce from the
    // network will be ignored; the Lance will stay in suspension mode until
    // ClockSync initiates a send request


    //----------------------------------------------------------------------
    //--- crash the Lance
    //----------------------------------------------------------------------

    virtual void Crash();
    // crash the module; that includes throwing away all messages that are
    // currently in the list of delays


    //----------------------------------------------------------------------
    //--- set associated objects
    //----------------------------------------------------------------------

    void SetClockSync (CClockSync *cs);
    // set the clockSync pointer to cs

    void SetClock (CClockInterface *clock);
    // set the clock


    //----------------------------------------------------------------------
    //--- request/free message buffer and envelope
    //----------------------------------------------------------------------

    void GetBuffer (INT16 msgsize, char **message, INT32& msgID);
    // returns a pointer to a data buffer; allocates four bytes more (to
    // contain the message ID), but returns a pointer to the address of byte
    // four in this buffer
    // note: this is *no* CI buffer, the buffer resides in local memory
    // (the data is copied into a CI buffer upon send)
    // assert: message != NULL

    void GetEnvelope (tenvx **tenv);
    // returns a pointer to a timestamped envelope
    // note: again, this is no CI envelope, but resides in local memory
    // (the data is copied into a CI envelope upon send)
    // assert: tenv != NULL
    // context: ISR

    void ReturnBuffer (char *message);
    // frees the pointer to a data buffer; assumes that the buffer was
    // allocated using GetBuffer() and that it starts at an address four
    // bytes before 'message'
    // buffers delivered to ClockSync are of that type
    // assert: message != NULL
    // context: ISR

    void ReturnEnvelope (tenvx *tenv);
    // frees the pointer to a timestamped envelope; assumes that the
    // envelope was allocated using GetEnvelope()
    // envelopes delivered to ClockSync are of that type
    // assert: tenv != NULL
    // context: ISR

    void GetReceiveBuffer (INT16 size, bufx **rbuf);
    // returns a pointer to a data buffer; this function just allocates the
    // amount of memory requested; you should call this function to obtain
    // a local receive buffer - Announce() assumes that its 'buf' parameter
    // was obtained via this function
    // assert: rbuf != NULL

    void GetReceiveEnvelope (renvx **renv);
    // just allocates the structure
    // assert: renv != NULL

    void RetReceiveBuffer (bufx *rbuf);
    // use this function to return buffers allocated with GetReceiveBuffer()
    // assert: rbuf != NULL

    void RetReceiveEnvelope (renvx *renv);
    // use this function to return mem allocated with GetReceiveEnvelope()
    // assert: renv != NULL


    //----------------------------------------------------------------------
    //--- send/receive messages
    //----------------------------------------------------------------------

    void Send (int mode, tenvx *tenv, char *message, INT16 msgsize,
               int if_num, int area);
    // delays the message, which is then sent via SendNow()
    // suspended is set to False, the crash status may change; if the
    // module is crashed, the message (and all currently in the delay list)
    // is discarded
    // the function assumes that 'tenv' was obtained via the GetEnvelope()
    // function, and 'message' via the GetBuffer() function

    int SendNow (int mode, tenvx *tenv, char *message, INT16 msgsize,
                 int if_num, int area);
    // delivers the message to the network, then deletes tenv and message;
    // tenv may be NULL
    // the function allocates CI buffers, copies the parameters into them
    // and calls ci_send()
    // the function returns a CI error code if one of the CI functions
    // fails, or 0 if the message was sent successfully
    // the Lance module may crash and msges (including the destination
    // address) may be scrambled in this function
    // note: *regardless* of the 'mode', tenv and message are always returned
    // (so ClockSync should never use a tenv and message pointer after it
    // has called Send())
    // assert: message != NULL
    // context: ISR

    void Announce (renvx *renv, bufx *buf, INT16 size, int if_num);
    // receives a message from the network and delays it before it is
    // delivered to ClockSync;
    // the function assumes that the 'renv' and 'buf' data reside in
    // local memory and must be disposed via the RetReceive...() functions

    void AnnounceNow (renvx *renv, bufx *buf, INT16 size, int if_num);
    // delivers the message to ClockSync, then deletes the pointers using
    // the RetReceive...() functions
    // context: ISR


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- initialization of the data members
    //----------------------------------------------------------------------

    virtual void Init ();
    // Lance is suspended

    virtual void Dispose ();


    //----------------------------------------------------------------------
    //--- delay messages
    //----------------------------------------------------------------------

    void DelayMessage (int mode, tenvx *tenv, char *message, INT16 msgsize,
                       int if_num, int area);
    // delays the outgoing message according to the delay data
    // assert: clock != NULL

    void DelayMessage (renvx *renv, bufx *buf, INT16 size, int if_num);
    // delays the incoming message according to the delay data
    // assert: clock != NULL


    //----------------------------------------------------------------------
    //--- scramble messages
    //----------------------------------------------------------------------
    
    void ScrambleMessage (addr address,char *msg,INT16 size,INT32 msgID);
    // scramble outgoing messages, notify evaluation
    // context: ISR

    void ScrambleMessage (addr address, STimeData *time, char *msg,
                          INT16 size, INT32 nodeID, int area, INT32 msgID);
    // scramble incoming messages, notify evaluation
    // context: ISR

    void Scramble (void *data, INT16 size);
    // just scramble the data, no notification


    //----------------------------------------------------------------------
    //--- fault injection
    //----------------------------------------------------------------------

    virtual void ReportCrashChange();
    // creates an event of type SApplCrashEvent, fills in all parameters
    // (type=CrashStatusChange) and sends it to the evaluation


    //----------------------------------------------------------------------
    //--- event reports
    //----------------------------------------------------------------------

    void ReportMessageEvent (EEvent type, INT32 msgID);
    // create a message event with given type, fill in the data and send it
    // to the evaluation; use this function for incoming messages
    // module, type, sender, senderArea, msgID are filled in (not 'receiver')
    // context: ISR

    void ReportMessageEvent (EEvent type,SINT32 nodeID,int area,INT32 msgID);
    // create a message event with given type, fill in the data and send it
    // to the evaluation; use this function for outgoing messages
    // module, type, msgID are filled in (not 'receiver')
    // context: ISR


    //----------------------------------------------------------------------
    //--- error reports
    //----------------------------------------------------------------------

    int ReportError (EReportType type,char *file,int line,char *expression);
    // call global report function with module Lance, nodeID and area


private:

    EBool suspended_;  // True if the Lance is currently suspended

    CClockSync *clockSync_;   // associated clock synchronization module
    CClockInterface *clock_;  // associated clock module

    int area_;  // the DMA area of the clock

    INT32 msgID_;  // identifier of the next message that is sent

    SRandomData distDataSend_;  // random data for delaying messages
    SRandomData distDataRecv_;

    double pSendScramble_;  // prob. for scrambling outgoing msgs
    double pRecvScramble_;  // prob. for scrambling incoming msgs

    double pSendDuplicate_;  // prob. for duplicating outgoing msgs
    double pRecvDuplicate_;  // prob. for duplicating incoming msgs

    CRandom *random_;
    // the random class used to generate random delays

    SDelayMcb *msg_;  // the delayed message that is processed next
    CDelayList *listOfDelayedMessages_;  // list of all other delayed msges


#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- ISR
    //----------------------------------------------------------------------

    virtual void isr();
    // send/deliver all messages due at the current time (called by DT B5)
    // context: ISR


    //----------------------------------------------------------------------
    //--- management of the message delay list
    //----------------------------------------------------------------------

    void Insert (SDelayMcb *mcb);
    // inserts the 'mcb' into the delay list, possibly sets DT B5
    // assert: mcb != NULL, clock_ != NULL

    void DeleteDelayedMessages ();
    // deletes all entries in the list, deletes member 'msg' (sets it to NULL)
    // if 'msg' is NULL, the function returns at once
};


#endif


