//
// listbase.hpp
//
// declaration of a abstract base list class of object pointers using
// arrays; optimized for repeated access and sequential access; uses binary
// search for other access; no duplicates; the sort key must be supplied
// in the derived class
// note: reimplement ACObjectList when porting to hardware, since the
//       software implementation cannot handle concurrent access
//
// created: 27.10.96
//
// last change: 11.02.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:37:28 $
//


#ifndef _LISTBASE_HPP_
#define _LISTBASE_HPP_

#include <stdlib.h>  // for the NULL macro

#include "bool.hpp"
#include "constant.hpp"


//**************************************************************************
// class ACObjectList: this abstract class manages a sorted array
//     it is assumed that most of the time the access is sequential; if
//     not, binary search is used
//     note: no duplicates are allowed and the type of the sorting
//     criterion has to be supplied in the derived class
//     the destructor and the Compare() and DeleteKey() functions are pure
//     virtual
//**************************************************************************

class ACObjectList
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    ACObjectList (INT32 aSize = ArrayAllocationSize);
    // allocate memory
    
    virtual ~ACObjectList () =0;
    // pure virtual, but nevertheless it deletes the _array


    //----------------------------------------------------------------------
    //--- get the array status
    //----------------------------------------------------------------------

    EBool Empty() const;
    // returns True if the array is empty, False otherwise
    
    EBool BeginOfList () const;
    // returns True if _current is at the beginning of
    // the array, False otherwise (or if the array is empty)

    EBool EndOfList () const;
    // returns True if _current is at the end of the
    // array, False otherwise (or if the array is empty)


    //----------------------------------------------------------------------
    //--- move in the array
    //----------------------------------------------------------------------

    void* GetFirst ();
    // get the first object in the array, or NULL if the array is empty

    void* GetLast ();
    // get the last object or NULL if the array is empty

    void* GetNext ();
    // get the next object in the array or NULL if the end of the array
    // has been reached

    void* GetPrev ();
    // return the previous object in the array or NULL if the beginning
    // of the array has been reached

    void* Find (void *key);
    // returns the object with this key or NULL, if the key was not found


    //----------------------------------------------------------------------
    //--- get information about the current element
    //----------------------------------------------------------------------

    void* GetCurrentObject () const;
    // returns a pointer to the object that is currently being accessed or
    // NULL, if no element is being accessed

    void* GetCurrentObjectKey () const;
    // returns the object key of the currently accessed element or NULL,
    // if no element is being accessed


    //----------------------------------------------------------------------
    //--- add/remove objects
    //----------------------------------------------------------------------

    EBool Insert (void *object, void *key);
    // insert a new element into the (sorted) array;
    // if an object with this index exists, False is returned

    void* Remove (void *key);
    // remove an object; if the index does not exist, NULL is returned


protected:

    //----------------------------------------------------------------------
    //--- sorting
    //----------------------------------------------------------------------

    virtual int Compare (void *key1, void* key2) const =0;
    // returns -1 if key1 < key2, 0 if key1 = key2, and +1 if key1 > key2;
    // this function has to be supplied by the derived class!


    //----------------------------------------------------------------------
    //--- deletion of the key pointer
    //----------------------------------------------------------------------

    virtual void DeleteKey (void *key) =0;
    // should free the memory of the key structure; this function is called
    // whenever an object is removed


    //----------------------------------------------------------------------
    //--- internal information concerning the current element
    //----------------------------------------------------------------------

    EBool CurrentIsValid () const;
    // returns True if _current is an occupied slot, False otherwise

    void* GetCurrentKey () const;
    // returns a pointer to the key of the current element; the current
    // element must exist!

    void* GetCurrent () const;
    // returns the object of the current element; the current element must
    // exist!

    int CompareToCurrent (void *key) const;
    // returns -1 if key < currentKey, = 0 if the keys are equal, and +1
    // if key > currentKey
    // the current element must exist!


    //----------------------------------------------------------------------
    //--- internal memory management
    //----------------------------------------------------------------------

    void UpdateArray();
    // allocates a new array of size _size and copies the data of _firstFree
    // elements from _array to the new array; _array is then deleted and set
    // to the new array
    // assert: _firstFree <= _size

    void AllocateMemory();
    // allocates a new array that is _allocSize elements bigger than the
    // current array size stored in _size; the data is copied to the new
    // array and _array is set to it; the old array is discarded

    void FreeMemory();
    // frees the last _allocSize elements of the current array; note that
    // the function will assume that these elements are not in use!
    // if they are, an assert will fail!


    //----------------------------------------------------------------------
    //--- internal array information
    //----------------------------------------------------------------------

    INT32 Size() const;
    // returns the current size of the array

    INT32 Occupied() const;
    // returns the number of array elements that are currently in use

    INT32 AllocationSize() const;
    // returns the additional size used for expansion (contraction) of
    // the array


    //----------------------------------------------------------------------
    //--- internal array manipulation
    //----------------------------------------------------------------------

    void InsertAtCurrent (void *object, void *key);
    // insert the data at the _current index; shifts the rest of the
    // array one element to the right; there has to be at least one free
    // element in the array
    // if the array is full (after insertion), new memory is allocated
    
    void RemoveAtCurrent ();
    // removes the data at the _current index; shifts the rest of the
    // array one element to the left
    // if more than _allocSize+1 elements are unused, memory is freed


private:

    //**********************************************************************
    // class CObjectElement: this class represents an element of the sorted
    //     array; it contains the data that is stored in the array
    //**********************************************************************
    
    class CObjectElement
    {
    public:

        //----------------------------------------------------------------------
        //--- constructor
        //----------------------------------------------------------------------

        CObjectElement ();  // constructor to initialize the pointer


        //----------------------------------------------------------------------
        //--- set the members of the class
        //----------------------------------------------------------------------

        void SetObject (void *object, void *key);


        //----------------------------------------------------------------------
        //--- get the members of the class
        //----------------------------------------------------------------------

        void* GetObject () const;
        void* GetKey () const;


    private:

        void* object_;  // pointer to the object that is stored
        void* key_;  // pointer to the key (sorting criterion)
    }; // CObjectElement

    CObjectElement *_array;  // pointer to the beginning of the sorted array
    
    INT32 _allocSize;  // number of new elements if array is expanded
    INT32 _size;  // size of the array

    INT32 _firstFree;  // index of the first free entry
    INT32 _current;   // index of the currently accessed entry
};


#endif

