//
// netacc.hpp
//
// declaration of the NetworkAccountant-class responsible for the management
// of the transmission delays
//
// created: 29.07.96
//
// last change: 28.09.98
//
// RCS version control information:
//     $Revision: 1.12 $$State: Exp $$Date: 1998/09/28 10:57:35 $
//


#ifndef _NETACC_HPP_
#define _NETACC_HPP_


#include "account.hpp"
#include "constant.hpp"
#include "general.hpp"


class CRandom;

struct SNetDistributionP2P;
struct SNetworkData;
struct SParameter;


//**************************************************************************
// class CNetworkAccountant: this class is a database for the network and
//      stores transmission delays; it can compute an arbitrary delay
//      between two nodes using the data stored
//**************************************************************************


class CNetworkAccountant : public CAccountant
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    CNetworkAccountant (SINT32 networkID);
    virtual ~CNetworkAccountant ();


    //----------------------------------------------------------------------
    //--- configuration of system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the given parameter to its new value; if the parameter does
    // not exist or if the value is not valid, False is returned
    // existing parameters: DistributionEarly, D..Timely, D..Late,
    // PEarly, PLate

    virtual EBool GetParameter (SParameter& param);// const;
    // returns the value of the given parameter in the appropriate field of
    // SParameter; if the parameter does not exist, False is returned


    //----------------------------------------------------------------------
    //--- management of transmission delays
    //----------------------------------------------------------------------

    double GetArbitraryDelay (SINT32 senderID, SINT32 receiverID,
                              EArrival& arrivalType);
    // returns a random transmission delay in the specified interval for
    // the given pair of nodes; if no interval is specified (EArrivalCount),
    // the interval is chosen according to the probabilities of early/late
    // messages
    // if no delays are stored for this pair of nodes and sender != receiver,
    // the method returns NoAccount; if sender == receiver, system constant
    // DefaultOwnDelay is returned


    //----------------------------------------------------------------------
    //--- management of nodes
    //----------------------------------------------------------------------

    EBool DeleteNode (SINT32 nodeID);
    // if the node exists, all data concerning the node (whether as sender
    // or as receiver) will be destroyed and the method will return True;
    // if the node doesn't exist, False is returned

    EBool IsConnected (SINT32 senderID, SINT32 receiverID);
    // returns True if an entry for the pair of nodes exists, False otherwise


protected:

#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- initialization and destruction of the data members
    //----------------------------------------------------------------------

    virtual void Init ();
    // initializes all class members to default values; a second call
    // (without Dispose() in between) has no effect

    virtual void Dispose ();
    // deletes all dynamic members; a second call (without Init() in
    // between) has no effect


    //----------------------------------------------------------------------
    //--- internal management of probabilities
    //----------------------------------------------------------------------

    EBool SetProbability (SNetworkData& param);
    // same as SetParameter(), but only for probabilities
    // set the stored arrival probability to the given value "prob"
    // if 0<=prob and probEarly+probLate<=1, the data is stored and True is
    // returned; else False is returned and the method has no effect

    double GetProbability (SNetworkData& param);
    // get the specified probability or NoAccount, if no data is stored


    //----------------------------------------------------------------------
    //--- internal management of data parameters (set/get function)
    //----------------------------------------------------------------------

    EBool SetDataParameter (const SNetDistributionP2P *param);
    // same as SetParameter(), but only for data parameters

    EBool GetDataParameter (SNetDistributionP2P *param);
    // same as GetParameter(), but only for data parameters


private:

    SINT32 networkID_;
    // for initializing the random generator

    CRandom *random_;
    // the random class used to generate random numbers

    EBool initCalled_;
    // set to True by Init(), and to False by Dispose()
};


#endif

