//
// netbase.hpp
//
// declaration of the network base class
//
// created: 02.07.96
//
// last change: 02.09.98
//
// RCS version control information:
//     $Revision: 1.8 $$State: Exp $$Date: 1998/09/02 16:04:33 $
//


#ifndef _NETBASE_HPP_
#define _NETBASE_HPP_

#include "ci.h"

#include "crashing.hpp"
#include "ctrled.hpp"
#include "distrib.hpp"
#include "error.hpp"
#include "netproc.hpp"


class CNetworkAccountant;
class CNodeIdList;

struct SMsg;


//**************************************************************************
// class CDelayer: this class is a process class and is needed to delay
//     messages (broadcast latency!) before they are sent
//**************************************************************************

class CDelayer : public ACNetworkProcess
{
public:

    //----------------------------------------------------------------------
    //--- message delivery
    //----------------------------------------------------------------------

    virtual void Deliver (SMsg *msg);
    // returns the message to the network again (calling Transmit()); the
    // network is identified by the if_num stored in SMsg
};



//**************************************************************************
// class CMessenger: a process class that is needed to deliver messages
//     to the nodes
//**************************************************************************

class CMessenger : public ACNetworkProcess
{
public:

    //----------------------------------------------------------------------
    //--- message delivery
    //----------------------------------------------------------------------

    virtual void Deliver (SMsg *msg);
    // returns the message to the network again (calling Deliver()); the
    // network is identified by the if_num stored in SMsg
};



//**************************************************************************
// class ACNetwork: this is an abstract class that provides methods for
//     sending data to receiving nodes;
//     a derived class should overload the SendMsg method to implement a
//     given network
//**************************************************************************

class ACNetwork : public CControlledObject, public CCrashingModule
{
    friend class CDelayer;  // needs to call Transmit()
    friend class CMessenger;  // needs to call Deliver()

public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    ACNetwork (CControllingObject *ctrl, SINT32 networkID);
    virtual ~ACNetwork();


    //----------------------------------------------------------------------
    //--- reset all members to their initial values
    //----------------------------------------------------------------------

    virtual void Reset ();
    // restores the initial state


    //----------------------------------------------------------------------
    //--- control the system parameters
    //----------------------------------------------------------------------

    virtual EBool SetParameter (const SParameter& param);
    // sets the requested parameter value
    // existing params: Lambda, DistributionLambda, Omega, DistributionOmega,
    // and all params of CControlledObject, CCrashingModule and of the
    // network accountant

    virtual EBool GetParameter (SParameter& param) const;
    // returns the requested parameter value
    // additional parameter: CrashStatus


    //----------------------------------------------------------------------
    //--- execution of system commands
    //----------------------------------------------------------------------

    virtual EBool ExecuteCommand (const SCommand& command);
    // executes the command; returns True if successfull, otherwise False
    // existing commands: (Dis-)ConnectNode and all commands of
    // CControlledObject and CCrashingModule


    //----------------------------------------------------------------------
    //--- control message transmission
    //----------------------------------------------------------------------

    virtual void SendMsg (tenvx *tenv, bufx *buf, INT16 size, int area);
    // send the message to the destination given in tenv->da (this function
    // just creates a SMsg and then calls SendMsg (SMsg *msg))
    // if the sending node (which is found by calling GetCurrentNodeID())
    // is not connected, the message is discarded
    // note: the function assumes that tenv->length_field = TS_PACKET if
    // the message is to be timestamped


protected:

    CNetworkAccountant *networkAccountant;
    // the network accountant; it manages the transmission delays

    CDelayer *delayer;
    // the delayer delays the message, then calls Transmit()

    CMessenger *messenger;
    // the messenger delays the message, then calls Deliver()

    CNodeIdList *listOfNodes;
    // the nodes that are connected to the network
    // note: in this class, the list is only used in SendMsg() to check
    // if the sender node is connected

    SRandomData distDataLambda;  // for computing latency
    SRandomData distDataOmega;  // for computing operation delay

    CRandom *random;  // for computing the random variables


#ifdef PUBLIC_IN_TEST
public:
#endif
    //----------------------------------------------------------------------
    //--- initialization and destruction of the data members
    //----------------------------------------------------------------------

    virtual void Init ();
    // initializes all class members to default values; a second call
    // (without Dispose() in between) has no effect

    virtual void Dispose();
    // deletes all dynamic members; a second call (without Init() in
    // between) has no effect


    //----------------------------------------------------------------------
    //--- query of the network ID
    //----------------------------------------------------------------------

    SINT32 GetNetworkID ();
    // returns the network ID of the network


    //----------------------------------------------------------------------
    //--- control network crashes
    //----------------------------------------------------------------------

    virtual void Crash ();
    // crash the net; if the net was already crashed, this method has no
    // effect


    //----------------------------------------------------------------------
    //--- control the connection of nodes
    //----------------------------------------------------------------------

    void Connect (SINT32 nodeID);
    // the node can send messages to this network

    void Disconnect (SINT32 nodeID);
    // the node cannot send messages to this network


    //----------------------------------------------------------------------
    //--- internal control of message transmission
    //----------------------------------------------------------------------

    virtual void SendMsg (SMsg *msg) =0;
    // send the message to a single recipient or to all nodes;
    // overload this function to get a working network

    virtual void Transmit (SMsg *msg);
    // transmit a single message (between a pair of nodes): compute the
    // transmission time and give the message to the messenger

    virtual void Deliver (SMsg *msg);
    // delivers the message to the Lances of the receiver node; the CI
    // buffers are copied into 'local' memory (of course, in pure simulation
    // all memory is local), and it is the task of each Lance module to
    // delete them
    // the local memory is obtained by CLance::GetReceiveBuffer() and
    // -Envelope() and should  be returned by the corresponding Ret-funtions
    // note: the function must be called by a process which has the nodeID
    // of the receiver node
    // the function may inject faults, and it checks the crash-status before
    // delivering the message


    //----------------------------------------------------------------------
    //--- fault injection
    //----------------------------------------------------------------------

    virtual void ReportCrashChange();
    // creates an event of type SNetworkCrashEvent, fills in all parameters
    // and sends it to the evaluation


    //----------------------------------------------------------------------
    //--- event reports
    //----------------------------------------------------------------------

    virtual void ReportEvent (EEvent eventType);
    // report a network event to the controlling object

    void ReportMessageEvent (EEvent type, const SMsg *msg);
    // create a message event of given type, fill in the message identifiers
    // from msg, send it to the evaluation


    //----------------------------------------------------------------------
    //--- error reports
    //----------------------------------------------------------------------

    virtual int ReportError (EReportType type, char *file, int line,
                             char *expression);


private:

    SINT32 networkID_;
    // ID of the network

    EBool initCalled_;  // set to True by Init(), and to False by Dispose()
};


#endif


