//
// params.hpp
//
// all structures, enums etc. that are needed to manage the system
// parameters
//
// created: 05.10.96
//
// last change: 22.01.99
//
// RCS version control information:
//     $Revision: 1.7 $$State: Exp $$Date: 1999/01/26 14:14:23 $
//


#ifndef _PARAMS_HPP_
#define _PARAMS_HPP_


//**************************************************************************
// system modules
//**************************************************************************

// these are all modules that can be accessed by the user;
// the enum is used for commands, parameters and event reports
enum EModule
{
    // controlling objects, IFC to EvalSys
    Controller=0, Supervisor,

    // controlled objects
    ClockSync, Lance, Clock,

    // pure simulation only
    Network
};


//**************************************************************************
// system parameters
//**************************************************************************

// these are all parameters in the system; if the evaluation system wants
// to change something in a module, it should specify the parameter type
// using one of these enums
enum ESystemParameter
{
    // Network
    DistributionEarly = 0,
    DistributionTimely,
    DistributionLate,  // distributions for early/timely/late messages
    DistributionLambda,  // distribution for broadcast latency
    DistributionOmega,  // distribution for broadcast operation delay
    PEarly,
    PLate,  // probability for early/late messages

    // Lance
    DistributionSend,  // distribution for delaying msges before send
    DistributionRecv,  // distribution for delaying msges before announce
    PScrambleSend,  // probability for scrambling outgoing messages
    PScrambleRecv,  // probability for scrambling incoming messages
    PDuplicateSend,  // probability for duplicating outgoing messages
    PDuplicateRecv,  // probability for duplicating incoming messages

    // Network, Clock, Lance
    PCrash,  // probability of a module crash
    PRecover,  // probability of a module crash
    CrashStatus,  // returns crash status - read only

    // ClockSync
    AssociatedNetwork,  // the network to which the CSA synchronizes
    ComputationDelayCompensation,  // Gamma (time from CV-alg. to resync.)
    DelayCompensation,  // Delta + Lambda + Omega (time from FME to alg.)
    DelayMeasurementCycles,  // =n --> every n-th FME does delay measurement
    Delta,
    EpsilonPlus,
    EpsilonMinus, // estimated msg transmission characteristics
    NumberOfFaultyNodes,  // maximum(!) number of faulty nodes allowed
    NumberOfNodes,  // number of nodes in the system
    RateSynchronizationCycles,  // rate synchronization (used like delay m.)
    RoundPeriod,  // the round period P (time between two FMEs)
    SynchronizationAlgorithm,  // the synchronization algorithm used
    ThetaPlus,
    ThetaMinus,  // computed bounds on the clock rate
    U,  // accumulated clock state corrections since last rate sync. round

    UserParameter,  // user defined parameter (synchronization algorithm)

    // Clock
    Accuracy,  // accuracy of the clock
    AccuracyTransmissionLoss,  // transmitted accuracy granularity
    AmortizationDuration,  // duration of the amortization phase
    ClockDrift,  // the clock drift (seconds/second) (pure simulation only!)
    ClockGranularity,  // transmitted clock granularity
    ClockValue,  // clock value (macrostamp, timestamp)
    DistributionFast,
    DistributionRegular,
    DistributionSlow,  // distributions for the clock drift
    RhoPlus,
    RhoMinus,  // bounds on the clock rate (read-only)
    PDriftChange,  // prob. for clock drift being changed
    PFast,
    PSlow,  // prob. for clock rate being faster/slower than regular
    PReadFault,
    PWriteFault,  // prob. for wrong reading/setting the clock
    WobbleDrift,  // data for wobbling the drift over a certain range

    // Controller, Supervisor, ClockSync
    TimerGranularity,  // duty or pSOS timer granularity - read only

    // Supervisor, Controller
    EventReport,  // what types of events should be reported
    EventClassReport,  // what classes of events should be reported
    InterfacePeriod   // process period (Supervisor: also snapshot period)
};


//**************************************************************************
// system commands
//**************************************************************************

// a list of commands that can be requested by the evaluation
enum ESystemCommand
{
    // Controller
    CreateApplication = 0,  // start a new synchronization application
    DeleteApplication,  // kill an existing synchronization appl.
    SuspendApplication,  // suspend an application (stop operation)
    ResumeApplication,  // resume operation of an application
    SuspendSimulation,  // suspend all applications (in HW: of the node)
    ResumeSimulation,  // resume all applications (in HW: of the node)
    ResetSimulation,  // reset all application modules and all networks
    KillSimulation,  // delete everything, stop program execution
    CreateNetwork,  // create a new network (pure simulation only!)
    DeleteNetwork,  // delete a network (pure simulation only!)

    // Supervisor
    EnableSnapshot,  // trigger a snapshot once every period
    DisableSnapshot,  // disable the triggering of periodic snapshots

    // Network
    ConnectNode,  // connect a node to the network
    DisconnectNode,  // remove an existing connection

    // ClockSync
    ConnectNetwork,  // add the network to the list of associated networks
    DisconnectNetwork,  // remove the network from the list of assoc. nets
    ReportTransmissions,  // send real transmission delays to EvalSys
    DontReportTransmissions,  // do not send real delays to EvalSys
    UserCommand,  // user defined command for the synchronization algorithm

    // Clock
    ConnectGps,  // connect a GPS receiver to a GPU of the UTCSU
    DisconnectGps,  // disconnect a GPS receiver from the UTCSU

    // Network, Lance, Clock
    CrashModule,  // crash the module
    RecoverModule,  // recover a crashed module

    // all modules
    ResetModule  // restore the module to its initial state
};


//**************************************************************************
// message replies
//**************************************************************************

// possible types of message replies
enum EReplyType
{
    ReplyBoolean = 0,  // reply contains a boolean value
    ReplyDouble,  // reply contains a double value
    ReplyStructure  // reply contains a whole structure
};


//**************************************************************************
// types of messages between evaluation and simulation; the first byte in
// a message contains the type
//**************************************************************************

// note: EMessageType is cast to INT8 in the socket interface class for
// stream sockets (CStreamInterface), so the values should be in the
// range [0,255]
enum EMessageType
{
    // EvalSys -> Sim
    MsgCommand = 0,  // execute a command
    MsgSetParameter,  // set a parameter
    MsgGetParameter,  // get a parameter

    // Sim -> EvalSys
    MsgEvent,  // report an event
    MsgReply,  // reply to a message
    MsgModule,  // send an INT8 with the module (Controller, Supervisor)
    MsgByteOrder,  // send the INT16 for checking the byte-order
    MsgEnumSizes,  // send a structure containing the sizes of all enums

    // EvalSys <-> Sim
    MsgResynchronize,  // request for resynchronization
    MsgSynchronizationByte = 0xFF // resynchronization byte
    // the last two commands are for resynchronization if the communication
    // has gone awry; if one side is out of sync, it sends a message
    // 'MsgResynchronize' to its partner which responds with the message
    // 'MsgSynchronizationByte'
};


#endif


