//
// process.hpp
//
// interface between the SimUTC simulation and the underlying simulation
// environment C++SIM; only the needed functions are provided
// note: this file is necessary for switching between the PC (no threads)
//       and Linux (testing the thread-specific things)
//
// created: 12.10.96
//
// last change: 09.10.98
//
// RCS version control information:
//     $Revision: 1.4 $$State: Exp $$Date: 1998/10/09 14:12:55 $
//


#ifndef _PROCESS_HPP_
#define _PROCESS_HPP_

#include "psos.h"

#include "bool.hpp"


extern const unsigned long DefaultProcessID;
// for processes that don't have an ID at the time of creation

extern const INT32 TICKS2SEC;
// one second is TICKS2SEC pSOS ticks


#ifndef LINUX_SIM
  #define NO_RESOURCE
  // this define is used to (un-)reference a process; if C++SIM is not
  // used, you should always define NO_RESOURCE
#endif

#ifdef LINUX_SIM

// use the class library
#ifndef PROCESS_H_
#  include <ClassLib/Process.h>
#endif


class ACProcess : public Process
{
public:

    //----------------------------------------------------------------------
    //--- constructor
    //----------------------------------------------------------------------

    ACProcess (unsigned long nodeID = DefaultProcessID) { nodeID_ = nodeID; }
    // sets the ID of the node for which the process works


    //----------------------------------------------------------------------
    //--- management of the node ID
    //----------------------------------------------------------------------

    virtual void SetNodeID (unsigned long nodeID) { nodeID_ = nodeID; }
    // set/change the node ID

    virtual unsigned long GetNodeID () const { return nodeID_; }
    // return the node ID


    //----------------------------------------------------------------------
    //--- pSOS functions
    //----------------------------------------------------------------------

    void tm_wkafter (INT32 time) { Hold ((double)time/TICKS2SEC); }


private:

    unsigned long nodeID_;
    // the ID of the node for which this process works
};

#else
// create a dummy process interface through ACProcess


//**************************************************************************
// class ACProcess: this class is the interface to the process class of
//     C++SIM; only the functions provided by this class are needed by
//     the SimUTC processes
//**************************************************************************

class ACProcess
{
public:

    //----------------------------------------------------------------------
    //--- constructor, destructor
    //----------------------------------------------------------------------

    ACProcess (unsigned long nodeID = DefaultProcessID);
    virtual ~ACProcess();


    //----------------------------------------------------------------------
    //--- from class Thread inherited thread scheduling funtions
    //----------------------------------------------------------------------

    virtual void Suspend();


    //----------------------------------------------------------------------
    //--- process (re-)scheduling or descheduling
    //----------------------------------------------------------------------

    void Activate ();
    void ReActivate ();
    void ReActivateAt (double AtTime = 0);
    void ReActivateDelay (double AtTime = 0);
    void Cancel();
    virtual void terminate();


    //----------------------------------------------------------------------
    //--- determining the process state
    //----------------------------------------------------------------------

    EBool idle () const;


    //----------------------------------------------------------------------
    //--- scheduling information
    //----------------------------------------------------------------------

    double evtime() const;
    static const ACProcess* current();
    // changed the return value to ACProcess, but that's okay


    //----------------------------------------------------------------------
    //--- simulation information
    //----------------------------------------------------------------------

    static double CurrentTime ();


#ifdef DOS_SIM

    //----------------------------------------------------------------------
    //--- simulation control; no part of CProcess, but needed for debugging
    //--- DO NOT USE!
    //----------------------------------------------------------------------

    static void SetCurrentTime (double time);
    // set the simulation time

    static void SetActiveProcess (ACProcess *process);
    // set the active process (that is returned by current())

#endif


    //----------------------------------------------------------------------
    //--- management of the node ID
    //----------------------------------------------------------------------

    virtual void SetNodeID (unsigned long nodeID);
    // set/change the node ID

    virtual unsigned long GetNodeID () const;
    // return the node ID


    //----------------------------------------------------------------------
    //--- the main function of the process
    //----------------------------------------------------------------------

    virtual void Body() = 0;


protected:
#ifdef PUBLIC_IN_TEST
public:
#endif

    //----------------------------------------------------------------------
    //--- deactivation and rescheduling
    //----------------------------------------------------------------------

    void Hold (double t);

#ifndef PSOS_SIM
    void tm_wkafter (INT32 time);  // simulate the pSOS-function
#endif


private:

    enum EState {Terminated, Suspended, Passive, Scheduled};

    double assocTime_;
    EState state_;

    unsigned long nodeID_;
    // the ID of the node for which this process works

    static double currTime_;  // current simulation time
    static ACProcess *currentProcess_;  // currently active process
};

#endif
// no LINUX_SIM

#endif


