//
// sockstre.hpp
//
// declaration of the basic stream socket interface class, and access
// classes for client and server stream sockets
//
// created: 07.02.97
//
// last change: 29.07.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/07/29 16:34:30 $
//


#ifndef _SOCKSTRE_HPP_
#define _SOCKSTRE_HPP_

#include "psos.h"

#include "bool.hpp"
#include "params.hpp"
#include "sockbase.hpp"


struct SBase;


//**************************************************************************
// class CStreamInterface: used as an interface to a stream socket;
//      uses the communication protocol specified for the IFC between
//      supervisors and the evaluation
//      can send and receive messages (which consist of a message header, a
//      message ID, and a structure derived from SBase (cf. structs.hpp)),
//      manages the creation/deletion of the socket, keeps track of message
//      IDs
//**************************************************************************

class CStreamInterface : public CSocket
{
public:

    //----------------------------------------------------------------------
    //--- constructor
    //----------------------------------------------------------------------

    CStreamInterface (ACSocketAddress *addr, INT32 sid);
    // assume that a stream socket with the ID 'sid' exists and is open


    //----------------------------------------------------------------------
    //--- message ID management
    //----------------------------------------------------------------------

    SINT32 GetMessageID () const;
    // returns the ID of the last message that has been received or
    // NoAccount, if no message has ever been received


    //----------------------------------------------------------------------
    //--- information about the byte order
    //----------------------------------------------------------------------

    EBool ByteOrderReversed () const;
    // returns True if the byte order of the communcation partners is not
    // the same (use Send/ReceiveByteOrder() to find out about that), False
    // if the order is the same or if it was never checked


    //----------------------------------------------------------------------
    //--- message transmission
    //----------------------------------------------------------------------

    INT32 SendMessage (EMessageType type, SBase *data);
    // transmits the type, ID and structure over the socket (the ID is
    // inserted automatically), returns the message ID or NoAccount (-1)
    // if the send operation failed
    // assert: data != NULL


    //----------------------------------------------------------------------
    //--- message reception
    //----------------------------------------------------------------------

    EBool ReceiveMessageHeader (EMessageType *type, double timeout = 0);
    // get the first datum of a message (default: non-blocking); the call
    // blocks according to 'timeout' (seconds); return True if successful;
    // assert: type != NULL

    EBool ReceiveMessage (SBase **data);
    // the function assumes that the message ID is the first item in the
    // socket, followed by a structure derived from SBase and that the
    // first member in SBase is 'INT32 size', denoting the size of the
    // structure (without the 'size' member); it allocates the appropriate
    // amount of memory, reads the structure and returns a pointer to the
    // memory in 'data'
    // note: the memory *must* be returned using ReturnMessageBuffer()!
    // assert: data != NULL
    

    //----------------------------------------------------------------------
    //--- message reply
    //----------------------------------------------------------------------

    EBool ReplyToMessage (SBase *data);
    // send a MsgReply to the last received message over the socket,
    // containing the result (data of type SBase) of a message;
    // returns True if successful, else False
    // assert: data != NULL
    // assert: a message has been received


    //----------------------------------------------------------------------
    //--- special messages
    //----------------------------------------------------------------------

    EBool SendSynchronizationByte();
    // send the resync byte to the evaluation; returns True if the message
    // was sent successfully, else False

    EBool Resynchronize();
    // ask the evaluation to resynchronize; returns True if the message
    // was sent successfully, else False

    EBool SendModule (EModule module);
    // send the module type (cast to INT8); returns True if the message
    // was sent successfully, else False
    // use ReceiveModule() to get the module of the communication partner

    EBool ReceiveModule (EModule& module, double timeout=0);
    // receive the module; reception of the message header blocks according
    // to 'timeout', reception of the INT8 uses the internal timeout;
    // if the message was received successfully, 'module' will be modified
    // to contain the module type that was received and True is returned;
    // else False is returned and 'module' is not changed

    EBool SendByteOrder();
    // send the byte order test; returns True if the message was sent
    // successfully, else False
    // use ReceiveByteOrder() to read the byte order message

    EBool ReceiveByteOrder (double timeout=0);
    // receive the byte order test; message header reception blocks
    // according to 'timeout' (in seconds), the reception of the INT16
    // blocks according to the internal timeout; if the order is reversed,
    // byteOrderReversed is set to True; returns True if the message was
    // received successfully, else False


    //----------------------------------------------------------------------
    //--- message buffer handling
    //----------------------------------------------------------------------

    static void *GetMessageBuffer (INT32 size);
    // allocates 'size' bytes and returns a pointer to the memory

    static void ReturnMessageBuffer (void *buffer);
    // deallocates the buffer; assumes that it has been allocated using
    // GetMessageBuffer()
    // note: buffers obtained by ReceiveMessage() *must* be returned using
    // this function


protected:

    //----------------------------------------------------------------------
    //--- internal constructor
    //----------------------------------------------------------------------

    CStreamInterface (ACSocketAddress *addr);
    // store the address of the socket


    //----------------------------------------------------------------------
    //--- internal member initialization
    //----------------------------------------------------------------------

    virtual void Init ();
    // set members to default values


    //----------------------------------------------------------------------
    //--- internal message transmission
    //----------------------------------------------------------------------

    EBool SendMessageHeader (EMessageType type);
    // send the first byte of the message (ie. the type); returns True if
    // the message was sent successfully, else False

    EBool SendMessageID ();
    // send the (default) message ID; returns True if the message was sent
    // successfully, else False

    EBool SendMessageID (INT32 id);
    // send a special message ID; returns True if the message was sent
    // successfully, else False

    EBool SendMessage (void *data, INT32 size);
    // send the size and data; returns True if the message was sent
    // successfully, else False


    //----------------------------------------------------------------------
    //--- internal message ID reception
    //----------------------------------------------------------------------

    EBool ReceiveMessageID (double timeout = 0);
    // receive the message ID (default: non-blocking); the call blocks
    // according to 'timeout'; returns True if successful


private:

    INT32 sendMsgID_;  // ID of the next message that is to be sent
    SINT32 receiveMsgID_;  // ID of the currently received message
    EBool byteOrderReversed_;
    // True if the byte-order of the communication partners is not the same
};



//**************************************************************************
// class CStreamServer: creates a stream socket, listens to it, can
//     accept connections
//**************************************************************************

class CStreamServer : public CStreamInterface
{
public:

    //----------------------------------------------------------------------
    //--- constructor
    //----------------------------------------------------------------------

    CStreamServer (ACSocketAddress *addr);
    // create the socket, listen to it

    virtual ~CStreamServer ();
    // close the socket (unlink it if necessary)


    //----------------------------------------------------------------------
    //--- accept connections
    //----------------------------------------------------------------------

    CStreamInterface *Accept();
    // accepts a connection; returns NULL if the accept has failed, calls
    // ReceiveByteOrder() if the accept has succeeded
};



//**************************************************************************
// class CStreamClient: identifies a stream socket, connects to it
//**************************************************************************

class CStreamClient : public CStreamInterface
{
public:

    //----------------------------------------------------------------------
    //--- constructor
    //----------------------------------------------------------------------

    CStreamClient (ACSocketAddress *addr);
    CStreamClient (ACSocketAddress *addr, int numberOfTrials);
    // specify the number of connection attempts before giving up
    // note: if the connection could not be established, an assert will
    // fail


protected:

    //----------------------------------------------------------------------
    //--- socket connection
    //----------------------------------------------------------------------

    EBool Connect (ACSocketAddress *addr, int numberOfTrials);
    // tries to identify and to connect to the socket; gives up after
    // 'numberOfTrials' attempts; if the connection could be established,
    // SendByteOrder() is called
};


#endif


