//
// structs.hpp
//
// all structures and classes that are needed to manage system parameters
// and events
// TAKE CARE: every class should set the size parameter to the size of the
// struct in its ctor
// NOTE: only use public members in classes derived from SBase (because
// ordering of the members of different access types is compiler specific)
//
// created: 24.01.97
//
// last change: 06.11.98
//
// RCS version control information:
//     $Revision: 1.17 $$State: Exp $$Date: 1998/11/06 10:54:37 $
//


#ifndef _STRUCTS_HPP_
#define _STRUCTS_HPP_

#include "bool.hpp"
#include "clockcfg.hpp"
#include "database.hpp"
#include "distrib.hpp"
#include "error.hpp"
#include "event.hpp"
#include "general.hpp"
#include "params.hpp"
#include "time.hpp"


//**************************************************************************
// base struct for events, parameters and commands
//**************************************************************************

struct SBase
{
    INT32 size;  // size of the structure; should be set in the ctors
    // note: the type *must* be INT32, that is expected in CStreamInterface

    SBase() { size = sizeof(*this); }
};

struct SModuleBase : public SBase
{
    EModule module;  // the module reporting the event
    SModuleBase() { size = sizeof(*this); }
};


//**************************************************************************
// structs needed for events
//**************************************************************************

// base struct for all events that are logged
struct SEvent : public SModuleBase
{
    EEvent type;  // the type of the event that is reported
    double time;  // global timestamp of the report (drawn by NoteEvent())
    SEvent() { size = sizeof(*this); }
};


//--------------------------------------------------------------------------
// network events
//--------------------------------------------------------------------------

// network events include the networkID
struct SNetEvent : public SEvent
{
    SINT32 networkID;
    SNetEvent() { size = sizeof(*this); }
};


// sent if the crash status of a network changes
struct SNetCrashEvent : public SNetEvent
{
    EBool crashed;  // True if network is now crashed, False if recovered
    SNetCrashEvent() { size = sizeof(*this); }
};


// sent by the network if something happens to a message
struct SNetMessageEvent : public SNetEvent
{
    SINT32 sender;  // the nodeID of the sender
    int area;  // the area of the sender
    SINT32 receiver;  // the nodeID of the receiver; -1 for broadcast msg
    INT32 msgID;  // the number of the message sent by this CSA
    SNetMessageEvent() { size = sizeof(*this); }
};


//--------------------------------------------------------------------------
// application events
//--------------------------------------------------------------------------

// base structure of all application events; contains the timestamp of the
// clock
struct SApplEvent : public SEvent
{
    STimestamp clockTime;  // current value of the clock
    SINT32 nodeID;  // ID of the application
    int area;  // area of the UTCSU
    SApplEvent() { size = sizeof(*this); }
};


// a network-specific event sent by the application
struct SApplNetEvent : public SApplEvent
{
    SINT32 networkID;
    SApplNetEvent() { size = sizeof(*this); }
};


// GPS update: contains the difference between GPS time and clock time
struct SApplGpsEvent : public SApplEvent
{
    double diff;  // difference of GPS time from clock time
    SApplGpsEvent() { size = sizeof(*this); }
};


// clock update: contains the new time as well as the current clock value
struct SApplClockEvent : public SApplEvent
{
    STimestamp newTime;  // new value of the clock if it is set anew
    SApplClockEvent() { size = sizeof(*this); }
};


struct SApplPreloadEvent : public SApplClockEvent
{
    double loadAt;  // the clock time at which the amortization starts
    SApplPreloadEvent() { size = sizeof(*this); }
};


// clock time interval: contains the time interval
struct SApplTimeEvent : public SApplEvent
{
    EBool input;  // True...input interval, False...result of the CV
    STimeInterval interval;  // the interval (low, high, refpoint) itself
    SApplTimeEvent() { size = sizeof(*this); }
};


// rate interval: contains the rate interval
struct SApplRateEvent : public SApplEvent
{
    EBool input;  // True...input interval, False...resulting local rate
    SRate rate;  // the interval (thetaMinus, thetaPlus, u) itself
    double refPoint;  // the reference point
    SApplRateEvent() { size = sizeof(*this); }
};


// sent if the crash status of lance/clock changes
struct SApplCrashEvent : public SApplEvent
{
    EBool crashed;  // True if network is now crashed, False if recovered
    SApplCrashEvent() { size = sizeof(*this); }
};


// sent by the sender CSA if something happens to a message
struct SApplSendMsgEvent : public SApplEvent
{
    INT32 msgID;  // the number of the message sent by this CSA
    SINT32 receiver;  // nodeID of the receiver (-1 for broadcast)
    SApplSendMsgEvent() { size = sizeof(*this); }
};


// sent by the receiver CSA if something happens to a message
struct SApplRecvMsgEvent : public SApplSendMsgEvent
{
    SINT32 sender;  // nodeID of the sender
    int senderArea;  // the area of the sender
    SApplRecvMsgEvent() { size = sizeof(*this); }
};

struct SApplDelayEvent : public SApplRecvMsgEvent
{
    STimestamp delaytime;
    SApplDelayEvent() { size = sizeof(*this); }
};

// is sent if the clock sync. algorithm has computed a new transmission
// delay estimation
struct SApplDelayEstimationEvent : public SApplEvent
{
    SINT32 networkID;  // network from which the data stems
    SINT32 sender;  // nodeID of the sender node
    SDelay delay;  // the new transmission delay estimations

    SApplDelayEstimationEvent() { size = sizeof(*this); }
};


//**************************************************************************
// structs for setting/getting system parameters
//**************************************************************************

// the base class needed for all parameters
struct SParameter : public SModuleBase
{
    ESystemParameter type;  // the parameter type
    SParameter() { size = sizeof(*this); }
};


//--------------------------------------------------------------------------
// network parameters
//--------------------------------------------------------------------------

// base struct for parameters concerning the network
struct SNetworkParameter : public SParameter
{
    SINT32 networkID;  // needed for determining which network is meant
    SNetworkParameter() { size = sizeof(*this); }
};

struct SNetDistribution : public SNetworkParameter
{
    SRandomData data;  // the data necessary to draw random values
    SNetDistribution() { size = sizeof(*this); }
};

struct SNetDistributionP2P : public SNetDistribution
{
    SINT32 sender;
    SINT32 receiver;
    SNetDistributionP2P() { size = sizeof(*this); }
};

// same as SSingleParameter, but for networks
struct SSingleNetworkParameter : public SNetworkParameter
{
    double value;
    SSingleNetworkParameter() { size = sizeof(*this); }
};


// this struct is used for transmission delay parameters (network)
// used by: CNetworkAccountant
struct SNetworkData : public SSingleNetworkParameter
{
    SINT32 sender;
    SINT32 receiver;
    SNetworkData() { size = sizeof(*this); }
};


//--------------------------------------------------------------------------
// application parameters
//--------------------------------------------------------------------------

// necessary for all GetParameter() and most SetParameter() Commands
struct SSingleParameter : public SParameter
{
    double value;  // the parameter value
    SSingleParameter() { size = sizeof(*this); }
};


// necessary for setting distribution functions and their parameters
struct SApplDistribution : public SParameter
{
    SRandomData data;  // the data necessary to draw random values
    SApplDistribution() { size = sizeof(*this); }
};


// necessary for setting all data needed for wobbling the drift
struct SApplWobbleDrift : public SApplDistribution
{
    // inherits 'SRandomData data' for setting distribution of rhoAdd
    double period;  // time in seconds between two wobble modifications
    double rhoMin;  // lower bound on drift during wobbling
    double rhoMax;  // upper bound on drift during wobbling
    SApplWobbleDrift() { size = sizeof(*this); }
};


// this struct is used for setting/obtaining the clock value
// used by: CClockInterface
struct STimeParameter : public SParameter
{
    INT32 macrostamp;  // the clock time value (macrostamp, timestamp)
    INT32 timestamp;
    STimeParameter() { size = sizeof(*this); }
};


// this struct is used for transmission delay parameters (delay measurement)
// used by: CApplicationDatabase
struct STransmissionData : public SParameter
{
    SINT32 networkID;  // needed for determining which network is meant
    SINT32 nodeID;  // specify the node itself
    double value;
    STransmissionData() { size = sizeof(*this); }
};


// this struct is used for measured clock rate parameters (rate sync.)
// used by: CApplicationDatabase
struct SRateData : public STransmissionData
{
    int area;  // area of the UTCSU
    SRateData() { size = sizeof(*this); }
};


// needed for changing the events and event classes that are reported
struct SEventReport : public SSingleParameter
{
    EBool set;  // True if the bit (mask) should be set; False if not
    INT32 masksize;  // size of the mask
    char mask[EventCount/8+1];  // stores the bitmask
/**/// note: it is assumed that there are more events than event classes!

    SEventReport ();

    void Set (EEvent event, EBool on);
    // set a single event
    void Set (EEventClass eventClass, EBool on);
    // set a single event class
    void Set (ESystemParameter eventType, void *map, INT32 s);
    // set a complete event (class) mask

    void Get (EBool& on) const;
    // get the state of a single event (class)
    void Get (SINT32& value) const;
    // find out what value (event or event class) is stored in the base
    // struct
};


//**************************************************************************
// structs for executing system commands
//**************************************************************************

// the base struct needed for all commands
struct SCommand : public SModuleBase
{
    ESystemCommand type;  // the command type
    SCommand() { size = sizeof(*this); }
};


struct SGpsConnectCommand : public SCommand
{
    EGpsUnit gpu;
    SGpsConnectCommand() { size = sizeof(*this); }
};


// needed for all commands to the Controller that concern one application
struct SApplCommand : public SCommand
{
    SINT32 nodeID;
    int area;
    SApplCommand() { size = sizeof(*this); }
};

// needed to create a network
struct SNetworkCreateCommand : public SCommand
{
    ENetworkType networkType;  // type of the network
    SNetworkCreateCommand() { size = sizeof(*this); }
};

// needed to delete a network
struct SNetworkCommand : public SCommand
{
    SINT32 networkID;
    SNetworkCommand() { size = sizeof(*this); }
};

// needed for node connection/disconnection
struct SNetworkConnectCommand : public SNetworkCommand
{
    SINT32 nodeID;
    SNetworkConnectCommand() { size = sizeof(*this); }
};


//**************************************************************************
// structs needed for replies
//**************************************************************************

// base structure for all replies
// also used as reply for GetParameter messages; if transmitting that reply
// over the socket, enclose the data directly behind the structure;
// set the size to contain the size of both SReply and the data structure
struct SReply : public SModuleBase
{
    EReplyType type;  // the type of the reply
    SReply () { type = ReplyStructure; size = sizeof(*this); }
};

// normal reply for Command and SetParameter messages
struct SBooleanReply : public SReply
{
    EBool result;

    SBooleanReply () { size = sizeof(*this); }
    SBooleanReply (EBool res)
        { type = ReplyBoolean; result = res; size = sizeof(*this); }
};

// used to return the networkID in reply to a CreateNetwork command
struct SDoubleReply : public SReply
{
    double value;

    SDoubleReply () { size = sizeof(*this); }
    SDoubleReply (double val)
        { type = ReplyDouble; value = val; size = sizeof(*this); }
};


//**************************************************************************
// struct containing the enum sizes
//**************************************************************************

// contains the sizes of all enums used in the structs above, plus the
// size of the basic data types with unknown size (int, double)
struct SEnumSizes : public SBase
{
    INT8 sizedouble;
    INT8 sizeint;

    INT8 sizeEBool;
    INT8 sizeEEvent;
    INT8 sizeEGpsUnit;
    INT8 sizeEModule;
    INT8 sizeENetworkType;
    INT8 sizeEReplyType;
    INT8 sizeESystemCommand;
    INT8 sizeESystemParameter;

    SEnumSizes ()
    {
        size = sizeof(*this);

        sizedouble = sizeof(double);
        sizeint = sizeof(int);

        sizeEBool = sizeof(EBool);
        sizeEEvent = sizeof(EEvent);
        sizeEGpsUnit = sizeof(EGpsUnit);
        sizeEModule = sizeof(EModule);
        sizeENetworkType = sizeof(ENetworkType);
        sizeEReplyType = sizeof(EReplyType);
        sizeESystemCommand = sizeof(ESystemCommand);
        sizeESystemParameter = sizeof(ESystemParameter);
    }
};


#endif


