//
// utcsu.hpp
//
// UTCSU memory layout for read and write access to the UTCSU registers
//
// created: 17.02.97
//
// last change: 24.03.98
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 09:57:06 $
//


#ifndef _UTCSU_HPP_
#define _UTCSU_HPP_


#include "psos.h"


const unsigned short NumberOfGPU = 3;
// number of GPU's of the UTCSU

const unsigned short NumberOfSSU = 6;
// number of SSU's of the UTCSU


//**************************************************************************
// struct SUtcsuRead: the register map of the UTCSU for read access (only
//      the necessary parts); taken from the UTCSU Block Diagram
//      (chapter 6, p. 73)
//**************************************************************************

// GPU1-3 (GPS Unit 1-3)
struct S_GPUread {
    INT32 MSGPS;  // macro stamp GPS
    INT32 TSGPS;  // time stamp GPS
    INT32 unused8[6];
};

// SSU1-6 (Synchronization Subnet Unit 1-6)
struct S_SSUread {
    INT32 MSRCV;  // macro stamp receive
    INT32 TSRCV;  // time stamp receive
    INT32 MSXMT;  // macro stamp transmit
    INT32 TSXMT;  // time stamp transmit
    INT32 ACCPXMT; // ACCPXMT [15:0] and ACCNXMT [31:16] (acc. P/N transmit)
    INT32 ACCNPLXMT;  // lower part of ACCP(N)XMT
    INT32 ACCNPHXMT;  // higher part of ACCP(N)XMT
    INT32 unused9;
};


struct SUtcsuRead  // read access
{
    // LTU (Local Time Unit)
    INT32 MSGET;   // macro stamp get [23:0] and checksum get [31:24]
    INT32 TSGETL;   // time stamp get latch
    INT32 USGET;   // micro stamp get
    INT32 TSGET;  // time stamp get
    INT32 NSGET;   // nano stamp get
    INT32 unused1;
    INT32 AMORTTIMEGETL;
    INT32 AMORTTIMEGET;

    // ACU (Accuracy Unit)
    INT32 STATEPGETL;  // STATEPGETL [15:0] and STATENGETL [31:16]
    INT32 STATEPGET;  // STATEPGET [15:0] and STATENGET [31:16]
    INT32 ALPHAPGETL;
    INT32 ALPHANGETL;
    INT32 ALPHAPGET;
    INT32 ALPHANGET;
    INT32 NALPHAPGET;  // NALPHAPGET [15:0] and NALPHANGET [31:16]
    INT32 unused2;

    // CLU (Clock Unit)
    INT32 unused3[8];

    // ITU (Interrupt Unit)
    INT32 IDENTIFICATION;
    INT32 UTCSTAT1;
    INT32 UTCSTAT2;
    INT32 UTCCONF1;
    INT32 UTCCONF2;
    INT32 UTCINTSTAT1;
    INT32 UTCINTSTAT2;
    INT32 unused4;

    // BTU (Built-in Test Unit)
    INT32 unused5[8];

    // SNU (Snapshot Unit)
    INT32 MSSNU;  // macro stamp snapshot
    INT32 TSSNU;  // time stamp snapshot
    INT32 ACCPSNU;  // ACCPSNU [15:0] and ACCNSNU [31:16] (acc. snapshot)
    INT32 unused6[5];

    // APU (Application Unit)
    INT32 MSAPP;  // macro stamp application
    INT32 TSAPP;  // time stamp application
    INT32 ACCPAPP;  // ACCPAPP [15:0] and ACCNAPP [31:16] (accuracy appl.)
    INT32 MSAPPX;  // macro stamp application X
    INT32 TSAPPX;  // time stamp application X
    INT32 ACCPAPPX;  // ACCPAPP [15:0] and ACCNAPP [31:16] (acc. appl. X)
    INT32 unused7[2];

    // GPU1-3 (GPS Unit 1-3)
    S_GPUread gpu[NumberOfGPU];

    // SSU1-6 (Synchronization Subnet Unit 1-6)
    S_SSUread ssu[NumberOfSSU];


    // the control bits of the UTCSU (read access)
    enum EConfiguration  // UTCCONF1/2 in ITU
    {
        // UTCCONF1
        CLEAPSECOND = 1<<0,  // do leap adjustment (0..no, 1..yes)
        CLEAPSINS   = 1<<1,  // 1..insert leap second; 0..delete leap sec.
        CSARINTS    = 1<<16,
        CSARINPU    = 1<<17,
        CSARSTRTS   = 1<<18,
        CSARSTRPU   = 1<<19,
        CSUMINTS    = 1<<20,
        CSUMINPU    = 1<<21,
        CSUMSTRTS   = 1<<22,
        CSUMSTRPU   = 1<<23,
        CPOLHWSNAP  = 1<<28,  // polarity (1..rising edge, 0..falling edge)
        CPOLSYNCRUN = 1<<29,

        // UTCCONF2
        CPOLAPP      = 1<<0,  // polarity (1/0 ... see above)
        CPOLAPP1     = 1<<1,
        CPOLAPP2     = 1<<2,
        CPOLAPP3     = 1<<3,
        CPOLAPP4     = 1<<4,
        CPOLAPP5     = 1<<5,
        CPOLAPP6     = 1<<6,
        CPOLAPP7     = 1<<7,
        CPOLAPP8     = 1<<8,
        CPOLGPS1PPS1 = 1<<12,
        CPOLGPS1PPS2 = 1<<13,
        CPOLGPS1PPS3 = 1<<14,
        CPOLGPSSTAT1 = 1<<15,
        CPOLGPSSTAT2 = 1<<16,
        CPOLGPSSTAT3 = 1<<17,
        CPOLRCV1     = 1<<20,
        CPOLRCV2     = 1<<21,
        CPOLRCV3     = 1<<22,
        CPOLRCV4     = 1<<23,
        CPOLRCV5     = 1<<24,
        CPOLRCV6     = 1<<25,
        CPOLXMT1     = 1<<26,
        CPOLXMT2     = 1<<27,
        CPOLXMT3     = 1<<28,
        CPOLXMT4     = 1<<29,
        CPOLXMT5     = 1<<30,
        CPOLXMT6     = 1<<31
    };
    enum EStatus  // UTCSTAT1/2 in ITU
    {
        // UTCSTAT1
        SCAMA    = 1<<0,  // continuous amort. phase active (1..active)
        SAPN     = 1<<2,
        SANN     = 1<<3,
        SAPPL    = 1<<4,
        SAPNL    = 1<<5,
        SANPL    = 1<<6,
        SANNL    = 1<<7,
        SBNDP    = 1<<8,
        SBNDN    = 1<<9,
        SPLLLOCK = 1<<10,
        IINTT    = 1<<12,
        IINTN    = 1<<13,
        IINTA    = 1<<14,
        SHWSNAP  = 1<<18,  // hardware snapshot
        SSWDUTY  = 1<<19,
        STSAPP   = 1<<20,
        STSAPP1  = 1<<21,
        STSAPP2  = 1<<22,
        STSAPP3  = 1<<23,
        STSAPP4  = 1<<24,
        STSAPP5  = 1<<25,
        STSAPP6  = 1<<26,
        STSAPP7  = 1<<27,
        STSAPP8  = 1<<28,
        SAPPDUTY = 1<<29,// 1..DUTY>=INTPBUS and E=1, 0..DUTY<INTPBUS or E=0
        SAPPREF  = 1<<30,

        // UTCSTAT2
        SGPS1PPS1 = 1<<0,  // 1 PPS from GPU1
        SGPS1PPS2 = 1<<1,
        SGPS1PPS3 = 1<<2,
        SGPSSTAT1 = 1<<3,  // status from GPU1
        SGPSSTAT2 = 1<<4,
        SGPSSTAT3 = 1<<5,
        STSRCV1   = 1<<8,
        STSRCV2   = 1<<9,
        STSRCV3   = 1<<10,
        STSRCV4   = 1<<11,
        STSRCV5   = 1<<12,
        STSRCV6   = 1<<13,
        STSXMT1   = 1<<14,
        STSXMT2   = 1<<15,
        STSXMT3   = 1<<16,
        STSXMT4   = 1<<17,
        STSXMT5   = 1<<18,
        STSXMT6   = 1<<19,
        SDUTYA1   = 1<<20,  // duty timer A
        SDUTYA2   = 1<<21,
        SDUTYA3   = 1<<22,
        SDUTYA4   = 1<<23,
        SDUTYA5   = 1<<24,
        SDUTYA6   = 1<<25,
        SDUTYB1   = 1<<26,  // duty timer B
        SDUTYB2   = 1<<27,
        SDUTYB3   = 1<<28,
        SDUTYB4   = 1<<29,
        SDUTYB5   = 1<<30,
        SDUTYB6   = 1<<31
    };
    enum EInterrupt  // UTCINTSTAT1/2 in ITU
    {
        // UTCINTSTAT1
        ICAMS    = 1<<0,  // rising edge, INTT: continuous amort. phase started
        ICAMT    = 1<<1,  // falling edge, INTT: cont. amort. phase terminated
        INTPO    = 1<<2,  // pulse, INTT: NTPTIME overrun (yr. 2036, leap sec.)
        IBNDP    = 1<<4,  // rising, INTT
        IBNDN    = 1<<5,  // rising, INTT
        IAPPL    = 1<<6,  // rising, INTT
        IAPNL    = 1<<7,  // rising, INTT
        IANPL    = 1<<8,  // rising, INTT
        IANNL    = 1<<9,  // rising, INTT
        IFAILED  = 1<<10,  // perm. low, INTT
        IPUREPH  = 1<<12,  // pulse, INTA
        IHWSNAP  = 1<<16,  // pulse, INTA: hardware snapshot event
        ISWSNAP  = 1<<17,  // pulse, INTA
        ISYNCRUN = 1<<18,  // pulse, INTA
        ITSAPP   = 1<<20,  // pulse, INTA
        ITSAPP1  = 1<<21,  // pulse, INTA
        ITSAPP2  = 1<<22,  // pulse, INTA
        ITSAPP3  = 1<<23,  // pulse, INTA
        ITSAPP4  = 1<<24,  // pulse, INTA
        ITSAPP5  = 1<<25,  // pulse, INTA
        ITSAPP6  = 1<<26,  // pulse, INTA
        ITSAPP7  = 1<<27,  // pulse, INTA
        ITSAPP8  = 1<<28,  // pulse, INTA
        IAPPDUTY = 1<<29,  // pulse, INTA: INTPBUS >= DUTY and E=1

        // UTCINTSTAT2
        IGPS1PPS1 = 1<<0,  // pulse, INTN: 1 PPS from GPU1
        IGPS1PPS2 = 1<<1,  // pulse, INTN
        IGPS1PPS3 = 1<<2,  // pulse, INTN
        IGPSSTAT1 = 1<<3,  // pulse, INTN: status information from GPU1
        IGPSSTAT2 = 1<<4,  // pulse, INTN
        IGPSSTAT3 = 1<<5,  // pulse, INTN
        ITSRCV1   = 1<<8,  // pulse, INTN
        ITSRCV2   = 1<<9,  // pulse, INTN
        ITSRCV3   = 1<<10,  // pulse, INTN
        ITSRCV4   = 1<<11,  // pulse, INTN
        ITSRCV5   = 1<<12,  // pulse, INTN
        ITSRCV6   = 1<<13,  // pulse, INTN
        ITSXMT1   = 1<<14,  // pulse, INTN
        ITSXMT2   = 1<<15,  // pulse, INTN
        ITSXMT3   = 1<<16,  // pulse, INTN
        ITSXMT4   = 1<<17,  // pulse, INTN
        ITSXMT5   = 1<<18,  // pulse, INTN
        ITSXMT6   = 1<<19,  // pulse, INTN
        IDUTYA1   = 1<<20,  // pulse, INTT: duty timer A
        IDUTYA2   = 1<<21,  // pulse, INTT
        IDUTYA3   = 1<<22,  // pulse, INTT
        IDUTYA4   = 1<<23,  // pulse, INTT
        IDUTYA5   = 1<<24,  // pulse, INTT
        IDUTYA6   = 1<<25,  // pulse, INTT
        IDUTYB1   = 1<<26,  // pulse, INTT: duty timer B
        IDUTYB2   = 1<<27,  // pulse, INTT
        IDUTYB3   = 1<<28,  // pulse, INTT
        IDUTYB4   = 1<<29,  // pulse, INTT
        IDUTYB5   = 1<<30,  // pulse, INTT
        IDUTYB6   = 1<<31   // pulse, INTT
    };
};


//**************************************************************************
// struct SUtcsuWrite: the register map of the UTCSU for write access
//      (again only the necessary parts)
//**************************************************************************

// SSU1-6 (Synchronization Subnet Unit 1-6)
struct S_SSUwrite {
    INT32 DUTYAH;  // duty timer A high preload
    INT32 DUTYAL;  // DUTYL (DT low preload) [15:0] and EA (enable bit) [16]
    INT32 DUTYBH;  // duty timer B high preload
    INT32 DUTYBL;  // DUTYL (DT low preload) [15:0] and EB (enable bit) [16]
    INT32 unused8[4];
};

struct SUtcsuWrite  // write access
{
    // LTU (Local Time Unit)
    INT32 MSSET;  // macro stamp set [23:0] and step low [31:24]
    INT32 TSSET;  // time stamp set
    INT32 USSET;  // micro stamp set
    INT32 NTPSET; // NTP-Time set
    INT32 STEPAMORT;  // step amortization phase
    INT32 STEPPURE;   // step pure phase
    INT32 AMORTTIMESET;
    INT32 STARTAMORT;

    // ACU (Accuracy Unit)
    INT32 LAMBDAPPURE;  // LAMBDAPPURE [15:0] and LAMBDAPAMORT [31:16]
    INT32 LAMBDANPURE;  // LAMBDANPURE [15:0] and LAMBDANAMORT [31:16]
    INT32 STATEPSET;
    INT32 STATENSET;
    INT32 BOUNDP;  // BOUNDP [15:0] and BOUNDN [31:16]
    INT32 ALPHASET;
    INT32 ALPHAPNSET;  // ALPHAPNSET [7:0]
    INT32 unused1;

    // CLU (Clock Unit)
    INT32 SWRESET;  // SWRESET [7:0]
    INT32 unused2[7];

    // ITU (Interrupt Unit)
    INT32 UTCCONF1;
    INT32 UTCCONF2;
    INT32 UTCINTEN1;  // interrupt enable 1
    INT32 UTCINTEN2;  // interrupt enable 2
    INT32 UTCINTCLEAR1;  // interrupt clear 1
    INT32 UTCINTCLEAR2;  // interrupt clear 2
    INT32 unused3[2];

    // BTU (Built-in Test Unit)
    INT32 unused4[8];

    // SNU (Snapshot Unit)
    INT32 SWSNAP;  // software snapshot
    INT32 DUTYH;  // duty timer high preload
    INT32 DUTYL;  // DUTYL (DT low preload) [15:0] and E (enable bit) [16]
    INT32 unused5[5];

    // APU (Application Unit)
    INT32 APPDUTYH;  // duty timer high preload
    INT32 APPDUTYL;  // DUTYL (DT low preload) [15:0] and E (enable bit) [16]
    INT32 APPSEL;  // application select [2:0]
    INT32 APPCLEAR;  // application clear [7:0]
    INT32 unused6[4];

    // GPU1-3 (GPS Unit 1-3)
    INT32 unused7[8*NumberOfGPU];

    // SSU1-6 (Synchronization Subnet Unit 1-6)
    S_SSUwrite ssu[NumberOfSSU];


    // the control bits of the UTCSU (write access)
    enum EConfiguration  // UTCCONF1/2 in ITU
    {
        // UTCCONF1
        CLEAPSECOND = 1<<0,  // do leap adjustment (0..no, 1..yes)
        CLEAPSINS   = 1<<1,  // 1..insert leap second; 0..delete leap sec.
        CSARINTS    = 1<<16,
        CSARINPU    = 1<<17,
        CSARSTRTS   = 1<<18,
        CSARSTRPU   = 1<<19,
        CSUMINTS    = 1<<20,
        CSUMINPU    = 1<<21,
        CSUMSTRTS   = 1<<22,
        CSUMSTRPU   = 1<<23,
        CPOLHWSNAP  = 1<<28,  // polarity (1..rising edge, 0..falling edge)
        CPOLSYNCRUN = 1<<29,

        // UTCCONF2
        CPOLAPP      = 1<<0,  // polarity (1/0 ... see above)
        CPOLAPP1     = 1<<1,
        CPOLAPP2     = 1<<2,
        CPOLAPP3     = 1<<3,
        CPOLAPP4     = 1<<4,
        CPOLAPP5     = 1<<5,
        CPOLAPP6     = 1<<6,
        CPOLAPP7     = 1<<7,
        CPOLAPP8     = 1<<8,
        CPOLGPS1PPS1 = 1<<12,
        CPOLGPS1PPS2 = 1<<13,
        CPOLGPS1PPS3 = 1<<14,
        CPOLGPSSTAT1 = 1<<15,
        CPOLGPSSTAT2 = 1<<16,
        CPOLGPSSTAT3 = 1<<17,
        CPOLRCV1     = 1<<20,
        CPOLRCV2     = 1<<21,
        CPOLRCV3     = 1<<22,
        CPOLRCV4     = 1<<23,
        CPOLRCV5     = 1<<24,
        CPOLRCV6     = 1<<25,
        CPOLXMT1     = 1<<26,
        CPOLXMT2     = 1<<27,
        CPOLXMT3     = 1<<28,
        CPOLXMT4     = 1<<29,
        CPOLXMT5     = 1<<30,
        CPOLXMT6     = 1<<31
    };
    enum EInterrupt  // UTCINTEN1/2 and UTCINTCLEAR1/2 in ITU
    {
        // UTCINTSTAT1
        ICAMS    = 1<<0,  // rising edge, INTT: continuous amort. phase started
        ICAMT    = 1<<1,  // falling edge, INTT: cont. amort. phase terminated
        INTPO    = 1<<2,  // pulse, INTT: NTPTIME overrun (yr. 2036, leap sec.)
        IBNDP    = 1<<4,  // rising, INTT
        IBNDN    = 1<<5,  // rising, INTT
        IAPPL    = 1<<6,  // rising, INTT
        IAPNL    = 1<<7,  // rising, INTT
        IANPL    = 1<<8,  // rising, INTT
        IANNL    = 1<<9,  // rising, INTT
        IFAILED  = 1<<10,  // perm. low, INTT
        IPUREPH  = 1<<12,  // pulse, INTA
        IHWSNAP  = 1<<16,  // pulse, INTA: hardware snapshot event
        ISWSNAP  = 1<<17,  // pulse, INTA
        ISYNCRUN = 1<<18,  // pulse, INTA
        ITSAPP   = 1<<20,  // pulse, INTA
        ITSAPP1  = 1<<21,  // pulse, INTA
        ITSAPP2  = 1<<22,  // pulse, INTA
        ITSAPP3  = 1<<23,  // pulse, INTA
        ITSAPP4  = 1<<24,  // pulse, INTA
        ITSAPP5  = 1<<25,  // pulse, INTA
        ITSAPP6  = 1<<26,  // pulse, INTA
        ITSAPP7  = 1<<27,  // pulse, INTA
        ITSAPP8  = 1<<28,  // pulse, INTA
        IAPPDUTY = 1<<29,  // pulse, INTA: INTPBUS >= DUTY and E=1

        // UTCINTSTAT2
        IGPS1PPS1 = 1<<0,  // pulse, INTN: 1 PPS from GPU1
        IGPS1PPS2 = 1<<1,  // pulse, INTN
        IGPS1PPS3 = 1<<2,  // pulse, INTN
        IGPSSTAT1 = 1<<3,  // pulse, INTN: status information from GPU1
        IGPSSTAT2 = 1<<4,  // pulse, INTN
        IGPSSTAT3 = 1<<5,  // pulse, INTN
        ITSRCV1   = 1<<8,  // pulse, INTN
        ITSRCV2   = 1<<9,  // pulse, INTN
        ITSRCV3   = 1<<10,  // pulse, INTN
        ITSRCV4   = 1<<11,  // pulse, INTN
        ITSRCV5   = 1<<12,  // pulse, INTN
        ITSRCV6   = 1<<13,  // pulse, INTN
        ITSXMT1   = 1<<14,  // pulse, INTN
        ITSXMT2   = 1<<15,  // pulse, INTN
        ITSXMT3   = 1<<16,  // pulse, INTN
        ITSXMT4   = 1<<17,  // pulse, INTN
        ITSXMT5   = 1<<18,  // pulse, INTN
        ITSXMT6   = 1<<19,  // pulse, INTN
        IDUTYA1   = 1<<20,  // pulse, INTT: duty timer A
        IDUTYA2   = 1<<21,  // pulse, INTT
        IDUTYA3   = 1<<22,  // pulse, INTT
        IDUTYA4   = 1<<23,  // pulse, INTT
        IDUTYA5   = 1<<24,  // pulse, INTT
        IDUTYA6   = 1<<25,  // pulse, INTT
        IDUTYB1   = 1<<26,  // pulse, INTT: duty timer B
        IDUTYB2   = 1<<27,  // pulse, INTT
        IDUTYB3   = 1<<28,  // pulse, INTT
        IDUTYB4   = 1<<29,  // pulse, INTT
        IDUTYB5   = 1<<30,  // pulse, INTT
        IDUTYB6   = 1<<31   // pulse, INTT
    };
};


#endif


