//
// errtask.cpp
//
// implementation of the error report class used for storing and processing
// error reports
// note: this implementation stores the message in a queue and periodically
// prints it to cerr
// note: the class can be used as a task in hardware (for printing to the
// terminal)
//
// created: 31.03.97
//
// last change: 11.03.98
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 10:49:31 $
//


#include <string.h>  // for strlen
#ifdef PSOS_SIM
  #include <stdlib.h>
  #if  __cplusplus
    extern "C" {
  #endif
    #include <util.h>  // for xp_strout()
  #if  __cplusplus
    }
  #endif
#endif

#include "psos.h"

#include "constant.hpp"
#include "errtask.hpp"


#ifndef PSOS_SIM
  #define xp_strout(x,message) \
	CErrorHandler::ProcessMessage (message)
  // xp_strout() is only known in pSOS (there, it prints to the XRAY
  // standard output screen); in all other cases, the error messages
  // is printed through the base class
#endif


//**************************************************************************
// class CErrorTask: stores strings (error messages) in a queue and
//      periodically prints them to cerr
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// create the queue, store its ID
CErrorTask::CErrorTask (char qname[4])
{
    q_create (qname, 0, Q_NOLIMIT | Q_LOCAL, &qid_);
}

// emergency flush and delete the queue
CErrorTask::~CErrorTask()
{
    EmergencyFlush();  // empty the queue
    q_delete (qid_);  // delete it
}


//--------------------------------------------------------------------------
//--- store a message
//--------------------------------------------------------------------------

// stores a copy of the message in the queue
void CErrorTask::NoteError (char *message)
{
    INT32 msg[4];

    if (message == NULL) return;  // nothing to store

    char *string = GetMessageString (strlen (message)+1);  // get memory
    if (string == NULL) return;  // cannot assert at this point
    // note: this error will be lost, but an Assert() might result in an
    // endless loop

    strcpy (string, message);  // copy data into string

    msg[0] = (INT32)string;  // store the pointer
    q_send (qid_, msg);  // put the message into the queue
}


//--------------------------------------------------------------------------
//--- message queue handling
//--------------------------------------------------------------------------

// reads at most 'MaxErrorMessages' messages from the message queue and
// processes them
void CErrorTask::Flush()
{
    INT32 msg[4];
    INT8 i=0;

    while ((i++<MaxErrorMessages) && (q_receive (qid_, Q_NOWAIT, 0, msg)==0))
    {
        ProcessMessage ((char *)msg[0]);  // print the message
        ReturnMessageString ((char *)msg[0]);  // return the memory
    }
}


//--------------------------------------------------------------------------
//--- message queue handling (emergency)
//--------------------------------------------------------------------------

// empties the queue, processes messages using EmergencyProcessMessage()
void CErrorTask::EmergencyFlush()
{
    INT32 msg[4];

    while (q_receive (qid_, 1, 0, msg) == 0)  // message exists
    {
        EmergencyProcessMessage ((char *)msg[0]);  // process the message
        ReturnMessageString ((char *)msg[0]);  // return the memory
    }
}


//--------------------------------------------------------------------------
//--- for the simulation scheduler only; do not use
//--------------------------------------------------------------------------

// periodically flush the queue
void CErrorTask::Body ()
{
    for (;;)
    {
        Flush();  // read some messages, process them
        tm_wkafter ((INT32)(ErrorReportPeriod*TICKS2SEC));
    }
}


//----------------------------------------------------------------------
//--- internal message output
//----------------------------------------------------------------------

// print the message (followed by a newline) to an output screen
// (in pSOS: the XRAY standard output screen; else: cerr)
void CErrorTask::ProcessMessage (char *message)
{
    static char newline[2] = "\n";
    xp_strout (0, message);
    xp_strout (0, newline);
}


//--------------------------------------------------------------------------
//--- internal message processing (emergency)
//--------------------------------------------------------------------------

// same as ProcessMessage()
void CErrorTask::EmergencyProcessMessage (char *message)
{
    ProcessMessage (message);
}


//--------------------------------------------------------------------------
//--- internal message buffer handling
//--------------------------------------------------------------------------

// allocates and returns a string of the given length; use only this
// function for obtaining memory for message strings
char *CErrorTask::GetMessageString (INT32 len)
{
    if (len == 0) return NULL;  // cannot assert
    return new char[len];
}

// deallocates the string; it is assumed that the string has been
// allocated using GetMessageString()
void CErrorTask::ReturnMessageString (char *string)
{
    if (string == NULL) return;  // again cannot assert
    delete[] string;
}

