//
// csyntest.cpp
//
// this is the file used to test the classes defined in the files as far
// as possible
// *) the tests were done using a debugger to verify the correct outcome
// *) all classes were compiled and debugged using Watcom 10.5 for Windows
//
// test the clock synchronization class in clocksyn.hpp
// note: the functions ReportEvent(), ComputeDelay(), ReplyToMessage()
//       and StoreTime() were made public for this test
//
// created: 14.11.96
//
// last change: 04.02.99
//


#include <iostream.h>  // for memcpy etc.
#include <string.h>  // for memcpy etc.

#include "..\algfetz.hpp"
#include "..\algfree.hpp"
#include "..\algfta.hpp"
#include "..\clockifc.hpp"
#include "..\clocksyn.hpp"
#include "..\constant.hpp"
#include "..\database.hpp"
#include "..\errbase.hpp"
#include "..\error.hpp"
#include "..\general.hpp"
#include "..\globals.hpp"
#include "..\interupt.hpp"
#include "..\lance.hpp"
#include "..\msgform.hpp"
#include "..\sim_glob.hpp"
#include "..\template.hpp"


#define Need_ACProcess
#define Need_CBitmask
#define Need_CClock
#define Need_CCrashingModule
#define Need_CSoftwareClock
#define Need_CSupervisor
#define Need_SMsg


const INT32 Infinity = 0xFFFFFFFF;
// infinite accuracy


// necessary for ACClockSyncAlgorithm
EBool rateAvailable = True;  // return value of GetLocalRate()
EBool setAssoc = True;  // return value of SetAssociation()

// control clock behaviour
INT32 clockAccuracy = 0;  // perfect clock
double clockTime = 0;  // timestamp of the clock, in seconds

// control behaviour of the clock interface
EBool pnode = False;  // return value of GpsConnected()


// necessary globals
CClockList listOfClocks;

INT32 DisableInterrupts()
  { cout << "DisableInterrupts() called" << endl; return 0; }
void EnableInterrupts (INT32)
  { cout << "EnableInterrupts() called" << endl; }


//--------------------------------------------------------------------------
// interrupt management functions

void CallISR (CNeedInterrupt *obj)
{
    obj->isr();  // call the isr of the object
}

void SetISR (int vector, CNeedInterrupt *obj, CallFunc caller)
{
    cout << "SetISR() called with vector = " << vector;
    cout << ", object = " << (long)obj;
    cout << ", callfunc = " << (long)caller << endl;
}
void RemoveISR (int vector, CNeedInterrupt *obj, CallFunc caller)
{
    cout << "RemoveISR() called with vector = " << vector;
    cout << ", object = " << (long)obj;
    cout << ", callfunc = " << (long)caller << endl;
}
void InvokeISR (int vector)
{
    cout << "InvokeISR() called with vector = " << vector << endl;
}


int main()
{
    //----------------------------------------------------------------------
    // test the behaviour of the class CClockSync
    //----------------------------------------------------------------------

    SSingleParameter singleParam;
    EBool result;
    int i;
    CClockSync *cs;
    CSupervisor sup ("QUEU", NULL, 1, 0, 25e6);
    CClockInterface *clock = new CClockInterface (&sup, 25e6, 0);
    CLance *lance = new CLance (&sup, 0);

    CErrorHandler *handler = new CErrorHandler;
    InstallErrorHandler (handler);
    InstallAlgorithms();


    //----------------------------------------------------------------------
    // test ctor and Init()

    cout << endl << "*** test ctor ***" << endl << endl;

    cs = new CClockSync (&sup, 2);
    // lance=clock = NULL, csaType=FreeRun,
    // area = 2, roundPeriod = DefaultRoundPeriod (100),
    // delayComp=1, computationDelayCompensation = 0,
    // delayCycles=delayCounter = 0, rateCycles=rateCounter =0,
    // suspended = True, initCalled = True, roundStage = 0,
    // resynchronizing = False, timeOfResync = -1, reportDelays = False
    // output: DisableInterrupts() called
    //         SetISR() called
    //         EnableInterrupts() called

    // tested on 21.09.98, 17:30: all tests okay


    //----------------------------------------------------------------------
    // test Init() and Dispose()

    cs->Init();
    // should return at once (trace into the code to verify that!)

    cs->Dispose();
    // initCalled = False
    cs->Dispose();
    // should return at once (trace into the code to verify that!)

    cs->Init();
    // initCalled = True

    // tested on 05.01.97, 12:00: all tests okay


    //----------------------------------------------------------------------
    // test ChangeAlgorithm()

    cout << endl << "*** test ChangeAlgorithm() ***" << endl << endl;

    result = cs->ChangeAlgorithm (EAlgorithmTypeCount);  // return False
    // should have no effect

    result = cs->ChangeAlgorithm (Midpoint);  // return True
    // csaType = Midpoint
    // output: CSA::operator=() called

    // tested on 07.08.97, 21:50: all tests okay


    //----------------------------------------------------------------------
    // test the functions for setting the objects

    cout << endl << "*** test SetClock() ***" << endl << endl;

    cs->SetClock (clock);
    // clock_ = clock
    // output: CSA::ClockChanged() called

    cs->SetLance (lance);
    // lance_ = lance

    // tested on 04.02.99, 12:10: all tests okay


    //----------------------------------------------------------------------
    // test ExecuteCommand() - continued below

    cout << endl << "*** test ExecuteCommand() ***" << endl << endl;

    SNetworkCommand ncom;
    ncom.type = ConnectNetwork;
    ncom.networkID = -2;
    result = cs->ExecuteCommand (ncom);  // return False

    ncom.networkID = 2;
    result = cs->ExecuteCommand (ncom);  // return True
    // output: CSA::ExecuteCommand called

    ncom.networkID = 5;
    result = cs->ExecuteCommand (ncom);  // return True
    // output: CSA::ExecuteCommand called

    ncom.type = DisconnectNetwork;
    ncom.networkID = 2;
    result = cs->ExecuteCommand (ncom);  // return True
    // output: CSA::ExecuteCommand called

    ncom.networkID = 5;
    result = cs->ExecuteCommand (ncom);  // return True
    // output: CSA::ExecuteCommand called

    // tested on 05.08.97, 10:40: all tests okay so far

    SCommand cmd;
    cmd.type = ReportTransmissions;
    result = cs->ExecuteCommand (cmd);  // return False
    // nothing should happen

    cmd.type = DontReportTransmissions;
    result = cs->ExecuteCommand (cmd);  // return True
    // reportDelays is (still) False

    // tested on 29.07.98, 16:00: all tests okay so far


    //----------------------------------------------------------------------
    // test Set/GetParameter()

    singleParam.type = ClockDrift;
    singleParam.value = 1;
    result = cs->SetParameter (singleParam);  // return False

    singleParam.type = DelayMeasurementCycles;
    singleParam.value = -1;
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = 11;
    result = cs->SetParameter (singleParam);  // return True
    // delayCycles = 11, delayCounter = 10

    singleParam.type = RateSynchronizationCycles;
    singleParam.value = -1;
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = 3;
    result = cs->SetParameter (singleParam);  // return True
    // rateCycles = 3, rateCounter = 2

    // tested on 14.05.97, 20:30: all tests okay so far

    singleParam.type = RoundPeriod;
    singleParam.value = 0;
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = 0.9;  // P < delay + compDelay
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = 50;
    result = cs->SetParameter (singleParam);  // return True
    // roundPeriod = 50

    singleParam.type = ComputationDelayCompensation;
    singleParam.value = -1;
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = 49.5;  // delay + compDelay > P
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = 4;
    result = cs->SetParameter (singleParam);  // return True
    // computationDelayComp_ = 4

    singleParam.type = DelayCompensation;
    singleParam.value = -1;
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = 47;  // delay + compDelay > P
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = 5;
    result = cs->SetParameter (singleParam);  // return True
    // delayComp_ = 5

    // tested on 05.08.97, 10:50: all tests okay so far


    //----------------------------------------------------------------------
    // test InterruptAtMult()

    cout << endl << "*** test InterruptAtMult() ***" << endl << endl;

    cs->InterruptAtMult();
    // timeOfResync = 59
    // output: DisableInterrupts() called
    //         InterruptAtMult called with 50
    //         EnableInterrupts() called

    // tested on 29.07.98, 16:00: all tests okay


    //----------------------------------------------------------------------
    // test Set/GetParameter() - continued

    cout << endl << "*** test Set/GetParameter() ***" << endl << endl;

    singleParam.type = RoundPeriod;
    singleParam.value = 60;
    result = cs->SetParameter (singleParam);  // return True
    // roundPeriod = 60, timeOfResync = 69 (1*60 + compensations),
    // output: InterruptAtMult (60)

    singleParam.type = DelayCompensation;
    singleParam.value = 2;
    result = cs->SetParameter (singleParam);  // return True
    // delayCompensation = 2, timeOfResync = 66

    cs->isr();  // just to increment the round stage; test is below

    singleParam.type = RoundPeriod;
    singleParam.value = 70;
    result = cs->SetParameter (singleParam);  // return True
    // roundPeriod = 70

    singleParam.type = DelayCompensation;
    singleParam.value = 5;
    result = cs->SetParameter (singleParam);  // return True
    // delayCompensation = 5

    singleParam.type = ComputationDelayCompensation;
    singleParam.value = 5;
    result = cs->SetParameter (singleParam);  // return True
    // compDelayCompensation = 5, timeOfResync = 67

    cs->isr();  // roundStage=2

    singleParam.type = RoundPeriod;
    singleParam.value = 80;
    result = cs->SetParameter (singleParam);  // return True
    // roundPeriod = 80

    singleParam.type = DelayCompensation;
    singleParam.value = 6;
    result = cs->SetParameter (singleParam);  // return True
    // delayCompensation = 6

    singleParam.type = ComputationDelayCompensation;
    singleParam.value = 2;
    result = cs->SetParameter (singleParam);  // return True
    // compDelayCompensation = 2

    // tested on 07.08.97, 22:30: all tests okay so far


    singleParam.type = SynchronizationAlgorithm;
    singleParam.value = EAlgorithmTypeCount;
    result = cs->SetParameter (singleParam);  // return False
    singleParam.value = Fetzer;
    result = cs->SetParameter (singleParam);  // return True
    // csaType = Fetzer
    // output: CSA::operator=() called


    singleParam.type = NumberOfNodes;
    singleParam.value = 3;
    result = cs->SetParameter (singleParam);  // return True
    // csa->#nodes = 3

    singleParam.type = NumberOfFaultyNodes;
    singleParam.value = 2;
    result = cs->SetParameter (singleParam);  // return True
    // csa->#faulty = 2

    singleParam.type = UserParameter; // none implemented at this time
    singleParam.value = 0;
    result = cs->SetParameter (singleParam);  // return False

    // tested on 05.08.97, 10:50: all tests okay so far


    singleParam.type = ClockDrift;
    result = cs->GetParameter (singleParam);  // return False

    singleParam.type = DelayMeasurementCycles;
    result = cs->GetParameter (singleParam);  // return True
    // value = 11

    singleParam.type = RateSynchronizationCycles;
    result = cs->GetParameter (singleParam);  // return True
    // value = 3

    // tested on 14.05.97, 20:30: all tests okay so far

    singleParam.type = RoundPeriod;
    result = cs->GetParameter (singleParam);  // return True
    // value = 80

    singleParam.type = DelayCompensation;
    result = cs->GetParameter (singleParam);  // return True
    // value = 6

    singleParam.type = ComputationDelayCompensation;
    result = cs->GetParameter (singleParam);  // return True
    // value = 2

    singleParam.type = SynchronizationAlgorithm;
    singleParam.value = EAlgorithmTypeCount;
    result = cs->GetParameter (singleParam);  // return True
    // value = 3 (Fetzer)


    singleParam.type = NumberOfNodes;
    result = cs->GetParameter (singleParam);  // return True
    // value = 3

    singleParam.type = NumberOfFaultyNodes;
    result = cs->GetParameter (singleParam);  // return True
    // value = 2

    singleParam.type = UserParameter;  // nothing implemented
    result = cs->GetParameter (singleParam);  // return False

    // tested on 05.08.97, 10:50: all tests okay

    singleParam.type = SynchronizationAlgorithm;
    singleParam.value = Midpoint;
    result = cs->SetParameter (singleParam);
    // cannot leave the type on 'Fetzer' because it depends
    // on correct initialization of the database; however,
    // ACClockSyncAlgorithm is only a dummy; so we have to use
    // Midpoint for any tests that cause ClockSync to call
    // StoreTime() (like the test of Announce()!)


    //----------------------------------------------------------------------
    // test Suspend()

    cout << endl << "*** test Suspend() ***" << endl << endl;

    cs->Suspend ();
    // roundStage=0, suspended_ = True, timeOfResync_ = -1 (NoAccount)
    // output: CClockIfc::DisableInterrupt() called
    // output: CSA::Suspend() called

    // tested on 17.12.97, 18:20: all tests okay


    //----------------------------------------------------------------------
    // test Resume()

    cout << endl << "*** test Resume() ***" << endl << endl;

    cs->Resume();
    // timeUntilResync_ = 88;
    // output: CClockIfc::EnableInterrupt() called
    // output: InterruptAtMult() called with 80

    // tested on 17.12.97, 18:20: all tests okay


    //----------------------------------------------------------------------
    // test BroadcastTime()

    cout << endl << "*** test BroadcastTime() ***" << endl << endl;

    rateAvailable = True;  // allow rate information

    cs->BroadcastTime();
    // suspended = False,
    // should return at once (no connected networks) [trace it!]

    ncom.type = ConnectNetwork;  // connect two networks
    ncom.networkID = 2;
    cs->ExecuteCommand (ncom);
    ncom.networkID = 3;
    cs->ExecuteCommand (ncom);
    // 2x output: CSA::ExecuteCommand() called

    cs->BroadcastTime();
    // delayCounter = 0, rateCounter = 0
    // output: CSA::GetLocalRate() called
    // 2x output: request buffer, (INT32)mode = 11, (SINT32)nodeID = 1

    singleParam.type = DelayMeasurementCycles;
    singleParam.value = 0;
    result = cs->SetParameter (singleParam);
    // delayCycles=0, delayCounter = (INT32)(-1)
    singleParam.type = RateSynchronizationCycles;
    result = cs->SetParameter (singleParam);
    // rateCycles=0, rateCounter = (INT32)(-1)

    cs->BroadcastTime();
    // delayCounter=rateCounter = -1 (unchanged)
    // 2x output: request buffer, (INT32)mode = 1, (SINT32)nodeID = 1

    // tested on 16.01.98, 17:00: all tests okay so far

    pnode = True;
    cs->BroadcastTime();
    // delayCounter=rateCounter = -1 (unchanged)
    // 2x output: request buffer, (INT32)mode = 17, (SINT32)nodeID = 1
    // for ADDITIONAL_PNODE_DATA times (compiler macro):
    //   2x output: request buffer, (INT32)mode = 17, (SINT32)nodeID = 1

    // tested on 29.07.98, 16:30: all tests okay so far

    pnode = False;
    singleParam.type = DelayMeasurementCycles;
    singleParam.value = 1;
    result = cs->SetParameter (singleParam);
    // delayCycles = 1
    singleParam.type = RateSynchronizationCycles;
    result = cs->SetParameter (singleParam);
    // rateCycles = 1

    cs->BroadcastTime();
    // delayCounter=rateCounter = 0, mode = Timestamp | DelayMeasurement |
    // RateSynchronization (11)
    // output: CSA::GetLocalRate() called
    // 2x output: request buffer, (INT32)mode = 11, (SINT32)nodeID = 1,
    //         (double)thetaPlus=0.1, (double)thetaMinus=0.2, (double)u=0

    rateAvailable = False;  // no rate information
    cs->BroadcastTime();
    // delayCounter=rateCounter = 0, mode = Timestamp | DelayMeasurement (3)
    // output: CSA::GetLocalRate() called
    // report: local rate not available
    // 2x output: request buffer, (INT32)mode = 3, (SINT32)nodeID = 1

    // tested on 16.01.98, 17:00: all tests okay so far


    // DEBUGGER: set resynchronize_ = True
    rateAvailable = True;  // rate information
    singleParam.type = DelayMeasurementCycles;
    singleParam.value = 2;
    result = cs->SetParameter (singleParam);
    // delayCycles = 2
    singleParam.type = RateSynchronizationCycles;
    singleParam.value = 3;
    result = cs->SetParameter (singleParam);
    // rateCycles = 3

    cs->BroadcastTime();
    // delayCounter=rateCounter = 0
    // output: CSA::GetLocalRate() called
    // 2x output: request buffer, (INT32)mode = 10, (SINT32)nodeID = 1,
    //         (double)thetaPlus=0.1, (double)thetaMinus=0.2, (double)u=0

    cs->BroadcastTime();
    // delayCounter=rateCounter = 1
    // output: none

    cs->BroadcastTime();
    // delayCounter=0, rateCounter = 2
    // 2x output: request buffer, (INT32)mode = 2, (SINT32)nodeID = 1

    cs->BroadcastTime();
    // delayCounter=1, rateCounter = 0
    // output: CSA::GetLocalRate() called
    // 2x output: request buffer, (INT32)mode = 8, (SINT32)nodeID = 1
    //         (double)thetaPlus=0.1, (double)thetaMinus=0.2, (double)u=0

    // DEBUGGER: set resynchronize_ = False

    // tested on 16.01.98, 17:00: all tests okay


    //----------------------------------------------------------------------
    // test ComputeDelay()
    // note: remember, the associated network is "2", the area is "2",
    //       cs is the node P, and the message comes from node Q

    cout << endl << "*** test ComputeDelay() ***" << endl << endl;

    renvx *renv = new renvx;
    STimeData nodePTime, nodeQTime;
    ti_acx usiTime;
    SReplyData replyData;
    INT32 dummy;

    singleParam.type = AssociatedNetwork;
    singleParam.value = 2;
    result = cs->SetParameter (singleParam);  // return True

    usiTime.acc = 0;
    SecondsToNtp (0, usiTime.ms, usiTime.ts);
    // initial message is sent from P at t=0 (seconds)
    nodePTime.ms = usiTime.ms;
    nodePTime.ts = usiTime.ts;
    nodePTime.acc = usiTime.acc;
    nodePTime.area = 2;

    SecondsToNtp (5, usiTime.ms, usiTime.ts);
    // the reply to the message is sent from Q to P at t=5
    nodeQTime.ms = usiTime.ms;
    nodeQTime.ts = usiTime.ts;
    nodeQTime.acc = usiTime.acc;
    nodeQTime.area = 0;

    replyData.msReceive = usiTime.ms;
    replyData.tsReceive = usiTime.ts;
    // the receive time on Q was t=5 as well
    replyData.nodeID = 2;
    replyData.thetaPlus = 0.3;
    replyData.thetaMinus = 0.4;


    for (i=0; i<ADDRCNT; i++) renv->sa[i] = 0;
    renv->sa[0] = 1;
    for (i=0; i<MAXASICS; i++)
        SecondsToNtp (10, renv->utcsu[i].rms, renv->utcsu[i].rts);
    // the receive time on P is t=10


    nodePTime.area = 1;  // wrong on purpose
    cs->ComputeDelay (renv, nodeQTime, replyData, nodePTime, 2);
    // should return at once (trace into the code!)

    nodePTime.area = 2;
    cs->ComputeDelay (renv, nodeQTime, replyData, nodePTime, 2);
    // output: CSA::StoreDelay() called with delayP/delayQ = 10/0,
    //         thetas = (0.4,0.3)

    // tested on 17.12.97, 18:20: all tests okay


    //----------------------------------------------------------------------
    // test ReplyToMessage()
    // note: cs is the node Q (with nodeID = 1) and replies to P

    cout << endl << "*** test ReplyToMessage() ***" << endl << endl;

    for (i=0; i<ADDRCNT; i++) renv->sa[i] = 0;
    renv->sa[0] = 1;
    for (i=0; i<MAXASICS; i++)
    {
        renv->utcsu[i].rms = 1;
        renv->utcsu[i].rts = 5;
    }
    nodePTime.area = 4;
    nodePTime.ms = 1; nodePTime.ts = 1; nodePTime.acc = 0;
    cs->ReplyToMessage (renv, nodePTime, 2);
    // output: CSA::GetLocalRate() called
    // output: request envelope, request buffer, call CLance::Send() with
    // (all but rate 4 bytes) [4(mode), ?(nodeID), 1(rms), 5(rts),
    // thetaMinus (8 bytes), thetaPlus (8 bytes), 4(area), 1(ms), 1(ts),
    // 0(acc)]  and address {1,0,..,0}

    // tested on 16.01.98, 17:00: all tests okay


    //----------------------------------------------------------------------
    // test StoreTime()
    // note: cs is the node P and got the time from Q
    // timeOfResync_ == 88

    cout << endl << "*** test StoreTime() ***" << endl << endl;

    INT32 dummy2;

    for (i=0; i<ADDRCNT; i++) renv->sa[i] = 0;
    renv->sa[0] = 1;
    SecondsToNtp (5, usiTime.ms, usiTime.ts);  // time = 5
    for (i=0; i<MAXASICS; i++)
        SecondsToNtp (10, renv->utcsu[i].rms, renv->utcsu[i].rts);
    SecondsToNtp (0.002, dummy, dummy2);  // to get the accuracy values
    usiTime.acc = (dummy2 & 0x01FFFE) >> 1;
    // mask out the relevant bits and shift them to get accPlus = 0.002 sec
    SecondsToNtp (0.001, dummy, dummy2);
    usiTime.acc += (dummy2 & 0x01FFFE) << 15;  // get accMinus = 0.001 sec
    nodeQTime.ms = usiTime.ms;
    nodeQTime.ts = usiTime.ts;
    nodeQTime.acc = usiTime.acc;
    nodeQTime.area = 0;

    clockTime = 3;  // timestamp of clock should be >= timestamp in renv

    cs->StoreTime (renv, nodeQTime, 2);
    // output: CSA::StoreTime() called with time until resync = 86
    // (timeOfResync(88) - renvTime(2))

    // tested on 07.08.97, 23:00: all tests okay so far

    clockTime = 256;  // timestamp of clock should be <= timestamp in renv

    for (i=0; i<MAXASICS; i++)
    {
        renv->utcsu[i].rms = 0;
        renv->utcsu[i].rts = 1<<24;  // 1 second
    }
    cs->StoreTime (renv, nodeQTime, 2);
    // output: CSA::StoreTime() called with time until resync = 87
    // (timeOfResync(88) - renvTime(1))

    // tested on 08.08.97, 12:00: all tests okay so far


    //----------------------------------------------------------------------
    // test Announce()
    // note: cs is the node P and got the message from Q

    cout << endl << "*** test Announce() ***" << endl << endl;

    char *message;  // message
    INT32 msgID = 5;

    cs->Announce (renv, nodeQTime, NULL, 0, 2, msgID);
    // report: original message was empty

    usiTime.acc = 0;
    SecondsToNtp (0, usiTime.ms, usiTime.ts);
    // initial message is sent from P at t=0 (seconds)
    nodePTime.ms = usiTime.ms;
    nodePTime.ts = usiTime.ts;
    nodePTime.acc = usiTime.acc;
    nodePTime.area = 2;

    SecondsToNtp (5, usiTime.ms, usiTime.ts);
    // the reply to the message is sent from Q to P at t=5
    nodeQTime.ms = 0xFFFFFFFF;  // CS bits should be masked out
    nodeQTime.ts = usiTime.ts;
    nodeQTime.acc = usiTime.acc;
    nodeQTime.area = 0;

    replyData.msReceive = 0x0FFFFFF;  // here, CS is already masked
    replyData.tsReceive = usiTime.ts;
    replyData.mode = 4; // DelayMeasurementReply
    replyData.nodeID = 2;
    // the receive time on Q was t=5 as well

    clockTime = 10;  // must be >= the timestamp in renv

    for (i=0; i<ADDRCNT; i++) renv->sa[i] = 0;
    renv->sa[0] = 1;
    for (i=0; i<MAXASICS; i++)
        SecondsToNtp (10, renv->utcsu[i].rms, renv->utcsu[i].rts);
    // the receive time on P is t=10

    INT16 size = sizeof (replyData)+sizeof (nodePTime);
    // correct the size

    message = new char[size];  // allocate the message buffer
    memcpy (message, &replyData, sizeof (replyData));
    memcpy (message+sizeof (replyData), &nodePTime, sizeof (nodePTime));

    // message contains a delay measurement reply
    cs->Announce (renv, nodeQTime, message, size, 2, msgID);
    // time.ms should be = 0x0FFFFFF (CS masked out) [trace it!]
    // should call ComputeDelay(), which should compute delayP=10, delayQ=0
    // output: CSA::StoreDelay() called with 10/0

    // message contains a delay measurement reply, but size is wrong
    cs->Announce (renv, nodeQTime, message, size-1, 2, msgID);
    // report: message size too small


    replyData.mode = 3;  // Timestamp | DelayMeasurement
    replyData.nodeID = 3;  // the time information should not be stored
    memcpy (message, &replyData, sizeof (SSendData));
    // only copy the relevant part (ie. the SSendData part of SReplyData)

    // message contains a delay measurement request and the timestamp
    dummy = sizeof (SSendData);  // need just the first part o.t. msg
    cs->Announce (renv, nodeQTime, message, dummy, 2, msgID);
    // should call ReplyToMessage() and StoreTime()
    // report: wrong mapping nodeID <-> network address (not with dummy-CSA)
    // output: CSA::GetLocalRate() called etc. (output of ReplyToMessage()),
    //         CSA::StoreTime() called (gps=0)

    // message contains a delay measurement request and the timestamp, but
    // it is not for the own area
    cs->Announce (renv, nodeQTime, message, dummy, 1, msgID);
    // output: CSA::StoreTime() called (gps=0)

    // message contains a delay measurement request and the timestamp,
    // but the size is wrong
    cs->Announce (renv, nodeQTime, message, dummy-1, 2, msgID);
    // report: message size too small

    delete[] message;  // free the message buffer

    SRate rate;
    rate.thetaPlus = 0.1;
    rate.thetaMinus = 0.2;
    rate.u = 3.4;

    size = sizeof(SSendData) + sizeof (SRate);
    message = new char[size];  // allocate the message buffer anew

    replyData.mode = 25;  // RateSynchronization, Timestamp, GpsTime
    memcpy (message, &replyData, sizeof (SSendData));
    // only copy the relevant part (ie. the SSendData part of SReplyData)
    memcpy (message+sizeof (SSendData), &rate, sizeof (SRate));

    // message contains rate information
    cs->Announce (renv, nodeQTime, message, size, 2, msgID);
    // report: wrong mapping nodeID <-> network address (not with dummy-CSA)
    // output: CSA::StoreRate() called
    // output: CSA::StoreTime() called (gps=1)

    cs->Announce (renv, nodeQTime, message, size-1, 2, msgID);
    // report: message size too small

    // tested on 16.01.98, 17:00: all tests okay so far


    // DEBUGGER: set reportDelays to True
    cs->Announce (renv, nodeQTime, message, size-1, 2, msgID);
    // should send event 'TransmissionDelay' [trace it!]
    // report: message size too small

    // tested on 29.07.98, 16:00: all tests okay


    //----------------------------------------------------------------------
    // test ApplyConvergenceFunction() and Resynchronize()

    cout << endl << "*** test ApplyCV() & Resynchronize() ***" << endl << endl;

    cs->ApplyConvergenceFunction();
    // output: CSA::CV called

    cs->Resynchronize();
    // output: CSA::SetNewClockValue() was called

    // tested on 02.08.97, 21:00: all tests okay


    //----------------------------------------------------------------------
    // test ExecuteCommand() - continued
    // note: calls Reset()

    SCommand command;
    command.type = UserCommand;  // nothing implemented
    result = cs->ExecuteCommand (command);  // return False

    cs->BroadcastTime();
    // suspended = False

    command.type = ResetModule;
    result = cs->ExecuteCommand (command);  // return True
    // networkID = -1, nodeID = 1, area = 2, timeOfResync_ = -1,
    // delayCycles = 0, cycleCounter = 0, suspended = True, roundStage = 0,
    // roundPeriod = DefaultRoundPeriod (100), csaType = FreeRun

    // tested on 21.09.98, 17:50: all tests okay


    //----------------------------------------------------------------------
    // test Announce() - continued
    // test StoreTime() - continued

    cs->Announce (renv, nodeQTime, message, size, 2, msgID);
    // report: msg in suspension state

    cs->BroadcastTime();
    // suspended = False

    cs->Announce (renv, nodeQTime, message, size, 2, msgID);
    // output: CSA::StoreRate() called
    // report: resynchronization time unknown (this is the output of
    // StoreTime(), so this function is tested as well)

    delete[] message;  // free the message buffer

    // tested on 06.01.98, 17:00: all tests okay


    //----------------------------------------------------------------------
    // test isr()
    // tested with FOR_GG defined and undefined in clocksyn.cpp; the final
    // version will be FOR_GG undefined!

    cout << endl << "*** test isr() ***" << endl << endl;

    clockTime = 10;  // just to make certain we know the time
    cs->InterruptAtMult();
    // timeOfResync = 101

    singleParam.type = DelayCompensation;
    singleParam.value = 0;
    cs->SetParameter (singleParam);
    // timeOfResync = 100
    singleParam.type = ComputationDelayCompensation;
    cs->SetParameter (singleParam);

    clockTime = 150;
    cs->isr();
    // roundStage = 0, timeOfResync = 200 (current value + next multiple)
    // should call BroadcastTime(),
    // ApplyConvergenceFunction() and Resynchronize()
    // output: ClearInterrupt() called, CFta::CV() called,
    #ifdef FOR_GG  // will become obsolete!
    // InterruptAtMult() called with 100, CSA::SetNewClockValue() called
    #else
    // CSA::PreloadNewClockValue() called, InterruptAtMult() called with 100
    #endif

    singleParam.type = DelayCompensation;
    singleParam.value = 4;
    cs->SetParameter (singleParam);
    // timeOfResync = 204
    singleParam.type = ComputationDelayCompensation;
    singleParam.value = 3;
    cs->SetParameter (singleParam);
    // timeOfResync = 207

    clockTime = 207;

    cs->isr();
    // roundStage = 1, should call BroadcastTime()
    // output: ClearInterrupt() called, InterruptAfter (4)

    cs->isr();
    // roundStage = 2, should call ApplyConvergenceFunction(),
    // and PreloadNewClockValue()
    // output: ClearInterrupt() called, InterruptAfter(3)
    #ifdef FOR_GG  // will become obsolete!
    // CFta::CV() called
    #else
    // CFta::CV() called, CSA::PreloadNewClockValue() called
    #endif

    cs->isr();
    // roundStage = 0, timeOfResync = 307
    // should just call InterruptAtMult()
    // output: ClearInterrupt() called, InterruptAtMult (100)
    #ifdef FOR_GG  // will become obsolete!
    // SetNewClockValue() called
    #else
    // nothing more
    #endif

    // tested on 06.01.98, 12:10: all tests okay


    //----------------------------------------------------------------------
    // test Suspend() - continued

    cs->isr();
    // roundStage = 1

    cs->Suspend();
    // roundStage = 0, suspended = True, timeOfResync = NoAccount

    cs->Resume();
    // timeOfResync = 307
    // output: InterruptAtMult (100)

    // tested on 08.08.97, 21:10: all tests okay


    //----------------------------------------------------------------------
    // test dtor()

    cout << endl << "*** test dtor ***" << endl << endl;

    delete cs;
    // output: CClockIfc::DisableInterrupt() called
    //         CClockIfc::DisableDutyTimer() called
    // output: DisableInterrupts() called
    //         RemoveISR() called
    //         EnableInterrupts() called

    // tested on 16.01.98, 17:00: all tests okay


    DeinstallErrorHandler();
    delete handler;

    //----------------------------------------------------------------------

    return 0;
}


//--------------------------------------------------------------------------
// ACClockSyncAlgorithm

ACClockSyncAlgorithm::ACClockSyncAlgorithm (CClockSync *) {}
ACClockSyncAlgorithm::~ACClockSyncAlgorithm () {}
ACClockSyncAlgorithm& ACClockSyncAlgorithm::operator=
    (const ACClockSyncAlgorithm&)
  {cout << "CSA::operator=() called" << endl; return *this; }
void ACClockSyncAlgorithm::Reset () {cout << "CSA::Reset() called" << endl;}
void ACClockSyncAlgorithm::Suspend () {cout << "CSA::Suspend() called" << endl;}
EBool ACClockSyncAlgorithm::SetParameter (const SParameter& param)
{
    switch (param.type)
    {
        case NumberOfNodes: numberOfNodes = ((SSingleParameter*)&param)->value;
            break;
        case NumberOfFaultyNodes: numberOfFaultyNodes = ((SSingleParameter*)&param)->value;
            break;
        case AssociatedNetwork:
        case Delta:
        case EpsilonPlus:
        case EpsilonMinus:
        case ThetaPlus:
        case ThetaMinus:
        case U: return True; break;
        default: return False;
    } // switch

    return True;
}
EBool ACClockSyncAlgorithm::GetParameter (SParameter& param) const
{
    switch (param.type)
    {
        case NumberOfNodes: ((SSingleParameter*)&param)->value = numberOfNodes;
            break;
        case NumberOfFaultyNodes: ((SSingleParameter*)&param)->value = numberOfFaultyNodes;
            break;
        case AssociatedNetwork:
        case Delta:
        case EpsilonPlus:
        case EpsilonMinus:
        case ThetaPlus:
        case ThetaMinus:
        case U: return True; break;
        default: return False;
    } // switch

    return True;
}
EBool ACClockSyncAlgorithm::ExecuteCommand (const SCommand&)
  { cout << "CSA::ExecuteCommand() called" << endl; return False; }
EBool ACClockSyncAlgorithm::StoreTime (const STimestamp&, double until,
    SINT32, const CAddr&, int, EBool gps)
  { cout << "CSA::StoreTime() called with time until resync = ";
    cout << until << ", gps = " << gps << endl; return True; }
EBool ACClockSyncAlgorithm::StoreDelay (double dmin, double dmax,
    const SRate& rate, SINT32, const CAddr&)
  { cout << "CSA::StoreDelay() called with delayP/delayQ = ";
    cout << dmin << "/" << dmax << ", thetas = " << rate.thetaMinus;
    cout << ", " << rate.thetaPlus << endl; return True; }
EBool ACClockSyncAlgorithm::StoreRate (const SRate&, const CNtpTime&,
    const CNtpTime&, SINT32, const CAddr&, int)
  { cout << "CSA::StoreRate() called" << endl; return False; }
EBool ACClockSyncAlgorithm::GetLocalRate (SRate& rate) const
  { cout << "CSA::GetLocalRate() called" << endl;
    rate.thetaPlus=0.1; rate.thetaMinus=0.2; rate.u=0; return rateAvailable; }
void ACClockSyncAlgorithm::ClockChanged ()
  {cout << "CSA::ClockChanged() called" << endl;}
EBool ACClockSyncAlgorithm::SetAssociation (SINT32, const CAddr&, SINT32)
  { cout << "CSA::SetAssociation() called" << endl; return setAssoc; }
EBool ACClockSyncAlgorithm::DeleteAssociation (SINT32, SINT32)
  { cout << "CSA::DeleteAssociation() called" << endl; return False; }
void ACClockSyncAlgorithm::SetNewClockValue ()
  { cout << "CSA::SetNewClockValue() was called" << endl; }
void ACClockSyncAlgorithm::PreloadNewClockValue (double)
  { cout << "CSA::PreloadNewClockValue() was called" << endl; }
EBool ACClockSyncAlgorithm::ComputeNewDelay (double, double, SINT32,
    const CAddr&) { return False; }
SINT32 ACClockSyncAlgorithm::GetAssociatedNetwork () const {return -1;}
EBool ACClockSyncAlgorithm::GetResynchronizationStatus () const {return False;}
void ACClockSyncAlgorithm::SetResynchronizationStatus (EBool) const {}
CClockInterface *ACClockSyncAlgorithm::GetClock () const { return NULL; }
void ACClockSyncAlgorithm::ReportEvent (SEvent *) {}
void ACClockSyncAlgorithm::ReportInterval (EBool, const STimeInterval&) {}
void ACClockSyncAlgorithm::ReportRate (EBool, const SRate&, double) {}
int ACClockSyncAlgorithm::ReportError (EReportType, char *, int,
    char *) { return 0; }
void ACClockSyncAlgorithm::Init () {}
void ACClockSyncAlgorithm::Dispose () {}
double ACClockSyncAlgorithm::GetRoundPeriod () const {return 100;}
double ACClockSyncAlgorithm::GetDelayCompensation() const {return 0;}
double ACClockSyncAlgorithm::GetComputationDelayCompensation() const {return 0;}
void ACClockSyncAlgorithm::GetOwnID (SINT32&, int&) const {}



//--------------------------------------------------------------------------
// templates, CFta, CMidpoint, CFreeRun

CClockTemplate *clockTemplate = NULL;
CGpsTemplate *gpsTemplate = NULL;

CFta::CFta (CClockSync *cs) : ACClockSyncAlgorithm (cs) {}
EBool CFta::ConvergenceFunction ()
  { cout << "CSA::ConvergenceFunction() called" << endl; return True; }
ACClockSyncAlgorithm *CFta::CreateFta (CClockSync *cs) { return new CFta(cs); }
ACClockSyncAlgorithm *CFreeRun::CreateFreeRun (CClockSync *cs) { return new CFta(cs); }


//--------------------------------------------------------------------------
// CControllingObject

CControllingObject::CControllingObject (char [4], ACSocketAddress *, SINT32) {}
CControllingObject::~CControllingObject () {}
void CControllingObject::Reset() {}
EBool CControllingObject::SetParameter (const SParameter&) { return False; }
EBool CControllingObject::GetParameter (SParameter&) const { return False; }
EBool CControllingObject::ExecuteCommand (const SCommand&) { return False; }
void CControllingObject::NoteEvent (SEvent *event)
{ cout << "received event with type " << event->type << endl; }
void CControllingObject::Body() {}
ACSocketAddress *CControllingObject::GetSocketAddress () { return 0; }
void CControllingObject::ReplyToMessage (double) {}
void CControllingObject::ReplyToMessage (EBool) {}
void CControllingObject::ReplyToMessage (SBase *) {}
SEvent *CControllingObject::GetEvent () { return 0; }
int CControllingObject::ReportError (EReportType, char *, int, char *) { return 0; }
void CControllingObject::EmptyQueue () {}
EBool CControllingObject::EventReportAllowed (EEvent) { return False; }
void CControllingObject::UpdateMask (CBitMask&, const SEventReport *) {}
void CControllingObject::GetBit (const CBitMask&, SEventReport *) const {}
EBool CControllingObject::InitEventTable () { return False; }
void CControllingObject::Resynchronize() {}
void CControllingObject::ProcessData (EMessageType) {}
void CControllingObject::TransmitEvent (SEvent *) {}
void CControllingObject::DoPeriodicActivities () {}


//--------------------------------------------------------------------------
// CClockInterface

CClockInterface::CClockInterface (CControllingObject *ctrl, double, int)
    : CApplControlledObject (ctrl) {}
CClockInterface::~CClockInterface () {}
EBool CClockInterface::SetParameter (const SParameter&) { return False; }
EBool CClockInterface::GetParameter (SParameter&) const { return False; }
EBool CClockInterface::ExecuteCommand (const SCommand&) {return False;}
CClock *CClockInterface::GetClock() {return NULL;}
double CClockInterface::GetFrequency () const {return 0;}
double CClockInterface::GetGranularityDT() const {return 0;}
double CClockInterface::GetGranularityOsc() const {return 0;}
double CClockInterface::GetGranularity() const {return 0;}
double CClockInterface::GetGranularityAcc() const {return 0;}
double CClockInterface::GetRhoPlus() const {return WcRhoPlus;}
double CClockInterface::GetRhoMinus() const {return WcRhoMinus;}
void CClockInterface::SetSteppure (INT32) {}
INT32 CClockInterface::GetSteppure () {return 0;}
void CClockInterface::SetSteplow (INT8) {}
void CClockInterface::SetLambdapure (INT16, INT16) {}
void CClockInterface::SetBounds (INT32) {}

EBool CClockInterface::SetAmortizationDuration (double) {return False;}
EBool CClockInterface::GpsConnected () {return pnode;}
INT32 CClockInterface::GetAccuracy() const {return clockAccuracy;}
EBool CClockInterface::GetTimestamp (STimestamp& time) {
  SecondsToNtp (clockTime, time.ms, time.ts);
  time.acc = clockAccuracy; return True; }
void CClockInterface::SetTimestamp (const SExtendedTimestamp& time) {
    SetTimestamp ((STimestamp)time); }
void CClockInterface::SetTimestamp (const STimestamp& time) {
  clockTime = NtpToSeconds (time.ms, time.ts);
  clockAccuracy = time.acc;
  cout << "CClockInterface::SetTimestamp() called with (ms,ts,acc) = (";
  cout << time.ms << "," << time.ts << "," << time.acc << ")";
  cout << " --> clock value = " << clockTime << endl; }
void CClockInterface::PreloadTimestamp (double at,
    const SExtendedTimestamp& time) {
  PreloadTimestamp (at, (STimestamp)time); }
void CClockInterface::PreloadTimestamp (double at, const STimestamp& time)
  { clockTime = NtpToSeconds (time.ms, time.ts);
  clockAccuracy = time.acc;
  cout << "CClockInterface::PreloadTimestamp() called with (ms,ts,acc) = (";
  cout << time.ms << "," << time.ts << "," << time.acc << ")";
  cout << " --> clock value = " << clockTime << ", for time " << at << endl; }
void CClockInterface::EnableInterrupt (EInterruptSource) {
  cout << "CClockInterface::EnableInterrupt() called" << endl; }
void CClockInterface::DisableInterrupt (EInterruptSource) {
  cout << "CClockInterface::DisableInterrupt() called" << endl; }
void CClockInterface::DisableDutyTimer (EInterruptSource) {
  cout << "CClockInterface::DisableDutyTimer() called" << endl; }
void CClockInterface::ClearInterrupt (EInterruptSource) {
    cout << "CClockInterface::ClearInterrupt() called" << endl; }
EBool CClockInterface::InterruptPending (EInterruptSource) {return True;}
EBool CClockInterface::InterruptPending (EClockInterrupt) {return True;}
void CClockInterface::ClearInterrupt (EClockInterrupt) {}
void CClockInterface::InterruptAfter (EInterruptSource, double after) {
  cout << "CClockInterface::InterruptAfter() called with " << after << endl; }
double CClockInterface::InterruptAtMult (EInterruptSource, double multTime) {
  cout << "CClockInterface::InterruptAtMult() called with " << multTime << endl;
  return ((int)(clockTime/multTime) + 1)*multTime; }
void CClockInterface::InterruptAt (EInterruptSource, double at) {
  cout << "CClockInterface::InterruptAt() called with " << at << endl; }
void CClockInterface::Init () {}
void CClockInterface::Dispose() {}
void CClockInterface::isr() {}
void CClockInterface::wobble_isr() {}
void CClockInterface::Amortize (double, INT32, INT32) {}
void CClockInterface::PreloadAmort (double, INT32, INT32) {}
void CClockInterface::Reset() {}
void CClockInterface::Crash() {}
void CClockInterface::Recover() {}
void CClockInterface::ReportCrashChange() {}
int CClockInterface::ReportError (EReportType, char *, int, char*)
  {return 0;}


//--------------------------------------------------------------------------
// CLance

CLance::CLance (CControllingObject *ctrl, int)
    : CApplControlledObject (ctrl) {}
CLance::~CLance() {}
EBool CLance::SetParameter (const SParameter&) {return False;}
EBool CLance::GetParameter (SParameter&) const {return False;}
EBool CLance::ExecuteCommand (const SCommand&) {return False;}
void CLance::Reset() {}
void CLance::Suspend() {}
void CLance::Crash() {}
void CLance::SetClockSync (CClockSync *) {}
void CLance::GetBuffer (INT16 size, char **msg, INT32& msgID) {
  *msg = new char[size]; msgID = 1;
  cout << "CLance::GetBuffer() returned " << (long)*msg << endl; }
void CLance::GetEnvelope (tenvx **tenv) {
  *tenv = new tenvx;
  cout << "CLance::GetEnvelope() returned " << (long)*tenv << endl; }
void CLance::ReturnBuffer (char *buf) {
  if (buf != NULL) delete[] buf;
  cout << "CLance::ReturnBuffer() called with " << (long)buf << endl; }
void CLance::ReturnEnvelope (tenvx *tenv) {
  if (tenv != NULL) delete tenv;
  cout << "CLance::ReturnEnvelope() called with " << (long)tenv << endl; }
void CLance::Send (int, tenvx *tenv, char *msg, INT16 size, int, int) {
  cout << "CLance::Send() was called with msg";
  if (msg == NULL) { cout << " = NULL" << endl; return;}
  cout << " value [";
  for (int i=0; i<size-1;i++) cout << (int)*(msg+i) << ",";
  cout << (int)*(msg+size-1) << "]";
  cout << ", send to address ";
  if (tenv == NULL) { cout << " = NULL" << endl; }
  else {
      cout << "{";
      for (i=0; i < ADDRCNT-1; i++) cout << (int)(tenv->da)[i] << ",";
      cout << (int)(tenv->da)[ADDRCNT-1] << "}";
  }
  if (tenv != NULL) delete tenv;
  if (msg != NULL) delete[] msg; cout << endl; }
void CLance::Announce (renvx *, bufx *, INT16, int) {}
void CLance::Init() {}
void CLance::Dispose() {}
void CLance::isr() {}
void CLance::ReportCrashChange() {}


#include "inlines.hpp"


