//
// RANDTST - TESTFILE FOR LIBRARY DISTRIB
//
// created: ??.??.98, Gerald Hummel
//
// last change: 25.05.98, Gerald Hummel
//
// last revision: 27.08.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/08/27 11:29:16 $
//

#include <malloc.h>
#include <math.h>
#include <stdio.h>

#include "distrib.hpp"


#define OUTPUT "ranvars.dat"
// the name of the output file


//--------------------------------------------------------------------------
// this is an example table of the exponential distribution
// for testing the generic function generator
//--------------------------------------------------------------------------

double *tab_ptr = NULL;  // pointer to the generic function table

// creates a table with the inversion function of exp()
void gen_tab()
{
  double *p,x;

  if (tab_ptr != NULL) free (tab_ptr);
  // discard previous data

  tab_ptr=(double *)malloc((2*(10*10*10+1))*sizeof(double));
  // allocate 101 entries, each consisting of a <x,y> pair

  p=tab_ptr;  // iterate through the table, starting with first element

  for(x=0; x<10; x+=0.01)  // x varies in [0,10), in steps of 0.01
  {
    *p = x;              // store the x component
    *(p+1) = 1-exp(-x);  // store the y component
    p+=2;  // go to next pair
  } // for
}


//--------------------------------------------------------------------------
// other globals
//--------------------------------------------------------------------------

void PrintHelp();
// print a short help text


//--------------------------------------------------------------------------
// main program
//
// let the user choose a function and enter parameters, then draw numbers
// from that random stream and store them in a file; the user should copy
// the file to some other name before selecting the next function
//
// the user can specify how many numbers are drawn, and in how many
// equivalence classes they are stored (the target interval is parted
// into several equivalence classes, and all numbers falling into one
// class are counted; with this method, you get a discreet approximation
// of the original function, which you can then plot with 'gnuplot')
//--------------------------------------------------------------------------

int main()
{
CRandom RCLASS;  /* Class of random number generators */ 

double lo,hi,stdev,E;

  FILE *f;  // the output file

  char fkt;  // the distribution function chosen for a run
  int anz_zahlen;  // how many random numbers are drawn in one run
  int divisions;  // number of equivalence classes
  double length;  // the length of one equivalence class

  double ranval;  // the random number obtained from the generator
  int *tst_field = NULL;  // the array with the equivalence classes

  double sum;  // the sum of all random numbers drawn in one rum
  double sum2;  // square sum of these random numbers

  int i;  // for iterations
  int *p;  // for iterating through tst_field


  // now execute runs until the user types 'x'
  for(;;)
  {
    printf("U>niform, E>xponential, N>ormal, G>amma, W>eibull,"
           " GeneriC>, H>elp, EX>it: ");
    scanf("\n%c",&fkt);  // get the user input

    if (fkt == 'x') goto exit;  // quit the program

    if (fkt == 'h')  // print a help text, then ask user again
    {
        PrintHelp();
        continue;
    } // if

    printf("lower limit: ");  // ask user for lower bound of interval
    scanf("%lf",&lo);

    printf("upper limit: ");  // ask user for upper bound of interval
    scanf("%lf",&hi);

    if (fkt != 'u')  // if the distr. is not uniform, we need the mean
    {
      printf("mean: ");
      scanf("%lf",&E);
    } // if

    switch(fkt)
    {
      case 'u':  // uniform distribution
        RCLASS.Set(Uniform,0,lo,hi,0,0);
        break;
      case 'e':  // exponential distribution
        RCLASS.Set(Exponential,0,lo,hi,E,0);
        break;
      case 'n':  // normal
        printf("stdev: ");  // we need to ask user for std. deviation
        scanf("%lf",&stdev);
        RCLASS.Set(Normal,0,lo,hi,E,stdev);
        break;
      case 'g':  // gamma
        RCLASS.Set(Gamma,0,lo,hi,E,0);
        break;
      case 'w':  // weibull
        RCLASS.Set(Weibull,0,lo,hi,E,0);
        break;
      case 'c':  // generic
        gen_tab();  // create the table
        RCLASS.SetGeneric(0,lo,hi,tab_ptr,(10*10*10));  
        break;
      default:
        break;
    } // switch

    printf("\nnumber of draws: ");  // ask how many numbers should be drawn
    scanf("%d",&anz_zahlen);

    printf("divisions/interval:");  // ask for number of equivalence classes
    scanf("%d",&divisions);

    length = (hi-lo)/divisions;  // length of one division
    tst_field = (int *)malloc((divisions)*sizeof(int));  // create target table
    if (tst_field==NULL) return 1;  // could not create table

    // initialize field with 0
    p=tst_field;
    for(i=0;i<divisions;i++) *p++=0;

    // now draw as many random numbers as requested
    p=tst_field;
    sum=0;  // set the sum of all numbers to 0
    sum2=0;  // set the square sum to 0
    for(i=0;i<anz_zahlen;i++)
    {
      switch(fkt)  // draw a random number
      {
        case 'u':ranval=RCLASS.RandomVar(Uniform);break;
        case 'e':ranval=RCLASS.RandomVar(Exponential,0,lo,hi,E,0);break;
        case 'n':ranval=RCLASS.RandomVar(Normal);break;
        case 'g':ranval=RCLASS.RandomVar(Gamma);break;
        case 'w':ranval=RCLASS.RandomVar(Weibull);break;
        case 'c':ranval=RCLASS.RandomVar(Generic);break;
      } // switch

      if((ranval>=lo)&&(ranval<hi))  // increment appropriate equiv. class
      {
        p[(int)((ranval-lo)/length)]++;
      }

      sum += ranval;     // update the sum of all drawn numbers
      sum2 += ranval*ranval;  // update the square sum
    } // for

    p=tst_field;
    ranval=0;

    // store the data for gnuplot
    f=fopen(OUTPUT,"w");
    for(i=0;i<divisions;i++)
    { 
      fprintf(f,"%lf %lf\n",(lo+(hi-lo)*i/(divisions-1)),(double)*p/anz_zahlen);
      p++;
    } // for
    fclose(f);

    // print the mean and std.dev. to stdout as control info for the user
    printf("\n\nArithm. Mean: %lf\n",sum/anz_zahlen);
    printf("Std. Deviation: %lf\n",sqrt(sum2/anz_zahlen-pow(sum/anz_zahlen,2)));

    free (tst_field);
    tst_field = NULL;
  } // for

exit:  // quit the program

  // free memory
  if (tst_field != NULL) free (tst_field);
  if (tab_ptr != NULL) free (tab_ptr);

  return 0;
}


//--------------------------------------------------------------------------
// other globals
//--------------------------------------------------------------------------

// print a short help text
void PrintHelp()
{
    const int DefaultLineLength = 65;  // default length of a line

    char helptext[] = "\nChoose one of the distribution functions by typing"
                      " the character before the '>' sign. The program"
                      " will ask you for the lower and upper bounds of"
        " the interval in which the random numbers should be drawn."
        " Depending on the function you have chosen, you may be prompted to"
        " enter a mean and standard deviation as well. At last, you have"
        " to specify how many random numbers the program should draw and"
        " in how many equivalence classes these numbers should be divided."
        " The program then draws random numbers, counts how many fall into"
        " each class, and at last prints this data to file ranvars.dat. The"
        " output contains of <x,y> pairs, where x is the center of the"
        " equivalence class and y is the cardinality of random numbers that"
        " are members of"
        " this class. The file is in gnuplot format, you can use the"
        " 'plot' command of gnuplot to draw the function.\n";

    char *p = helptext;  // pointer to current position in help text
    long pos;  // current position in the help text
    int linelen = DefaultLineLength;  // default length of a line

    // print the lines; try to make the newline at spaces
    for (pos = 0; pos < strlen (helptext); pos+=linelen, p+=linelen)
    {
        linelen = DefaultLineLength;  // set back to default value
        while ((*(p+linelen) != ' ') && (linelen < 80))  // print more
        {
            linelen++;
        }
        if (*p == ' ') { p++; pos++; linelen--; } // do not print leading space
        printf ("%.*s\n", linelen, p);
    } // for
}

