//
// account.cpp
//
// declaration of the accountant class that is used as a base class for
// the transmission data accountant of the network; it stores the network
// transmission delay data (and later on, also failure probabilities)
//
// created: 15.11.96
//
// last change: 23.06.98
//
// RCS version control information:
//     $Revision: 1.5 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <string.h>  // for memcmp etc.

#include "account.hpp"
#include "acctemp.hpp"
#include "error.hpp"


//**************************************************************************
// class CAccountant: this class is a database for the storage of network
//      transmission delays; the delays are associated with a pair of
//      keys <SINT32, SINT32>
//      key1 is the entry to the primary list CTransmissionList which
//      stores secondary lists CTransmissionDelayList with the index key2
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CAccountant::CAccountant ()
{
    initCalled_ = False;
    CAccountant::Init();  // initialize the members
}


CAccountant::~CAccountant ()
{
    CAccountant::Dispose();
}


//--------------------------------------------------------------------------
//--- reset all members to their initial values
//--------------------------------------------------------------------------

// deletes all entries, restores the initial state of the class
// it's like a delete with subsequent new
void CAccountant::Reset()
{
    Dispose();  // delete the dynamic members
    Init();  // set all values to their defaults again
}


//--------------------------------------------------------------------------
//--- access to the transmission delay data
//--------------------------------------------------------------------------

// the delay data is stored; if there was a former entry, its data is
// overwritten
void CAccountant::SetDelay (SNetworkDelay delay, SINT32 key1, SINT32 key2)
{
    SetupDataEntry (key1, key2);  // prepare the data_ structure

    *data_ = delay;  // copy the information
}


// if the entry exists, the delay data is written into delay and True is
// returned, else False is returned
// assert: delay != NULL
EBool CAccountant::GetDelay (SNetworkDelay *delay, SINT32 key1, SINT32 key2)
{
    Assert (delay != NULL);

    if (!GetInformation (key1, key2)) return False;  // no entry exists
    else
    {
        *delay = *data_;  // copy the information
        return True;
    }
}


//--------------------------------------------------------------------------
//--- deletion of entries
//--------------------------------------------------------------------------

// deletes all entries with key1; if at least one entry exists, True is
// returned, else False is returned
EBool CAccountant::DeleteKey1 (SINT32 key1)
{
    if (key1_ == key1) // the data is now invalid
    {
        ResetDataEntry ();
    }

    return listOfDelays->Delete (key1);
    // delete the entry from the primary list
}


// deletes all entries with key2; if at least one entry exists, True is
// returned, else False is returned
// if a primary list is empty after deletion of the entry, it is deleted
// as well
EBool CAccountant::DeleteKey2 (SINT32 key2)
{
    CTransmissionDelayList *list;
    SINT32 key1 = 0;
    SINT32 newKey = 0;
    EBool keyExisted = False;
    EBool done = False;

    if (key2_ == key2) // the data is now invalid
    {
        ResetDataEntry ();
    }

    // search all entries where key2 is the secondary index
    list = listOfDelays->GetLast();
    // note: the primary list is searched in reverse order (from end to
    // beginning); this is due to the fact that elements might have to be
    // deleted and SortedList sets the current pointer to the element
    // *following* the deleted one; so if a forward search is conducted,
    // one could not simply say "GetNext" after an element was deleted;
    // with backward search a "GetPrev" is okay in every case but where the
    // last element is deleted
    
    if (list == NULL) return keyExisted;
    // there are no entries

    do
    {
        if (listOfDelays->BeginOfList()) done = True;
        // first element of primary list is reached

        Assert (list != NULL);  // the secondary list should exist

#ifdef LINUX_SIM
        operator |= (keyExisted, list->Delete (key2));
        // delete the entry
        // note: the g++ does not use the operator |= if it is called
        // directly; but if is is called explicitly, then the operator is
        // used (tested by stepping into the function)
#else
        keyExisted |= list->Delete (key2);
        // note: WATCOM does not complain about this statement (and
        // hopefully executes it correctly), and in pSOS the type
        // EBool is defined to be 'int' anyway
#endif

        if (list->Empty())  // node was the last in this list
        {
            listOfDelays->GetCurrentObjectKey (key1); // get key1
            listOfDelays->Delete (key1);  // delete this list
            list = listOfDelays->GetCurrentObject();
            // get a pointer to the current element
        }

        if (!done)  // set ptr to the previous element
        {
            listOfDelays->GetCurrentObjectKey (newKey);

            if (newKey > key1)
            {
                // the current pointer is not on the new element
                list = listOfDelays->GetPrev();
                // get the previous secondary list

                key1 = 0;  // reset the index
            }
        } // if
    } while (!done);
    
    return keyExisted;
}


//--------------------------------------------------------------------------
//--- initialization and destruction of the data members
//--------------------------------------------------------------------------

// initializes all class members to default values; a second call
// (without Dispose() in between) has no effect
void CAccountant::Init ()
{
    if (initCalled_) return;  // Init() has already been called

    initCalled_ = True;

    ResetDataEntry ();  // no data is being accessed

    listOfDelays = new CTransmissionList;
    // make a list to store future data

    Assert (listOfDelays != NULL);  // paranoia dictates that
}


// deletes all dynamic members; a second call (without Init() in
// between) has no effect
void CAccountant::Dispose ()
{
    if (!initCalled_) return;  // Dispose() has already been called
    delete listOfDelays;
    ResetDataEntry ();
    initCalled_ = False;
}


//--------------------------------------------------------------------------
//--- internal management of the data entry (key1, key2, data)
//--------------------------------------------------------------------------

// resets the data entries to invalid values
void CAccountant::ResetDataEntry ()
{
    data_ = NULL;
}


//--------------------------------------------------------------------------
//--- internal management of transmission data
//--------------------------------------------------------------------------

// tries to get the data for the given pair of nodes; if the data is
// available, key1_, key2_ and data_ will be updated accordingly
// and the function will return True;
// if no entry was found, the method has no effect and False is returned
EBool CAccountant::GetInformation (SINT32 key1, SINT32 key2)
{
    CTransmissionDelayList *list;
    SNetworkDelay *data;

    if ((key1_ == key1) && (key2_ == key2) && (data_ != NULL)) return True;
    // the requested data is already loaded


    list = listOfDelays->Find (key1);
    // search the primary list for key1

    if (list == NULL) return False;    // no account for key1
    else
    {
        data = list->Find (key2);  // search for key2
        
        if (data == NULL) return False;  // no account for key2
        else
        {
            key1_ = key1;
            key2_ = key2;
            data_ = data;
            return True;
        }
    }
}


// take over the information stored in key1_, key2_ and data_
// and save it in the list
// take care: SetInformation() requires that no entry for the given pair
// of nodes exists
// assert: data_ != NULL
// assert: entry does not exist
void CAccountant::SetInformation ()
{
    CTransmissionDelayList *list;
    SNetworkDelay *data;
    
    Assert (data_ != NULL);

    list = listOfDelays->Find (key1_);
    // search the primary list for key1

    if (list != NULL)  // key1 found
    {
        data = list->Find (key2_);  // search for key2
        Assert (data == NULL);  // the entry must not exist
        list->Insert (data_, key2_); // store the new data
    }
    else
    {
        list = new CTransmissionDelayList;  // create a new secondary list
        list->Insert (data_, key2_);  // store the secondary key and data
        listOfDelays->Insert (list, key1_);
        // store the primary key and list
    }
}
    

// search for the transmission data of the pair of nodes; if no data exist,
// create a new data entry for the pair of nodes and insert it into
// the list; in any case store the data in key1_, key2_ and data_
void CAccountant::SetupDataEntry (SINT32 key1, SINT32 key2)
{
    if (!GetInformation (key1, key2))  // no account
    {
        data_ = new SNetworkDelay;
        key1_ = key1;
        key2_ = key2;

        SetInformation();  // store the data in the list
    }
}

