//
// addrman.cpp
//
// address manager to store triples <networkID,nodeID,nodeAddress>;
// searches can be conducted for <networkID,nodeID> (returning the
// nodeAddress) or for <networkID,nodeAddress> (returning the nodeID)
//
// created: 03.01.97
//
// last change: 01.08.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include "address.hpp"
#include "addrman.hpp"
#include "addrtemp.hpp"
#include "error.hpp"
#include "idtemp.hpp"


//**************************************************************************
// class CAddressManager: this class is used as a database by the delay
//      accountant to store pairs of nodeIDs (used for identification of
//      a node by the user) and node addresses (used by the network);
//      the pairs have to be unique in a given (sub-)network
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CAddressManager::CAddressManager ()
{
    initCalled_ = False;
    CAddressManager::Init();
}


CAddressManager::~CAddressManager ()
{
    CAddressManager::Dispose();
}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

void CAddressManager::Reset ()
{
    Dispose();
    Init();
}


//--------------------------------------------------------------------------
//--- access to the database
//--------------------------------------------------------------------------

// stores the <networkID,nodeID,nodeAddress> triple if possible;
// return values:
// NodeIDInUse:  the nodeID is already associated with another
//               nodeAddress (in the given network)
// AddressInUse: the nodeAddress is already associated with another
//               nodeID (in the given network)
// AssocStored:  nodeID and nodeAddress were not associated (in that
//               case, they are now associated with each other) or
//               they are already (with each other)
EResult CAddressManager::Insert (SINT32 networkID, SINT32 nodeID,
                                 const CAddr& nodeAddress)
{
    EResult result;

    result = CheckAddress (networkID, nodeID, nodeAddress);

    if (result == NoAssociation)  // make new entries in the lists
    {
        SINT32 *storedID = new SINT32 (nodeID);
        CAddr *storedAddress = new CAddr (nodeAddress);

        listOfAddresses_->Insert (storedAddress, networkID, nodeID);
        listOfNodes_->Insert (storedID, networkID, *storedAddress);
        return AssocStored;
    }
    else return result;
}


// returns True, if an associated network address exists; in that case,
// the address is copied into nodeAddress; if the assoc. does not exist,
// False is returned and nodeAddress is not modified
EBool CAddressManager::FindAddress (SINT32 networkID, SINT32 nodeID,
                                   CAddr& nodeAddress)
{
    CAddr *storedAddress = listOfAddresses_->Find (networkID,nodeID);
    if (storedAddress == NULL) return False;  // no association
    else
    {
        nodeAddress = *storedAddress;
        return True;
    }
}


// returns True, if an associated nodeID exists; in that case, the
// value is copied into nodeID; if the assoc. does not exist, False is
// returned and nodeID is not modified
// assert: nodeID != NULL
EBool CAddressManager::FindNodeID (SINT32 networkID, const CAddr& nodeAddress,
                                  SINT32 *nodeID)
{
    Assert (nodeID != NULL);

    SINT32 *storedID = listOfNodes_->Find (networkID, nodeAddress);
    if (storedID == NULL) return False;  // no association
    else
    {
        *nodeID = *storedID;
        return True;
    }
}


//--------------------------------------------------------------------------
//--- management of the database
//--------------------------------------------------------------------------

// if the nodeAddress is associated with a nodeID, the association is
// deleted and True is returned; else False is returned
EBool CAddressManager::DeleteAddress (SINT32 networkID, const CAddr& nodeAddress)
{
    SINT32 *storedID = listOfNodes_->Find (networkID, nodeAddress);
    if (storedID != NULL)
    {
        listOfAddresses_->Delete (networkID, *storedID);
        listOfNodes_->Delete (networkID, nodeAddress);
        // note: database integrity is not checked at this point; it is
        // assumed that Insert() keeps the database intact
        return True;
    }
    else return False;  // address did not exist
}

// if the nodeID is associated with a node address, the association is
// deleted and True is returned; else False is returned
EBool CAddressManager::DeleteNodeID (SINT32 networkID, SINT32 nodeID)
{
    CAddr *storedAddress = listOfAddresses_->Find (networkID, nodeID);
    if (storedAddress != NULL)
    {
        listOfNodes_->Delete (networkID, *storedAddress);
        // note: the nodeID *must* be deleted first, because if the address
        // were deleted first, the memory pointed to by storedAddress
        // would be invalid!
        listOfAddresses_->Delete (networkID, nodeID);
        // note: database integrity is not checked at this point; it is
        // assumed that Insert() keeps the database intact
        return True;
    }
    else return False;  // ID did not exist
}


//--------------------------------------------------------------------------
//--- initialization and destruction of the data members
//--------------------------------------------------------------------------

// initializes all class members to default values
void CAddressManager::Init()
{
    if (initCalled_) return;  // Init() has already been called
    initCalled_ = True;

    listOfAddresses_ = new CAddressList;
    listOfNodes_ = new CNodeList;

    Assert (listOfAddresses_ != NULL);
    Assert (listOfNodes_ != NULL);
}


// deletes all dynamic members
void CAddressManager::Dispose()
{
    if (!initCalled_) return;  // Dispose() has already been called
    delete listOfAddresses_;
    delete listOfNodes_;
    initCalled_ = False;
}


//--------------------------------------------------------------------------
//--- management of the address list
//--------------------------------------------------------------------------

// searches for <networkID,nodeID> in the address list and for <networkID,
// nodeAddress> in the nodeID list
// return values:
// NoAssociation: both entries do not exist
// AssocStored:   both entries exist and correlate (ie. the stored values
//                are equal to the appropriate parameter values)
// NodeIDInUse:   the first entry (key nodeID) exists, the second does not
// AddressInUse:  the second entry exists, the first does not
// assertions are used to guarantee the integrity of the two lists
EResult CAddressManager::CheckAddress (SINT32 networkID, SINT32 nodeID,
                                       const CAddr& nodeAddress)
{
    CAddr *storedAddress = listOfAddresses_->Find (networkID, nodeID);
    SINT32 *storedID = listOfNodes_->Find (networkID, nodeAddress);

    if ((storedAddress == NULL) && (storedID == NULL))
        return NoAssociation;  // the address and ID are not in the lists

    if ((storedAddress != NULL) && (storedID == NULL))
        return NodeIDInUse;  // nodeID already associated

    if ((storedID != NULL) && (storedAddress == NULL))
        return NodeAddressInUse;  // address already associated

    if ((nodeAddress == *storedAddress) && (nodeID == *storedID))
        return AssocStored;  // there is an association

    Assert ((nodeAddress != *storedAddress) && (nodeID != *storedID));
    // if one list stores the pair, but the other does not, Insert() has
    // done something wrong

    return BothInUse;  // both address and ID exist, but no assoc.
}


