//
// algbase.cpp
//
// implementation of the (abstract) base class of the clock synchronization
// algorithm
//
// created: 12.02.97
//
// last change: 22.01.99
//
// RCS version control information:
//     $Revision: 1.12 $$State: Exp $$Date: 1999/01/22 10:59:41 $
//

#include "address.hpp"
#include "algbase.hpp"
#include "clockifc.hpp"
#include "clocksyn.hpp"
#include "constant.hpp"
#include "database.hpp"
#include "error.hpp"
#include "idlist.hpp"
#include "structs.hpp"


//**************************************************************************
// module global functions (needed for ComputeNewDelay)
//**************************************************************************

// returns the minimum of two double values 
static inline double Min (double a, double b)
{
    if (a < b) return a;
    else return b;
}

// returns the maximum of two double values 
static inline double Max (double a, double b)
{
    if (a > b) return a;
    else return b;
}


//**************************************************************************
// class ACClockSyncAlgorithm: abstract base class for implementing the
//      currently used clock synchronization algorithm; it can store data
//      of other nodes (time intervals, transmission delays, rate info)
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assert: cs != NULL
ACClockSyncAlgorithm::ACClockSyncAlgorithm (CClockSync *cs)
{
    Assert (cs != NULL);
    cs_ = cs;
    rate = new SRate;
    Assert (rate != NULL);

    ClockChanged();  // set the rate to default values

    newClockValue = new SExtendedTimestamp;
    Assert (newClockValue != NULL);

    initCalled_ = False;
    ACClockSyncAlgorithm::Init();
    // initialize the members that are modified during reset
}

ACClockSyncAlgorithm::~ACClockSyncAlgorithm ()
{
    ACClockSyncAlgorithm::Dispose();
    delete rate;
    delete newClockValue;
}


//--------------------------------------------------------------------------
//--- operators
//--------------------------------------------------------------------------

// copies all interesting things (#nodes, #faulty, assocNetwork, cs)
ACClockSyncAlgorithm& ACClockSyncAlgorithm::operator= 
    (const ACClockSyncAlgorithm& alg)
{
    if (this == &alg) return *this;  // no need to copy from the own object

    numberOfNodes = alg.numberOfNodes;
    numberOfFaultyNodes = alg.numberOfFaultyNodes;
    assocNetwork_ = alg.assocNetwork_;

    return *this;
}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

// discard all data
void ACClockSyncAlgorithm::Reset ()
{
    Dispose();  // delete everything
    Init();  // and create it anew
}


//--------------------------------------------------------------------------
//--- suspend the synchronization
//--------------------------------------------------------------------------

// discard the collected interval data
void ACClockSyncAlgorithm::Suspend ()
{
    delete database;
    database = new CApplicationDatabase;
    Assert (database != NULL);
    timeValid = False;
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
// existing parameters: NumberOfNodes, NumberOfFaultyNodes, Delta,
// EpsilonPlus, EpsilonMinus, ThetaPlus, ThetaMinus, U, AssociatedNetwork
EBool ACClockSyncAlgorithm::SetParameter (const SParameter& param)
{
    SSingleParameter *p = (SSingleParameter *)&param;
    // easier access in the code below

    switch (param.type)
    {
      case NumberOfNodes:
          numberOfNodes = (INT32)p->value;
          break;
      case NumberOfFaultyNodes:
          numberOfFaultyNodes = (INT32)p->value;
          break;
      case AssociatedNetwork:
          if (!cs_->listOfConnectedNetworks_->Find ((SINT32)p->value))
              return False;
          // associated network must be in the list of connected networks
          assocNetwork_ = (SINT32)p->value;
          break;
      case ThetaPlus:
          if (p->value < 0) return False;
          rate->thetaPlus = p->value;
          break;
      case ThetaMinus:
          if ((p->value < 0) || (p->value >= 1)) return False;
          rate->thetaMinus = p->value;
          break;
      case U:
          if (p->value < 0) return False;
          rate->u = p->value;
          break;
      case Delta:
      case EpsilonPlus:
      case EpsilonMinus:
          return database->SetParameter (param);
          break;

      default:  // unknown parameter type
          return False;
    } // switch

    return True;
}

// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned
EBool ACClockSyncAlgorithm::GetParameter (SParameter& param) const
{
    SSingleParameter *p = (SSingleParameter *)&param;
    // easier access in the code below

    switch (param.type)
    {
      case NumberOfNodes:
          p->value = numberOfNodes;
          break;
      case NumberOfFaultyNodes:
          p->value = numberOfFaultyNodes;
          break;
      case AssociatedNetwork:
          p->value = assocNetwork_;
          break;
      case ThetaPlus:
          p->value = rate->thetaPlus;
          break;
      case ThetaMinus:
          p->value = rate->thetaMinus;
          break;
      case U:
          p->value = rate->u;
          break;
      case Delta:
      case EpsilonPlus:
      case EpsilonMinus:
          return database->GetParameter (param);
          break;

      default:  // unknown parameter type
          return False;
    } // switch

    return True;
}


//----------------------------------------------------------------------
//--- execution of system commands
//----------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False
// existing commands: ConnectNetwork, DisconnectNetwork
EBool ACClockSyncAlgorithm::ExecuteCommand (const SCommand& command)
{
    SNetworkCommand *netc = (SNetworkCommand*)&command;  // easier access

    switch (command.type)
    {
        case ConnectNetwork:  // maybe set the associated network
          if (assocNetwork_ == NoAccount)  // set the assoc. network
              return SetAssociatedNetwork (netc->networkID);
          break;
        case DisconnectNetwork:  // maybe change the associated network
          if (netc->networkID==assocNetwork_)  // change the assoc. network
              return SetAssociatedNetwork (cs_->listOfConnectedNetworks_->GetFirst());
          break;

        default:
          return False;
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- add time intervals
//--------------------------------------------------------------------------

// computes the current time interval from the timestamp and puts it
// into the database; returns True if successful, False if an entry was
// already there; 'resync' is the (local) time until the next
// resynchronization
EBool ACClockSyncAlgorithm::StoreTime (const STimestamp& time, double resync,
    SINT32 networkID, const CAddr& netAddress, int area, EBool)
{
    CClockInterface *clock = GetClock();  // get the clock
    Assert (clock != NULL);  // clock must exist

    SDelay delay;

    if (!database->GetDelay (delay, networkID, netAddress))
        return False;
    // no delay data stored

    STimeInterval interval;  // the resulting interval
    if (database->GetTime (interval, networkID, netAddress, area))
        return False;
    // entry exists

    CNtpTime timestamp (time.ms, time.ts);
    CNtpTime accPlus (0, (time.acc & 0x0000FFFF) << 1);
    CNtpTime accMinus (0, ((time.acc) >> 16 & 0x0000FFFF) << 1);
    // acc contains both positive and negative accuracy (bit -8..-23);
    // accuracy starts from bit -23, timestamp from -24 on
    // note: "<< 1" will never produce an overflow, because negative values
    // are set to zero (--> the msb is always zero)

    interval.refpoint = timestamp;  // store the reference point
    interval.low = timestamp - accMinus;  // compute the accuracy int.
    interval.high = timestamp + accPlus;

    // do delay compensation --> add wc. transmission delay
    interval.refpoint += CNtpTime (delay.delta);  // add deterministic part
    interval.low  += CNtpTime (delay.delta - delay.epsilonMinus);
    interval.high += CNtpTime (delay.delta + delay.epsilonPlus);

    // compensate the time until the next resynchronization (global time)
    interval.refpoint += CNtpTime (resync); // again, add deterministic part
    interval.low += CNtpTime (resync*(1-rate->thetaMinus));
    interval.high += CNtpTime (resync*(1+rate->thetaPlus));

    // add uncertainties caused by granularities
    interval.low -= CNtpTime (clock->GetGranularity() +
                              clock->GetGranularityAcc() +
                              clock->GetGranularityOsc());
    interval.high += CNtpTime (clock->GetGranularityAcc() +
                               clock->GetGranularityOsc());
 
    database->SetTime (interval, networkID, netAddress, area); // store data
    ReportInterval (True, interval);  // report input interval to EvalSys
    return True;
}


//--------------------------------------------------------------------------
//--- add transmission delay intervals
//--------------------------------------------------------------------------

// process the delays (do drift compensation for both), compute new
// transmission delay data and store it in the database;
// the function returns False if the upper bound for the transmission delay
// is negative; otherwise it calls ComputeDelay() and returns its value
EBool ACClockSyncAlgorithm::StoreDelay (double localDelay, double remoteDelay,
    const SRate& data, SINT32 networkID, const CAddr& netAddress)
{
    double dmin, dmax;
    // the upper and lower bound of the transmission delay

    // localDelay-remoteDelay is the transmission delay, but in the time of
    // the local/remote clock!; we want the real (global) delay, so the
    // clock rate uncertainties have to be considered as well

    dmin = (localDelay*(1-rate->thetaMinus) - remoteDelay*(1+data.thetaPlus))/2;
    // complete lower bound of the transmission time from local to remote

    dmax = (localDelay*(1+rate->thetaPlus) - remoteDelay*(1-data.thetaMinus))/2;
    // complete upper bound of the transmission time from local to remote

    if (dmax < 0) return False;
    // note: the condition is needed, because faulty clocks/messages
    // might induce "negative" delays, which have to be discarded

    if (dmin < 0)
    // uncertainties might result in a negative lower bound, but this cannot
    // be true, so set it to zero and adjust dmax so that the computed delta
    // is still in the middle of the interval
    // note: tests have revealed that setting delta = 60e-6, epsPlus=epsMinus=0
    // and lambda=300ms (much higher than delta; kmschoss has computed the bound
    // to be delta <= lambda*theta) results in a negative dmin;
    // the original solution was to simply set dmin=0 if it is negative; but
    // this makes a symmetrical interval [dmin,dmax] asymmetrical, so delta
    // is estimated to be higher than it really is; therefore, if we are setting
    // dmin=0, then we should adjust dmax as well
    // but is the adjustment of dmax justified? well, it is only done if dmin<0,
    // so when is dmin<0? I believe that dmin is only <0 if we have a situation
    // like described above, a fast network but slow access time; since in this
    // case dmax is highly overestimating the w.c. transmission delay anyway,
    // we can safely reduce it and stil have a valid bound on the transmission
    // delay of this message
    // note: this may be wrong if the thetas are not symmetrical; if there are
    // problems or if you get transmission delay estimations which appear to be
    // wrong, check if the thetas are equal; if they are not, then maybe dmax
    // is reduced more than it should be
    {
        dmax = dmax+dmin;  // adding the negative value makes dmax smaller
        dmin = 0;  // dmin is set to zero
        if (dmax < 0) Report (adjusting dmin has made dmax negative);
        // if you get this report, then the current solution for handling a
        // negative dmin is not correct; you should try to find a better
        // solution
    } // if

    if (dmin > dmax) return False;
    // if this happens, then either remoteDelay > localDelay (obviously wrong),
    // or the rates are negative (is that possible?), or we have made an error
    // when adjusting dmax (in this case, the report for dmax<0 should
    // have been printed)

    return ComputeNewDelay (dmin, dmax, networkID, netAddress);
    // take over the new delays and compute new transmission data
}


//--------------------------------------------------------------------------
//--- manage rate synchronization data
//--------------------------------------------------------------------------

// Since no rate algorithm is implemented, the rate information of the remote
// node is not needed and thus ignored. The function always returns True.
// overload the function in the derived class and implement the rate
// algorithm!
EBool ACClockSyncAlgorithm::StoreRate (const SRate&, const CNtpTime&,
    const CNtpTime&, SINT32, const CAddr&, int)
{
    return True;
}


// if ClockSync::clock != NULL, the current rate information of the
// local clock (stored in member rate) is copied into 'data' (the data
// is broadcast to the other nodes) and True is returned; if clock=NULL,
// False is returned
EBool ACClockSyncAlgorithm::GetLocalRate (SRate& data) const
{
    if (GetClock() == NULL) return False;
    data = *rate;
    return True;
}


//--------------------------------------------------------------------------
//--- manage clock changes
//--------------------------------------------------------------------------

// sets clock dependent data to default values
// the local rate is set to default values (wc. bounds of the clock, or global
// wc. bounds if GetClock() returns NULL)
// derived classes should overload the function to implement their
// own clock dependent initializations
// call this function in ClockSync after the clock pointer is set to a
// new clock
void ACClockSyncAlgorithm::ClockChanged()
{
    CClockInterface *clock = GetClock();

    if (clock != NULL)
    {
        rate->thetaMinus = clock->GetRhoMinus();  // get the w.c. bounds
        rate->thetaPlus = clock->GetRhoPlus();
    }
    else
    {
        rate->thetaMinus = WcRhoMinus;  // default w.c. lower bound
        rate->thetaPlus = WcRhoPlus;  // default w.c. upper bound
    }
    rate->u = 0;
}


//--------------------------------------------------------------------------
//--- manage the association nodeID <--> network address
//--------------------------------------------------------------------------

// stores the association <networkID, address, nodeID>; if an association
// already exists, the new data is ignored; overload the function in the
// derived class to obtain different behaviour; the function returns True
// if the triple could be stored, else False
EBool ACClockSyncAlgorithm::SetAssociation (SINT32 networkID,
    const CAddr& address, SINT32 nodeID)
{
    return database->SetAssociation (networkID, address, nodeID);
}

// deletes the nodeAddress associated with the <networkID,nodeID> key
// from the list of addresses; if it was not found, False is returned,
// else True is returned
EBool ACClockSyncAlgorithm::DeleteAssociation (SINT32 networkID, SINT32 nodeID)
{
    return database->DeleteAssociation (networkID, nodeID);
}


//--------------------------------------------------------------------------
//--- compute and set the new clock interval
//--------------------------------------------------------------------------

// if 'timeValid' is True, the function sets the clock to the new value
// stored in 'newClockValue'; if it is False, the clock value is not
// changed; in any case, 'timeValid' is set to False
// assert: ClockSync::clock != NULL
void ACClockSyncAlgorithm::SetNewClockValue()
{
    if (timeValid)
    {
        CClockInterface *clock = GetClock();  // get the clock
        Assert (clock != NULL);

        clock->SetTimestamp (*newClockValue);  // set it to the new value

        timeValid = False;  // invalidate the clock value member
    }
}

void ACClockSyncAlgorithm::PreloadNewClockValue (double loadAt)
{
    if (timeValid)
    {
        CClockInterface *clock = GetClock();  // get the clock
        Assert (clock != NULL);

        clock->PreloadTimestamp (loadAt, *newClockValue);
        // preload the new value

        timeValid = False;  // invalidate the clock value member
    }
}


//--------------------------------------------------------------------------
//--- compute the new transmission delay data (interval)
//--------------------------------------------------------------------------

// the function is called by StoreDelay() to add new transmission delays
// (given by the bounds (dmin, dmax)) to the database;
// the new delta is the mean of any stored delta and the mean of dmin,
// dmax; epsilonPlus, epsilonMinus are chosen so that the resulting
// interval encompasses both any old interval stored and (dmin, dmax);
// the function always returns True
// overload the function in the derived class to implement a different
// strategy for the computation of the delay data
EBool ACClockSyncAlgorithm::ComputeNewDelay (double dmin, double dmax,
    SINT32 networkID, const CAddr& netAddress)
{
    // now store the delay
    SDelay newDelay;  // new transmission delay for the delay accountant
    SDelay measured;  // measured delay (delta is the mean of dmin, dmax)

    measured.delta = (dmin + dmax)/2;  // compute the transmission data
    measured.epsilonPlus = dmax - measured.delta;
    measured.epsilonMinus = measured.epsilonPlus;
    // note: if you have set eps=0 in the network and the measured delta is
    // nevertheless not equal to the real delta, then the reason may be that
    // the local or the remote clock (or both) have a non-zero drift; this
    // makes the [dmin,dmax] interval "asymmetric" (i.e., not centered around
    // the real delta)

    if (database->GetDelay (newDelay, networkID, netAddress))
    // old entry for that area exists
    {
        double newMean;
        newMean = (measured.delta + newDelay.delta)/2;

        // set epsilon minus
        newDelay.epsilonMinus = newMean -
            Min (newDelay.delta - newDelay.epsilonMinus, dmin);

        // set epsilon plus
        newDelay.epsilonPlus = 
            Max (dmax, newDelay.delta + newDelay.epsilonPlus) - newMean;

        // set delta
        newDelay.delta = newMean;
    }
    else  // this is a new entry
    {
        newDelay = measured;  // simply use the new data
    }

    // store the new delay
    database->SetDelay (newDelay, networkID, netAddress);

    SINT32 nodeID;
    if (database->GetAssociation (networkID, netAddress, &nodeID))
    // the entry for this <networkID, netAddress> should already be in
    // the database
    {
        SApplDelayEstimationEvent *event;
        event = new SApplDelayEstimationEvent;  // notify EvalSys
        event->module = ClockSync;
        event->type = DelayEstimation;
        event->networkID = networkID;
        event->sender = nodeID;
        event->delay = newDelay;
        cs_->ReportEvent (event);
    } // if
    else
        Report (cannot send transmission delay estimation - unknown address);

    return True;
}


//--------------------------------------------------------------------------
//--- internal handling of the CSA identification
//--------------------------------------------------------------------------

// returns the nodeID and area of ClockSync
void ACClockSyncAlgorithm::GetOwnID (SINT32& nodeID, int& area) const
{
    nodeID = cs_->GetNodeID();
    area = cs_->area_;
}


//--------------------------------------------------------------------------
//--- internal handling of the associated network
//--------------------------------------------------------------------------

// returns the ID of the associated network of ClockSync, or NoAccount
// if no associated network exists
SINT32 ACClockSyncAlgorithm::GetAssociatedNetwork () const
{
    return assocNetwork_;
}

// if 'networkID' is in the list of connected networks of ClockSync,
// the function sets the associated network to 'networkID' and returns
// True; otherwise, False is returned and the function has no effect
// the change is reported to EvalSys with the event 'ChangeAssociation'
EBool ACClockSyncAlgorithm::SetAssociatedNetwork (SINT32 networkID)
{
    if ((networkID != NoAccount) &&
        !cs_->listOfConnectedNetworks_->Find (networkID)) return False;

    assocNetwork_ = networkID;  // change the associated network

    SApplNetEvent *event = new SApplNetEvent;  // notify EvalSys
    event->module = ClockSync;
    event->type = ChangeAssociation;
    event->networkID = networkID;
    cs_->ReportEvent (event);

    return True;
}


//--------------------------------------------------------------------------
//--- internal handling of the resynchronization mode of CS
//--------------------------------------------------------------------------

// returns the value of ClockSync::resynchronizing
EBool ACClockSyncAlgorithm::GetResynchronizationStatus() const
{
    return cs_->resynchronizing_;
}

// sets the value of ClockSync::resynchronizing to 'status'
void ACClockSyncAlgorithm::SetResynchronizationStatus (EBool status) const
{
    cs_->resynchronizing_ = status;
}


//--------------------------------------------------------------------------
//--- internal handling of the clock
//--------------------------------------------------------------------------

// returns the clock of ClockSync
CClockInterface *ACClockSyncAlgorithm::GetClock () const
{
    return cs_->clock_;
}


//--------------------------------------------------------------------------
//--- internal handling of synchronization round duration data
//--------------------------------------------------------------------------

// returns the RoundPeriod of ClockSync
double ACClockSyncAlgorithm::GetRoundPeriod () const
{
    return cs_->roundPeriod_;
}

// returns the rate period of ClockSync
// returns 0 if there is no rate synchronization
double ACClockSyncAlgorithm::GetRatePeriod () const
{
    return cs_->roundPeriod_*cs_->rateCycles_;
}

// returns the delay compensation duration of ClockSync
double ACClockSyncAlgorithm::GetDelayCompensation() const
{
    return cs_->delayCompensation_;
}

// returns the computation delay compensation duration of ClockSync
double ACClockSyncAlgorithm::GetComputationDelayCompensation() const
{
    return cs_->computationDelayCompensation_;
}


//--------------------------------------------------------------------------
//--- internal event reports
//--------------------------------------------------------------------------

// pass the event to Clocksync (set module=ClockSync, type=UserEvent)
// assert: event != NULL
void ACClockSyncAlgorithm::ReportEvent (SEvent *event)
{
    Assert (event != NULL);
    event->module = ClockSync;
    event->type = UserEvent;
    cs_->ReportEvent (event);
}

// creates an event of type SApplIntervalEvent, copies 'interval' into it,
// sets the event module to 'ClockSync', the type to 'IntervalEvent' and
// its member 'input' to the parameter 'input' (which should be True if the
// interval is used as input to the CV, False if it is its result)
void ACClockSyncAlgorithm::ReportInterval (EBool input,
    const STimeInterval& interval)
{
    SApplTimeEvent *event = new SApplTimeEvent;
    Assert (event != NULL);

    event->type = IntervalEvent;
    event->module = ClockSync;
    event->input = input;
    event->interval = interval;
    cs_->ReportEvent (event);
}

// creates an event of type SApplRateEvent, copies 'rate' and 'refPoint'
// into it, sets the module to 'ClockSync', the type to 'RateEvent' and
// its member 'input' to the parameter 'input' (which should be True if
// the rate is used as input to the rate algorithm, False if it is its result)
void ACClockSyncAlgorithm::ReportRate (EBool input, const SRate& rate,
    double refPoint)
{
    SApplRateEvent *event = new SApplRateEvent;
    Assert (event != NULL);

    event->type = RateEvent;
    event->module = ClockSync;
    event->input = input;
    event->rate = rate;
    event->refPoint = refPoint;
    cs_->ReportEvent (event);
}



//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int ACClockSyncAlgorithm::ReportError (EReportType type, char *file,
                                       int line, char *expression)
{
    return cs_->ReportError (type, file, line, expression);
}


//--------------------------------------------------------------------------
//--- create/destroy the members that are modified during reset
//--------------------------------------------------------------------------

// creates the members that are modified during a reset; a second call
// (without Dispose() in between) has no effect; sets #nodes=1, #faulty=0
void ACClockSyncAlgorithm::Init ()
{
    if (initCalled_) return;  // Init() has already been called
    initCalled_ = True;
    numberOfNodes = 1;  // at least the own node must exist
    numberOfFaultyNodes = 0;  // no nodes may exhibit faults
    database = new CApplicationDatabase;
    Assert (database != NULL);
    timeValid = False;
    assocNetwork_ = NoAccount;
}

// deletes the dynamic members created by Init(); a second call
// (without Init() in between), has no effect
void ACClockSyncAlgorithm::Dispose ()
{
    if (!initCalled_) return;  // Dispose() has already been called
    delete database;
    initCalled_ = False;
}


