//
// algfetz.cpp
//
// uses FTA (reference point), but handles GPS nodes according to Fetzer's
// algorithm
//
// BUG: the current implementation does not check the networkID (in a final
// implementation, the P-Node msges should be kept in a list and only the
// data of the associated network should be used in the convergence function)
//
// BUG: numberOfFaultyNodes is not considered for P-Nodes in the convergence
// function
//
// BUG: the convergence functions calls that of CFta to compute the result
// for the S-Nodes, and this function uses numberOfNodes to determine if
// it has enough data for computation; but since P-Nodes do not enter in
// this computation, the function needs an additional parameter that tells
// how many P-Nodes exist
// --> algorithm needs two more params (#PNodes, #faultyPNodes)
// currently, we do not have faulty nodes, so the code works
//
// BUG: the convergence function for P-Nodes does not consider #faultyNodes
//
// BUG: the computation of the parameter M (of the Fetzer algorithm) just
// uses WcRhoPlus instead of the maximum of WcRhoMinus and WcRhoPlus
//
// NOTE: to avoid sorting the external times, the average over all values
// was computed; the origial Fetzer Algorithm just uses the highest and
// lowest values (after removing #faultyNodes values from both ends)
//
// created: 15.04.98
//
// last change: 15.06.98
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <math.h>  // for fabs()

#include "algfetz.hpp"
#include "clockifc.hpp"
#include "constant.hpp"
#include "database.hpp"
#include "time.hpp"


static const INT32 NumberOfPNodes = 10;
// must be greater or equal to the number of P-Node messages expected in
// one round



//**************************************************************************
// class CFtaWithFetzer: uses CFta as base class, but stores P-Node data
//      separatly and only merges the two after FTA has been computed
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assert: cs != NULL
CFtaWithFetzer::CFtaWithFetzer (CClockSync *cs) : CFta (cs)
{
    initCalled_ = False;
    CFtaWithFetzer::Init();
}

CFtaWithFetzer::~CFtaWithFetzer ()
{
    CFtaWithFetzer::Dispose();
}


//--------------------------------------------------------------------------
//--- suspend the synchronization
//--------------------------------------------------------------------------

// discard the collected interval data
void CFtaWithFetzer::Suspend ()
{
    CFta::Suspend();  // remove data from the S-Nodes

    delete [] gpsdata;  // remove data from the P-Nodes
    gpsdata = new STimeInterval[NumberOfPNodes];
    Assert (gpsdata != NULL);
    gpsdataIndex = 0;

    timeValid = False;  // mark an existing result interval as invalid
}


//--------------------------------------------------------------------------
//--- add time intervals
//--------------------------------------------------------------------------

// if 'gps' is True, then the time interval of this node is kept
// in a separate list; if 'gps' is False, the function of the base class
// is called to handle the time information
// note: if the array for P-Node messages is full, further messages are
// rejected
EBool CFtaWithFetzer::StoreTime (const STimestamp& time, double resync,
    SINT32 networkID, const CAddr& netAddress, int area, EBool gps)
{
    if (gps == False)
        return CFta::StoreTime (time,resync,networkID,netAddress,area,gps);
    // let the base class handle it

    // note: the following code is taken over from ACClockSyncAlgorithm;
    // the only change is that the database is not checked for an existing
    // entry (therefore, multiple msges from the same P-Node are possible),
    // and the resulting interval is entered into the 'gpsdata' array

    if (gpsdataIndex == NumberOfPNodes) return False;
    // the array for keeping P-Node messages is full, reject further msges

    CClockInterface *clock = GetClock();  // get the clock
    Assert (clock != NULL);  // clock must exist

    SDelay delay;

    if (!database->GetDelay (delay, networkID, netAddress))
        return False;
    // no delay data stored

    STimeInterval interval;  // the resulting interval

    CNtpTime timestamp (time.ms, time.ts);
    CNtpTime accPlus (0, (time.acc & 0x0000FFFF) << 1);
    CNtpTime accMinus (0, ((time.acc) >> 16 & 0x0000FFFF) << 1);
    // acc contains both positive and negative accuracy (bit -8..-23);
    // accuracy starts from bit -23, timestamp from -24 on
    // note: "<< 1" will never produce an overflow, because negative values
    // are set to zero (--> the msb is always zero)

    interval.refpoint = timestamp;  // store the reference point
    interval.low = timestamp - accMinus;  // compute the accuracy int.
    interval.high = timestamp + accPlus;

    // do delay compensation --> add wc. transmission delay
    interval.refpoint += CNtpTime (delay.delta);  // add deterministic part
    interval.low  += CNtpTime (delay.delta - delay.epsilonMinus);
    interval.high += CNtpTime (delay.delta + delay.epsilonPlus);

    // compensate the time until the next resynchronization (global time)
    interval.refpoint += CNtpTime (resync); // again, add deterministic part
    interval.low += CNtpTime (resync*(1-rate->thetaMinus));
    interval.high += CNtpTime (resync*(1+rate->thetaPlus));

    // add uncertainties caused by granularities
    interval.low -= CNtpTime (clock->GetGranularity() +
                              clock->GetGranularityAcc() +
                              clock->GetGranularityOsc());
    interval.high += CNtpTime (clock->GetGranularityAcc() +
                               clock->GetGranularityOsc());

    gpsdata[gpsdataIndex] = interval;  // store the interval
    gpsdataIndex++;  // set index to next free entry

    ReportInterval (True, interval);  // report input interval to EvalSys
    return True;
}


//--------------------------------------------------------------------------
//--- compute and set the new clock interval
//--------------------------------------------------------------------------

// returns True if a new interval could be computed (depends on the
// number of nodes in the system and the maximum number of faulty
// nodes); in that case 'newTime' is set to the new clock value; if
// no valid interval could be computed, False is returned and 'newTime'
// is not changed
EBool CFtaWithFetzer::ConvergenceFunction ()
{
    if (!CFta::ConvergenceFunction ())
    {
        // computation of the S-Nodes did not work -> forget it
        gpsdataIndex = 0;  // throw away the node data
        return False;
    }

    // now merge with the data from the P-Nodes
    if (gpsdataIndex == 0) return True;  // no data from P-Nodes

    double externalTime=0;  // the time value of the P-Nodes
    double internalTime;  // the time value of the S-Nodes
    double difference;  // the difference between internal and external time
    double M = GetRoundPeriod()*WcRhoPlus;  // compute the parameter M
    int i;

    if (M==0) goto finished;  // use only internal synchronization

    internalTime = NtpToSeconds (*newClockValue);  // get internal time

    // compute average of all P-Node data -> external time
    for (i=0; i<gpsdataIndex; i++)
    {
        externalTime += NtpToSeconds (gpsdata[i].refpoint);
    } // for
    externalTime = externalTime/gpsdataIndex;

    // merge the two times
    difference = externalTime - internalTime;  // get the time difference

    if (fabs(difference) < M) SecondsToNtp (externalTime, *newClockValue);
    // the difference is small enough to set directly to the external time

    else if (difference < 0) SecondsToNtp (internalTime-M, *newClockValue);
    else SecondsToNtp (internalTime+M, *newClockValue);
    // just adapt the internal time into the direction of the external time

finished:
    gpsdataIndex = 0;  // throw away the P-node data
    return True;
}


//--------------------------------------------------------------------------
//--- create/destroy the members that are modified during reset
//--------------------------------------------------------------------------

// creates the members that are modified during a reset; a second call
// (without Dispose() in between) has no effect
void CFtaWithFetzer::Init ()
{
    if (initCalled_) return;  // Init() has already been called
    CFta::Init();
    gpsdata = new STimeInterval[NumberOfPNodes];
    Assert (gpsdata != NULL);
    gpsdataIndex = 0;
    initCalled_ = True;
}

// deletes the dynamic members created by Init(); a second call
// (without Init() in between), has no effect
void CFtaWithFetzer::Dispose ()
{
    if (!initCalled_) return;  // Dispose() has already been called
    delete [] gpsdata;
    CFta::Dispose();
    initCalled_ = False;
}


//--------------------------------------------------------------------------
//--- create function of class CFtaWithFetzer
//--------------------------------------------------------------------------

// creates and returns a new object of type CFtaWithFetzer
ACClockSyncAlgorithm *CFtaWithFetzer::CreateFtaWithFetzer (CClockSync *cs)
{
    return new CFtaWithFetzer (cs);
}



