//
// algfta.cpp
//
// implementation of the FTA (fault tolerant average) clock synchronization
// algorithm
//
// created: 12.02.97
//
// last change: 16.09.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Exp $$Date: 1998/09/16 16:26:41 $
//


#include <string.h>

#include "algfta.hpp"
#include "clockifc.hpp"
#include "constant.hpp"
#include "database.hpp"
#include "error.hpp"
#include "time.hpp"


const INT32 Infinity = 0xFFFFFFFF;
/**/// infinite accuracy; all bits=1, okay? - No!

const INT32 HighestAccuracyBit = 16;
// highest bit in timestamp that is covered by the accuracy, as seen from
// the lsb of timestamp (lsb = -24, highest bit = -8)

const INT32 HighestAccuracyValue = (1 << HighestAccuracyBit+1)-1;
// the highest value in timestamp that is covered by the accuracy


//**************************************************************************
// module global functions
//**************************************************************************

static SExtendedTimestamp Mean (const CNtpTime& low, const CNtpTime& high);
// compute the arithmetic mean of two timestamps, set the accuracy to the
// end of the interval




//**************************************************************************
// class CFta: provides the FTA algorithm
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assert: cs != NULL (done by the ctor of ACClockSyncAlgorithm)
CFta::CFta (CClockSync *cs) : ACClockSyncAlgorithm (cs)
{
    midpoint_ = False;  // CFta uses the reference point
}


//--------------------------------------------------------------------------
//--- compute and set the new clock interval
//--------------------------------------------------------------------------

// returns True if a new interval could be computed (depends on the
// number of nodes in the system and the maximum number of faulty
// nodes); in that case 'newTime' is set to the new clock value; if
// no valid interval could be computed, False is returned and 'newTime'
// is not changed
EBool CFta::ConvergenceFunction ()
{
    STimeInterval interval;  // the new time interval
    timeValid = False;  // mark the time as invalid

    if (GetAssociatedNetwork() == NoAccount) return False;
    // no associated network

    if (!FTA (interval, GetAssociatedNetwork())) return False;
/**/// right now, only the associated network is used

    // compute the arithmetic mean
    *newClockValue = Mean (interval.low, interval.high);
    ReportInterval (False, interval);// report resulting interval to EvalSys

    timeValid = True;  // the new timestamp is valid

    return True;
}


//--------------------------------------------------------------------------
//--- create function of class CFta
//--------------------------------------------------------------------------

// creates and returns a new object of type CFta
ACClockSyncAlgorithm *CFta::CreateFta (CClockSync *cs)
{
    return new CFta (cs);
}


//--------------------------------------------------------------------------
//--- FTA algorithm
//--------------------------------------------------------------------------

// computes the new clock interval for the network; returns True if the
// computation was successful, False otherwise
EBool CFta::FTA (STimeInterval& interval, SINT32 networkID)
{
    STimeInterval data;
    SExtendedTimestamp time;
    double *tsArray;
    INT32 size;  // array size
    double ts;  // the timestamp that is inserted into tsArray
    int free;  // first free element in tsArray
    int up, low, curr;  // for binary search
    EBool entryExists;  // exit condition of the for-loop below

    if (numberOfNodes <= 3*numberOfFaultyNodes) // n > 3f is needed for FTA
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        // delete all data of the network
        return False;
    }

    size = database->Cardinality (CApplicationDatabase::Time, networkID);
    if (size <= 2*numberOfFaultyNodes)
    // need at least 2*f + 1 timestamps, because 2*f are thrown away
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }

    tsArray = new double[size];
    // need to sort the timestamps

    free = 0;
    low = up = curr = 0;

    for (entryExists = database->GetFirstTime (data, networkID);
         entryExists;
         entryExists = database->GetNextTime (data, networkID))
    {
        if (midpoint_) time = Mean (data.low, data.high);  // get the mean
        else  // use the reference point
        {
            time.ms = data.refpoint.ms;
            time.ts = data.refpoint.ts;
            time.us = data.refpoint.us;
            time.acc = 0;
        }
        ts = NtpToSeconds (time); // convert to seconds

        if (free != 0)  // find the entry
        {
            up = (free>0) ? free-1 : 0;  // prepare for binary search
            low = 0;
            curr = (up+low)/2;

            while ((ts != tsArray[curr]) && (low < up))
            {
                if (ts < tsArray[curr]) up = (curr>0) ? curr-1 : 0;
                else if (ts > tsArray[curr]) low = curr+1;
    
                curr = (low+up)/2;
            }

            if (ts > tsArray[curr]) curr++;
            // didn't find the entry, need to insert behind the current one
        } // if

        Assert (curr < (SINT32)size);
        Assert (curr >= 0);

        memmove (tsArray+curr+1, tsArray+curr,  // move everything
                 (size-curr-1)*sizeof(double));

        tsArray[curr] = ts;  // enter the timestamp

        free++;  // increment the free index
    } // for

    // okay, all timestamps are in the sorted list, now just use FTA
    ts = 0;
    curr = 0;
    for (INT32 i=numberOfFaultyNodes; i < size-numberOfFaultyNodes; i++)
    {
        ts = ts + tsArray[i];  // accumulate the timestamps
        curr++;  // count the number of timestamps
        // note: no check for overflow, after all it's just for testing
    }

    Assert (curr != 0);  // should always be the case
    ts = ts/curr;  // take the mean value

    delete[] tsArray;
    database->DeleteAll (CApplicationDatabase::Time, networkID);
    // the data is not needed anymore

    SecondsToNtp (ts, interval.low.ms, interval.low.ts, interval.low.us);
    interval.high = interval.low;
    interval.refpoint = interval.low;

    return True;
}



//**************************************************************************
// module global functions
//**************************************************************************

// compute the arithmetic mean of two timestamps, set the accuracy to the
// end of the interval
// note: the computation is a bit confusing, because in order to avoid
// overflows I first divide by two and then add; but because the 64 bit NTP
// time is broken into two 32 bit parts, I have to take care of the msb of
// ts and the lsb of ms myself, which results in the code below;
// easier ways exist?
static SExtendedTimestamp Mean (const CNtpTime& low, const CNtpTime& high)
{
    Assert ((high.ms>low.ms) || ((high.ms==low.ms) && (high.ts>=low.ts)));
    // make sure that the upper bound is >= the lower bound

    SExtendedTimestamp result;

    CNtpTime addition = low + high;  // just add the time values

    result.us = (addition.us >> 1) & 0x7FFFFFFF;
    // divide by two (shift right)

    result.ts = (addition.ts >> 1) & 0x7FFFFFFF;
    if (addition.ts & 1) result.us += (1<<31);  // shift into us

    result.ms = (addition.ms >> 1) & 0x7FFFFFFF;
    if (addition.ms & 1) result.ts += (1<<31);  // shift into ts


    // now compute the accuracy
    CNtpTime diff = high - CNtpTime (result.ms, result.ts, result.us);
    // that's already the accuracy, but the format is still wrong (need
    // bits (-8,-23) for the accuracy, timestamp contains bits (7,-24))

    // see if there's an overflow
    if ((diff.ms != 0) || (diff.ts >= HighestAccuracyValue))
    {
        result.acc = Infinity;  // accuracy overflow
    }
    else result.acc = (diff.ts>>1 & 0x0FFFF) + (diff.ts<<15 & 0xFFFF0000);
    // the accuracies are symmetric;

    return result;
}


