//
// algfti.cpp
//
// implementation of the FTI "clock synchronization algorithm"
//
// created: 03.07.97, Klaus Schossmaier
//
// last change: 03.07.97, Klaus Schossmaier
//
// last revision: 04.02.99
//
// RCS version control information:
//     $Revision: 1.13 $$State: Exp $$Date: 1999/02/04 14:17:44 $
//

#include <fstream.h>
#include <string.h>

#include "address.hpp"
#include "algfti.hpp"
#include "alggen.hpp"
#include "clockifc.hpp"
#include "constant.hpp"
#include "database.hpp"
#include "error.hpp"
#include "time.hpp"

// enable output files
//#define GNU_ACC_FTI

extern char *ftoa (double f, int width, char *string);  // from ftoa.cpp
// returns 'f' as a string with at least 'width' characters in %e format;
// stores the string in 'string', which must be big enough to hold the
// converted string; if 'string' is NULL, NULL is returned


//**************************************************************************
// module global functions
//**************************************************************************
 
static EBool FTIntersection (SAsymInterval* input, SAsymInterval& output,
                             INT32 size, int f);


//**************************************************************************
// class CFti: provides the FTI algorithm
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assert: cs != NULL (done by the ctor of ACClockSyncAlgorithm)
CFti::CFti (CClockSync *cs) : ACClockSyncAlgorithm (cs)
{
    ClockChanged();
    // initialize the clock data
}


//--------------------------------------------------------------------------
//--- compute and set the new clock interval
//--------------------------------------------------------------------------

// returns True if a new interval could be computed (depends on the
// number of nodes in the system and the maximum number of faulty
// nodes); in that case 'newTime' is set to the new clock value; if
// no valid interval could be computed, False is returned and 'newTime'
// is not changed
EBool CFti::ConvergenceFunction ()
{
    CNtpTime point;                  // new reference point
    double alpha_minus, alpha_plus;  // new accuracies
    STimeInterval evalState;         // state interval for EvalSys

    timeValid = False;  // mark the time as invalid

    if (GetAssociatedNetwork() == NoAccount) return False;
    // no associated network

    if (!Fti (point, alpha_minus, alpha_plus, GetAssociatedNetwork())) 
        return False;
    // FTI function failed

    *newClockValue = IntervalToClock (point, alpha_minus, alpha_plus);
    // set new clock interval

    evalState.refpoint = point;
    evalState.low  = point - CNtpTime (alpha_minus);
    evalState.high = point + CNtpTime (alpha_plus);
    ReportInterval (False, evalState);
    // report new state interval to EvalSys as output interval

    timeValid = True;  
    // the new timestamp is valid

    return True;
}


//--------------------------------------------------------------------------
//--- manage clock changes
//--------------------------------------------------------------------------

// computes initialization data (clock deterioration, accuracy) for 
// the clock
void CFti::ClockChanged()
{
    CClockInterface *clock;  // pointer to the clock
    STimestamp time;         // current time
    double fnom;             // nominal clock frequency


    ACClockSyncAlgorithm::ClockChanged();  // initialize the local rate

    // initialize the clock
    // beware the order: 1st the deterioration, 2nd the accuracy

    clock = GetClock();
    if (clock == NULL) return;
    // the clock is not available, so we cannot initialize it
    // note: we could report that the clock is NULL, since the user may not
    // have intended this; but the report might come into the way of object
    // destruction

    fnom = clock->GetFrequency ();
    // lambda = rho/fnom and convert to binary and select S[2^-37,...,2^51]
    // e.g. rho=1^-6, fnom = 25Mhz, set lambda to 2^-44, thus 0x0080
    // e.g. rho=10^-6, thus 0x0500
    clock->SetLambdapure (0x00D0, 0x00D0); // 0x0080
    // intitialize the deterioration

    clock->GetTimestamp (time);
    time.acc = 0x00500050;
    // upper 2 bytes = left acc, lower 2 bytes = right acc
    clock->SetTimestamp (time);
    // initialize accuracy
}


//--------------------------------------------------------------------------
//--- create function of class CFreeRun
//--------------------------------------------------------------------------

// creates and returns a new object of type CFreeRun
ACClockSyncAlgorithm *CFti::CreateFti (CClockSync *cs)
{
    return new CFti (cs);
}


//--------------------------------------------------------------------------
//--- FTI algorithm
//--------------------------------------------------------------------------

// computes the new clock interval for the network; returns True if the
// computation was successful, False otherwise
EBool CFti::Fti (CNtpTime& point, double& alpha_minus, double& alpha_plus,
               SINT32 networkID)
{
    STimeInterval data;
    int i; //
    EBool entryExists;  // 
    INT32 size;  // array size
    SAsymInterval *accuracyIntervals; // 
    SAsymInterval accIntResult;  
    filebuf Facc, Fpre, Fadj, Fdet;

    if (numberOfNodes <= 3*numberOfFaultyNodes) // n > 3f is needed for OP
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        // delete all data of the network
        return False;
    }

    size = database->Cardinality (CApplicationDatabase::Time, networkID);
    if (size <= 2*numberOfFaultyNodes)
    // need at least 2*f + 1 timestamps, because 
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }

    accuracyIntervals = new SAsymInterval[size];
    // these should be the input intervals represented as doubles

    for (entryExists = database->GetFirstTime (data, networkID), i = 0;  
         entryExists;
         entryExists = database->GetNextTime (data, networkID), i++)
    {
         accuracyIntervals[i].left  = NtpToSeconds (data.low);
         accuracyIntervals[i].ref   = NtpToSeconds (data.refpoint);
         accuracyIntervals[i].right = NtpToSeconds (data.high);
    }

    // now we have the accuracy intervals at hand in accuracyIntervals[]

#ifdef GNU_ACC_FTI
    if (Facc.open("../gnuplot/acc.dat",output) == 0)
    {
        delete[] accuracyIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }
    ostream acc(&Facc); 
       
    for(i=0;i<size;i++)
    {
        ftoa(i,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].ref,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].left,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].right,WidthOfChars,val);
        acc << val << endl;
    }
#endif

    if (!FTIntersection (accuracyIntervals, accIntResult, size, 
                         numberOfFaultyNodes))
    {
        delete[] accuracyIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
#ifdef GNU_ACC_FTI
        Facc.close();
#endif
        return False;
    }    
    // FTI function upon accuracy intervals

#ifdef GNU_ACC_FTI
    acc << "6.0 ";
    ftoa(accIntResult.ref,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntResult.left,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntResult.right,WidthOfChars,val);
    acc << val << endl;

    Facc.close();
#endif

    SecondsToNtp (accIntResult.ref, point.ms, point.ts, point.us);
    // hand back the new reference point

    alpha_minus = accIntResult.ref - accIntResult.left;
    Assert( alpha_minus >= 0);
    // hand back the new left edge

    alpha_plus = accIntResult.right - accIntResult.ref;
    Assert( alpha_plus >= 0);
    // hand back the new right edge
    
    delete[] accuracyIntervals;
    database->DeleteAll (CApplicationDatabase::Time, networkID);
    // the data is not needed anymore

    return True;
}


//**************************************************************************
// module global functions
//**************************************************************************

//--------------------------------------------------------------------------
//--- FTI function
//--------------------------------------------------------------------------

EBool FTIntersection (SAsymInterval* input, SAsymInterval& output, 
                      INT32 size, int f)
{
    int i;
    int minCut;
    double *sortedArray;
    double *unsortedArray;

    minCut = size-f;
    if (minCut < 0 ) return False;
    // number of required cuts

    sortedArray = new double[size];
    unsortedArray = new double[size];

    // determine the left edge

    for (i=0;i<size;i++)
        unsortedArray[i] = input[i].left;
    
    SortArray (unsortedArray, sortedArray, size);

    output.left = sortedArray[size-1-f];   // skip the f innermost edges

    // determine the right edge
    
    for (i=0;i<size;i++)
        unsortedArray[i] = input[i].right;
    
    SortArray (unsortedArray, sortedArray, size);
    
    output.right = sortedArray[f];

    output.ref = (output.left + output.right)/2;
    // take the midpoint as default
 
    delete[] sortedArray;
    delete[] unsortedArray;

    return True;
}

