//
// alggen.cpp
//
// contains constants and functions that are needed by several algorithms
//
// created: 19.01.99
//
// last change: 21.01.99
//
// RCS version control information:
//     $Revision: 1.8 $$State: Exp $$Date: 1999/01/21 12:42:22 $
//

#include <stdlib.h>  // for NULL
#include <string.h>  // for memmove()

#include "alggen.hpp"
#include "error.hpp"


//**************************************************************************
// functions
//**************************************************************************

//--------------------------------------------------------------------------
//--- Marzullo function
//--------------------------------------------------------------------------

EBool ApplyMarzullo (SAsymInterval* input, SAsymInterval& output, INT32 size,
                     int f)
{
    Assert (input != NULL);

    int i, j;
    int cut;
    int minCut;
    double sweep;
    double *sortedArray;
    double *unsortedArray;

    minCut = size - f;
    if (minCut < 0 ) return False;
    // number of required intersections

    sortedArray = new double[size];
    unsortedArray = new double[size];

    // --- determine the left edge

    for (i=0;i<size;i++)
        unsortedArray[i] = input[i].left;
    // fetch left edges

    SortArray (unsortedArray, sortedArray, size);
    // sort left edges ascending
    for (i=0; i<size; i++)
    {
        sweep = sortedArray[i];
        // go out-side-in

        cut = 0;
        for(j=0;j<size;j++)
        {
            if( (input[j].left <= sweep) && (sweep <= input[j].right))
                cut++;
        }
        // count the intersections with the intervals

        if (cut >= minCut) break;
        // stop when suffiencient many intersections
    }

    if (cut < minCut)
    {
        delete[] sortedArray;
        delete[] unsortedArray;
        return False;
    }
    else
        output.left = sweep;
    // hand back the left edge

    // --- determine the right edge

    for (i=0;i<size;i++)
        unsortedArray[i] = input[i].right;
    // fetch right edges

    SortArray (unsortedArray, sortedArray, size);
    // sort right edges ascending

    for (i=size-1; i>=0; i--)
    {
        sweep = sortedArray[i];
        // go out-side-in

        cut = 0;
        for(j=0;j<size;j++)
        {
            if( (input[j].left <= sweep) && (sweep <= input[j].right))
                cut++;
        }
        // count the intersection with the intervals

        if (cut >= minCut) break;
        // stop when suffiencient many intersections
    }

    if (cut < minCut)
    {
        delete[] sortedArray;
        delete[] unsortedArray;
        return False;
    }
    else
        output.right = sweep;
    // hand back the right edge

    output.ref = (output.left + output.right)/2;
    // take the midpoint, but has no meaning

    delete[] sortedArray;
    delete[] unsortedArray;

    return True;
}


//--------------------------------------------------------------------------
//--- sorting an array of doubles
//--------------------------------------------------------------------------

void SortArray (double* unsortedArray, double* sortedArray, INT32 size)
{
    Assert (unsortedArray != NULL);
    Assert (sortedArray != NULL);
    Assert (size > 0);

    int i;
    int free;
    int up, low, curr;
    double item;

    free = 0;
    low = up = curr = 0;

    for (i=0;i<size;i++)
    {
        item = unsortedArray[i];

        if (free != 0)  // find the entry
        {
            up = (free>0) ? free-1 : 0;  // prepare for binary search
            low = 0;
            curr = (up+low)/2;

            while ((item != sortedArray[curr]) && (low < up))
            {
                if (item < sortedArray[curr]) up = (curr>0) ? curr-1 : 0;
                else if (item > sortedArray[curr]) low = curr+1;

                curr = (low+up)/2;
            }

            if (item > sortedArray[curr]) curr++;
            // didn't find the entry, need to insert behind the current one
        } // if

        Assert (curr < (SINT32)(size));
        Assert (curr >= 0);

        memmove (sortedArray+curr+1, sortedArray+curr,  // move everything
                 (size-curr-1)*sizeof(double));

        sortedArray[curr] = item;  // enter the timestamp

        free++;  // increment the free index
    } // for
}


//--------------------------------------------------------------------------
//--- intersection of two intervals
//--------------------------------------------------------------------------

EBool Intersect (const SAsymInterval& In1, const SAsymInterval& In2,
    SAsymInterval& Out)
{
    if (In1.left > In2.left)
        Out.left = In1.left;
    else
        Out.left = In2.left;
    // left edge

    if (In1.right > In2.right)
        Out.right = In2.right;
    else
        Out.right = In1.right;
    // right edge

    if (Out.left > Out.right)
        return False;
    // valid interval?

    Out.ref = (Out.left + Out.right)/2;
    // take the midpoint, but has no meaning

    return True;
}

//--------------------------------------------------------------------------
//--- convert computed state interval to adjusted clock
//--------------------------------------------------------------------------

SExtendedTimestamp IntervalToClock (CNtpTime point, double alpha_minus,
    double alpha_plus)
{
    SExtendedTimestamp clock;         // resulting clock
    CNtpTime a_ntp_plus, a_ntp_minus; // converted accuracies

    clock = point;
    // set the reference point, = is overloaded

    SecondsToNtp (alpha_minus, a_ntp_minus.ms, a_ntp_minus.ts, a_ntp_minus.us);
    SecondsToNtp (alpha_plus, a_ntp_plus.ms, a_ntp_plus.ts, a_ntp_plus.us);
    // map both accuracy intervals into registers

    if ((a_ntp_plus.ms != 0) || (a_ntp_plus.ts >= HighestAccuracyValue))
        a_ntp_plus.ts = Infinity;
    // prepare accuracy overflow

    if ((a_ntp_minus.ms != 0) || (a_ntp_minus.ts >= HighestAccuracyValue))
        a_ntp_minus.ts = Infinity;
    // prepare accuracy overflow

    clock.acc = (a_ntp_plus.ts>>1 & 0x0FFFF) +
                (a_ntp_minus.ts<<15 & 0xFFFF0000);
    // compose the accuracies
    // positive: .ts is (7,-24), need (-8,-23) in lower 2 bytes, hence >>1
    // negative: .ts is (7,-24), need (-8,-23) in upper 2 bytes, hence <<15

    return clock;
}

