//
// algmarz.cpp
//
// implementation of Marzullo's "clock synchronization algorithm"
//
// created: 03.07.97, Klaus Schossmaier
//
// last change: 03.07.97, Klaus Schossmaier
//
// last revision: 04.02.99
//
// RCS version control information:
//     $Revision: 1.13 $$State: Exp $$Date: 1999/02/04 14:17:44 $
//

#include <fstream.h>
#include <string.h>

#include "address.hpp"
#include "alggen.hpp"
#include "algmarz.hpp"
#include "clockifc.hpp"
#include "constant.hpp"
#include "database.hpp"
#include "error.hpp"
#include "time.hpp"


// enable output files
//#define GNU_ACC_MARZ

extern char *ftoa (double f, int width, char *string);  // from ftoa.cpp
// returns 'f' as a string with at least 'width' characters in %e format;
// stores the string in 'string', which must be big enough to hold the
// converted string; if 'string' is NULL, NULL is returned


//**************************************************************************
// module global functions
//**************************************************************************

static SExtendedTimestamp IntervalToClock (STimeInterval& interval);
// converts the time interval into the clock register format and returns
// this value


//**************************************************************************
// class CMarzullo: features the Marzullo algorithm
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assert: cs != NULL (done by the ctor of ACClockSyncAlgorithm)
CMarzullo::CMarzullo (CClockSync *cs) : ACClockSyncAlgorithm (cs)
{
    ClockChanged();
    // initialize the clock data
}

//--------------------------------------------------------------------------
//--- compute and set the new clock interval
//--------------------------------------------------------------------------

// returns True if a new interval could be computed (depends on the
// number of nodes in the system and the maximum number of faulty
// nodes); in that case 'newTime' is set to the new clock value; if
// no valid interval could be computed, False is returned and 'newTime'
// is not changed
EBool CMarzullo::ConvergenceFunction ()
{
    STimeInterval interval;  // the new time interval
    timeValid = False;  // mark the time as invalid


    if (GetAssociatedNetwork() == NoAccount) return False;
    // no associated network

    if (!Marzullo (interval, GetAssociatedNetwork())) return False;

    *newClockValue = IntervalToClock (interval);
    // set new clock interval

    ReportInterval (False, interval);
    // report resulting interval to EvalSys

    timeValid = True;  
    // the new timestamp is valid

    return True;
}


//--------------------------------------------------------------------------
//--- manage clock changes
//--------------------------------------------------------------------------

// computes initialization data (clock deterioration, accuracy) for
// the clock, and some clock dependent parameters needed by the algorithm
void CMarzullo::ClockChanged()
{
    CClockInterface *clock;  // pointer to the clock
    STimestamp time;         // current time
    double fnom;             // nominal clock frequency


    ACClockSyncAlgorithm::ClockChanged();  // initialize the local rate

    // initialize the clock
    // beware the order: 1st the deterioration, 2nd the accuracy

    clock = GetClock();
    if (clock == NULL) return;
    // the clock is not available, so we cannot initialize it
    // note: we could report that the clock is NULL, since the user may not
    // have intended this; but the report might come into the way of object
    // destruction

    fnom = clock->GetFrequency ();
    // lambda = rho/fnom and convert to binary and select S[2^-37,...,2^51]
    // e.g. rho=1^-6, fnom = 25Mhz, set lambda to 2^-44, thus 0x0080
    // e.g. rho=10^-6, thus 0x0500 
    clock->SetLambdapure (0x00D0, 0x00D0); // 0x0080
    // intitialize the deterioration

    clock->GetTimestamp (time);
    time.acc = 0x00500050; 
    // upper 2 bytes = left acc, lower 2 bytes = right acc
    clock->SetTimestamp (time);
    // initialize accuracy to zero
}


//--------------------------------------------------------------------------
//--- create function of class CFreeRun
//--------------------------------------------------------------------------

// creates and returns a new object of type CMarzullo
ACClockSyncAlgorithm *CMarzullo::CreateMarzullo (CClockSync *cs)
{
    return new CMarzullo (cs);
}


//--------------------------------------------------------------------------
//--- Marzullo's algorithm
//--------------------------------------------------------------------------

// computes the new clock interval for the network; returns True if the
// computation was successful, False otherwise
EBool CMarzullo::Marzullo (STimeInterval& interval, SINT32 networkID)
{
    STimeInterval data;
    int i; //
    EBool entryExists;  // 
    INT32 size;  // array size
    SAsymInterval *accuracyIntervals; // 
    SAsymInterval accIntResult;  
    filebuf Facc, Fpre, Fadj, Fdet;

    if (numberOfNodes <= 3*numberOfFaultyNodes) // n > 3f is needed for OP
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        // delete all data of the network
        return False;
    }

    size = database->Cardinality (CApplicationDatabase::Time, networkID);
    if (size <= 2*numberOfFaultyNodes)
    // need at least 2*f + 1 timestamps, because 
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }

    accuracyIntervals = new SAsymInterval[size];
    // these should be the input intervals represented as doubles

    for (entryExists = database->GetFirstTime (data, networkID), i = 0;  
         entryExists;
         entryExists = database->GetNextTime (data, networkID), i++)
    {
         accuracyIntervals[i].left  = NtpToSeconds (data.low);
         accuracyIntervals[i].ref   = NtpToSeconds (data.refpoint);
         accuracyIntervals[i].right = NtpToSeconds (data.high);
    }

    // now we have the accuracy intervals at hand in accuracyIntervals[]

#ifdef GNU_ACC_MARZ
    if (Facc.open("../gnuplot/acc.dat",output) == 0)
    {
        delete[] accuracyIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }
    ostream acc(&Facc); 
       
    for(i=0;i<size;i++)
    {
        ftoa(i,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].ref,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].left,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].right,WidthOfChars,val);
        acc << val << endl;
    }
#endif

    if (!ApplyMarzullo (accuracyIntervals, accIntResult, size,
                        numberOfFaultyNodes))
    {
        delete[] accuracyIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
#ifdef GNU_ACC_MARZ
        Facc.close();
#endif
        return False;
    }    
    // Marzullo's function upon accuracy intervals

#ifdef GNU_ACC_MARZ
    acc << "6.0 ";
    ftoa(accIntResult.ref,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntResult.left,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntResult.right,WidthOfChars,val);
    acc << val << endl;

    Facc.close();
#endif

    SecondsToNtp (accIntResult.ref,
    interval.refpoint.ms, interval.refpoint.ts, interval.refpoint.us);
    // hand back the new reference point

    SecondsToNtp (accIntResult.left,  
    interval.low.ms, interval.low.ts, interval.low.us);
    // hand back the new left edge

    SecondsToNtp (accIntResult.right, 
    interval.high.ms, interval.high.ts, interval.high.us);
    // hand back the new right edge
    
    delete[] accuracyIntervals;
    database->DeleteAll (CApplicationDatabase::Time, networkID);
    // the data is not needed anymore

    return True;
}



//**************************************************************************
// module global functions
//**************************************************************************

//--------------------------------------------------------------------------
//--- convert computed interval to adjusted clock 
//--------------------------------------------------------------------------

// converts the time interval into the clock register format and returns
// this value
static SExtendedTimestamp IntervalToClock (STimeInterval& interval)
{
    SExtendedTimestamp clock;
    CNtpTime alpha_plus, alpha_minus;

    clock = interval.refpoint;
    // set the reference point, = is overloaded

    Assert ( (interval.high.ms > interval.low.ms) ||
             ((interval.high.ms == interval.low.ms)) &&
             ((interval.high.ts >= interval.low.ts))    );
    // make sure that the upper bound is >= the lower bound

    alpha_plus = interval.high - interval.refpoint;
    alpha_minus = interval.refpoint - interval.low;
    // comute both accuracy intervals

    if ((alpha_plus.ms != 0) || (alpha_plus.ts >= HighestAccuracyValue))
        alpha_plus.ts = Infinity;
    // prepare accuracy overflow

    if ((alpha_minus.ms != 0) || (alpha_minus.ts >= HighestAccuracyValue))
        alpha_minus.ts = Infinity;
    // prepare accuracy overflow

    clock.acc = (alpha_plus.ts>>1 & 0x0FFFF)+(alpha_minus.ts<<15 & 0xFFFF0000);
    // compose the accuracies
    // positive: .ts is (7,-24), need (-8,-23) in lower 2 bytes, hence >>1
    // negative: .ts is (7,-24), need (-8,-23) in upper 2 bytes, hence <<15

    return clock;
}


