//
// algop.cpp
//
// implementation of the OP "clock synchronization algorithm", which
// does not synchronize at all, but lets the clock drift
//
// created: 01.09.98, Klaus Schossmaier
//
// last change: 07.10.98, Klaus Schossmaier
//
// last revision: 04.02.99
//
// RCS version control information:
//     $Revision: 1.10 $$State: Exp $$Date: 1999/02/04 14:17:44 $
//

#include <fstream.h>
#include <string.h>

#include "address.hpp"
#include "algop.hpp"
#include "clockifc.hpp"
#include "constant.hpp"
#include "database.hpp"
#include "error.hpp"
#include "time.hpp"


// need the following algorithmic parameters
#define PI_H_MINUS 5e-3 // 208e-6
#define PI_H_PLUS 5e-3 // 212e-6
#define PI_O_MINUS 5e-3 // 207e-6 
#define PI_O_PLUS 5e-3 // 207e-6
#define UPSILON_MAX 300e-6 // 201e-6

// enable output files
//#define GNU_ACC
//#define GNU_PRE
//#define GNU_DET
//#define GNU_ADJ

// file names
#define NAME_ACC "acc.dat"
#define NAME_PRE "pre.dat"
#define NAME_DET "det.dat"
#define NAME_ADJ "adj.dat"

// provide some macros for closing these files
#ifdef GNU_ACC
  #define CloseAcc() (Facc.close())
#else
  #define CloseAcc() ((void)0)
#endif

#ifdef GNU_PRE
  #define ClosePre() (Fpre.close())
#else
  #define ClosePre() ((void)0)
#endif

#ifdef GNU_DET
  #define CloseDet() (Fdet.close())
#else
  #define CloseDet() ((void)0)
#endif

#ifdef GNU_ADJ
  #define CloseAdj() (Fadj.close())
#else
  #define CloseAdj() ((void)0)
#endif



//**************************************************************************
// module global constants
//**************************************************************************

const int StartSynchronization = 5;
// number of delay measurements needed before the algorithm starts
// using the delay data in StoreTime()

const double EpsilonPlusInitial = 0.0; // 20e-6;
const double EpsilonMinusInitial = 0.0; // 20e-6;
// for the computation of the precision intervals we need
// some initial values for EpsilonPlus/Minus
// they can be refined after some delay measurements

extern char *ftoa (double f, int width, char *string);  // from ftoa.cpp
// returns 'f' as a string with at least 'width' characters in %e format;
// stores the string in 'string', which must be big enough to hold the
// converted string; if 'string' is NULL, NULL is returned


//**************************************************************************
// class COP: features the OP algorithm
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assert: cs != NULL (done by the ctor of ACClockSyncAlgorithm)
COP::COP (CClockSync *cs) : ACClockSyncAlgorithm (cs)
{
    cntDelay = 0;  // no delay measurements
    SetResynchronizationStatus (True);  // own time is not valid

    ClockChanged();  // initialize clock dependent data
}


//--------------------------------------------------------------------------
//--- compute and set the new clock interval
//--------------------------------------------------------------------------

EBool COP::ConvergenceFunction ()
{
    CNtpTime point;                  // new reference point
    double alpha_minus, alpha_plus;  // new accuracies
    STimeInterval evalState;         // state interval for EvalSys
  
    timeValid = False;  // mark the time as invalid

    if (GetAssociatedNetwork() == NoAccount) 
        return False;
    // no associated network

    if (!OP (point, alpha_minus, alpha_plus, GetAssociatedNetwork())) 
        return False;
    // OP function failed

    *newClockValue = IntervalToClock(point, alpha_minus, alpha_plus);
    // set new clock interval

    evalState.refpoint = point;    
    evalState.low  = point - CNtpTime (alpha_minus);    
    evalState.high = point + CNtpTime (alpha_plus);    
    ReportInterval (False, evalState);    
    // report new state interval to EvalSys as output interval

    timeValid = True;  
    // the new timestamp is valid

    return True;
}


//--------------------------------------------------------------------------
//--- add time intervals
//--------------------------------------------------------------------------

// if CS:resync is True, the time is not stored; so the algorithm
// only starts working after resync is False; then the time is
// passed on to StoreTime() of the base class
EBool COP::StoreTime (const STimestamp& time, double resync,
        SINT32 networkID, const CAddr& netAddress, int area, EBool gps)
{
    if (GetResynchronizationStatus ()) return False;
    // do not store the time as long as CS is resynchronizing

    return ACClockSyncAlgorithm::StoreTime (time, resync, networkID,
                                            netAddress, area, gps);
}


//--------------------------------------------------------------------------
//--- add transmission delay intervals
//--------------------------------------------------------------------------

// counts to N before setting CS::resync to False, calls StoreDelay
// of base class to process delays
EBool COP::StoreDelay (double localDelay, double remoteDelay,
     const SRate& remoteRate, SINT32 networkID, const CAddr& netAddress)
{
    if (cntDelay >= 0) cntDelay++;

    if (cntDelay == StartSynchronization)
    {
        SetResynchronizationStatus (False);
        // the own timestamp is valid

        cntDelay = -1;  // stop counting the delay measurement rounds
    }

    // store the new delay
    return ACClockSyncAlgorithm::StoreDelay (localDelay, remoteDelay,
                remoteRate, networkID, netAddress);
}


//--------------------------------------------------------------------------
//--- manage clock changes
//--------------------------------------------------------------------------

// computes initialization data (clock deterioration, accuracy) for
// the clock, and some clock dependent parameters needed by the algorithm
void COP::ClockChanged()
{
    CClockInterface *clock;    // pointer to the clock
    STimestamp time;           // current time
    double fnom;               // nominal clock frequency
    SINT32 nodeID;             // own node ID
    int area;                  // own area
    double pi0Minus, pi0Plus;  // precision interval


    ACClockSyncAlgorithm::ClockChanged();  // initialize the local rate

    // initialize the clock
    // beware the order: 1st the deterioration, 2nd the accuracy

    clock = GetClock();
    if (clock == NULL) return;
    // the clock is not available, so we cannot initialize it
    // note: we could report that the clock is NULL, since the user may not
    // have intended this; but the report might come into the way of object
    // destruction

    GetOwnID (nodeID, area);
    cout << "node " << nodeID << " uses:" << endl;

    fnom = clock->GetFrequency ();
    // lambda = rho/fnom and convert to binary and select S[2^-37,...,2^51]
    // e.g. rho=1e-6, fnom = 25Mhz, set lambda to 2^-44, thus 0x0080
    // e.g. rho=2e-6, thus 0x00C0
    // e.g. rho=10e-6, thus 0x0500 
    clock->SetLambdapure (0x010F, 0x010F); 
    // intitialize the deterioration

    clock->GetTimestamp (time);
    time.acc = 0x00500050;  
    // upper 2 bytes = left acc, lower 2 bytes = right acc
    // e.g. 9.5us -> 0050
    clock->SetTimestamp (time);
    // initialize accuracy to zero
   
    // precision interval pi_0 = [pi0Minus,pi0Plus]:

    pi0Minus = (EpsilonPlusInitial + EpsilonMinusInitial) + 
          4 * clock->GetGranularity () + 
          0.5 * (GetRoundPeriod () + GetDelayCompensation () + 
                 2 * GetComputationDelayCompensation ())  
              * (clock->GetRhoPlus () + clock->GetRhoMinus ()) + 
          0.5 * GetRoundPeriod () 
              * (clock->GetRhoPlus () - clock->GetRhoMinus ());

    pi0Plus = (EpsilonPlusInitial + EpsilonMinusInitial) + 
          4 * clock->GetGranularity () + 
          0.5 * (GetRoundPeriod () + GetDelayCompensation () + 
                 2 * GetComputationDelayCompensation ())   
              * (clock->GetRhoPlus () + clock->GetRhoMinus ()) - 
          0.5 * GetRoundPeriod () 
              * (clock->GetRhoPlus () - clock->GetRhoMinus ());

    // precision interval pi^H = [piHMinus,piHPlus]:


#ifdef PI_H_MINUS
    piHMinus = PI_H_MINUS; //240e-6 + 0;
#else
    piHMinus = pi0Minus + EpsilonMinusInitial + 3 * clock->GetGranularity () +
          (GetRoundPeriod () + GetComputationDelayCompensation ()) 
          * clock->GetRhoMinus ();
#endif

#ifdef PI_H_PLUS
    piHPlus = PI_H_PLUS; //240e-6 + 0;
#else
    piHPlus = pi0Plus + EpsilonPlusInitial + 2 * clock->GetGranularity () +
          (GetRoundPeriod () + GetComputationDelayCompensation ()) 
          * clock->GetRhoPlus ();
#endif

    cout << "    piH =   [" << piHMinus << ", " << piHPlus << "]" << endl;

    // precision interval pi^o = [piOwnMinus,piOwnPlus]:

#ifdef PI_O_MINUS
    piOwnMinus = PI_O_MINUS; //240e-6 + 0;
#else
    piOwnMinus = pi0Minus + clock->GetGranularity () +
          GetRoundPeriod () * clock->GetRhoMinus ();
#endif

#ifdef PI_O_PLUS
    piOwnPlus = PI_O_PLUS; //240e-6 + 0;
#else
    piOwnPlus = pi0Plus + clock->GetGranularity () +
          GetRoundPeriod () * clock->GetRhoPlus ();
#endif

    cout << "    piOwn = [" << piOwnMinus << ", " << piOwnPlus << "]" << endl;

    // maximal correction Upsilon_max:

#ifdef UPSILON_MAX 
    Ups = UPSILON_MAX; //500e-6;
#else
    Ups = 2 * clock->GetGranularity () +
        GetRoundPeriod () * (clock->GetRhoPlus () + clock->GetRhoMinus ());
#endif

    cout << "    Ups = " << Ups << endl;
}


//--------------------------------------------------------------------------
//--- create function of class COP
//--------------------------------------------------------------------------

// creates and returns a new object of type COP
ACClockSyncAlgorithm *COP::CreateOP (CClockSync *cs)
{
    return new COP (cs);
}

//--------------------------------------------------------------------------
//--- OP algorithm
//--------------------------------------------------------------------------

// computes the new clock interval for the network; returns True if the
// computation was successful, False otherwise
EBool COP::OP (CNtpTime& point, double& alpha_minus, double& alpha_plus,
               SINT32 networkID)
{
    STimeInterval data;
    int i; 
    EBool entryExists;  
    INT32 size;  // array size
    SINT32 nodeID;
    int area;
    CAddr netAddress;
    SAsymInterval *accuracyIntervals; 
    SAsymInterval *precisionIntervals;  
    SAsymInterval preIntMarz;
    SAsymInterval preIntResult;
    SAsymInterval accIntMarz;
    SAsymInterval accIntResult;  
    double adjustment;
    filebuf Facc, Fpre, Fadj, Fdet;

    if (numberOfNodes <= 3*numberOfFaultyNodes) // n > 3f is needed for OP
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        // delete all data of the network
        return False;
    }

    size = database->Cardinality (CApplicationDatabase::Time, networkID);
    if (size <= 2*numberOfFaultyNodes)
    // need at least 2*f + 1 timestamps, because 
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }

    accuracyIntervals = new SAsymInterval[size];
    precisionIntervals = new SAsymInterval[size];
    // these should be the input intervals represented as doubles

    // the own accuracy/precision interval

    GetOwnID (nodeID, area);
    // for identification
    if (!database->GetAssociation (networkID, nodeID, netAddress)) 
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }    
    // obtain netAddress
    if (!database->RemoveTime (data, networkID, netAddress, area))
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }
    // obtain own accuracy interval

    accuracyIntervals[0].left  = NtpToSeconds (data.low);
    accuracyIntervals[0].ref   = NtpToSeconds (data.refpoint);
    accuracyIntervals[0].right = NtpToSeconds (data.high);
    // own accuracy interval represented as doubles, index = 0

    precisionIntervals[0].left  = accuracyIntervals[0].ref - piOwnMinus;
    precisionIntervals[0].ref   = accuracyIntervals[0].ref; 
    precisionIntervals[0].right = accuracyIntervals[0].ref + piOwnPlus;
    // own precision interval represented as doubles, index = 0

    // the remote accuracy/precision intervals

    for (entryExists = database->GetFirstTime (data, networkID), i = 1;  
         entryExists;
         entryExists = database->GetNextTime (data, networkID), i++)
    {
         accuracyIntervals[i].left  = NtpToSeconds (data.low);
         accuracyIntervals[i].ref   = NtpToSeconds (data.refpoint);
         accuracyIntervals[i].right = NtpToSeconds (data.high);

         precisionIntervals[i].left  = accuracyIntervals[i].ref - piHMinus;
         precisionIntervals[i].ref   = accuracyIntervals[i].ref; 
         precisionIntervals[i].right = accuracyIntervals[i].ref + piHPlus;
    }

    // now we have the accuracy and precision intervals at hand
    // in accuracyIntervals[] and precisionIntervals[] respectively
    // where in [0] are the own ones

#ifdef GNU_ACC
    if (Facc.open(NAME_ACC,output) == 0)
    {
        Report (cannot open accuracy file);
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }
    ostream acc(&Facc); 
       
    for(i=0;i<size;i++)
    {
        ftoa(i,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].ref,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].left,WidthOfChars,val);
        acc << val << " ";
        ftoa(accuracyIntervals[i].right,WidthOfChars,val);
        acc << val << endl;
    }
#endif

#ifdef GNU_PRE
    if (Fpre.open(NAME_PRE,output) == 0)
    {
        Report (cannot open precision file);
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        CloseAcc();
        return False;
    }
    ostream pre(&Fpre); 
       
    for(i=0;i<size;i++)
    {
        ftoa(i,WidthOfChars,val);
        pre << val << " ";
        ftoa(precisionIntervals[i].ref,WidthOfChars,val);
        pre << val << " ";
        ftoa(precisionIntervals[i].left,WidthOfChars,val);
        pre << val << " ";
        ftoa(precisionIntervals[i].right,WidthOfChars,val);
        pre << val << endl;
    }
#endif
    
    if (!ApplyMarzullo (precisionIntervals, preIntMarz, size,
                        numberOfFaultyNodes))
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        CloseAcc();
        ClosePre();
        return False;
    }
    // Marzullo's function upon associated precision intervals
    // normaly use f = numberOfFaultyNodes

#ifdef GNU_PRE
    pre << "6.0 ";
    ftoa(preIntMarz.ref,WidthOfChars,val);
    pre << val << " ";
    ftoa(preIntMarz.left,WidthOfChars,val);
    pre << val << " ";
    ftoa(preIntMarz.right,WidthOfChars,val);
    pre << val << endl;
#endif

    if (!Intersect (preIntMarz, precisionIntervals[0], preIntResult) )
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        CloseAcc();
        ClosePre();
        return False;
    }
    // intersection with local precision interval

    preIntResult.ref =
    (piOwnMinus * preIntResult.right + piOwnPlus * preIntResult.left) /
    (piOwnMinus + piOwnPlus);
    // candidat reference point

#ifdef GNU_PRE
    pre << "7.0 ";
    ftoa(preIntResult.ref,WidthOfChars,val);
    pre << val << " ";
    ftoa(preIntResult.left,WidthOfChars,val);
    pre << val << " ";
    ftoa(preIntResult.right,WidthOfChars,val);
    pre << val << endl;
#endif

    adjustment = preIntResult.ref - precisionIntervals[0].ref;
 
    if (adjustment > Ups)
        preIntResult.ref = precisionIntervals[0].ref + Ups;
    if (adjustment < - Ups)
        preIntResult.ref = precisionIntervals[0].ref - Ups;
    // new reference point 

#ifdef GNU_PRE
    pre << "8.0 ";
    ftoa(preIntResult.ref,WidthOfChars,val);
    pre << val << " ";
    ftoa(preIntResult.left,WidthOfChars,val);
    pre << val << " ";
    ftoa(preIntResult.right,WidthOfChars,val);
    pre << val << endl;
#endif

#ifdef GNU_ADJ
    if (Fadj.open(NAME_ADJ,append) == 0)
    {
        Report (cannot open adjustments file);
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        CloseAcc();
        ClosePre();
        return False;
    }
    ostream adj(&Fadj); 

    if (nodeID == 4)
    {    
        ftoa(round++,WidthOfChars,val);
        adj << val << " ";
        ftoa(adjustment,WidthOfChars,val);
        adj << val << endl;
    }
#endif

    if (!ApplyMarzullo (accuracyIntervals, accIntMarz, size,
                        numberOfFaultyNodes))
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }    
    // Marzullo's function upon accuracy intervals
    // normaly use f = numberOfFaultyNodes

#ifdef GNU_ACC
    acc << "6.0 ";
    ftoa(accIntMarz.ref,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntMarz.left,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntMarz.right,WidthOfChars,val);
    acc << val << endl;
#endif

    if (!Intersect (accIntMarz, accuracyIntervals[0], accIntResult) )
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        CloseAcc();
        ClosePre();
        return False;
    }     
    // intersection with local accuracy interval

#ifdef GNU_ACC
    acc << "7.0 ";
    ftoa(accIntResult.ref,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntResult.left,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntResult.right,WidthOfChars,val);
    acc << val << endl;
#endif

    if (preIntResult.ref > accIntResult.right)
        accIntResult.right = preIntResult.ref;
    if (preIntResult.ref < accIntResult.left)
        accIntResult.left = preIntResult.ref;
    // new edges of accuracy interval

#ifdef GNU_ACC
    acc << "8.0 ";
    ftoa(accIntResult.ref,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntResult.left,WidthOfChars,val);
    acc << val << " ";
    ftoa(accIntResult.right,WidthOfChars,val);
    acc << val << endl;
#endif


#ifdef GNU_DET
    if (Fdet.open(NAME_DET,append) == 0)
    {
        Report (cannot open deteriorated accuracies file);
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        CloseAcc();
        ClosePre();
        CloseAdj();
        return False;
    }
    ostream det(&Fdet); 

    if (nodeID == 1)
    {    
        
        ftoa(accIntResult.ref,WidthOfChars,val);
        det << val << " ";
        det << "0.0 ";
        ftoa(accIntResult.left - accIntResult.ref,WidthOfChars,val);
        det << val << " ";
        ftoa(accIntResult.right - accIntResult.ref,WidthOfChars,val);
        det << val << endl;
    }
#endif

    SecondsToNtp (preIntResult.ref, point.ms, point.ts, point.us);
    // hand back the new reference point

    alpha_minus = preIntResult.ref - accIntResult.left;
    Assert( alpha_minus >= 0);
    // hand back the new left edge

    alpha_plus = accIntResult.right - preIntResult.ref;
    Assert( alpha_plus >= 0);
    // hand back the new right edge
    
    delete[] accuracyIntervals;
    delete[] precisionIntervals;
    database->DeleteAll (CApplicationDatabase::Time, networkID);
    // the data is not needed anymore

    CloseAcc();
    ClosePre();
    CloseAdj();
    CloseDet();

    return True;
}


//--------------------------------------------------------------------------
//--- set members that are modified during reset
//--------------------------------------------------------------------------

// sets cntDelay=0, resync. status to True, calls Init() of base class
void COP::Init ()
{
    ACClockSyncAlgorithm::Init();

    cntDelay = 0;  // no delay measurements
    SetResynchronizationStatus (True);  // own time is not valid
}

