//
// algrate.cpp
//
// implementation of the "clock rate synchronization algorithm"
//
// created: 07.10.98, Klaus Schossmaier
//
// last change: 23.11.98, Klaus Schossmaier
//
// last revision: 04.02.99
//
// RCS version control information:
//     $Revision: 1.13 $$State: Exp $$Date: 1999/02/04 14:24:33 $
//

#include <fstream.h>
#include <string.h>   // for memmove()

#include "address.hpp"
#include "addrman.hpp"
#include "alggen.hpp"
#include "algrate.hpp"
#include "clockifc.hpp"
#include "constant.hpp"
#include "database.hpp"
#include "error.hpp"
#include "time.hpp"


// enable terminal output during a run
#define RUN_SETTING
//#define RUN_RATE

// enable output files for gnuplot
//#define GNU_ACC
//#define GNU_PRE
//#define GNU_DRIFT
//#define GNU_CONS

// file names
#define NAME_ACC   "acc.dat"
#define NAME_CONS  "cons.dat"
#define NAME_DRIFT "drift.dat"
#define NAME_PRE   "pre.dat"

// provide some macros for closing these files
#ifdef GNU_ACC
  #define CloseAcc() (Facc.close())
#else
  #define CloseAcc() ((void)0)
#endif

#ifdef GNU_PRE
  #define ClosePre() (Fpre.close())
#else
  #define ClosePre() ((void)0)
#endif

#ifdef GNU_CONS
  #define CloseCons() (Fcons.close())
#else
  #define CloseCons() ((void)0)
#endif

#ifdef GNU_DRIFT
  #define CloseDrift() (Fdrift.close())
#else
  #define CloseDrift() ((void)0)
#endif


//**************************************************************************
// module global constants
//**************************************************************************

// --- default algorithmic parameters
// when undefined they are computed
#define PI_H        220e-6
#define PI_O        220e-6
#define UPSILON_MAX 300e-6 
#define SIGMA_MAX   0.01e-6 // 0.01e-6
#define GAMMA_H     4e-6
#define GAMMA_O     4e-6
#define THETA_MAX   10e-6 

// --- rate interval boundaries
#define RATE_MAX 2.0
#define RATE_MIN 0.0

// ---- initial accuracies of the clocks
// a delayed start could render them as not correct!
const double accPlusInitial  = 50e-6;
const double accMinusInitial = 50e-6;
 
// --- initial rate drifts of the clocks
// a delayed start could render them as not correct!
const double thetaPlusInitial  = 2e-6;
const double thetaMinusInitial = 2e-6;

// --- initial values for EpsilonPlus/Minus
// for the computation of the precision intervals 
// they can be refined after some delay measurements
const double epsilonPlusInitial  = 10e-6;
const double epsilonMinusInitial = 10e-6;

// --- number of delay measurements 
// before the algorithm starts using the delay data in StoreTime()
const int StartSynchronization = 14;

extern char *ftoa (double f, int width, char *string);  // from ftoa.cpp
// returns 'f' as a string with at least 'width' characters in %e format;
// stores the string in 'string', which must be big enough to hold the
// converted string; if 'string' is NULL, NULL is returned


//**************************************************************************
// module global functions
//**************************************************************************

static INT16 ThetaToLambda (double theta, double fnom);
// convert computed rate drift for a LAMBDA register

static void PlotInterval (ostream& dataStream, double x,
                          const SAsymInterval& In);
// write an (asym)interval into an .dat file for GNUPLOT


//**************************************************************************
// class CRate: features the Rate algorithm
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assert: cs != NULL (done by the ctor of ACClockSyncAlgorithm)
// for initialization beware the order: 
// 1st the deterioration, 2nd the accuracy
CRate::CRate (CClockSync *cs) : ACClockSyncAlgorithm (cs)
{
    cntDelay = 0;  
    // no delay measurements

    SetResynchronizationStatus (True);  
    // own time is not valid
   
    cout.setf (ios::fixed, ios::floatfield);

    rateRound = False;
    // initial no rate round

    GetOwnID (nodeID, area);
    // determine the ID and store it in the member data

    ClockChanged();  // initialize clock dependent data
}


//--------------------------------------------------------------------------
//--- add time intervals
//--------------------------------------------------------------------------

// if CS:resync is True, the time is not stored; so the algorithm
// only starts working after resync is False; then the time is
// passed on to StoreTime() of the base class
EBool CRate::StoreTime (const STimestamp& time, double resync,
        SINT32 networkID, const CAddr& netAddress, int area, EBool gps)
{
    if (GetResynchronizationStatus ()) 
        return False;
    // do not store the time as long as CS is resynchronizing

    return ACClockSyncAlgorithm::StoreTime (time, resync, networkID,
        netAddress, area, gps);
}


//--------------------------------------------------------------------------
//--- add transmission delay intervals
//--------------------------------------------------------------------------

// counts to N before setting CS::resync to False, calls StoreDelay
// of base class to process delays
EBool CRate::StoreDelay (double localDelay, double remoteDelay,
     const SRate& remoteRate, SINT32 networkID, const CAddr& netAddress)
{    
    if (cntDelay >= 0) cntDelay++;

    if (cntDelay == StartSynchronization)
    {
        SetResynchronizationStatus (False);
        // the own timestamp is valid

        cntDelay = -1;  // stop counting the delay measurement rounds
    }
    cntDelay++;

    if (cntDelay == StartSynchronization)
    {
        SetResynchronizationStatus (False);
    }
    // the own timestamp is valid

    return ACClockSyncAlgorithm::StoreDelay (localDelay, remoteDelay,
        remoteRate, networkID, netAddress);
}


//--------------------------------------------------------------------------
//--- add rate intervals
//--------------------------------------------------------------------------

// called when rate intervals are arriving my node
// do the preprocessing and fill the database with remote rate intervals 
EBool CRate::StoreRate (const SRate& remoteRate, const CNtpTime& localTime,
    const CNtpTime& remoteTime, SINT32 networkID, const CAddr& netAddress, 
    int remoteArea)
{
    SINT32 remoteID;           // nodeID of peer node
    SRateTime *rateData;       // pointer to the rate interval
    SRate evalRate;            // rate interval for EvalSys 
    double deltaRemoteTime;    // expired time on the remote node
    double deltaLocalTime;     // expired time on the local node
    SAsymInterval quotientInt; // quotient rate interval 
    SAsymInterval rateInt;     // input rate interval     
    double lengthInt;          // interval length, half side
    SDelay delay;              // transmission delay parameters

    database->GetAssociation (networkID, netAddress, &remoteID);
    // get ID of the remote node   

    // --- compute the quotient rate intervals

    quotientInt.ref   = 1.0;
    quotientInt.left  = 1.0;
    quotientInt.right = 1.0;
    // initialize the quotient rate interval

    rateData = (SRateTime *) database->GetRate (networkID, netAddress, 
        remoteArea);
    // get rate data from last rate round, if any

    if (rateData != NULL)   // there was a previous rate round
    {
        // --- compute the refpoint of the quotient rate interval

        deltaRemoteTime = 
            NtpToSeconds (remoteTime) - NtpToSeconds (rateData->remoteTime);
        // expired logical time on remote clock

        deltaLocalTime = 
            NtpToSeconds (localTime) - NtpToSeconds (rateData->localTime);
        // expired logical time on local clock
   
        quotientInt.ref = (deltaRemoteTime - remoteRate.u) / 
            (deltaLocalTime - rate->u);
       // take the ratio

        Assert (quotientInt.ref > RATE_MIN);
        Assert (quotientInt.ref < RATE_MAX);
        // relative rate measurement not useful

        // --- compute the lengths of the quotient rate interval

        if (!database->GetDelay (delay, networkID, netAddress))
        {
            return False; 
            // alternatively use initial values!
        }
        // get transmission delay uncertainty

        lengthInt = sigma * (deltaRemoteTime - remoteRate.u) + 
            (delay.epsilonMinus + delay.epsilonPlus) / 
            (deltaLocalTime - rate->u);
        // length of each side

        quotientInt.left  = quotientInt.ref - lengthInt;
        quotientInt.right = quotientInt.ref + lengthInt;
        // set the quotient rate interval symmetricaly

        Assert (quotientInt.left > RATE_MIN);
        Assert (quotientInt.right < RATE_MAX);
        // relative rate measurement not useful

        rateRound = True;
        // informs the algorithm that this is a rate round
        // but not the very first time
    }
    else
    {
        rateData = new SRateTime;
        database->SetRate (rateData, networkID, netAddress, remoteArea);
        // when there is no former entry, put pointer into the database
    }

    // --- compute remote rate intervals
    // note that the very first one is not used

    if ((nodeID == remoteID) && (area == remoteArea))  
    {
        // this is just the local rate interval
        rateInt.ref   = 1.0;
        rateInt.left  = 1.0 - remoteRate.thetaMinus;
        rateInt.right = 1.0 + remoteRate.thetaPlus;
    }
    else
    {
        rateInt.ref = quotientInt.ref;
        // take over the reference point 

        rateInt.left = quotientInt.left * (1 - remoteRate.thetaMinus) -
            2 * sigma * ( GetDelayCompensation () + 
                GetComputationDelayCompensation () ); 
        // left edge of remote rate interval

        rateInt.right = quotientInt.right * (1 + remoteRate.thetaPlus) +
            2 * sigma * ( GetDelayCompensation () + 
                GetComputationDelayCompensation () ); 
        // right edge of remote rate interval
    } 
 
    Assert (rateInt.left  > RATE_MIN);
    Assert (rateInt.right < RATE_MAX);
    // relative rate measurement or remote rate interval not useful
 
    // --- save the current rate data in the database; has two functions: 
    // 1) remember timestamps for the next rate synchronization
    // 2) hand over rate interval to the CV

    rateData->thetaPlus = rateInt.right - rateInt.ref;
    rateData->thetaMinus = rateInt.ref - rateInt.left;
    rateData->u = remoteRate.u;
    rateData->ref = rateInt.ref;
    rateData->localTime = localTime;
    rateData->remoteTime = remoteTime;
    // compose the entry

    evalRate.thetaPlus = rateData->thetaPlus;            
    evalRate.thetaMinus = rateData->thetaMinus;
    evalRate.u = remoteRate.u;
    ReportRate (True, evalRate, quotientInt.ref);        
    // report rate interval to EvalSys, as input 

    return True;
}


//--------------------------------------------------------------------------
//--- manage clock changes
//--------------------------------------------------------------------------

// computes initialization data (clock deterioration, accuracy) for
// the clock, and some clock dependent parameters needed by the algorithm
void CRate::ClockChanged()
{
    CClockInterface *clock;        // access to the clock
    double fnom;                   // oscillator frequency
    STimestamp state;              // current clock state
    SExtendedTimestamp state0;     // clock state with desired accuray
    double pi0Minus, pi0Plus;      // initial precisions
    INT16 lambdaMinus, lambdaPlus; // deteriorations of accuracies 


    ACClockSyncAlgorithm::ClockChanged();  // initialize the local rate

    clock = GetClock();
    if (clock == NULL) return;
    // the clock is not available, so we cannot initialize it
    // note: we could report that the clock is NULL, since the user may not
    // have intended this; but the report might come into the way of object
    // destruction


    #ifdef RUN_SETTING
        cout << "----- Setting of node " << nodeID << " -----" << endl;
    #endif

    // --- initial rate interval R_init = [thetaMinusInitial,thetaPlusInitial]

    rate->thetaMinus = thetaMinusInitial;  
    rate->thetaPlus  = thetaPlusInitial;  
    // set the initial local rate interval

    #ifdef RUN_SETTING
        cout << "R_init   = [" 
             << rate->thetaMinus * 1e6 << ", " 
             << rate->thetaPlus * 1e6 << "] ppm" << endl;
    #endif

    fnom = clock->GetFrequency ();
    lambdaMinus = ThetaToLambda (thetaMinusInitial, fnom);
    lambdaPlus  = ThetaToLambda (thetaPlusInitial, fnom);
    clock->SetLambdapure (lambdaMinus, lambdaPlus); 
    // set the initial deterioration according to R_init  

    // --- initial accuracy interval A_init = [accMinusInitial,accPlusInitial]

    state0 = IntervalToClock (0, accMinusInitial, accPlusInitial);
    // compute the accuracy values, reference point is unused

    clock->GetTimestamp (state);
    // get the current clock state

    state.acc = state0.acc;  
    clock->SetTimestamp (state);
    // set the initial accuracies

    #ifdef RUN_SETTING
        cout << "A_init   = [" 
             << accMinusInitial << ", " << accPlusInitial << "] s" << endl;
    #endif

    // --- precision interval pi_0 = [pi0Minus,pi0Plus]

    pi0Minus = (epsilonPlusInitial + epsilonMinusInitial) + 
          4 * clock->GetGranularity () + 
          0.5 * (GetRoundPeriod () + GetDelayCompensation () + 
                 2 * GetComputationDelayCompensation ())  
              * (clock->GetRhoPlus () + clock->GetRhoMinus ()) + 
          0.5 * GetRoundPeriod () 
              * (clock->GetRhoPlus () - clock->GetRhoMinus ());

    pi0Plus = (epsilonPlusInitial + epsilonMinusInitial) + 
          4 * clock->GetGranularity () + 
          0.5 * (GetRoundPeriod () + GetDelayCompensation () + 
                 2 * GetComputationDelayCompensation ())   
              * (clock->GetRhoPlus () + clock->GetRhoMinus ()) - 
          0.5 * GetRoundPeriod () 
              * (clock->GetRhoPlus () - clock->GetRhoMinus ());

    // --- precision interval pi^H = [piHMinus,piHPlus]

    #ifdef PI_H
        piHMinus = piHPlus = PI_H;  
    #else
        piHMinus = pi0Minus + epsilonMinusInitial + 
            3 * clock->GetGranularity () + 
            (GetRoundPeriod () + GetComputationDelayCompensation ()) 
            * clock->GetRhoMinus ();

        piHPlus = pi0Plus + epsilonPlusInitial + 
            2 * clock->GetGranularity () +
            (GetRoundPeriod () + GetComputationDelayCompensation ()) 
            * clock->GetRhoPlus ();
    #endif

    #ifdef RUN_SETTING
        cout << "piH      = [" 
             << piHMinus << ", " << piHPlus << "] s" << endl;
    #endif

    // --- precision interval pi^o = [piOwnMinus,piOwnPlus]

    #ifdef PI_O
        piOwnMinus = piOwnPlus = PI_O;
    #else
        piOwnMinus = pi0Minus + clock->GetGranularity () +
              GetRoundPeriod () * clock->GetRhoMinus ();

        piOwnPlus = pi0Plus + clock->GetGranularity () +
              GetRoundPeriod () * clock->GetRhoPlus ();
    #endif

    #ifdef RUN_SETTING
        cout << "piOwn    = [" 
             << piOwnMinus << ", " << piOwnPlus << "] s" << endl;
    #endif

    // --- maximal state correction Upsilon_max:
 
    #ifdef UPSILON_MAX
        ups = UPSILON_MAX;
    #else
        ups = 2 * clock->GetGranularity () +
            GetRoundPeriod () * (clock->GetRhoPlus () + clock->GetRhoMinus ());
    #endif

    #ifdef RUN_SETTING
        cout << "ups      = " << ups << " s " << endl;
    #endif

    // --- oscillator stability

    sigma = SIGMA_MAX;  // no calculation, no measurement

    #ifdef RUN_SETTING
        cout << "sigma    = " << sigma * 1e6 << " ppm/s " << endl;    
    #endif

    // --- consonance interval gamma_H = [gammaHMinus,gammaHPlus]

    #ifdef GAMMA_H
        gammaHMinus = gammaHPlus = GAMMA_H;
    #else
        gammaHMinus = gammaHPlus = sigma * ( 3 * GetRatePeriod () + 
            10 * GetDelayCompensation () + 
            6 * GetComputationDelayCompensation () ) + 
            3 * (epsilonPlusInitial + epsilonMinusInitial) / 
                ( GetRatePeriod () - GetDelayCompensation () - 
                epsilonPlusInitial - epsilonMinusInitial );
    #endif

    #ifdef RUN_SETTING
        cout << "gammaH   = [" 
             << gammaHMinus * 1e6 << ", " << gammaHPlus * 1e6 << "] ppm " 
             << endl;
    #endif

    // --- consonance interval gamma_O = [gammaOwnMinus,gammaOwnPlus]

    #ifdef GAMMA_O
        gammaOwnMinus = gammaOwnPlus = GAMMA_O;
    #else
        gammaOwnMinus = gammaOwnPlus = sigma * ( 3 * GetRatePeriod () + 
            7 * GetDelayCompensation () + 
            4 * GetComputationDelayCompensation () ) + 
            2 * (epsilonPlusInitial + epsilonMinusInitial) / 
                ( GetRatePeriod () - GetDelayCompensation () - 
                epsilonPlusInitial - epsilonMinusInitial );
    #endif

    #ifdef RUN_SETTING
        cout << "gammaOwn = [" 
             << gammaOwnMinus * 1e6 << ", " << gammaOwnPlus * 1e6 << "] ppm " 
             << endl;
    #endif

    // --- maximal rate correction Theta_max

    #ifdef THETA_MAX
        thetaMax = THETA_MAX;
    #else
        thetaMax = 2 * sigma * GetRatePeriod ();       
    #endif

    #ifdef RUN_SETTING
        cout << "thetaMax = " << thetaMax * 1e6 << " ppm " << endl;
    #endif

    cout << endl;    
}


//--------------------------------------------------------------------------
//--- create function of class CRate
//--------------------------------------------------------------------------

// creates and returns a new object of type CRate
ACClockSyncAlgorithm *CRate::CreateRate (CClockSync *cs)
{
    return new CRate (cs);
}


//--------------------------------------------------------------------------
//--- compute and set the new clock interval and new rate interval
//--------------------------------------------------------------------------

// first synchronize the rate and then the state, otherwise
// the accumulation of the state adjustments is not correct
EBool CRate::ConvergenceFunction ()
{        
    CClockInterface *clock;          // access to the clock
    double fnom;                     // oscillator frequency
    CNtpTime point;                  // new reference point
    double alpha_minus, alpha_plus;  // new accuracies
    double theta_minus, theta_plus;  // new rate drifts
    double multStep;                 // factor for rate adjustment
    double stepNew;                  // new coupling factor
    INT32 step;                      // UTCSU step register 
    INT16 lambdaMinus, lambdaPlus;   // UTCSU deterioration registers 
    STimeInterval evalState;         // state interval for EvalSys
    SRate evalRate;                  // rate interval for EvalSys 
 
    timeValid = False;  
    // first of all mark the time as invalid

    if (GetAssociatedNetwork() == NoAccount) 
        return False;
    // no associated network

    clock = GetClock();
    Assert (clock != NULL);  
    // clock must exist

    // --- synchronize the clock rate

    if (rateRound) // this a rate round as well 
    {
        if (!Rate (multStep, theta_minus, theta_plus, GetAssociatedNetwork())) 
            return False;
        // Rate() function failed

        step = clock->GetSteppure();
        // the current coupling factor

        stepNew = step * multStep;
        // new coupling factor by multiplication 

        Assert (stepNew <= 0xFFFFFFFF);
        // coupling factor is not useful

        clock->SetSteppure ((INT32)(stepNew));
        // change the clock rate 

        rate->thetaMinus = theta_minus;
        rate->thetaPlus  = theta_plus;
        // set new local rate interval, without deterioration

        rateRound = False;
        // next round is not for rate

        evalRate.thetaPlus = theta_plus;   
        evalRate.thetaMinus = theta_minus;
        evalRate.u = 0.0;
        ReportRate (False, evalRate, multStep);        
        // report new rate interval to EvalSys, as output   
    } // if

    rate->thetaMinus += (sigma * GetRoundPeriod ());
    rate->thetaPlus  += (sigma * GetRoundPeriod ());
    // deteriorate local rate interval for the next state round

    fnom = clock->GetFrequency ();
    lambdaMinus = ThetaToLambda (rate->thetaMinus, fnom);
    lambdaPlus  = ThetaToLambda (rate->thetaPlus, fnom);
    clock->SetLambdapure (lambdaMinus, lambdaPlus); 
    // set the new accuracy interval deteriorations

    // --- synchronize the clock state

    if (!State (point, alpha_minus, alpha_plus, GetAssociatedNetwork())) 
        return False;
    // State() function failed

    *newClockValue = IntervalToClock (point, alpha_minus, alpha_plus);
    // set new clock interval

    evalState.refpoint = point; 
    evalState.low  = point - CNtpTime (alpha_minus);
    evalState.high = point + CNtpTime (alpha_plus);
    ReportInterval (False, evalState);            
    // report new state interval to EvalSys, as output           

    timeValid = True;  
    // the new timestamp is valid
    
    return True;
}


//--------------------------------------------------------------------------
//--- state algorithm
//--------------------------------------------------------------------------

// computes the new clock interval for the network; 
// returns True if the computation was successful, False otherwise
EBool CRate::State (CNtpTime& point, double& alpha_minus, double& alpha_plus,
    SINT32 networkID)
{
    STimeInterval data; // read out the database of state intervals 
    EBool entryExists;  // is there a state interval in the database
    INT32 sizeState;    // number of state intervals in the database
    int i; 
    CAddr netAddress;
    SAsymInterval *accuracyIntervals;  // set of accuracy intervals
    SAsymInterval *precisionIntervals; // set of precision intervals
    SAsymInterval preIntMarz;          // precision interval after Marzullo
    SAsymInterval preIntResult;        // precision interval towards final 
    SAsymInterval accIntMarz;          // accuracy interval after Marzullo
    SAsymInterval accIntResult;        // accuracy interval towards final
    double adjustment;
    filebuf Facc, Fpre; // write intervals to .dat files for gnuplot

    // --- enough state intervals

    if (numberOfNodes <= 3*numberOfFaultyNodes) // n > 3f is needed for OP
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }

    sizeState = database->Cardinality (CApplicationDatabase::Time, networkID);
    if (sizeState <= 2*numberOfFaultyNodes)
    // need at least 2*f + 1 timestamps
    {
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }

    accuracyIntervals = new SAsymInterval[sizeState];
    precisionIntervals = new SAsymInterval[sizeState];
    // these should be the input state intervals represented in doubles

    // --- get own accuracy/precision interval

    if (!database->GetAssociation (networkID, nodeID, netAddress)) 
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }    
    // obtain netAddress

    if (!database->RemoveTime (data, networkID, netAddress, area))
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }
    // obtain own accuracy interval

    accuracyIntervals[0].left  = NtpToSeconds (data.low);
    accuracyIntervals[0].ref   = NtpToSeconds (data.refpoint);
    accuracyIntervals[0].right = NtpToSeconds (data.high);
    // own accuracy interval represented in doubles, index = 0

    precisionIntervals[0].left  = accuracyIntervals[0].ref - piOwnMinus;
    precisionIntervals[0].ref   = accuracyIntervals[0].ref; 
    precisionIntervals[0].right = accuracyIntervals[0].ref + piOwnPlus;
    // own precision interval represented in doubles, index = 0

    // --- get remote accuracy/precision intervals

    for (entryExists = database->GetFirstTime (data, networkID), i = 1;  
         entryExists;
         entryExists = database->GetNextTime (data, networkID), i++)
    {
         accuracyIntervals[i].left  = NtpToSeconds (data.low);
         accuracyIntervals[i].ref   = NtpToSeconds (data.refpoint);
         accuracyIntervals[i].right = NtpToSeconds (data.high);
         // preprocessed accuracy interval

         precisionIntervals[i].left  = accuracyIntervals[i].ref - piHMinus;
         precisionIntervals[i].ref   = accuracyIntervals[i].ref; 
         precisionIntervals[i].right = accuracyIntervals[i].ref + piHPlus;
         // computed precision interval
    } // for

    // now we have the accuracy and precision intervals at hand
    // in accuracyIntervals[] and precisionIntervals[] respectively,
    // where in [0] are the own ones

#ifdef GNU_PRE
    if (Fpre.open(NAME_PRE,output) == 0)
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        return False;
    }
    ostream pre(&Fpre); 
    for(i=0;i<sizeState;i++)
        PlotInterval (pre, i, precisionIntervals[i]);
#endif
    
    // --- apply convergence function OP upon the precision intervals
 
    if (!ApplyMarzullo (precisionIntervals, preIntMarz, sizeState, 
        numberOfFaultyNodes))
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        ClosePre();
        return False;
    }
    // Marzullo's function upon associated precision intervals

#ifdef GNU_PRE
    PlotInterval (pre, 6, preIntMarz);
#endif

    if (!Intersect (preIntMarz, precisionIntervals[0], preIntResult))
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        ClosePre();
        return False;
    }
    // intersection with local precision interval

    preIntResult.ref =
        (piOwnMinus * preIntResult.right + piOwnPlus * preIntResult.left) /
        (piOwnMinus + piOwnPlus);
    // candidat reference point

#ifdef GNU_PRE
    PlotInterval (pre, 7, preIntResult);
#endif

    adjustment = preIntResult.ref - precisionIntervals[0].ref;
 
    if (adjustment > ups)
        preIntResult.ref = precisionIntervals[0].ref + ups;
    if (adjustment < - ups)
        preIntResult.ref = precisionIntervals[0].ref - ups;
    // new reference point 

    rate->u += (preIntResult.ref - precisionIntervals[0].ref);
    // accumulate state adjustment for relative rate measurement

#ifdef GNU_PRE
    PlotInterval (pre, 8, preIntResult);
#endif

    // --- apply convergence function OP upon the accuracy intervals

#ifdef GNU_ACC
    if (Facc.open(NAME_ACC,output) == 0)
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        ClosePre();
        return False;
    }         
    ostream acc(&Facc); 
    for(i=0;i<sizeState;i++)
        PlotInterval (acc, i, accuracyIntervals[i]);
#endif

    if (!ApplyMarzullo (accuracyIntervals, accIntMarz, sizeState, 
        numberOfFaultyNodes))
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        ClosePre();
        CloseAcc();
        return False;
    }    
    // Marzullo's function upon accuracy intervals

#ifdef GNU_ACC
    PlotInterval (acc, 6, accIntMarz);
#endif

    if (!Intersect (accIntMarz, accuracyIntervals[0], accIntResult))
    {
        delete[] accuracyIntervals;
        delete[] precisionIntervals;
        database->DeleteAll (CApplicationDatabase::Time, networkID);
        ClosePre();
        CloseAcc();
        return False;
    }     
    // intersection with own accuracy interval

#ifdef GNU_ACC
    PlotInterval (acc, 7, accIntResult);
#endif

    if (preIntResult.ref > accIntResult.right)
        accIntResult.right = preIntResult.ref;
    if (preIntResult.ref < accIntResult.left)
        accIntResult.left = preIntResult.ref;
    // new edges of accuracy interval

#ifdef GNU_ACC
    PlotInterval (acc, 8, accIntResult);
#endif

    // --- compute the resulting values
 
    SecondsToNtp (preIntResult.ref, point.ms, point.ts, point.us);
    // hand back the new reference point, from precision intervals

    alpha_minus = preIntResult.ref - accIntResult.left;
    Assert (alpha_minus >= 0);
    // hand back the new left edge

    alpha_plus = accIntResult.right - preIntResult.ref;
    Assert (alpha_plus >= 0);
    // hand back the new right edge
    
    delete[] accuracyIntervals;
    delete[] precisionIntervals;
    database->DeleteAll (CApplicationDatabase::Time, networkID);
    // the data is not needed anymore

    ClosePre();
    CloseAcc();
    // close the debug files if necessary

    return True;
}


//--------------------------------------------------------------------------
//--- rate algorithm
//--------------------------------------------------------------------------

// compute the new coupling factor (for STEP register) and 
// local rate interval (for LAMBDA registers)
// returns True if the computation was successful, False otherwise
EBool CRate::Rate (double& multStep, double& theta_minus, double& theta_plus, 
    SINT32 networkID)
{
    SRate *rateEntryExists; // pointer to rate intervals in the database
    SRate *rateEntryOwn;    // pointer to own rate interval in the database
    INT32 sizeRate;         // number of rate intervals in the database
    int i;
    CAddr netAddress;
    SAsymInterval *rateIntervals;       // set of rate intervals
    SAsymInterval *consonanceIntervals; // set of consonance intervals
    SAsymInterval consIntMarz;          // consonance intervals after Marzullo
    SAsymInterval consIntResult;        // consonance intervals towards final
    SAsymInterval rateIntMarz;          // rate intervals after Marzullo
    SAsymInterval rateIntResult;        // rate interval towards final
    filebuf Fcons, Fdrift; // write intervals to .dat files for GNUPLOT

    rate->u = 0;
    // clear state adjustments for next relative rate measurement

    // --- enough rate intervals

    if (numberOfNodes <= 3*numberOfFaultyNodes) // n > 3f is needed for OP
    {
        // database->DeleteAll (CApplicationDatabase::Rate, networkID);
        return False;
    }

    sizeRate = database->Cardinality (CApplicationDatabase::Rate, networkID);
    if (sizeRate <= 2*numberOfFaultyNodes)
    // need at least 2*f + 1 timestamps
    {
        // database->DeleteAll (CApplicationDatabase::Rate, networkID);
        return False;
    }

    rateIntervals = new SAsymInterval[sizeRate];
    consonanceIntervals = new SAsymInterval[sizeRate]; 
    // these should be the input rate intervals represented in doubles

    // --- get own rate/consonance interval

    if (!database->GetAssociation (networkID, nodeID, netAddress)) 
    {
        delete[] rateIntervals;
        delete[] consonanceIntervals;
        // database->DeleteAll (CApplicationDatabase::Rate, networkID);
        return False;
    }    
    // obtain netAddress

    rateEntryOwn = database->RemoveRate (networkID, netAddress, area);
    if (rateEntryOwn == NULL)
    {
        delete[] rateIntervals;
        delete[] consonanceIntervals;
        // database->DeleteAll (CApplicationDatabase::Rate, networkID);
        return False;
    }        
    // obtain own rate interval

    rateIntervals[0].ref   = ((SRateTime *)rateEntryOwn)->ref;
    rateIntervals[0].left  = rateIntervals[0].ref - rateEntryOwn->thetaMinus;
    rateIntervals[0].right = rateIntervals[0].ref + rateEntryOwn->thetaPlus;
    // own rate interval represented in  doubles, index = 0

    consonanceIntervals[0].left  = rateIntervals[0].ref - gammaOwnMinus;
    consonanceIntervals[0].ref   = rateIntervals[0].ref;
    consonanceIntervals[0].right = rateIntervals[0].ref + gammaOwnPlus;
    // own consonance interval represented in doubles, index = 0

    // --- get remote rate/consonance intervals

    for (rateEntryExists = database->GetFirstRate (networkID), i = 1;  
         rateEntryExists != NULL;
         rateEntryExists = database->GetNextRate (networkID), i++)
    {
        rateIntervals[i].ref = ((SRateTime *)rateEntryExists)->ref;  
        rateIntervals[i].left = rateIntervals[i].ref - 
            rateEntryExists->thetaMinus;
        rateIntervals[i].right = rateIntervals[i].ref + 
            rateEntryExists->thetaPlus;
        // preprocessed rate interval

        consonanceIntervals[i].left  = rateIntervals[i].ref - gammaHMinus;
        consonanceIntervals[i].ref   = rateIntervals[i].ref;
        consonanceIntervals[i].right = rateIntervals[i].ref + gammaHPlus;
        // computed consonance interval
    } // for

    database->SetRate (rateEntryOwn, networkID, netAddress, area);
    // put the own rate interval back into the database
    // necessary to carry out the relative rate measurement

    // now we have the rate and consonance intervals at hand
    // in rateIntervals[] and consonanceIntervals[] respectively,
    // where in [0] are the own ones

    // --- apply convergence function OP upon the consonance intervals

#ifdef GNU_CONS
    if (Fcons.open(NAME_CONS,output) == 0)
    {
        delete[] consonanceIntervals;
        delete[] rateIntervals;
        return False;
    }
    ostream cons(&Fcons);   
    for(i=0;i<sizeRate;i++)
        PlotInterval (cons, i, consonanceIntervals[i]);
#endif

    if (!ApplyMarzullo (consonanceIntervals, consIntMarz, sizeRate, 
                        numberOfFaultyNodes))
    {
        delete[] consonanceIntervals;
        delete[] rateIntervals;
        CloseCons();
        return False;
    }
    // Marzullo's function upon the consonance intervals

#ifdef GNU_CONS
    PlotInterval (cons, 6, consIntMarz);
#endif

    if (!Intersect (consIntMarz, consonanceIntervals[0], consIntResult))
    {
        delete[] consonanceIntervals;
        delete[] rateIntervals;
        CloseCons();
        return False;
    } 
    // intersection with own precision interval 

    consIntResult.ref =
        (gammaOwnMinus * consIntResult.right + 
            gammaOwnPlus * consIntResult.left) /
        (gammaOwnMinus + gammaOwnPlus);
    // candidat reference point

#ifdef GNU_CONS
    PlotInterval (cons, 7, consIntResult);
#endif
 
    if ((consIntResult.ref - 1) > thetaMax)
        consIntResult.ref = 1.0 + thetaMax;
    if ((consIntResult.ref - 1) < - thetaMax)
        consIntResult.ref = 1.0 - thetaMax;
    // new reference point 

#ifdef GNU_CONS
    PlotInterval (cons, 8, consIntResult);
#endif

    multStep = consIntResult.ref;
    // hand back the relative rate change of the local clock

    // --- apply convergence function OP upon the rate intervals

#ifdef GNU_DRIFT
    if (Fdrift.open(NAME_DRIFT,output) == 0)
    {
        delete[] consonanceIntervals;
        delete[] rateIntervals;
        CloseCons();
        return False;
    }
    ostream drift(&Fdrift);   
    for(i=0;i<sizeRate;i++)
        PlotInterval (drift, i, rateIntervals[i]);
#endif

    if (!ApplyMarzullo (rateIntervals, rateIntMarz, sizeRate, 
                        numberOfFaultyNodes))
    {
        delete[] consonanceIntervals;
        delete[] rateIntervals;
        CloseCons();
        CloseDrift();
        return False;
    }
    // Marzullo's function upon the rate intervals

#ifdef GNU_DRIFT
    PlotInterval (drift, 6, rateIntMarz);
#endif

    if (!Intersect (rateIntMarz, rateIntervals[0], rateIntResult))
    {
        delete[] consonanceIntervals;
        delete[] rateIntervals;
        CloseCons();
        CloseDrift();
        return False;
    } 
    // intersection with own rate interval 

#ifdef GNU_DRIFT
    PlotInterval (drift, 7, rateIntResult);
#endif

    if (consIntResult.ref > rateIntResult.right)
        rateIntResult.right = consIntResult.ref;
    if (consIntResult.ref < rateIntResult.left)
        rateIntResult.left = consIntResult.ref;
    // new edges of rate interval

#ifdef GNU_DRIFT
    PlotInterval (drift, 8, rateIntResult);  
#endif

    // --- compute the resulting values

    Assert (consIntResult.ref > 0);

    theta_minus = (consIntResult.ref - rateIntResult.left)/consIntResult.ref;
    // normalize the negative rate drift with the consonance refpoint 
    // no deterioration at here

    Assert (theta_minus >= 0);
    Assert (theta_minus <= 1);
    // negative rate drift is not useful

    theta_plus = (rateIntResult.right - consIntResult.ref)/consIntResult.ref;
    // normalize the positive rate drift with the consonance refpoint 
    // no deterioration at here

    Assert (theta_plus >= 0);
    // positive rate drift is not useful

#ifdef RUN_RATE
    cout << "drift rates for node " << nodeID << ": " 
         << theta_minus * 1e6 << ", " << theta_plus * 1e6<< endl;
#endif

    delete[] rateIntervals;
    delete[] consonanceIntervals;
    // the data is not needed anymore

    CloseCons();
    CloseDrift();
    // close the debug files if necessary

    return True;
}


//--------------------------------------------------------------------------
//--- set members that are modified during reset
//--------------------------------------------------------------------------

// sets cntDelay=0, resync. status to True, calls Init() of base class
void CRate::Init ()
{
    ACClockSyncAlgorithm::Init();

    cntDelay = 0;  // no delay measurements
    SetResynchronizationStatus (True);  // own time is not valid

    rateRound = False;  // initial no rate round
}



//**************************************************************************
// module global functions
//**************************************************************************

//--------------------------------------------------------------------------
//--- convert computed rate drift for a LAMBDA register 
//--------------------------------------------------------------------------
    
// for fnom = 25Mhz
// theta = 0.1e-6 yields 0x0009
// theta =   1e-6 yields 0x005A
// theta =  10e-6 yields 0x0384
// theta = 100e-6 yields 0x232F
// theta = 360e-6 yields 0x7EA9
static INT16 ThetaToLambda (double theta, double fnom)
{
    double lambda;
    CNtpTime lambdaRegister;

    Assert (theta > 0);
    Assert (fnom > 0);
    lambda = theta / fnom;
    // deterioration per tick, in double

    SecondsToNtp (lambda, 
        lambdaRegister.ms, lambdaRegister.ts, lambdaRegister.us);
    // convert to binary where .us is (-25,-56)
  
    if (lambdaRegister.ms > 0) 
        return 0x8FFF;
    if (lambdaRegister.ts > 0) 
        return 0x8FFF;
    if ((lambdaRegister.us & 0xFFF00000) > 0)
        return 0x8FFF;
    // in case of an overflow hand back the possible maximum

    return (INT16)(lambdaRegister.us >> 5);
    // need to select SIGN(-37,-51), hence >>5
}


//--------------------------------------------------------------------------
//--- write an (asym)interval into an .dat file for GNUPLOT 
//--------------------------------------------------------------------------

static void PlotInterval (ostream& dataStream, double x, 
    const SAsymInterval& In)
{
    char val[40];

    ftoa(x,14,val);
    dataStream << val << " ";
    ftoa(In.ref,14,val);
    dataStream << val << " ";
    ftoa(In.left,14,val);
    dataStream << val << " ";
    ftoa(In.right,14,val);
    dataStream << val << endl;
}

