//
// bitmask.cpp
//
// implementation of a class that provides an array of bits that can be set
// using the place of the bit (instead of the value, which is a power of
// two); useful for enumerations
//
// created: 23.01.97
//
// last change: 04.05.97
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 10:49:31 $
//


#include <string.h>

#include "bitmask.hpp"
#include "error.hpp"


//**************************************************************************
// class CBitMask: allows access to the bits specifying the place instead
//    of the value (places start with 0)
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// allocate enough memory to store aSize bits;
// assert: aSize > 0
CBitMask::CBitMask (INT32 aSize)
{
    Assert (aSize > 0);  // array size must be positive
    Allocate (aSize/8+1);  // allocate the array
}

// allocate the memory needed to store 'size' bytes of mask;
// assert: size > 0, mask != NULL
CBitMask::CBitMask (void *mask, INT32 size)
{
    Assert (size > 0);
    Assert (mask != NULL);  // make sure the information is valid

    Allocate (size);
    memcpy (bitmask_, mask, size_);
}

// copy the information in bitmask
CBitMask::CBitMask (const CBitMask& bitmask)
{
    Allocate (bitmask.size_);  // allocate the array
    memcpy (bitmask_, bitmask.bitmask_, size_);
}

CBitMask::~CBitMask ()
{
    delete[] bitmask_;  // assumes that there is always something allocated
}


//--------------------------------------------------------------------------
//--- set (store a mask) and reset (set all bits to zero)
//--------------------------------------------------------------------------

// discard all old data, store 'size' bytes of the mask;
// assert: size > 0
void CBitMask::Set (const void *mask, INT32 size)
{
    Assert (size > 0);
    Assert (mask != NULL);  // make sure the information is valid
    ReAllocate (size);  // allocate an appropriate array
    memcpy (bitmask_, mask, size_);
}


// set all bits to zero, keep enough memory to store aSize bits;
// assert: aSize > 0
void CBitMask::Reset (INT32 aSize)
{
    Assert (aSize > 0);  // array size must be positive
    ReAllocate (aSize/8+1);  // allocate new array
}


//--------------------------------------------------------------------------
//--- operators
//--------------------------------------------------------------------------

int CBitMask::operator& (const INT32 place) const
{
    return GetBit (place);
}


CBitMask& CBitMask::operator|= (const INT32 place)
{
    SetBit (place);

    return *this;
}


CBitMask& CBitMask::operator^= (const INT32 place)
{
    if (size_*8 <= place) SetBit (place);  // bit is zero
    else
    {
        bitmask_[GetIndex (place)] ^= (INT8)((1 << GetOffset (place)) & 0xFF);
        // toggle the bit
    }

    return *this;
}


//--------------------------------------------------------------------------
//--- compute the index and offset of a place
//--------------------------------------------------------------------------

// returns the array index of the INT8 in which place is located
INT32 CBitMask::GetIndex (INT32 place) const
{
    return place/8;
}

// returns the offset in the INT8 in which place is located
INT32 CBitMask::GetOffset (INT32 place) const
{
    return (place - 8*(place/8));
}


//--------------------------------------------------------------------------
//--- set/get a bit
//--------------------------------------------------------------------------

// adds the bit to the bitmask
void CBitMask::SetBit (INT32 place)
{
    if (size_*8 <= place)  // have to reallocate to store the new bit
    {
        INT8 *oldmask = bitmask_;  // keep a pointer to the old mask
        INT32 oldsize = size_;  // store the old size (for copying)

        Allocate (place/8+1);  // allocate a new array
        memcpy (bitmask_, oldmask, oldsize);  // copy the old data
        delete[] oldmask;  // delete the old array
    }

    // now the bit is in the current bitmask, just set it
    bitmask_[GetIndex (place)] |= (INT8)((1 << GetOffset (place)) & 0xFF);
}

// returns 0 if the bit is not set, otherwise != 0
int CBitMask::GetBit (INT32 place) const
{
    if (size_*8 <= place) return 0;  // bit not stored
    else return bitmask_[GetIndex (place)] &
                (INT8)((1 << GetOffset (place)) & 0xFF);
}


//--------------------------------------------------------------------------
//--- internal array management
//--------------------------------------------------------------------------

// assumes that the array is deleted, allocates a new one with size 'size',
// fills the array with 0
void CBitMask::Allocate (INT32 size)
{
    size_ = size;
    bitmask_ = new INT8[size_];
    memset (bitmask_, 0, size_);
}

// deletes the old array, allocates a new with size 'size'; stores the
// new size in size_, and fills the array with 0
void CBitMask::ReAllocate (INT32 size)
{
    delete[] bitmask_;
    Allocate (size);
}



