//
// ci.cpp
//
// implementation of the CI functions
//
// note: when porting to the hardware -> remove the object file and link
// again using the appropriate file from the driver of Ms. Richter
// note: take care not to use ci_buf_size() in modules that are also used
// in the hardware simulation!
//
// created: 02.10.96
//
// last change: 30.10.97
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 10:49:31 $
//


#include <stdlib.h>  // for NULL

#include "error.h"

#include "ci.hpp"
#include "sim_glob.hpp"
#include "idlist.hpp"


#define CI_OKAY   0


static CNodeIdList listOfInitializedNetworks;
// this is a list of if_num's for which ci_init() has been called


//--------------------------------------------------------------------------
// this function is not part of the CI; therefore it is not mentioned
// in ci.hpp and has to be declared as "extern" in every file where it
// is needed (right now (16.11.96), only SMsg in sim_gen.hpp needs it)
//--------------------------------------------------------------------------

INT32 ci_buf_size (INT16 size);
// number of bytes allocated if a buffer size of 'size' is requested in
// ci_get_tbuf(); it is the least multiple of sizeof(bufx) >= 'size'


//--------------------------------------------------------------------------
// CI function dispatcher
//--------------------------------------------------------------------------

int ci_init (int if_num, ci_fn ci_ann_addr, ci_efn ci_ann_err_addr);
int ci_get_tbuf (INT16 size, bufx **buf, int if_num, int area);
int ci_get_tenv (int mode, tenvx **tenv, int if_num, int area);
int ci_ret_tbuf (bufx *buf);
int ci_ret_tenv (tenvx *tenv);
int ci_send (int mode,tenvx *tenv,bufx *buf,INT16 size,int if_num,int area);

// calls the appropriate CI function and returns its value
// 0               ... everything okay
// INVALID_COMMAND ... invalid function code
// INVALID_POINTER ... pointer for arguments (param) was NULL
int ci_dispatcher(int functionID, ciparam *param)
{
    if (param == NULL) return INVALID_POINTER;

    switch (functionID)  // find the appropriate function
    {
        case CI_INIT:
            return ci_init (param->inip.if_num, param->inip.ci_announce_address,
                            param->inip.ci_announce_err_add);
            break;
        case CI_GTB:
            return ci_get_tbuf(param->gtbp.size, param->gtbp.buf,
                               param->gtbp.if_num, param->gtbp.area);
            break;
        case CI_GTE:
            return ci_get_tenv(param->gtep.mode, param->gtep.tenv,
                               param->gtep.if_num, param->gtep.area);
            break;
        case CI_RTB:
            return ci_ret_tbuf(param->rtbp.buf);
            break;
        case CI_RTE:
            return ci_ret_tenv(param->rtep.tenv);
            break;
        case CI_SEND:
            return ci_send(param->sendp.mode, param->sendp.tenv,
                           param->sendp.buf, param->sendp.size,
                           param->sendp.if_num, param->sendp.area);
            break;

        case CI_RRB:  // not supported in pure simulation
        case CI_RRE:
        case CI_MCON:
            return INVALID_COMMAND;
            break;

        default:  // unknown function ID
            return INVALID_COMMAND;
    } // switch

    return 0;
}


//--------------------------------------------------------------------------
// IFC initialization
//--------------------------------------------------------------------------

// return values:
// 0                 ... everything okay
// INVALID_IF_NUMBER ... the ifc does not exist
// INVALID_POINTER   ... one of the function pointers was NULL
int ci_init (int if_num, ci_fn ci_announce_address, ci_efn ci_announce_err_addr)
{
    if (ci_announce_address == NULL) return INVALID_POINTER;
    if (ci_announce_err_addr == NULL) return INVALID_POINTER;

    // see if a network with this IF-number exists
    if (listOfNetworks.Find (if_num) != NULL)  // the network exists
    {
        listOfInitializedNetworks.Insert (if_num);
        // mark the network as initialized
        return CI_OKAY;
    }
    else return INVALID_IF_NUMBER;  // no network with this ID was found
}


//--------------------------------------------------------------------------
// for computing the size of a tbuf structure
//--------------------------------------------------------------------------

// number of bytes allocated if a buffer size of 'size' is requested in
// ci_get_tbuf(); it is the least multiple of sizeof(bufx) >= 'size'
// note: this is *no* CI function! It is only needed to make computation
// of the number of bufx structures needed easier and less error prone;
// since both ci_get_tbuf and SMsg need to do such a computation, this
// function is provided.
INT32 ci_buf_size (INT16 size)
{
    if (size == 0) return 0;
    else return ((size-1)/B_SIZE + 1)*sizeof(bufx);
    // compute the number of buffers needed
}


//--------------------------------------------------------------------------
// memory management
//--------------------------------------------------------------------------

// allocates memory for bufx; return values:
// 0                 ... everything okay
// INVALID_IF_NUMBER ... ifc (or area) does not exist / is not initialized
// INVALID_POINTER   ... pointer for returning pointer was NULL
// NO_BUF_FREE       ... no buffer is available (probably all in use)
int ci_get_tbuf (INT16 size, bufx **buf, int if_num, int area)
{
    if (area >= MAXASICS) return INVALID_IF_NUMBER;

    if (!listOfInitializedNetworks.Find (if_num)) return INVALID_IF_NUMBER;
    // network has not been initialized

    if (buf == NULL) return INVALID_POINTER;

    if (size > 0)  // allocate bytes (multiple of the size of bufx)
    {
        *buf = (bufx *)new unsigned char[ci_buf_size(size)];

        if (*buf == NULL) return NO_BUF_FREE;
    }
    else *buf = NULL;

    return CI_OKAY;
}

// the function places the mode into tenv->length_field, which is an INT16,
// so let's hope that is okay
// return values:
// 0                 ... everything okay
// INVALID_IF_NUMBER ... ifc (or area) does not exist / is not initialized
// INVALID_POINTER   ... pointer for returning pointer was NULL
// NO_ENV_FREE       ... no envelope is available (probably all in use)
int ci_get_tenv (int mode, tenvx **tenv, int if_num, int area)
{
    if (area >= MAXASICS) return INVALID_IF_NUMBER;

    if (!listOfInitializedNetworks.Find (if_num)) return INVALID_IF_NUMBER;
    // network has not been initialized

    if (tenv == NULL) return INVALID_POINTER;

    *tenv = new tenvx;  // make a new structure

    if (*tenv != NULL)
    {
        (*tenv)->length_field = (INT16)mode;
        return CI_OKAY;
    }
    else return NO_ENV_FREE;
}

// note: you can return an rbuf with ci_ret_tbuf(), but never return a tbuf
// with ci_ret_rbuf()!
// return values:
// 0                   ... everything okay
// INVALID_BUF_ADDRESS ... address doesn't exist
int ci_ret_tbuf (bufx *buf)
{
    if (buf == NULL) return INVALID_BUF_ADDRESS;

    delete[] (unsigned char *)buf;  // just free the memory again

    return CI_OKAY;
}

// return values:
// 0                   ... everything okay
// INVALID_ENV_ADDRESS ... address doesn't exist
int ci_ret_tenv (tenvx *tenv)
{
    if (tenv == NULL) return INVALID_ENV_ADDRESS;

    delete tenv;

    return CI_OKAY;
}


//--------------------------------------------------------------------------
// message transmission
//--------------------------------------------------------------------------

// sends a message to the network with the appropriate if_num
// return values:
// 0                    ... everything okay
// INVALID_IF_NUMBER    ... ifc (or area) does not exist/is not initialized
// NO_ENVELOPE_ATTACHED ... pointer to the envelope was empty
int ci_send (int mode, tenvx *tenv, bufx *buf, INT16 size, int if_num,
             int area)
{
    ACNetwork *network;

    if (area >= MAXASICS) return INVALID_IF_NUMBER;

    if (!listOfInitializedNetworks.Find (if_num))
    // the network has not been initialized
    {
        return INVALID_IF_NUMBER;
    }

    if (tenv == NULL) return NO_ENVELOPE_ATTACHED;

    network = listOfNetworks.Find (if_num);  // get the network

    if (network == NULL)
    {
        return INVALID_IF_NUMBER;  // unknown network
        // note: have to check for the existance of the network, because
        // it may have been deleted since ci_init() was called
    }
    else
    {
        network->SendMsg (tenv, buf, size, area);
        // give the message to the network

        if ((mode & RET_PKB_IMM) == RET_PKB_IMM)  // return buffers at once
        {
            ci_ret_tbuf (buf);
            ci_ret_tenv (tenv);
        }
    
        return CI_OKAY;
    } // else
}

