//
// clock_sw.cpp
//
// definition of the base class used for accessing the software clock
//
// created: 19.02.97
//
// last change: 24.09.98
//
// RCS version control information:
//     $Revision: 1.7 $$State: Exp $$Date: 1998/09/24 12:34:43 $
//


#include <iostream.h> // DEBUG

#include <string.h>

#include "clock_sw.hpp"
#include "int128.hpp"
#include "interupt.hpp"
#include "irqtemp.hpp"
#include "structs.hpp"
#include "time.hpp"


// DEBUG
// print out the value, use precision (if 0, just print the value using
// system precision)
ostream& precout (ostream& s, double val, int precision = 0)
{
    static char unit[5][8];// = {"seconds", "ms", "us", "ns", "ps"};
    // the unit of the precision output
    strcpy (unit[0], "seconds");
    strcpy (unit[1], "ms");
    strcpy (unit[2], "us");
    strcpy (unit[3], "ns");
    strcpy (unit[3], "ps");

    int i;
    for (i=0; i<4 && (val < 1); i++) val = val * 1000;

    if (precision == 0) return s << val << " " << unit[i];
    else
    {
        s << (int)val << ".";
        val = (val - (int)val)*10;  // get the fractional part
        while (precision > 0) { s << (int)val; val = (val - (int)val)*10;
            precision--; }
    }
    return s << " " << unit[i];
}


static const INT32 DutyEnabled = 0x10000;  // int. enable bit in DUTYL
static const INT8 PolHWSnap = 28;  // bit in UTCCONF1
static const INT8 IntHwSnapshot = 16;  // interrupt bit of the HW snapshot
static const INT8 IntGPS1 = 32;  // interrupt bit of the GPS1PPS1
static const INT8 IntGPS2 = 33;  // interrupt bit of the GPS1PPS2
static const INT8 IntGPS3 = 34;  // interrupt bit of the GPS1PPS3
static const INT8 IntDutyB1 = 58;  // interrupt bit of duty timer B on SSU1



//**************************************************************************
// class CClockTemplate: the template class used for obtaining an instance
//      of the currently used clock (software)
//**************************************************************************

//--------------------------------------------------------------------------
//--- return a new instance of the class
//--------------------------------------------------------------------------

// returns a new object of the software clock
CClock *CClockTemplate::NewClock (double frequency, SINT32 nodeID, int area)
{
    return new CSoftwareClock (frequency, nodeID, area);
}



//**************************************************************************
// struct SInternalRegisters: internal clock registers that have to be
//      simulated in the software clock
//**************************************************************************

struct SInternalRegisters
{
    // NTPTIME
    INT32 ms;  // CS and MS [+31,+8]
    INT32 ts;  // [+7,-24]
    INT32 us;  // [-25,-56]
    INT32 ns;  // [-57,-59]
    EBool overflow;  // set to True by UpdateNtp() if an overflow occurred

    // STEP value during pure phase
    INT32 STEPPUREACT;  // [-20,-51]

    // deterioration values for the accuracies during pure phase
    INT16 LAMBDAPPUREACT;  // S:[-37,-51]
    INT16 LAMBDANPUREACT;

    INT32 ALPHAP;  // accuracy envelopes S:[-8,-38]
    INT32 ALPHAN;
    INT16 NALPHAP;  // fractional part of the accuracies [-39,-51]
    INT16 NALPHAN;  // (only 13 bits!)

    // amortization timer
    INT32 AMORTTIMER;
};


// the values of utcsuWrite_->STARTAMORT; used to determine the current
// phase (the ACU makes a difference between the first amort. cycle, where
// state correction of the envelope is done, and the remaining cycles,
// during which a step value is added)
enum EAmortizationTypes {
    NoAmortization = 0,  // not in amortization phase
    BeginAmortization,  // start of amortization phase (1st amort. cycle)
    MiddleOfAmortization  // somewhere after the first amort. cycle
};


//**************************************************************************
// module global functions
//**************************************************************************

static EClockInterrupt GetInterruptType (INT8 interruptBit);
// returns the interrupt type (INTA, INTN, INTT) of the bit; interruptBit
// is in the range of 0..63
// assert: interrupt bit is in use

static double DutyToSeconds (INT32 dutyh, INT32 dutyl);
// converts the contents of the duty timer registers to seconds
// note: dutyh contains NTP bits [+31,0], duty low contains [-1,-16] in
// bits [0,15]; the timestamp is expanded with 0 in bits [-17,-24]

static double GetTime (const CNtpTime& cycles, double frequency);
// returns the global time that corresponds to 'cycles' clock cycles of
// a clock with the frequency 'frequency'


//**************************************************************************
// class CSoftwareClock: interface to the software clock
//**************************************************************************

static const double TwoPow16 = 65536.;
static const double TwoPow21 = 2097152.;
static const double TwoPow29 = 536870912.;
static const double TwoPow30 = 1073741824.;
static const double TwoPow32 = 4294967296.;
// comma is only necessary after the last (else the result is 0 (overflow))


//--------------------------------------------------------------------------
// constructor, destructor
//--------------------------------------------------------------------------

CSoftwareClock::CSoftwareClock (double frequency, SINT32 nodeID, int area)
    : CClock (frequency, new SUtcsuRead, new SUtcsuWrite)
{
    nodeID_ = nodeID;  // store the node ID
    area_ = area;  // store the area

    internals_ = new SInternalRegisters;
    Assert (internals_ != NULL);

    CSoftwareClock::Init();

    // initialize all clock registers that always have valid values
    utcsuRead_->IDENTIFICATION = 0x49435410;  // "ICT"+0x10
    utcsuRead_->UTCSTAT1 = 0;  // status registers must be correct
    utcsuRead_->UTCSTAT2 = 0;
    utcsuRead_->UTCINTSTAT1 = 0;  // also interrupt status registers
    utcsuRead_->UTCINTSTAT2 = 0;
    utcsuRead_->UTCCONF1 = 0;  // no configuration data
    utcsuRead_->UTCCONF2 = 0;
    utcsuWrite_->UTCINTEN1 = 0;  // no interrupts are enabled
    utcsuWrite_->UTCINTEN2 = 0;
    utcsuWrite_->UTCCONF1 = 0;  // no configuration data
    utcsuWrite_->UTCCONF2 = 0;
}

CSoftwareClock::~CSoftwareClock ()
{
    Assert (utcsuWrite_->UTCINTEN1 == 0);
    Assert (utcsuWrite_->UTCINTEN2 == 0);
    // all interrupts must have been disabled at this point

    CSoftwareClock::Dispose();
    delete utcsuRead_;
    delete utcsuWrite_;
    delete internals_;
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
// known params: ClockDrift
EBool CSoftwareClock::SetParameter (const SParameter& param)
{
    switch (param.type)
    {
        case ClockDrift:
            if (((SSingleParameter*)&param)->value <= -1) return False;
            // frequencies must be positive, so clock drifts must be > -1

            Update ();
            // set the clock to its current value (using the old rate)

            fOsc_ = GetFrequency() +
                    ((SSingleParameter*)&param)->value * GetFrequency();
            // set the real oscillator frequency
            // note: don't compute f*(1+drift), since |drift| << 1

            RescheduleInterrupts ();  // reschedule any interrupts
            break;

        default:  // unknown parameter - pass on to the clock
            return CClock::SetParameter (param);
    } // switch

    return True;
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned
EBool CSoftwareClock::GetParameter (SParameter& param) const
{
    switch (param.type)
    {
        case ClockDrift:
            ((SSingleParameter*)&param)->value =
                (fOsc_-GetFrequency())/GetFrequency();
            // return the clock drift
/**/        // problems with rounding error can occur?
            break;

        default:  // unknown parameter
            return CClock::GetParameter (param);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- read access functions that have to be simulated
//--------------------------------------------------------------------------

// LTU (Local Time Unit)
INT32 CSoftwareClock::GetTSGETL()
{
    GetSTATEPGETL();  // update the clock, latch the accuracies

    GetALPHAPGETL();  // latch the accuracy envelopes
    GetALPHANGETL();

    // now latch the NTPTIME registers

    utcsuRead_->MSGET = (internals_->ms & 0x00FFFFFF);  // CS always zero
    utcsuRead_->TSGETL = internals_->ts;
    utcsuRead_->TSGET = internals_->ts;
    utcsuRead_->USGET = internals_->us;
    utcsuRead_->NSGET = internals_->ns;
    return CClock::GetTSGETL();
}

INT32 CSoftwareClock::GetAMORTTIMEGETL()
{
    Update();  // that computes the current value of the amort. timer
    utcsuRead_->AMORTTIMEGETL = internals_->AMORTTIMER;  // latch the data
    utcsuRead_->AMORTTIMEGET = internals_->AMORTTIMER;
    return CClock::GetAMORTTIMEGETL();
}


// ACU (Accuracy Unit)
INT32 CSoftwareClock::GetSTATEPGETL()
{
    Update();  // compute the current accuracy envelopes
    utcsuRead_->STATEPGET = GetAccuracy();
    utcsuRead_->STATEPGETL = utcsuRead_->STATEPGET;
    return CClock::GetSTATEPGETL();
}

INT32 CSoftwareClock::GetALPHAPGETL()
{
    Update();  // compute current envelopes
    utcsuRead_->ALPHAPGETL = internals_->ALPHAP;
    utcsuRead_->ALPHAPGET = internals_->ALPHAP;
    utcsuRead_->NALPHAPGET &= 0xFFFF0000;  // latch the fractional P part
    utcsuRead_->NALPHAPGET += (internals_->NALPHAP & 0x1FFF);
    return CClock::GetALPHAPGETL();
}

INT32 CSoftwareClock::GetALPHANGETL()
{
    Update();  // compute current envelopes
    utcsuRead_->ALPHANGETL = internals_->ALPHAN;
    utcsuRead_->ALPHANGET = internals_->ALPHAN;
    utcsuRead_->NALPHAPGET &= 0x0000FFFF;  // latch the fractional N part
    utcsuRead_->NALPHAPGET += (INT32)internals_->NALPHAN<<16 & 0x1FFF0000;
    return CClock::GetALPHANGETL();
}


// ITU (Interrupt Unit)
INT32 CSoftwareClock::GetUTCSTAT1()
{
    Update();
    // set the status registers
    if (utcsuWrite_->STARTAMORT) utcsuRead_->UTCSTAT1 |= SUtcsuRead::SCAMA;
    else utcsuRead_->UTCSTAT1 &= ~SUtcsuRead::SCAMA;

    return CClock::GetUTCSTAT1();
}

INT32 CSoftwareClock::GetUTCCONF1()
{
    utcsuRead_->UTCCONF1 = utcsuWrite_->UTCCONF1;
    return CClock::GetUTCCONF1();
}

INT32 CSoftwareClock::GetUTCCONF2()
{
    utcsuRead_->UTCCONF2 = utcsuWrite_->UTCCONF2;
    return CClock::GetUTCCONF2();
}


//--------------------------------------------------------------------------
//--- write access functions that have to be simulated
//--------------------------------------------------------------------------

// LTU (Local Time Unit)
void CSoftwareClock::SetMSSET(INT32 val)
{
    short change = (val & 0xFF000000) != (utcsuWrite_->MSSET & 0xFF000000);
    // update and rescheduling of interrupts is only necessary if STEPLOW
    // is changed

    if (change) Update();
    // STEPLOW has changed - must update so that the new STEPLOW can be
    // taken over

    CClock::SetMSSET(val);  // store the new value of MSSET

    if (change) RescheduleInterrupts();  // react to the new STEPLOW
}

void CSoftwareClock::SetNTPSET(INT32)
{
    Update();  // compute the current clock state, store the global time

    // now take over the new data (ms,ts,us) into the internal regs
    internals_->ms = (utcsuWrite_->MSSET & 0x00FFFFFF); // mask out STEPLOW
    internals_->ts = utcsuWrite_->TSSET;
    internals_->us = utcsuWrite_->USSET;
    internals_->ns = 0;

/**/// note: if the new NTPTIME is less than the old, an interrupt should
    // be raised; can't do this with the current interrupt implementation

    RescheduleInterrupts();  // react to the new clock state
}

void CSoftwareClock::SetSTEPAMORT(INT32 val)
{
    Update();
    CClock::SetSTEPAMORT (val);
    if (utcsuWrite_->STARTAMORT != NoAmortization)
    // should never be called during amortization phase (cf. Loy, p. 87)!
    {
        Report (STEPAMORT set during amortization phase);
        RescheduleInterrupts();  // have to use the new stepamort value
    }
}

void CSoftwareClock::SetSTARTAMORT(INT32)
{
    Update();  // do any old amortizations
    internals_->AMORTTIMER = utcsuWrite_->AMORTTIMESET;
    // take over the amortization time
    internals_->STEPPUREACT = utcsuWrite_->STEPPURE;
    // take over the step value for pure phase
    internals_->LAMBDAPPUREACT = (INT16)(utcsuWrite_->LAMBDAPPURE & 0x0FFFF);
    internals_->LAMBDANPUREACT = (INT16)(utcsuWrite_->LAMBDANPURE & 0x0FFFF);
    // take over the accuracy envelope deterioration values for pure phase

    if (internals_->AMORTTIMER > 0) CClock::SetSTARTAMORT(BeginAmortization);
    else CClock::SetSTARTAMORT(NoAmortization);
    // only start the amort. phase if the amorttimer has a value >= 0

    RescheduleInterrupts();  // react to the new clock state
}


// ACU (Accuracy Unit)
void CSoftwareClock::SetALPHAPNSET(INT32)
{
    Update();  // get the envelopes up to date
    internals_->ALPHAP = utcsuWrite_->ALPHASET;  // load the new envelopes
    internals_->ALPHAN = utcsuWrite_->ALPHASET;
    internals_->NALPHAP = 0x1FFF;  // set the fractional part to all one
    internals_->NALPHAN = 0x1FFF;
}


// CLU (Clock Unit)
void CSoftwareClock::SetSWRESET(INT32)
{
    memset (internals_, 0, sizeof(*internals_));
    internals_->overflow = False;
    assocSimTime_ = GetGlobalTime();
    memset (utcsuRead_, 0, sizeof(*utcsuRead_));
    memset (utcsuWrite_, 0, sizeof(*utcsuWrite_));

    CClock::SetBOUNDP (0xFFFFFFFF);  // set to all 1 after reset
    utcsuRead_->IDENTIFICATION = 0x49435410;  // "ICT"+0x10

    // disable and deschedule the interrupts
    for (int i=0; i<64; i++)
    {
        DisableInterrupt (i);  // disable the interrupt
        if ((i < 32+GPU1) || (i >= 32+NumberOfGPU))  // if not GPS
            DescheduleInterrupt (i);  // remove it from scheduler queue
    }
}


// ITU (Interrupt Unit)
void CSoftwareClock::SetUTCINTEN1(INT32 val)
{
    Update();
    for (INT8 i=0; i<32; i++)  // enable/disable the interrupts 0..31
    {
        if (val & 1<<i)
        {
            EnableInterrupt (i);  // enable the interrupt
            ScheduleInterrupt (i);  // try to schedule it
        }
        else
        {
            DisableInterrupt (i);  // disable the interrupt
            DescheduleInterrupt (i);  // remove it from scheduler queue
        }
    }
    CClock::SetUTCINTEN1(val);
}

void CSoftwareClock::SetUTCINTEN2(INT32 val)
{
    INT8 i;
    Update();
    for (i=32; i<35; i++)  // enable/disable interrupts 32..34 (GPS1PPS)
    {
        if (val & 1<<(i-32))
        {
            EnableInterrupt (i);
        }
        else
        {
            DisableInterrupt (i);
        }
    } // for
    // note: a GPS1PPS interrupt is handled differently; it is created by
    // CSoftwareGps and schedules itself; so it is just enabled, and if it
    // is disabled later on, the process is not descheduled;
    // GPS1PPS interrupts must not be descheduled!

    for (i=35; i<64; i++)  // enable/disable the interrupts 35..63
    {
        if (val & 1<<(i-32))
        {
            EnableInterrupt (i);
            ScheduleInterrupt (i);
        }
        else
        {
            DisableInterrupt (i);
            DescheduleInterrupt (i);
        }
    } // for
    CClock::SetUTCINTEN2(val);
}

void CSoftwareClock::SetUTCINTCLEAR1(INT32 val)
{
    utcsuRead_->UTCINTSTAT1 &= ~val;  // clear the IP bit(s) 0..31
}

void CSoftwareClock::SetUTCINTCLEAR2(INT32 val)
{
    utcsuRead_->UTCINTSTAT2 &= ~val;  // clear the IP bit(s) 32..63
}


// SNU (Snapshot Unit)
void CSoftwareClock::SetSWSNAP(INT32)
{
    SwSnapshot();  // capture all dynamic registers
}


// APU (Application Unit)
void CSoftwareClock::SetAPPDUTYH (INT32 val)
{
    Update();
    CClock::SetAPPDUTYH (val);
    ModifiedDutyTimer (DutyTimerAPP, CNtpTime (val, utcsuWrite_->APPDUTYL));
}

void CSoftwareClock::SetAPPDUTYL(INT32 val)
{
    Update();
    CClock::SetAPPDUTYL (val);
    ModifiedDutyTimer (DutyTimerAPP, CNtpTime (utcsuWrite_->APPDUTYH, val));
}


// SSU1-6 (Synchronization Subnet Unit 1-6)
void CSoftwareClock::SetDUTYAH (ESSUnit ssu, INT32 val)
{
    Update();
    CClock::SetDUTYAH (ssu, val);
    ModifiedDutyTimer ((EDutyTimer)(DutyTimerASSU1+ssu),
                       CNtpTime (val, utcsuWrite_->ssu[ssu].DUTYAL));
}

void CSoftwareClock::SetDUTYAL (ESSUnit ssu, INT32 val)
{
    Update();
    CClock::SetDUTYAL (ssu, val);
    ModifiedDutyTimer ((EDutyTimer)(DutyTimerASSU1+ssu),
                       CNtpTime (utcsuWrite_->ssu[ssu].DUTYAH, val));
}

void CSoftwareClock::SetDUTYBH (ESSUnit ssu, INT32 val)
{
    Update();
    CClock::SetDUTYBH (ssu, val);
    ModifiedDutyTimer ((EDutyTimer)(DutyTimerBSSU1+ssu),
                       CNtpTime (val, utcsuWrite_->ssu[ssu].DUTYBL));
}

void CSoftwareClock::SetDUTYBL (ESSUnit ssu, INT32 val)
{
    Update();
    CClock::SetDUTYBL (ssu, val);
    ModifiedDutyTimer ((EDutyTimer)(DutyTimerBSSU1+ssu),
                       CNtpTime (utcsuWrite_->ssu[ssu].DUTYBH, val));
}


//--------------------------------------------------------------------------
//--- GPS 1PPS handling; FOR CGpsInterrupt ONLY!
//--------------------------------------------------------------------------

// copies the new clock value into MSGPS and TSGPS
// note: a drift of the GPS receiver (which would be very small anyway)
// might be simulated by setting the bits to a value slightly different
// from zero; sets 'gpu_' to 'gpu'
// assert: gpu < EGpsUnitCount
void CSoftwareClock::CaptureTime (EGpsUnit gpu)
{
    Assert (gpu < EGpsUnitCount);

    Update();

    utcsuRead_->gpu[gpu].MSGPS = internals_->ms;  // GPS healthy
    utcsuRead_->gpu[gpu].TSGPS = internals_->ts;
/**/// note: right now, GPS is always healthy

    gpu_ = gpu;  // store the GPS unit
}


//----------------------------------------------------------------------
//--- GPS interrupt handling; FOR CSoftwareGps ONLY!
//----------------------------------------------------------------------

// returns the GPS unit that has last called CaptureTime()
EGpsUnit CSoftwareClock::GetGpsUnit()
{
    return gpu_;
}


// returns the GPS interrupt process of the given GPU; if the process
// does not exist, it is created
// assert: gpu < EGpsUnitCount
CGpsInterrupt *CSoftwareClock::GetGpsInterrupt (EGpsUnit gpu)
{
    Assert (gpu < EGpsUnitCount);

    return (CGpsInterrupt *)CreateInterrupt (gpu+32);
}


//--------------------------------------------------------------------------
//--- initialization and destruction of the data members
//--------------------------------------------------------------------------

// initialize all members
void CSoftwareClock::Init ()
{
    listOfInterrupts_ = new CInterruptList;
    Assert (listOfInterrupts_ != NULL);
    assocSimTime_ = GetGlobalTime();
    fOsc_ = GetFrequency();
    gpu_ = EGpsUnitCount;  // invalid value
}

// delete all dynamic members
void CSoftwareClock::Dispose()
{
    delete listOfInterrupts_;  // that deletes or unref's all processes
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int CSoftwareClock::ReportError (EReportType type, char *file, int line,
                                 char *expression)
{
    return ::ReportError (type,file,line,expression,Clock,nodeID_,area_);
}


//--------------------------------------------------------------------------
//--- snapshots
//--------------------------------------------------------------------------

// make a hardware snapshot, generate an interrupt (if enabled)
void CSoftwareClock::HwSnapshot()
{
    Update();
    utcsuRead_->MSSNU = internals_->ms;
    utcsuRead_->TSSNU = internals_->ts;
    utcsuRead_->ACCPSNU = GetAccuracy();

    if (utcsuWrite_->UTCINTEN1 & SUtcsuWrite::IHWSNAP)  // snapshot enabled
    {
        CClockInterrupt *clockint = CreateInterrupt (IntHwSnapshot);
        clockint->RaiseAfter(0);  // raise interrupt at once
        // note: scheduleTime = -1, so the interrupt cannot be rescheduled
        // in ScheduleInterrupt(); but that is quite alright, since the
        // interrupt is always raised at once and therefore rescheduling is
        // never necessary
    }
}

// make a software snapshot (captures all dynamically changing registers)
void CSoftwareClock::SwSnapshot()
{
    GetTSGETL();
    // that should have the same effect as the software snapshot
}


//--------------------------------------------------------------------------
//--- internal access functions for status/configuration bits
//--------------------------------------------------------------------------

// sets the bit in UTCSTAT1 (0..31) or UTCSTAT2 (32..63)
// assert: signalBit < 64
void CSoftwareClock::SetSignal (INT8 signalBit)
{
    Assert (signalBit < 64);

    CClockInterrupt *clockint = listOfInterrupts_->Find (signalBit);
/**/// note: that assumes that the signal has an interrupt associated with
    // it; that's not always the case (pe. SCAMA)!

    if ((clockint != NULL) && !clockint->idle()) return;
    // interrupt is scheduled (or active) and will set the signal

    // no interrupt or not scheduled; set the signal at once
    if (signalBit < 32) utcsuRead_->UTCSTAT1 |= 1<<signalBit;
    else utcsuRead_->UTCSTAT2 |= 1 << (signalBit-32);
}

// returns True if the bit in UTCSTAT1 (0..31) or UTCSTAT2 (32..63) is
// set, otherwise False
// assert: signalBit < 64
EBool CSoftwareClock::GetSignal (INT8 signalBit)
{
    Assert (signalBit < 64);

    if ((signalBit < 32) && (utcsuRead_->UTCSTAT1 & 1<<signalBit))
        return True;
    if ((signalBit > 31) && (utcsuRead_->UTCSTAT2 & 1<<(signalBit-32)))
        return True;
    return False;
}

// clears the bit in UTCSTAT1 (0..31) or UTCSTAT2 (32..63)
// assert: signalBit < 64
void CSoftwareClock::ClearSignal (INT8 signalBit)
{
    Assert (signalBit < 64);

    if (signalBit < 32) utcsuRead_->UTCSTAT1 &= ~(1<<signalBit);
    else utcsuRead_->UTCSTAT2 &= ~(1 << (signalBit-32));
}

// returns True if the bit in UTCCONF1 (0..31) or UTCCONF2 (32..63) is
// set, otherwise False
// assert: signalBit < 64
EBool CSoftwareClock::GetPolarity (INT8 bitnumber)
{
    Assert (bitnumber < 64);

    if ((bitnumber < 32) && (utcsuWrite_->UTCCONF1 & 1<<bitnumber))
        return True;
    if ((bitnumber > 31) && (utcsuWrite_->UTCCONF2 & 1<<(bitnumber-32)))
        return True;
    return False;
}


//--------------------------------------------------------------------------
//--- internal clock update functions
//--------------------------------------------------------------------------

// update the clock value
void CSoftwareClock::Update()
{
    if (assocSimTime_ >= GetGlobalTime()) return;
    // clock is already up to date

    CNtpTime cycles;
    GetClockCycles (GetGlobalTime() - assocSimTime_, cycles);
    // get the number of cycles since the last update

    if ((cycles.ms == 0) && (cycles.ts == 0)) return;
    // difference too small, does not fit into 1 cycle

    double cycleDiff = GetTime (cycles, fOsc_);
    // get the real difference that is updated

    assocSimTime_ += cycleDiff;  // store the current global time

    if (internals_->AMORTTIMER > 0)  // have to enter amortization phase
    {
        CNtpTime amortCycles;
        amortCycles.ts = GetAmortCycles (cycles);

        UpdateNtp (utcsuWrite_->STEPAMORT, amortCycles);
        // do the amortization of the clock value

        UpdateEnvelopes (False, amortCycles);

        cycles = cycles - amortCycles;
        internals_->AMORTTIMER -= amortCycles.ts;

        if (internals_->AMORTTIMER==0) utcsuWrite_->STARTAMORT=NoAmortization;
        // amortization phase is finished
        else utcsuWrite_->STARTAMORT=MiddleOfAmortization;
        // at least the first cycle is past
    }

    if (cycles.ms==0 && cycles.ts==0) return;
    // finished, no more cycles left

    // now do the pure phase
    UpdateNtp (internals_->STEPPUREACT, cycles);
    UpdateEnvelopes (True, cycles);
}


// add the step-value for 'cycles' cycles to the NTPTIME, take STEPLOW
// as the [-52,-59] portion of step
// note: for doing the mutliplication and adding to the NTPTIME, an
// INT128 object is used
void CSoftwareClock::UpdateNtp (INT32 step, const CNtpTime& cycles)
{
    INT8 steplow = utcsuWrite_->MSSET>>24 & 0x0FF;

    INT128 stepreg1;
    INT128 cycle1;

    stepreg1.SetBinary (0, 0, step>>24 & 0x0FF,
                        (step<<8 & 0xFFFFFF00) + steplow);
    cycle1.SetBinary (0, 0, cycles.ms, cycles.ts);

    stepreg1 = stepreg1*cycle1;  // this is the value that is added to NTPTIME
    stepreg1 += INT128(internals_->ms, internals_->ts, internals_->us,
                       internals_->ns);  // so let us add the NTPTIME

    INT8 overflow;
    // now, stepreg contains the new NTPTIME registers
    stepreg1.GetNtp (internals_->ms, internals_->ts, internals_->us, 
                     (INT8&)internals_->ns, overflow);

    SetOverflow (overflow);
}

// assumes that 'alpha' is S:[-8,-38] and 'nalpha' is [-39,-51] (just 13 bit!),
// sets the INT128 accordingly
static void EnvelopeToINT128 (INT128& i, INT32 alpha, INT16 nalpha)
{
    if (alpha & 0x80000000)  // convert two's compl. to absolute value and sign
    {
        nalpha = (~nalpha & 0x1FFF) + 1;  // make nalpha positive (keep carry)
        if (nalpha > 0x1FFF) alpha = ~alpha + 1;  // add carry from nalpha
        else alpha = ~alpha;  // no carry -> just invert
        nalpha &= 0x1FFF;  // remove carry

        i.SetBinary (0, 0, alpha>>19 & 0x1FFF,
                     (alpha<<13 & 0xFFFFE000) + (nalpha & 0x1FFF));
        i = -i;
    }
    else
    {
        i.SetBinary (0, 0, alpha>>19 & 0x0FFF,
                     (alpha<<13 & 0xFFFFE000) + (nalpha & 0x1FFF));
    }
}

// if pure=True, LAMBDAP(N)PUREACT is added to ALPHAP(N) for 'cycles'
// cycles; if pure=False, LAMBDAP(N)AMORT is used, and if STARTAMORT is
// set to 'BeginAmortization', the envelopes are realigned using STATEP(N)SET
void CSoftwareClock::UpdateEnvelopes (EBool pure, const CNtpTime& cycles)
{
    INT16 stepreg;  // for determining the step value that is added
    INT128 cycle1;  // the number of cycles
    INT128 stepreg1;  // the step value
    INT128 alpha;  // the value of the envelope
    INT32 align;  // value for aligning envelope in first amort. cycle
    INT128 align1;  // alignment value
    INT128 upper;  // upper bound on envelopes
    INT128 lower;  // lower bound on envelopes
    INT32 temp;  // to store the ALPHAP(N) reg that should be modified
    INT16 ntemp;  // store the NALPHAP(N) reg that should be modified
    static const short EnvP = 0;  // for the for-loop below
    EBool positive = True;  // needed for easier code below


    upper.SetBinary (0, 0, 0xFFF, 0xFFFFFFFF);
    lower.SetBinary (0, 0, 0x1000, 0);
    lower = -lower;  // negative bound

    // first cycle of amortization is used for realigning the envelopes
    if (!pure && (utcsuWrite_->STARTAMORT==BeginAmortization))
    {
        Assert (cycles.ms == 0);
        // during amort. phase, cycles.ms should be zero

        if (cycles.ts == 0) return;  // nothing to do

        cycle1.SetBinary (0, 0, cycles.ms, cycles.ts-1);
    }
    else  // all cycles are used for adding the step value
    {
        cycle1.SetBinary (0, 0, cycles.ms, cycles.ts);
    }


    for (int i=EnvP; i<EnvP+2; i++)  // do computation for P and N env.
    {
        if (i==EnvP) positive=True;
        else positive=False;

        align = 0;  // default: no alignment value is added

        switch (pure)
        {
            case True:
                if (i==EnvP) stepreg = internals_->LAMBDAPPUREACT;
                else stepreg = internals_->LAMBDANPUREACT;
                break;
            case False:
                if (utcsuWrite_->STARTAMORT==BeginAmortization)
                // get the value for realigning the envelope
                {
                    if (i==EnvP) align = utcsuWrite_->STATEPSET;
                    else align = utcsuWrite_->STATENSET;
                }

                if (i==EnvP) stepreg = utcsuWrite_->LAMBDAPPURE>>16 & 0x0FFFF;
                else stepreg = utcsuWrite_->LAMBDANPURE>>16 & 0x0FFFF;
                break;
        } // switch

        if (stepreg & 0x8000)
        {
            stepreg1.SetBinary (0, 0, 0, (~stepreg+1) & 0x0FFFF);  // set step value
            //stepreg1 = -stepreg1;  // value is negative
        }
        else stepreg1.SetBinary (0, 0, 0, stepreg & 0x7FFF);  // set step value

        stepreg1 = stepreg1*cycle1;  // get the complete update value

        EnvelopeToINT128 (align1, align, 0);  // store the alignment value

        GetEnvelope (positive, temp, ntemp);  // read the envelope (P or N)

        // store the envelope in alpha
        EnvelopeToINT128 (alpha, temp, ntemp);


        //--- add the alignment value

        alpha += align1;
        // should rather subtract 'align1'?!?

        if (alpha > upper) alpha = upper;  // overflow
        else if (alpha < lower) alpha = lower;  // underflow


        //--- add the step value

        alpha += stepreg1;       // add the step register

        //--- check for overflow, store result
        if (alpha >= upper) SetUpperLimit (positive);
        else if (alpha <= lower) SetLowerLimit (positive);
        else
        {
            if (alpha < 0) alpha = ~alpha - 1;  // convert to two's complement
            INT32 a, b, c, d;  // just temporaries
            alpha.GetBinary (a, b, c, d);
            temp = ((c & 0x0FFF)<<19) + (d>>13 & 0x7FFFF);
            ntemp = (INT16)d & 0x1FFF;
            if (alpha < 0) temp |= 0x80000000;  // set the sign bit

            SetEnvelope (positive, temp, ntemp);  // store result
        } // else
    } // for
}

// sets internals_->ALPHAP(N) and internals_->NALPHAP(N) to the upper limit
// (which is 0:1...1); if 'positive' is True, the P-envelope is used,
// else the N-envelope is set
/**/// right now, no interrupts are raised
void CSoftwareClock::SetUpperLimit (EBool positive)
{
    if (positive)
    {
        internals_->ALPHAP = 0x7FFFFFFF;
        internals_->NALPHAP = 0x1FFF;
/**/    // here, INTPLALPHAP should be raised
    }
    else
    {
        internals_->ALPHAN = 0x7FFFFFFF;
        internals_->NALPHAN = 0x1FFF;
/**/    // here, INTPLALPHAN should be raised
    }
}

// sets internals_->ALPHAP(N) and internals_->NALPHAP(N) to the lower limit
// (which is 1:0...0); if 'positive' is True, the P-envelope is used,
// else the N-envelope is set
/**/// right now, no interrupts are raised
void CSoftwareClock::SetLowerLimit (EBool positive)
{
    if (positive)
    {
        internals_->ALPHAP = 0x80000000;
        internals_->NALPHAP = 0;
/**/    // here, INTNLALPHAP should be raised
    }
    else
    {
        internals_->ALPHAN = 0x80000000;
        internals_->NALPHAN = 0;
/**/    // here, INTNLALPHAN should be raised
    }
}


// reads internals_->ALPHAP(N) into alpha and internals_->NALPHAP(N) into
// nalpha; if 'positive' is True, the P-envelope is used, else the
// N-envelope is read
void CSoftwareClock::GetEnvelope (EBool positive, INT32& alpha, INT16& nalpha)
{
    if (positive)
    {
        alpha = internals_->ALPHAP;
        nalpha = internals_->NALPHAP;
    }
    else
    {
        alpha = internals_->ALPHAN;
        nalpha = internals_->NALPHAN;
    }
}

// stores alpha into internals_->ALPHAP(N) and nalpha into
// internals_->NALPHAP(N); if 'positive' is True, the P-envelope is used,
// else the N-envelope is modified
void CSoftwareClock::SetEnvelope (EBool positive, INT32 alpha, INT16 nalpha)
{
    if (positive)
    {
        internals_->ALPHAP = alpha;
        internals_->NALPHAP = nalpha & 0x1FFF;
    }
    else
    {
        internals_->ALPHAN = alpha;
        internals_->NALPHAN = nalpha & 0x1FFF;
    }
}

// returns the global time (in seconds) that will pass until the
// internal time 'delta' (in seconds) has passed; if 'delta' is no
// multiple of the granularity, the next multiple is used instead
double CSoftwareClock::GetRealDelay (double delta)
{
    if (delta <= 0) return 0;  // clock has reached the value

    Update ();  // set the clock to its correct value

    double realdelta=0;  // the global delta

    if (internals_->AMORTTIMER > 0)  // have to consider amort. phase
    {
        double amortRate = GetRate (utcsuWrite_->STEPAMORT);
        // compute the amortization rate

        double amortClockTime = (internals_->AMORTTIMER * amortRate)/fOsc_;
        // that's the time (in clock seconds) that the amortization phase
        // will last

        if (amortClockTime >= delta) return delta/amortRate;
        // the whole 'delta' is in the amortization phase, only amortRate
        // is needed
/**/// note: should use next multiple of granularity, instead of delta

        realdelta = amortClockTime/amortRate;  // real duration of amort.
        delta -= amortClockTime;  // rest of 'delta' is in pure phase
    } // if

    realdelta += delta/GetRate (internals_->STEPPUREACT);
    // use the rate of the pure phase for the rest of 'delta'
/**/// note: should use next multiple of granularity, instead of delta

    return realdelta;
}


// computes the number of clock cycles that occur during a duration of
// 'delta' real seconds and stores it in 'cycles'
// note: two INT32 should be sufficient, since one INT32 can store 4e9,
// so two are 16e18; assuming frequencies in the range of 1e6..1e9 and
// deltas in the range of 1..100 secs, there's ample space left
void CSoftwareClock::GetClockCycles (double delta, CNtpTime& cycles)
{
    if (delta <= 0)  // no cycles during a non-positive delta
    {
        cycles.ts = 0;
        cycles.ms = 0;
        return;
    }
    
    double high = (fOsc_*delta)/TwoPow32;
    // that contains the high portion of the cycles

    cycles.ms = (INT32)high;  // round down
    if (cycles.ms==0) cycles.ts = (INT32)(fOsc_*delta);
    else cycles.ts = (INT32)((high-cycles.ms)*TwoPow32);
/**/// note: if high and cycles.ms are big, the rounding error may become
    // significant; but I don't expect high to get bigger than 10
}

// returns the number of amortization cycles in the next 'cycles' clock
// cycles
INT32 CSoftwareClock::GetAmortCycles (const CNtpTime& cycles)
{
    if (cycles.ms==0 && cycles.ts<internals_->AMORTTIMER) return cycles.ts;
    else return internals_->AMORTTIMER;
}


// returns the real clock rate (clock seconds per second) if step and
// STEPLOW is used
double CSoftwareClock::GetRate (INT32 step)
{
    INT8 steplow = (INT8)(utcsuWrite_->MSSET>>24 & 0x0FF);
    return (fOsc_/TwoPow30)*((double)step/TwoPow21 + (double)steplow/TwoPow29);
/**/// note: do *not* use 1/2^31 or more, since the result is negative
    // (even though the Win95 calculator computes a positive result; maybe
    // a bug in the WATCOM library?)
    // anyway: rounding errors may occur for small values of step and
    // steplow (cf. to the test file for measurements); if the errors are
    // not tolerable, two algorithms (one for large and one for small
    // values) should be implemented
}


//--------------------------------------------------------------------------
//--- internal interrupt handling
//--------------------------------------------------------------------------

// enable the interrupt; interruptBit is in the range from 0..63
// (0..31: UTCINTEN1; 32..63: UTCINTEN2)
void CSoftwareClock::EnableInterrupt (INT8 interruptBit)
{
    CClockInterrupt *clockint = listOfInterrupts_->Find (interruptBit);
    if (clockint != NULL) clockint->Enable ();  // allow call to ISR
}

// if the interrupt of the source was enabled, it is turned off; else
// nothing happens
void CSoftwareClock::DisableInterrupt (INT8 interruptBit)
{
    CClockInterrupt *clockint = listOfInterrupts_->Find (interruptBit);
    if (clockint != NULL) clockint->Disable ();
}


// create and return an interrupt process; if an interrupt with the ID
// already exists, that interrupt is returned instead
// a new interrupt is put into the interrupt list, enabled according to
// the UTCINTENx register, and its vector is set according to its ID
// assert: interruptBit < 64
CClockInterrupt *CSoftwareClock::CreateInterrupt (INT8 interruptBit)
{
    Assert (interruptBit < 64);

    CClockInterrupt *clockint = listOfInterrupts_->Find (interruptBit);
    if (clockint != NULL) return clockint;

    EGpsUnit gpu;  // for GPS1PPS interrupts

    switch (interruptBit)
    {
        case IntGPS1:
        case IntGPS2:
        case IntGPS3:
            gpu = (EGpsUnit)(interruptBit-IntGPS1);
            // get the gpu (gpu=0 --> GPU1)
            clockint = new CGpsInterrupt (nodeID_, this, gpu);
            break;
        case IntDutyB1:
            clockint = new CDutyB1Interrupt (nodeID_, this);
            break;

        default:
            clockint = new CClockInterrupt (nodeID_);  // normal interrupt
            break;
    } // switch

    Assert (clockint != NULL);

    // set the interrupt vector
    clockint->SetVector (GetClockIntvec (nodeID_, area_) |
                         GetInterruptType (interruptBit));

    listOfInterrupts_->Insert (clockint, interruptBit);

    // store the status register that is modified (IP bit set) on interrupt
    if (interruptBit > 31)
    {
        clockint->SetStatusRegister (utcsuRead_->UTCINTSTAT2);
        interruptBit -= 32;  // modify the interruptBit for the process

        // enable/disable the interrupt according to the UTCINTENx register
        if (utcsuWrite_->UTCINTEN2 & 1<<interruptBit) clockint->Enable();
        else clockint->Disable();
    }
    else
    {
        clockint->SetStatusRegister (utcsuRead_->UTCINTSTAT1);

        // enable/disable the interrupt according to the UTCINTENx register
        if (utcsuWrite_->UTCINTEN1 & 1<<interruptBit) clockint->Enable();
        else clockint->Disable();
    }

    clockint->SetPendingBit (interruptBit);  // store the IP bit number

#ifndef NO_RESOURCE
    Resource::ref(clockint);  // register the process in C++SIM
#endif

    return clockint;
}

// reschedules all interrupt processes after the clock state has been
// changed; processes that were not scheduled will be ignored
void CSoftwareClock::RescheduleInterrupts ()
{
    for (CClockInterrupt *clockint = listOfInterrupts_->GetFirst();
         clockint != NULL;
         clockint = listOfInterrupts_->GetNext())
    {
        if (!clockint->idle()) ScheduleInterrupt (clockint);
        // interrupt is scheduled (or active) --> reschedule it
        // note: active processes may be (pseudo-)scheduled using
        // RaiseAfter()
    }
}

// (re-)schedule the process so that the interrupt occurs at its stored
// scheduleTime; the function assumes that the clock is up to date
void CSoftwareClock::ScheduleInterrupt (INT8 interruptBit)
{
    CClockInterrupt *clockint = listOfInterrupts_->Find (interruptBit);
    if (clockint != NULL) ScheduleInterrupt (clockint);
}

// assert: clockint != NULL
void CSoftwareClock::ScheduleInterrupt (CClockInterrupt *clockint)
{
    Assert (clockint != NULL);

    double time, scheduledTime;

    scheduledTime = clockint->GetScheduleTime ();  // scheduled clock value
    if (scheduledTime < 0) return;  // process does not need scheduling

    time = GetCurrentClockValue ();  // current clock value

    if (time >= scheduledTime)  // interrupt should occur at once!
        clockint->RaiseAfter (0);
    else  // still some time left till interrupt
        clockint->RaiseAfter (GetRealDelay (scheduledTime-time));
// DEBUG
/*
cout << " node " << nodeID_ << " is raising interrupt at sim.time ";
precout (cout, GetGlobalTime(), 12); cout << " + ";
cout << GetRealDelay (scheduledTime-time) << " seconds" << endl;
cout << " that corresponds to clock time ";
precout (cout, time, 12); cout << " + ";
cout << scheduledTime - time << endl;
*/
}

// removes the interrupt process from the scheduler queue; if the process
// does not exist or is not scheduled, nothing happens
void CSoftwareClock::DescheduleInterrupt (INT8 interruptBit)
{
    CClockInterrupt *clockint = listOfInterrupts_->Find (interruptBit);
    if (clockint != NULL) clockint->CancelInterrupt();
}


//--------------------------------------------------------------------------
//--- internal timestamp/accuracy access
//--------------------------------------------------------------------------

// returns the current clock value in seconds
double CSoftwareClock::GetCurrentClockValue()
{
    static const double BitMinus59 = 1.734723475977e-18;
    // the value of bit -59 of the NTP time in seconds; the exact
    // value was obtained using the Win95 calculator (2^-59)

    return NtpToSeconds (internals_->ms & 0x00FFFFFF, internals_->ts,
                         internals_->us) + (double)internals_->ns*BitMinus59;
}


// returns the current accuracy (from INTPBUS)
INT32 CSoftwareClock::GetAccuracy()
{
    INT32 acc;

    // remember: negative accuracies are read as zero
    if (internals_->ALPHAN & 1<<31) acc = 0;
    else acc = ((internals_->ALPHAN & 0x7FFF8000) << 1);
    // store bits [-8,-23] in the upper two bytes (leave out the sign bit)
    if (!(internals_->ALPHAP & 1<<31)) acc += (internals_->ALPHAP>>15 & 0x0FFFF);

    return acc;
}


// if overflow != 0, internals_->overflow is set to True (and the overflow
// bits are removed from the macrostamp); else it is set to False
void CSoftwareClock::SetOverflow (int overflow)
{
    if (overflow)
    {
        internals_->overflow = True;
/**/    // note: should raise an interrupt if enabled
        internals_->ms &= 0x00FFFFFF;  // mask out the overflow bits
    }
    else internals_->overflow = False;
}


//--------------------------------------------------------------------------
//--- internal duty timer handling
//--------------------------------------------------------------------------

// if the duty timer (duty.ms = DutyHigh, duty.ts = DutyLow) is enabled
// the clock interrupt process is scheduled accordingly; else the process
// is descheduled
// if duty timer is enabled, assert: 0 <= dutyBit <= 63
void CSoftwareClock::ModifiedDutyTimer (EDutyTimer dutyBit, const CNtpTime& duty)
{
    CClockInterrupt *clockint;

    if (duty.ts & DutyEnabled)
    // timer is enabled -> create and schedule the interrupt process
    {
        clockint = CreateInterrupt (dutyBit);  // make or get the interrupt
        clockint->SetScheduleTime (DutyToSeconds (duty.ms, duty.ts));
        ScheduleInterrupt (clockint);  // try to schedule the interrupt
    }
    else  // deschedule the interrupt; clear timer signal in UTCSTAT
    {
        clockint = listOfInterrupts_->Find (dutyBit);
        if (clockint != NULL)
        {
            clockint->SetScheduleTime (-1);
            clockint->CancelInterrupt();
        }
        if (dutyBit < 32) utcsuRead_->UTCSTAT1 &= ~(1 << dutyBit);
        else utcsuRead_->UTCSTAT2 &= ~(1 << (dutyBit-32));
        // mask out the signal bit
    }
}


#ifndef NDEBUG
//--------------------------------------------------------------------------
//--- debugging
//--------------------------------------------------------------------------

static void PrintHex (const char *str, INT32 reg)
{
    char strOutput[80];
    sprintf (strOutput, "%-14s  %x", str, reg);
    cout << strOutput << endl;
}

// prints all registers to cout
void CSoftwareClock::DumpRegisters()
{
    int i;  // for iterating through replicated units

    cout << endl << "*** register dump for the UTCSU on node/area ";
    cout << nodeID_ << "/" << area_ << " ***" << endl;
    cout << "the dump is made at simulation time " << (INT32)GetGlobalTime();
    cout << " + " << GetGlobalTime() - (INT32)GetGlobalTime();
    cout << ", diff to current assoc. time is " << GetGlobalTime() - assocSimTime_;
    cout << ", clock time is ";
    cout << (INT32)NtpToSeconds (internals_->ms, internals_->ts, internals_->us);
    cout << " + " << NtpToSeconds (internals_->ms, internals_->ts, internals_->us) - (INT32)NtpToSeconds (internals_->ms, internals_->ts, internals_->us);
    cout << endl;

    // ------- internal registers -------

    cout << endl << "*** internal registers ***" << endl;

    PrintHex ("ms", internals_->ms);
    PrintHex ("ts", internals_->ts);
    PrintHex ("us", internals_->us);
    PrintHex ("ns", internals_->ns);
    PrintHex ("overflow", internals_->overflow);
    PrintHex ("STEPPUREACT", internals_->STEPPUREACT);
    PrintHex ("LAMBDAPPUREACT", internals_->LAMBDAPPUREACT);
    PrintHex ("LAMBDANPUREACT", internals_->LAMBDANPUREACT);
    PrintHex ("ALPHAP", internals_->ALPHAP);
    PrintHex ("ALPHAN", internals_->ALPHAN);
    PrintHex ("NALPHAP", internals_->NALPHAP);
    PrintHex ("NALPHAN", internals_->NALPHAN);
    PrintHex ("AMORTTIMER", internals_->AMORTTIMER);


    // ------- read registers -------

    cout << endl << "*** read registers ***" << endl;

    // LTU
    PrintHex ("MSGET", utcsuRead_->MSGET);
    PrintHex ("TSGETL", utcsuRead_->TSGETL);
    PrintHex ("USGET", utcsuRead_->USGET);
    PrintHex ("TSGET", utcsuRead_->TSGET);
    PrintHex ("NSGET", utcsuRead_->NSGET);
    PrintHex ("AMORTTIMEGETL", utcsuRead_->AMORTTIMEGETL);
    PrintHex ("AMORTTIMEGET", utcsuRead_->AMORTTIMEGET);

    // ACU
    PrintHex ("STATEPGETL", utcsuRead_->STATEPGETL);
    PrintHex ("STATEPGET", utcsuRead_->STATEPGET);
    PrintHex ("ALPHAPGETL", utcsuRead_->ALPHAPGETL);
    PrintHex ("ALPHANGETL", utcsuRead_->ALPHANGETL);
    PrintHex ("ALPHAPGET", utcsuRead_->ALPHAPGET);
    PrintHex ("ALPHANGET", utcsuRead_->ALPHANGET);

    // ITU
    PrintHex ("IDENTIFICATION", utcsuRead_->IDENTIFICATION);
    PrintHex ("UTCSTAT1", utcsuRead_->UTCSTAT1);
    PrintHex ("UTCSTAT2", utcsuRead_->UTCSTAT2);
    PrintHex ("UTCCONF1", utcsuRead_->UTCCONF1);
    PrintHex ("UTCCONF2", utcsuRead_->UTCCONF2);
    PrintHex ("UTCINTSTAT1", utcsuRead_->UTCINTSTAT1);
    PrintHex ("UTCINTSTAT2", utcsuRead_->UTCINTSTAT2);

    // SNU
    PrintHex ("MSSNU", utcsuRead_->MSSNU);
    PrintHex ("TSSNU", utcsuRead_->TSSNU);
    PrintHex ("ACCPSNU", utcsuRead_->ACCPSNU);

    // APU
    PrintHex ("MSAPP", utcsuRead_->MSAPP);
    PrintHex ("TSAPP", utcsuRead_->TSAPP);
    PrintHex ("ACCPAPP", utcsuRead_->ACCPAPP);
    PrintHex ("MSAPPX", utcsuRead_->MSAPPX);
    PrintHex ("TSAPPX", utcsuRead_->TSAPPX);
    PrintHex ("ACCPAPPX", utcsuRead_->ACCPAPPX);

    // GPU1-3
    for (i=0; i<3; i++) {
    cout << "GPU" << i+1 << ":" << endl;
    PrintHex ("  MSGPS", utcsuRead_->gpu[i].MSGPS);
    PrintHex ("  TSGPS", utcsuRead_->gpu[i].TSGPS); }

    // SSU1-6
    for (i=0; i<6; i++) {
    cout << "SSU" << i+1 << ":" << endl;
    PrintHex ("  MSRCV", utcsuRead_->ssu[i].MSRCV);
    PrintHex ("  TSRCV", utcsuRead_->ssu[i].TSRCV);
    PrintHex ("  MSXMT", utcsuRead_->ssu[i].MSXMT);
    PrintHex ("  TSXMT", utcsuRead_->ssu[i].MSXMT);
    PrintHex ("  ACCPXMT", utcsuRead_->ssu[i].ACCPXMT);
    PrintHex ("  ACCNPLXMT", utcsuRead_->ssu[i].ACCNPLXMT);
    PrintHex ("  ACCNPHXMT", utcsuRead_->ssu[i].ACCNPHXMT); }


    // ------- write registers -------

    cout << endl << "*** write registers ***" << endl;

    // LTU
    PrintHex ("MSSET", utcsuWrite_->MSSET);
    PrintHex ("TSSET", utcsuWrite_->TSSET);
    PrintHex ("USSET", utcsuWrite_->USSET);
    PrintHex ("NTPSET", utcsuWrite_->NTPSET);
    PrintHex ("STEPAMORT", utcsuWrite_->STEPAMORT);
    PrintHex ("STEPPURE", utcsuWrite_->STEPPURE);
    PrintHex ("AMORTTIMESET", utcsuWrite_->AMORTTIMESET);
    PrintHex ("STARTAMORT", utcsuWrite_->STARTAMORT);

    // ACU
    PrintHex ("LAMBDAPPURE", utcsuWrite_->LAMBDAPPURE);
    PrintHex ("LAMBDANPURE", utcsuWrite_->LAMBDANPURE);
    PrintHex ("STATEPSET", utcsuWrite_->STATEPSET);
    PrintHex ("STATENSET", utcsuWrite_->STATENSET);
    PrintHex ("BOUNDP", utcsuWrite_->BOUNDP);
    PrintHex ("ALPHASET", utcsuWrite_->ALPHASET);
    PrintHex ("ALPHAPNSET", utcsuWrite_->ALPHAPNSET);

    // CLU
    PrintHex ("SWRESET", utcsuWrite_->SWRESET);

    // ITU
    PrintHex ("UTCCONF1", utcsuWrite_->UTCCONF1);
    PrintHex ("UTCCONF2", utcsuWrite_->UTCCONF2);
    PrintHex ("UTCINTEN1", utcsuWrite_->UTCINTEN1);
    PrintHex ("UTCINTEN2", utcsuWrite_->UTCINTEN2);
    PrintHex ("UTCINTCLEAR1", utcsuWrite_->UTCINTCLEAR1);
    PrintHex ("UTCINTCLEAR2", utcsuWrite_->UTCINTCLEAR2);

    // SNU
    PrintHex ("SWSNAP", utcsuWrite_->SWSNAP);
    PrintHex ("DUTYH", utcsuWrite_->DUTYH);
    PrintHex ("DUTYL", utcsuWrite_->DUTYL);

    // APU
    PrintHex ("APPDUTYH", utcsuWrite_->APPDUTYH);
    PrintHex ("APPDUTYL", utcsuWrite_->APPDUTYL);
    PrintHex ("APPSEL", utcsuWrite_->APPSEL);
    PrintHex ("APPCLEAR", utcsuWrite_->APPCLEAR);

    // SSU1-6
    for (i=0; i<6; i++) {
    cout << "SSU" << i+1 << ":" << endl;
    PrintHex ("  DUTYAH", utcsuWrite_->ssu[i].DUTYAH);
    PrintHex ("  DUTYAL", utcsuWrite_->ssu[i].DUTYAL);
    PrintHex ("  DUTYBH", utcsuWrite_->ssu[i].DUTYBH);
    PrintHex ("  DUTYBL", utcsuWrite_->ssu[i].DUTYBL); }
}
#endif


//**************************************************************************
// module global functions
//**************************************************************************

// returns the interrupt type (INTA, INTN, INTT) of the bit; interruptBit
// is in the range of 0..63
// assert: interrupt bit is in use
static EClockInterrupt GetInterruptType (INT8 interruptBit)
{
    if (interruptBit < 32)
    {
        if (1<<interruptBit & INTAReg1) return INTA;
        if (1<<interruptBit & INTTReg1) return INTT;
    }
    else if (interruptBit < 64)
    {
        if (1<<(interruptBit-32) & INTNReg2) return INTN;
        if (1<<(interruptBit-32) & INTTReg2) return INTT;
    }

    Assert (False);
    // the interrupt bit is not in use

    return INTA;  // just to satisfy the compiler
}

// converts the contents of the duty timer registers to seconds
// note: dutyh contains NTP bits [+31,0], dutyl contains [-1,-16] in
// bits [0,15]; the timestamp is expanded with 0 in bits [-17,-24]
static double DutyToSeconds (INT32 dutyh, INT32 dutyl)
{
    return dutyh + (double)(dutyl & 0x0FFFF)/TwoPow16;
}

// returns the global time that corresponds to 'cycles' clock cycles of
// a clock with the frequency 'frequency'
static double GetTime (const CNtpTime& cycles, double frequency)
{
    const double TwoPow32 = 4294967296.;
    return cycles.ts/frequency + (cycles.ms*TwoPow32)/frequency;
}


