//
// clockifc.cpp
//
// implementation of the clock interface
//
// possible bug in function SetStepRegisters(): should an over-/underflow
//      in steplow in-/decrement steppure? (consider: maybe overflow
//      can not happen?)
//
// created: 26.10.96
//
// last change: 03.02.99
//
// RCS version control information:
//     $Revision: 1.28 $$State: Exp $$Date: 1999/02/03 11:43:59 $
//

#include <math.h>
#include <stdlib.h>

#include "clockbas.hpp"
#include "clockcfg.hpp"
#include "clockifc.hpp"
#include "constant.hpp"
#include "distrib.hpp"
#include "gpsbase.hpp"
#include "params.hpp"
#include "template.hpp"
#include "time.hpp"


static const double TwoPow51 = 4294967296.*524288.;
// needed for computing the value of the STEP and STEPLOW registers
static const double TwoPow38 = 4294967296.*64.;
// needed in DragAccuracy() for preloads

static const SINT32 UpperAccuracyBound = 0x7FFFFFFF;
static const SINT32 LowerAccuracyBound = 0x80000000;
// bounds for the accuracy values; needed in DragAccuracy()

static const double GpsAmort = 0.99;
// amortization time of GPS


//**************************************************************************
// module global functions
//**************************************************************************

static INT32 SourceToMask (EInterruptSource source);
// get interrupt bit for an interrupt source
// if source<0 or source>63, an assert fails



//**************************************************************************
// class CClockInterface: this class is the interface to the clock (both
//      hardware and software)
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CClockInterface::CClockInterface (CControllingObject *ctrl, double frequency,
    int area)
    : CApplControlledObject (ctrl)
{
    clock_ = clockTemplate->NewClock (frequency, GetNodeID(), area);
    Assert (clock_ != NULL);

    INT32 old = DisableInterrupts();
    SetISR (GetClockIntvec (GetNodeID(), area) | INTA, this);
    // clock interrupt INTA (hardware snapshot) calls the default ISR
    // of CClockInterface

    SetISR (GetClockIntvec (GetNodeID(), area) | INTT, this, CallWobbleISR);
    // clock interrupt INTT (duty timer B6) calls the wobble ISR
    EnableInterrupts (old);

    area_ = area;  // store the area
    gps_ = NULL;  // default is S-Node

    CClockInterface::Init();
}

CClockInterface::~CClockInterface ()
{
    Dispose();
    // this also disabled all interrupts and duty timers used
    // by this module

    // remove the ISRs
    INT32 old = DisableInterrupts();
    RemoveISR (GetClockIntvec (GetNodeID(), area_) | INTA, this);
    RemoveISR (GetClockIntvec (GetNodeID(), area_) | INTT, this, CallWobbleISR);
    EnableInterrupts (old);

    delete clock_;
}


//--------------------------------------------------------------------------
//--- reset all members to their initial values
//--------------------------------------------------------------------------

// restores the initial state
void CClockInterface::Reset ()
{
    CCrashingModule::Reset();  // reset crash state
    CApplControlledObject::Reset();  // reset other things
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
// known params: AmortizationDuration, ClockValue, WobbleDrift,
// Accuracy, PFast, PSlow, DistributionFast, DistributionRegular,
// DistributionSlow, PReadFault, PWriteFault, PDriftChange,
// and all of CClock (HW or SW), CApplControlledObject and CCrashingModule
EBool CClockInterface::SetParameter (const SParameter& param)
{
    SSingleParameter *p = (SSingleParameter*)&param;  // makes code easier
    SApplDistribution *dist = (SApplDistribution*)&param;
    SApplWobbleDrift *wobble = (SApplWobbleDrift*)&param;

    switch (param.type)
    {
        case PCrash:
            return CCrashingModule::SetProbability (True, p->value);
            break;
        case PRecover:
            return CCrashingModule::SetProbability (False, p->value);
            break;
        case AmortizationDuration:
            return SetAmortizationDuration (p->value);
            break;
        case ClockValue:
            clock_->SetMSSET (steplow_|(((STimeParameter*)&param)->macrostamp &
                                        0x00FFFFFF));
            clock_->SetTSSET (((STimeParameter*)&param)->timestamp);
            clock_->SetUSSET (0);
            clock_->SetNTPSET (0);  // take over the new time value
            break;
        case Accuracy:
            clock_->SetALPHASET ((INT32)p->value);
            clock_->SetALPHAPNSET (0);  // take over the new accuracies
            break;

        case DistributionFast:
        // draw random variables in (c+l, c+h) with mean c+m
            // conditions: l>=0, c+l>=0, m within (l,h), c+h <= WcRhoPlus
            if (!dist->data.DataValid()) return False;
            if (dist->data.constant+dist->data.high > WcRhoPlus) return False;

            distDataFast_ = dist->data;  // store the random data
            rhoPlus_ = distDataFast_.constant + distDataFast_.low;

            if (wobbleData_.rhoMax>rhoPlus_) wobbleData_.rhoMax = rhoPlus_;
            // adjust the wobble interval to the new drift bound

            distDataRegular_.high = rhoPlus_ - distDataRegular_.constant;
            // compute the upper bound of the regular interval

            break;

        case DistributionRegular:
            // conditions: c+m within (-rhoMinus, rhoPlus)
            if (-rhoMinus_ > dist->data.constant+dist->data.mean) return False;
            if (rhoPlus_ < dist->data.constant+dist->data.mean) return False;

            distDataRegular_.dist = dist->data.dist;
            distDataRegular_.mean = dist->data.mean + dist->data.constant-
                                    distDataRegular_.constant;
            distDataRegular_.freeparam = dist->data.freeparam;
            break;

        case DistributionSlow:
            // conditions: l>=0, c+l>=0, m within (l,h), c+h <= WcRhoMinus
            if (!dist->data.DataValid()) return False;
            if (dist->data.constant+dist->data.high > WcRhoMinus) return False;

            distDataSlow_ = dist->data;
            rhoMinus_ = distDataSlow_.constant + distDataSlow_.low;
            // rhoMinus is the lower bound of the resulting interval

            if (wobbleData_.rhoMin>rhoMinus_) wobbleData_.rhoMin = rhoMinus_;
            // adjust the wobble interval to the new drift bound

            distDataRegular_.high = rhoMinus_ + rhoPlus_;
            distDataRegular_.mean = distDataRegular_.mean +
                                    distDataRegular_.constant + rhoMinus_;
            distDataRegular_.low = 0;
            distDataRegular_.constant = -rhoMinus_;
            // the regular interval has to be computed again (its lower
            // bound is always zero), because the constant has changed

            break;

        case PFast:
            if ((p->value < 0) || (p->value+pSlow_ > 1)) return False;
            pFast_ = p->value;
            break;
        case PSlow:
            if ((p->value < 0) || (p->value+pFast_ > 1)) return False;
            pSlow_ = p->value;
            break;
        case PDriftChange:
            if ((p->value < 0) || (p->value > 1)) return False;
            pDriftChange_ = p->value;
            break;
        case PReadFault:
            if ((p->value < 0) || (p->value > 1)) return False;
            pReadFault_ = p->value;
            break;
        case PWriteFault:
            if ((p->value < 0) || (p->value > 1)) return False;
            pWriteFault_ = p->value;
            break;

        case WobbleDrift:
            // period >= 0, 0 <= rhoMax <= rhoPlus_, 0 <= rhoMin <= rhoMinus_,
            // distribution data is valid
            if (wobble->period < 0) return False;
            if (wobble->rhoMax < 0) return False;
            if (wobble->rhoMax > rhoPlus_) return False;
            if (wobble->rhoMin < 0) return False;
            if (wobble->rhoMin > rhoMinus_) return False;
            if (!wobble->data.DataValid()) return False;

            // take over the new values
            // note: the direction of the wobbling stays the same
            wobbleData_.distrib= wobble->data;
            wobbleData_.period = wobble->period;
            wobbleData_.rhoMax = wobble->rhoMax;
            wobbleData_.rhoMin = wobble->rhoMin;

            // set the duty timer accordingly
            if (wobbleData_.period == 0) DisableDutyTimer (IntDutyB6);
            // stop wobbling
            else InterruptAtMult (IntDutyB6, wobbleData_.period);
            // set the timer to the next multiple of the period

            break;

        default:  // unknown parameter - pass on to the clock
            if (CApplControlledObject::SetParameter (param)) return True;
            else return clock_->SetParameter (param);
    } // switch

    return True;
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned
// additional parameter: CrashStatus, RhoMinus, RhoPlus
EBool CClockInterface::GetParameter (SParameter& param) const
{
    SSingleParameter *p = (SSingleParameter*)&param;  // makes code easier
    SApplDistribution *dist = (SApplDistribution*)&param;
    SApplWobbleDrift *wobble = (SApplWobbleDrift*)&param;

    switch (param.type)
    {
        case CrashStatus:
            p->value = Crashed();
            break;
        case PCrash:
            p->value = CCrashingModule::GetProbability (True);
            break;
        case PRecover:
            p->value = CCrashingModule::GetProbability (False);
            break;
        case RhoMinus:
            p->value = rhoMinus_;
            break;
        case RhoPlus:
            p->value = rhoPlus_;
            break;
        case AmortizationDuration:
            p->value = amorttime_;
            break;
        case ClockValue:
            ((STimeParameter*)&param)->timestamp = clock_->GetTSGETL();
            ((STimeParameter*)&param)->macrostamp = clock_->GetMSGET() &
                                                    0x0FFFFFF;
            break;
        case Accuracy:
            p->value = GetAccuracy();
            break;
        case DistributionFast:
            dist->data = distDataFast_;
            break;
        case DistributionRegular:
            dist->data = distDataRegular_;
            break;
        case DistributionSlow:
            dist->data = distDataSlow_;
            break;
        case PFast:
            p->value = pFast_;
            break;
        case PSlow:
            p->value = pSlow_;
            break;
        case PDriftChange:
            p->value = pDriftChange_;
            break;
        case PReadFault:
            p->value = pReadFault_;
            break;
        case PWriteFault:
            p->value = pWriteFault_;
            break;
        case WobbleDrift:
            wobble->data = wobbleData_.distrib;
            wobble->period = wobbleData_.period;
            wobble->rhoMax = wobbleData_.rhoMax;
            wobble->rhoMin = wobbleData_.rhoMin;
            break;

        default:  // unknown parameter - pass on to the clock
            if (CApplControlledObject::GetParameter (param)) return True;
            else return clock_->GetParameter (param);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// returns True if the command was executed, else False
// known commands: ConnectGps, DisconnectGps, and all commands of
// CApplControlledObject and CCrashingModule
EBool CClockInterface::ExecuteCommand (const SCommand& command)
{
    SGpsConnectCommand *cmd = (SGpsConnectCommand *)&command;

    switch (command.type)
    {
        case ConnectGps:
            if (cmd->gpu >= EGpsUnitCount) return False;
            AddGps ((EGpsUnit)cmd->gpu);
            break;
        case DisconnectGps:
            if (cmd->gpu >= EGpsUnitCount) return False;
            RemoveGps ((EGpsUnit)cmd->gpu);
            break;

        default:  // pass on to the base class
            if (CApplControlledObject::ExecuteCommand (command)) return True;
            else return CCrashingModule::ExecuteCommand (command);
    } // switch

    return True;
}


//----------------------------------------------------------------------
//--- crash/recover
//----------------------------------------------------------------------

// stop the clock (STEP=STEPLOW=0)
void CClockInterface::Crash()
{
    SetStepRegisters (0, 0);  // this sets STEPPURE and STEPPLOW to zero
    CCrashingModule::Crash();
}

// set step and steplow to old values
void CClockInterface::Recover()
{
    SetStepRegisters (steppure_, steplow_);  // restore old step values
    CCrashingModule::Recover();
}


//--------------------------------------------------------------------------
//--- handle GPS receivers
//--------------------------------------------------------------------------

// returns True if there is at least one GPS receiver connected to the
// clock, otherwise False
EBool CClockInterface::GpsConnected ()
{
    if (gps_ != NULL) return True;
    else return False;
}


//--------------------------------------------------------------------------
//--- configuration of the amortization duration
//--------------------------------------------------------------------------

// stores the maximum amortization duration (only if 'amorttime' >= 0; and
// if the clock has GPS, amorttime must be in (0,1))
// the function returns True if the amortization duration was changed,
// otherwise False
// amorttime==0 causes function SetTimestamp() to set the clock
// instantaneously
EBool CClockInterface::SetAmortizationDuration (double amorttime)
{
    if ((amorttime < 0) ||
        ((gps_ != NULL) && ((amorttime <= 0) || (amorttime>=1)))) return False;
    // do not change the amortization duration

    amorttime_ = amorttime;
    return True;
}


//--------------------------------------------------------------------------
//--- access clock data
//--------------------------------------------------------------------------

// returns the rate uncertainty (upper bound)
double CClockInterface::GetRhoPlus () const
{
    return rhoPlus_;
}

// returns the rate uncertainty (lower bound)
double CClockInterface::GetRhoMinus () const
{
    return rhoMinus_;
}

// returns the current accuracy of the clock
INT32 CClockInterface::GetAccuracy () const
{
    return clock_->GetSTATEPGETL();  // latch and return the accuracies
}

// returns the ideal clock frequency
double CClockInterface::GetFrequency () const
{
    return clock_->GetFrequency();
}

// returns the external granularity of the clock
double CClockInterface::GetGranularity () const
{
    return clock_->GetGranularity();
}

// returns the duty timer granularity of the clock
double CClockInterface::GetGranularityDT () const
{
    return clock_->GetGranularityDT();
}

// returns the internal granularity of the clock
double CClockInterface::GetGranularityS () const
{
    return clock_->GetGranularityS();
}

// returns the accuracy transmission loss of the clock
double CClockInterface::GetGranularityAcc () const
{
    return clock_->GetGranularityAcc();
}

// returns the oscillator granularity of the clock
double CClockInterface::GetGranularityOsc () const
{
    return clock_->GetGranularityOsc();
}


//--------------------------------------------------------------------------
//--- clock access functions
//--------------------------------------------------------------------------

// return the NTP time and the accuracy
// injects faults; if there is a read fault, 'time' is set to arbitrary
// values and True is returned
// if there are no faults, the return value of the function
// depends on the number of GPS receivers connected to the clock: if
// there are no receivers, the function always returns True; if there
// are GPS receivers connected, the function returns the value of
// 'gpsStatus'
EBool CClockInterface::GetTimestamp (STimestamp& time)
{
    InjectFaults();  // inject all kinds of general faults
    if (random_->CastDice (pReadFault_))  // just return garbage
    {
        ReportEvent (Clock, ClockReadFault);
        time.ms = (INT32)random_->RandomVar (0, 0x1000000);
        time.ts = (INT32)random_->RandomVar (0, 0xFFFFFFFF);
        time.acc = (INT32)random_->RandomVar (0, 0xFFFFFFFF);
        return True;
    }

    // no errors (or crash, but that just stops the clock), therefore read
    // the normal clock value
    time.ts = clock_->GetTSGETL();  // latch everything
    time.ms = clock_->GetMSGET() & 0x0FFFFFF;
    time.acc = clock_->GetSTATEPGET();

    if (gps_ == NULL) return True;  // S-Node timestamps are always okay
    else return gpsStatus_;  // P-Node times are okay if GPS is okay
}

// set the NTP time and the accuracy; if amortTime=0, the clock is directly
// set to the new values (in that case, the accuracies are both set to the
// higher value in time.acc); if amortTime > 0, the differences are
// amortized (and the accuracies can be non-symetrical)
void CClockInterface::SetTimestamp (const STimestamp& time)
{
    SExtendedTimestamp xtime;
    xtime = time;  // us is automatically set to zero
    SetTimestamp (xtime); 
}

// may inject faults (crash, drift change); if there is a write fault,
// ms,ts,us and ns are set to random values; if the clock is crashed,
// nothing happens
void CClockInterface::SetTimestamp (const SExtendedTimestamp& time)
{
    if (gps_ != NULL) return;  // GPS clocks may not be set

    SExtendedTimestamp xtime;  // for faulty values

    InjectFaults();  // inject all kinds of general faults
    if (Crashed())
    {
        ReportEvent (Clock, IgnoreSetting);
        return;  // ignore the value
    }

    if (random_->CastDice (pWriteFault_))  // just set garbage
    {
        ReportEvent (Clock, ClockWriteFault);
        xtime.ms = (INT32)random_->RandomVar (0, 0x1000000);
        xtime.ts = (INT32)random_->RandomVar (0, 0xFFFFFFFF);
        xtime.us = (INT32)random_->RandomVar (0, 0xFFFFFFFF);
        xtime.acc = (INT32)random_->RandomVar (0, 0xFFFFFFFF);
    }
    else xtime = time;  // us is automatically set to zero

    // first report the update (-->old time is used for the event timestamp)
    SApplClockEvent *event = new SApplClockEvent;
    event->module = Clock;
    event->type = ClockUpdate;
    event->newTime = (STimestamp)xtime;
    ReportEvent (event);

    if (amorttime_ == 0)  // set instantaneously
    {
        // preload the time value
        clock_->SetMSSET (steplow_ | (xtime.ms & 0x00FFFFFF));
        clock_->SetTSSET (xtime.ts);
        clock_->SetUSSET (xtime.us);

        // preload the accuracy value (set both accuracies to the higher
        // value)
        if ((xtime.acc & 0x0FFFF) > (xtime.acc>>16 & 0x0FFFF))
            clock_->SetALPHASET (xtime.acc<<15 & 0x7FFF0000);
        else clock_->SetALPHASET (xtime.acc>>1 & 0x7FFF0000);

        // take over the preloaded values
        clock_->SetNTPSET (0);  // take over the new time value
        clock_->SetALPHAPNSET (0);  // take over the new accuracy value
        return;
    }


    INT32 alphaPlus;  // for the accuracy differences
    INT32 alphaMinus;

    // compute the accuracy differences to the current clock values
    ComputeAccuracyCorrection (xtime.acc, 0, alphaPlus, alphaMinus);

    SExtendedTimestamp currTime;
    GetTimestamp (currTime);  // obtain the current clock value
    currTime.us = clock_->GetUSGET();

    // try to amortize the time and accuracy differences
    Amortize (NtpToSeconds (xtime) - NtpToSeconds (currTime),
              (SINT32)alphaPlus, (SINT32)alphaMinus);
}

// if 'amortTime' is zero, the functions call SetTimestamp() to set the
// timestamp at once; otherwise, PreloadAmort() is called to preload
// the amortization data, and then duty timer B1 is programmed to fire
// at time 'loadAt', thus starting the amortization phase
void CClockInterface::PreloadTimestamp (double loadAt, const STimestamp& time)
{
    SExtendedTimestamp xtime;
    xtime = time;  // us is automatically set to zero
    PreloadTimestamp (loadAt, xtime); 
}

void CClockInterface::PreloadTimestamp (double loadAt,
    const SExtendedTimestamp& time)
{
    if (gps_ != NULL) return;  // GPS clocks may not be changed

    SExtendedTimestamp xtime;  // for faulty values

    InjectFaults();  // inject all kinds of general faults
    if (Crashed())
    {
        ReportEvent (Clock, IgnoreSetting);
        return;  // ignore the value
    }

    if (random_->CastDice (pWriteFault_))  // just set garbage
    {
        ReportEvent (Clock, ClockWriteFault);
        xtime.ms = (INT32)random_->RandomVar (0, 0x1000000);
        xtime.ts = (INT32)random_->RandomVar (0, 0xFFFFFFFF);
        xtime.us = (INT32)random_->RandomVar (0, 0xFFFFFFFF);
        xtime.acc = (INT32)random_->RandomVar (0, 0xFFFFFFFF);
    }
    else xtime = time;  // us is automatically set to zero

    if (amorttime_ == 0)  // set instantaneously
    {
        SetTimestamp (xtime);
        return;
    }


    // first report the update
    SApplPreloadEvent *event = new SApplPreloadEvent;
    event->module = Clock;
    event->type = ClockPreload;
    event->newTime = (STimestamp)xtime;
    event->loadAt = loadAt;
    ReportEvent (event);


    INT32 alphaPlus;  // for the accuracy differences
    INT32 alphaMinus;

    SExtendedTimestamp currTime;  // the current clock time
    GetTimestamp (currTime);  // obtain the current clock value
    currTime.us = clock_->GetUSGET();  // get us as well

    // compute the accuracy differences to the clock values at time 'loadAt'
    ComputeAccuracyCorrection (xtime.acc, loadAt-NtpToSeconds (currTime),
                               alphaPlus, alphaMinus);

    // load the amortization registers
    PreloadAmort (NtpToSeconds (xtime) - loadAt,
                  (SINT32)alphaPlus, (SINT32)alphaMinus);

    InterruptAt (IntDutyB1, loadAt);
    // program the amortization duty timer (B1)
}


//--------------------------------------------------------------------------
//--- access to other clock registers
//--------------------------------------------------------------------------

// sets the STEPPURE register; if the clock is crashed, the value is not set
void CClockInterface::SetSteppure (INT32 step)
{
    if (Crashed())  // ignore the set request
    {
        ReportEvent (Clock, IgnoreSetting);
        return;
    }

    steppure_ = step;  // take over the new value

    if (deltaRho_ != 0) SetFaultyDrift (deltaRho_);
    // the faulty drift will change, because it depends on the steppure
    // value; calling SetFaultyDrift() also calls SetStepRegisters()
    else SetStepRegisters (steppure_, steplow_);  // take over new values
}

// returns the current value of the STEPPURE register
INT32 CClockInterface::GetSteppure ()
{
    return steppure_;
}

// sets the STEPLOW register; if the clock is crashed, the value is not set 
void CClockInterface::SetSteplow (INT8 steplow)
{
    if (Crashed())  // ignore the set request
    {
        ReportEvent (Clock, IgnoreSetting);
        return;
    }

    steplow_  = (INT32)steplow<<24 & 0xFF000000;  // take over the new value

    if (deltaRho_ != 0) SetFaultyDrift (deltaRho_);
    // the faulty drift will change, because it depends on the steplow
    // value; calling SetFaultyDrift() also calls SetStepRegisters()
    else SetStepRegisters (steppure_, steplow_);  // take over new values
}

// sets the LAMBDAPURE registers; if the clock is crashed, the values are
// not set
// note: if you want to use PreloadTimestamp() to preload a clock
// value at a given time in the future, then you should only change
// the LAMBDA values of the clock through this function; setting the
// registers directly will result in wrong computation of the
// correction values for the accuracy!
void CClockInterface::SetLambdapure (INT16 lambdaP, INT16 lambdaN)
{
    if (Crashed())
    {
        ReportEvent (Clock, IgnoreSetting);
        return;
    }
    clock_->SetLAMBDAPPURE (lambdaP);
    clock_->SetLAMBDANPURE (lambdaN);
    lambdaP_ = lambdaP;  // store the values for preloaded accuracies
    lambdaN_ = lambdaN;
}

// sets the BOUNDP(N) register (the upper 16 bits of 'bounds' are the
// value of BOUNDN, the lower 16 bits the value of BOUNDP); 
// if the clock is crashed, the value is not set
void CClockInterface::SetBounds (INT32 bounds)
{
    if (Crashed())
    {
        ReportEvent (Clock, IgnoreSetting);
        return;
    }
    clock_->SetBOUNDP (bounds);
}


//--------------------------------------------------------------------------
//--- interrupt handling
//--------------------------------------------------------------------------

// enable the interrupt of the source; that assumes that the vector
// has already been set
void CClockInterface::EnableInterrupt (EInterruptSource source)
{
    Assert ((source >= 0) && (source < 64));

    if (source < 32)  // source is in UTCINTEN1
    {
        clock_->SetUTCINTCLEAR1 (SourceToMask (source));  // clear IP
        clock_->EnableInterrupt (area_, 1<<source, 0);  // enable interrupt
    }
    else  // source is in UTCINTEN2
    {
        clock_->SetUTCINTCLEAR2 (SourceToMask (source));
        clock_->EnableInterrupt (area_, 0, 1<<(source-32));
    }
}

// if the interrupt of the source was enabled, it is turned off; else
// nothing happens
void CClockInterface::DisableInterrupt (EInterruptSource source)
{
    Assert ((source >= 0) && (source < 64));

    if (source < 32)  // source is in UTCINTEN1
    {
        clock_->DisableInterrupt (area_, 1<<source, 0);// disable interrupt
    }
    else  // source is in UTCINTEN2
    {
        clock_->DisableInterrupt (area_, 0, 1<<(source-32));
    }
}


// returns True if the IP bit of the source is set, else False
EBool CClockInterface::InterruptPending (EInterruptSource source)
{
    Assert ((source >= 0) && (source < 64));

    if (source < 32)  // source is in UTCINTSTAT1
    {
        if (clock_->GetUTCINTSTAT1() & SourceToMask(source)) return True;
    }
    else  // source is in UTCINTSTAT2
    {
        if (clock_->GetUTCINTSTAT2() & SourceToMask(source)) return True;
    }

    return False;
}

// returns True if an interrupt of the given type is pending, else
// False
EBool CClockInterface::InterruptPending (EClockInterrupt inttype)
{
    switch (inttype)
    {
        case INTA:
            if (clock_->GetUTCINTSTAT1() & INTAReg1) return True;
            break;
        case INTN:
            if (clock_->GetUTCINTSTAT2() & INTNReg2) return True;
            break;
        case INTT:
            if (clock_->GetUTCINTSTAT1() & INTTReg1) return True;
            if (clock_->GetUTCINTSTAT2() & INTTReg2) return True;
            break;

        default:
            Assert (False);
    } // switch

    return False;
}


// if the interrupt of the source was set, it is cleared; else nothing
// happens
void CClockInterface::ClearInterrupt (EInterruptSource source)
{
    Assert ((source >= 0) && (source < 64));

    if (source < 32)
    {
        clock_->SetUTCINTCLEAR1 (SourceToMask(source));
    }
    else
    {
        clock_->SetUTCINTCLEAR2 (SourceToMask(source));
    }
}

// clears all interrupts of the given interrupt type (INTA, INTN, INTT)
void CClockInterface::ClearInterrupt (EClockInterrupt inttype)
{
    switch (inttype)
    {
        case INTA:
           clock_->SetUTCINTCLEAR1 (INTAReg1);
           break;
        case INTN:
           clock_->SetUTCINTCLEAR2 (INTNReg2);
           break;
        case INTT:
           clock_->SetUTCINTCLEAR1 (INTTReg1);
           clock_->SetUTCINTCLEAR2 (INTTReg2);
           break;

       default:
           Assert (False);  // should never happen
    } // switch
}


// produces a duty timer interrupt at clock time 'at' (in seconds)
void CClockInterface::InterruptAt (EInterruptSource source, double at)
{
    CNtpTime time;
    SecondsToNtp (at, time);  // 'at' in NTPTIME
    SetDutyTimer (source, time);  // program the duty timer
}

// produces a duty timer interrupt in 'after' (clock) seconds
void CClockInterface::InterruptAfter (EInterruptSource source, double after)
{
    CNtpTime time = GetCurrentClockTime ();
    CNtpTime addvalue;
    SecondsToNtp (after, addvalue);  // 'after' in NTPTIME
    time = time + addvalue;  // add 'after' to the current clock value
    SetDutyTimer (source, time);  // program the duty timer
}

// produces a duty timer interrupt when the clock time is a multiple of
// 'multTime' (in seconds);
// example: if multTime is 100 sec, and clock time is 185, the clock
// will raise an interrupt at clock time 200
// the function returns the (clock) time at which the interrupt is
// triggered
double CClockInterface::InterruptAtMult (EInterruptSource source,
    double multTime)
{
    // get the current clock time
    CNtpTime time = GetCurrentClockTime();
    double currTime = NtpToSeconds (time);

    // compute the next multiple
    int mult = (int)(currTime/multTime) + 1;

    // set the duty timer to that value
    SecondsToNtp (mult*multTime, time.ms, time.ts);
    SetDutyTimer (source, time);  // program the duty timer

    if (time.ms & 0xFF000000) Report (overflow in InterruptAtMult);
    // currently, overflows are only reported, but not handled

    // return the next multiple, but adapted to the duty timer granularity
    modf (mult*multTime/GetGranularityDT(), &currTime);
    // use 'currTime' as temporary to store the integral part (cutting off
    // the fractional part is necessary to get a multiple of the duty timer
    // granularity)
    return currTime*GetGranularityDT();
}

// if 'source' is a duty timer, DUTYL is set to zero (thus disabling
// the timer)
void CClockInterface::DisableDutyTimer (EInterruptSource source)
{
    ESSUnit ssu;

    switch (source)
    {
        case IntDutyA1: ssu = SSU1; break;
        case IntDutyA2: ssu = SSU2; break;
        case IntDutyA3: ssu = SSU3; break;
        case IntDutyA4: ssu = SSU4; break;
        case IntDutyA5: ssu = SSU5; break;
        case IntDutyA6: ssu = SSU6; break;

        case IntDutyB1: ssu = SSU1; break;
        case IntDutyB2: ssu = SSU2; break;
        case IntDutyB3: ssu = SSU3; break;
        case IntDutyB4: ssu = SSU4; break;
        case IntDutyB5: ssu = SSU5; break;
        case IntDutyB6: ssu = SSU6; break;

        default: return;  // other timers are not supported
    } // switch

    // disable the timer
    switch (source)  // determine if A or B is used
    {
        case IntDutyA1: case IntDutyA2: case IntDutyA3:
        case IntDutyA4: case IntDutyA5: case IntDutyA6:
            clock_->SetDUTYAL (ssu, 0);  // disable the timer
            break;

        case IntDutyB1: case IntDutyB2: case IntDutyB3:
        case IntDutyB4: case IntDutyB5: case IntDutyB6:
            clock_->SetDUTYBL (ssu, 0);  // disable the timer
            break;

        default: Assert (False);  // currently not supported
    } // switch
}


//--------------------------------------------------------------------------
//--- get the clock itself
//--------------------------------------------------------------------------

// returns the clock itself; only to be used if direct access to the
// clock registers is absolutely necessary!
CClock *CClockInterface::GetClock()
{
    return clock_;
}


//--------------------------------------------------------------------------
//--- initialization and destruction of the data members
//--------------------------------------------------------------------------

// initialize all members; enable the snapshot and DTB6 interrupts
void CClockInterface::Init ()
{
    Assert (gps_ == NULL);

    random_ = new CRandom (GetNodeID());  // init. random stream with nodeID

    deltaSteppure_ = 0;
    deltaSteplow_ = 0;
    deltaRho_ = 0;

    rhoMinus_ = WcRhoMinus;
    rhoPlus_ = WcRhoPlus;
    amorttime_ = DefaultAmortizationDuration;
    gpsStatus_ = False;

    pReadFault_ = 0;
    pWriteFault_ = 0;
    pDriftChange_ = 0;
    pFast_ = 0;
    pSlow_ = 0;

    distDataFast_.dist = Uniform;
    distDataFast_.constant = 0;
    distDataFast_.low = rhoPlus_;
    distDataFast_.high = rhoPlus_;
    distDataFast_.mean = rhoPlus_;
    distDataFast_.freeparam = 0;

    distDataSlow_.dist = Uniform;
    distDataSlow_.constant = 0;
    distDataSlow_.low = rhoMinus_;
    distDataSlow_.high = rhoMinus_;
    distDataSlow_.mean = rhoMinus_;
    distDataSlow_.freeparam = 0;

    distDataRegular_.dist = Uniform;
    distDataRegular_.constant = 0;
    distDataRegular_.low = 0;
    distDataRegular_.high = 0;
    distDataRegular_.mean = 0;
    distDataRegular_.freeparam = 0;

    // note: at this time, the wobble duty timer is already disabled,
    // either because Init() was called from the ctor and the timers are
    // not enabled anyway, or because it was called from Reset(), which
    // executes Dispose() before calling Init(); so we can reset the wobble
    // data without harm
    wobbleData_.period = 0;  // no wobbling
    wobbleData_.addDrift = True;  // start with adding rhoAdd
    wobbleData_.rhoMax = 0;
    wobbleData_.rhoMin = 0;
    wobbleData_.distrib.dist = Uniform;
    wobbleData_.distrib.constant = 0;
    wobbleData_.distrib.low = 0;
    wobbleData_.distrib.high = 0;
    wobbleData_.distrib.mean = 0;
    wobbleData_.distrib.freeparam = 0;

    InitializeClock();
    // initialize important clock registers

    EnableInterrupt (IntSnapshot);  // enable the snapshot interrupt
    EnableInterrupt (IntDutyB6);  // enable the wobble interrupt
}

// delete 'gps' -> go back to S-Node
// all duty timers and interrupts enabled by this class are disabled;
// other interrupts and timers should be disabled in the classes that
// have enabled them!
void CClockInterface::Dispose ()
{
    if (gps_ != NULL) delete gps_;  // that disables all enabled GPS receivers
    gps_ = NULL;
    delete random_;

    // disable all interrupts used by this class (Snapshot and DTB6);
    // also disable all duty timers used by this class (B1 and B6)

    DisableInterrupt (IntSnapshot);  // disable the snapshot interrupt
    DisableInterrupt (IntDutyB6);  // disable the wobble interrupt
    // only disable the interrupts that are used by this module
/**/// note: for safety purposes, EnableInterrupt() could store which
    // interrupts were enabled by the application; disable all that
    // are still enabled when the dtor is called

    DisableDutyTimer (IntDutyB1);
    // disable the amortization duty timer B1, even if it was not used

    DisableDutyTimer (IntDutyB6);
    // disable the wobble duty timer B6

}


// initialize the clock; the function does not change the
// UTCINTEN registers
void CClockInterface::InitializeClock ()
{
    clock_->SetAMORTTIMESET (0);  // stop an ongoing amortization phase
    clock_->SetSTARTAMORT (1);

    clock_->SetLAMBDAPPURE (0);  // no deterioration of accuracies
    clock_->SetLAMBDANPURE (0);
    lambdaP_ = 0;  // set the internal variables to the LAMBDA values
    lambdaN_ = 0;
    clock_->SetSTATEPSET (0);  // no state correction during amort.
    clock_->SetSTATENSET (0);
    clock_->SetBOUNDP (0xFFFFFFFF);  // accuracy bounds are disabled

    clock_->SetALPHASET (0x7FFFFFFF);  // accuracies set to maximum
    clock_->SetALPHAPNSET (0);  // take over the accuracies


    clock_->SetUTCCONF1 (clock_->GetUTCCONF1() & ~SUtcsuWrite::CPOLHWSNAP);
    // set the polarity of the HW-Snapshot to falling edge (hw-based
    // simulation uses wired-or)

    // compute the initial values for step and steplow, according to
    // the "ideal" frequency of the clock
    double f = clock_->GetFrequency();
    double dummy;
    INT8 steplow = (INT8)(modf (TwoPow51/f, &dummy) * 256);
    steplow_  = (INT32)steplow<<24 & 0xFF000000;
    steppure_ = (INT32)dummy;

    SetStepRegisters (steppure_, steplow_);
    // set the step registers of the clock to steppure_ and steplow_
    // note: at this time we never have a faulty drift, because
    // InitializeClock() is always called as the last statement of
    // Init(), and Init() resets the faulty drift to zero

    // set the initial clock value
    INT32 ms, ts, us;
    SecondsToNtp (GetGlobalTime(), ms, ts, us);
    // set the initial clock value to the current global time
    // note: this avoids problems during the initial joining phase, when
    // a new node is integrated into the system; normally, it is the
    // responsibility of the algorithm to join its node, that is to
    // extract a valid clock value from the clock synchronization data of
    // the other nodes and to set its own clock to this value; setting
    // the clock value to the current global time, which is the same as
    // the "GPS time" makes the joining phase easier to handle, because
    // the node will automatically be more or less synchronized to the
    // other nodes
    // note: this also works in the hardware-based simulation if there
    // is a master clock which offers a global time that is related to
    // the external time
    // if you do not have a master clock, then GetGlobalTime() should
    // return 0 or some other constant value, and in this case, it is the
    // responsibility of the algorithm to integrate the node into the
    // running system
    // if the master clock is not related to external time, the effect
    // is much the same as that described above: again, the algorithm
    // must take care of the joining problem

    clock_->SetMSSET (steplow_ | (ms & 0x00FFFFFF));
    // note: here, we take advantage of the fact that there is no
    // faulty drift; otherwise, we would override it through the access
    // to MSSET
    clock_->SetTSSET (ts);
    clock_->SetUSSET (us);
    clock_->SetNTPSET (0);  // take over the new time value
}


//--------------------------------------------------------------------------
//--- rate changes
//--------------------------------------------------------------------------

// sets STEPPURE and STEPLOW to the parameters (bits 0-23 of steplow are
// masked out before setting STEPLOW in MSSET); if steppure or steplow
// are not 0, the delta values of 'deltaSteppure' and 'deltaSteplow' are
// added and the clock registers are set to the result; if both steppure
// and steplow are 0, then the registers are set to 0 as well (stopping
// the clock)
// note: steppure_ and steplow_ are not changed
void CClockInterface::SetStepRegisters (INT32 steppure, INT32 steplow)
{
    steplow = steplow & 0xFF000000;  // mask out macrostamp bits

    if ((steplow!=0) || (steppure!=0))
    {
        // add deltaSteppure to the requested steppure value
        if (deltaSteppure_ != 0)  // add a faulty drift in STEPPURE
        {
            if (deltaSteplow_ & 0x200)  // subtract delta from steppure
            {
                if (deltaSteppure_ > steppure) steppure = 0;
                else steppure = steppure-deltaSteppure_;
            }
            else
            {
                if (deltaSteppure_+steppure < steppure) steppure=0xFFFFFFFF;
                else steppure += deltaSteppure_;
            }
        } // if

        // add deltaSteplow to the requested steplow value
        if ((deltaSteplow_&0x0FF) != 0)  // add a faulty drift in STEPLOW
        {
            if (deltaSteplow_ & 0x100)  // subtract delta from steplow
            {
                if ((deltaSteplow_<<24 & 0xFF000000) > steplow) steplow = 0;
/**/// shouldn't there be a carry taken from steppure? can such an underflow
    // even happen?
                else steplow = steplow-(deltaSteplow_<<24 & 0xFF000000);
            }
            else
            {
                if ((deltaSteplow_<<24 & 0xFF000000)+steplow < steplow)
                    steplow = 0xFF000000;
/**/// shouldn't there be an overflow into steppure? can such an overflow
    // even happen?
                else steplow += deltaSteplow_<<24 & 0xFF000000;
            }
        } // if
    } // if steplow or steppure != 0

    clock_->SetMSSET (steplow);
    clock_->SetSTEPPURE (steppure);  // preload step (pure phase)
    clock_->SetAMORTTIMESET (1);  // short amortization
    clock_->SetSTARTAMORT (1);  // to take over the step pure value

    clock_->SetAMORTTIMESET (0);  // turn off amortization
    clock_->SetSTARTAMORT (1);
    // note: explicitly turning off the amortization is only needed for
    // pure simulation; in hardware-based simulation, the amortization
    // phase is already completed when the function returns
    // in pure simulation, a problem occurs if the clock is set to a new
    // value afterwards and if after the state change CSoftwareClock::Update()
    // is called; in that case, the amortization cycle is updated even
    // though in reality it would already be finished)
}


//--------------------------------------------------------------------------
//--- fault injection
//--------------------------------------------------------------------------

// create an event of type 'CrashStatusChange', send it to the evaluation
void CClockInterface::ReportCrashChange ()
{
    SApplCrashEvent *event = new SApplCrashEvent;
    event->module = Clock;
    event->type = CrashStatusChange;
    event->crashed = Crashed();
    CControlledObject::ReportEvent (event);
}

// modify crash status, maybe change state and rate
void CClockInterface::InjectFaults()
{
    double rho;  // the new drift (negative if clock is slower than 1)

    //--- crash/recover the module
    ModifyCrashStatus();

    //--- change the rate
    if (!random_->CastDice (pDriftChange_)) return;  // no drift change

    if (random_->CastDice (pSlow_))  // very slow
    {
        rho = -random_->RandomVar (distDataSlow_);
    }
    else if (random_->CastDice (pFast_/(1-pSlow_)))  // very fast
    {
        rho = random_->RandomVar (distDataFast_);
    }
    else  // within expected bounds
    {
        rho = random_->RandomVar (distDataRegular_);
    }

    Assert (rho >= -1);  // this should always be the case

    SetFaultyDrift (rho);  // set the new drift
}

// sets deltaSteppure and deltaSteplow so that the clock drifts by 'rho'
// and calls SetStepRegisters() to take over this drift
void CClockInterface::SetFaultyDrift (double rho)
{
    // compute the values for step and steplow

    double faultySteppure;  // faulty steppure value
    INT16 faultySteplow;  // faulty steplow value
    // note: adding the delta values to steppure_ and steplow_ should
    // result in the faultyStep... values

    faultySteppure = (1+rho)*steppure_;
    // faulty steppure adds the drift to the current steppure
    // note: we have to use steppure_ instead of computing the value
    // from GetFrequency(), because the user may have changed the step
    // value due to rate correction and we have to keep this change

    faultySteplow = (INT8)(steplow_>>24);
    faultySteplow = (INT16)(faultySteplow*(1+rho));
    // faulty steplow adds the drift to the current steplow

    // check overflow in faultySteplow
    if (faultySteplow & 0x0FF00)  // overflow occured
    {
        faultySteppure += (faultySteplow >> 8) & 0x0FF;
        // put the overflow into faultySteppure
        faultySteplow = faultySteplow & 0x0FF;
        // remove it from faultySteplow
    } // if

    // check overflow in faultySteppure
    if (faultySteppure > 0xFFFFFFFF)  // overflow -> set to maximum
    {
        faultySteppure = 0xFFFFFFFF;
        faultySteplow = 0x0FF;
    }
    else faultySteppure = (INT32)(faultySteppure);
    // no overflow -> just cut off the fractional part


    // we have the faultyStep values; now we compute the delta values

    if ((INT8)(steplow_>>24) <= faultySteplow)
        deltaSteplow_ = faultySteplow - (INT8)(steplow_>>24);
    else deltaSteplow_ = ((INT8)(steplow_>>24) - faultySteplow) + 0x100;
    // store absolute difference and set the sign bit

    if (steppure_ <= (INT32)faultySteppure)
        deltaSteppure_ = (INT32)faultySteppure - steppure_;
    else
    {
        deltaSteppure_ = (steppure_ - (INT32)faultySteppure);
        deltaSteplow_ += 0x200;  // store the sign bit in deltaSteplow
    }


    SetStepRegisters (steppure_, steplow_);
    // set step registers to take over the new deltas

    deltaRho_ = rho;  // store the drift in its double value
}

// returns the current value of the faulty drift
double CClockInterface::GetFaultyDrift ()
{
    return deltaRho_;
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int CClockInterface::ReportError (EReportType type, char *file, int line,
                                  char *expression)
{
    return ::ReportError (type, file, line, expression, Clock,
                          GetNodeID(), area_);
}


//--------------------------------------------------------------------------
//--- snapshot ISR
//--------------------------------------------------------------------------

// called if a HW snapshot interrupt (INTA) occurs; reads the NTP-time and
// sends an event to the Supervisor; clears the interrupt
void CClockInterface::isr()
{
    if (!InterruptPending (IntSnapshot)) return;  // no supported interrupt
    ClearInterrupt (IntSnapshot);  // clear the snapshot bit

    SApplClockEvent *event = new SApplClockEvent;
    event->module = Clock;
    event->type = Snapshot;
    event->newTime.ms = (clock_->GetMSSNU() & 0x0FFFFFF); // mask out checksum
    event->newTime.ts = clock_->GetTSSNU();
    event->newTime.acc = clock_->GetACCPSNU();

    ReportEvent (event);
}


//--------------------------------------------------------------------------
//--- drift wobbling ISR
//--------------------------------------------------------------------------

// called if a duty timer B6 interrupt (INTT) occurs; gets a value to add to
// the drift and calls SetFaultyDrift() with the new value; if the current
// drift is increased or decreased is determined by wobbleData.addDrift;
// if increasing the drift goes above wobbleData.RhoMax, then the drift
// is decreased; if decreasing it goes below -wobbleData.RhoMin, it is
// increased; if both increasing or decreasing the drift would set the
// drift outside the wobble bounds, the drift is not modified at all;
// if the current drift is already outside the wobble bounds, then it is
// modified so that it goes into the direction of the nearest bound
// clears the interrupt
// context: ISR
void CClockInterface::wobble_isr()
{
    if (!InterruptPending (IntDutyB6)) return;  // no supported interrupt
    ClearInterrupt (IntDutyB6);  // clear the DTB6 bit

    ReportEvent (Clock, TimerEvent);

    double rho = GetFaultyDrift();  // get the current drift value
    double rhoAdd = random_->RandomVar (wobbleData_.distrib);
    // get the incremental drift value

    // now check if we should add or subtract the new value from the
    // current drift; also check if we should add it at all (it can
    // happen that the bounds are so tight that we can neither add nor
    // subtract the value without stepping outside; in this case, we
    // do not modify the drift at all)
    // on the other hand, if we are already outside the bounds due to
    // fault injection, then we should step into the direction of the
    // bound (alternative: set the drift to the bound)

    INT8 dirChange = 0;  // number of direction changes in code below
    // we need it to determine if the new value oversteps both bounds

    if (wobbleData_.addDrift && (rho+rhoAdd > wobbleData_.rhoMax))
    // the new drift would be over the upper bound -> go down
    {
        wobbleData_.addDrift = False;
        dirChange++;
    } // if

    if (!wobbleData_.addDrift && (rho-rhoAdd < -wobbleData_.rhoMin))
    // the new drift would be under the lower bound -> go up
    {
        wobbleData_.addDrift = True;
        dirChange++;
    } // if

    if (dirChange > 1)  // we were changing direction twice
    // so either the current drift is within the bounds and we would
    // overstep both when applying rhoAdd, or the current drift is
    // outside the bounds and applying rhoAdd does not bring it into
    // the bounds either
    {
        if ((rho <= wobbleData_.rhoMax) && (rho >= -wobbleData_.rhoMin))
        // the current drift is within the bounds
        {
            rhoAdd = 0;  // do not add anything at all
        }
        else  // the current drift is outside the bounds
        // -> we have to determine the wobble direction again
        {
            if (rho > wobbleData_.rhoMax)
            // current drift is above upper bound -> go down
            {
                wobbleData_.addDrift = False;
                // check if going down oversteps the lower bound,
                // if it does, then stop at the lower bound
                if (rho-rhoAdd < -wobbleData_.rhoMin)
                    rhoAdd = rho+wobbleData_.rhoMin;
            }
            else
            // current drift is below lower bound -> go up
            {
                wobbleData_.addDrift = True;
                // check if going up oversteps the upper bound,
                // if it does, then stop at the upper bound
                if (rho+rhoAdd > wobbleData_.rhoMax)
                    rhoAdd = wobbleData_.rhoMax-rho;
            }
        } // else
    } // if

    if (wobbleData_.addDrift) SetFaultyDrift (rho+rhoAdd);
    else SetFaultyDrift (rho-rhoAdd);
    // set the new drift

    if (wobbleData_.period == 0)
    // should never be the case because SetParameter() disables the
    // duty timer if the period is zero, but it does not hurt to check it
    {
        Report (wobble period is zero in wobble ISR);  // report it
        DisableDutyTimer (IntDutyB6);  // stop wobbling
    }
    else if ((wobbleData_.rhoMax == 0) && (wobbleData_.rhoMin == 0))
    // this can happen if the bounds for rhoPlus/Minus are changed to 0
    // after the WobbleDrift has been set, or if they were originally
    // set to zero; the SetParameter() code does not check for this,
    // it relies on the fact that this case will be recognized here
    // note: we should only check if the bounds are both zero, not
    // if they are equal, because rhoMin is a bound for negative drifts
    // and RhoMax a bound for positive drifts; therefore, equal values
    // do not imply that the interval is a point; only if both values
    // are zero then the interval is a point, and in this case we cannot
    // wobble the drift until the interval is set again with parameter
    // WobbleDrift, so we can safely turn off the duty timer and stop
    // drift wobbling
    {
        DisableDutyTimer (IntDutyB6);  // stop wobbling
    }
    else
    // produce interrupt at the next multiple of the period
    {
        // note: producing the next interrupt requires different code
        // for hardware-based and for pure simulation:
        // in pure simulation, we have the problem that the interrupt
        // process may be scheduled a little bit before the time at
        // which it should occur (this is due to rounding errors when
        // setting the scheduling time, and I believe that these errors
        // are due to the fact that not all numbers can be represented
        // by a 'double'; so if you happen to have an interrupt scheduled
        // at a time that cannot be represented, then it will occur
        // a little bit (1e-11 secs or less) before the time you wanted
        // it to occur); but InterruptAtMult() is designed to produce
        // an interrupt at the next multiple of the given period, seen
        // from the current time on; since the current time is a little
        // bit *before* the multiple of the current interrupt, the next
        // multiple is the same as the currently handled one!
        // there are two solutions to this problem: (1) you always
        // get the current time and check if the return value of
        // InterruptAtMult() is one period away from the current time,
        // or (2) you simply call InterruptAfter() with the period;
        // both solutions have advantages and disadvantages:
        // (1) has the advantage, that it occurs exactly at a multiple of
        // the period if the time of the current interrupt could be
        // represented; the disadvantage is that needs to get the current
        // clock time and that it may have to call InterruptAt() to correct
        // the time set by InterruptAtMult()
        // (2) is only one line long, but the occurence of interrupts will
        // drift away from the multiple of the period, because whenever a
        // time cannot be represented, the interrupt will occur a little bit
        // earlier that the full period, and using InterruptAfter() you
        // cannot compensate for this by making the next period a little
        // bit longer, like the call to InterruptAt() does

#ifdef PSOS_SIM
        InterruptAtMult (IntDutyB6, wobbleData_.period);
        // in hardware-based simulation, this will always work because
        // the interrupt is raised directly by the UTCSU
        // note: even if it happens that the desired period is not a
        // multiple of the duty timer granularity, the time for executing
        // the above code is long enough that we will have passed the
        // current multiple by the time we call InterruptAtMult()
#else
        // we are in pure simulation; you should define at least one of
        // WOBBLE_USE_VERSION_1 or WOBBLE_USE_VERSION_2
        // note that it is not harmful to uncomment both; in this case,
        // the last version will be executed, overriding the previous version

  //#define WOBBLE_USE_VERSION_1  // sometimes uses InterruptAt()
  #define WOBBLE_USE_VERSION_2  // always uses InterruptAfter()

  #ifdef WOBBLE_USE_VERSION_1
        // this is the implementation of version (1)

        double nextInterrupt;  // the time of the next interrupt
        STimestamp time;  // for the current clock time
        GetTimestamp (time);  // get current clock time

        nextInterrupt = InterruptAtMult (IntDutyB6, wobbleData_.period);
        // set the interrupt, but get the time to check if it is okay

        if (nextInterrupt < NtpToSeconds (time) + wobbleData_.period)
        // this is not the interrupt we want! we want the next one
        {
            InterruptAt (IntDutyB6, nextInterrupt + wobbleData_.period);
            // manually set the interrupt to the next period
        }
  #endif // VERSION_1

  #ifdef WOBBLE_USE_VERSION_2
        // this is the implementation of version (2)

        InterruptAfter (IntDutyB6, wobbleData_.period);
        // produce an interrupt in 'period' seconds
  #endif // VERSION_2

#endif // not PSOS_SIM
    } // else
}


//--------------------------------------------------------------------------
//--- connect/disconnect GPS receivers
//--------------------------------------------------------------------------

// enable interrupts from the GPS receiver on GPS unit 'gpu'; the CSA will
// from now on be a P-Node, so the clock will refuse to take over new
// clock values set by the clock synchronization algorithm
// assert: gpu < EGpsUnitCount
void CClockInterface::AddGps (EGpsUnit gpu)
{
    Assert (gpu < EGpsUnitCount);

    if (gps_ == NULL)  // create the GPS class
    {
        gps_ = gpsTemplate->NewGPS (this, area_);  // get a GPS object
        Assert (gps_ != NULL);

        if (amorttime_ >= 1) amorttime_ = GpsAmort;  // use short amort. phase
    }

    INT32 old = DisableInterrupts();
    SetISR (GetGpsIntvec (GetNodeID(), area_, gpu), gps_);
    // the GPS interrupt calls the default ISR of the GPS class (which
    // reads the GPS time)
    EnableInterrupts (old);

    gps_->EnableGps (gpu);  // enable the receiver
}

// disables interrupts from the GPS receiver in GPS unit 'gpu'; if there
// are no further GPS receivers enabled, the node type will switch to
// S-Node and the clock will accept new clock values set by the clock
// synchronization algorithm
void CClockInterface::RemoveGps (EGpsUnit gpu)
{
    if (gps_ == NULL) return;  // nothing to remove

    gps_->DisableGps (gpu);  // disable the receiver

    INT32 old = DisableInterrupts();
    RemoveISR (GetGpsIntvec (GetNodeID(), area_, gpu), gps_);
    // the GPS interrupt calls the default ISR of the GPS class (which
    // reads the GPS time)
    EnableInterrupts (old);

    if (!gps_->GpsEnabled())  // clock does not have GPS
    {
        delete gps_;
        gps_ = NULL;  // go back to S-Node mode
    }
}


//--------------------------------------------------------------------------
//--- clock amortization
//--------------------------------------------------------------------------

// computes the accuracy after 'afterTime' clock seconds if the current
// accuracy is 'currAlpha' (in S:[-8,-38] format) and the deterioration
// value is 'lambda' (in S:[-37,-51] format); the value returned is in
// S:[-8,-38] format
// assert: afterTime >= 0
SINT32 CClockInterface::DragAccuracy (SINT32 currAlpha, SINT16 lambda,
                                      double afterTime)
{
    Assert (afterTime >= 0);
    // we cannot drag accuracies into the past

    if (lambda == 0) return currAlpha;  // no deterioration

    if ((steppure_ == 0) && ((steplow_ & 0xFF000000) == 0))
    // 'afterTime' is only useful if the clock advances, and only
    // in this case can we divide by the step value; if the clock does
    // not advance, then the new accuracy would be "infinite" (that is, it
    // would be at a bound for lambda != 0 and stay the same for
    // lambda = 0); so that is what we do: we return either the bound
    // or the current accuracy; but since this situation is quite likely
    // not intended, we inform the user through a report
    {
        Report (dragging accuracy of non-advancing clock);

        if (lambda == 0) return currAlpha;
        // lambda is zero -> return the current accuracy (this case is
        // already handled above, but it does not hurt to check it again)
        else if (lambda < 0) return LowerAccuracyBound;
        // lambda is negative -> return the lower bound
        else return UpperAccuracyBound;
        // lambda is positive -> return the upper bound
    }

    // if we come here, then either steplow or step are not zero, so we
    // can divide by the step value
    // the formula for computing the accuracy difference that accumulates
    // in 'afterTime' clock seconds is:
    //     (afterTime*lambda*2^-51)/(steppure*2^-51 + steplow*2^-59) =
    //     (afterTime*lambda)/(steppure + steplow/256)
    // (the unit of the expression is seconds)

    double stepvalue;  // for easier computation below
    stepvalue = (double)steppure_ + (double)(steplow_>>24 & 0x0FF)/256.;
    // this is the value in the denominator; it is already adjusted
    // to the format of LambdaP(N) (that is, the 2^-51 is removed)

    double dAlphaDiff;  // the deterioration value for alpha
    // note: to avoid problems with overflows, we use a double value
    // for all computations and only convert into the SINT32 at the end
    // of the computation

    dAlphaDiff = afterTime*(double)lambda/stepvalue;
    // compute the value that is added; this value is given in seconds

    dAlphaDiff = dAlphaDiff*TwoPow38;
    // now dAlphaDiff==1 means that bit -38 is set

    double dAlphaDragged;
    // double representation of the accuracy after 'afterTime' seconds

    dAlphaDragged = (double)(currAlpha) + dAlphaDiff;
    // the dragged accuracy is the current accuracy plus the deterioration

    // now convert the double value back into a SINT32 accuracy
    // check if the accuracy steps over its bounds (remember: if it goes over
    // 0x7FFFFFFF or under 0x80000000, then it should not wrap but stay at
    // these bounds)
    if (dAlphaDragged >= 0x7FFFFFFF) return UpperAccuracyBound;
    // stay on upper bound
    else if (dAlphaDragged < -(double)0x7FFFFFFF) return LowerAccuracyBound;
    // stay on lower bound
    else return (SINT32)dAlphaDragged;
    // the value is already okay -> just cast it

    Assert (False);  // we should never come here
    return currAlpha;  // to satisfy stupid compilers
}


// computes the accuracy difference that one should amortize in 'afterTime'
// (clock) seconds (through a call to Amortize() or PreloadAmort())
void CClockInterface::ComputeAccuracyCorrection (INT32 acc,
    double afterTime, INT32& diffPlus, INT32& diffMinus)
{
    SINT32 alphaP = (SINT32)clock_->GetALPHAPGETL();
    SINT32 alphaN = (SINT32)clock_->GetALPHANGETL();
    // latch the current accuracies (bits S:[-8,-38])

    if (afterTime > 0)  // must compensate for accuracy deterioration
    {
        // alphaP/N contain the current accuracies; after the time 'afterTime'
        // they will have changed by lambdaP(N)
        // note: this assumes that members lambdaP(N) contain the current
        // values of the LAMBDAP(N)PURE registers; if the user has set these
        // values directly through the register clock obtained by GetClock(),
        // then the computation will be wrong!
        // --> it would be nice if the current values of LAMBDA could be
        // read

        alphaP = DragAccuracy (alphaP, lambdaP_, afterTime);
        alphaN = DragAccuracy (alphaN, lambdaN_, afterTime);
        // compute the accuracies after 'afterTime' clock seconds have passed
    } // if

    // now, alphaP and alphaN contain the accuracies of the clock in
    // 'afterTime' seconds

    // compute the accuracy differences (acc contains 2x bits [-8,-23]
    // (unsigned!), and the positive alpha is in the lower two bytes)
    diffPlus = (SINT32)(acc<<15 & 0x7FFF8000) - alphaP;
    diffMinus = (SINT32)(acc>>1 & 0x7FFF8000) - alphaN;
}


// programs the clock to amortize the given time difference (in seconds);
// if the time cannot be fully amortized, as much as possible is amortized;
// 'diffPlus', 'diffMinus' are the differences of the accuracies,
// given in format S:[-8,-38]
// assert: amorttime_ > 0
void CClockInterface::PreloadAmort (double diff, INT32 diffPlus,
                                    INT32 diffMinus)
{
    static const double Ceiling = 0xffffffff;  // highest step value

    Assert (amorttime_ > 0);  // should be checked by the caller

    double CyclesPerAmortPhase = clock_->GetFrequency()*amorttime_;
    if (CyclesPerAmortPhase > Ceiling) CyclesPerAmortPhase = Ceiling;
    // the amortization duration cannot be longer than Ceiling
    else CyclesPerAmortPhase =(INT32)CyclesPerAmortPhase; 

    INT32 stepamort = 0;
    if ((diff*TwoPow51)/CyclesPerAmortPhase + 0.5 + steppure_ > Ceiling)
        stepamort = (INT32)Ceiling;  // amortize as much as possible
    else if ((diff*TwoPow51)/CyclesPerAmortPhase + 0.5 + steppure_ > 0)
        stepamort = (INT32)((diff*TwoPow51)/CyclesPerAmortPhase+0.5+steppure_);
    // note: STEPLOW does not have to be considered, because STEP and
    // STEPLOW are what is needed to pass one clock cycle in pure phase;
    // STEPLOW cannot be changed, so the question is, what do we have to
    // add to STEP to amortize the difference; and that value should be
    // as close as possible

    clock_->SetSTATEPSET (diffPlus);
    clock_->SetSTATENSET (diffMinus);  // amortize the accuracies at once

/**/// note: should I check if the amortization phase is active?
    clock_->SetSTEPAMORT (stepamort);
    clock_->SetAMORTTIMESET ((INT32)CyclesPerAmortPhase);
}

// calls PreloadAmort() to load everything for the amortization phase,
// then writes to STARTAMORT to start the amortization phase
// the function is also called by the associated GPS class (if the 
// clock has GPS functionality)
void CClockInterface::Amortize (double diff, INT32 diffPlus, INT32 diffMinus)
{
    PreloadAmort (diff, diffPlus, diffMinus);
    clock_->SetSTARTAMORT (0);  // start the amortization
}


//--------------------------------------------------------------------------
//--- odds and ends
//--------------------------------------------------------------------------

// return the current clock time (macrostamp and timestamp)
CNtpTime CClockInterface::GetCurrentClockTime ()
{
    CNtpTime time;

    // get current clock value
    time.ts = clock_->GetTSGETL();
    time.ms = clock_->GetMSGET() & 0x0FFFFFF;

    return time;
}


// program the duty timer to raise an interrupt at NTPTIME time
void CClockInterface::SetDutyTimer (EInterruptSource source,
    const CNtpTime& time)
{
    ESSUnit ssu;

    switch (source)
    {
        case IntDutyA1: ssu = SSU1; break;
        case IntDutyA2: ssu = SSU2; break;
        case IntDutyA3: ssu = SSU3; break;
        case IntDutyA4: ssu = SSU4; break;
        case IntDutyA5: ssu = SSU5; break;
        case IntDutyA6: ssu = SSU6; break;

        case IntDutyB1: ssu = SSU1; break;
        case IntDutyB2: ssu = SSU2; break;
        case IntDutyB3: ssu = SSU3; break;
        case IntDutyB4: ssu = SSU4; break;
        case IntDutyB5: ssu = SSU5; break;
        case IntDutyB6: ssu = SSU6; break;

        default: Assert (False);  // currently not supported
    } // switch

    // disable the timer
    switch (source)  // determine if A or B is used
    {
        case IntDutyA1: case IntDutyA2: case IntDutyA3:
        case IntDutyA4: case IntDutyA5: case IntDutyA6:
            clock_->SetDUTYAL (ssu, 0);  // disable the timer
            break;

        case IntDutyB1: case IntDutyB2: case IntDutyB3:
        case IntDutyB4: case IntDutyB5: case IntDutyB6:
            clock_->SetDUTYBL (ssu, 0);  // disable the timer
            break;

        default: Assert (False);  // currently not supported
    } // switch


    CNtpTime duty (0, 1<<8);  // for round-to-nearest

    if (time.ts & 1<<7) duty = duty + time;  // rest>=1/2 lsb --> round up
    else duty = time;  // rest < 1/2 lsb --> round down


    // set the timer
    switch (source)  // again determine if A or B is used
    {
        case IntDutyA1: case IntDutyA2: case IntDutyA3:
        case IntDutyA4: case IntDutyA5: case IntDutyA6:
            clock_->SetDUTYAH (ssu, (duty.ms<<8 & 0xFFFFFF00) +
                                    (duty.ts>>24 & 0x0FF));
            clock_->SetDUTYAL (ssu, (duty.ts>>8 & 0x0FFFF) + 0x10000);
            break;

        case IntDutyB1: case IntDutyB2: case IntDutyB3:
        case IntDutyB4: case IntDutyB5: case IntDutyB6:
            clock_->SetDUTYBH (ssu, (duty.ms<<8 & 0xFFFFFF00) +
                                    (duty.ts>>24 & 0x0FF));
            clock_->SetDUTYBL (ssu, (duty.ts>>8 & 0x0FFFF) + 0x10000);
            break;

        default: Assert (False);  // currently not supported
    } // switch
}



//**************************************************************************
// module global functions
//**************************************************************************

// get interrupt bit for an interrupt source
// if source<0 or source>63, an assert fails
static INT32 SourceToMask (EInterruptSource source)
{
    Assert ((source >= 0) && (source < 64));

    if (source < 32) return 1<<source;
    else return 1<<(source-32);
}


//**************************************************************************
// friend of CClockInterface
//**************************************************************************

// assumes that 'obj' is of type CClockInterface and invokes the wobble ISR
void CallWobbleISR (CNeedInterrupt *obj)
{
    ((CClockInterface *)obj)->wobble_isr();  // call the wobble ISR
}

