//
// clockirq.cpp
//
// GPS interrupt process
//
// created: 29.03.97
//
// last change: 25.08.97
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/07/29 16:19:34 $
//


#include <stdlib.h>  // for NULL

#include "clock_sw.hpp"
#include "clockirq.hpp"
#include "constant.hpp"
#include "error.hpp"
#include "general.hpp"
#include "interupt.hpp"


//**************************************************************************
// class CDutyB1Interrupt: an interrupt of duty timer B on SSU1; this duty
//    timer is hardwired to AMORTDUTY and starts an amortization phase
//    (i.e., writes to register STARTAMORT)
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

// store a pointer to the clock; needed for setting STARTAMORT
CDutyB1Interrupt::CDutyB1Interrupt (unsigned long nodeID, CSoftwareClock *clock)
    : CClockInterrupt (nodeID)
{
    Assert (clock != NULL);
    clock_ = clock;
}


//--------------------------------------------------------------------------
//--- interrupt specific work
//--------------------------------------------------------------------------

// is executed prior to calling the ISR; sets the IP bit, writes to
// register STARTAMORT
void CDutyB1Interrupt::PreExecution ()
{
//cout << "preexecution" << endl;
    CClockInterrupt::PreExecution();  // set the IP bit
    clock_->SetSTARTAMORT(1);  // start the amortization phase
//cout << "amortization is started" << endl;
//clock_->DumpRegisters();
// DEBUG
}



//**************************************************************************
// class CGpsInterrupt: an interrupt of the GPS receiver (every full
//      second); captures the clock value, calls the 1PPS ISR (if enabled),
//      then waits for 'GpsDelay' ms and calls the ISR of the GPS-receiver
//      (time information ready)
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

// store a pointer to the clock that is "connected" to the GPS receiver
// assert: gpu  < EGpsUnitCount
CGpsInterrupt::CGpsInterrupt (unsigned long nodeID, CSoftwareClock *clock,
    EGpsUnit gpu) : CClockInterrupt (nodeID)
{
    Assert (clock != NULL);
    clock_ = clock;
    
    Assert (gpu < EGpsUnitCount);
    gpu_ = gpu;

    gpstime_ = new CInterruptProcess;
    Assert (gpstime_ != NULL);
    gpstime_->SetNodeID (nodeID);

#ifndef NO_RESOURCE
    Resource::ref(gpstime_);  // register the process in C++SIM
#endif

    RaiseAfter (1 - GetGlobalTime() + (INT32)GetGlobalTime());
    // schedule the interrupt for the next full second
    // note: the interrupt is never raised at once
}

CGpsInterrupt::~CGpsInterrupt ()
{
#ifndef NO_RESOURCE
    Resource::unref(gpstime_);  // either unref the process in C++SIM
#else
    delete gpstime_;  // or delete it
#endif
}


//--------------------------------------------------------------------------
//--- handling of the GPS time interrupt
//--------------------------------------------------------------------------

// sets the interrupt vector of the GPS time ISR
void CGpsInterrupt::SetTimeVector (int area)
{
    gpstime_->SetVector (GetGpsIntvec (GetNodeID(), area, gpu_));
}

// the GPS time ISR is called in PostExecution()
void CGpsInterrupt::EnableTime ()
{
    gpstime_->Enable();
}

// the GPS time ISR is not called in PostExecution()
void CGpsInterrupt::DisableTime()
{
    gpstime_->Disable();
}


//--------------------------------------------------------------------------
//--- interrupt specific work
//--------------------------------------------------------------------------

// is executed prior to calling the ISR; prompts the
// clock to capture the 1PPS pulse
void CGpsInterrupt::PreExecution ()
{
//    CClockInterrupt::PreExecution();  // set the IP bit

    clock_->CaptureTime (gpu_);
}

// is executed after calling the ISR; it schedules the GPS time interrupt
// for activation in 'GpsDelay' seconds from now, then schedules itself
// for the next full seconds
void CGpsInterrupt::PostExecution ()
{
    CClockInterrupt::PostExecution();  // empty right now

    gpstime_->RaiseAfter (GpsDelay);  // schedule the GPS time ISR

    RaiseAfter (1 - GetGlobalTime() + (INT32)GetGlobalTime());
    // schedule the interrupt for the next full second
    // note: the interrupt is never raised at once, which is absolutely
    // necessary in this case
}

