//
// clocksyn.cpp
//
// implementation of the clock synchronization class
//
// note: if you want ClockSync to send its clock data more than once per
//       round (only in case of a P-Node), set macro ADDITIONAL_PNODE_DATA=nn
//       and compile again; P-Node data is then sent nn+1 times
//
// note: since it's always hard to remember the message formats of all those
//       synchronization messages going around, here's a complete list:
//
// ************ msg format ***********
// Timestamp, DelayMeasurement (Node P [FME] -> Node Q [Announce]):
//    Node P: INT32 mode, SINT32 nodeID (P)
//
// RateSynchronization (Node P [FME] -> Node Q [Announce]):
//    Node P: INT32 mode, SINT32 nodeID (P), double thetaPlus (P),
//            double thetaMinus (P), double u (P)
//
// DelayMeasurementReply (Node Q [ReplyToMessage] -> Node P [Announce]):
//    Node Q: INT32 mode, SINT32 nodeID (Q), INT32 msReceive (Q),
//            INT32 tsReceive (Q), double thetaMinus (Q),
//            double thetaPlus (Q), INT32 area (P), ms/ts/acc (P)
// ***********************************
//
// *** msg format using structures ***
// Timestamp, DelayMeasurement (Node P [FME] -> Node Q [Announce]):
//    Node P: SSendData (P)
// RateSynchronization (Node P [FME] -> Node Q [Announce]):
//    Node P: SSendData (P), SRate (P)
// DelayMeasurementReply (Node Q [Announce] -> Node P [Announce]):
//    Node Q: SReplyData (Q), STimeData (P)
// ***********************************
//
// created: 09.11.96
//
// last change: 22.01.99
//
// RCS version control information:
//     $Revision: 1.12 $$State: Exp $$Date: 1999/01/22 09:54:39 $
//


#include <string.h>  // for memcpy etc.

#include "address.hpp"
#include "algbase.hpp"
#include "clockifc.hpp"
#include "clocksyn.hpp"
#include "database.hpp"
#include "error.hpp"
#include "globals.hpp"
#include "idlist.hpp"
#include "interupt.hpp"
#include "lance.hpp"
#include "msgform.hpp"
#include "template.hpp"
#include "time.hpp"


const int CI_Mode = (TS_PACKET | RET_PKB_IMM);
// return the packet buffer at once (no call to ci_ret_tbuf() necessary)

// the message modes; the value denotes the number of the bit that has to
// be set
enum EMessageMode {
    Timestamp=0,  // the send timestamp of the msg is valid
    DelayMeasurement,  // the msg requests a reply for delay meas. purposes
    DelayMeasurementReply,  // the msg is a reply to the above request
    RateSynchronization,  // the msg contains rate sync. information
    GpsTime  // in conjunction with 'Timestamp'; the node has GPS
};

enum ERoundStage {BeforeSend=0, BeforeCV, BeforeResync};
// the stages of one round; needed for the ISR to determine which function
// to call; the values must be continuous


//**************************************************************************
// class CClockSync: this class is the main clock synchronization
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CClockSync::CClockSync (CControllingObject *ctrl, int area)
    : CApplControlledObject (ctrl)
{
    area_ = area;

    lance_ = NULL;
    clock_ = NULL;

    resynchronizing_ = False;  // actively take part in the synchronization

    csa_ = NULL;  // the CSA will be set by the Init() function

    INT32 old = DisableInterrupts();
    SetISR (GetClockIntvec (GetNodeID(), area_) | INTT, this); 
    // clock interrupt INTT (duty timer B on SSU4) calls the default ISR
    EnableInterrupts (old);

    initCalled_ = False;
    CClockSync::Init();
    // initialize the members that are modified during reset
}

// the clock must not have been deleted when the dtor is called
CClockSync::~CClockSync ()
{
    clock_->DisableInterrupt (IntDutyB4);  // disable the interrupt
    clock_->DisableDutyTimer (IntDutyB4);  // disable the timer itself

    INT32 old = DisableInterrupts();
    RemoveISR (GetClockIntvec (GetNodeID(), area_) | INTT, this);
    EnableInterrupts (old);

    CClockSync::Dispose();
}


//--------------------------------------------------------------------------
//--- suspend/resume the synchronization
//--------------------------------------------------------------------------

// discard the collected interval data, don't accept messages
// assert: clock != NULL
void CClockSync::Suspend ()
{
    Assert (clock_ != NULL);

    clock_->DisableInterrupt (IntDutyB4);  // disable the interrupt
    suspended_ = True;  // no new data is accepted
    csa_->Suspend ();  // let CSA do its own suspend actions
    roundStage_ = BeforeSend;  // operation is resumed with a new FME
    timeOfResync_ = NoAccount;
}

// trigger an interrupt at the next multiple of the round period
// assert: clock != NULL
void CClockSync::Resume()
{
    Assert (clock_ != NULL);

    clock_->EnableInterrupt (IntDutyB4);  // enable the interrupt
    InterruptAtMult ();  // program the interrupt
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the requested parameter value; returns True if successful,
// otherwise False
// existing parameters: DelayMeasurementCycles, RateSynchronizationCycles,
// ComputationDelayCompensation, DelayCompensation, SynchronizationAlgorithm,
// RoundPeriod, and all params of CApplControlledObject
// parameters to CSA: NumberOfNodes, NumberOfFaultyNodes, AssociatedNetwork,
// Delta, EpsilonPlus, EpsilonMinus, ThetaPlus, ThetaMinus, U, UserParameter
EBool CClockSync::SetParameter (const SParameter& param)
{
    SSingleParameter *sparam = (SSingleParameter*)&param;  // easy access

    switch (param.type)
    {
      case DelayMeasurementCycles:
          if (sparam->value < 0) return False;
          delayCycles_ = (INT32)sparam->value;
          delayCounter_ = delayCycles_-1;  // cycleCounter is reset as well
          break;
      case RateSynchronizationCycles:
          if (sparam->value < 0) return False;
          rateCycles_ = (INT32)sparam->value;
          rateCounter_ = rateCycles_-1;  // cycleCounter is reset as well
          break;
      case RoundPeriod:
          if (sparam->value <= 0) return False;
          if (sparam->value < delayCompensation_+computationDelayCompensation_)
              return False;
          roundPeriod_ = sparam->value;
          if ((timeOfResync_ != NoAccount) && (roundStage_ == BeforeSend))
          // we are before the FME(), and an interrupt is scheduled
          {
              InterruptAtMult();
              // reschedule the interrupt according to the new round period
          }
          clock_->SetAmortizationDuration (roundPeriod_*0.02);
          // amortization duration must be less or equal to the round period
          // note: 2% is just a rule of thumb; Clock also has a parameter for
          // the amortization duration
          break;
      case ComputationDelayCompensation:
          if (sparam->value < 0) return False;
          if (sparam->value + delayCompensation_ > roundPeriod_)
              return False;
          if ((timeOfResync_ != NoAccount) && (roundStage_ < BeforeResync))
          // we are before the resynchronization, an interrupt is scheduled
          {
              timeOfResync_ += sparam->value-computationDelayCompensation_;
              // adapt timeOfResync_ to the difference
          }
          computationDelayCompensation_ = sparam->value;
          break;
      case DelayCompensation:
          if (sparam->value < 0) return False;
          if (sparam->value + computationDelayCompensation_ > roundPeriod_)
              return False;
          if ((timeOfResync_ != NoAccount) && (roundStage_ < BeforeCV))
          // we are before the CV, an interrupt is scheduled
          {
              timeOfResync_ += sparam->value - delayCompensation_;
              // adapt timeOfResync_ to the difference
          }
          delayCompensation_ = sparam->value;
          break;
      case SynchronizationAlgorithm:
          if ((sparam->value < 0) || (sparam->value >= EAlgorithmTypeCount))
              return False;
          return ChangeAlgorithm ((EAlgorithmType)sparam->value);
          break;

      case AssociatedNetwork:
      case NumberOfNodes:
      case NumberOfFaultyNodes:
      case Delta:
      case EpsilonPlus:
      case EpsilonMinus:
      case ThetaPlus:
      case ThetaMinus:
      case U:
      case UserParameter:  // for the clock sync. algorithm
          return csa_->SetParameter (param);
          break;

      default:  // unknown parameter type, pass on to the base class
          return CApplControlledObject::SetParameter (param);
    } // switch

    return True;
}

// returns the requested parameter value; returns True if successful,
// otherwise False
EBool CClockSync::GetParameter (SParameter& param) const
{
    SSingleParameter *sparam = (SSingleParameter*)&param;  // easy access

    switch (param.type)
    {
      case DelayMeasurementCycles:
          sparam->value = delayCycles_;
          break;
      case RateSynchronizationCycles:
          sparam->value = rateCycles_;
          break;
      case RoundPeriod:
          sparam->value = roundPeriod_;
          break;
      case ComputationDelayCompensation:
          sparam->value = computationDelayCompensation_;
          break;
      case DelayCompensation:
          sparam->value = delayCompensation_;
          break;
      case TimerGranularity:
          if (clock_ == NULL) return False;
          sparam->value = clock_->GetGranularityDT();
          break;
      case SynchronizationAlgorithm:
          sparam->value = csaType_;
          break;

      case AssociatedNetwork:
      case NumberOfNodes:
      case NumberOfFaultyNodes:
      case Delta:
      case EpsilonPlus:
      case EpsilonMinus:
      case ThetaPlus:
      case ThetaMinus:
      case U:
      case UserParameter:  // for the clock sync. algorithm
          return csa_->GetParameter (param);
          break;

      default:  // unknown parameter type, pass on to the base class
          return CApplControlledObject::GetParameter (param);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False
// existing commands: ConnectNetwork, DisconnectNetwork, and all commands
// of CApplControlledObject
// commands to CSA: ConnectNetwork (also processed in this class),
// DisconnectNetwork (also processed in this class), UserCommand
EBool CClockSync::ExecuteCommand (const SCommand& command)
{
    SNetworkCommand *netc = (SNetworkCommand*)&command;  // easier access

    switch (command.type)
    {
        case ConnectNetwork:  // connect a network
          if (netc->networkID < 0) return False;
          listOfConnectedNetworks_->Insert (netc->networkID);
          csa_->ExecuteCommand (command);  // pass on the information
          break;
        case DisconnectNetwork:  // remove a connected network
          listOfConnectedNetworks_->Remove (netc->networkID);
          csa_->ExecuteCommand (command);  // pass on the information
          break;
        case ReportTransmissions:  // report measured transmission times
#ifdef PSOS_SIM
          reportDelays_ = True;
#else
          return False;  // ignore the command in pure simulation
#endif
          break;
        case DontReportTransmissions:  // do not report times
          reportDelays_ = False;
          break;

        case UserCommand:  // pass it to the clock sync. algorithm
          return csa_->ExecuteCommand (command);
          break;

        default:  // unknown command - pass on to the base class
          return CApplControlledObject::ExecuteCommand (command);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- set associated objects
//--------------------------------------------------------------------------

// set the lance pointer to the lance
void CClockSync::SetLance (CLance *lance)
{
    lance_ = lance;
}

// set the clock pointer to the clock (also in CSA)
void CClockSync::SetClock (CClockInterface *clock)
{
    clock_ = clock;
    csa_->ClockChanged ();
}


//--------------------------------------------------------------------------
//--- clock synchronization functions
//--------------------------------------------------------------------------

// sends a broadcast message to the Lance
// assert: lance != NULL
// *** msg format ***
// Timestamp, DelayMeasurement (Node P [FME]):
//    Node P: SSendData
//        ==> INT32 mode, SINT32 nodeID (P)
// RateSynchronization (Node P [FME]):
//    Node P: SSendData (P), SRate (P)
//        ==> INT32 mode, SINT32 nodeID (P), double thetaPlus (P),
//            double thetaMinus (P), double u (P)
// ******************
void CClockSync::BroadcastTime ()
{
    char *buf;  // the message buffer
    SSendData data;  // the data that is sent in the buffer
    INT32 bufsize;  // size of the buffer requested from Lance
    SRate rateData;  // the rate data that is to be sent
    INT32 msgID;  // ID of the message, returned by GetBuffer()

    suspended_ = False;  // suspension is canceled if FME is initiated

    if (listOfConnectedNetworks_->Empty()) return;
    // don't have a network to send to

    // in resynchronization mode, messages are sent/received normally, but
    // messages sent by FME don't have 'Timestamp' in the mode; that tells
    // the receiver that the timestamp sent in the message is invalid

    if (!resynchronizing_) data.mode = (1<<Timestamp);
    else data.mode = 0;

    data.nodeID = GetNodeID();
    bufsize = sizeof(SSendData);

    if (delayCycles_ > 0)  // delay measurement is turned on
    {
        delayCounter_++;  // count this call to FME

        if (delayCounter_ >= delayCycles_)  // send a delay measurement msg
        {
            delayCounter_ = 0;
            data.mode |= (1<<DelayMeasurement);
        }
    }

    if (rateCycles_ > 0)  // rate synchronization is turned on
    {
        rateCounter_++;  // again count this call to FME

        if (rateCounter_ >= rateCycles_)  // send the rate information
        {
            rateCounter_ = 0;
            if (csa_->GetLocalRate (rateData))  // local rate is stored
            {
                data.mode |= (1<<RateSynchronization);
                bufsize += sizeof (SRate);
            }
            else  // the message does not contain rate information
            {
                Report (local rate not available);
            }
        } // if
    } // if

    if (data.mode == 0) return;
    // the message would only contain an invalid timestamp, so don't send
    // it at all

    if ((data.mode & (1<<Timestamp)) && clock_->GpsConnected())
        data.mode |= (1<<GpsTime);  // the clock is influenced by GPS

    Assert (lance_ != NULL);

    for (SINT32 networkID_ = listOfConnectedNetworks_->GetFirst();
         networkID_ != NoAccount;
         networkID_ = listOfConnectedNetworks_->GetNext())
    {
        lance_->GetBuffer (bufsize, &buf, msgID);  // request a buffer
        memcpy (buf, &data, sizeof(SSendData));   // copy the data

        // copy the rate data as last entry
        if (data.mode & (1<<RateSynchronization))
        {
            memcpy (buf+sizeof(SSendData), &rateData, sizeof(SRate));
        }

        // report to EvalSys that a message is sent
        SApplSendMsgEvent *event = new SApplSendMsgEvent;
        event->module = ClockSync;
        event->type = SendMessage;
        event->msgID = msgID;
        event->receiver = -1;  // broadcast message
        ReportEvent (event);

        lance_->Send (CI_Mode, NULL, buf, bufsize, networkID_, area_);
        // send the message

#ifdef ADDITIONAL_PNODE_DATA
// check result on Fetzer if P-Node data is sent more than once -> Fetzer (and
// also original FTA) gets better (in original FTA, w.c. precision is halved)
        if (data.mode & (1<<GpsTime))  // send P-Node data more than once
        {
            if (data.mode & (1<<RateSynchronization))
                data.mode &= ~(1<<RateSynchronization);  // no rate sync.

            for (int i=0; i<ADDITIONAL_PNODE_DATA; i++)
            {
                lance_->GetBuffer (bufsize, &buf, msgID);  // request a buffer
                memcpy (buf, &data, sizeof(SSendData));   // copy the data

                lance_->Send (CI_Mode, NULL, buf, bufsize, networkID_, area_);
            } // for
        } // if
#endif
    } // for
}

// take over the data, react to it and/or store it
// assert: renv != NULL
//
// *** msg format ***
// Timestamp, DelayMeasurement (Node P [FME] -> Node Q [Announce]):
//    Node Q: SSendData (P)
//        ==> INT32 mode, SINT32 nodeID (P)
// RateSynchronization (Node P [FME] -> Node Q [Announce]):
//    Node Q: SSendData (P), SRate (P)
//        ==> INT32 mode, SINT32 nodeID (P), double thetaPlus (P),
//            double thetaMinus (P), double u (P)
// DelayMeasurementReply (Node Q [ReplyToMessage] -> Node P [Announce]):
//    Node P: SReplyData (Q), STimeData (P)
//        ==> INT32 mode, SINT32 nodeID(Q), INT32 msReceive (Q),
//            INT32 tsReceive (Q), double thetaPlus (Q),
//            double rhoMinus (Q), INT32 area (P), ms/ts/acc time (P)
// ******************
void CClockSync::Announce (renvx *renv, STimeData sendtime, char *message,
    INT16 msgsize, int if_num, INT32 msgID)
{
    if (suspended_)  // ignore messages while suspended
    {
        Report (received message in suspension state);
        Report (--> CClockSync::suspended is *not* obsolete);
        // should never happen if "suspended" is really obsolete
        return;
    }

    SSendData data;  // includes the mode of the message

    Assert (renv != NULL);

    if (message == NULL)
    {
        Report (original message was empty);
        return;
    }

    if (msgsize < sizeof(SSendData))
    // mode and nodeID have to be the first items in the message
    {
        Report (message size too small);
        return;
    }

    memcpy (&data, message, sizeof(SSendData));
    // extract the data

    // report to EvalSys that a message was received
    SApplRecvMsgEvent *event = new SApplRecvMsgEvent;
    event->module = ClockSync;
    event->type = ReceivedMessage;
    event->sender = data.nodeID;
    event->senderArea = sendtime.area;
    event->msgID = msgID;
    ReportEvent (event);

    if (reportDelays_)  // have to report the transmission delay
    {
        STimestamp starttime;
        Assert (clock_ != NULL);
        SApplDelayEvent *ev = new SApplDelayEvent;
        ev->module = ClockSync;
        ev->type = TransmissionDelay;
        ev->sender = data.nodeID;
        ev->senderArea = sendtime.area;
        ev->msgID = msgID;
        GetTransmitTimestamp (&renv->sa, sendtime.area, clock_->GetClock(),
                              starttime);
        ev->delaytime.ms = renv->utcsu[sendtime.area].rms - starttime.ms;
        ev->delaytime.ts = renv->utcsu[sendtime.area].rts - starttime.ts;
        ev->delaytime.acc = starttime.acc;
        // note: currently, the clock drift is ignored
        ReportEvent (ev);
    } // if

    sendtime.ms = sendtime.ms & 0x00FFFFFF;
    // mask out the checksum bits
/**/// note: should I do something with the checksum?

    // store the association between the network address and the nodeID
    // of the remote node
    if (!csa_->SetAssociation (if_num, &(renv->sa), data.nodeID))
        Report (wrong mapping nodeID <-> network address);


    // now check the modes
    if ((data.mode & (1<<RateSynchronization)) == (1<<RateSynchronization))
    // extract the rate sync. data
    {
        SRate remoteRate;

        if (msgsize < sizeof(SSendData)+sizeof(SRate))
        // the mode and nodeID have to be the first items in the message,
        // followed by the rate information of the remote clock
        {
            Report (message size too small);
            return;
        }

        // extract the rate
        memcpy (&remoteRate, message + sizeof(SSendData), sizeof(SRate));

        CNtpTime localTime (renv->utcsu[area_].rms & 0x00FFFFFF,
                            renv->utcsu[area_].rts);  // time of message reception
        CNtpTime remoteTime (sendtime.ms, sendtime.ts); // message transmission

        csa_->StoreRate (remoteRate, localTime, remoteTime, if_num,
                         &(renv->sa), sendtime.area);
        // store the rate information
    }

    if ((data.mode & (1<<DelayMeasurementReply)) == (1<<DelayMeasurementReply))
    // evaluate the transmission delay data
    {
        Assert (data.mode == (1<<DelayMeasurementReply));
        // note: it is assumed that DMRs are *additional* messages and
        // therefore dedicated (important for the interpretation of the
        // message)

        STimeData myTime;  // the original send time of this node
        SReplyData replyData; // there is more information than just "data"

        if (msgsize < sizeof(STimeData)+sizeof(SReplyData))
        // the remote data is the first entry in the message, followed by
        // the original local data
        {
            Report (message size too small);
            return;
        }

        memcpy (&replyData, message, sizeof(SReplyData));
        memcpy (&myTime, message+sizeof(SReplyData), sizeof(STimeData));
        // note: not necessary to mask out CS in myTime, since that was
        // done in the sender

        ComputeDelay (renv, sendtime, replyData, myTime, if_num);
        // compute the delay and store it
    } // if


    if (((data.mode & (1<<DelayMeasurement)) == (1<<DelayMeasurement)) &&
        listOfConnectedNetworks_->Find (if_num))
    // came from a connected network
    // this is a request for delay measurement --> have to reply to it
    {
        ReplyToMessage (renv, sendtime, if_num);
    } // if

    if ((data.mode & (1<<Timestamp)) == (1<<Timestamp))
    // this is a timestamped message - store the time information
    {
        if ((data.mode & (1<<GpsTime)) == 0) StoreTime (renv, sendtime, if_num);
        else StoreTime (renv, sendtime, if_num, True);
        // store the time (add information about node type of the sender)
    } // if
}

// prompt CSA to use the collected data to compute a new clock value
void CClockSync::ApplyConvergenceFunction ()
{
    // invoke the clock synchronization algorithm
    if (!csa_->ConvergenceFunction ())
    {
        ReportEvent (ClockSync, ConvergenceFailed);
        // report the failure
    }
}

// prompt CSA to set the clock to the new value
void CClockSync::Resynchronize ()
{
    csa_->SetNewClockValue ();  // set the clock to the new value
}


//--------------------------------------------------------------------------
//--- interrupt service routine
//--------------------------------------------------------------------------

// called if a duty timer interrupt from DTB4 (INTT) occurs; calls one of
// BroadcastTime(), ApplyConvergenceFunction() or Resynchronize(),
// and sets the duty timer to a new value (depending on the round period)
// assert: timeOfResync != NoAccount
void CClockSync::isr ()
{
    if (!clock_->InterruptPending (IntDutyB4)) return;  // not supported
    clock_->ClearInterrupt (IntDutyB4);  // clear the DT bit

    ReportEvent (ClockSync, TimerEvent);

    Assert (timeOfResync_ != NoAccount);  // that should not happen anymore

    switch (roundStage_)
    {
        case BeforeSend:
            if (delayCompensation_ > 0)  // schedule for next stage
                clock_->InterruptAfter (IntDutyB4, delayCompensation_);

            BroadcastTime ();  // call FME
            roundStage_++;  // set the next stage

            if (delayCompensation_ > 0) break;  // leave the ISR

        case BeforeCV:
            if (computationDelayCompensation_ > 0)// schedule for next stage
                clock_->InterruptAfter (IntDutyB4, computationDelayCompensation_);

            ApplyConvergenceFunction ();  // compute the new clock value
            roundStage_++;  // set the next stage

// DEBUG
#ifndef NO_PRELOAD
            csa_->PreloadNewClockValue (timeOfResync_);
#endif

            if (computationDelayCompensation_ > 0) break;  // leave the ISR

        case BeforeResync:
            InterruptAtMult ();  // schedule for next FME
// DEBUG
#ifdef NO_PRELOAD
            Resynchronize ();  // resynchronize
#endif
            roundStage_ = BeforeSend;  // go back to the first stage
            break;

        default:
            Assert (False);  // should never happen
    } // switch
}


//--------------------------------------------------------------------------
//--- create/destroy the members that are modified during reset
//--------------------------------------------------------------------------

// initialize the members that are modified during reset
void CClockSync::Init ()
{
    if (initCalled_) return;  // Init() has already been called
    initCalled_ = True;

    computationDelayCompensation_ = DefaultComputationDelayCompensation;
    delayCompensation_ = DefaultDelayCompensation;

    listOfConnectedNetworks_ = new CNodeIdList;
    Assert (listOfConnectedNetworks_ != NULL);

    delayCycles_ = 0;  // no delay measurement
    delayCounter_ = 0;
    rateCycles_ = 0;  // no rate synchronization
    rateCounter_ = 0;
    roundPeriod_ = DefaultRoundPeriod;
    suspended_ = True;  // is the default after startup or reset
    roundStage_ = BeforeSend;  // start with FME

    timeOfResync_ = NoAccount;  // the next resync. time is not known

    reportDelays_ = False;  // do not report measured delays

    Assert(ChangeAlgorithm (FreeRun));  // FreeRun must exist
    // use FreeRun as default algorithm
}

// delete the dynamic members
void CClockSync::Dispose ()
{
    if (!initCalled_) return;  // Dispose() has already been called
    delete csa_;
    csa_ = NULL;
    delete listOfConnectedNetworks_;
    initCalled_ = False;
}


//--------------------------------------------------------------------------
//--- management of interrupts at multiples of the round period
//--------------------------------------------------------------------------

// prompts the duty timer of the clock to generate an interrupt at
// the next multiple of the round period, sets timeOfResync_
// accordingly
// assert: clock != NULL
void CClockSync::InterruptAtMult ()
{
    Assert (clock_ != NULL);

    clock_->DisableInterrupt (IntDutyB4);  // disable the interrupt
    // note: disabling the interrupt is important because if it occurs
    // before this function is finished, isr() is called with a wrong value
    // for timeOfResync_

    timeOfResync_ = clock_->InterruptAtMult (IntDutyB4, roundPeriod_);
    // wait for the next multiple of the round period

    timeOfResync_ += computationDelayCompensation_ + delayCompensation_;
    // add the lengths of stages 'BeforeCV' and 'BeforeResync' to get
    // the time of the next resynchronization

    clock_->EnableInterrupt (IntDutyB4);  // enable the interrupt
}


//--------------------------------------------------------------------------
//--- management of the synchronization algorithm
//--------------------------------------------------------------------------

// change the synchronization algorithm (change csa and csaType)
// the algorithm can be changed on the fly (no invalid pointers)
// assert: the algorithm has been installed (the create function exists)
EBool CClockSync::ChangeAlgorithm (EAlgorithmType newType)
{
    ACClockSyncAlgorithm *newCsa, *tmp;

    AlgFunction func = algorithmManager.GetFunction (newType);
    // get the create function

    if (func==NULL) return False;  // algorithm type is not installed

    newCsa = func (this);  // get an object of the new algorithm class
    Assert (newCsa != NULL);  // check if the creation was successful

    if (csa_ != NULL) *newCsa = *csa_;  // copy relevant configuration data

    csaType_ = newType;  // store the new type
    tmp = csa_;  // store a pointer to the old algorithm class
    csa_ = newCsa;  // store the new class (it is used from now on)
    if (tmp != NULL) delete tmp;  // delete the old class

    return True;
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int CClockSync::ReportError (EReportType type, char *file, int line,
                             char *expression)
{
    return ::ReportError (type, file, line, expression, ClockSync,
                          GetNodeID(), area_);
}


//--------------------------------------------------------------------------
//--- actions taken in Announce()
//--------------------------------------------------------------------------

// compute the transmission delay
// note: 'myTime' contains the original send time, 'renv' stores the receive
// timestamp of the reply; 'data' contains the remote receive time of the
// original message and the clock drift of the remote node, and 'time' the
// send timestamp of its reply
void CClockSync::ComputeDelay (renvx *renv, STimeData& time,
        SReplyData& data, STimeData& myTime, int if_num)
{
    double delayP;  // the measured round trip delay in seconds
    double delayQ;  // the measured delay of the remote node Q

    // check if the message is for this area
    if (myTime.area != (INT32)area_) return;  // ignore the message


    // now compute the round trip delay (in seconds)
    delayP = NtpToSeconds (renv->utcsu[area_].rms, renv->utcsu[area_].rts) -
             NtpToSeconds (myTime.ms, myTime.ts);
    // this is the complete round trip delay, including the delay at the
    // remote node


    // compute the delay added by the remote node Q
    delayQ = NtpToSeconds (time.ms, time.ts) -
             NtpToSeconds (data.msReceive, data.tsReceive);


    SRate remoteRate;  // extract the remote rate
    remoteRate.thetaMinus = data.thetaMinus;
    remoteRate.thetaPlus = data.thetaPlus;

    csa_->StoreDelay (delayP, delayQ, remoteRate, if_num, &(renv->sa));
}


// reply to a request for transmission delay measurement
// assert: lance != NULL, clock != NULL
// *** msg format ***
// DelayMeasurementReply (Node Q [ReplyToMessage] -> Node P [Announce]):
//    Node Q: SReplyData (Q), STimeData (P)
//        ==> INT32 mode, SINT32 nodeID (Q), INT32 msReceive (Q),
//            INT32 tsReceive (Q), INT32 area (P), ms/ts/acc (P)
// ******************
void CClockSync::ReplyToMessage (renvx *renv, STimeData& time, int if_num)
{
    SReplyData replyData;
    tenvx *tenv;
    char *replyBuf;
    SRate rate;
    INT32 msgID;

    Assert (clock_ != NULL);
    Assert (lance_ != NULL);

    replyData.mode = 1<<DelayMeasurementReply;
    replyData.nodeID = GetNodeID();
    replyData.msReceive = renv->utcsu[area_].rms;
    replyData.tsReceive = renv->utcsu[area_].rts;
    // the own receive timestamp is returned as well

    if (csa_->GetLocalRate (rate))  // local rate is stored
    {
        replyData.thetaMinus = rate.thetaMinus;
        replyData.thetaPlus = rate.thetaPlus;
    }
    else Report (no rate information stored);  // should never happen


    // request buffer and envelope
    lance_->GetEnvelope (&tenv);
    lance_->GetBuffer (sizeof(STimeData)+sizeof(SReplyData),
                       &replyBuf, msgID);

    memcpy (replyBuf, &replyData, sizeof(SReplyData));
    // put the mode, nodeID and own timestamp first
    memcpy (replyBuf+sizeof(SReplyData), &time, sizeof(STimeData));
    // then return the area and time of the sender

    memcpy (tenv->da, renv->sa, sizeof(addr));
    // this is no broadcast, but a dedicated reply to the sender!


    // report to EvalSys that a message is sent
    SApplSendMsgEvent *event = new SApplSendMsgEvent;
    event->module = ClockSync;
    event->type = SendMessage;
    event->msgID = msgID;
    event->receiver = -2;
/**/// just for now!
    ReportEvent (event);

    lance_->Send (CI_Mode, tenv, replyBuf, sizeof(STimeData)+
                  sizeof(SReplyData), if_num, area_);
    // send message
}


// store the time information
void CClockSync::StoreTime (renvx *renv, STimeData& time, int if_num,
                            EBool gps)
{
    if (timeOfResync_ < 0)  // cannot compute time until resync --> discard
    {
        Report (resynchronization time unknown);
        return;
    }

    Assert (clock_ != NULL);

    double timeUntilResync;  // time from receive time until resync.
    STimestamp remoteTime;  // for extracting the time from STimeData

    timeUntilResync = timeOfResync_ -
            NtpToSeconds (renv->utcsu[area_].rms, renv->utcsu[area_].rts);

    if (timeUntilResync < 0)  // should never happen?
    {
        Report (time until resynchronization is negative?);
        return;  // discard everything
    }

    remoteTime.ms = time.ms;  // copy the time and accuracy
    remoteTime.ts = time.ts;
    remoteTime.acc = time.acc;

    csa_->StoreTime (remoteTime, timeUntilResync, if_num, &(renv->sa),
                     time.area, gps);
    // deliver the time information to the CSA
}


