//
// constant.cpp
//
// definition of all constants in the system
// edit this file and link again to change the default values for
// system constants
// note: constants are devided into three groups (changeable by user,
// by evaluation, by simulation), ordering in the groups is alphabetically
//
// created: 28.09.96
//
// last change: 28.09.98
//
// RCS version control information:
//     $Revision: 1.6 $$State: Exp $$Date: 1998/09/28 10:58:37 $
//


#include "constant.hpp"


//**************************************************************************
// the following constants are system constants; the user may change them
// through the evaluation system
//**************************************************************************

//--------------------------------------------------------------------------
// synchronization round parameters
//--------------------------------------------------------------------------

const double DefaultRoundPeriod = 100;  // the constant should be > 0!
// the default round period of the clock synchronization application (in
// seconds)
// used by: CClockSync

const double DefaultDelayCompensation = 0.01*DefaultRoundPeriod;
// the constant should be >= 0!
// the default wait period from FME to application of the convergence
// function (in seconds)
// used by: CClockSync

const double DefaultComputationDelayCompensation = 0;  // should be >= 0!
// the default wait period from application of the convergence function to
// resynchronization (in seconds)
// used by: CClockSync


//--------------------------------------------------------------------------
// Supervisor/Controller parameters
//--------------------------------------------------------------------------

const double DefaultIfcPeriod = 1;  // the constant should be > 0!
// the default period (in seconds) of the interface modules responsible for
// communication with the evaluation; in the case of the Supervisor, this
// is also the period of the (periodic) system snapshot
// used by: CControllingObject


//--------------------------------------------------------------------------
// network parameters
//--------------------------------------------------------------------------

const double DefaultOwnDelay = 0;  // the constant should be >= 0!
// the default delay of a message that is transmitted to the sender node
// itself (in seconds)
// if you do not specify transmission delay data from a node to itself,
// then this default is used; in consequence, messages to oneself are
// always delivered if the node is connected, even if no delay data is
// stored
// used by: CNetworkAccountant


//--------------------------------------------------------------------------
// clock parameters
//--------------------------------------------------------------------------

const double WcRhoPlus = 1e-4;  // the constant should be >= 0!
// the upper bound on the worst case clock drift (global, for all clocks)
/**/// note: that's just a dummy and probably much too high
// used by: CClockInterface, ACClockSyncAlgorithm

const double WcRhoMinus = 1e-4;  // the constant should be >= 0 and < 1!
// the lower bound on the worst case clock drift (global, for all clocks)
/**/// note: that's just a dummy and probably much too high
// used by: CClockInterface, ACClockSyncAlgorithm

const double DefaultAmortizationDuration = 0.01*DefaultRoundPeriod;
// the constant should be >= 0! (=0 means instantaneous setting of the clock)
// the default duration of the amortization phase
// used by: CClockInterface

const double DefaultClockGranularity = 1./16777216.;  // should be > 0!
// the external clock granularity (1 lsb of the timestamp, = 2^-24)
// used by: CClock

const double DefaultAccGranularity = 1./8388608.;  // should be > 0!
// the accuracy transmission loss (1 lsb of the transmitted acc, = 2^-23)
// used by: CClock



//**************************************************************************
// the following constants are used internally and in communication with the
// evaluation; they may be changed by the programmer of the simulation or
// by the evaluation, but have no effect on the user
//**************************************************************************

// right now, no constants fall into this group


//**************************************************************************
// the following constants are simulation details; they may be changed by
// the programmer of the simulation, but have no (direct) effect on the
// evaluation or the user
//**************************************************************************

//--------------------------------------------------------------------------
// general constants
//--------------------------------------------------------------------------

const SINT32 NoAccount = -1;  // the constant should be < 0!
// this constant is generally used to indicate that the requested data
// was not found
// used by: a lot of classes
// note: if the value is changed, at least assure that it is a negative
// value!

const INT32 ArrayAllocationSize = 16;  // the constant should be > 0!
// number of array elements that are freshly allocated whenever there is
// not enough allocated memory to store a new object
// used by: the list classes (ACObjectList, ...)


//--------------------------------------------------------------------------
// node topology constants
//--------------------------------------------------------------------------

const INT32 NumberOfAreas = 6;  // the constant should be > MAXASICS!
// number of slots on the carrier board
// the number of areas on one node; the constant is determined by the
// number of NTI modules that the carrier board can hold
// used by: several classes that check if the 'area' parameter requested
// is within acceptable bounds
// note: the simulation assumes that every area within [0, NumberOfAreas)
// can be used; if a slot on the carrier board is not occupied, the user
// must not create an application on that area, since the simulation does
// not check if a slot is occupied
// note: the user can only create applications on the areas [0, MAXASICS),
// and there should be at least one more slot for the master clock


//--------------------------------------------------------------------------
// socket constants
//--------------------------------------------------------------------------

const int MaxTrials = 5;  // the constant should be > 0!
// default number of s_ident() calls before giving up
// used by: CClientSocket

const double SocketTimeout = 2;
// the default time an s_receive() call may block (in seconds); set it
// to a value < 0 if you wish to block until a message is received
// used by: CSocket

const INT32 MaxMessageSize = 1024;  // the constant should be >= 256
// maximum size of a message (useful for avoiding allocation of too
// much memory if the byte order conversion fails)
// used by: CStreamSocket

const int EvalsysPort = 3000;
// the default port of the evaluation system (stream socket)
// used by: main()

const int ErrorPort = 4000;
// the default port of the evaluation system error handler (datagram socket)
// used by: main()


//--------------------------------------------------------------------------
// GPS constants
//--------------------------------------------------------------------------

const double GpsDelay = 0.01;  // the constant should be >= 0!
// the delay (in seconds) from the GPS 1PPS signal to calling the ISR
// (pure simulation only)
// note: set it to 10-100 ms (?)
// used by: CGpsInterrupt

const double GpsEpsilon = 5*1e-7;  // the constant should be >= 0!
// the clock value of a P-Node must differ from the GPS time by more
// than this value if the difference is to be amortized
// used by: ACGps


//--------------------------------------------------------------------------
// error report constants
//--------------------------------------------------------------------------

const double ErrorReportPeriod = 1;  // the constant should be > 0!
// the period (in seconds) of the error report task/process
// used by: CErrorTask (and derived classes)

const unsigned int MaxErrorMessages = 100; // the constant should be > 0!
// the number of error messages the error report task/process processes
// in one round
// used by: CErrorTask (and derived classes)

const unsigned int MaxErrorLength = 256;  // the constant should be > 0!
// the maximum length of an error string
// note: the fixed part of an error message has at least 33 bytes, and
// more if additional information like the module is reported; to this
// you have to add the information you want to print (expression, file
// name, line number, module, ...)
// used by: error.cpp, CErrorTask

const unsigned int MaxErrorNumber = 200;  // the constant should be > 0!
// the maximum number of error messages that the error queue can hold
// if there are more errors than the queue can hold, they will be discarded!
// used by: CErrorTask


