//
// cont_sw.cpp
//
// implementation of the synchronization controller (pure simulation)
//
// created: 25.01.97
//
// last change: 23.06.98
//
// RCS version control information:
//     $Revision: 1.4 $$State: Stable $$Date: 1998/07/29 16:23:44 $
//


#include "ci.hpp"
#include "cont_sw.hpp"
#include "error.hpp"
#include "globals.hpp"
#include "network.hpp"
#include "sim_glob.hpp"
#include "supervis.hpp"
#include "template.hpp"


extern EBool g_finished;
// currently the 'tester' process is notified of the end of the simulation
// (as requested from the EvalSys with the command KillSimulation) through
// this variable


#ifndef NO_RESOURCE
extern void ClearISRs (SINT32 nodeID, int area);
// the function is defined in interupt.cpp and only available in pure
// simulation if NO_RESOURCE is not defined; in that case, ISRs from
// deleted CSAs have to be removed from the vector table by hand
#endif


// dummy functions for ci_init(); should never be called
int dummy_announce (renvx *, bufx *, INT16, int)
{
    Assert (False);
    return 0;
}

void dummy_error_announce (tenvx *, bufx *, INT16, int, int)
{
    Assert (False);
}

extern int ci_init (int if_num, ci_fn ci_ann_addr, ci_efn ci_ann_err_addr);
// ci_init() is called directly, since the module is only needed in
// pure simulation, where direct calls are possible


//**************************************************************************
// class CSoftwareController: this class controls all clock synchronization
//     applications and the network; for the network it is the link to
//     the simulation
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

CSoftwareController::CSoftwareController (char qname[4], ACSocketAddress *addr)
    : ACController (qname, addr)
{
    networkID_ = 1;  // network IDs start with 1
}

CSoftwareController::~CSoftwareController ()
{
    DeleteAllSimulationObjects();
    // note: the function must be called in every derived class, because
    // it calls StopApplication() if there are any supervisors in the
    // listOfSupervisors, and this function would not be available in the
    // dtor of the base class
}


//--------------------------------------------------------------------------
//--- reset of the simulation
//--------------------------------------------------------------------------

void CSoftwareController::Reset()
{
    networkID_ = 1;  // network IDs again start with 1
    ACController::Reset();  // reset the base class
}


// reset the simulation (set all parameters in all applications and
// networks to their default values); returns True if at least one
// object exists, False if no simulation objects exist
EBool CSoftwareController::ResetTheSimulation()
{
    EBool result;

    for (ACNetwork *network = listOfNetworks.GetFirst();
         network != NULL;
         network = listOfNetworks.GetNext())
    {
        network->Reset();
    } // for

    result = ACController::ResetTheSimulation();  // reset supervisors etc.

    if (!result && !listOfNetworks.Empty()) result = True;
    return result;
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
EBool CSoftwareController::SetParameter (const SParameter& param)
{
    ACNetwork *network;

    switch (param.module)
    {
        case Controller:
            return ACController::SetParameter (param);
            break;
        case Network:
            // find out the network
            network = listOfNetworks.Find (((SNetworkParameter*)&param)->networkID);
            if (network == NULL) return False;
            else return network->SetParameter (param);
            break;

        default:
            return False;  // unknown parameter
    } // switch
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned (the
// parameter-fields may nevertheless be changed!)
EBool CSoftwareController::GetParameter (SParameter& param) const
{
    ACNetwork *network;

    switch (param.module)
    {
        case Controller:
            return ACController::GetParameter (param);
            break;
        case Network:
            // again, find out the network
            network = listOfNetworks.Find (((SNetworkParameter*)&param)->networkID);
            if (network == NULL) return False;
            else return network->GetParameter (param);
            break;

        default:
            return False;  // unknown parameter
    } // switch
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False
// existing commands: all Controller-Commands of ECommandType in params.hpp
EBool CSoftwareController::ExecuteCommand (const SCommand& command)
{
    switch (command.module)  // check the module (Controller or Network)
    {
        case Controller:  // is handled below
            break;
        case Network:  // deliver to the network
            return ExecuteNetworkCommand (*(SNetworkCommand*)&command);
            break;

        default:
            return False;
    } // switch

    switch (command.type)  // check the type of the Controller command
    {
        case CreateNetwork:
            return CreateNewNetwork (*(SNetworkCreateCommand*)&command);
            break;
        case DeleteNetwork:
            return DeleteANetwork (*(SNetworkCommand*)&command);
            break;

        default:  // deliver to the base class
            return ACController::ExecuteCommand (command);
    } // switch
}


//--------------------------------------------------------------------------
//--- starting/stopping applications
//--------------------------------------------------------------------------

// activates the supervisor process
// assert: the supervisor must exist
EBool CSoftwareController::StartApplication (SINT32 nodeID, int area)
{
    CSupervisor *sup = listOfSupervisors.Find (nodeID, area);
    Assert (sup != NULL);  // should never happen

    listOfClocks.Insert (sup->GetClock(), nodeID, area);
    // insert the clock into the list of clocks

    listOfLances.Insert (sup->GetLance(), nodeID, area);
    // insert the lance into the list of lances

#ifndef NO_RESOURCE
    Resource::ref (sup);  // ref the process in C++SIM
#endif
    sup->Activate();  // activate the process

    return True;
}

// cancels the supervisor process, terminates it
// assert: the supervisor must exist
EBool CSoftwareController::StopApplication (SINT32 nodeID, int area)
{
    CSupervisor *sup = listOfSupervisors.Find (nodeID, area);
    Assert (sup != NULL);  // should never happen

    sup->SuspendOperation();  // stop working (no new interrupts)
    sup->Reset();  // reset everything (that also kills the clock processes)

    sup->Cancel();
    sup->terminate();  // terminate the process

    listOfClocks.Remove (nodeID, area);  // remove objects from the lists
    listOfLances.Remove (nodeID, area);
    // note: it is assumed that the supervisor is deleted (or unref'd) in
    // the base class (by calling listOfSupervisors.Delete())

#ifndef NO_RESOURCE
    ClearISRs (nodeID, area);
    // remove the ISRs of that CSA from the vector table (because the
    // CSA is not deleted at once and therefore cannot do it, and we
    // cannot know if EvalSys does not create another CSA with the same
    // nodeID and area)
#endif

    return True;
}


//--------------------------------------------------------------------------
//--- stop program execution
//--------------------------------------------------------------------------

// empty clock list, delete network list, call base class function
void CSoftwareController::DeleteAllSimulationObjects ()
{
    listOfClocks.RemoveAll();  // remove all clock entries
    // note: the clocks will be deleted during deletion of the supervisors

    listOfLances.RemoveAll();  // remove all lance entries
    // note: lances will be deleted during deletion of the supervisors

    ACController::DeleteAllSimulationObjects();  // delete all supervisors

    listOfNetworks.DeleteAll();  // delete all networks
    // has to come after the deletion of the supervisor, because otherwise
    // the application may still try to send messages to a network that is
    // not there anymore (it's no problem, but it induces an error report)

    // note: these statements assume that there are never two controllers
    // in the same context (using the same lists)!
}


// deschedule the process (the function will not return)
void CSoftwareController::StopExecution()
{
    g_finished = True;  // notify the tester
    Cancel();  // stop execution
}


//--------------------------------------------------------------------------
//--- reaction to network commands
//--------------------------------------------------------------------------

// find out the network and call its ExecuteCommand() function
EBool CSoftwareController::ExecuteNetworkCommand (SNetworkCommand& command)
{
    ACNetwork *network;
    network = listOfNetworks.Find (command.networkID);
    if (network == NULL) return False;
    return network->ExecuteCommand (command);
}

// create a new network, return the networkID to EvalSys
// if the network type is not installed, False is returned
EBool CSoftwareController::CreateNewNetwork (SNetworkCreateCommand& command)
{
    NetworkFunction func = networkManager.GetFunction (command.networkType);
    // get the create function

    if (func==NULL) return False;  // check if the network type is installed

    func (this, networkID_);  // create an object of the network

    Assert (ci_init (networkID_, dummy_announce, dummy_error_announce) == 0);
    // initialize the IFC

    ReplyToMessage ((double)networkID_);  // send the networkID to EvalSys
    networkID_++;  // increment the first free ID for networks
    return True;
}

// if the network exists, it is deleted and True is returned; else False
// is returned
EBool CSoftwareController::DeleteANetwork (SNetworkCommand& command)
{
    return listOfNetworks.Delete (command.networkID);
}



