//
// crashing.cpp
//
// implementation of a class that can crash and recover
//
// created: 11.08.97
//
// last change: 12.08.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include "crashing.hpp"
#include "general.hpp"
#include "params.hpp"
#include "structs.hpp"


//**************************************************************************
// class CCrashingModule: can crash and recover
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

// calls Reset() to set members to default values
CCrashingModule::CCrashingModule ()
{
    Reset();
}


//--------------------------------------------------------------------------
//--- reset all members to their initial values
//--------------------------------------------------------------------------

// sets/restores the initial state
void CCrashingModule::Reset ()
{
    crashed_ = False;
    probCrash_ = 0;
    probRecover_ = 0;
}


//--------------------------------------------------------------------------
//--- control the system parameters
//--------------------------------------------------------------------------

// if 'probCrash' is True, member 'probCrash_' is changed, else member
// 'probRecover_'; the function returns True if 'value' is in [0, 1],
// and in that case the member is set to 'value'
EBool CCrashingModule::SetProbability (EBool probCrash, double prob)
{
    if ((prob < 0) || (prob > 1)) return False;

    if (probCrash) probCrash_ = prob;
    else probRecover_ = prob;

    return True;
}

// if 'probCrash' is True, the value of member 'probCrash_' is returned,
// else the value of member 'probRecover_' is returned
double CCrashingModule::GetProbability (EBool probCrash) const
{
    if (probCrash) return probCrash_;
    else return probRecover_;
}

// returns the current crash status of the module
EBool CCrashingModule::Crashed () const
{
    return crashed_;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False
// existing commands: CrashModule, RecoverModule,
EBool CCrashingModule::ExecuteCommand (const SCommand& command)
{
    switch (command.type)
    {
        case CrashModule:
            Crash();
            break;
        case RecoverModule:
            Recover();
            break;

        default:
            return False;
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- control module crashes
//--------------------------------------------------------------------------

// crash the module; if the module was already crashed, this method has
// no effect
void CCrashingModule::Crash ()
{
    crashed_ = True;
    ReportCrashChange();  // report the new state
}

// recover the module; if the module wasn't crashed, this method has no
// effect
void CCrashingModule::Recover ()
{
    crashed_ = False;
    ReportCrashChange();  // report the new state
}


//--------------------------------------------------------------------------
//--- fault injection
//--------------------------------------------------------------------------

// may change the crash status of the module (according to the probabilities);
// returns True if the status was changed
EBool CCrashingModule::ModifyCrashStatus ()
{
    EBool change;

    if (crashed_) change = CastDice (probRecover_);// should module recover?
    else change = CastDice (probCrash_);  // should module crash?

    if (change)  // crash status should change
    {
        if (crashed_) Recover();  // recover a crashed module
        else Crash();  // crash a recovered module
    }

    return change;
}

// derived modules should send an event with type 'CrashStatusChange' to
// the evaluation
void CCrashingModule::ReportCrashChange()
{
}


