//
// ctrled.cpp
//
// implementation of the base class used for all modules that are controlled
// by some other module
//
// created: 13.01.97
//
// last change: 18.08.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <stdlib.h>  // for the NULL macro

#include "ctrled.hpp"
#include "ctrling.hpp"
#include "error.hpp"


//**************************************************************************
// class CControlledObject: this class stores the link to the controlling
//    object and provides a function for reporting events
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CControlledObject::CControlledObject (CControllingObject *ctrl)
{
    Assert (ctrl != NULL);
    // no sense in having no controller to report to

    ctrl_ = ctrl;
}

CControlledObject::~CControlledObject () {}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

// calls Dispose() and Init()
void CControlledObject::Reset()
{
    Dispose();
    Init();
}


//--------------------------------------------------------------------------
//--- get processID of controlling object (= nodeID for CSAs)
//--------------------------------------------------------------------------

// returns the process ID of the controlling object
SINT32 CControlledObject::GetNodeID()
{
    return ctrl_->GetNodeID();
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// always returns False
EBool CControlledObject::SetParameter (const SParameter&)
{
    return False;
}


// always returns False
EBool CControlledObject::GetParameter (SParameter&) const
{
    return False;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// returns True if the command was executed, else False
// known commands: ResetModule
EBool CControlledObject::ExecuteCommand (const SCommand& command)
{
    switch (command.type)
    {
        case ResetModule:
            Reset();
            break;

        default:
            return False;
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- event reports
//--------------------------------------------------------------------------

// report an event to the controlling object
void CControlledObject::ReportEvent (EModule module, EEvent eventType)
{
    SEvent *event;

    event = new SEvent;
    event->module = module;
    event->type = eventType;
    ctrl_->NoteEvent (event);
}


// pass the event to the controlling object
void CControlledObject::ReportEvent (SEvent *event)
{
    ctrl_->NoteEvent (event);
}


//--------------------------------------------------------------------------
//--- create/destroy the members that are modified during reset
//--------------------------------------------------------------------------

// creates and/or initializes the members that are modified during
// reset; empty in this base class
void CControlledObject::Init ()
{
}

// deletes all dynamic members; empty in this base class
void CControlledObject::Dispose ()
{
}



//**************************************************************************
// class CApplControlledObject: like CControlledObject, but uses only
//    SApplicationEvent's; use it as a base for all application specific
//    modules
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

CApplControlledObject::CApplControlledObject (CControllingObject *ctrl) :
    CControlledObject (ctrl) { }


//--------------------------------------------------------------------------
//--- event reports
//--------------------------------------------------------------------------

// report an event to the controlling object
void CApplControlledObject::ReportEvent (EModule module, EEvent eventType)
{
    SApplEvent *event;

    event = new SApplEvent;
    event->module = module;
    event->type = eventType;
    CControlledObject::ReportEvent (event);
}

void CApplControlledObject::ReportEvent (SEvent *event)
{
    CControlledObject::ReportEvent (event);
}


