//
// ctrling.cpp
//
// implementation of the base class used for all modules that are
// controlling some other module
//
// created: 13.01.97
//
// last change: 26.01.99
//
// RCS version control information:
//     $Revision: 1.7 $$State: Exp $$Date: 1999/01/26 18:00:08 $
//


#include <string.h>  // for memset

#include "ctrling.hpp"
#include "error.hpp"
#include "sockstre.hpp"


//**************************************************************************
// module global functions
//**************************************************************************

static INT32 SecondsToTicks (double period);
// converts 'period' (in seconds) into pSOS ticks
/*
static inline INT32 SecondsToTicks (double period)
{
    if (period*TICKS2SEC < 1) return 1;  // wait at least one tick!
    else return (INT32)(period*TICKS2SEC);
}
*/


//**************************************************************************
// class CControllingObject: the class is a basic interface between modules
//    and the evaluation; it provides a function for reporting events, and
//    executes the base loop that services the sockets to the evaluation
//**************************************************************************


EBool CControllingObject::initialized_ = False;
EEventClass CControllingObject::table_[EventCount];


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// creates the event queue and the stream socket to which it should connect
CControllingObject::CControllingObject (char qname[4], ACSocketAddress *addr,
    SINT32 nodeID)
    : ACProcess (nodeID)
{
    const INT32 QFlags = Q_LOCAL | Q_NOLIMIT | Q_FIFO | Q_SYSBUF;

    socket_ = new CStreamClient (addr);
    Assert (socket_ != NULL);

    q_create (qname, 0, QFlags, &eventQueueID_);

    // initialize the event classification table
    if (!initialized_ && !InitEventTable ())
    {
        Report (event table has leak(s));
    }

    Assert (sizeof (void*) <= sizeof (INT32));
    // this condition is needed because the queue stores event pointers
    // using INT32 variables

    resync_ = False;

    waitPeriod_ = DefaultIfcPeriod;  // set the default wait period
}

CControllingObject::~CControllingObject ()
{
    classMask_.Reset();  // reset the class mask (no more events)
    EmptyQueue ();  // empty the event queue
    q_delete (eventQueueID_);
    delete socket_;
}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

// empty the event queue, reset event report masks
void CControllingObject::Reset()
{
    eventMask_.Reset();  // reset event mask
    classMask_.Reset();  // reset the class mask
    EmptyQueue ();  // discard all events in the queue
    waitPeriod_ = DefaultIfcPeriod;  // set the default wait period
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
// existing parameters: EventReport, EventClassReport, InterfacePeriod
EBool CControllingObject::SetParameter (const SParameter& param)
{
    switch (param.type)
    {
        case EventReport:
            UpdateMask (eventMask_, (SEventReport *) &param);
            break;
        case EventClassReport:
            UpdateMask (classMask_, (SEventReport *) &param);
            break;
        case InterfacePeriod:
            if (((SSingleParameter *)&param)->value <= 0) return False;
            // periods have to be positive
            waitPeriod_ = ((SSingleParameter *)&param)->value;
            break;

        default:  // unknown parameter
            return False;
    }  // switch

    return True;
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned
// existing parameters: EventReport, EventClassReport, InterfacePeriod,
// TimerGranularity
EBool CControllingObject::GetParameter (SParameter& param) const
{
    switch (param.type)
    {
        case EventReport:
            GetBit (eventMask_, (SEventReport *) &param);
            break;
        case EventClassReport:
            GetBit (classMask_, (SEventReport *) &param);
            break;
        case InterfacePeriod:
            ((SSingleParameter *)&param)->value = waitPeriod_;
            break;
        case TimerGranularity:
            ((SSingleParameter *)&param)->value = 1/TICKS2SEC;
            break;

        default:  // unknown parameter
            return False;
    }  // switch

    return True;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False
// existing command: ResetModule
EBool CControllingObject::ExecuteCommand (const SCommand& command)
{
    switch (command.type)
    {
        case ResetModule:
            Reset ();
            break;

        default:
            return False;  // unknown command
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- event logging
//--------------------------------------------------------------------------

// stores the event until it can be processed
void CControllingObject::NoteEvent (SEvent *event)
{
    INT32 msg[4];

    Assert (event != NULL);

    event->time = GetGlobalTime ();  // get the global time, store it

    // now check if the event should be reported
    if (!EventReportAllowed (event->type))  // don't report
    {
        delete event;  // discard the event
        return;
    }

    if (event->type == Snapshot)
        event->time = GetSnapshotTime ();  // get time of last snapshot

    msg[0] = (INT32)event;  // store the event pointer in the message!

    q_send (eventQueueID_, msg);  // store the event in the queue
}


//--------------------------------------------------------------------------
//--- the main function
//--------------------------------------------------------------------------

void CControllingObject::Body()
{
    SEvent *event;  // for the handling of events
    EMessageType type;  // first byte in a socket msg should be the type

    for (;;)
    {
        // check the input socket for new messages from the evaluation
        while (socket_->ReceiveMessageHeader (&type))
        // something came from the socket
        {
            if (resync_ && (type != MsgSynchronizationByte)) continue;
            // ignore all data until the sync byte is received

            switch (type)
            {
                case MsgCommand:
                case MsgSetParameter:
                case MsgGetParameter:  // all three expect further data
                    ProcessData (type);
                    break;
                case MsgResynchronize:  // send resync byte
                    socket_->SendSynchronizationByte();
                    break;
                case MsgSynchronizationByte:  // turn off resync mode
                    resync_ = False;
                    break;

                default:  // oops, something went wrong --> resynchronize
                    Resynchronize();
                    break;
            } // switch
        } // while

        // send all events to the evaluation
        while ((event = GetEvent()) != NULL)
        {
            TransmitEvent (event);  // send the event
            delete event;  // delete the event
        }  // while

        DoPeriodicActivities();
        // give the derived class a chance to do own periodic work

        tm_wkafter (SecondsToTicks (waitPeriod_));  // wait some time
    } // for
}


//--------------------------------------------------------------------------
//--- internal socket handling
//--------------------------------------------------------------------------

// returns the address of the socket
ACSocketAddress *CControllingObject::GetSocketAddress ()
{
    return socket_->GetSocketAddress();
}


// send a MsgReply (containing the value) over the socket
void CControllingObject::ReplyToMessage (double value)
{
    SDoubleReply reply (value);
    socket_->ReplyToMessage (&reply);
}

// send a MsgReply over the socket, containing the result of a message
void CControllingObject::ReplyToMessage (EBool result)
{
    SBooleanReply reply (result);
    socket_->ReplyToMessage (&reply);
}

// send a MsgReply over the socket, followed by the data structure
void CControllingObject::ReplyToMessage (SBase *data)
{
    Assert (data != NULL);

    SReply reply;
    INT32 replysize = reply.size;
    char *buf;

    reply.size = replysize+data->size;  // that's the whole message size

    buf = (char*)socket_->GetMessageBuffer (reply.size);
    Assert (buf != NULL);

    memcpy (buf, &reply, replysize);  // copy the reply into the buffer
    memcpy (buf+replysize, data, data->size); // followed by the data itself
    // note: the first item in the buffer is the reply, so the buffer can
    // be cast to SBase

    socket_->ReplyToMessage ((SBase*)buf);
    socket_->ReturnMessageBuffer (buf);
}

// send the module type over the socket
// assert: module is one of {Controller, Supervisor}
void CControllingObject::SendModule (EModule module)
{
    Assert ((module == Controller) || (module == Supervisor));
    if (!socket_->SendModule (module)) Report (could not send module type);
}

// send the byte-order message over the socket
void CControllingObject::SendByteOrder()
{
    if (!socket_->SendByteOrder()) Report (could not send byte-order);
}

// send the enum-size message over the socket
void CControllingObject::SendEnumSizes()
{
    SEnumSizes msg;  // the sizes are set automatically in the ctor
    if (socket_->SendMessage (MsgEnumSizes, &msg)==(INT32)NoAccount)
        Report (could not send enum-sizes);
}


//--------------------------------------------------------------------------
//--- internal event retrieval
//--------------------------------------------------------------------------

// returns the next event from the queue, or NULL if the queue is empty
SEvent *CControllingObject::GetEvent ()
{
    INT32 msg[4];

    if (q_receive (eventQueueID_, Q_NOWAIT, 0, msg) == 0)
    {
        // msg[0] is the pointer to the event
        return (SEvent*) msg[0];
    }
    else return NULL;  // no more events in the queue
}


//--------------------------------------------------------------------------
//--- periodic activities of derived classes
//--------------------------------------------------------------------------

// is called once every period; derived classes should overload this
// function if they have own periodic work to do
void CControllingObject::DoPeriodicActivities()
{
    // all activities of the base class are carried out in Body()
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int CControllingObject::ReportError (EReportType type, char *file, int line,
                                     char *expression)
{
    return ::ReportError (type, file, line, expression);
    // just call the global function
    // note: cannot make this a pure virtual function because in that case
    // the compiler (GNU) complains about its use in the ctor
}


//--------------------------------------------------------------------------
//--- internal queue management
//--------------------------------------------------------------------------

// deletes all events that are stored in the queue
// you should set the event class mask to zero before calling this function
void CControllingObject::EmptyQueue ()
{
    SEvent *event;
    INT32 msg[4];

    while (q_receive (eventQueueID_, Q_NOWAIT, 0, msg) == 0)
    {
        // msg[0] is the pointer to the event
        event = (SEvent*) msg[0];
        delete event;  // delete the event
    }
    // let's hope this will never result in an endless loop
}


//--------------------------------------------------------------------------
//--- internal event management
//--------------------------------------------------------------------------

// returns True if the event may be reported, else False
EBool CControllingObject::EventReportAllowed (EEvent event)
{
    if (!(eventMask_ & event) ||  // event report turned off
        !(classMask_ & table_[event]))  // whole class turned off
    {
        return False;
    }
    else return True;  // both event and class are turned on
}


//--------------------------------------------------------------------------
//--- internal management of the bitmasks
//--------------------------------------------------------------------------

void CControllingObject::UpdateMask (CBitMask& bitmask, const SEventReport *report)
{
    if (report->masksize <= 0)  // have to set a single event
    {
        SINT32 bit;

        report->Get (bit);
        if (report->set) bitmask |= bit;  // set the bit
        else if (bitmask & bit) bitmask ^= bit;
        // bit is set -> clear the bit
    }
    else  // have to set the whole mask
    {
        bitmask.Set (report->mask, report->masksize);
    }
}

// store the value of a bit in 'report'
void CControllingObject::GetBit (const CBitMask& bitmask,
                                 SEventReport *report) const
{
    SINT32 bit;

    report->Get (bit);  // obtain the bit that is of interest
    if (bitmask & bit) report->set = True;  // the bit is set
    else report->set = False;  // the bit is not set
}


//--------------------------------------------------------------------------
//--- internal management of event classifications
//--------------------------------------------------------------------------

EBool CControllingObject::InitEventTable ()
{
    if (initialized_) return True;
    // table has already been initialized

    int i;

    initialized_ = True;

    for (i=0; i<EventCount; i++) table_[i] = NoClassification;
    // initialize the table to allow checking for leaks (if events are
    // added later on)


    // below follows a simple table that may be put into an own file event.cpp
    // in future versions

    // Clock
    table_[ClockUpdate] = SystemReport;
    table_[ClockPreload] = SystemReport;
    table_[Snapshot] = InterruptEvent;
    table_[GpsEvent] = InterruptEvent;
    table_[ClockReadFault] = InterruptEvent;
    table_[ClockWriteFault] = InterruptEvent;
    table_[IgnoreSetting] = InterruptEvent;

    // Network
    table_[MessageLost] = FaultInjection;
    table_[DeliverMessage] = SystemReport;
    table_[EarlyMessage] = FaultInjection;
    table_[LateMessage] = FaultInjection;
    table_[CrashStatusChange] = FaultInjection;

    // Lance
    table_[DiscardSendMessage] = FaultInjection;
    table_[DiscardRecvMessage] = FaultInjection;
    table_[DuplicateSendMessage] = FaultInjection;
    table_[DuplicateRecvMessage] = FaultInjection;
    table_[ScrambleSendMessage] = FaultInjection;
    table_[ScrambleRecvMessage] = FaultInjection;

    // ClockSync
    table_[ChangeAssociation] = SystemReport;
    table_[ConvergenceFailed] = SystemReport;
    table_[DelayEstimation] = SystemReport;
    table_[IntervalEvent] = SystemReport;
    table_[RateEvent] = SystemReport;
    table_[ReceivedMessage] = InterruptEvent;
    table_[SendMessage] = InterruptEvent;
    table_[TimerEvent] = InterruptEvent;
    table_[TransmissionDelay] = SystemReport;
    table_[UserEvent] = SystemReport;


    // now go through the table once more, search for leaks;
    // if a leak is found: check the above table against EEvent, make sure
    // that all events of EEvent are listed above
    // if the leak occurs because the values of EEvent are not continuous,
    // and if this is intended, skip the hole in the code below
    for (i=0; i<EventCount; i++)
    {
        if (table_[i] == NoClassification) return False;  // found a leak
    } // for

    return True;
}


//--------------------------------------------------------------------------
//--- internal resynchronization
//--------------------------------------------------------------------------

// ask the evaluation to resynchronize, go into resync mode
void CControllingObject::Resynchronize()
{
    socket_->Resynchronize();
    resync_ = True;
}


//--------------------------------------------------------------------------
//--- internal socket data handling
//--------------------------------------------------------------------------

// process data from the socket
void CControllingObject::ProcessData (EMessageType msgtype)
{
    SBase *msgbuf;  // rest of the data, structure SBase is assumed

    // get the message size and body
    if (!socket_->ReceiveMessage (&msgbuf))
    {
        Resynchronize();
        return;
    }

    switch (msgtype)
    {
        case MsgCommand:
            ReplyToMessage (ExecuteCommand (*(SCommand*)msgbuf));
            break;
        case MsgSetParameter:
            ReplyToMessage (SetParameter (*(SParameter*)msgbuf));
            break;
        case MsgGetParameter:
            if (!GetParameter (*(SParameter*)msgbuf))
                ReplyToMessage (False);
            else
                ReplyToMessage (msgbuf);
            break;

        default:
            Assert (False);  // should never happen
    } // switch

    socket_->ReturnMessageBuffer (msgbuf);  // not needed anymore
}

// send the event to the evaluation
// assert: event != NULL
void CControllingObject::TransmitEvent (SEvent *event)
{
    Assert (event != NULL);
    socket_->SendMessage (MsgEvent, event);
}



//**************************************************************************
// module global functions
//**************************************************************************

// converts 'period' (in seconds) into pSOS ticks
static INT32 SecondsToTicks (double period)
{
    if (period*TICKS2SEC < 1) return 1;  // wait at least one tick!
    else return (INT32)(period*TICKS2SEC);
}


