//
// database.cpp
//
// implementation of the database class that is used to store the
// information collected about the network transmission times and about
// other nodes
//
// created: 29.04.97
//
// last change: 08.10.98
//
// RCS version control information:
//     $Revision: 1.4 $$State: Exp $$Date: 1998/10/08 16:19:49 $
//


#include "address.hpp"
#include "addrman.hpp"
#include "database.hpp"
#include "datatemp.hpp"
#include "error.hpp"
#include "structs.hpp"


//**************************************************************************
// class CApplicationDatabase: stores information about other nodes (rates,
//      current time intervals, transmission characteristics); storing
//      additional information is possible
//      data is accessed with the key <if_num, netAddress[, area]>
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CApplicationDatabase::CApplicationDatabase()
{
    initCalled_ = False;
    CApplicationDatabase::Init();
}

CApplicationDatabase::~CApplicationDatabase()
{
    CApplicationDatabase::Dispose();
}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

// discard all data
void CApplicationDatabase::Reset ()
{
    Dispose();
    Init();
}


//--------------------------------------------------------------------------
//--- configuration of system parameters - IFC to evalsys
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or the entry is not found or the value is not valid, False
// is returned
// parameters: Delta, EpsilonPlus, EpsilonMinus, ThetaPlus, ThetaMinus, U
EBool CApplicationDatabase::SetParameter (const SParameter& param)
{
    SDelay delay;
    SRate *rate;
    CAddr netAddress;

    STransmissionData *data = (STransmissionData *)&param; // easier access
    int area;  // for the rate parameters

    switch (param.type)
    {
        case Delta:  // all three access the delay list
        case EpsilonPlus:
        case EpsilonMinus:
            // try to convert the nodeID into a network address
            if (!addressManager_->FindAddress (data->networkID,
                data->nodeID, netAddress)) return False;

            if (!GetDelay (delay, data->networkID, netAddress)) return False;

            switch (param.type)  // do the type specific things
            {
                case Delta:
                    delay.delta = data->value;
                    break;
                case EpsilonPlus:
                    delay.epsilonPlus = data->value;
                    break;
                case EpsilonMinus:
                    delay.epsilonMinus = data->value;
                    break;
                default:
                    Assert (False);  // should never happen
            } // switch

            SetDelay (delay, data->networkID, netAddress);  // store data
            break;

        case ThetaPlus:  // all three access the rate list
        case ThetaMinus:
        case U:
            // try to convert the nodeID into a network address
            if (!addressManager_->FindAddress (data->networkID,
                data->nodeID, netAddress)) return False;
            area = ((SRateData*)data)->area;  // makes access easier

            rate = GetRate (data->networkID, netAddress, area);
            if (rate == NULL) return False;

            switch (param.type)  // now do the type specific things
            {
                case ThetaPlus:
                    rate->thetaPlus = data->value;
                    break;
                case ThetaMinus:
                    rate->thetaMinus = data->value;
                    break;
                case U:
                    rate->u = data->value;
                    break;
                default:
                    Assert (False);  // should never happen
            } // switch

            break;

        default:
            return False;
    } // switch

    return True;
}

// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter or entry does not exist, False is
// returned
EBool CApplicationDatabase::GetParameter (SParameter& param) const
{
    SDelay delay;
    SRate *rate;
    CAddr netAddress;

    STransmissionData *data = (STransmissionData *)&param; // easier access
    int area;  // for the rate parameters

    switch (param.type)
    {
        case Delta:  // all three access the delay list
        case EpsilonPlus:
        case EpsilonMinus:
            // try to convert the nodeID into a network address
            if (!addressManager_->FindAddress (data->networkID,
                data->nodeID, netAddress)) return False;

            if (!GetDelay (delay, data->networkID, netAddress)) return False;

            switch (param.type)  // do the type specific things
            {
                case Delta:
                    data->value = delay.delta;
                    break;
                case EpsilonPlus:
                    data->value = delay.epsilonPlus;
                    break;
                case EpsilonMinus:
                    data->value = delay.epsilonMinus;
                    break;
                default:
                    Assert (False);  // should never happen
            } // switch
            break;

        case ThetaPlus:  // all three access the rate list
        case ThetaMinus:
        case U:
            // try to convert the nodeID into a network address
            if (!addressManager_->FindAddress (data->networkID,
                data->nodeID, netAddress)) return False;
            area = ((SRateData*)data)->area;  // makes access easier

            rate = GetRate (data->networkID, netAddress, area);
            if (rate == NULL) return False;

            switch (param.type)  // now do the type specific things
            {
                case ThetaPlus:
                    data->value = rate->thetaPlus;
                    break;
                case ThetaMinus:
                    data->value = rate->thetaMinus;
                    break;
                case U:
                    data->value = rate->u;
                    break;
                default:
                    Assert (False);  // should never happen
            } // switch
            break;

        default:
            return False;
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- general access to the data of all types
//--------------------------------------------------------------------------

// returns the number of entries for the given type and network
INT32 CApplicationDatabase::Cardinality (EDataType type, SINT32 networkID)
{
    CDelaySet *dset;
    CRateSet *rset;
    CTimeSet *tset;

    switch (type)
    {
        case Delay:
            dset = listOfDelays_->Find (networkID);
            if (dset != NULL) return dset->Cardinality();
            break;
        case Rate:
            rset = listOfRates_->Find (networkID);
            if (rset != NULL) return rset->Cardinality();
            break;
        case Time:
            tset = listOfTimes_->Find (networkID);
            if (tset != NULL) return tset->Cardinality();
            break;
        default:
            Assert (False);  // should not happen
    } // switch

    return 0;  // no entry for the given type and network
}

// deletes all entries of the given type and network
void CApplicationDatabase::DeleteAll (EDataType type, SINT32 networkID)
{
    switch (type)
    {
        case Delay:
            listOfDelays_->Delete (networkID);
            break;
        case Rate:
            listOfRates_->Delete (networkID);
            break;
        case Time:
            listOfTimes_->Delete (networkID);
            break;
        default:
            Assert (False);  // should not happen
    } // switch
}

// deletes all entries of the given type (for all networks)
void CApplicationDatabase::DeleteAll (EDataType type)
{
    switch (type)
    {
        case Delay:
            listOfDelays_->DeleteAll ();
            break;
        case Rate:
            listOfRates_->DeleteAll ();
            break;
        case Time:
            listOfTimes_->DeleteAll ();
            break;
        default:
            Assert (False);  // should not happen
    } // switch
}

// deletes the entry; if type==Delay, the area parameter is ignored
// note: area is no default param to avoid errors if Rate or Time is
// accessed; inconvenient use for Delay is better than error prone use
// for the other types
void CApplicationDatabase::Delete (EDataType type, SINT32 networkID,
                                   const CAddr& netAddress, int area)
{
    CDelaySet *dset;
    CRateSet *rset;
    CTimeSet *tset;

    switch (type)
    {
        case Delay:
            dset = listOfDelays_->Find (networkID);
            if (dset != NULL) dset->Delete (netAddress);
            break;
        case Rate:
            rset = listOfRates_->Find (networkID);
            if (rset != NULL) rset->Delete (netAddress, area);
            break;
        case Time:
            tset = listOfTimes_->Find (networkID);
            if (tset != NULL) tset->Delete (netAddress, area);
            break;
        default:
            Assert (False);  // should not happen
    } // switch
}


//--------------------------------------------------------------------------
//--- access to the transmission characteristics
//--------------------------------------------------------------------------

// store the data; any old data is overwritten
void CApplicationDatabase::SetDelay (const SDelay& delay, SINT32 networkID,
    const CAddr& netAddress)
{
    CDelaySet *set = listOfDelays_->Find (networkID);  // get the set
    if (set == NULL)  // set does not exist
    {
        set = new CDelaySet;  // make a new set
        listOfDelays_->Insert (set, networkID);  // store the set
    }  // if

    SDelay *data = set->Find (netAddress);  // search the set
    if (data == NULL)
    {
        data = new SDelay;  // create a new entry
        set->Insert (data, netAddress);  // store the entry
    } // if

    *data = delay;  // copy the delay data
}

// if the entry exists, it is copied into 'delay' and True is returned;
// else False is returned
EBool CApplicationDatabase::GetDelay (SDelay& delay, SINT32 networkID,
    const CAddr& netAddress) const
{
    CDelaySet *set = listOfDelays_->Find (networkID);  // get the set
    if (set == NULL) return False;

    SDelay *data = set->Find (netAddress);  // search the set
    if (data == NULL) return False;  // no data stored

    delay = *data;  // copy the delay data
    return True;
}

// remove the entry; if it is found, it is copied into 'delay' and True
// is returned, else False is returned
EBool CApplicationDatabase::RemoveDelay (SDelay& delay, SINT32 networkID,
                                        const CAddr& netAddress)
{
    CDelaySet *set = listOfDelays_->Find (networkID);  // get the set
    if (set == NULL) return False;

    SDelay *data = set->Remove (netAddress);  // search the set
    if (data == NULL) return False;  // no data stored

    if (set->Empty()) listOfDelays_->Delete (networkID);  // delete the set

    delay = *data;  // copy the delay data
    delete data;
    return True;
}


// if the list is not empty, the first entry is copied into 'delay'
// and True is returned; else False is returned
EBool CApplicationDatabase::GetFirstDelay (SDelay& delay, SINT32 networkID)
{
    CDelaySet *set = listOfDelays_->Find (networkID);  // get the set
    if (set == NULL) return False;
    SDelay *ptr = set->GetFirst();
    Assert (ptr != NULL);  // empty sets should not happen
    delay = *ptr;
    return True;
}

// if the list is not empty and the last element is not reached, the
// next entry is copied into 'rate' and True is returned; else False
// is returned
EBool CApplicationDatabase::GetNextDelay (SDelay& delay, SINT32 networkID)
{
    CDelaySet *set = listOfDelays_->Find (networkID);  // get the set
    if (set == NULL) return False;
    SDelay *ptr = set->GetNext();
    if (ptr == NULL) return False;
    delay = *ptr;
    return True;
}


//--------------------------------------------------------------------------
//--- access to the rate information
//--------------------------------------------------------------------------

// store the data; if there is already an entry stored, it is deleted
// assert: rate != NULL
void CApplicationDatabase::SetRate (SRate *rate, SINT32 networkID,
    const CAddr& netAddress, int area)
{
    Assert (rate != NULL);

    CRateSet *set = listOfRates_->Find (networkID);  // get the set
    if (set == NULL)  // set does not exist
    {
        set = new CRateSet;  // make a new set
        listOfRates_->Insert (set, networkID);  // store the set
    }  // if

    SRate *oldRate = set->Remove (netAddress, area);  // remove old entry
    if (oldRate != NULL) delete oldRate;  // old entry exists -> delete it

    set->Insert (rate, netAddress, area);  // store the new entry
}

// if the entry exists, it is returned; otherwise, the function returns NULL
SRate *CApplicationDatabase::GetRate (SINT32 networkID,
    const CAddr& netAddress, int area) const
{
    CRateSet *set = listOfRates_->Find (networkID);  // get the set
    if (set == NULL) return NULL;  // the set does not exist

    return set->Find (netAddress, area);  // return the entry or NULL
}

// remove the entry; if it is found, the pointer to the data is returned;
// otherwise, NULL is returned
SRate *CApplicationDatabase::RemoveRate (SINT32 networkID,
    const CAddr& netAddress, int area)
{
    CRateSet *set = listOfRates_->Find (networkID);  // get the set
    if (set == NULL) return NULL;  // the set does not exist

    SRate *data = set->Remove (netAddress, area);  // search the set
    if (data == NULL) return NULL;  // no data stored

    if (set->Empty()) listOfRates_->Delete (networkID);  // delete empty set

    return data;  // return the data
}

// if the list is not empty, the first entry is returned; else NULL is
// returned
SRate *CApplicationDatabase::GetFirstRate (SINT32 networkID)
{
    CRateSet *set = listOfRates_->Find (networkID);  // get the set
    if (set == NULL) return NULL;  // the set does not exist
    SRate *data = set->GetFirst();  // get the first entry
    Assert (data != NULL);  // it must exist (empty sets should not happen)
    return data;  // return the first entry
}

// if the list is not empty and the last element is not reached, the
// next entry is returned; else NULL is returned
SRate *CApplicationDatabase::GetNextRate (SINT32 networkID)
{
    CRateSet *set = listOfRates_->Find (networkID);  // get the set
    if (set == NULL) return NULL;  // the set does not exist
    return set->GetNext();  // return the next entry
}


//--------------------------------------------------------------------------
//--- access to the time intervals
//--------------------------------------------------------------------------

// store the data; any old data is overwritten
void CApplicationDatabase::SetTime (const STimeInterval& time,
    SINT32 networkID, const CAddr& netAddress, int area)
{
    CTimeSet *set = listOfTimes_->Find (networkID);  // get the set
    if (set == NULL)  // set does not exist
    {
        set = new CTimeSet;  // make a new set
        listOfTimes_->Insert (set, networkID);  // store the set
    }  // if

    STimeInterval *data = set->Find (netAddress, area);  // search the set
    if (data == NULL)
    {
        data = new STimeInterval;  // create a new entry
        set->Insert (data, netAddress, area);  // store the entry
    } // if

    *data = time;  // copy the time data
}

// if the entry exists, it is copied into 'time' and True is returned;
// else False is returned
EBool CApplicationDatabase::GetTime (STimeInterval& time, SINT32 networkID,
    const CAddr& netAddress, int area) const
{
    CTimeSet *set = listOfTimes_->Find (networkID);  // get the set
    if (set == NULL) return False;

    STimeInterval *data = set->Find (netAddress, area);  // search the set
    if (data == NULL) return False;  // no data stored

    time = *data;  // copy the time data
    return True;
}

// remove the entry; if it is found, it is copied into 'time' and True
// is returned, else False is returned
EBool CApplicationDatabase::RemoveTime (STimeInterval& time, SINT32 networkID,
                                       const CAddr& netAddress, int area)
{
    CTimeSet *set = listOfTimes_->Find (networkID);  // get the set
    if (set == NULL) return False;

    STimeInterval *data = set->Remove (netAddress, area);  // search the set
    if (data == NULL) return False;  // no data stored

    if (set->Empty()) listOfTimes_->Delete (networkID);  // delete the set

    time = *data;  // copy the time data
    delete data;
    return True;
}

// if the list is not empty, the first entry is copied into 'time'
// and True is returned; else False is returned
EBool CApplicationDatabase::GetFirstTime (STimeInterval& time, SINT32 networkID)
{
    CTimeSet *set = listOfTimes_->Find (networkID);  // get the set
    if (set == NULL) return False;
    STimeInterval *ptr = set->GetFirst();
    Assert (ptr != NULL);  // empty sets should not happen
    time = *ptr;
    return True;
}

// if the list is not empty and the last element is not reached, the
// next entry is copied into 'time' and True is returned; else False
// is returned
EBool CApplicationDatabase::GetNextTime (STimeInterval& time, SINT32 networkID)
{
    CTimeSet *set = listOfTimes_->Find (networkID);  // get the set
    if (set == NULL) return False;
    STimeInterval *ptr = set->GetNext();
    if (ptr == NULL) return False;
    time = *ptr;
    return True;
}


//--------------------------------------------------------------------------
//--- manage association nodeID <-> network address - for IFC to Evalsys
//--------------------------------------------------------------------------

// stores the association <networkID, address, nodeID>; if another
// association already exists, the new data is ignored and False is
// returned, else the triple is stored and True is returned; overload
// the function in the derived class to obtain different behaviour
EBool CApplicationDatabase::SetAssociation (SINT32 networkID,
    const CAddr& address, SINT32 nodeID)
{
    if (addressManager_->Insert (networkID, nodeID, address) != AssocStored)
        return False;
    else return True;
}

// returns True, if an associated network address exists; in that case,
// the address is copied into netAddress; if the assoc. does not exist,
// False is returned and netAddress is not modified
EBool CApplicationDatabase::GetAssociation (SINT32 networkID, SINT32 nodeID,
                                           CAddr& netAddress)
{
    return addressManager_->FindAddress (networkID, nodeID, netAddress);
}

// returns True, if an associated nodeID exists; in that case, the
// value is copied into nodeID; if the assoc. does not exist, False is
// returned and nodeID is not modified
EBool CApplicationDatabase::GetAssociation (SINT32 networkID,
    const CAddr& netAddress, SINT32* nodeID)
{
    return addressManager_->FindNodeID (networkID, netAddress, nodeID);
}

// deletes the netAddress associated with the <networkID,nodeID> key
// from the list of addresses; if there is no entry, False is returned,
// else True is returned
EBool CApplicationDatabase::DeleteAssociation (SINT32 networkID, SINT32 nodeID)
{
    return addressManager_->DeleteNodeID (networkID, nodeID);
}


//--------------------------------------------------------------------------
//--- create/destroy the members that are modified during reset
//--------------------------------------------------------------------------

// creates the members that are modified during a reset; a second call
// (without Dispose() in between) has no effect
void CApplicationDatabase::Init ()
{
    if (initCalled_) return;  // Init() has already been called
    initCalled_ = True;

    addressManager_ = new CAddressManager;
    Assert (addressManager_ != NULL);

    listOfDelays_ = new CDelaySetList;
    Assert (listOfDelays_ != NULL);

    listOfRates_ = new CRateSetList;
    Assert (listOfRates_ != NULL);

    listOfTimes_ = new CTimeSetList;
    Assert (listOfTimes_ != NULL);
}

// deletes the dynamic members created by Init(); a second call
// (without Init() in between), has no effect
void CApplicationDatabase::Dispose ()
{
    if (!initCalled_) return;  // Dispose() has already been called
    delete addressManager_;
    delete listOfDelays_;
    delete listOfRates_;
    delete listOfTimes_;
    initCalled_ = False;
}


