//
// DISTRIB.CPP - Distribution Classes
//
// Random Number Generators for 32-bit Machines
//
// created: ??.??.98, Gerald Hummel
//
// last change: 25.05.98, Gerald Hummel
//
// last revision: 03.02.99
//
// RCS version control information:
//     $Revision: 1.16 $$State: Exp $$Date: 1999/02/03 15:57:51 $
//


#include <stdio.h>
#include <math.h>

#include "distrib.hpp"
#include "error.hpp"

#ifdef DOS_SIM
    #define PI         3.14159265358979323846
    #define MAXDOUBLE  1.701412e+38
    #define MINDOUBLE (2.938736e-39)
    // both double bounds are taken from the Linux include file values.h,
    // they are for VAX and recommended because they are very conservative

int signgam = 0;

#endif

#ifndef PSOS_SIM
  extern unsigned long g_streamOffset;
  // the random streams start at a random offset if SimUTC is called
  // with the -i parameter (indeterministic simulation behaviour)
#else
  const unsigned long g_streamOffset = 0;
  // in hardware-based simulation, the hardware causes enough variations,
  // so an indeterministic start of the random streams is not necessary
#endif


//**************************************************************************
// data structure for drawing random numbers
//**************************************************************************

// checks if low>=0, low+constant>=0, low<=mean<=high; if all three
// conditions are fulfilled, then the function returns True, otherwise False
EBool SRandomData::DataValid() const
{
    if (low < 0) return False;
    if (low+constant < 0) return False;
    if (low > mean) return False;
    if (mean > high) return False;
    return True;  // everything is okay
}


//**************************************************************************
// distribution functions
//**************************************************************************

/* Constants for Random Generators */
static const int limit_gam=500;
static const int wei_tabsize=463; /* depends on the number of for-Iterations */

// prototype declarations from gnuplot functions 
static double lngamma(double z);
static double lgampos(double z);
static double lgamneg(double z);
const unsigned long StreamLength = 1000;
// length of a random stream, used for selecting a stream


distributions::distributions (long MGSeed, long LCGSeed)
{
    // Clean up input parameters
    if ((MGSeed&1) == 0) MGSeed--;
    if (MGSeed<0) MGSeed = -MGSeed;
    if (LCGSeed<0) LCGSeed = -LCGSeed;

    // Initialise state
    MSeed = MGSeed;
    LSeed = LCGSeed;

    oldMSeed = MSeed;
    oldLSeed = LSeed;
    difference = 0;

    for (unsigned int i=0; i < (sizeof(series)/sizeof(double)); i++)
        series[i] = MGen();

    low = 0;
    range = 1;
}

void distributions::SetSeeds (long mgSeed, long lcgSeed, long distance)
{
    Assert ((mgSeed > 0) && (lcgSeed > 0));
    unsigned int i;
    MSeed = mgSeed;
    LSeed = lcgSeed;
    difference = 0;

    for (i=0; i< (sizeof(series)/sizeof(double)); i++)
        series[i] = MGen();
    // well, that's the state of the series, if MSeed is the initial seed
/**/// note: if I understand the principles of random generators correctly,
    // the initialization should never be necessary if distance=SizeOfSeries;
    // check this!

    // now try to restore the changes that were made to series
    for (i=0; (long)i<distance; i++) GetUniform();
    // draw enough random numbers to restore the state
}

// return the seeds of the generators
void distributions::GetSeeds (long& mgSeed, long& lcgSeed,
                              long& distance) const
{
    mgSeed = oldMSeed;
    lcgSeed = oldLSeed;
    distance = difference;
}


//--------------------------------------------------------------------------
//--- linear congruential generator
//--------------------------------------------------------------------------

double distributions::GetUniform () 
{
    // A linear congruential generator based on the algorithm from
    // "Algorithms", R. Sedgewick, Addison-Wesley, Reading MA, 1983.
    // pp. 36-38.
    const long m=100000000;
    const long b=31415821;
    const long m1=10000;

    // Do the multiplication in pieces to avoid overflow
    long    p0 = LSeed%m1,
            p1 = LSeed/m1,
            q0 = b%m1,
            q1 = b/m1;

    LSeed = (((((p0*q1+p1*q0)%m1)*m1+p0*q0)%m) + 1) % m;

    // The results of the LC generator are shuffled with
    // the multiplicative generator as suggested by
    // Maclaren and Marsaglia (See Knuth Vol2, Seminumerical Algorithms)

    long choose = LSeed % (sizeof(series)/sizeof(double));

    double result = series[choose];
    series[choose] =  MGen();

    if (difference < SizeOfSeries) difference++;
    // keep the old seeds, but increment the distance to the current seeds
    else
    {
        oldMSeed = MSeedGen (oldMSeed);  // compute the next seeds
        oldLSeed = LSeedGen (oldLSeed);
    }
    return low + range * result;
}


//--------------------------------------------------------------------------
//--- multiplicative generator
//--------------------------------------------------------------------------

double distributions::MGen ()
{
    // A multiplicative generator, courtesy I. Mitrani 1992,
    // private correspondence
    // Y[i+1] = Y[i] * 5^5 mod 2^26
    // period is 2^24, initial seed must be odd

    const long int two2the26th = 67108864;      // 2**26

    MSeed = (MSeed * 25) % two2the26th;
    MSeed = (MSeed * 25) % two2the26th;
    MSeed = (MSeed * 5) % two2the26th;

    return (double) MSeed / (double) two2the26th;
}


//--------------------------------------------------------------------------
//--- restore generator to a previous state
//--------------------------------------------------------------------------

// returns the MSeed that would result in a call to MGen() with the
// seed mSeed
long distributions::MSeedGen (long mSeed) const
{
    // A multiplicative generator, courtesy I. Mitrani 1992,
    // private correspondence
    // Y[i+1] = Y[i] * 5^5 mod 2^26
    // period is 2^24, initial seed must be odd

    const long int two2the26th = 67108864;      // 2**26

    mSeed = (mSeed * 25) % two2the26th;
    mSeed = (mSeed * 25) % two2the26th;
    mSeed = (mSeed * 5) % two2the26th;

    return mSeed;
}

// returns the LSeed that would result in a call to GetUniform() with the
// seed lSeed
long distributions::LSeedGen (long lSeed) const
{
    // A linear congruential generator based on the algorithm from
    // "Algorithms", R. Sedgewick, Addison-Wesley, Reading MA, 1983.
    // pp. 36-38.
    const long m=100000000;
    const long b=31415821;
    const long m1=10000;

    // Do the multiplication in pieces to avoid overflow
    long    p0 = lSeed%m1,
            p1 = lSeed/m1,
            q0 = b%m1,
            q1 = b/m1;

    lSeed = (((((p0*q1+p1*q0)%m1)*m1+p0*q0)%m) + 1) % m;

    return lSeed;
}


//**************************************************************************
// class CUniform: interval bounds are not fixed but can be changed
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

CUniform::CUniform (double low, double high, int streampos)
{
    SetRange (0, low, high);
    for (unsigned int ss=0; ss<StreamLength*streampos+g_streamOffset; ss++)
        (void) GetUniform();  // draw numbers until we get to start position
}


//--------------------------------------------------------------------------
//--- access to the random generator
//--------------------------------------------------------------------------

double CUniform::operator() ()
{
    return con + GetUniform();
}


//--------------------------------------------------------------------------
//--- manipulate lower and upper bound
//--------------------------------------------------------------------------

void CUniform::SetRange (double c, double l, double h)
{
    low = l;
    range = h-l;
    con=c;
}

//----------------------- E X P O N E N T I A L ------------------------------

C_exponential::C_exponential(double c, double low, double high, double E,
                             int streampos)
{
    set(c,low,high,E);
    for (unsigned int ss=0; ss<StreamLength*streampos+g_streamOffset; ss++)
        (void) GetUniform();  // draw numbers until we get to start position
}

double C_exponential::operator() ()
{   double uni_dev;
    uni_dev=GetUniform();
    if (uni_dev<MINDOUBLE) return hi;
    else return con+(lo+((hi-lo)*(-log(1-uni_dev))*mean)/limit);
}

void C_exponential::set(double c, double low, double high,
                        double E)
{   con=c;lo=low;hi=high;
    limit=-log(MAXDOUBLE);
    mean=(E-lo)*limit/(hi-lo);
}

//--------------------------- N O R M A L ----------------------------------- 

C_normal::C_normal(double c, double low, double high, double E, double StDv,
                   int streampos)
{
    set(c,low,high,E,StDv);
    for (unsigned int ss=0; ss<StreamLength*streampos+g_streamOffset; ss++)
        (void) GetUniform();  // draw numbers until we get to start position
}

double C_normal::operator() ()
{   double fac,r,v1,v2,val1,val2;
    for(;;)
    {   for(;;)
        {   v1=2*GetUniform()-1;v2=2*GetUniform()-1;  /* get x and y values */
            r=v1*v1+v2*v2;
            if(r<1) break;  /* accept values => inside unit circle */
            if(buf_idx>-1)   /* if variate_buffer not empty... */
            return con+norm_var_buf[buf_idx--];
            /* ...take value from buffer */
        }
        fac=sqrt(-2*log(r)/r);val1=mean+stdev*v1*fac;val2=mean+stdev*v2*fac;
        if((lo<=val1)&&(val1<=hi))
        {   if(((buf_idx<9)&&(lo<=val2))&&(val2<=hi))
            norm_var_buf[++buf_idx]=val2;   /* buffer full ? */
            return con+(val1);
        }
        if((lo<=val2)&&(val2<=hi)) return con+(val2);
        if(buf_idx>-1) return con+(norm_var_buf[buf_idx--]);
    }
}

void C_normal::set(double c, double low, double high, double E,
                   double StDv)
{   con=c;lo=low;hi=high;
    buf_idx=-1; /* buffer empty */
    mean=E;stdev=StDv;
}

//---------------------------- G A M M A ------------------------------------

C_gamma::C_gamma(double c, double low, double high, double E, int streampos)
{
    set(c,low,high,E);
    for (unsigned int ss=0; ss<StreamLength*streampos+g_streamOffset; ss++)
        (void) GetUniform();  // draw numbers until we get to start position
}

double C_gamma::operator() ()
{   double am,e,s,v1,v2,x,y;
    int j;
 A: if(mean<6)
    {   x=1;
        for(j=1;j<=mean;j++) x=x*GetUniform();
        x=-log(x);
    }
    else
    {   for(;;)
        {   for(;;)
            {   for(;;)
                {   v1=2*GetUniform()-1;v2=2*GetUniform()-1;
                    if((v1*v1+v2*v2)<=1) break;
                }
                y=v2/v1;am=mean-1;s=sqrt(2*am+1);x=s*y+am;
                if(x>0) break;
            }
            e=(1+y*y)+exp(am*log(x/am)-s*y);
            if(GetUniform()<=e) break;
        }
    }
    if(x>(2*limit_gam)) goto A;
    else
    {   if(flag==1) return con+hi-(hi-lo)*x/(2*limit_gam);
        else return con+lo+(hi-lo)*x/(2*limit_gam);
    }
}

void C_gamma::set(double c, double low, double high, double E)
{   con=c;lo=low;hi=high;
    mean=(E-lo)*(2*limit_gam)/(hi-lo);
    if(flag=mean>limit_gam) mean=limit_gam-(mean-limit_gam);

}

//------------------ W E I B U L L - G E N E R A T O R ----------------------

class C_generate_weibull
{ public:

// Constructor -> initialize the Weibull-table:
  C_generate_weibull();

  protected:
  double Mean(double l,double a);  /* Compute Mean for Weibullfunction */
  double Limit(double l,double a); /* Compute Limit for Weibullfunctions */
  double gamma(double a); /* Gammafunction taken from Gnuplot */
};


struct wei_par { double E,l,a,limit,rel; }; /* Weibull_Datastruct */
static wei_par wei_tab[wei_tabsize];  /* Buffer for Weibull_Datafile */

double C_generate_weibull::Mean(double l,double a)/* Compute Mean */
{ return pow((1/l),(1/a))*(gamma(1+1/a)); }

double C_generate_weibull::Limit(double l,double a)/* Compute Limit */ 
{ return pow((-log(MINDOUBLE))/l,(1/a)); }

//Constructor:

C_generate_weibull::C_generate_weibull()
{ int *p,*start,i,j,k; /* table for the weibull parameters E,lamdba,alpha */
  double x,l,a,add;
  Assert ((wei_tab!=NULL)&&(wei_tabsize>0));  /* assure that table exists */
  i=0;l=1;
  /* asymmetric bellshapes */ 
  add=0.001;a=l=1;
  for(;;)
  { a+=add;
    if(a>100) break;
    wei_tab[i].E=Mean(l,a);wei_tab[i].l=l;
    wei_tab[i].a=a;wei_tab[i].limit=Limit(l,a);
    wei_tab[i].rel=wei_tab[i].E*100/wei_tab[i].limit;
    add=0.01*a;i++;
  }
}

double C_generate_weibull::gamma(double a)
{  return(exp(lngamma(a)));
}

//------------------------- W E I B U L L -----------------------------------

// Constructor:

C_weibull::C_weibull(double c, double low, double high, double E,
                     int streampos)
{
    set(c,low,high,E);
    for (unsigned int ss=0; ss<StreamLength*streampos+g_streamOffset; ss++) 
        (void) GetUniform();  // draw numbers until we get to start position 
}

static C_generate_weibull wei_gen;

double C_weibull::operator() ()
{ double x;
next:
  x=pow((-(log(1-GetUniform())))/l,(1/a));
  if(x<weibull_limit) return con+lo+(hi-lo)*x/weibull_limit;
  else goto next;
}

void C_weibull::set(double c, double low, double high, double E)
{ double relation;
  int i,j;
  Assert (wei_tab[0].rel!=0);  /* table must have been created */
  con=c;lo=low;hi=high;
  relation=(E-lo)*100/(hi-lo);
  i=0;  /* search for suitable table entry for E(x) */
  while((wei_tab[i].rel<relation)&&((i+1)<wei_tabsize)) i++;
  mean=wei_tab[i].E;
  l=wei_tab[i].l;
  a=wei_tab[i].a;
  weibull_limit=wei_tab[i].limit;
}

//---------------------- G E N E R I C  -------------------------------------

C_generic::C_generic(int streampos)
{
    set(0,0,1,NULL,0);
    for (unsigned int ss=0; ss<StreamLength*streampos+g_streamOffset; ss++)  
        (void) GetUniform();  // draw numbers until we get to start position  
}

double C_generic::operator() ()
{ double *p=inv_funct,uni=GetUniform();
  while((p<(inv_funct+tabsize-1))&&(*(p+1)<uni)) p+=2;
  return con+lo+(*p)*(hi-lo)/(*(inv_funct+tabsize-2));
}

void C_generic::set(double c, double low, double high,
                    double *inv, unsigned long size)
{ inv_funct=inv;tabsize=size;
  con=c;lo=low;hi=high;
}

//------------------------- R A N D O M -------------------------------------

CRandom::CRandom (int streampos)
  :  normal(0,0,1,0.5,1, streampos), uniform(0,1, streampos),
     exponential(0,0,1,0.5, streampos), gamma(0,0,1,0.5, streampos),
     weibull(0,0,1,0.5, streampos), generic(streampos)
{
}

// returns 'True' with probability 'prob'
// assert: 'prob' is in [0, 1]
EBool CRandom::CastDice (double prob)
{
    Assert ((prob >= 0) && (prob <= 1));  // check if interval correct

    double var = GetUniform();  // get a uniform number
    if (var <= prob) return True;  // check if result is True or False
    else return False;
}


// get a random variable of the selected distribution
double CRandom::RandomVar(EDistribution dist)
{ double random_number=0;
  switch (dist)
  {   case Uniform:      random_number=uniform();break;
      case Exponential:  random_number=exponential();break;
      case Normal:       random_number=normal();break;
      case Gamma:        random_number=gamma();break;
      case Weibull:      random_number=weibull();break;
      case Generic:      random_number=generic();break; 
      default:           random_number=0;break;
  }
  return random_number;
}

// get a random variable of the selected distribution with previous set
double CRandom::RandomVar(EDistribution dist, double c, double low,
                          double high, double mean, double free_par)
{ double random_number=0;

  if (low == high) return low;  // no need to draw anything
  Assert ((low<=mean)&&(mean<=high));  // assure, that mean inside interval
  switch (dist)
  {   case Uniform:      uniform.SetRange(c,low,high);
                         random_number=uniform();break;
      case Exponential:  exponential.set(c,low,high,mean);
                         random_number=exponential();break;
      case Normal:       normal.set(c,low,high,mean,free_par);
                         random_number=normal();break;
      case Gamma:        gamma.set(c,low,high,mean);
                         random_number=gamma();break;
      case Weibull:      weibull.set(c,low,high,mean);
                         random_number=weibull();break; 
      default:           random_number=0;break;
  }
  return random_number;
}

// same as above, but all data is in one structure
double CRandom::RandomVar(SRandomData& data)
{
    return RandomVar (data.dist, data.constant, data.low, data.high,
                      data.mean, data.freeparam);
}

// return uniform value
double CRandom::RandomVar(double low, double high)
{
    return RandomVar (Uniform, 0, low, high, 0, 0);
}

// change the settings of the selected distribution
void CRandom::Set(EDistribution dist, double c, double low,
                  double high, double mean, double free_par)
{ Assert ((low<=mean)&&(mean<=high));  // assure that mean inside interval
  switch (dist)
  {   case Uniform:      uniform.SetRange(c, low, high);break;
      case Exponential:  exponential.set(c, low, high, mean);break;
      case Normal:       normal.set(c,low, high, mean, free_par);break;
      case Gamma:        gamma.set(c,low, high, mean);break;
      case Weibull:      weibull.set(c, low, high, mean);break;
  }
}

void CRandom::SetGeneric(double c, double low, double high,
                         double *inv_funct, unsigned long tabsize)
{ Assert ((inv_funct!=NULL)&&(tabsize>0));  /* assure that table exists */
  Assert (low<=high);  // assure that interval is not empty
  generic.set(c, low, high, inv_funct, tabsize);
}

//--------------------------------------------------------------------------
//--- manipulate the seeds directly
//--------------------------------------------------------------------------

// takes over the seeds for the generators
// assert: mgSeed, lcgSeed > 0
// assert: the distribution is supported
void CRandom::SetSeeds (EDistribution dist, long mgSeed, long lcgSeed,
                        long distance)
{
    Assert ((mgSeed > 0) && (lcgSeed > 0));
    switch (dist)
    {
        case Normal:
            normal.SetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Uniform:
            uniform.SetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Exponential:
            exponential.SetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Gamma:
            gamma.SetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Weibull:
            weibull.SetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Generic:
            generic.SetSeeds (mgSeed, lcgSeed, distance);
            break;
        default:
            Assert (0);
            // is only called if the distribution is not supported
            break;
    } // switch
}

// return the seeds of the generators
// assert: the distribution is supported
void CRandom::GetSeeds (EDistribution dist, long& mgSeed, long& lcgSeed,
                        long& distance) const
{
    switch (dist)
    {
        case Normal:
            normal.GetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Uniform:
            uniform.GetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Exponential:
            exponential.GetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Gamma:
            gamma.GetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Weibull:
            weibull.GetSeeds (mgSeed, lcgSeed, distance);
            break;
        case Generic:
            generic.GetSeeds (mgSeed, lcgSeed, distance);
            break;
        default:
            Assert (0);
            // is only called if the distribution is not supported
            break;
    } // switch
}

//---- G A M M A F U N C T I O N   F R O M   G N U P L O T ------------------

#define ITMAX   100
#ifdef FLT_EPSILON
#define MACHEPS FLT_EPSILON /* 1.0E-08 */
#else
#define MACHEPS 1.0E-08
#endif
#ifdef FLT_MIN_EXP
#define MINEXP  FLT_MIN_EXP /* -88.0 */
#else
#define MINEXP  -88.0
#endif
#ifdef FLT_MAX
#define OFLOW   FLT_MAX /* 1.0E+37 */
#else
#define OFLOW   1.0E+37
#endif
#ifdef FLT_MAX_10_EXP
#define XBIG    FLT_MAX_10_EXP /* 2.55E+305 */
#else
#define XBIG    2.55E+305
#endif

/*
 * Mathematical constants
 */
#define LNPI 1.14472988584940016
#define LNSQRT2PI 0.9189385332046727
#define PNT68 0.6796875
#define SQRTPI 0.9189385332046727417803297
#define SQRT_TWO 1.41421356237309504880168872420969809   /* JG */

#ifndef min /* GCC ST uses inline functions */
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

static double lngamma();
static double lgamneg();
static double lgampos();

/**
 * from statlib, Thu Jan 23 15:02:27 EST 1992 ***
 *
 * This file contains two algorithms for the logarithm of the gamma function.
 * Algorithm AS 245 is the faster (but longer) and gives an accuracy of about
 * 10-12 significant decimal digits except for small regions around X = 1 and
 * X = 2, where the function goes to zero.
 * The second algorithm is not part of the AS algorithms.   It is slower but
 * gives 14 or more significant decimal digits accuracy, except around X = 1
 * and X = 2.   The Lanczos series from which this algorithm is derived is
 * interesting in that it is a convergent series approximation for the gamma
 * function, whereas the familiar series due to De Moivre (and usually wrongly
 * called Stirling's approximation) is only an asymptotic approximation, as
 * is the true and preferable approximation due to Stirling.
 * 
 * Uses Lanczos-type approximation to ln(gamma) for z > 0. Reference: Lanczos,
 * C. 'A precision approximation of the gamma function', J. SIAM Numer.
 * Anal., B, 1, 86-96, 1964. Accuracy: About 14 significant digits except for
 * small regions in the vicinity of 1 and 2.
 * 
 * Programmer: Alan Miller CSIRO Division of Mathematics & Statistics
 * 
 * Latest revision - 17 April 1988
 * 
 * Additions: Translated from fortran to C, code added to handle values z < 0.
 * The global variable signgam contains the sign of the gamma function.
 * 
 * IMPORTANT: The signgam variable contains garbage until AFTER the call to
 * lngamma().
 * 
 * Permission granted to distribute freely for non-commercial purposes only
 * Copyright (c) 1992 Jos van der Woude, jvdwoude@hut.nl
 */

/* Local data, not visible outside this file 
static double   tb[] =
{
    0.9999999999995183E+00,
    0.6765203681218835E+03,
    -.1259139216722289E+04,
    0.7713234287757674E+03,
    -.1766150291498386E+03,
    0.1250734324009056E+02,
    -.1385710331296526E+00,
    0.9934937113930748E-05,
    0.1659470187408462E-06,
};   */

/* from Ray Toy */
static double tb[] = {
        .99999999999980993227684700473478296744476168282198,
     676.52036812188509856700919044401903816411251975244084,
   -1259.13921672240287047156078755282840836424300664868028,
     771.32342877765307884865282588943070775227268469602500,
    -176.61502916214059906584551353999392943274507608117860,
      12.50734327868690481445893685327104972970563021816420,
       -.13857109526572011689554706984971501358032683492780,
        .00000998436957801957085956266828104544089848531228,
        .00000015056327351493115583383579667028994545044040,
};

static double   lgamneg(double z)
{
    double          tmp;

    /* Use reflection formula, then call lgampos() */
    tmp = sin(z * PI);

    if (fabs(tmp) < MACHEPS) {
        tmp = 0.0;
    } else if (tmp < 0.0) {
        tmp = -tmp;
        signgam = -1;
    }
    return LNPI - lgampos(1.0 - z) - log(tmp);

}

static double   lgampos(double z)
{
    double          sum;
    double          tmp;
    int             i;

    sum = tb[0];
    for (i = 1, tmp = z; i < 9; i++) {
        sum += tb[i] / tmp;
        tmp++;
    }

    return log(sum) + LNSQRT2PI - z - 6.5 + (z - 0.5) * log(z + 6.5);
}

static double lngamma(double z)
{
    double signgam = 1.0;
    if (z <= 0.0)
        return lgamneg(z);
    else
        return lgampos(z);
}

