//
// duplist.cpp
//
// definition of an abstract list class using an array; duplicates are
// possible
//
// created: 25.07.96
//
// last change: 15.11.96
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <string.h>  // for memmove etc.

#include "duplist.hpp"
#include "error.hpp"


//**************************************************************************
// class CArrayElement: this class represents an element of the sorted array;
//                      it contains the data that is stored in the array
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

ACSortedArray::CArrayElement::CArrayElement ()
{
    object_ = NULL;
    objectID_ = NoAccount;
}


//--------------------------------------------------------------------------
//--- set the members of the class
//--------------------------------------------------------------------------

// set the object-pointer and object ID
void ACSortedArray::CArrayElement::SetObject (void *object, double objectID)
{
    object_ = object;
    objectID_ = objectID;
}


//--------------------------------------------------------------------------
//--- read the members of the class
//--------------------------------------------------------------------------

// get the object-pointer
void* ACSortedArray::CArrayElement::GetObject () const
{
    return object_;
}


// get the objectID
double ACSortedArray::CArrayElement::GetObjectID () const
{
    return objectID_;
}


//**************************************************************************
// class ACSortedArray: this abstract class manages a sorted array
//     it is assumed that most of the time the access is sequential; if
//     not, binary search is used; duplicates are possible
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// initialize the array
ACSortedArray::ACSortedArray (EBool duplicates, INT32 aSize)
{
    Assert (aSize != 0);  // creating an array with no size is ridiculous!
    
    allocSize_ = aSize;
    size_ = aSize;
    array_ = new CArrayElement[size_];  // create a new array

    firstFree_ = 0;  // no elements in the array
    current_ = 0;
    duplicates_ = duplicates;
}

// has to be there, although the destructor is pure virtual, because
// destructors of base classes are always called
// since the constructor of the class allocates an array, the destructor
// frees the memory
// note that the destructor is pure virtual, because it can't delete the
// objects that are stored in the array; so a derived class *must* exist
// and handle the deletion of the objects
ACSortedArray::~ACSortedArray ()
{
    delete [] array_;
}



//--------------------------------------------------------------------------
//--- get the array status
//--------------------------------------------------------------------------

// returns True if array is empty, False otherwise
EBool ACSortedArray::Empty() const
{
    if (firstFree_ == 0) return True;
    else return False;
}


// returns True if the position pointer is standing at the beginning
// of the array, False otherwise (or if the array is empty)
EBool ACSortedArray::BeginOfList () const
{
    if (!Empty() && (current_ == 0)) return True;
    else return False;
}


// returns True if the position pointer is standing at the end of the array,
// False otherwise (or if the array is empty)
EBool ACSortedArray::EndOfList () const
{
    if (!Empty() && (current_ == firstFree_-1)) return True;
    else return False;
}


//--------------------------------------------------------------------------
//--- move in the array
//--------------------------------------------------------------------------

// get the first element, or NULL if the array is empty
void* ACSortedArray::GetFirst ()
{
    if (Empty()) return NULL;
    else
    {
        current_ = 0;
        return GetCurrent();
    }
}


// get the last element, or NULL if the array is empty
void* ACSortedArray::GetLast ()
{
    if (Empty()) return NULL;
    else
    {
        current_ = firstFree_-1;
        return GetCurrent();
    }
}


// get the next element in the array, or NULL if the end of the array
// has been reached;
// if current_ is no valid index (because the array is empty, or because
// Find() was invoked with an unknown index), NULL is returned
void* ACSortedArray::GetNext ()
{
    if (!CurrentIsValid()) return NULL;
    if (EndOfList()) return NULL;  // last element has been reached

    current_++;

    return GetCurrent();
}


// return the previous element in the array, or NULL if the beginning
// of the array has been reached
// if current_ is no valid index (because the array is empty, or because
// Find() was invoked with an unknown index), NULL is returned
void* ACSortedArray::GetPrev ()
{
    if (!CurrentIsValid()) return NULL;
    if (BeginOfList()) return NULL;  // first element has been reached

    current_--;

    return GetCurrent();
}


// returns the element with this index; if no such element exists,
// NULL is returned
// if duplicates are allowed, the first object in the array with this index
// is returned; note that in this case binary search is used to find any
// occurence of the index, but that from then on linear search is used to
// find the first occurence!
// if the element was found, current_ contains its index; if it
// was not found, then current_ contains the index where it
// should be inserted; note that current_ might be the same as
// firstFree_, so calling GetCurrentID() would not be a good idea!
void* ACSortedArray::Find (double index)
{
    INT32 lowerBound, upperBound;  // bounds for the binary search

    if (Empty()) return NULL;  // no elements in array

    if (!CurrentIsValid()) current_=firstFree_/2;
    // if current is not valid, set it to the middle of the array to help
    // with the search
    
    // if current_ is not the element, set it to one of its neighbours
    // (this makes sequential access faster)
    if ((index < GetCurrentID()) && !BeginOfList()) current_--;
    else if ((index > GetCurrentID()) && !EndOfList()) current_++;

    // now check if current_ points to the element searched
    if (GetCurrentID() == index)  // the element was found
    {
        if (duplicates_)  // duplicates may exist
        {
            while (!BeginOfList() &&
                    ((array_+current_-1)->GetObjectID() == index))
                current_--;  // the prev element has the same index
        }
        
        return GetCurrent();  // return the object
    }


    // since current_ has not found the element, access is neither
    // repeated nor sequential, so binary search is used
    if (index < GetCurrentID())
    {
        lowerBound = 0;
        upperBound = (current_>0) ? (current_-1) : 0;
    }
    else
    {
        lowerBound = current_+1;
        upperBound = firstFree_-1;
    }

    current_ = (lowerBound+upperBound)/2;
    
    while ((GetCurrentID() != index) && (lowerBound < upperBound))
    {
        if (index < GetCurrentID())
            upperBound = (current_>0) ? (current_-1) : 0;
        else lowerBound = current_+1;

        current_ = (lowerBound+upperBound)/2;
    }

    // okay, now either the element was found, or it's not in the array
    if (GetCurrentID() == index)  // the element was found
    {
        if (duplicates_)  // duplicates may exist
        {
            while (!BeginOfList() &&
                    ((array_+current_-1)->GetObjectID() == index))
                current_--;  // the prev element has the same index
        }
        
        return GetCurrent();  // return the object
    }
    else
    {
        if (GetCurrentID() < index) current_++;
        // this is the place where the element should be inserted
        
        return NULL;
    }
}


// returns the next object with the same index as the current_ index;
// if no more objects with this index exist, NULL is returned and
// the function has no effect
void* ACSortedArray::FindNext ()
{
    if (!duplicates_) return NULL;  // there can't be any duplicates
    if (!CurrentIsValid()) return NULL;  // current is no valid element
    if (EndOfList()) return NULL;  // end of the array has been reached

    if ((array_+current_+1)->GetObjectID() == GetCurrentID())
    {
        current_++;  // the next object has the same index
        return GetCurrent();  // return it
    }
    else return NULL;  // no more objects with this ID
}


//--------------------------------------------------------------------------
//--- get information about the current element
//--------------------------------------------------------------------------

// returns a pointer to the member that is currently being accessed or
// NULL, if no element is being accessed
void* ACSortedArray::GetCurrentObject () const
{
    if (CurrentIsValid()) return GetCurrent();
    else return NULL;
}


// returns the index of the currently accessed element or NoAccount,
// if no element is being accessed
double ACSortedArray::GetCurrentObjectID () const
{
    if (CurrentIsValid()) return GetCurrentID();
    else return NoAccount;
}


//--------------------------------------------------------------------------
//--- add/remove elements
//--------------------------------------------------------------------------


// insert a new element into the (sorted) array;
// if an object with this index exists but duplicates are not allowed,
// False is returned
// if duplicates are allowed and elements with the index exist, the new
// element is inserted as the first
EBool ACSortedArray::Insert (void *object, double index)
{
    if (object == NULL) return False;
    // only existing objects are inserted

    if ((Find (index) != NULL) && !duplicates_) // index does already exist
    {                                      // and duplicates are not allowed
        return False;
    }
    
    // it is inserted at the current_
    // position
    InsertAtCurrent (object, index);

    return True;
}


// remove an object; if duplicates are allowed, the first object in the
// array with the appropriate index will be removed
void* ACSortedArray::Remove (double index)
{
    void* object;

    object = Find (index);  // search for the object

    if (object != NULL) RemoveAtCurrent ();
    // remove the current_ data entry

    return object;
}


//--------------------------------------------------------------------------
//--- internal information concerning the current element
//--------------------------------------------------------------------------

// returns True if current_ is an occupied slot, False otherwise
EBool ACSortedArray::CurrentIsValid () const
{
    if (current_ < firstFree_) return True;
    else return False;
}


// returns the object ID of the current element; the element must exist!
double ACSortedArray::GetCurrentID () const
{
    Assert (CurrentIsValid());  // the element has to exist
    
    return (array_ + current_)->GetObjectID();
}


// returns the object of the current element; the element must exist!
void* ACSortedArray::GetCurrent () const
{
    Assert (CurrentIsValid());  // the element has to exist
    
    return (array_ + current_)->GetObject();
}


//--------------------------------------------------------------------------
//--- internal memory management
//--------------------------------------------------------------------------

// allocates a new array of size size_ and copies the data of firstFree_
// elements from array_ to the new array; array_ is then deleted and set
// to the new array
// assert: firstFree_ <= size_
void ACSortedArray::UpdateArray()
{
    CArrayElement *newArray;

    Assert (firstFree_ <= size_);  // the data has to fit the new array
    
    newArray = new CArrayElement[size_];  // make a new array
    memcpy (newArray, array_, firstFree_*sizeof(CArrayElement));
    // copy the data to the new array
    
    delete[] array_;  // delete the old array
    array_ = newArray;  // set the array_-Pointer to the new array
}


// allocates a new array that is allocSize_ elements bigger than the
// current array size stored in size_; the data is copied to the new
// array and array_ is set to it; the old array is discarded
void ACSortedArray::AllocateMemory()
{
    size_ = size_ + allocSize_;  // have to allocate more space
    UpdateArray();  // make an array of the new size and copy the data
}


// frees the last allocSize_ elements of the current array; note that
// the function will assume that these elements are not in use! if
// they are, the call to UpdateArray will fail!
// assert: size>allocSize
void ACSortedArray::FreeMemory()
{
    size_ = size_ - allocSize_;  // reduce the size of the array
    Assert (size_ > 0);  // array must have a positive size
    UpdateArray();  // make an array of the new size and copy the data
}


//--------------------------------------------------------------------------
//--- internal array information
//--------------------------------------------------------------------------

// returns the current size of the array
INT32 ACSortedArray::Size() const
{
    return size_;
}


// returns the number of array elements that are currently in use
INT32 ACSortedArray::Occupied() const
{
    return firstFree_;
}


// returns the additional size used for expansion (contraction) of the array
INT32 ACSortedArray::AllocationSize() const
{
    return allocSize_;
}


// returns True if duplicate entries are allowed, False otherwise
EBool ACSortedArray::Duplicates() const
{
    return duplicates_;
}


//--------------------------------------------------------------------------
//--- internal array manipulation
//--------------------------------------------------------------------------

// insert the new data at current_;
// if current_ is not the end of the array, the following members are
// moved back using memmove()
// if the new data has occupied the last free space, a new array
// with bigger size (determined by allocSize_) is created and data is
// copied using memcpy();
// assert: the array must have at least one element free!
//         current_ must be a valid index or the first free index!
void ACSortedArray::InsertAtCurrent (void *object, double index)
{
    Assert (firstFree_ < size_);  // a free element must exist
    Assert (current_ <= firstFree_);  // current_ must be a valid slot
    
    if (current_ != firstFree_)  // shift the following entries to the right
    {
        memmove ((array_+current_+1), (array_ + current_),
                 (firstFree_-current_)*sizeof(CArrayElement));
        // move all elements from current_ on one item back in the array
        // note: here, memmove() is used because it works properly with
        // overlapping objects (as opposed to memcpy()!)
    }

    (array_ + current_)->SetObject (object, index);
    // insert the new data at the current_ position

    firstFree_++;  // set _free to the next free element

    if (firstFree_ == size_)  // the end of the free buffer is reached
    {
        AllocateMemory();  // allocate new memory
    }
}


// remove the element defined by current_ from the array;
// if the element is not the last entry, the following items are moved
// left using memmove;
// if the removal of the objects leaves more than allocSize_+1 places free,
// allocSize_ elements are freed;
// assert: current_ must be a valid index!
void ACSortedArray::RemoveAtCurrent ()
{
    Assert (CurrentIsValid());  // current_ must be valid

    if (current_ != (firstFree_-1))  // have to move some elements
    {
        memmove ((array_+current_), (array_+current_+1),
                 (firstFree_-current_-1)*sizeof(CArrayElement));
        // move all elements after current_ one slot to the left
    }
    else if (current_ != 0)
        current_--;  // set current to the new last element
    
    firstFree_--;  // the size of the used array was reduced

    if ((size_-firstFree_) > allocSize_+1)
    // note: with this approach, at least two free elements are saved
    // from deallocation; therefore repeated insertion and removal of
    // an element won't result in continuous allocation/disallocation
    // of the memory
    {
        FreeMemory();  // delete the last portion of the array
    }
}



