//
// error.cpp
//
// implementation of the global error report functions
// they report the error to an error report handler
//
// created: 06.01.97
//
// last change: 28.04.98
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 10:49:31 $
//


#ifdef PSOS_SIM
  #include <psos.h>  // for k_fatal()
  #define ABORTFUNC k_fatal
  #define ABORTPARAMS 0x10000, K_LOCAL
  // first param is the error code, second the flags (only local node is
  // shutdown
#else
  #include <stdlib.h>  // for abort()
  #define ABORTFUNC abort
  #define ABORTPARAMS
  // abort does not have any parameters
#endif
#include <string.h>

#include "error.hpp"
#include "errbase.hpp"
#include "general.hpp"


static CErrorHandler *errorHandler = NULL;
// the error handler


//--------------------------------------------------------------------------
// use these functions to (de-)install an error report object; if none is
// installed, the macros will still behave as described below (Assert()
// causes program abortion, Check() returns the value of the expression),
// but there won't be any output
//--------------------------------------------------------------------------

// set the internal error handler pointer to the handler
void InstallErrorHandler (CErrorHandler *handler)
{
    errorHandler = handler;
}

// set the internal error handler pointer to NULL
void DeinstallErrorHandler ()
{
    errorHandler = NULL;
}


//--------------------------------------------------------------------------
// the following function just reports a message
//--------------------------------------------------------------------------

void ReportMessage (const char *msg)
{
    if (errorHandler != NULL) errorHandler->NoteError ((char*)msg);
}


//--------------------------------------------------------------------------
// the following functions report the expression plus file and line number
// of the call to the error reporter (if installed); if they return (which
// is only the case if type != Asserting), they always return 0
// please call these functions from the ReportError() function of the class
//--------------------------------------------------------------------------

static const int MessageLen = 256;
// maximum number of characters the error message may have


// report the expression plus file and linenumber (into the msg string)
// note: there is no check if the length of the assembled string exceeds
// the size of the message buffer!
static void PrintFileSpecs (EReportType type, char *expression,
                            char *file, int line, char *msg)
{
    if (msg == NULL) return;  // can't assert at this point

    char buf[2*sizeof(int)+1];  // for converting the line

    switch (type)
    {
        case Asserting: strcpy (msg, "assert failed: ");
            break;
        case Checking: strcpy (msg, "check failed: ");
            break;
        case Reporting: strcpy (msg, "report: ");
            break;
    } // switch

    // now add the string "<expression> in file <file> at line <line>"
    strcat (msg, expression);
    strcat (msg, " in file ");
    strcat (msg, file);
    strcat (msg, " at line ");
    strcat (msg, itoa (line, buf, 10));
}


// aborts program execution if type=Asserting, else returns 0
static int GetReturnValue (EReportType type)
{
    if (type == Asserting)
    {
        if (errorHandler != NULL) delete errorHandler;
        // calling the dtor forces output before the program is stopped

       ABORTFUNC (ABORTPARAMS);  // stop program execution
    } // if

    return 0;
}


// converts a module number into a string constant
static const char *ModuleToString (EModule module)
{
    switch (module)
    {
        case Controller: return "Controller"; break;
        case Network: return "Network"; break;
        case Supervisor: return "Supervisor"; break;
        case ClockSync: return "ClockSync"; break;
        case Clock: return "Clock"; break;
        case Lance: return "Lance"; break;

        default: return "Unknown"; break;
    } // switch
}


//--------------------------------------------------------------------------
// the following functions report the expression plus file and line number
// of the call to the evaluation system; if they return (which is only
// the case if type != Asserting), they always return 0
//--------------------------------------------------------------------------

int ReportError (EReportType type, char *file, int line, char *expression)
{
    char msg[MessageLen];
    PrintFileSpecs (type, expression, file, line, msg); // assemble the msg
    if (errorHandler != NULL) errorHandler->NoteError (msg);  // report message
    return GetReturnValue (type);
}

int ReportError (EReportType type, char *file, int line, char *expression,
                 EModule module)
{
    char msg[MessageLen];
    PrintFileSpecs (type, expression, file, line, msg);
    strcat (msg, "\n");  // make a new line
    strcat (msg, "module = ");  // add the module
    strcat (msg, ModuleToString(module));
    if (errorHandler != NULL) errorHandler->NoteError (msg);  // report message
    return GetReturnValue (type);
}

int ReportError (EReportType type, char *file, int line, char *expression,
                 EModule module, SINT32 moduleID)
{
    char msg[MessageLen];
    char buf[2*sizeof(int)+1];  // for converting the moduleID

    PrintFileSpecs (type, expression, file, line, msg);
    strcat (msg, "\n");  // make a new line
    strcat (msg, "module = ");  // add the module
    strcat (msg, ModuleToString(module));
    strcat (msg, ", moduleID = ");  // add the module ID
    strcat (msg, itoa (moduleID, buf, 10));
    if (errorHandler != NULL) errorHandler->NoteError (msg);  // report message
    return GetReturnValue (type);
}

int ReportError (EReportType type, char *file, int line, char *expression,
                 EModule module, SINT32 nodeID, int area)
{
    char msg[MessageLen];
    char buf[2*sizeof(int)+1];  // for converting nodeID and area

    PrintFileSpecs (type, expression, file, line, msg);
    strcat (msg, "\n");  // make a new line
    strcat (msg, "module = ");  // add the module
    strcat (msg, ModuleToString(module));
    strcat (msg, ", nodeID = ");  // add the node ID
    strcat (msg, itoa (nodeID, buf, 10));
    strcat (msg, ", area = ");  // add the node ID
    strcat (msg, itoa (area, buf, 10));
    if (errorHandler != NULL) errorHandler->NoteError (msg);  // report message
    return GetReturnValue (type);
}


