//
// gps_sw.cpp
//
// implementation of the GPS class (for pure simulation)
//
// created: 29.03.97
//
// last change: 01.08.97
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 10:49:31 $
//


#include "clock_sw.hpp"
#include "clockifc.hpp"
#include "clockirq.hpp"
#include "general.hpp"
#include "gps_sw.hpp"
#include "time.hpp"

#define OFFSET 0//10000
// just for testing


//**************************************************************************
// class CGpsTemplate: the template class used for obtaining an instance
//      of the urrently used GPS class (software or hardware)
//**************************************************************************

//--------------------------------------------------------------------------
//--- return a new instance of the currently used CSA class
//--------------------------------------------------------------------------

// returns a new object of the class; this function is needed so that
// CClockInterface can obtain the correct GPS class
// assert: clock != NULL (done by the ctor of CSoftwareGps)
ACGps *CGpsTemplate::NewGPS (CClockInterface *clock, int area)
{
    return new CSoftwareGps (clock, area);
}



//**************************************************************************
// class CSoftwareGps: the GPS time is taken from the global time
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assert: clock != NULL (through the ctor of ACGps)
CSoftwareGps::CSoftwareGps (CClockInterface *clock, int area)
    : ACGps (clock, area)
{
}

// disables all GPS receivers that were enabled
CSoftwareGps::~CSoftwareGps ()
{
    for (int i=0; enableMask_ != 0; i++)  // disable all enabled GPS rec.
    {
        if (enableMask_ & (1<<i))
        {
            DisableGps ((EGpsUnit)i);  // disable the GPU
            enableMask_ &= ~(1<<i);  // mask out the bit
        }
    } // for
}


//--------------------------------------------------------------------------
//--- handling GPS time
//--------------------------------------------------------------------------

// enables the GPS time interrupt of the GPS interrupt process
// calls SetGpsStatus() to set the gpsStatus of clock to False
void CSoftwareGps::EnableGps (EGpsUnit gpu)
{
    CSoftwareClock *clock = (CSoftwareClock*)clock_->GetClock ();
    CGpsInterrupt *gps = clock->GetGpsInterrupt (gpu);
    // obtain (maybe create) the interrupt process

    Assert (gps != NULL);

    gps->SetTimeVector (area_);  // set the interrupt vector
    gps->EnableTime();  // enable the interrupt
    enableMask_ |= (1<<gpu);  // store the GPU in the mask
    SetGpsStatus (gpu, False);  // 'gpsStatus' of the clock is set to False
}

// disables the GPS time interrupt of the GPS interrupt process
// clears the status bit in 'statusMask'
void CSoftwareGps::DisableGps (EGpsUnit gpu)
{
    CSoftwareClock *clock = (CSoftwareClock*)clock_->GetClock ();
    CGpsInterrupt *gps = clock->GetGpsInterrupt (gpu);
    // obtain the interrupt process

    Assert (gps != NULL);

    gps->DisableTime();  // disable the interrupt
    enableMask_ &= ~(1<<gpu);  // mask out the GPU
    statusMask_ &= ~(1<<gpu);  // clear the status bit
    // note: do not call SetGpsStatus(), because at that time the GPU must
    // still be enabled, so clock->gpsStatus would be wrongly set to False;
    // clearing the bit keeps the gpsStatus, which is exactly what we want
}

// obtain the NTP time of the GPS receiver that has triggered the
// interrupt
void CSoftwareGps::GetGpsTime (CNtpTime& time)
{
    SecondsToNtp ((INT32)GetGlobalTime()+OFFSET, time);
    // pure simulation: use the global time, but round off to the
    // full second
}

// return the identifier of the GPU to which the receiver that has
// triggered the interrupt is connected
EGpsUnit CSoftwareGps::GetGpsNumber ()
{
    CSoftwareClock *clock = (CSoftwareClock*)clock_->GetClock ();
    return clock->GetGpsUnit();
}

// clear the interrupt that was triggered by the end of the GPS time
// information
void CSoftwareGps::ClearGpsInterrupt()
{
    // nothing to do in software
}


