//
// gpsbase.cpp
//
// implementation of the GPS class (for pure simulation)
//
// note: when porting to hardware, derive a class and overload the
// virtual functions
//
// created: 29.03.97
//
// last change: 13.04.98
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <stdlib.h>  // for NULL

#include "clockbas.hpp"
#include "clockifc.hpp"
#include "constant.hpp"
#include "error.hpp"
#include "gpsbase.hpp"
#include "time.hpp"


CGpsTemplate *gpsTemplate = new CGpsTemplate;
// set the template; in order to use it, one has to define its NewGPS()
// function


//**************************************************************************
// class ACGps: abstract base class providing an ISR for the GPS1PPS
//      interrupt of the UTCSU; it modifies the clock time according to the
//      time obtained from the GPS receiver
//**************************************************************************

static const INT32 GpsNotHealthy = (INT32)1<<31;
// bit 31 in the MSGPS contains the status information of the receiver
/**/// assumption: bit=0 --> healthy


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// stores the clock that is used
// assert: clock != NULL
ACGps::ACGps (CClockInterface *clock, int area)
{
    Assert (clock != NULL);
    clock_ = clock;
    area_ = area;
    enableMask_ = 0;
    statusMask_ = 0;
}

// assert: enableMask_ is 0
ACGps::~ACGps ()
{
    Assert (!enableMask_);  // there must not be any enabled GPS receivers
}


//--------------------------------------------------------------------------
//--- managing GPS receivers
//--------------------------------------------------------------------------

// returns True if at least one GPS receiver is enabled ('enableMask'
// is different from zero), otherwise False
EBool ACGps::GpsEnabled()
{
    if (enableMask_ != 0) return True;
    else return False;
}


//--------------------------------------------------------------------------
//--- GPS status
//--------------------------------------------------------------------------

// notifies the clock interface of the GPS status; set 'status' to
// True if the GPS receiver is healthy, to False if it is not
// assert: the receiver connected to unit 'gpu' is enabled
void ACGps::SetGpsStatus (EGpsUnit gpu, EBool status)
{
    Assert (enableMask_ & (1<<gpu));  // receiver is enabled
    if (!status) statusMask_ &= ~(1<<gpu);
    else statusMask_ |= (1<<gpu);

    if (enableMask_ == statusMask_) clock_->gpsStatus_ = True;  // all healthy
    else clock_->gpsStatus_ = False;
}


//--------------------------------------------------------------------------
//--- internal event reports
//--------------------------------------------------------------------------

// pass the event to the clock (also set module and type of the event)
// assert: event != NULL
void ACGps::ReportEvent (SEvent *event)
{
    Assert (event != NULL);
    event->module = Clock;
    event->type = GpsEvent;
    clock_->ReportEvent (event);
}


//--------------------------------------------------------------------------
//--- get clock time (GPS)
//--------------------------------------------------------------------------

// reads the MSGPS and TSGPS registers of the clock; uses the GPU
// specified in variable 'gpu'
// assert: gpu < EGpsUnitCount
void ACGps::GetClockGps (EGpsUnit gpu, CNtpTime& clockTime)
{
    CClock *clock = clock_->GetClock();  // need direct access to the regs
    Assert (clock != NULL);

    Assert (gpu < EGpsUnitCount);
    clockTime.ms = (clock->GetMSGPS (gpu) & (0x0FFFFFF | GpsNotHealthy));
    clockTime.ts = clock->GetTSGPS (gpu);
    // read the clock value (mask out the checksum, leave the status bit)
    // the clock value is automatically latched on the rising edgle of
    // the GPS1PPS pulse
}


//--------------------------------------------------------------------------
//--- GPS ISRs
//--------------------------------------------------------------------------

// called after the time information from GPS is available; reads the
// GPS time and compares it with the clock value stored; if the value
// difference is more than 'GpsEpsilon', the clock is amortized
// clears the interrupt
void ACGps::isr()
{
    ClearGpsInterrupt ();  // clear the IP bit of the GPS time reception

    CNtpTime clockTime;  // time value at the time of the 1PPS
    CNtpTime gpsTime;  // the time information of the GPS receiver
    EGpsUnit gpu = GetGpsNumber();  // the number of the GPS receiver

    GetClockGps (gpu, clockTime);  // read the clock value (MSGPS, TSGPS)
    GetGpsTime (gpsTime);  // get the GPS value

    if (clockTime.ms & GpsNotHealthy)
    {
        SetGpsStatus (gpu, False);  // notify the clock of unhealthy GPS
        return;  // ignore the GPS time information
    }
    else SetGpsStatus (gpu, True);  // GPS receiver is healthy

    clockTime.ms &= ~GpsNotHealthy;  // mask out the healthy bit

    double diff = NtpToSeconds (gpsTime) - NtpToSeconds (clockTime);
    // compute the time difference between the current clock value
    // and its correct value (instead of using the current values,
    // those at the time of the 1PPS pulse are used; the clock drift
    // is ignored)

    if ((diff > GpsEpsilon) || (-diff > GpsEpsilon))
    {
        SApplGpsEvent *event = new SApplGpsEvent;
        event->diff = diff;
        ReportEvent (event);

        clock_->Amortize (diff, 0, 0);  // amortize time difference
    }
}


