//
// idlist.cpp
//
// definition of the nodelist class
//
// created: 18.10.96
//
// last change: 07.02.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <string.h>  // for memcpy etc.

#include "error.hpp"
#include "idlist.hpp"


//**************************************************************************
// class CNodeIdList: this class stores a sorted array of SINT32
//     it is assumed that most of the time the access is sequential; if
//     not, binary search is used
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// initialize the array
CNodeIdList::CNodeIdList (INT32 aSize)
{
    Assert (aSize != 0);  // creating an array with no size is ridiculous!
    
    allocSize_ = aSize;
    size_ = aSize;
    array_ = new SINT32[size_];  // alloc memory

    firstFree_ = 0;  // no elements in the array
    current_ = 0;
}

CNodeIdList::~CNodeIdList ()
{
    delete [] array_;
}



//--------------------------------------------------------------------------
//--- get the array status
//--------------------------------------------------------------------------

// returns True if array is empty, False otherwise
EBool CNodeIdList::Empty() const
{
    if (firstFree_ == 0) return True;
    else return False;
}


// returns True if the position pointer is standing at the beginning
// of the array, False otherwise (or if the array is empty)
EBool CNodeIdList::BeginOfList () const
{
    if (!Empty() && (current_ == 0)) return True;
    else return False;
}


// returns True if the position pointer is standing at the end of the array,
// False otherwise (or if the array is empty)
EBool CNodeIdList::EndOfList () const
{
    if (!Empty() && (current_ == firstFree_-1)) return True;
    else return False;
}


//--------------------------------------------------------------------------
//--- move in the array
//--------------------------------------------------------------------------

// get the first element, or NoAccount if the array is empty
SINT32 CNodeIdList::GetFirst ()
{
    if (Empty()) return NoAccount;
    else
    {
        current_ = 0;
        return GetCurrent();
    }
}


// get the last element, or NoAccount if the array is empty
SINT32 CNodeIdList::GetLast ()
{
    if (Empty()) return NoAccount;
    else
    {
        current_ = firstFree_-1;
        return GetCurrent();
    }
}


// get the next element in the array, or NoAccount if the end of the
// array has been reached;
// if current_ is no valid index (because the array is empty, or because
// Find() was invoked with an unknown index), NoAccount is returned
SINT32 CNodeIdList::GetNext ()
{
    if (!CurrentIsValid()) return NoAccount;
    if (EndOfList()) return NoAccount;

    current_++;
    return GetCurrent();
}


// returns True if the index is in the list
// if the element was found, current_ contains its index; if it
// was not found, then current_ contains the index where it
// should be inserted; note that current_ might be the same as
// firstFree_, so calling GetCurrentID() would not be a good idea!
EBool CNodeIdList::Find (SINT32 index)
{
    INT32 lowerBound, upperBound;  // bounds for the binary search


    if (Empty()) return False;  // no elements in array

    if (!CurrentIsValid()) current_=firstFree_/2;
    // if current is not valid, set it to the middle of the array to help
    // with the search
    
    // if current_ is not the element, set it to one of its neighbours
    // (this makes sequential access faster)
    if ((index < GetCurrentID()) && !BeginOfList()) current_--;
    else if ((index > GetCurrentID()) && !EndOfList()) current_++;

    // now check if current_ points to the element searched
    if (GetCurrentID() == index)  // the element was found
    {
        return True;
    }


    // since current_ has not found the element, access is neither
    // repeated nor sequential, so binary search is used
    if (index < GetCurrentID())
    {
        lowerBound = 0;
        upperBound = (current_>0) ? (current_-1) : 0;
    }
    else
    {
        lowerBound = current_+1;
        upperBound = firstFree_-1;
    }

    current_ = (lowerBound+upperBound)/2;

    while ((GetCurrentID() != index) && (lowerBound < upperBound))
    {
        if (index < GetCurrentID())
            upperBound = (current_>0) ? (current_-1) : 0;
        else lowerBound = current_+1;

        current_ = (lowerBound+upperBound)/2;
    }

    // okay, now either the element was found, or it's not in the array
    if (GetCurrentID() == index)  // the element was found
    {
        return True;
    }
    else
    {
        if (GetCurrentID() < index) current_++;
        // this is the place where the element should be inserted
        
        return False;
    }
}


//--------------------------------------------------------------------------
//--- add/remove elements
//--------------------------------------------------------------------------


// insert a new element into the (sorted) array;
EBool CNodeIdList::Insert (SINT32 index)
{
    EBool result = Find (index);


    if (result) // index does already exist
    {
        return False;
    }

    // it is inserted at the current_
    // position
    InsertAtCurrent (index);

    return True;
}


// remove an index
EBool CNodeIdList::Remove (SINT32 index)
{
    EBool result = Find (index);  // search for the index

    if (result) RemoveAtCurrent ();
    // remove the current_ data entry

    return result;
}


//--------------------------------------------------------------------------
//--- internal information concerning the current element
//--------------------------------------------------------------------------

// returns True if current_ is an occupied slot, False otherwise
EBool CNodeIdList::CurrentIsValid () const
{
    if (current_ < firstFree_) return True;
    else return False;
}


// returns the object ID of the current element; the element must exist!
SINT32 CNodeIdList::GetCurrentID () const
{
    Assert (CurrentIsValid());  // the element has to exist

    return array_[current_];
}


// returns the object of the current element; the element must exist!
SINT32 CNodeIdList::GetCurrent () const
{
    Assert (CurrentIsValid());  // the element has to exist

    return array_[current_];
}


//--------------------------------------------------------------------------
//--- internal memory management
//--------------------------------------------------------------------------

// allocates a new array of size size_ and copies the data of firstFree_
// elements from array_ to the new array; array_ is then deleted and set
// to the new array
// assert: firstFree_ <= size_
void CNodeIdList::UpdateArray()
{
    SINT32 *newArray;

    Assert (firstFree_ <= size_);  // the data has to fit the new array

    newArray = new SINT32[size_];  // make a new array
    memcpy (newArray, array_, firstFree_*sizeof(SINT32));
    // copy the data to the new array

    delete[] array_;  // delete the old array
    array_ = newArray;  // set the array_-Pointer to the new array
}


// allocates a new array that is allocSize_ elements bigger than the
// current array size stored in size_; the data is copied to the new
// array and array_ is set to it; the old array is discarded
void CNodeIdList::AllocateMemory()
{
    size_ = size_ + allocSize_;  // have to allocate more space
    UpdateArray();  // make an array of the new size and copy the data
}


// frees the last allocSize_ elements of the current array; note that
// the function will assume that these elements are not in use! if
// they are, the call to UpdateArray will fail!
// assert: size>allocSize
void CNodeIdList::FreeMemory()
{
    size_ = size_ - allocSize_;  // reduce the size of the array
    Assert (size_ > 0);  // array must have a positive size
    UpdateArray();  // make an array of the new size and copy the data
}


//--------------------------------------------------------------------------
//--- internal array information
//--------------------------------------------------------------------------

// returns the current size of the array
INT32 CNodeIdList::Size() const
{
    return size_;
}


// returns the number of array elements that are currently in use
INT32 CNodeIdList::Occupied() const
{
    return firstFree_;
}


// returns the additional size used for expansion (contraction) of the array
INT32 CNodeIdList::AllocationSize() const
{
    return allocSize_;
}


//--------------------------------------------------------------------------
//--- internal array manipulation
//--------------------------------------------------------------------------

// insert the new data at current_;
// if current_ is not the end of the array, the following members are
// moved back using memmove()
// if the new data has occupied the last free space, a new array
// with bigger size (determined by allocSize_) is created and data is
// copied using memcpy();
// assert: the array must have at least one element free!
//         current_ must be a valid index or the first free index!
void CNodeIdList::InsertAtCurrent (SINT32 index)
{
    Assert (firstFree_ < size_);  // a free element must exist
    Assert (current_ <= firstFree_);  // current_ must be a valid slot
    
    if (current_ != firstFree_)  // shift the following entries to the right
    {
        memmove ((array_+current_+1), (array_ + current_),
                 (firstFree_-current_)*sizeof(SINT32));
        // move all elements from current_ on one item back in the array
        // note: here memmove is used because it works properly with
        // overlapping objects (as opposed to memcpy!)
    }

    array_[current_] = index;
    // insert the new data at the current_ position

    firstFree_++;  // set _free to the next free element

    if (firstFree_ == size_)  // the end of the free buffer is reached
    {
        AllocateMemory();  // allocate new memory
    }
}


// remove the element defined by current_ from the array;
// if the element is not the last entry, the following items are moved
// left using memmove;
// if the removal of the objects leaves more than allocSize_+1 places free,
// allocSize_ elements are freed;
// assert: current_ must be a valid index!
void CNodeIdList::RemoveAtCurrent ()
{
    Assert (CurrentIsValid());  // current_ must be valid

    if (current_ != (firstFree_-1))  // have to move some elements
    {
        memmove ((array_+current_), (array_+current_+1),
                 (firstFree_-current_-1)*sizeof(SINT32));
        // move all elements after current_ one slot to the left
    }
    else if (current_ != 0)
        current_--;  // set current to the new last element
    
    firstFree_--;  // the size of the used array was reduced

    if ((size_-firstFree_) > allocSize_+1)
    // note: with this approach, at least two free elements are saved
    // from deallocation; therefore repeated insertion and removal of
    // an element won't result in continuous allocation/disallocation
    // of the memory
    {
        FreeMemory();  // delete the last portion of the array
    }
}



