//
// impl_sw.cpp
//
// function implementations that differ in hardware and software; in this
// file, the software implementations are given
//
// created: 30.03.97
//
// last change: 24.07.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/07/28 09:49:46 $
//


#include <stdlib.h>  // for NULL

#include "ci.h"
#include "psos.h"

#include "address.hpp"
#include "clock_sw.hpp"
#include "clockcfg.hpp"
#include "error.hpp"
#include "sim_glob.hpp"
#include "msgform.hpp"
#include "nodetemp.hpp"
#include "process.hpp"
#include "time.hpp"



//**************************************************************************
// obtaining interrupt vectors and levels
//**************************************************************************

// returns the clock interrupt vector for the area on the given node;
// assert: area < NumberOfAreas
// assert: area < 16
// note: when porting to hardware, return the appropriate MOD_x_INT_VEC
// and ignore the nodeID
int GetClockIntvec (SINT32 nodeID, int area)
{
    Assert (area < NumberOfAreas);
    Assert (area < 16);
    return nodeID*0x1000 + 0xA0 + area*0x10;
    // note: this computation limits NumberOfAreas, but the limit of
    // ACController::CreateNewApplication() (which is 16) is the tighter
    // bound
}

// returns the GPS-receiver interrupt vector for the area on the given node
// assert: area < NumberOfAreas, area < 16, gpu < EGpsUnitCount
// note: when porting to hardware, return the appropriate vector for the
// area and gpu and ignore the nodeID
int GetGpsIntvec (SINT32 nodeID, int area, EGpsUnit gpu)
{
    Assert (gpu < EGpsUnitCount);
    return GetClockIntvec (nodeID, area) + 0x40000000;
/**/// this assumes that sizeof(int) >= 4
    // note: this computation limits clock interrupt vectors to 0x03FFFFFF;
    // since the number of nodes is limited to 2^12 anyway (the limit
    // comes from ACController::CreateNewApplication()), that's no problem
}

// returns the interrupt level of the UTCSU and LANCE interrupts; the levels
// should be equal (if they are not, correct behaviour of the simulation
// cannot be guaranteed).
INT8 GetClockIntLevel()
{
    return 6;
    // note: the level is not important in SW; the HW version should return
    // the appropriate macro
}


//**************************************************************************
// get global time
//**************************************************************************

// returns the global time (in seconds)
// note: when porting to hardware, set global variable g_masterClock to a
// value different from NULL in the ctor of CHardwareController and return
// the value of the MSGET and TSGET registers of the master clock (mask
// out the checksum!)
double GetGlobalTime ()
{
    return ACProcess::CurrentTime();
}


//**************************************************************************
// get time of last snapshot
//**************************************************************************

// also just returns the global time (in seconds), because in pure
// simulation this is also the time of the last snapshot
// note: when porting to hardware, return the value of the TSSNU and MSSNU
// registers of the master clock (mask out the checksum!)
double GetSnapshotTime ()
{
    return GetGlobalTime();
}


//**************************************************************************
// get the transmit timestamp
//**************************************************************************

// not supported in pure simulation, so ms/ts are set to zero
// note: in hardware-based simulation, you have to find out which UTCSU
// registers belong to the sender with the network address 'address' and
// the area 'area', read the MS/TS registers and set 'time'.
void GetTransmitTimestamp (const CAddr&, int, const CClock*, STimestamp& time)
{
    time.ms = 0;
    time.ts = 0;
}


//**************************************************************************
// get the position of network time value in message
//**************************************************************************

static ACNodeList listOfPositions_;
// the list of position data

// store the length and position of all data that is automatically added
// to a message by the network into the appropriate fields of 'data'
void SetPosition (const SMsgPosition& data, SINT32 nodeID, SINT32 networkID)
{
    SMsgPosition *pos;
    pos = (SMsgPosition *)listOfPositions_.Find (networkID, nodeID);

    if (pos != NULL) *pos = data;  // just store the new value
    else
    {
        pos = new SMsgPosition;
        *pos = data;
        listOfPositions_.Insert (pos, networkID, nodeID);
    }
}


static void GetDefaultPosition (SMsgPosition& data);

// puts the length and position of all data that is automatically added
// to a message by the network into the appropriate fields of 'data'
// if there is no data stored (using SetPosition()), GetDefaultPosition()
// is used to obtain the default position data and this data is returned
// note: in hardware-based simulation, the 'nodeID' is useless, because
// at compile-time it is not known; instead, you could ignore the nodeID
// and compile a dedicated version for every target node (ie, implement
// the function for every target node)
void GetPosition (SMsgPosition& data, SINT32 nodeID, SINT32 networkID)
{
    SMsgPosition *pos;
    pos = (SMsgPosition *)listOfPositions_.Find (networkID, nodeID);

    if (pos==NULL) GetDefaultPosition (data);  // use the default position
    else data = *pos;
}


// returns the default position data (the added data starts at byte 0 in
// the message buffer, has the structure
//     INT32 area; ti_acx time;
static void GetDefaultPosition (SMsgPosition& data)
{
    ti_acx netTime;  // this structure is put in front of the message
    INT16 offset = 0;  // the offset of this structure

    data.position[SMsgPosition::Area] = 0;  // the area is the first item
    data.length[SMsgPosition::Area] = sizeof(INT32);
    offset = data.length[SMsgPosition::Area];

    data.position[SMsgPosition::Macrostamp] = 
        offset + (INT32)&(netTime.ms) - (INT32)&netTime;
    data.position[SMsgPosition::Timestamp] =
        offset + (INT32)&(netTime.ts) - (INT32)&netTime;
    data.position[SMsgPosition::Accuracy] =
        offset + (INT32)&(netTime.acc) - (INT32)&netTime;

    data.length[SMsgPosition::Macrostamp] = sizeof(netTime.ms);
    data.length[SMsgPosition::Timestamp] = sizeof(netTime.ts);
    data.length[SMsgPosition::Accuracy] = sizeof(netTime.acc);
}


//**************************************************************************
// set/get the clock frequencies
//**************************************************************************

static ACNodeList listOfFrequencies_;
// the list of clock frequencies

// stores the frequency of a clock; the frequencies of all clocks should be
// set after the program is started (when calling the other initialization
// functions
// note: in hardware-based simulation, you can ignore the nodeID
void SetClockFrequency (SINT32 nodeID, int area, double frequency)
{
    double *freq = (double *)listOfFrequencies_.Find (nodeID, area);
    if (freq != NULL) *freq = frequency;  // just store the new value
    else
    {
        freq = new double (frequency);
        listOfFrequencies_.Insert (freq, nodeID, area);
    }
}

// returns the frequency of a clock; the function is called by the Controller
// whenever a new CSA is created
// if no frequency is stored, a default frequency (25 MHz) is returned
// note: in hardware-based simulation, you can ignore the nodeID
double GetClockFrequency (SINT32 nodeID, int area)
{
    static const double FREQUENCY = 25e6;
/**/// default clock frequency; make an entry in constant.hpp!

    double *freq = (double *)listOfFrequencies_.Find (nodeID, area);
    if (freq != NULL) return *freq;  // return the stored frequency
    else return FREQUENCY;  // return the default frequency
}


//**************************************************************************
// trigger hardware snapshots on all UTCSUs
//**************************************************************************

// initiates a hardware snapshot on all UTCSUs in the system (that assumes
// that all UTCSUs are hardwired together, regardless of their subnets)
// note: in hardware-based simulation, an output register is accessed; just
// set the bit and clear it again
void MakeSystemSnapshot()
{
    for (CClockInterface *clock = listOfClocks.GetFirst();
         clock != NULL;
         clock = listOfClocks.GetNext())
    {
        ((CSoftwareClock*)clock->GetClock())->HwSnapshot();
        // initiate the snapshot
    } // for
}



