//
// int128.cpp
// 
// implementation of an INT128 class based on GNU class Integer
//
// created: 06.11.97
//
// last change: 15.06.98
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 10:49:31 $
//


#include "int128.hpp"


//**************************************************************************
// class INT128
//**************************************************************************

//--------------------------------------------------------------------------
// set/get the value directly
//--------------------------------------------------------------------------

// a0 is the lowest INT32, a3 the highest
void INT128::SetBinary (INT32 a3, INT32 a2, INT32 a1, INT32 a0)
{
    unsigned short reg[8];  // register containing the init. value
    *((INT32*)reg)   = a0;
    *((INT32*)reg+1) = a1;
    *((INT32*)reg+2) = a2;
    *((INT32*)reg+3) = a3;

    // create a new IntRep and store it in the inherited member
    rep = Ialloc(rep, reg, 8, 1, 8);

    *this = *this*1;  // let Integer modify the array length
    // note: this is necessary because some Integer operators (comparisons
    // for example) assume that there are no unused array entries (that is,
    // no leading entries which are 0)
}

// a0 is the lowest INT32, a3 the highest
void INT128::GetBinary (INT32& a3, INT32& a2, INT32& a1, INT32& a0)
{
    INT8 len = rep->len;  // for the switch below
    if (len > 8) len = 8;  // all higher numbers are ignored

    a3 = 0; a2 = 0; a1 = 0; a0 = 0;  // default value if len==0

    switch (len)  // copy everything
    {
        case 8:  // upper 16 bits of a3 exist
            a3 = rep->s[7]<<16 & 0xFFFF0000;
        case 7:  // lower 16 bits of a3 exist
            a3 += rep->s[6];
        case 6:  // upper 16 bits of a2 exist
            a2 = rep->s[5]<<16 & 0xFFFF0000;
        case 5:  // lower 16 bits of a2 exist
            a2 += rep->s[4];
        case 4:  // upper 16 bits of a1 exist
            a1 = rep->s[3]<<16 & 0xFFFF0000;
        case 3:  // lower 16 bits of a1 exist
            a1 += rep->s[2];
        case 2:  // upper 16 bits of a0 exist
            a0 = rep->s[1]<<16 & 0xFFFF0000;
        case 1:  // lower 16 bits of a0 exist
            a0 += rep->s[0];
        default:  // nothing stored -> is same as 0
            return;
    } // switch
}


//--------------------------------------------------------------------------
// set/get the value in UTCSU format
//--------------------------------------------------------------------------

// the parameters are in the UTCSU format; that is, only the lowest
// three bits of the 'ns' are considered valid, and the fourth bit
// of the internal value is the lsb of 'us'
void INT128::SetNtp (INT32 ms, INT32 ts, INT32 us, INT8 ns)
{
    unsigned short reg[6];  // register containing the init. value
    // note: this assumes that a 'short' is at least two bytes long

    ms &= 0x0FFFFFF;  // cut off at NTP +32

    *((INT32*)reg)   = (ns & 0x07) + (us<<3);      // bits [-28:-59]
    *((INT32*)reg+1) = (us>>29 & 0x07) + (ts<<3);  // bits [+4 :-27]
    *((INT32*)reg+2) = (ts>>29 & 0x07) + (ms<<3);  // bits [+36: +5]
    // this copies bits [+36:-59], but since the ms was cut off at
    // bit +32, we have in fact the NTP bits [+31:-59] of the UTCSU
    // now, we have: reg[0] = bits [-44:-59]
    //               reg[1] = bits [-28:-43]
    //               reg[2] = bits [-12:-27]
    //               reg[3] = bits [+04:-11]
    //               reg[4] = bits [+20:+05]
    //               reg[5] = bits [+36:+21]

    // create a new IntRep and store it in the inherited member
    rep = Ialloc(rep, reg, 6, 1, 6);
    // 6 shorts are copied from reg into an array with length 6
    // note: I hope I interpreted the code correctly: the function
    // is Ialloc (oldrep, src, srclen_places, sign, newlength_places)

    *this = *this*1;  // let Integer modify the length
    // note: this is necessary because some Integer operators (comparisons
    // for example) assume that there are no unused array entries (that is,
    // no leading entries which are 0)
}

// reads the value into the ms, ts, us and ns parameters; cuts off
// at NTP bit +32, but sets 'overflow' to 1 if the internal number
// is larger (if it is smaller, 'overflow' is set to 0)
void INT128::GetNtp (INT32& ms, INT32& ts, INT32& us, INT8& ns, INT8& overflow)
{
    INT8 len = rep->len;  // for the switch below

    ms = 0; ts = 0; us = 0; ns = 0;  // initialize the values
    overflow = 0;  // assume there are no overflows

    if (len > 6)
    {
        len = 6;       // all higher numbers are ignored
        overflow = 1;  // but the overflow is reported
    }


    // convert the internal representation to the UTCSU format
    // note: the following switch does not contains any breaks; this
    // is intended (if one case is true, all statements below should
    // be executed as well)
    switch (len)  // copy everything
    {
        case 6:  // "middle" 16 bits of ms exist

            ms = rep->s[5];  // get ms bits 13..28
            ms <<= 13;       // shift everything to correct place
            if (ms & 0xFF000000) overflow = 1;  // oops, overflow
            ms &= 0x0FFFFFF; // cut off at NTP +32

        case 5:  // upper 3 bits of ts exist, lower 13 bits of ms

            ms += rep->s[4] >> 3 & 0x1FFF; // get ms bits 0..12
            ts = rep->s[4] & 0x07;         // get ts bits 29..31
            ts <<= 16;
            // shift ts for two bytes (the next case will shift the rest)

        case 4:  // "middle" 16 bits of ts exist

            ts += rep->s[3];  // get ts bits 13..28
            ts <<= 13;        // shift everything to correct place

        case 3:  // upper 3 bits of us exist, lower 13 bits of ts

            ts += rep->s[2] >> 3 & 0x1FFF;  // get ts bits 0..12
            us = rep->s[2] & 0x07;          // get us bits 29..31
            us <<= 16;  // shift us for two bytes

        case 2:  // "middle" 16 bits of us exist

            us += rep->s[1];  // get us bits 13..28
            us <<= 13;       // shift everything to correct place

        case 1:  // ns exists, and lower 13 bits of the us

            us += rep->s[0] >> 3 & 0x1FFF;  // get us bits 0..12
            ns = rep->s[0] & 0x07;         // get ns bits 0..3

        default:  // nothing stored -> interpreted as zero
            return;
    } // switch
}


//--------------------------------------------------------------------------
// necessary operators
//--------------------------------------------------------------------------

// take care: the Integer::operator~ () is faulty! it inverts and
// reverts(!) the highest two byte
// note: this operator~ inverts in 16-bit steps; so if the number stored
// is 0x1FFFF, then ~0x2FFFF = 0xFFFD0000 (this is a difference to the
// Integer::operator~, where (if it were not for the error) ~0x2FFFF would
// be 0x10000 (the error makes the result 0x20000))
INT128& INT128::operator~ ()
{
    // invert everything
    int i;
    int last = rep->len-1;  // last non-zero array entry
    if (last <= 0) return *this;  // better use Assert()

    for (i=last; i>=0; i--)
    {
        rep->s[i] = ~rep->s[i];
    } // for

    return *this;
}


