//
// interupt.cpp
//
// functions and types needed for interrupt handling
//
// use SetISR() to set an interrupt vector to an ISR and to obtain the vector
// use RemoveISR() to remove an ISR from an interrupt vector
// use InvokeISR() to call the ISR for an interrupt vector
//
// created: 18.01.97
//
// last change: 24.06.97
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include "ci.h"

#include "bool.hpp"
#include "error.hpp"
#include "interupt.hpp"
#include "isrtemp.hpp"
#include "process.hpp"  // for NO_RESOURCE


//**************************************************************************
// friend of CNeedInterrupt
//**************************************************************************

// invokes the default ISR of the object
void CallISR (CNeedInterrupt *obj)
{
    obj->isr();  // call the isr of the object
}



//**************************************************************************
// class CInterruptHandler: manages the interrupt vector table
//**************************************************************************

class CInterruptHandler
{
public:

    //----------------------------------------------------------------------
    //--- management of the vector table
    //----------------------------------------------------------------------

    void Insert (int vector, CNeedInterrupt *obj, CallFunc caller);
    // stores the ISR with this vector

    EBool Find (int vector, CNeedInterrupt **obj, CallFunc *caller);
    // searches the first object and ISR that corresponds to this vector
    // if an entry is found, True is returned, else False
    // assert: obj != NULL, caller != NULL

    EBool FindNext (CNeedInterrupt **obj, CallFunc *caller);
    // searches the next object abd ISR that corresponds to the vector of
    // the last call to Find()
    // if an entry is found, True is returned, else False
    // assert: obj != NULL, caller != NULL

    void Remove (int vector, CNeedInterrupt *obj, CallFunc caller);
    // removes all ISRs belonging to the 'obj' and 'caller', which are
    // corresponding to this vector

#ifndef NO_RESOURCE
    void ClearTable () { table_.DeleteAll(); }
    // remove all entries from the vector table

    void ClearTable (SINT32 nodeID, int area);
    // remove all entries from that particular CSA from the vector table
#endif


private:

    CVectorTable table_;  // interrupt vector table
};


static CInterruptHandler handler_;


#ifndef NO_RESOURCE
//**************************************************************************
// manual clearing of the vector table
//**************************************************************************

// only use these functions in the software controller, and only if the
// C++SIM macro NO_RESOURCE is not defined

void ClearISRs () { handler_.ClearTable(); }
// remove all ISRs from the vector table

void ClearISRs (SINT32 nodeID, int area)
  { handler_.ClearTable(nodeID, area); }
// removes all ISRs of a particular CSA from the vector table

// note: clearing the vector table manually is necessary because objects
// are not deleted (and thus the ISRs are not removed) if resource
// management is turned on; so if a node is created and later 'deleted',
// the node is only removed from the lists, but its interrupt service
// routines stay in the table until its dtor is called
// this scenario only occurs in pure simulation (using C++SIM), and only
// if the macro NO_RESOURCE is not defined; process.hpp ensures that
// NO_RESOURCE is defined if we are not in pure simulation

#endif


//**************************************************************************
// use only these functions for interrupt handling
//**************************************************************************

// stores the object with this vector in the interrupt vector table; if the
// vector is in use, the object is simply added to a list of ISRs that
// are called by InvokeISR()
// the function caller() will be invoked by the interrupt
// solely use this function to set the object (isr) for a vector!
// note: make sure that the interrupt in question is disabled when you call
// this function, even if there is already an objectISR stored
void SetISR (int vector, CNeedInterrupt *obj, CallFunc caller)
{
    handler_.Insert (vector, obj, caller);
}


// removes this particular object from the interrupt vector table
// solely use this function to remove the object (isr) for a vector!
// note: make sure that the interrupt in question is disabled when you call
// this function
void RemoveISR (int vector, CNeedInterrupt *obj, CallFunc caller)
{
    handler_.Remove (vector, obj, caller);
}


// calls the ISR of the object(s) associated with this vector
// assert: an entry must exist (i.e., at least one ISR must have been
// installed)
void InvokeISR (int vector)
{
    CallFunc caller;  // the call function
    CNeedInterrupt* obj;  // the object whose ISR should be called

#ifdef PSOS_SIM
    if (!handler_.Find (vector, &obj, &caller)) return;
    // in pSOS, it is easier to call InvokeISR() whenever an interrupt
    // occurs and let InvokeISR() sort out if there are any ISRs installed
#else
    Assert (handler_.Find (vector, &obj, &caller)); // vector must be in use
    // since there are no other interrupts, a vector is only used if there
    // should be an object ISR for it; so in this case, a missing object
    // ISR is a programming error
#endif
    do {
        Assert (obj != NULL);
        Assert (caller != NULL);
        (*caller)(obj);  // call the ISR
    } while (handler_.FindNext (&obj, &caller));
}


//**************************************************************************
// class CInterruptHandler: manages the interrupt vector table and can
//     appoint an interrupt vector to an ISR
//**************************************************************************

//--------------------------------------------------------------------------
//--- management of the vector table
//--------------------------------------------------------------------------

// stores the ISR with this vector
void CInterruptHandler::Insert (int vector, CNeedInterrupt *obj,
                                CallFunc caller)
{
    SInterrupt *info = new SInterrupt (obj, caller);
    table_.Insert (info, vector);  // store the new ISR
}

// searches the first object and ISR that corresponds to this vector
// if an entry is found, True is returned, else False
// assert: obj != NULL, caller != NULL
EBool CInterruptHandler::Find (int vector, CNeedInterrupt **obj,
                              CallFunc *caller)
{
    Assert (obj != NULL);
    Assert (caller != NULL);

    SInterrupt *info = table_.Find (vector);
    if (info == NULL) return False;
    else
    {
        *obj = info->object;
        *caller = info->caller;
        return True;
    }
}

// searches the next object abd ISR that corresponds to the vector of
// the last call to Find()
// if an entry is found, True is returned, else False
// assert: obj != NULL, caller != NULL
EBool CInterruptHandler::FindNext (CNeedInterrupt **obj, CallFunc *caller)
{
    Assert (obj != NULL);
    Assert (caller != NULL);

    SInterrupt *info = table_.FindNext ();
    if (info == NULL) return False;
    else
    {
        *obj = info->object;
        *caller = info->caller;
        return True;
    }
}

// removes all ISRs belonging to the 'obj' and 'caller', which are
// corresponding to this vector
void CInterruptHandler::Remove (int vector, CNeedInterrupt *obj,
                                CallFunc caller)
{
    if ((obj == NULL) || (caller == NULL)) return;
    // NULL entries cannot be inserted anyway

    SInterrupt *info = table_.Find (vector);
    while (info != NULL)
    {
        if ((info->object == obj) && (info->caller == caller))
        {
            table_.RemoveAtCurrent();  // found the entry
            delete info;  // delete the wrapper structure
            return;
        }
        info = table_.FindNext ();
    } // while
}


#ifndef NO_RESOURCE

// delete all entries from that particular CSA from the vector table
void CInterruptHandler::ClearTable (SINT32 nodeID, int area)
{
    int vector;

    vector = GetClockIntvec (nodeID, area);
    table_.Delete (vector | INTA);  // remove all INTA entries
    table_.Delete (vector | INTT);  // remove all INTT entries
    table_.Delete (vector | INTN);  // remove all INTN entries
    // note: INTN are not used by SimUTC, but in pure simulation
    // there is nobody else and therefore it does not hurt to delete
    // them as well

    vector = GetGpsIntvec (nodeID, area, GPU1);
    table_.Delete (vector);  // remove entries for GPU1

    vector = GetGpsIntvec (nodeID, area, GPU2);
    table_.Delete (vector);  // remove entries for GPU2

    vector = GetGpsIntvec (nodeID, area, GPU3);
    table_.Delete (vector);  // remove entries for GPU3
}

#endif


