//
// intproc.cpp
//
// implementation of an interrupt process
// note: may become obsolete, since interrupt processes are only used in
//       the software clock, so CClockInterrupt may be merged with this
//       class
//
// created: 18.01.97
//
// last change: 26.06.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <stdlib.h>  // for NULL

#include "constant.hpp"
#include "error.hpp"
#include "interupt.hpp"
#include "intproc.hpp"


//**************************************************************************
// CInterruptProcess: a simple process generating an interrupt (calling
//    InvokeISR()) when it is activated and then either scheduling itself
//    for the next interrupt or cancelling itself
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

// set the interrupt vector
CInterruptProcess::CInterruptProcess ()
{
    vector_ = NoAccount;  // set the vector to an invalid number
    waitTime_ = NoAccount;  // no automatic rescheduling in Body()
    enabled_ = False;  // by default interrupts are disabled
}


//----------------------------------------------------------------------
//--- interrupt management
//----------------------------------------------------------------------

// sets the interrupt vector; *must* be called before the first
// interrupt is generated
void CInterruptProcess::SetVector (int vector)
{
    vector_ = vector;  // store the vector of the interrupt
}

// enable the interrupt; if it is not enabled at the time the interrupt
// should be generated, the interrupt service routine will not be called
// assert: interrupt vector has been set
void CInterruptProcess::Enable ()
{
    Assert (vector_ != NoAccount);

    enabled_ = True;
}

// disable the interrupt (till the next call to Enable())
void CInterruptProcess::Disable ()
{
    enabled_ = False;
}

// returns True if the interrupt is enabled, False if not
EBool CInterruptProcess::Enabled ()
{
    return enabled_;
}

// returns True if the process is active, False if not
EBool CInterruptProcess::Active ()
{
    if (ACProcess::current() == this) return True;  // process is active
    else return False;
}


//--------------------------------------------------------------------------
//--- create an interrupt
//--------------------------------------------------------------------------

// generate an interrupt in 'delay' seconds from now
void CInterruptProcess::RaiseAfter (double delay)
{
    // note: we have to differentiate whether rescheduling takes place in
    // the context of the process or not; if not --> have to reactivate it;
    // if yes: let it schedule itself in Body() using waitTime_
    
    if (Active())  // in context of process
    {
        waitTime_ = delay;  // turn on auto-scheduling in Body()
    }
    else  // normal activation
    {
        waitTime_ = NoAccount;  // turn off auto-scheduling in Body()
        ReActivateDelay (delay);  // schedule the process
    }
}


//--------------------------------------------------------------------------
//--- deschedule an interrupt
//--------------------------------------------------------------------------

// deschedules the interrupt (if it is scheduled) and sets waitTime to
// NoAccount
void CInterruptProcess::CancelInterrupt ()
{
    waitTime_ = NoAccount;  // turn off auto-scheduling in Body()
    if (!Active() && !idle()) Cancel();  // deschedule process
}


//--------------------------------------------------------------------------
//--- for the simulation scheduler only; do not use
//--------------------------------------------------------------------------

// call the ISR (if enabled), maybe schedule for the next interrupt
void CInterruptProcess::Body ()
{
    for (;;)
    {
        PreExecution ();  // execute interrupt specific things
        if (enabled_) InvokeISR (vector_); // call the ISR (only if enabled)
        PostExecution ();  // again do interrupt specific things

        if (waitTime_ >= 0)  // there is another interrupt to generate
        {
             Hold (waitTime_);  // reschedule
             waitTime_ = NoAccount;  // turn off further interrupts
        }
        else Cancel();  // suspend the process
    }
}


//--------------------------------------------------------------------------
//--- interrupt specific work
//--------------------------------------------------------------------------

// is executed prior to calling the ISR; a derived class can overload
// the function to implement some interrupt specific things
void CInterruptProcess::PreExecution ()
{
}

// is executed after calling the ISR
void CInterruptProcess::PostExecution ()
{
}


