//
// lance.cpp
//
// implementation of the Lance interface
//
// created: 21.10.96
//
// last change: 22.01.99
//
// RCS version control information:
//     $Revision: 1.10 $$State: Exp $$Date: 1999/01/22 12:15:13 $
//


#include <iostream.h>  // DEBUG

#include <stdlib.h>  // for the NULL macro
#include <string.h>  // for memcpy etc.

#include "ci.hpp"
#include "clockifc.hpp"
#include "clocksyn.hpp"
#include "distrib.hpp"
#include "dmsgtemp.hpp"
#include "error.hpp"
#include "general.hpp"
#include "lance.hpp"
#include "msgform.hpp"


static const addr BroadcastAddress = BROADCAST_ADDRESS;
// needed for setting the default address to broadcast

static const int AddedBufferLength = sizeof(INT32);
// length of the additional buffer that the Lance allocates


//**************************************************************************
// module global functions
//**************************************************************************

static bufx *BufToBufx (char *message);
// assumes that 'message' is the result of a call to GetBuffer(), and that
// the real bufx structure has 'AddedBufferLength' more bytes in front of
// 'message'; returns a pointer to this bufx structure

static char *BufxToBuf (bufx *buf);
// pendant to the above functions; returns a pointer to the address of the
// byte with number 'AddedBufferLength'


static void ExtractMessage (bufx *buf, INT16 size, SINT32 nodeID, int if_num,
                            STimeData *time, char **message, INT16 *msgsize);
// extracts the original message: the sender data (the macrostamp,
// timestamp, accuracy and area of the sender) is put into 'time', the
// rest is put into 'message' (an array of char is allocated accordingly)
// and the size of the 'message' buffer is stored in 'msgsize'

static int CI_GetBuffer (INT16 size, bufx **buf, int if_num, int area);
static int CI_GetEnvelope (int mode, tenvx **tenv, int if_num, int area);
static int CI_RetBuffer (bufx *buf);
static int CI_RetEnvelope (tenvx *tenv);
static int CI_Send (int mode, tenvx *tenv, bufx *buf, INT16 size,
                    int if_num, int area);
// just for easier access to the CI functions
// context: ISR


//**************************************************************************
// class CLance: this class is the interface between the simulation and
//    the CI functions; it can scramble and/or delay messages
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

CLance::CLance (CControllingObject *ctrl, int area)
    : CApplControlledObject (ctrl)
{
    clockSync_ = NULL;
    clock_ = NULL;
    area_ = area;

    INT32 old = DisableInterrupts();
    SetISR (GetClockIntvec (GetNodeID(), area_) | INTT, this); 
    // clock interrupt INTT (duty timer B on SSU5) calls the default ISR
    EnableInterrupts (old);
    
    CLance::Init();
}

CLance::~CLance()
{
    if (clock_ != NULL)  // disable the DT B5
    {
        clock_->DisableInterrupt (IntDutyB5);
        clock_->DisableDutyTimer (IntDutyB5);  // disable timer itself
    }

    INT32 old = DisableInterrupts();
    RemoveISR (GetClockIntvec (GetNodeID(), area_) | INTT, this);
    EnableInterrupts (old);

    Dispose();
}


//--------------------------------------------------------------------------
//--- reset all members to their initial values
//--------------------------------------------------------------------------

// restores the initial state
void CLance::Reset ()
{
    CCrashingModule::Reset();  // reset crash state
    CApplControlledObject::Reset();  // reset other things
}


//--------------------------------------------------------------------------
//--- control the system parameters
//--------------------------------------------------------------------------

// sets the requested parameter value
// existing params: PCrash, PRecover, DmaxSend, DistributionSend,
// DmaxRecv, DistributionRecv,
// and all params of CApplControlledObject
EBool CLance::SetParameter (const SParameter& param)
{
    SSingleParameter *sparam = (SSingleParameter *)&param;  // easier access
    SApplDistribution *dist = (SApplDistribution *)&param;

    switch (param.type)
    {
        case PCrash:
            return CCrashingModule::SetProbability (True, sparam->value);
            break;
        case PRecover:
            return CCrashingModule::SetProbability (False, sparam->value);
            break;
        case DistributionSend:
            if (!dist->data.DataValid()) return False;
            distDataSend_ = dist->data;
            break;
        case DistributionRecv:
            if (!dist->data.DataValid()) return False;
            distDataRecv_ = dist->data;
            break;
        case PScrambleSend:
            if ((sparam->value < 0) || (sparam->value > 1)) return False;
            pSendScramble_ = sparam->value;
            break;
        case PScrambleRecv:
            if ((sparam->value < 0) || (sparam->value > 1)) return False;
            pRecvScramble_ = sparam->value;
            break;
        case PDuplicateSend:
            if ((sparam->value < 0) || (sparam->value > 1)) return False;
            pSendDuplicate_ = sparam->value;
            break;
        case PDuplicateRecv:
            if ((sparam->value < 0) || (sparam->value > 1)) return False;
            pRecvDuplicate_ = sparam->value;
            break;

        default:  // pass on
            return CApplControlledObject::SetParameter (param);
    } // switch

    return True;
}

// returns the requested parameter value
// additional parameter: CrashStatus
EBool CLance::GetParameter (SParameter& param) const
{
    SSingleParameter *sparam = (SSingleParameter *)&param;  // easier access
    SApplDistribution *dist = (SApplDistribution *)&param;

    switch (param.type)
    {
        case CrashStatus:
            sparam->value = Crashed();
            break;
        case PCrash:
            sparam->value = CCrashingModule::GetProbability (True);
            break;
        case PRecover:
            sparam->value = CCrashingModule::GetProbability (False);
            break;
        case DistributionSend:
            dist->data = distDataSend_;
            break;
        case DistributionRecv:
            dist->data = distDataRecv_;
            break;
        case PScrambleSend:
            sparam->value = pSendScramble_;
            break;
        case PScrambleRecv:
            sparam->value = pRecvScramble_;
            break;
        case PDuplicateSend:
            sparam->value = pSendDuplicate_;
            break;
        case PDuplicateRecv:
            sparam->value = pRecvDuplicate_;
            break;


        default:  // pass on
            return CApplControlledObject::GetParameter (param);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False;
// existing commands: all of CApplControlledObject and CCrashingModule
EBool CLance::ExecuteCommand (const SCommand& command)
{
    switch (command.type)
    {
        default:
            if (CApplControlledObject::ExecuteCommand (command)) return True;
            else return CCrashingModule::ExecuteCommand (command);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- suspend the Lance
//--------------------------------------------------------------------------

// go into suspension mode; in that mode, all messages that are currently
// hold back will be discarded, and message announces from the network will
// be ignored; the Lance will stay in suspension mode until ClockSync
// initiates a send request
void CLance::Suspend()
{
    suspended_ = True;
    DeleteDelayedMessages();  // delete all delayed messages
}


//--------------------------------------------------------------------------
//--- crash the Lance
//--------------------------------------------------------------------------

// crash the module; that includes throwing away all messages that are
// currently in the list of delays
void CLance::Crash()
{
    CCrashingModule::Crash();  // crash the module
    DeleteDelayedMessages();  // delete all delayed messages
}


//--------------------------------------------------------------------------
//--- set associated objects
//--------------------------------------------------------------------------

void CLance::SetClockSync (CClockSync *cs)
{
    clockSync_ = cs;
}

void CLance::SetClock (CClockInterface *clock)
{
    if (clock_ != NULL)  // disable DT B5 on old clock
    {
        clock_->DisableInterrupt (IntDutyB5);
        clock_->DisableDutyTimer (IntDutyB5);  // disable timer itself
    }

    clock_ = clock;  // set to the new clock
/**/// note: it would be correct to set the vector of the ISR to the
    // nodeID/area used by this clock, instead of the own nodeID/area

    if (clock_ != NULL)
    {
        clock_->EnableInterrupt (IntDutyB5);  // enable DT B5 on new clock
        if (msg_!=NULL) clock_->InterruptAt (IntDutyB5, msg_->deliveryTime);
        // schedule the next delayed message
    }
/**/// note: should discard all delayed messages if clock == NULL? Or
    // wait for a new clock (currently done)?
}


//--------------------------------------------------------------------------
//--- request message buffer/timestamped envelope
//--------------------------------------------------------------------------

// returns a pointer to a data buffer; allocates four bytes more (to
// contain the message ID), but returns a pointer to the address of byte
// four in this buffer
// note: the buffer resides in local memory; this is *no* CI buffer
// (the data is copied into a CI buffer upon send)
// assert: message != NULL
void CLance::GetBuffer (INT16 msgsize, char **message, INT32& msgID)
{
    Assert (message != NULL);  // pointer must exist
    Assert (msgsize+AddedBufferLength > 0);  // we have to alloc something
/**/// note: maybe set *message=NULL if size+AddedBufLen==0 ?

    bufx *buf = (bufx*)new char[msgsize+AddedBufferLength];
    // allocate more bytes for the msgID
    *message = BufxToBuf (buf);  // set 'message' behind internal part

    memcpy (buf, &msgID_, sizeof(msgID_));  // store the message ID
    msgID = msgID_;  // return it to the caller as well
    msgID_++;  // increment to next free ID
}

// returns a pointer to a timestamped envelope
// note: again, this is no CI envelope, but resides in local memory
// (the data is copied into a CI envelope upon send)
// assert: tenv != NULL
void CLance::GetEnvelope (tenvx **tenv)
{
    *tenv = new tenvx;
    Assert (*tenv != NULL);
}

// frees the pointer to a data buffer; assumes that the buffer was
// allocated using GetBuffer() and that it starts at an address four
// bytes before 'message'
// buffers delivered to ClockSync are of that type
void CLance::ReturnBuffer (char *message)
{
    Assert (message != NULL);
    delete[] (char *)BufToBufx (message);
}

// frees the pointer to a timestamped envelope; assumes that the
// envelope was allocated using GetEnvelope()
// envelopes delivered to ClockSync are of that type
void CLance::ReturnEnvelope (tenvx *tenv)
{
    Assert (tenv != NULL);
    delete tenv;
}

// returns a pointer to a data buffer; this function just allocates the
// amount of memory requested; you should call this function to obtain
// a local receive buffer - Announce() assumes that its 'buf' parameter
// was obtained via this function
// assert: rbuf != NULL
void CLance::GetReceiveBuffer (INT16 size, bufx **rbuf)
{
    Assert (rbuf != NULL);
    *rbuf = (bufx*)new char[size];
}

// just allocates the structure
// assert: renv != NULL
void CLance::GetReceiveEnvelope (renvx **renv)
{
    Assert (renv != NULL);
    *renv = new renvx;
}

// use this function to return buffers allocated with GetReceiveBuffer()
// assert: rbuf != NULL
void CLance::RetReceiveBuffer (bufx *rbuf)
{
    Assert (rbuf != NULL);
    delete[] rbuf;
}

// use this function to return mem allocated with GetReceiveEnvelope()
// assert: renv != NULL
void CLance::RetReceiveEnvelope (renvx *renv)
{
    Assert (renv != NULL);
    delete renv;
}


//--------------------------------------------------------------------------
//--- send/receive messages
//--------------------------------------------------------------------------

// delays the message
// suspended is set to False, the crash status may change; if the module is
// crashed, the message (and all currently in the delay list) is discarded;
// the function assumes that 'tenv' was obtained via the GetEnvelope()
// function, and 'message' via the GetBuffer() function
// 'tenv' may be NULL
void CLance::Send (int mode, tenvx *tenv, char *message, INT16 msgsize,
                   int if_num, int area)
{
    suspended_ = False;
    // a send request is the signal to resume operation

    ModifyCrashStatus();
    // this may delete the messages in the delay list

    if (Crashed())  // return the buffers at once
    {
        ReportMessageEvent (DiscardSendMessage, NoAccount);
/**/    // msgID=NoAccount - this is just for now; should obtain the
        // msgID from 'message'

        if (tenv != NULL) ReturnEnvelope (tenv);
        if (message != NULL) ReturnBuffer (message);
        return;
    }

    // check if the message should be duplicated
    if (random_->CastDice (pSendDuplicate_))  // yes, duplicate it
    {
        tenvx *dupTenv;    // duplicate of the tenv
        char *dupMessage;  // duplicate of the message
        INT32 msgID;       // message ID of the duplicate (not needed)

        GetBuffer (msgsize, &dupMessage, msgID);
        memcpy (dupMessage, message, msgsize);

        if (tenv != NULL)  // only copy tenv if it exists
        {
            GetEnvelope (&dupTenv);
            memcpy (dupTenv, tenv, sizeof (tenvx));
        }
        else dupTenv = NULL;

        DelayMessage (mode, dupTenv, dupMessage, msgsize, if_num, area);
        // delay the duplicate

        ReportMessageEvent (DuplicateSendMessage, msgID);
        // report the duplication (the ID is that of the duplicated msg,
        // not that of the original message!)
    } // if

    DelayMessage (mode, tenv, message, msgsize, if_num, area);
}


// delivers the message to the network, then deletes tenv and message;
// tenv may be NULL
// the function allocates CI buffers, copies the parameters into them
// and calls ci_send()
// the function returns a CI error code if one of the CI functions
// fails, or 0 if the message was sent successfully
// the Lance module may crash and msges (including the destination address)
// may be scrambled in this function
// note: *regardless* of the 'mode', tenv and message are always returned
//   (so ClockSync should never use a tenv and message pointer after it
//   has called Send())
// note: the function may call DeleteDelayedMessages(), so take care in the
//   calling function to set msg_ to the *next* message (don't forget to
//   delete the Mcb structure of the current msg; the data itself is
//   deleted in this function); otherwise, the data of the current msg will
//   be deleted twice
// note: the function contains a goto statement; I believe that it is the
//   easiest way to conditionally return the parameter memory from several
//   locations in the function before returning, so don't try to talk me
//   out of it
// assert: message != NULL
int CLance::SendNow (int mode, tenvx *tenv, char *message, INT16 msgsize,
                     int if_num, int area)
{
    int ci_error = 0;  // for checking the error code of the CI functions
    INT32 msgID;
    bufx *cibuf;  // the CI buffer
    INT16 buflen;  // length of the CI buffer
    tenvx *citenv;  // the CI envelope


    Assert (message != NULL);  // need the sender data

    memcpy (&msgID, BufToBufx(message), sizeof (msgID_));  // get the msg ID

    if (suspended_) Report (Suspended in SendNow());
/**/// note: should never be suspended at this point
    suspended_ = False;
    // a send request is the signal to resume operation

    ModifyCrashStatus();
    // this may delete the messages in the delay list, so take care in the
    // calling function to set msg_ to the *next* message (don't forget to
    // delete the Mcb structure of the current msg; the data itself is
    // deleted in this function)

    if (Crashed())
    // note: if the lance is crashed, the msg will not be sent; nevertheless
    // the SendNow() function will return 0 (success), since crashes are part
    // of the fault injection
    {
        ReportMessageEvent (DiscardSendMessage, msgID);

        ci_error = 0;  // return value of the function
        goto DeleteAndReturn;
        // delete 'message' and 'tenv', then return ci_error
    }


    if (tenv == NULL)  // obtain a local tenv structure
    {
        GetEnvelope (&tenv);
        memcpy (tenv->da, BroadcastAddress, sizeof(addr));
        // default destination is broadcast message
    }
    // this is done because the address may be scrambled and an access
    // to CI memory is quite costly in time; so the address is scrambled
    // locally and afterwards copied into the CI envelope


    ScrambleMessage (tenv->da, message, msgsize, msgID);
    // here the msg and receiver address might get scrambled!


    // if there's no crash, the message is sent and CI memory is requested

    buflen = msgsize+AddedBufferLength;  // compute the real buffer length

    // allocate the CI buffer
    ci_error = CI_GetBuffer (buflen, &cibuf, if_num, area);
    if (ci_error != 0) goto DeleteAndReturn;  // error occured, return at once

    // allocate the CI envelope
    ci_error = CI_GetEnvelope (mode, &citenv, if_num, area);
    if (ci_error != 0)  // error has occured, return at once
    {
        CI_RetBuffer (cibuf);  // return the already allocated buffer
        goto DeleteAndReturn;
    }

    // CI structures were allocated, now copy the data

    memcpy (cibuf, BufToBufx(message), buflen);  // copy the message
    memcpy (citenv->da, tenv->da, sizeof(addr));  // copy the envelope


    // now send the message
    ci_error=CI_Send (mode|RET_PKB_IMM, citenv, cibuf, buflen, if_num, area);
    // always return the CI buffers at once

    if (ci_error != 0)  // CI didn't return the buffers -> have to return them
    {
        CI_RetBuffer (cibuf);
        CI_RetEnvelope (citenv);
    }


DeleteAndReturn:

    // here, the parameters 'tenv' and 'message' are returned
    // afterwards, the value of ci_error is returned, so don't forget to assign
    // the return value of the function to ci_error before jumping to this label

    if (tenv != NULL) ReturnEnvelope (tenv);
    if (message != NULL) ReturnBuffer (message);

    return ci_error;
}


// receives a message from the network, and delays it before it is
// delivered to ClockSync;
// the function assumes that the 'renv' and 'buf' data reside in
// local memory and must be disposed via the RetReceive...() functions
void CLance::Announce (renvx *renv, bufx *buf, INT16 size, int if_num)
{
    ModifyCrashStatus();
    // this may delete the messages in the delay list

    if (Crashed() || suspended_)  // return the buffers at once
    {
        ReportMessageEvent (DiscardRecvMessage, NoAccount,
                            NoAccount, NoAccount);
/**/    // most data is NoAccount - this is just for now; should extract
        // the data from 'message', store the extracted data

        if (renv != NULL) RetReceiveEnvelope (renv);
        if (buf != NULL) RetReceiveBuffer (buf);
        return;
    }

    // check if the message should be duplicated
    if (random_->CastDice (pRecvDuplicate_))  // yes, duplicate it
    {
        renvx *dupRenv;    // duplicate of the renv
        bufx *dupBuf;      // duplicate of the buffer

        GetReceiveBuffer (size, &dupBuf);
        GetReceiveEnvelope (&dupRenv);

        memcpy (dupRenv, renv, sizeof (renvx));
        memcpy (dupBuf, buf, size);

        DelayMessage (dupRenv, dupBuf, size, if_num);
        // delay the duplicate

        // report the duplication
        ReportMessageEvent (DuplicateRecvMessage, NoAccount,
                            NoAccount, NoAccount);
    } // if

    DelayMessage (renv, buf, size, if_num);
}


// delivers the message to ClockSync, then deletes the pointers
void CLance::AnnounceNow (renvx *renv, bufx *buf, INT16 size, int if_num)
{
    INT32 msgID;
    SSendData *data;
    char *message;  // the original message
    INT16 msgsize;
    STimeData sendTime;  // data added by the LANCE

    if (clockSync_ == NULL)  // no recipient
    {
        ReportMessageEvent (DiscardRecvMessage, NoAccount, NoAccount,
                            NoAccount);
        // most data is NoAccount, so we don't waste time extracting the
        // msg just for getting the msgID
/**/    // this may change in the future (put this if-statement below
        // the extraction)

        RetReceiveEnvelope (renv);
        RetReceiveBuffer (buf);
        return;
    }

    ExtractMessage (buf, size, GetNodeID(), if_num, &sendTime, &message,
                    &msgsize);
    // get the sender time/area and the original message

    Assert (message != NULL);  // need the sender data

    memcpy (&msgID, message, sizeof (msgID_));  // get the message ID
    data = (SSendData*)(message+AddedBufferLength);
    // first entry in the user-buffer is the sender data
/**/// note: here, the Lance must know the position of data that is
    // solely managed in ClockSync; this is not very flexible!

    ModifyCrashStatus();

    if (suspended_ | Crashed())  // do not accept any messages
    {
        ReportMessageEvent (DiscardRecvMessage, data->nodeID,
                            sendTime.area, msgID);
        delete[] message;
        RetReceiveEnvelope (renv);
        RetReceiveBuffer (buf);
        return;
    }


    ScrambleMessage (renv->sa, &sendTime,
                     message+AddedBufferLength, msgsize-AddedBufferLength,
                     data->nodeID, sendTime.area, msgID);
    // here the msg might get scrambled!


    clockSync_->Announce (renv, sendTime, message+AddedBufferLength,
                          msgsize-AddedBufferLength, if_num, msgID);
    // deliver to ClockSync

    // delete the pointers
    delete[] message;
    RetReceiveEnvelope (renv);
    RetReceiveBuffer (buf);
}


//--------------------------------------------------------------------------
//--- initialization and destruction of the data members
//--------------------------------------------------------------------------

// Lance is suspended
void CLance::Init ()
{
    random_ = new CRandom (GetNodeID());  // init. random stream with nodeID
    listOfDelayedMessages_ = new CDelayList;
    msg_ = NULL;
    msgID_ = 0;
    distDataSend_.dist = Uniform;
    distDataSend_.low = 0;
    distDataSend_.high = 0;
    distDataSend_.mean = 0;
    distDataSend_.constant = 0;
    distDataSend_.freeparam = 0;

    distDataRecv_ = distDataSend_;

    pSendScramble_ = 0;
    pRecvScramble_ = 0;
    pSendDuplicate_ = 0;
    pRecvDuplicate_ = 0;

    Suspend();  // default after creation and reset is suspended mode
    // take care: never call Suspend() before initializing msg_!
}

void CLance::Dispose()
{
    delete random_;
    DeleteDelayedMessages();  // delete all delayed messages
    delete listOfDelayedMessages_;  // delete the empty list
}


//--------------------------------------------------------------------------
//--- delay messages
//--------------------------------------------------------------------------

// delays the outgoing message according to the delay data
// assert: clock != NULL
void CLance::DelayMessage (int mode, tenvx *tenv, char *message, INT16 msgsize,
                           int if_num, int area)
{
    // get the delay
    double delay = random_->RandomVar (distDataSend_);

    Assert (clock_ != NULL);

    STimestamp time;
    clock_->GetTimestamp (time);  // get the current clock time

    SSendDelayMcb *mcb = new SSendDelayMcb;

    mcb->send = True;  // the message is outgoing
    mcb->deliveryTime = NtpToSeconds(time) + delay;  // compute delivery time
    mcb->mode = mode;
    mcb->tenv = tenv;
    mcb->msg = message;
    mcb->msgsize = msgsize;
    mcb->if_num = if_num;
    mcb->area = area;

    Insert (mcb);  // store in the list of delayed messages
}

// delays the incoming message according to the delay data
// assert: clock != NULL
void CLance::DelayMessage (renvx *renv, bufx *buf, INT16 size, int if_num)
{
    // get the delay
    double delay = random_->RandomVar (distDataRecv_);

    Assert (clock_ != NULL);

    STimestamp time;
    clock_->GetTimestamp (time);  // get the current clock time

    SRecvDelayMcb *mcb = new SRecvDelayMcb;

    mcb->send = False;  // the message is incoming
    mcb->deliveryTime = NtpToSeconds(time) + delay;  // compute delivery time
    mcb->renv = renv;
    mcb->buf = buf;
    mcb->size = size;
    mcb->if_num = if_num;

    Insert (mcb);  // store in the list of delayed messages
}


//--------------------------------------------------------------------------
//--- scramble messages
//--------------------------------------------------------------------------

// scramble incoming messages, notify evaluation
void CLance::ScrambleMessage (addr address,char *msg,INT16 size,INT32 msgID)
{
    if (!random_->CastDice (pSendScramble_)) return;
    // check if the message should be scrambled

    ReportMessageEvent (ScrambleSendMessage, msgID);

    Scramble (address, sizeof(addr));
    Scramble (msg, size);
}

// scramble outgoing messages, notify evaluation
void CLance::ScrambleMessage (addr address, STimeData *time, char *msg,
    INT16 size, INT32 nodeID, int area, INT32 msgID)
{
    if (!random_->CastDice (pRecvScramble_)) return;
    // check if the message should be scrambled

    ReportMessageEvent (ScrambleRecvMessage, nodeID, area, msgID);

    Scramble (address, sizeof(addr));
    Scramble (time, sizeof(*time));
    Scramble (msg, size);
}

// just scramble the data, no notification
void CLance::Scramble (void *data, INT16 size)
{
    int i;
    char *ptr = (char *)data;
    // pointer to the next element that is to be scrambled

    // get the scramble value
    INT32 scramble = (INT32)random_->RandomVar (0, 0xFFFFFFFF);

    // XOR the value to the data
    for (i=0; i < size/4; i++)
    {
        ((INT32*)ptr)[i] ^= scramble;
        // use INT32 because that access is faster than INT8 (or char) access
    } // for
    // now scramble the last 0-3 bytes
    for (i=size - size%4; i < size; i++)
    {
        ptr[i] ^= (INT8)scramble;
    } // for
}


//--------------------------------------------------------------------------
//--- fault injection
//--------------------------------------------------------------------------

// create an event of type 'CrashStatusChange', send it to the evaluation
void CLance::ReportCrashChange ()
{
    SApplCrashEvent *event = new SApplCrashEvent;
    event->module = Lance;
    event->type = CrashStatusChange;
    event->crashed = Crashed();
    CControlledObject::ReportEvent (event);
}


//--------------------------------------------------------------------------
//--- event reports
//--------------------------------------------------------------------------

// create a message event of given type, fill in the data and send it
// to the evaluation; use this function for incoming messages
void CLance::ReportMessageEvent (EEvent type, SINT32 nodeID, int area,
                                 INT32 msgID)
{
    SApplRecvMsgEvent *event = new SApplRecvMsgEvent;
    event->module = Lance;
    event->type = type;
    event->sender = nodeID;
    event->senderArea = area;
    event->receiver = GetNodeID();
    event->msgID = msgID;
    CControlledObject::ReportEvent (event);
}

// create a message event of given type, fill in the data and send it
// to the evaluation; use this function for outgoing messages
void CLance::ReportMessageEvent (EEvent type, INT32 msgID)
{
    SApplSendMsgEvent *event = new SApplSendMsgEvent;
    event->module = Lance;
    event->type = type;
    event->receiver = NoAccount;
    event->msgID = msgID;
    CControlledObject::ReportEvent (event);
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int CLance::ReportError (EReportType type, char *file, int line,
                         char *expression)
{
    return ::ReportError (type, file, line, expression, Lance,
                          GetNodeID(), area_);
}


//--------------------------------------------------------------------------
//--- ISR
//--------------------------------------------------------------------------

// send/deliver all messages due at the current time (called by DT B5)
void CLance::isr()
{
    static const double Epsilon = NtpToSeconds (0,1<<8);  // duty timer tick
    // this is needed for pure simulation, because the delays of the
    // messages do not consider the duty timer granularity; this causes
    // an endless loop if the delivery time is rounded up in the duty
    // timer, so we need to include the duty timer granularity in the
    // check of the delivery time
    // it is not necessary in hardware-based simulation, but does not hurt

    Assert (clock_ != NULL);

    if (!clock_->InterruptPending (IntDutyB5)) return;  // not supported
    clock_->ClearInterrupt (IntDutyB5);  // clear the DT bit

    ReportEvent (Lance, TimerEvent);

    STimestamp time;
    clock_->GetTimestamp (time);  // get the current time
    double current = NtpToSeconds (time);  // current time in seconds

    Assert (msg_ != NULL);  // a delayed message must exist

    SSendDelayMcb *send;  // for easier access in the if-statement below
    SRecvDelayMcb *recv;

    while ((msg_ != NULL) && (msg_->deliveryTime <= (current+Epsilon)))
    {
        // process the msg
        if (msg_->send)
        {
            send = (SSendDelayMcb *)msg_;
            msg_ = listOfDelayedMessages_->RemoveFirst();// remove next msg
            SendNow (send->mode, send->tenv, send->msg, send->msgsize,
                     send->if_num, send->area);
            delete send;  // free the memory of the Mcb
        }
        else
        {
            recv = (SRecvDelayMcb *)msg_;
            msg_ = listOfDelayedMessages_->RemoveFirst();// remove next msg
            AnnounceNow (recv->renv, recv->buf, recv->size, recv->if_num);
            delete recv;  // free the memory of the Mcb
        }
    } // while

    if (msg_ != NULL) clock_->InterruptAt (IntDutyB5, msg_->deliveryTime);
    else clock_->DisableDutyTimer (IntDutyB5);
    // set DT B5 to the next delivery time or turn it off
}


//--------------------------------------------------------------------------
//--- management of the message delay list
//--------------------------------------------------------------------------

// inserts the 'mcb' into the delay list, possibly sets DT B5
void CLance::Insert (SDelayMcb *mcb)
{
    Assert (mcb != NULL);  // only valid messages are stored
    Assert (clock_ != NULL);  // can't delay without a clock

    if (msg_ == NULL)  // set DT B5 to delivery time of current message
    {
        msg_ = mcb;  // store the new msg as the current one
        clock_->InterruptAt (IntDutyB5, msg_->deliveryTime); // schedule msg
    }
    else  // there may be other messages that are earlier
    {
        if (msg_->deliveryTime <= mcb->deliveryTime)
        // the message is not earlier than the scheduled one
        {
            listOfDelayedMessages_->Insert (mcb, mcb->deliveryTime);
            // store the message in the message queue
        }
        else // the new msg is earlier than the scheduled one
        {
            listOfDelayedMessages_->Insert (msg_, msg_->deliveryTime);
            // store the current message in the message queue

            msg_ = mcb;
            clock_->InterruptAt (IntDutyB5, msg_->deliveryTime);
            // schedule new msg
        }
    } // else
}


// deletes all entries in the list, deletes member 'msg' (sets it to NULL)
// if 'msg' is NULL, the function returns at once
void CLance::DeleteDelayedMessages ()
{
    if (msg_ == NULL) return;

    clock_->DisableDutyTimer (IntDutyB5);
    // disable the duty timer for the msg

    while (msg_ != NULL)  // there is still a message
    {
        if (msg_->send)
        {
            if (((SSendDelayMcb*)msg_)->msg != NULL)
                ReturnBuffer (((SSendDelayMcb*)msg_)->msg);
            if (((SSendDelayMcb*)msg_)->tenv != NULL)
                ReturnEnvelope (((SSendDelayMcb*)msg_)->tenv);
        }
        else
        {
            if (((SRecvDelayMcb*)msg_)->buf != NULL)
                RetReceiveBuffer (((SRecvDelayMcb*)msg_)->buf);
            if (((SRecvDelayMcb*)msg_)->renv != NULL)
                RetReceiveEnvelope (((SRecvDelayMcb*)msg_)->renv);
        }
        delete msg_;
        msg_ = listOfDelayedMessages_->RemoveFirst();
        // remove the first message from the list
    } // while

    msg_ = NULL;
}



//**************************************************************************
// module global functions
//**************************************************************************

//--------------------------------------------------------------------------
//--- CS buffer from/to La buffer conversion
//--------------------------------------------------------------------------

// assumes that 'message' is the result of a call to GetBuffer(), and that
// the real bufx structure has 'AddedBufferLength' more bytes in front of
// 'message'; returns a pointer to this bufx structure
static bufx *BufToBufx (char *message)
{
    if (message==NULL) return NULL;
    else return (bufx *)(message-AddedBufferLength);
}

// pendant to the above functions; returns a pointer to the address of the
// byte with number 'AddedBufferLength'
static char *BufxToBuf (bufx *buf)
{
    if (buf==NULL) return NULL;
    else return (char *)buf + AddedBufferLength;
}


//--------------------------------------------------------------------------
//--- message extraction
//--------------------------------------------------------------------------

// extracts the original message: the sender data (the macrostamp,
// timestamp, accuracy and area of the sender) is put into 'time', the
// rest is put into 'message' (an array of char is allocated accordingly)
// and the size of the 'message' buffer is stored in 'msgsize'
static void ExtractMessage (bufx *buf, INT16 size, SINT32 nodeID,
    int if_num, STimeData *time, char **message, INT16 *msgsize)
{
    SMsgPosition pos;

    Assert (buf != NULL);  // check the arguments
    Assert (time != NULL);
    Assert (message != NULL);
    Assert (msgsize != NULL);

    GetPosition (pos, nodeID, if_num);  // get the message position data

    int i;
    INT16 netsize=0;  // the size of all sender data

    for (i=0; i<SMsgPosition::EItemCount; i++)
    {
        Assert (pos.position[i] < size);
        // the position of the sender data  must be in the buffer

        Assert (pos.length[i] <= sizeof(((INT32*)time)[i]));
        // the member of 'time' must be able to hold the sender data

        netsize+=pos.length[i];  // add the size to the sum
    }

    Assert (netsize <= size);  // all sender data must be in buffer


/**/// note: the following code is preliminary; hopefully, a more
    // readable implementation will suggest itself

    char *ptr;  // pointer to the position of an element
    INT8 numberOfEntries = SMsgPosition::EItemCount;
    int j;

    // go through all entries of the STimeData (they must have the
    // type INT32!), extract their values from the buffer 'buf'
    for (i=0; i < numberOfEntries; i++)
    {
        ptr = (char*)buf+pos.position[i];
        // this is a pointer to the first byte of the element in 'buf'

        switch (pos.length[i])
        {
          case 0:  // set to zero
            ((INT32*)time)[i] = 0;
            break;
          case 1:  // size is 1 byte --> use INT8
            Assert (i!=SMsgPosition::Accuracy);  // always send both alphas
            ((INT32*)time)[i] = *(INT8*)ptr;
            break;
          case 2:  // size is 2 bytes --> use INT16
            ((INT32*)time)[i] = *(INT16*)ptr;
            if (i==SMsgPosition::Accuracy)  // need to shift to position
            {
                time->acc <<= 8;  // now alphaPlus is correct
                time->acc = (time->acc & 0x0FF00)+(time->acc<<8 & 0xFF000000);
            }
            break;
          case 3:  // size is 3 bytes --> use INT16 and INT8
            Assert (i!=SMsgPosition::Accuracy);  // always send both alphas
            ((INT32*)time)[i] = *(INT8*)ptr;
            ((INT32*)time)[i] += (INT32)(*((INT16*)(ptr+1))) << 8;
            break;
          case 4:  // size is 4 bytes --> use INT32
            ((INT32*)time)[i] = *(INT32*)ptr;
            break;

          default:  // no supported size
            Assert (False);

        } // switch
    } // for


    // now sort the entries in 'pos' according to their position; uses
    // bubble sort

    INT32 tmppos, tmplen;

    for (i=0; i < numberOfEntries; i++)
    for (j=1; j < numberOfEntries-i; j++)
    {
        if (pos.position[j-1] > pos.position[j])
        {
            // swap the entries
            tmppos = pos.position[j-1];
            tmplen = pos.length[j-1];
            pos.position[j-1] = pos.position[j];
            pos.length[j-1] = pos.length[j];
            pos.position[j] = tmppos;
            pos.length[j] = tmplen;
        }
    } // for


    // now extract the data

    if (size-netsize <= 0)  // no more data
    {
        *msgsize = 0;
        *message = NULL;
        return;
    }

    // allocate the memory
    *msgsize = size-netsize;
    *message = new char[*msgsize];


    INT16 bufpos = 0;  // the position in 'buf' from which is copied
    INT16 msgpos = 0;  // the position in 'message' into which is copied

    INT16 itempos, itemlen;
    // position and length of the current sender item

    for (i=0; i < numberOfEntries; i++)
    {
        itempos = pos.position[i];
        itemlen = pos.length[i];

        Assert (bufpos <= itempos);
        // if bufpos is greater than itempos, the sort alg. has failed

        if (bufpos < itempos)
        // copy all memory till the next sender data
        {
            memcpy (*message+msgpos, (char*)buf+bufpos, itempos-bufpos);
            msgpos += itempos-bufpos;
        }

        bufpos = itempos + itemlen;  // go to next chunk of data
    } // for

    // check if there is something left after the last sender item
    if (bufpos < size)  // yes, there is something
    {
        memcpy (*message+msgpos, (char*)buf+bufpos, size-bufpos);
    }

    // the sender data has been copied into 'time', the rest of the
    // message is in 'message', and 'msgsize' denotes the size of that
    // array
}


//--------------------------------------------------------------------------
//--- CI function wrappers
//--------------------------------------------------------------------------

// just for easier access to the ci_dispatcher() function
static int CI_GetBuffer (INT16 size, bufx **buf, int if_num, int area)
{
    ciparam param;  // param structure for the CI dispatcher

    param.gtbp.size = size;
    param.gtbp.buf = buf;
    param.gtbp.if_num = if_num;
    param.gtbp.area = area;

    return ci_dispatcher (CI_GTB, &param);
}

static int CI_GetEnvelope (int mode, tenvx **tenv, int if_num, int area)
{
    ciparam param;

    param.gtep.mode = mode;
    param.gtep.tenv = tenv;
    param.gtep.if_num = if_num;
    param.gtep.area = area;

    return ci_dispatcher (CI_GTE, &param);
}

static int CI_RetBuffer (bufx *buf)
{
    ciparam param;
    param.rtbp.buf = buf;
    return ci_dispatcher (CI_RTB, &param);
}

static int CI_RetEnvelope (tenvx *tenv)
{
    ciparam param;
    param.rtep.tenv = tenv;
    return ci_dispatcher (CI_RTE, &param);
}

static int CI_Send (int mode, tenvx *tenv, bufx *buf, INT16 size,
                    int if_num, int area)
{
    ciparam param;
    param.sendp.mode = mode;
    param.sendp.tenv = tenv;
    param.sendp.buf = buf;
    param.sendp.size = size;
    param.sendp.if_num = if_num;
    param.sendp.area = area;
    return ci_dispatcher (CI_SEND, &param);
}


