//
// listbase.cpp
//
// definition of the abstract base list class
// note: reimplement ACObjectList when porting to hardware, since the
//       software implementation cannot handle concurrent access
//
// created: 27.10.96
//
// last change: 11.02.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <string.h>  // for memcpy etc.

#include "listbase.hpp"
#include "error.hpp"


//**************************************************************************
// class CObjectElement: this class represents an element of the sorted
//      array; it contains the data that is stored in the array
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

ACObjectList::CObjectElement::CObjectElement ()
{
    object_ = NULL;
    key_ = NULL;
}


//--------------------------------------------------------------------------
//--- set the members of the class
//--------------------------------------------------------------------------

// set the object-pointer and object ID
void ACObjectList::CObjectElement::SetObject (void *object, void *key)
{
    object_ = object;
    key_ = key;
}


//--------------------------------------------------------------------------
//--- read the members of the class
//--------------------------------------------------------------------------

// get the object-pointer
void* ACObjectList::CObjectElement::GetObject () const
{
    return object_;
}


// get the object key
void* ACObjectList::CObjectElement::GetKey () const
{
    return key_;
}


//**************************************************************************
// class ACObjectList: this abstract class manages a sorted array
//     it is assumed that most of the time the access is sequential; if
//     not, binary search is used
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// initialize the array
ACObjectList::ACObjectList (INT32 aSize)
{
    Assert (aSize != 0);  // creating an array with no size is ridiculous!
    
    _allocSize = aSize;
    _size = aSize;
    _array = new CObjectElement[_size];  // create a new array

    _firstFree = 0;  // no elements in the array
    _current = 0;
}

// has to be there, although the destructor is pure virtual, because
// destructors of base classes are always called
// since the constructor of the class allocates an array, the destructor
// frees the memory
// note that the destructor is pure virtual, because it can't delete the
// objects that are stored in the array; so a derived class *must* exist
// and handle the deletion of the objects
ACObjectList::~ACObjectList ()
{
    delete [] _array;
}



//--------------------------------------------------------------------------
//--- get the array status
//--------------------------------------------------------------------------

// returns True if array is empty, False otherwise
EBool ACObjectList::Empty() const
{
    if (_firstFree == 0) return True;
    else return False;
}


// returns True if the position pointer is standing at the beginning
// of the array, False otherwise (or if the array is empty)
EBool ACObjectList::BeginOfList () const
{
    if (!Empty() && (_current == 0)) return True;
    else return False;
}


// returns True if the position pointer is standing at the end of the array,
// False otherwise (or if the array is empty)
EBool ACObjectList::EndOfList () const
{
    if (!Empty() && (_current == _firstFree-1)) return True;
    else return False;
}


//--------------------------------------------------------------------------
//--- move in the array
//--------------------------------------------------------------------------

// get the first element, or NULL if the array is empty
void* ACObjectList::GetFirst ()
{
    if (Empty()) return NULL;
    else
    {
        _current = 0;
        return GetCurrent();
    }
}


// get the last element, or NULL if the array is empty
void* ACObjectList::GetLast ()
{
    if (Empty()) return NULL;
    else
    {
        _current = _firstFree-1;
        return GetCurrent();
    }
}


// get the next element in the array, or the last element if the end of the
// array has been reached;
// if _current is no valid index (because the array is empty, or because
// Find() was invoked with an unknown index), NULL is returned
void* ACObjectList::GetNext ()
{
    if (!CurrentIsValid()) return NULL;
    if (EndOfList()) return NULL;

    _current++;

    return GetCurrent();
}


// return the previous element in the array, or the first if the beginning
// of the array has been reached
// if _current is no valid index (because the array is empty, or because
// Find() was invoked with an unknown index), NULL is returned
void* ACObjectList::GetPrev ()
{
    if (!CurrentIsValid()) return NULL;
    if (BeginOfList()) return NULL;

    _current--;

    return GetCurrent();
}


// returns the element with this key or NULL, if no such element exists;
// if the element was found, _current contains its array index; if it
// was not found, then _current contains the index where it
// should be inserted; note that _current might be the same as
// _firstFree, so calling GetCurrentKey() would not be a good idea!
void* ACObjectList::Find (void *key)
{
    INT32 lowerBound, upperBound;  // bounds for the binary search

    if (key == NULL) return NULL;
    // no key to search for

    if (Empty()) return NULL;  // no elements in array

    if (!CurrentIsValid()) _current=_firstFree/2;
    // if current is not valid, set it to the middle of the array to help
    // with the search
    
    // if _current is not the element, set it to one of its neighbours
    // (this makes sequential access faster)
    if ((CompareToCurrent (key) < 0) && !BeginOfList())
        _current--;
    else if ((CompareToCurrent (key) > 0) && !EndOfList())
        _current++;

    // now check if _current points to the element searched
    if (CompareToCurrent (key) == 0)  // the element was found
    {
        return GetCurrent();  // return the object
    }


    // since _current has not found the element, access is neither
    // repeated nor sequential, so binary search is used
    if (CompareToCurrent (key) < 0)
    {
        lowerBound = 0;
        upperBound = (_current>0) ? (_current-1) : 0;
    }
    else
    {
        lowerBound = _current+1;
        upperBound = _firstFree-1;
    }

    _current = (lowerBound+upperBound)/2;
    
    while ((CompareToCurrent (key) != 0) &&
           (lowerBound < upperBound))
    {
        if (CompareToCurrent (key) < 0)
            upperBound = (_current>0) ? (_current-1) : 0;
        else lowerBound = _current+1;

        _current = (lowerBound+upperBound)/2;
    }

    // okay, now either the element was found, or it's not in the array
    if (CompareToCurrent (key) == 0)  // the element was found
    {
        return GetCurrent();  // return the object
    }
    else
    {
        if (CompareToCurrent (key) > 0) _current++;
        // this is the place where the element should be inserted
        
        return NULL;
    }
}


//--------------------------------------------------------------------------
//--- get information about the current element
//--------------------------------------------------------------------------

// returns a pointer to the member that is currently being accessed or
// NULL, if no element is being accessed
void* ACObjectList::GetCurrentObject () const
{
    if (CurrentIsValid()) return GetCurrent();
    else return NULL;
}


// returns the key of the currently accessed element or NULL, if no element
// is being accessed
void* ACObjectList::GetCurrentObjectKey () const
{
    if (CurrentIsValid()) return GetCurrentKey();
    else return NULL;
}


//--------------------------------------------------------------------------
//--- add/remove elements
//--------------------------------------------------------------------------


// insert a new element into the (sorted) array;
// if an object with this index exists, False is returned
EBool ACObjectList::Insert (void *object, void *key)
{
    if (object == NULL) return False;
    // only existing objects are inserted

    if (key == NULL) return False;
    // key has to exist as well

    if (Find (key) != NULL) // index does already exist
    {
        return False;
    }
    
    // it is inserted at the _current position
    InsertAtCurrent (object, key);

    return True;
}


// remove an object; if the object does not exist, NULL is returned
void* ACObjectList::Remove (void *key)
{
    void* object;

    if (key == NULL) return NULL;
    // no key to search for

    object = Find (key);  // search for the object

    if (object != NULL)  // object was found
    {
        RemoveAtCurrent ();  // remove the data entry
    }

    return object;  // return the object
}


//--------------------------------------------------------------------------
//--- internal information concerning the current element
//--------------------------------------------------------------------------

// returns True if _current is an occupied slot, False otherwise
EBool ACObjectList::CurrentIsValid () const
{
    if (_current < _firstFree) return True;
    else return False;
}


// returns the object ID of the current element; the element must exist!
void* ACObjectList::GetCurrentKey () const
{
    Assert (CurrentIsValid());  // the element has to exist
    
    return (_array + _current)->GetKey();
}


// returns the object of the current element; the element must exist!
void* ACObjectList::GetCurrent () const
{
    Assert (CurrentIsValid());  // the element has to exist
    
    return (_array + _current)->GetObject();
}


// returns -1 if key < current, = 0 if the keys are equal,
// and +1 if key > current; the current element must exist!
int ACObjectList::CompareToCurrent (void *key) const
{
    Assert (CurrentIsValid());  // the element has to exist

    // compare the keys; the function is provided by the derived class
    return Compare (key, GetCurrentKey());
}


//--------------------------------------------------------------------------
//--- internal memory management
//--------------------------------------------------------------------------

// allocates a new array of size _size and copies the data of _firstFree
// elements from _array to the new array; _array is then deleted and set
// to the new array
// assert: _firstFree <= _size
void ACObjectList::UpdateArray()
{
    CObjectElement *newArray;

    Assert (_firstFree <= _size);  // the data has to fit the new array
    
    newArray = new CObjectElement[_size];  // make a new array
    memcpy (newArray, _array, _firstFree*sizeof(CObjectElement));
    // copy the data to the new array
    
    delete[] _array;  // delete the old array
    _array = newArray;  // set the _array-Pointer to the new array
}


// allocates a new array that is _allocSize elements bigger than the
// current array size stored in _size; the data is copied to the new
// array and _array is set to it; the old array is discarded
void ACObjectList::AllocateMemory()
{
    _size = _size + _allocSize;  // have to allocate more space
    UpdateArray();  // make an array of the new size and copy the data
}


// frees the last _allocSize elements of the current array; note that
// the function will assume that these elements are not in use! if
// they are, the call to UpdateArray will fail!
// assert: size>allocSize
void ACObjectList::FreeMemory()
{
    _size = _size - _allocSize;  // reduce the size of the array
    Assert (_size > 0);  // array must have a positive size
    UpdateArray();  // make an array of the new size and copy the data
}


//--------------------------------------------------------------------------
//--- internal array information
//--------------------------------------------------------------------------

// returns the current size of the array
INT32 ACObjectList::Size() const
{
    return _size;
}


// returns the number of array elements that are currently in use
INT32 ACObjectList::Occupied() const
{
    return _firstFree;
}


// returns the additional size used for expansion (contraction) of the array
INT32 ACObjectList::AllocationSize() const
{
    return _allocSize;
}


//--------------------------------------------------------------------------
//--- internal array manipulation
//--------------------------------------------------------------------------

// insert the new data at _current;
// if _current is not the end of the array, the following members are
// moved back using memmove()
// if the new data has occupied the last free space, a new array
// with bigger size (determined by _allocSize) is created and data is
// copied using memcpy();
// assert: the array must have at least one element free!
//         _current must be a valid index or the first free index!
void ACObjectList::InsertAtCurrent (void *object, void *key)
{
    Assert (_firstFree < _size);  // a free element must exist
    Assert (_current <= _firstFree);  // _current must be a valid slot
    
    if (_current != _firstFree)  // shift the following entries to the right
    {
        memmove ((_array+_current+1), (_array + _current),
                 (_firstFree-_current)*sizeof(CObjectElement));
        // move all elements from _current on one item back in the array
        // note: here, memmove() is used because it works properly with
        // overlapping objects (as opposed to memcpy()!)
    }

    (_array + _current)->SetObject (object, key);
    // insert the new data at the _current position

    _firstFree++;  // set _free to the next free element

    if (_firstFree == _size)  // the end of the free buffer is reached
    {
        AllocateMemory();  // allocate new memory
    }
}


// remove the element defined by _current from the array and delete the key;
// if the element is not the last entry, the following items are moved
// left using memmove;
// if the removal of the objects leaves more than _allocSize+1 places free,
// _allocSize elements are freed;
// assert: _current must be a valid index!
void ACObjectList::RemoveAtCurrent ()
{
    Assert (CurrentIsValid());  // _current must be valid

    DeleteKey (GetCurrentKey());  // delete the key
    // note: the key is never removed, only the object

    if (_current != (_firstFree-1))  // have to move some elements
    {
        memmove ((_array+_current), (_array+_current+1),
                 (_firstFree-_current-1)*sizeof(CObjectElement));
        // move all elements after _current one slot to the left
    }
    else if (_current != 0)
        _current--;  // set current to the new last element
    
    _firstFree--;  // the size of the used array was reduced

    if ((_size-_firstFree) > _allocSize+1)
    // note: with this approach, at least two free elements are saved
    // from deallocation; therefore repeated insertion and removal of
    // an element won't result in continuous allocation/disallocation
    // of the memory
    {
        FreeMemory();  // delete the last portion of the array
    }
}



