//
// main.cpp
//
// the main function of the simulation; checks options and starts the
// first simulation process
//
// created: 19.10.96
//
// last change: 03.02.99
//
// RCS version control information:
//     $Revision: 1.16 $$State: Exp $$Date: 1999/02/05 16:57:19 $
//


#ifndef PROCESS_H_
#  include <ClassLib/Process.h>
#endif

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "error.hpp"
#include "tester.hpp"


//**************************************************************************
// SimUTC version information
//**************************************************************************

#define NAME		"SimUTC"
#define IFC_VERSION	"2.3"
#define INTL_VERSION	"1.0"
#define COMMENTS	"added new fault types"
// the name and current version numbers of SimUTC


//**************************************************************************
// globals
//**************************************************************************

double g_delayTime = 0;
// the time that the sleeper process sleeps per second (thus delaying the
// whole simulation);
// it is set by the main() function (user definable via an option), is given
// in seconds (granularity is microseconds), and is needed to delay the
// simulation, because on a fast processor the events are so close together
// that they clog up the evaluation (XWindow-Server does not update screen
// anymore)
// it is used in tester.cpp by class CSleeper (the sleeper process)

double g_sleeperPeriod = 1;
// the period of the sleeper process
// this value specifies how often the process calls usleep() to delay
// the simulation; a big value causes the sleeper to run without delay
// for this time, and it is delayed for rather long when the sleeper
// sleeps; a small value lets the sleeper delay more often for shorter
// times - use this if you experience problems with sockets, if the
// evaluation is too slow, or if snapshot events are processed out of
// order

unsigned long g_streamOffset = 0;
// the offset in a random stream
// when a random generator object is created, it draws and discards a
// certain number of random variables to get to a desired offset in the
// random stream; this offset is normally fixed and will be the same in
// different simulation runs
// in order to get an indeterministic behaviour, g_streamOffset is added
// to the fixed offset; g_streamOffset holds a random number in the range
// [0,1000], the number is attained with the built-in random generator of C


//**************************************************************************
// local functions
//**************************************************************************

static void PrintUsage (const char *progname);
// prints the help message

static void PrintVersion();
// prints the current version (option -v)


//**************************************************************************
// main function
//**************************************************************************

int main (int argc, char** argv)
{
    int opt;
    char *nextarg;  // for reading the delay time

    while ((opt = getopt(argc,(char**)argv,"vhd:p:i")) != -1)
    {
        switch (opt)
        {
            case 'v':  // version
                PrintVersion();
                return 0;
                break;
            case 'h':  // help
                PrintUsage (argv[0]);
                return 0;
                break;
            case 'd':  // simulation delay
                g_delayTime = strtod (optarg, &nextarg);
                break;
            case 'p':  // sleeper process period
                g_sleeperPeriod = strtod (optarg, &nextarg);
                if (g_sleeperPeriod < 1.0/TICKS2SEC)
                    g_sleeperPeriod = 1.0/TICKS2SEC;
                // enforce the pSOS timer resolution
                if (g_sleeperPeriod < 1e-6) g_sleeperPeriod = 1e-6;
                // enforce at least 1 us resolution
                break;
            case 'i':  // make random generators indeterministic
                g_streamOffset = ((unsigned long)time(NULL)*3) % 1000;
                break;
        } // switch
    } // while


    Thread::Initialize();

    CTester t;

    t.Await();
    t.Exit();
    
    return 0;
}


//**************************************************************************
// local functions (implementation)
//**************************************************************************

// converts the input 'value', which is assumed to be in seconds, into
// a value >= 1, which is returned in 'norm'; the unit of the value is
// returned in 'unit', which must be big enough to contains the string
// and the trailing \0
// assert: unit != NULL
static void Normalize (double value, double& norm, char *unit)
{
    Assert (unit != NULL);

    static char units[5][3] = {"s", "ms", "us", "ns", "ps"};
    // all recognized units
    int index = 0;  // the index in the units-array

    // multiply 'value' by thousand until it is either >= 1, or we have
    // reached the last of our units, or the value is equal to zero
    for (index = 0;
         (index < 5) && (value < 1) && (value != 0);
         index++)
    {
        value = value*1000;
    } // for

    norm = value;  // copy the modified value into 'norm'
    strcpy (unit, units[index]);  // copy the unit
}

// prints the help message
static void PrintUsage (const char *progname)
{
    double granularity;  // the granularity of delay and period
    char unit[8];  // the unit of the granularity


    //--- usage

    cout << endl;
    cout << "usage: " << progname;
    cout << " -v | -h | [-d <delay> [-p <period>]] [-i]";
    cout << endl;
    cout << endl;


    //--- option v

    cout << "   v:          print version" << endl;
    cout << endl;


    //--- option h

    cout << "   h:          this message" << endl;
    cout << endl;


    //--- option d

    cout << "   d <delay>:  delay the simulation for <delay> seconds";
    cout <<                 " per simulated second" << endl;
    cout << "               (with a granularity of ";
    Normalize (1.0/CLOCKS_PER_SEC, granularity, unit);
    if (CLOCKS_PER_SEC > 1000000)  // less than 1 us
    // note: usleep() takes microseconds as argument, but is bound by the
    // clock granularity CLOCKS_PER_SEC
    {
        granularity = 1;
        strcpy (unit, "us");
    }
    cout << granularity << " " << unit << ")" << endl;
    cout << "               default: " << g_delayTime << " s" << endl;
    cout << endl;


    //--- option p

    cout << "   p <period>: the simulation is delayed every <period> seconds";
    cout << endl;
    cout << "               (with a granularity of ";
    Normalize (1.0/TICKS2SEC, granularity, unit);
    if (TICKS2SEC > 1000000)  // less than 1 us
    // note: even if the granularity of the pSOS timer allows higher
    // resolution than 1 us, we cut off at this point
    {
        granularity = 1;
        strcpy (unit, "us");
    }
    cout << granularity << " " << unit << ")" << endl;
    cout << "               default: " << g_sleeperPeriod << " s" << endl;
    cout << endl;


    //--- option i

    cout << "   i:          indeterministic random generators" << endl;
    cout << "               specifying this option makes every run of simutc ";
    cout << "unique" << endl;
    cout << "               (without it, runs are repeatable)" << endl;
    cout << endl;
}

// prints the current version (option -v)
static void PrintVersion()
{
    cout << "This is " << NAME;
    cout << " version " << IFC_VERSION << "." << INTL_VERSION << endl;
    cout << " -> interface version " << IFC_VERSION;
    cout << ", internal version " << INTL_VERSION;
    cout << " (" << COMMENTS << ")" << endl;
}


