//
// netacc.cpp
//
// definition of the NetworkAccountant class responsible for the management
// of the transmission delays
//
// created: 29.07.96
//
// last change: 24.09.98
//
// RCS version control information:
//     $Revision: 1.14 $$State: Exp $$Date: 1998/09/28 10:59:06 $
//


#include <stdlib.h>  // for the NULL macro

#include "constant.hpp"
#include "distrib.hpp"
#include "error.hpp"
#include "netacc.hpp"
#include "structs.hpp"


// returns NoAccount; will be replaced by NoAccount in future versions
static inline SINT32 GetNoAccount () { return NoAccount; }


//**************************************************************************
// class CNetworkAccountant: this class is a database for the network and
//                           stores transmission delays;
// it can compute an arbitrary delay between two nodes using the data stored
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CNetworkAccountant::CNetworkAccountant (SINT32 networkID)
{
    Assert (Early < EArrivalCount);
    Assert (InTime < EArrivalCount);
    Assert (Late < EArrivalCount);
    // we need these assumptions later in Set/GetParameter where they
    // would have to be checked every time; this is better

    networkID_ = networkID;  // stays the same for the whole life o.t. class

    initCalled_ = False;
    CNetworkAccountant::Init ();
}


CNetworkAccountant::~CNetworkAccountant ()
{
    CNetworkAccountant::Dispose ();
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
EBool CNetworkAccountant::SetParameter (const SParameter& param)
{
    SSingleNetworkParameter *sparam = ((SSingleNetworkParameter*)&param);
    // just to make the following code shorter
    SNetDistributionP2P *dist = (SNetDistributionP2P *)&param;

    switch (param.type)
    {
        case DistributionEarly:
        case DistributionTimely:
        case DistributionLate:
            if (!dist->data.DataValid()) return False;
            return SetDataParameter (dist);
            break;
        case PEarly:
        case PLate:
            return SetProbability (*(SNetworkData*)&param);
            break;

        default:  // unknown parameter
            return False;
    } // switch

    return True;
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned
EBool CNetworkAccountant::GetParameter (SParameter& param)// const
{
    SSingleNetworkParameter *sparam = ((SSingleNetworkParameter*)&param);
    SNetDistributionP2P *dist = (SNetDistributionP2P *)&param;
    double temp;

    switch (param.type)
    {
        case DistributionEarly:
        case DistributionTimely:
        case DistributionLate:
            return GetDataParameter (dist);
            break;
        case PEarly:
        case PLate:
            temp = GetProbability (*(SNetworkData*)&param);
            if (temp == GetNoAccount()) return False;
            sparam->value = temp;
            break;

        default:  // unknown parameter
            return False;
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- management of transmission delays
//--------------------------------------------------------------------------

// returns a random transmission delay in the specified interval for the
// pair of nodes;
// if no delays for this pair of nodes are stored and sender != receiver,
// the method returns NoAccount; if sender == receiver, constant
// DefaultOwnDelay is returned
double CNetworkAccountant::GetArbitraryDelay (SINT32 senderID,
        SINT32 receiverID, EArrival& arrivalType)
{
    SNetworkDelay delay;

    if (!GetDelay (&delay, senderID, receiverID))  // no entry
    {
        if (senderID == receiverID) return DefaultOwnDelay;
        // if no transmission delay is specified, DefaultOwnDelay is assumed
        else return GetNoAccount();
    }

    if (arrivalType == EArrivalCount)  // choose the interval
    {
        if (random_->CastDice (delay.probEarly)) arrivalType = Early;
        else if (random_->CastDice (delay.probLate/(1-delay.probEarly)))
            arrivalType = Late;
        else arrivalType = InTime;
    } // if

    SRandomData dist;  // the data that is used

    switch (arrivalType)
    {
        case Early:
            dist = delay.slow;
            if (!dist.DataValid())  // inconsistent data
            // this can only happen if the regular interval has been set
            // after the early interval
            {
                if (dist.low > dist.high) return GetNoAccount();
                // cannot use the data

                if (dist.mean > dist.high)  // set mean into middle
                {
                    Report (have to adapt mean of early transmission);
                    dist.mean = (dist.low + dist.high)/2;
                }
            } // if
            break;

        case InTime:
            dist = delay.regular;
            // check bounds against the early/late bounds, adapt if necessary
            if (dist.low+dist.constant <
                delay.slow.low+delay.slow.constant)
                dist.low = delay.slow.low+delay.slow.constant-dist.constant;
            if (dist.high+dist.constant >
                delay.fast.high+delay.slow.constant)
                dist.high = delay.slow.high+delay.slow.constant-dist.constant;

            if (!dist.DataValid()) return GetNoAccount();
            // mortally inconsistent data

            break;

        case Late:
            dist = delay.fast;
            if (!dist.DataValid())  // inconsistent data
            // this can only happen if the regular interval has been set
            // after the late interval, or if the late interval has
            // not been set at all
            {
                if (dist.low > dist.high) return GetNoAccount();
                // cannot use the data

                if (dist.mean > dist.high)  // set mean into middle
                {
                    Report (have to adapt mean of early transmission);
                    dist.mean = (dist.low + dist.high)/2;
                }
            } // if
            break;

        default:
            Assert (arrivalType != arrivalType);  // should never happen
    } // switch

    return random_->RandomVar (dist);  // return the random variable
}


//--------------------------------------------------------------------------
//--- management of nodes
//--------------------------------------------------------------------------

// if the node exists, all data concerning the node (whether as sender or
// as receiver) will be destroyed and the method will return True
// if the node doesn't exist, False is returned
EBool CNetworkAccountant::DeleteNode (SINT32 nodeID)
{
    EBool nodeExisted;
    
    nodeExisted = DeleteKey1 (nodeID);
    // delete the node from the list of senders

#ifdef LINUX_SIM
    operator |= (nodeExisted, DeleteKey2 (nodeID));
    // delete the node from the lists of receivers
    // note: g++ does not use the operator |= if it is called directly;
    // but if is is called explicitly, then the operator is used
#else
    nodeExisted |= DeleteKey2 (nodeID);
#endif

 
    return nodeExisted;
}


// returns True if an entry for the pair of nodes exists, False otherwise
// if the entry exists, it is stored in the buffer for future access
EBool CNetworkAccountant::IsConnected (SINT32 senderID, SINT32 receiverID)
{
    return GetInformation (senderID, receiverID);
}


//--------------------------------------------------------------------------
//--- initialization and destruction of the data members
//--------------------------------------------------------------------------

// initializes all class members to default values; a second call
// (without Dispose() in between) has no effect
void CNetworkAccountant::Init ()
{
    if (initCalled_) return;  // Init() has already been called
    initCalled_ = True;

    CAccountant::Init();

    // select the random stream according to the network ID
    random_ = new CRandom (networkID_);
    Assert (random_ != NULL);
}


// deletes all dynamic members; a second call (without Init() in
// between) has no effect
void CNetworkAccountant::Dispose ()
{
    if (!initCalled_) return;  // Dispose() has already been called
    delete random_;
    CAccountant::Dispose();
    initCalled_ = False;
}


//--------------------------------------------------------------------------
//--- internal management of probabilities
//--------------------------------------------------------------------------

// set the stored arrival probability to the given value
// if 0<=prob and probEarly+probLate<=1, the data is stored and True is
// returned; else False is returned and the method has no effect
EBool CNetworkAccountant::SetProbability (SNetworkData& param)
{
    SNetworkDelay delay;
    
    if (param.value < 0) return False;  // prob must be >= zero
 
    GetDelay (&delay, param.sender, param.receiver);
    // if entry exists, get it

    switch (param.type)
    {
        case PEarly:
            if (param.value + delay.probLate > 1) return False;
            delay.probEarly = param.value;  // store the probability
            break;
        case PLate:
            if (param.value + delay.probEarly > 1) return False;
            delay.probLate = param.value;  // store the probability
            break;

        default:
            return False;
    } // switch

    SetDelay (delay, param.sender, param.receiver);  // store the data
    return True;
}

// get the specified probability or NoAccount, if no data is stored
double CNetworkAccountant::GetProbability (SNetworkData& param)
{
    SNetworkDelay delay;

    if (!GetDelay (&delay, param.sender, param.receiver))  // no account
        return GetNoAccount();
    else
        switch (param.type)
        {
            case PEarly:
                return delay.probEarly;  // return the data
                break;
            case PLate:
                return delay.probLate;  // return the data
                break;
            default:
               return GetNoAccount();
        } // switch

    return GetNoAccount();  // never executed
}


// sets the given (data) parameter to its new value; if the parameter does
// not exist or if the value is not valid, False is returned
EBool CNetworkAccountant::SetDataParameter (const SNetDistributionP2P *param)
{
    SNetworkDelay delay;
    double temp;  // for consistency checks

    GetDelay (&delay, param->sender, param->receiver);
    // if entry exists, get it

    switch (param->type)
    {
        case DistributionEarly:
            // the upper bound is taken from the regular interval
            temp = delay.regular.low + delay.regular.constant-
                   param->data.constant;
            // this is the new upper bound of the early interval

            if (temp < param->data.low) return False;
            // mortally inconsistent - deny set request

            delay.slow = param->data;  // store the data
            delay.slow.high = temp;  // store upper bound

            if (temp < param->data.mean) // inconsistent, but we can repair it
            {
                Report (have to adapt mean of early transmission);
                delay.slow.mean = (delay.slow.low + delay.slow.high)/2;
                // we set the mean into the middle of the interval
            }

            break;

        case DistributionTimely:
            // timely data is always accepted, even if it clashes with the
            // bounds; checks are only done when the data is used
            // the fast/slow intervals are adapted as necessary; if the
            // modifications to them exceed setting the upper (slow interval)
            // or lower (fast interval) bound, an error report is issued
            // to notify the user (such extensive changes were probably not
            // intended)
            if (!param->data.DataValid()) return False;
            delay.regular = param->data;  // store the regular data

            // adapt upper bound of slow
            delay.slow.high = delay.regular.low+delay.regular.constant-
                              delay.slow.constant;

            // check for inconsistencies in the slow interval
            if (delay.slow.high < delay.slow.low)
            // the change has made the slow interval inconsistent!
            {
                if (delay.slow.high < 0)  // the constant is now too big!
                {
                    // add the constant to high to get the required upper
                    // bound and then set the constant to zero
                    delay.slow.high += delay.slow.constant;
                    delay.slow.constant = 0;
                }
                // fix the interval, reduce it to a point at the
                // lower bound of the regular interval, and report
                // that the interval has changed
                delay.slow.low = delay.slow.high;
                delay.slow.mean = delay.slow.high;
                Report (DistributionTimely has reduced slow interval to point);
            } // if
            else if (delay.slow.high < delay.slow.mean)
            // the mean is outside the bounds
            {
                delay.slow.mean = (delay.slow.low + delay.slow.high)/2;
                Report (DistributionTimely has changed mean of slow interval);
                // we set the mean into the middle of the interval
            }

            // adapt lower bound of fast
            delay.fast.low = delay.regular.high+delay.regular.constant-
                             delay.fast.constant;

            // check for inconsistencies in the fast interval
            if ((delay.fast.low > delay.fast.high) || (delay.fast.low < 0))
            // the change has made the fast interval inconsistent!
            // note: here, we have to check for the <0 condition explicitly,
            // because it is not contained in the first check (in the case
            // of the slow interval, the first and only condition is also
            // true if the value is <0)
            {
                if (delay.fast.low < 0)  // the constant is now too big!
                {
                    // add the constant to low to get the required upper
                    // bound and then set the constant to zero
                    delay.fast.low += delay.fast.constant;
                    delay.fast.constant = 0;
                }
                // fix the interval, reduce it to a point at the
                // upper bound of the regular interval, and report
                // that the interval has changed
                delay.fast.high = delay.fast.low;
                delay.fast.mean = delay.fast.low;
                Report (DistributionTimely has reduced fast interval to point);
            } // if
            else if (delay.fast.low > delay.fast.mean)
            // the mean is outside the bounds
            {
                delay.fast.mean = (delay.fast.low + delay.fast.high)/2;
                Report (DistributionTimely has changed mean of fast interval);
                // we set the mean into the middle of the interval
            }

            break;

        case DistributionLate:
            // the lower bound is taken from the regular interval
            temp = delay.regular.high+delay.regular.constant-
                   param->data.constant;
            // this is the new lower bound of the late interval

            if (temp > param->data.high) return False;
            // mortally inconsistent - deny set request

            delay.fast = param->data;  // store the data
            delay.fast.low = temp;

            if (temp > param->data.mean) // inconsistent, but we can repair it
            {
                Report (have to adapt mean of late transmission);
                delay.fast.mean = (delay.fast.low + delay.fast.high)/2;
                // we set the mean into the middle of the interval
            }

            break;
        default:
            return False;
    } // switch

    SetDelay (delay, param->sender, param->receiver);  // store the data
    return True;
}

// returns the value of the given (network data) parameter in the
// appropriate field of SParameter; if the parameter does not exist, False
// is returned and 'param' is not changed
EBool CNetworkAccountant::GetDataParameter (SNetDistributionP2P *param)
{
    SNetworkDelay delay;

    if (!GetDelay (&delay, param->sender, param->receiver)) return False;
    // if entry exists, get it

    switch (param->type)
    {
        case DistributionEarly:
            param->data = delay.slow;
            break;
        case DistributionTimely:
            param->data = delay.regular;
            break;
        case DistributionLate:
            param->data = delay.fast;
            break;
        default:
            return False;
    } // switch

    return True;
}

