//
// netbase.cpp
//
// implementation of the abstract network class
//
// created: 04.07.96
//
// last change: 28.09.98
//
// RCS version control information:
//     $Revision: 1.11 $$State: Exp $$Date: 1998/09/28 11:05:37 $
//


#include <string.h>  // for memcmp etc.

#include "error.hpp"
#include "globals.hpp"
#include "idlist.hpp"
#include "netacc.hpp"
#include "netbase.hpp"
#include "netgen.hpp"
#include "sim_gen.hpp"
#include "sim_glob.hpp"
#include "structs.hpp"


static const addr broadcastAddress = BROADCAST_ADDRESS;
// need this to compare with the destination address in SendMsg()


//**************************************************************************
// class CDelayer: a process class that is needed to delay messages before
//      they are processed by the network or delivered by the messenger
//**************************************************************************

//--------------------------------------------------------------------------
//--- delay the message
//--------------------------------------------------------------------------

// returns the message to the network again (calling Transmit); the
// network is identified by the if_num stored in SMsg
void CDelayer::Deliver (SMsg *msg)
{
    ACNetwork *network;

    network = listOfNetworks.Find (msg->if_num);  // find the network
    Assert (network != NULL);  // can only happen if the ID was wrong
    network->Transmit (msg);  // tell the network to transmit the msg
}



//**************************************************************************
// class CMessenger: a process class that is needed to deliver (dedicated)
//      messages to all Lance objects of a node
//**************************************************************************

//--------------------------------------------------------------------------
//--- message delivery
//--------------------------------------------------------------------------

// delivers the message to the lances of the receiving node
void CMessenger::Deliver (SMsg *msg)
{
    ACNetwork *network;

    SetNodeID (msg->receiver);
    // the messenger process runs in the "context" of the receiving node
    // note: necessary, since ClockSync may reply to a message at once,
    // calling ACNetwork::SendMsg() in the context of this process

    network = listOfNetworks.Find (msg->if_num);  // find the network
    Assert (network != NULL);  // can only happen if the ID was wrong
    network->Deliver (msg);  // tell the network to deliver the msg
}



//**************************************************************************
// class ACNetwork: this is an abstract class that provides methods for
//     sending data to receiving nodes;
//     a derived class should overload the SendMsg method to implement a
//     given network
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// init all variables
ACNetwork::ACNetwork (CControllingObject *ctrl, SINT32 networkID)
    : CControlledObject (ctrl)
{
    initCalled_ = False;
    networkID_ = networkID;
    ACNetwork::Init();  // initialize all members
    // note: specifying the scope is not necessary, because in the ctor
    // the derived class does not exist
    Assert (listOfNetworks.Insert (this, networkID_));
}

ACNetwork::~ACNetwork ()
{
    listOfNetworks.Remove (networkID_);
    ACNetwork::Dispose();  // delete all dynamic members
    // again, specifying the scope is not necessary but doesn't hurt
}


//--------------------------------------------------------------------------
//--- reset all members to their initial values
//--------------------------------------------------------------------------

// restores the initial state
void ACNetwork::Reset ()
{
    Dispose();
    Init();
    CCrashingModule::Reset();
}


//--------------------------------------------------------------------------
//--- control the system parameters
//--------------------------------------------------------------------------

// sets the requested parameter value
// existing params: Lambda, DistributionLambda, Omega, DistributionOmega,
// and all params of CControlledObject, CCrashingModule and of the network
// accountant
EBool ACNetwork::SetParameter (const SParameter& param)
{
    SSingleNetworkParameter *sparam = ((SSingleNetworkParameter*)&param);
    SNetDistribution *dist = (SNetDistribution *)&param;

    switch (param.type)
    {
      case DistributionLambda:
          if (!dist->data.DataValid()) return False;
          distDataLambda = dist->data;
          break;
      case DistributionOmega:
          if (!dist->data.DataValid()) return False;
          distDataOmega = dist->data;
          break;
      case PCrash:
          return CCrashingModule::SetProbability (True, sparam->value);
          break;
      case PRecover:
          return CCrashingModule::SetProbability (False, sparam->value);
          break;

      default:  // unknown parameter type, pass along to base or to NA
          if (CControlledObject::SetParameter (param)) return True;
          else return networkAccountant->SetParameter (param);
    } // switch

    return True;
}

// returns the requested parameter value
EBool ACNetwork::GetParameter (SParameter& param) const
{
    SSingleNetworkParameter *sparam = ((SSingleNetworkParameter*)&param);
    SNetDistribution *dist = (SNetDistribution *)&param;

    switch (param.type)
    {
      case DistributionLambda:
          dist->data = distDataLambda;
          break;
      case DistributionOmega:
          dist->data = distDataOmega;
          break;
      case CrashStatus:
          sparam->value = Crashed();
          break;
      case PCrash:
          sparam->value = CCrashingModule::GetProbability (True);
          break;
      case PRecover:
          sparam->value = CCrashingModule::GetProbability (False);
          break;

      default:  // unknown parameter type, pass along to base or to NA
          if (CControlledObject::GetParameter (param)) return True;
          else return networkAccountant->GetParameter (param);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command; returns True if successfull, otherwise False
// existing commands: (Dis-)ConnectNode and all commands of
// CControlledObject and CCrashingModule
EBool ACNetwork::ExecuteCommand (const SCommand& command)
{
    SNetworkConnectCommand *conn = (SNetworkConnectCommand *)&command;

    switch (command.type)
    {
        case ConnectNode:
            Connect (conn->nodeID);
            break;
        case DisconnectNode:
            Disconnect (conn->nodeID);
            break;

        default:  // pass on to the base classes
            if (CControlledObject::ExecuteCommand (command)) return True;
            else return CCrashingModule::ExecuteCommand (command);
    } // switch

    return True;
}


//--------------------------------------------------------------------------
//--- control message transmission
//--------------------------------------------------------------------------

// send the message to the destination given in tenv->da (this function
// just creates a SMsg and then calls SendMsg (SMsg *msg))
// if the sending node (which is found by calling GetCurrentNodeID()) is
// not connected, the message is discarded
// note: the function assumes that tenv->length_field = TS_PACKET if the
// message is to be timestamped
void ACNetwork::SendMsg (tenvx *tenv, bufx *buf, INT16 size, int area)
{
    if (tenv == NULL)  // user of the class did something wrong
    {
        Report (tenv == NULL);  // report it
        return;
    }

    SMsg *msg = new SMsg (buf, size, GetNetworkID(), area);

    Assert (msg != NULL);

    msg->sender = GetCurrentNodeID();
    // the sender nodeID is the ID of the currently active process

    msg->receiver = NoAccount;
    // right now, the message is not dedicated

    Assert (buf != NULL);
    Assert (size >= sizeof(INT32));  // the msgID must be in the buffer

    memcpy (&msg->msgID, msg->buf, sizeof(INT32));
    // the Lance puts the message ID as first item in the buffer

    if (!listOfNodes->Find (msg->sender))
    {
        Report (sending node is not connected to the network);
        delete msg;
        return;
    }

    if (memcmp (tenv->da, broadcastAddress, sizeof(addr)) != 0)
    {
        // this is a dedicated message!
        NetworkToNode (&(tenv->da), msg->receiver);
        msg->broadcast = False;
    }
    else msg->broadcast = True;  // this is a broadcast message

    if ((tenv->length_field & TS_PACKET) == TS_PACKET)
        msg->timestamped = True;
    else
    {
        Report (message is not timestamped);  // most likely an oversight
        msg->timestamped = False;
    }

    SendMsg (msg);  // send the message

    delete msg;  // the structure is not needed anymore
}


//--------------------------------------------------------------------------
//--- initialization and destruction of the data members
//--------------------------------------------------------------------------

// initialize all members
void ACNetwork::Init ()
{
    if (initCalled_) return;  // Init() has already been called
    initCalled_ = True;

    networkAccountant = new CNetworkAccountant (networkID_);
    listOfNodes = new CNodeIdList;
    messenger = new CMessenger;
    delayer = new CDelayer;

    Assert (networkAccountant != NULL);
    Assert (listOfNodes != NULL);
    Assert (messenger != NULL);
    Assert (delayer != NULL);

#ifndef NO_RESOURCE
    Resource::ref(messenger);  // register the processes in C++SIM
    Resource::ref(delayer);
#endif

    distDataLambda.dist = Uniform;
    distDataLambda.low = 0;
    distDataLambda.high = 0;
    distDataLambda.mean = 0;
    distDataLambda.constant = 0;
    distDataLambda.freeparam = 0;

    distDataOmega = distDataLambda;

    random = new CRandom (networkID_);
    Assert (random != NULL);
}

// delete all dynamic members
void ACNetwork::Dispose()
{
    if (!initCalled_) return;  // Dispose() has already been called
    initCalled_ = False;

    delete listOfNodes;
    delete networkAccountant;

    messenger->Cancel();  // stop the messenger
    messenger->terminate();  // terminate the process

    delayer->Cancel();  // stop the delayer
    delayer->terminate();  // terminate the process

#ifndef NO_RESOURCE
    Resource::unref(messenger);  // either unref the processes in C++SIM
    Resource::unref(delayer);
#else
    delete messenger;  // or delete them
    delete delayer;
#endif

    delete random;
}


//--------------------------------------------------------------------------
//--- query of the network ID
//--------------------------------------------------------------------------

SINT32 ACNetwork::GetNetworkID ()
{
    return networkID_;
}


//--------------------------------------------------------------------------
//--- control network crashes
//--------------------------------------------------------------------------

// crash the net; if the net was already crashed, this method has no effect
void ACNetwork::Crash ()
{
    for (SMsg *msg = messenger->GetFirst();
         msg != NULL;
         msg = messenger->GetNext())
    {
        ReportMessageEvent (MessageLost, msg);
        // notify EvalSys that the messages on the net will be discarded
    }

    messenger->Reset ();  // discard any messages that are on the net
    // note: messages of the delayer are not touched, because there are
    // situations in which they must be kept (pe. the initial broadcast
    // delay, which is modelled in the network but which rather belongs to
    // the network controller and driver)

    CCrashingModule::Crash();  // do other crash activities
}


//--------------------------------------------------------------------------
//--- control the connection of nodes
//--------------------------------------------------------------------------

// the node can send messages to this network
void ACNetwork::Connect (SINT32 nodeID)
{
    listOfNodes->Insert (nodeID);
}

// the node cannot send messages to this network
void ACNetwork::Disconnect (SINT32 nodeID)
{
    listOfNodes->Remove (nodeID);
}


//--------------------------------------------------------------------------
//--- internal control of message transmission
//--------------------------------------------------------------------------

// transmit a single message (between a pair of nodes): compute the
// transmission time and give the message to the messenger
void ACNetwork::Transmit (SMsg *msg)
{
    double d;  // the transmission time
    EArrival arrival = EArrivalCount;  // let netacc choose the type

    d = networkAccountant->GetArbitraryDelay (msg->sender, msg->receiver,
                                              arrival);

    if (d != NoAccount)  // transmit the message
    {
        if (arrival == Early) ReportMessageEvent (EarlyMessage, msg);
        else if (arrival == Late) ReportMessageEvent (LateMessage, msg);
        else ReportMessageEvent (SendMessage, msg);
        // note: the last statement is called if the arrival=InTime, or
        // if netacc has not changed arrival (if there was no delay data
        // stored and it used DefaultOwnDelay)

        messenger->Insert (new SMsg (*msg), d);
    }
    else Report (computation of transmission delay has failed);
}

// delivers the message to the Lances of the receiver node; the CI buffers
// are copied into 'local' memory (of course, in pure simulation all memory
// is local), and it is the task of each Lance module to delete them
// the local memory is obtained by CLance::GetReceiveBuffer() and -Envelope()
// and should be returned by the corresponding Ret-functions
// note: the function must be called by a process which has the nodeID
// of the receiver node
// the function may inject faults, and it checks the crash-status before
// delivering the message
void ACNetwork::Deliver (SMsg *msg)
{
    STimestamp time;  // the receiver timestamp
    CLance *lance;  // the receiving lance
    renvx *renv = new renvx;  // return buffer

    Assert (renv != NULL);

    ModifyCrashStatus ();  // fault injection

    if (Crashed())
    {
        ReportMessageEvent (MessageLost, msg);
        return;  // do not deliver the message
    }

    ReportMessageEvent (DeliverMessage, msg);

    // the sender address must be an internet address
    NodeToNetwork (msg->sender, &(renv->sa));

    // get the receive timestamps
    int area;
    for (area=0; area<MAXASICS; area++)
    {
        // check if the area is in use
        if (!GetClockvalue (msg->receiver, area, time)) continue;
/**/    // this could change to a break (the usage of areas is continuous)

        renv->utcsu[area].rts = time.ts;
        renv->utcsu[area].rms = time.ms;
    }

    // deliver the message to the Lances

    bufx *localBuf;  // a local copy of the CI buffer
    renvx *localRenv;  // local copy of CI envelope
    // note: in hardware-based simulation, it is the task of the ci_announce()
    // function to request local memory using the CLance::GetReceive...()
    // functions and to copy the CI data into this memory before delivering
    // it to the Lance modules; each Lance must have its own copy of the data
    // the ci_announce() function is also responsible for calling ci_mcon()
    // if necessary

    for (area=0; area<MAXASICS; area++)
    {
        lance = listOfLances.Find (msg->receiver, area);
        if (lance != NULL)
        {
            lance->GetReceiveBuffer (msg->bufsize, &localBuf);  // get the buffer
            memcpy (localBuf, msg->buf, msg->bufsize);  // copy the buf structure

            lance->GetReceiveEnvelope (&localRenv);
            *localRenv = *renv;  // copy the renv structure

            lance->Announce (localRenv, localBuf, msg->bufsize, msg->if_num);
            // this function is responsible for returning the local memory
            // after it is not needed anymore
        }
    }

    delete renv;  // the structure is not needed anymore
}


//--------------------------------------------------------------------------
//--- fault injection
//--------------------------------------------------------------------------

// create an event of type 'CrashStatusChange', send it to the evaluation
void ACNetwork::ReportCrashChange ()
{
    SNetCrashEvent *event = new SNetCrashEvent;
    event->module = Network;
    event->networkID = GetNetworkID();
    event->type = CrashStatusChange;
    event->crashed = Crashed();
    CControlledObject::ReportEvent (event);
}


//--------------------------------------------------------------------------
//--- event reports
//--------------------------------------------------------------------------

// report a network event to the controlling object
void ACNetwork::ReportEvent (EEvent eventType)
{
    SNetEvent *event = new SNetEvent;
    event->module = Network;
    event->type = eventType;
    event->networkID = GetNetworkID();
    CControlledObject::ReportEvent (event);
}

// create a message event of given type, fill in the message identifiers
// from msg, send it to the evaluation
void ACNetwork::ReportMessageEvent (EEvent type, const SMsg *msg)
{
    SNetMessageEvent *event = new SNetMessageEvent;
    event->module = Network;
    event->type = type;
    event->networkID = GetNetworkID();
    event->sender = msg->sender;
    event->area = msg->area;
    event->receiver = msg->receiver;
    event->msgID = msg->msgID;
    CControlledObject::ReportEvent (event);
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int ACNetwork::ReportError (EReportType type, char *file, int line,
                            char *expression)
{
    return ::ReportError (type, file, line, expression, Network,
                          GetNetworkID());
    // call the global error report with module and module ID
}


