//
// netproc.cpp
//
// implementation of the base class of network processes that are used to
// delay messages at the sender/receiver 
//
// created: 09.10.96
//
// last change: 25.08.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include "error.hpp"
#include "msgtemp.hpp"
#include "netproc.hpp"


//**************************************************************************
// class ACNetworkProcess: this class is a process class and is needed as
//     a base class for CDelayer and CMessenger; it can store a message
//     together with a simulation time and reschedule itself for the
//     earliest of these times
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

ACNetworkProcess::ACNetworkProcess()
{
    initCalled_ = False;
    ACNetworkProcess::Init();
}


ACNetworkProcess::~ACNetworkProcess()
{
    ACNetworkProcess::Dispose();
}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

// discards all messages, suspends process if necessary
void ACNetworkProcess::Reset()
{
    Dispose();
    Init();
}


//--------------------------------------------------------------------------
//--- message delivery
//--------------------------------------------------------------------------

// inserts the message into its internal message queue
// assert: msg != NULL
void ACNetworkProcess::Insert (SMsg *msg, double afterTime)
{
    Assert (msg != NULL);  // only valid messages are stored

    SMcb *newMsg = new SMcb;

    newMsg->msg = msg;
    newMsg->deliveryTime = CurrentTime() + afterTime;
    // prepare the new message structure

    if (msg_ == NULL)  // have to schedule at once
    {
        msg_ = newMsg;  // store the new msg as the current one
        ReActivateAt (msg_->deliveryTime);  // schedule the message
    }
    else // there may be other messages that are earlier
    {
        if (msg_->deliveryTime <= newMsg->deliveryTime)
        // the message is not earlier than the scheduled one
        {
            listOfMessages->Insert (newMsg, newMsg->deliveryTime);
            // store the message in the message queue
        }
        else // the new msg is earlier than the scheduled one
        {
            listOfMessages->Insert (msg_, msg_->deliveryTime);
            // store the current message in the message queue

            msg_ = newMsg;
            ReActivateAt (msg_->deliveryTime);  // schedule the new message
        }
    } // else
}


//--------------------------------------------------------------------------
//--- go through the list of messages
//--------------------------------------------------------------------------

SMsg *ACNetworkProcess::GetFirst()
{
    SMcb *mcb = listOfMessages->GetFirst();
    if (mcb != NULL) return mcb->msg;
    else return NULL;
}


SMsg *ACNetworkProcess::GetNext()
{
    SMcb *mcb = listOfMessages->GetNext();
    if (mcb != NULL) return mcb->msg;
    else return NULL;
}


//--------------------------------------------------------------------------
//--- the main function of the process
//--------------------------------------------------------------------------

void ACNetworkProcess::Body()
{
    for (;;)  // if the body ever returns, the thread is destroyed
    {
        while ((msg_ != NULL) && (msg_->deliveryTime <= CurrentTime()))
        {
            Deliver (msg_->msg);  // do whatever has to be done with the msg
            delete msg_;  // free the memory

            msg_ = listOfMessages->RemoveFirst();  // remove the next msg
        }

        if (msg_ == NULL)  // nothing to deliver
            Cancel ();  // passivate the thread
        else
            Hold (msg_->deliveryTime-CurrentTime());
            // schedule the next message
    } // for
}


//--------------------------------------------------------------------------
//--- init, dispose
//--------------------------------------------------------------------------

// initialize everything
void ACNetworkProcess::Init()
{
    if (initCalled_) return;  // Init() has already been called
    initCalled_ = True;

    listOfMessages = new CMessageList;
    if (!Active()) msg_ = NULL;

    Assert (listOfMessages != NULL);
}

// discard dynamic objects
void ACNetworkProcess::Dispose()
{
    if (!initCalled_) return;  // Dispose() has already been called
    initCalled_ = False;

    if (!Active())  // the process is not active
    {
        Cancel();  // remove the process from the scheduler queue
        if (msg_ != NULL) delete msg_;  // active process deletes it itself
    }
    delete listOfMessages;
}


//--------------------------------------------------------------------------
//--- process state
//--------------------------------------------------------------------------

// returns True if process is active, otherwise False
EBool ACNetworkProcess::Active()
{
   if (this == current()) return True;
   else return False;
}


