//
// network.cpp
//
// definition of the derived network class
//
// note: since ci_announce_error () has not been implemented, the sender
// is not notified if a message can't be delivered
//
// created: 02.10.96
//
// last change: 02.09.98
//
// RCS version control information:
//     $Revision: 1.6 $$State: Exp $$Date: 1998/09/02 14:08:47 $
//


#include "error.hpp"
#include "idlist.hpp"
#include "network.hpp"
#include "sim_gen.hpp"
#include "sim_glob.hpp"


//**************************************************************************
// class CBroadcastNetwork: this class implements a broadcast network
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

CBroadcastNetwork::CBroadcastNetwork (CControllingObject *ctrl,
    SINT32 networkID) : ACNetwork (ctrl, networkID)
{
}


//--------------------------------------------------------------------------
//--- creation function of class CBroadcastNetwork
//--------------------------------------------------------------------------

// creates and returns a new object of type CBroadcastNetwork
ACNetwork *CBroadcastNetwork::CreateNetwork (CControllingObject *ctrl, 
    SINT32 networkID)
{
    return new CBroadcastNetwork (ctrl, networkID);
}


//--------------------------------------------------------------------------
//--- internal control of message transmission
//--------------------------------------------------------------------------

// send the message to all nodes; first delay the message for at most
// lambda seconds
// assert: msg != NULL
void CBroadcastNetwork::SendMsg (SMsg *msg)
{
    Assert (msg != NULL);

    if (distDataLambda.high + distDataLambda.constant > 0)
    {
        double l = 0;  // the initial delay due to the broadcast delay
        SMsg *delayMsg = new SMsg (*msg);  // copy the message

        l = random->RandomVar (distDataLambda);
        // compute the broadcast delay for this message

        delayer->Insert (delayMsg, l);  // delay the msg for l seconds
    }
    else Transmit (msg);  // pass on the message if there's no delay
}


// send or broadcast the message after it was delayed; it is assumed that
// listOfNodes from the base class contains all receiver nodes
// assert: msg != NULL
void CBroadcastNetwork::Transmit (SMsg *msg)
{
    SMsg *newMsg;  // the message to a specific receiver
    EBool all = False;
    STimestamp time;  // for getting the timestamps

    Assert (msg != NULL);

    ModifyCrashStatus ();  // fault injection

    if (Crashed())  // do not transmit the message
    {
        ReportMessageEvent (MessageLost, msg);
        return;
    }

    if (listOfNodes->Empty()) return;  // no nodes are connected

    if (msg->timestamped)  // the message should be timestamped
    {
        GetClockvalue (msg->sender, msg->area, time);
        // get the time of the send operation

        msg->SetTime (time);  // store the time in the msg-structure
    }

    for (SINT32 receiverID=listOfNodes->GetFirst();  // go through all nodes
         !all;
         receiverID=listOfNodes->GetNext())
    {
        if (!msg->broadcast)  // this is a dedicated message
        {
            receiverID = msg->receiver;  // ignore the node list
            all = True;  // only the receiver node gets the message
        }

        newMsg = new SMsg (*msg);  // make a copy of the message

        for (int area=0; area<MAXASICS; area++)
        {
            // this is for the test of delay measurement
            if (GetClockvalue (receiverID, area, time))
                newMsg->rtime[area] = time.ts;
        } // for

        newMsg->receiver = receiverID;
        ACNetwork::Transmit (newMsg);
        // transmit this message to the receiving node

        delete newMsg;  // the message is not needed anymore

        if (listOfNodes->EndOfList()) all = True;
    } // for
}


