//
// process.cpp
//
// interface between the SimUTC simulation and the underlying simulation
// environment C++SIM; only the needed functions are provided
// note: this file is necessary for switching between the PC (no threads)
//       and Linux (testing the thread-specific things)
//
// created: 9.11.96
//
// last change: 09.10.98
//
// RCS version control information:
//     $Revision: 1.4 $$State: Exp $$Date: 1998/10/09 14:11:14 $
//

#include <stdlib.h>  // for the NULL macro
#include "process.hpp"


const unsigned long DefaultProcessID = 0;
// for processes that don't have an ID at the time of creation

const INT32 TICKS2SEC = 100;
// one second is 100 pSOS ticks (or 1 tick is 10ms)


#ifndef LINUX_SIM
// create a dummy process interface through ACProcess


//**************************************************************************
// class ACProcess: this class is the interface to the process class of
//     C++SIM; only the functions provided by this class are needed by
//     the SimUTC processes
//**************************************************************************

//--------------------------------------------------------------------------
//--- definition of the static members
//--------------------------------------------------------------------------

double ACProcess::currTime_ = 0;
// current simulation time
ACProcess *ACProcess::currentProcess_ = NULL;
// the currently active process




//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

ACProcess::ACProcess (unsigned long nodeID)
{
    nodeID_ = nodeID;
    assocTime_ = 0;
    state_ = Passive;
}

ACProcess::~ACProcess() {}


//--------------------------------------------------------------------------
//--- from class Thread inherited thread scheduling funtions
//--------------------------------------------------------------------------

void ACProcess::Suspend()
{
    state_ = Suspended;
}


//--------------------------------------------------------------------------
//--- process (re-)scheduling or descheduling
//--------------------------------------------------------------------------

void ACProcess::Activate ()
{
    if (state_ != Passive) return;
    // only passive processes can be activated

    assocTime_ = CurrentTime();
    state_ = Scheduled;
}

void ACProcess::ReActivate ()
{
    if (state_ == Terminated) return;
    // terminated processes may not be scheduled

    if (state_ == Suspended) return;
    // suspended processes may not be scheduled either (?)

    assocTime_ = CurrentTime();
    state_ = Scheduled;
}

void ACProcess::ReActivateAt (double AtTime)
{
    if (state_ == Terminated) return;
    // terminated processes may not be scheduled

    if (state_ == Suspended) return;

    if (CurrentTime() < AtTime) assocTime_ = AtTime;
    // schedule at that time
    else assocTime_ = CurrentTime();  // schedule at once

    state_ = Scheduled;
}

void ACProcess::ReActivateDelay (double AtTime)
{
    if (state_ == Terminated) return;
    // terminated processes may not be scheduled

    if (state_ == Suspended) return;

    if (AtTime < 0) AtTime = 0;
    // delay must be non-negative

    assocTime_ = CurrentTime() + AtTime;
    // schedule after the specified delay

    state_ = Scheduled;
}

void ACProcess::Cancel()
{
    state_ = Passive;
}

void ACProcess::terminate()
{
    state_ = Terminated;
}


//--------------------------------------------------------------------------
//--- determining the process state
//--------------------------------------------------------------------------

EBool ACProcess::idle () const
{
    if ((state_ == Scheduled) || (ACProcess::current() == this))
        return False;  // process scheduled or active
    else return True;
}


//--------------------------------------------------------------------------
//--- scheduling information
//--------------------------------------------------------------------------

double ACProcess::evtime() const
{
    return assocTime_;
}

const ACProcess* ACProcess::current()
{
    return currentProcess_;
}
// changed the return value to ACProcess, but that's okay


//--------------------------------------------------------------------------
//--- simulation information
//--------------------------------------------------------------------------

double ACProcess::CurrentTime()
{
    return currTime_;
}


#ifdef DOS_SIM

//--------------------------------------------------------------------------
//--- simulation control; no part of CProcess, but needed for debugging
//--------------------------------------------------------------------------

// set the simulation time
void ACProcess::SetCurrentTime (double time)
{
    currTime_ = time;
}

// set the active process (that is returned by current())
void ACProcess::SetActiveProcess (ACProcess *process)
{
    currentProcess_ = process;
}

#endif


//--------------------------------------------------------------------------
//--- deactivation and rescheduling
//--------------------------------------------------------------------------

void ACProcess::Hold (double t)
{
    assocTime_ += t;
    // assuming the process is active right now
}

#ifndef PSOS_SIM
void ACProcess::tm_wkafter (INT32 time)
{
    Hold ((double)time/TICKS2SEC);
    // just hold the process for the specified time
}
#endif


//--------------------------------------------------------------------------
//--- management of the node ID
//--------------------------------------------------------------------------

// set/change the node ID
void ACProcess::SetNodeID (unsigned long nodeID)
{
    nodeID_ = nodeID;
}

// return the node ID
unsigned long ACProcess::GetNodeID () const
{
    return nodeID_;
}


#endif
// no LINUX_SIM


