//
// psos.cpp
//
// implementation of the classes and system calls necessary to implement the
// pSOS+ system calls that are used in the hardware based simulation and
// that have to be simulated in the software based sim.
// remove the psos.cpp and psos.hpp files and the psos.h file and use the
// real psos.h and object file when porting to the hardware
//
// created: 14.09.96
//
// last change: 09.01.98
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <stdlib.h>  // for NULL
#include <string.h>

#include "error.hpp"
#include "psos.hpp"


//**************************************************************************
// class CQueue: a FIFO queue
//               it stores a pointer to the data
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------


// set the pointers to NULL
CQueue::CQueue()
{
    first_ = NULL;
    last_ = NULL;
}

// empty the queue
CQueue::~CQueue()
{
    INT32 dummy[4];

    // delete all elements in the queue
    while (!Empty())
    {
        Receive (dummy);  // remove the message
    }
}


//--------------------------------------------------------------------------
//--- access functions
//--------------------------------------------------------------------------

// append the data to the queue
INT32 CQueue::Send (INT32 data[4], EBool first)
{
    SLink *newElement = new SLink;
    
    for (int i=0; i<4; i++) newElement->data[i] = data[i];
    newElement->next = NULL;

    if (first_ == NULL)  // the queue is empty
    {
        first_ = newElement;
        last_ = newElement;
    }
    else if (first)  // put the message as first element
    {
        newElement->next = first_;
        first_ = newElement;
    }
    else  // append the new link
    {
        last_->next = newElement;
        last_ = newElement;
    }

    return Q_NoError;  // a send always succeeds
}


// remove and return the first element in the queue;
// returns ERR_NOMSG if the queue is empty
INT32 CQueue::Receive (INT32 data[4])
{
    if (first_ != NULL)  // remove the first element from the list
    {
        SLink *obsoleteLink;

        for (int i=0; i<4; i++) data[i] = first_->data[i];
        // extract the data
        obsoleteLink = first_;  // store the pointer to the link
        first_ = first_->next;  // set first to the next element
        delete obsoleteLink;  // delete the link pointer

        if (first_ == NULL) last_ = NULL;  // set the last-pointer to NULL
        
        return Q_NoError;
    }
    else return ERR_NOMSG;
}


//--------------------------------------------------------------------------
//--- status functions
//--------------------------------------------------------------------------

// returns True if the queue is empty, False otherwise
EBool CQueue::Empty()
{
    if (first_ == NULL) return True;
    else return False;
}


//**************************************************************************
// class CQueueManager: manages named Queues and allows access to them
//**************************************************************************


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// start with an empty list
CQueueManager::CQueueManager()
{
    firstQueue_ = NULL;
}
    
// delete all queues in the list
CQueueManager::~CQueueManager()
{
    SNamedLink *obsoleteLink;
    
    while (firstQueue_ != NULL)
    {
        delete firstQueue_->queue;
        obsoleteLink = firstQueue_;
        firstQueue_ = firstQueue_->next;
        delete obsoleteLink;
    }
}


//--------------------------------------------------------------------------
//--- access functions
//--------------------------------------------------------------------------

// create a queue with that name and return its ID
// there is no check for duplicates
INT32 CQueueManager::Create (char name[4], INT32 *qid)
{
    SNamedLink *newLink = new SNamedLink;  // the new element

    strncpy (newLink->name, name, 4);  // store the name (at most 4 chars)
    newLink->name[4] = '\0';  // terminate with '\0'
    newLink->queue = new CQueue;  // create a queue

    if (firstQueue_ == NULL)  // this is the first queue that is created
    {
        newLink->queueID = 1;  // queue IDs start with 1
        newLink->next = NULL;
    }
    else  // there are other queues
    {
        newLink->queueID = firstQueue_->queueID + 1;
        // just add 1 to the last queue ID

        newLink->next = firstQueue_;
    }

    firstQueue_ = newLink;
    // store the new queue as first element in the list

    *qid = firstQueue_->queueID;

    return Q_NoError; // creation of a queue never fails
}


// return the ID of the queue with this name
// if the name is not unique, the first queue with this name is returned
INT32 CQueueManager::Ident (char name[4], INT32 *qid)
{
    SNamedLink *link = firstQueue_;
    
    while ((link != NULL) && (strncmp (link->name, name, 4) != 0))
        link = link->next;  // search for an entry with that name

    if (link == NULL) return ERR_OBJNF; // the name was not found
    else
    {
        *qid = link->queueID;
        return Q_NoError;
    }
}


// delete the queue
INT32 CQueueManager::Delete (INT32 qid)
{
    SNamedLink *link = firstQueue_;
    
    if (firstQueue_ == NULL) return ERR_OBJID;  // queue not found
    if (firstQueue_->queueID == qid)  // queue is first element
    {
        firstQueue_ = firstQueue_->next;
        delete link;
        return Q_NoError;
    }

    while ((link->next != NULL) && (link->next->queueID > qid))
        link = link->next;  // search for an entry with that ID

    if ((link->next == NULL) || (link->next->queueID != qid))
        return ERR_OBJID;  // there is no queue with this ID
    else  // delete the queue
    {
        SNamedLink *obsoleteLink = link->next;
        link->next = link->next->next;
        delete obsoleteLink;
        return Q_NoError;
    }
}


// store the message in the appropriate queue
INT32 CQueueManager::Send (INT32 qid, INT32 msg[4], EBool first)
{
    CQueue *queue;

    queue = Find (qid);

    if (queue == NULL) return ERR_OBJID;  // there is no queue with this ID
    else return queue->Send (msg, first);  // store the message
}


// remove the message from the appropriate queue and return it
INT32 CQueueManager::Receive (INT32 qid, INT32 msg[4])
{
    CQueue *queue;

    queue = Find (qid);

    if (queue == NULL) return ERR_OBJID;  // there is no queue with this ID
    else return queue->Receive (msg);  // get the message
}


//--------------------------------------------------------------------------
//--- internal access functions
//--------------------------------------------------------------------------

// search for the queue with the given ID, return a pointer to it
// if the queue is not found, NULL is returned
CQueue *CQueueManager::Find (INT32 qid)
{
    SNamedLink *link = firstQueue_;
    
    while ((link != NULL) && (link->queueID > qid))
        link = link->next;  // search for an entry with that ID

    if ((link == NULL) || (link->queueID != qid)) return NULL;
    // there is no queue with this ID
    else return link->queue;
}


//**************************************************************************
// pSOS functions - implementation
//**************************************************************************


INT32 t_mode (INT32, INT32, INT32 *old_mode)
{
    Assert (old_mode != NULL);  // input checks never hurt
    *old_mode = 0xFFFFF1FF;  // for testing purposes
    return 0;  // always succeeds
}

static CQueueManager manager;


INT32 q_create (char name[4], INT32, INT32, INT32 *qid)
{
    // create a queue, set the ID
    return manager.Create (name, qid);
}
    
INT32 q_ident (char name[4], INT32, INT32 *qid)
{
    // find the queue with this name, set its ID
    return manager.Ident (name, qid);
}

INT32 q_receive (INT32 qid, INT32, INT32, INT32 msg[4])
{
    // take out a message from the appropriate queue
    return manager.Receive (qid, msg);
}

INT32 q_send (INT32 qid, INT32 msg[4])
{
    // put a message into the appropriate queue (append)
    return manager.Send (qid, msg);
}

INT32 q_urgent (INT32 qid, INT32 msg[4])
{
    // put a message into the appropriate queue (insert in front)
    return manager.Send (qid, msg, True);
}

INT32 q_delete (INT32 qid)
{
    return manager.Delete (qid);
}

