//
// sim_gen.cpp
//
// implementation of global functions that are only needed in the simulation
//
// created: 02.10.96
//
// last change: 11.03.98
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include <string.h>  // for memcpy etc.

#include "ci.hpp"
#include "error.hpp"
#include "general.hpp"
#include "msgform.hpp"
#include "process.hpp"
#include "sim_gen.hpp"
#include "sim_glob.hpp"


extern INT32 ci_buf_size (INT16 size);
// number of bytes allocated if a buffer size of 'size' is requested in
// ci_get_tbuf(); it is the least multiple of sizeof(bufx) >= 'size'

extern int ci_get_tbuf (INT16 size, bufx **buf, int if_num, int area);
extern int ci_ret_tbuf (bufx *buf);
// the functions are called directly, since the module is only needed in
// pure simulation, where direct calls are possible


//**************************************************************************
// global functions
//**************************************************************************


// returns the ID of the node that is currently active or NoAccount if
// no node is active
// used by: ACNetwork
SINT32 GetCurrentNodeID ()
{
    const ACProcess *process = (const ACProcess *)ACProcess::current ();
    if (process != NULL) return process->GetNodeID ();  // return its ID
    else return NoAccount;  // no current ID
}


// returns the timestamp of the specified area on the given node;
// if the (node,area) is not in use, False is returned
EBool GetClockvalue (SINT32 nodeID, int area, STimestamp& time)
{
    CClockInterface *clock = listOfClocks.Find (nodeID, area);
    // find the clock

    if (clock != NULL)  // found the clock
    {
        clock->GetTimestamp (time);  // read the clock
        return True;
    }
    else return False;
}


//**************************************************************************
// structure SMsg
//**************************************************************************

SMsg::SMsg ()
{
    bufsize = 0;
    buf = NULL;  // can't request a buffer without an if_num
    broadcast = True;  // default is broadcasting
    timestamped = False;  // should always be set to True by the application
}

SMsg::SMsg (bufx *b, INT16 size, int ifcNumber, int a)
{
    broadcast = True;  // default is broadcasting
    timestamped = False;
    if_num = ifcNumber;
    area = a;
    if (b != NULL) bufsize = size;
    else bufsize = 0;
    // ignore the size if no data buffer was specified

    Assert (ci_get_tbuf (bufsize, &buf, ifcNumber, a) == 0);
    // allocate the buffer

    if ((b != NULL) && (size != 0))
    {
        memcpy (buf, b, size);
    }
}
    
SMsg::SMsg (const SMsg& m)
{
    sender = m.sender;
    area = m.area;
    msgID = m.msgID;
    receiver = m.receiver;
    broadcast = m.broadcast;
    timestamped = m.timestamped;
    if_num = m.if_num;
    bufsize = m.bufsize;

    for (int i=0; i<MAXASICS; i++) rtime[i] = m.rtime[i];

    Assert (ci_get_tbuf (bufsize, &buf, if_num, area) == 0);
    // allocate the buffer

    if (m.buf != NULL)
    {
        memcpy (buf, m.buf, bufsize);
    }
}

SMsg::~SMsg()
{
    if (buf != NULL) ci_ret_tbuf (buf);
}


SMsg& SMsg::operator= (const SMsg& m)
{
    if (&m != this)
    {
        sender = m.sender;
        area = m.area;
        msgID = m.msgID;
        receiver = m.receiver;
        broadcast = m.broadcast;
        timestamped = m.timestamped;
        if_num = m.if_num;

        if (buf == NULL)  // have to allocate a new buffer
        {
            bufsize = m.bufsize;  // store the new size
            Assert (ci_get_tbuf (bufsize, &buf, if_num, area) == 0);
        }
        else if (ci_buf_size(m.bufsize) != ci_buf_size(bufsize))
        {
            ci_ret_tbuf (buf);  // return the old buffer
            bufsize = m.bufsize;  // store the new size
            Assert (ci_get_tbuf (bufsize, &buf, if_num, area) == 0);
            // request a new buffer with the correct size
        }


        for (int i=0; i<MAXASICS; i++) rtime[i] = m.rtime[i];

        if (m.buf != NULL)
        {
            memcpy (buf, m.buf, bufsize);
        }
    } // if

    return *this;
}

// put the area and time into the appropriate positions in buf (the
// position is obtained from GetPosition()), shifts the rest of the data;
// if timestamped is False, the method has no effect
// 'sender', 'if_num' and 'timestamped' must be set correctly
void SMsg::SetTime (STimestamp time)
{
    if (!timestamped) return;
    // message should not be timestamped

    SMsgPosition pos;  // position and length of time and area
    INT16 newSize=0;  // new data buffer size
    int i, j;

    GetPosition (pos, sender, if_num);
    // get the position of the area and clock value

    if ((pos.length[SMsgPosition::Macrostamp] > 0) &&
        (pos.length[SMsgPosition::Timestamp] < 4))  // will leave a hole!
        Report (NTP-time is not contiguous);

    // compute the new size of the buffer (including the new data)
    for (i=0; i < SMsgPosition::EItemCount; i++)
    {
        newSize += pos.length[i];
    }
    newSize += bufsize;


    if (buf == NULL)  // allocate a buffer of the needed size
    {
        Assert (ci_get_tbuf (newSize, &buf, if_num, area) == 0);
    }
    else if (newSize > ci_buf_size (bufsize))
    // buf is too small to take all information
    {
        bufx *oldBuf = buf;  // store the old buffer
        Assert (ci_get_tbuf (newSize, &buf, if_num, area) == 0);
        // request a new buffer with the correct size
        memcpy (buf, oldBuf, bufsize);  // copy bufsize characters
        ci_ret_tbuf (oldBuf);  // return the old buffer
    }


    // copy the data into the buffer

    INT8 order[4];  // contains the order of the items
    INT32 tmppos, tmplen, tmporder;

    for (i=0; i<SMsgPosition::EItemCount; i++) order[i] = i;
    // original order, as in STimeData

    for (i=0; i < SMsgPosition::EItemCount; i++)
    for (j=1; j < SMsgPosition::EItemCount-i; j++)
    {
        if (pos.position[j-1] > pos.position[j])
        {
            // swap the entries
            tmppos = pos.position[j-1];
            tmplen = pos.length[j-1];
            tmporder = order[j-1];
            pos.position[j-1] = pos.position[j];
            pos.length[j-1] = pos.length[j];
            order[j-1] = order[j];
            pos.position[j] = tmppos;
            pos.length[j] = tmplen;
            order[j] = tmporder;
        } // if
    } // for


    INT16 itempos, itemlen;  // position and length of the data
    INT32 val32;  // the value that is to be copied
    INT8 val8;   // only needed if itemlen < 4
    INT16 val16;

    for (i=0; i<4; i++)
    {
        // set the appropriate position and length
        itempos = pos.position[i];
        itemlen = pos.length[i];

        // move the data to the back, make room for the new item
        memmove ((char*)buf+itempos+itemlen, (char*)buf+itempos,
                 bufsize-itempos);

        switch (order[i])  // extract the value
        {
          case SMsgPosition::Area:
            val32 = area;
            break;
          case SMsgPosition::Macrostamp:
            val32 = time.ms;
            break;
          case SMsgPosition::Timestamp:
            val32 = time.ts;
            break;
          case SMsgPosition::Accuracy:
            val32 = time.acc;
            break;
          default: Assert (False);  // should not happen
        } // switch

        switch (itemlen)  // convert into the data format and copy
        {
          case 0:  // no data is put into the buffer
            break;
          case 1:
            Assert (order[i] != SMsgPosition::Accuracy);
            // accuracies must have an even number of bytes
            val8 = (INT8)val32;  // use a 1 byte variable
            memcpy ((char*)buf+itempos, &val8, itemlen);
            break;
          case 2:
            if (order[i] == SMsgPosition::Accuracy)  // register is split!
            {
                val16 = (INT16)((val32>>8 & 0x0FF) + (val32>>16 & 0x0FF00));
                // use the msb of the upper and lower accuracy
            }
            else val16 = (INT16)val32;  // use a 2 byte variable
            memcpy ((char*)buf+itempos, &val16, itemlen);
            break;
          case 3:  // break into the lower two bytes and the msb
            Assert (order[i] != SMsgPosition::Accuracy);
            // accuracies must have an even number of bytes
            val16 = (INT16)val32;  // use a 2 byte variable for lower bytes
            val8 = (INT8)(val32>>16);  // use a 1 byte variable for the msb
            memcpy ((char*)buf+itempos, &val16, 2);
            memcpy ((char*)buf+itempos+2, &val8, 1);
            break;
          case 4:  // use the 4 byte variable
            memcpy ((char*)buf+itempos, &val32, itemlen);
            break;
          default: Assert (False);  // should not happen
        } // switch

        bufsize+=itemlen;  // include the new item in the buffer size
    } // for

    Assert (bufsize == newSize);
    // the buffer should be full
}



