//
// sockbase.cpp
//
// implementation of a basic socket interface class for stream/datagram
// sockets
//
// created: 07.02.97
//
// last change: 06.04.97
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include "constant.hpp"
#include "error.hpp"
#include "sockbase.hpp"
#include "socket.hpp"


//--------------------------------------------------------------------------
//--- necessary functions
//--------------------------------------------------------------------------

// returns the timeout mode for s_receive calls
static ESocketMode GetSocketMode (double timeout)
{
    if (timeout < 0) return SNoTimeout;  // block indefinitely
    else return STimeout;  // use the timeout facility
}


//**************************************************************************
// class CSocket: used as an interface to a (stream or datagram) socket;
//      can send and receive messages
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

// assume that a socket with the ID 'sid' exists and is open
// assert: addr != NULL
CSocket::CSocket (ACSocketAddress* addr, INT32 sid, EBool stream)
{
    Assert (addr != NULL);
    stream_ = stream;
    addr_ = addr;
    sid_ = sid;
    Init ();
}

// closes the socket
CSocket::~CSocket ()
{
    s_delete (sid_);
}


//--------------------------------------------------------------------------
//--- reset
//--------------------------------------------------------------------------

// reset timeout
void CSocket::Reset ()
{
    Init ();
}


//--------------------------------------------------------------------------
//--- socket information
//--------------------------------------------------------------------------

// returns the socket address
ACSocketAddress *CSocket::GetSocketAddress()
{
    return addr_;
}

// returns the file descriptor of the socket
INT32 CSocket::GetSocketID()
{
    return sid_;
}

// returns True if the socket is a stream socket, False else
EBool CSocket::IsStream ()
{
    return stream_;
}


//--------------------------------------------------------------------------
//--- control timeouts
//--------------------------------------------------------------------------

// sets the timeout (in seconds) for message (not header!) reception;
// use a negative value to block indefinitely
void CSocket::SetTimeout (double timeout)
{
    timeout_ = timeout;
}

// returns the currently used timeout (in seconds)
double CSocket::GetTimeout ()
{
    return timeout_;
}


//--------------------------------------------------------------------------
//--- internal constructor
//--------------------------------------------------------------------------

// store the address and type of the socket
// assert: addr != NULL
CSocket::CSocket (ACSocketAddress* addr, EBool stream)
{
    Assert (addr != NULL);
    stream_ = stream;  // store the socket type
    addr_ = addr;  // store the address pointer
    Init ();  // initialize the params
}


//--------------------------------------------------------------------------
//--- internal member initialization
//--------------------------------------------------------------------------

void CSocket::Init ()
{
    timeout_ = SocketTimeout;  // default timeout
}


//--------------------------------------------------------------------------
//--- internal message transmission
//--------------------------------------------------------------------------

// send the data using the appropriate s_send() function; returns True
// if the send operation was successful, else False
// assert: data != NULL
EBool CSocket::Send (void *data, INT32 size)
{
    Assert (data != NULL);

    INT32 err;  // return value of s_send()

    if (stream_) err = s_send (sid_, data, size);  // send to stream socket
    else err = s_send (sid_, addr_, data, size); // send to datagram socket

    if (err != SOkay) return False;  // send operation has failed
    else return True;  // everything okay
}


//--------------------------------------------------------------------------
//--- internal message reception
//--------------------------------------------------------------------------

// reads the data from the socket (assumes that 'data' points to a
// buffer of size '*size'); blocks according to the internal timeout;
// modifies '*size' to reflect the number of bytes read; returns True
// if the receive operation was successful, else False
// assert: data, size != NULL
EBool CSocket::Receive (void *data, INT32 *size)
{
    return CSocket::Receive (data, size, timeout_);
}

// same as Receive() above, but timeout is controlled through
// 'timeout'; please see SetTimeout() for a description of 'timeout'
EBool CSocket::Receive (void *data, INT32 *size, double timeout)
{
    Assert (data != NULL);
    Assert (size != NULL);

    INT32 len = *size;  // length of data packet read (set it to max. size)

    // read the data
    if (s_receive (sid_,GetSocketMode(timeout),timeout,data,&len) != SOkay)
        return False;  // receive operation has failed

    *size = len;  // modify size to contain the real buffer length
    return True;
}


//--------------------------------------------------------------------------
//--- internal socket ID handling
//--------------------------------------------------------------------------

// sets the socket ID
void CSocket::SetSocketID (INT32 sid)
{
    sid_ = sid;
}


