//
// socket.cpp
//
// socket interface in pSOS style
//
// created: 30.01.97
//
// last change: 02.07.98
//
// RCS version control information:
//     $Revision: 1.3 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include "error.hpp"
#include "socket.hpp"


#ifndef DOS_SIM
// sockets do not exist in DOS

#ifdef PSOS_SIM
  #include <pna.h>
#else
  #include <netinet/protocols.h>  // for IP_TCP
  #include <sys/socket.h>
  #include <sys/time.h>
  #include <sys/types.h>
  #include <sys/un.h>
  #include <fcntl.h>  // for open
  #include <stdio.h>  // for fdopen, fflush
  #include <unistd.h>  // for read, write, close and unlink
#endif

#include <string.h>  // for bzero

#include "sockaddr.hpp"

#ifdef PSOS_SIM
  #define ADDRTYPE sockaddr_in  // socket address structure
  #define DATATYPE char         // type of the data field
  #define IP_PROTNAME IPPROTO_TCP  // IP protocol name for setsockopt()
#else
  #define ADDRTYPE sockaddr
  #define DATATYPE void
  #define IP_PROTNAME IP_TCP  // IP protocol name for setsockopt()
#endif


//**************************************************************************
// useful functions
//**************************************************************************

static int Convert (ESocketType type);
// convert an enum of type ESocketType to the corresponding 'define';
// note: the enums are not set to the defines in the header file to avoid
// including any socket specific files in socket.hpp

static void SetSocketOptions (int family, int sid);
// sets options of socket 'sid' for the given family (AF_UNIX or AF_INET)
// unix options: none
// internet options: SO_REUSEADDR, TCP_NODELAY


//**************************************************************************
// Socket "System" Calls
//**************************************************************************


// create a socket, bind address to it; listen if it's of type SStream
INT32 s_create (const ACSocketAddress *addr, ESocketType type, INT32 *sid)
{
    Assert (addr != NULL);
    Assert (sid != NULL);

    int id;  // temporary (signed!) socket ID for error checks

    // create socket
    id = socket (addr->Family(), Convert (type), 0);
    if (id < 0)
    {
        Report (error creating socket);
        return SFailed;
    }


    SetSocketOptions (addr->Family(), id);
    // configure the socket


    // name socket
    if (bind (id, (ADDRTYPE *)addr->Addr(), addr->Size()) < 0)
    {
        Report (error binding socket);
        return SFailed;
    }

    // start accepting connections
    if (type != SStream)  // datagram sockets don't have to 'listen'
    {
        *sid = id;
        return SOkay;
    }

    // listen to the socket
    if (listen (id, 5) < 0)
    {
        s_delete (id, addr);  // delete the socket again
        Report (error listening to socket);
        return SFailed;
    }

    *sid = id;
    return SOkay;
}


// create a socket; connect it to the socket specified by 'addr' if it's
// of type SStream
INT32 s_ident (const ACSocketAddress *addr, ESocketType type, INT32 *sid)
{
    Assert (addr != NULL);
    Assert (sid != NULL);

    int id;  // temporary (signed!) socket ID for error checks

    // create a local socket
    id = socket (addr->Family(), Convert (type), 0);
    if (id < 0)
    {
        Report (error creating local socket);
        return SFailed;
    }


    SetSocketOptions (addr->Family(), id);
    // configure the socket


    if (type != SStream)  // datagram sockets don't have to 'connect'
    {
        *sid = id;
        return SOkay;
    }

    // connect to the socket
    if (connect (id, (ADDRTYPE *)addr->Addr(), addr->Size()) < 0)
    {
        Report (error connecting to socket);
        return SFailed;
    }

    *sid = id;
    return SOkay;
}


// accept a new connection
INT32 s_accept (INT32 sid, INT32 *accsid)
{
    int id;  // temporary (signed!) socket ID for error checks

    id = accept (sid, (ADDRTYPE *)0, (int *)0);
    if (id == -1)
    {
        Report (error accepting a connection);
        return SFailed;
    }

    *accsid = id;
    return SOkay;
}


// send 'len' bytes from the buffer pointed to by 'data'; assume a stream
// socket
INT32 s_send (INT32 sid, const void *data, INT32 len)
{
    Assert (data != NULL);

    if (send (sid, (DATATYPE *)data, len, 0) < 0)
    {
        Report (error writing to socket);
        return SFailed;
    }

    return SOkay;
}


// send 'len' bytes from the buffer pointed to by 'data'; assume
// a datagram socket with address 'addr'
INT32 s_send (INT32 sid, const ACSocketAddress *addr, const void *data,
              INT32 len)
{
    Assert (addr != NULL);
    Assert (data != NULL);

    if (sendto (sid, (DATATYPE *)data, len, 0, (ADDRTYPE *)addr->Addr(),
                addr->Size()) < 0)
    {
        Report (error sending to socket);
        return SFailed;
    }
    else return SOkay;
}


// read at most 'len' bytes into the buffer pointed to by 'data';
// the 'mode' and 'timeout' allow non-blocking calls: if mode==STimeout,
// timeout specifies the timeout in seconds (timeout==0 means polling)
// with a resolution of microseconds;
// len is modified to represent the real size that was read
INT32 s_receive (INT32 sid, ESocketMode mode, double timeout, void *data,
                 INT32 *len)
{
    Assert (data != NULL);
    Assert (len != NULL);

    int realLen;  // the real data size that was read

    if (mode == STimeout)
    // wait at most 'timeout' seconds for the socket to get ready
    {
        fd_set fds;  // for select()
        timeval tv;  // for the timeout
        int err;

        FD_ZERO (&fds);
        FD_SET (sid, &fds);

        tv.tv_sec  = (long)timeout;  // extract the seconds
        tv.tv_usec = (long)(timeout*1e6) % (long)1e6;
        // extract the microseconds

        // now wait for the socket to become readable
        err = select (sid+1, &fds, 0, 0, &tv);

        if (err == -1)
        {
            Report (error using select);
            return SFailed;
        }
        if (err == 0) return SNoPendingData;  // timeout
    } // if

    // okay, data is available or call should be blocking anyway -
    // so just read from the socket

    if ((realLen = recv (sid, (DATATYPE *)data, *len, 0)) < 0)
    {
        Report (error reading from socket);
        return SFailed;
    }
    if (realLen == 0) return SEOF;  // EOF was encountered

    *len = realLen;  // store the size of the data
    return SOkay;
}


// close an existing socket; if addr != NULL and it's AF_UNIX, unlink it
INT32 s_delete (INT32 sid, const ACSocketAddress *addr)
{
    close (sid);  // close the socket

#ifndef PSOS_SIM
    if ((addr != NULL) && (addr->Family() == AF_UNIX))
    // have to unlink the socket as well
    {
        unlink (((sockaddr_un*)addr->Addr())->sun_path);
        // unlink the socket
    }
#endif

    return SOkay;
}


//**************************************************************************
// useful functions
//**************************************************************************


// convert an enum of type ESocketType to the corresponding 'define'
static int Convert (ESocketType type)
{
/**/// explicit convertion from ESocketType to the defines; better solution?
    switch (type)
    {
        case SStream:
            return SOCK_STREAM;
            break;
        case SDatagram:
            return SOCK_DGRAM;
            break;

        default:
            Assert (False);  // unknown type
    } // switch

    return -1;  // to satisfy pedantic compilers, will never execute
}


// sets options of socket 'sid' for the given family (AF_UNIX or AF_INET)
// unix options: none
// internet options: SO_REUSEADDR, TCP_NODELAY
static void SetSocketOptions (int family, int sid)
{
#ifndef PSOS_SIM  // pSOS does not know UNIX sockets

    // currently, there are no UNIX socket options we want to set
    if (family == AF_UNIX)
    {
    } // if

#endif

    // internet sockets are more difficult to use, since there are lots
    // of timeouts before somethings happens; fortunately, many properties
    // can be configured via setsockopt()
    if (family == AF_INET)
    {
        // set socket options to allow reusing a socket that is still in its
        // FIN_WAIT state
        char sopt = 1;
        if (setsockopt(sid,SOL_SOCKET,SO_REUSEADDR,&sopt,sizeof(sopt)) < 0)
        {
            Report (error setting socket option SO_REUSEADDR);
        }

        // disable the Nagle algorithm -> messages are not kept until there
        // is a sufficient amount of data, but sent at once
        // note: this may solve the problem that messages are sometimes
        // delayed until the receive timeout occurs; but the problem has not
        // occured for some time anyway; if it still occurs even with this
        // code, add a note here!
        sopt = 1;
        if (setsockopt(sid, IP_PROTNAME, TCP_NODELAY, &sopt, sizeof(sopt)) < 0)
        {
            Report (error setting socket option TCP_NODELAY);
        }
    } // if
}


#else
// DOS_SIM: we do not use the sockets in winsock.h

#include <string.h>  // for memcpy

// use queues to store the data
static int sockname = 0;  // for obtaining a unique socket name


//**************************************************************************
// Socket "System" Calls for DOS
// these calls are only useful for testing the CSocketInterface class, they
// do *not* establish a real connection
//**************************************************************************

INT32 s_create (const ACSocketAddress *, ESocketType, INT32 *sid)
{
    char sname[2*sizeof(int)+2];  // name of the socket
    // note: the name is only unique for #sockets < 16^3 (the first place
    // is always 'S' to avoid confusion with the real queues)

    sname[0] = 'S';
    itoa (sockname, sname+1, 16);  // get the socket name
    q_create (sname, 0, 0, sid);  // create the appropriate queue

    return SOkay;
}

INT32 s_ident (const ACSocketAddress *addr, ESocketType type, INT32 *sid)
{
    return s_create (addr, type, sid);
    // creation and identification both make a new socket
}

INT32 s_accept (INT32, INT32 *accsid)
{
    return s_create (NULL, SStream, accsid);
    // make a new socket; relies on the fact that the 'addr' is not used!
/**/// should maybe check if 'sid' is a valid socket ID?
}

INT32 s_send (INT32 sid, const void *data, INT32 len)
{
    Assert (data != NULL);

    INT32 msg[4];
    char *buf = new char[len];
    memcpy (buf, data, len);
    msg[0] = len;
    msg[1] = (INT32)buf;
    q_send (sid, msg);  // store the data in the queue

    return SOkay;
}

INT32 s_send (INT32 sid, const ACSocketAddress *, const void *data,
              INT32 len)
{
    return s_send (sid, data, len);
    // no difference between datagram and stream 'sockets' in DOS
}

INT32 s_receive (INT32 sid, ESocketMode mode, double, void *data,
                 INT32 *len)
{
    INT32 msg[4];

    if (q_receive (sid, 0, 0, msg) != 0)
    {
        Assert (mode != SNoTimeout);
        // empty queue should never happen(?)
        return SNoPendingData;
    }
    else
    {
        INT32 buflen;  // length of the data that is being copied
        if (*len < msg[0]) buflen = *len;
        else buflen = msg[0];
        // just copy as much of the data as is possible
        memcpy (data, (char*)msg[1], buflen);  // copy the data
        *len = buflen;  // return the size
        if (buflen < msg[0]) // there is some data left - put it back
        {
            char *buf = new char[msg[0] - buflen];
            memcpy (buf, (char*)msg[1] + buflen, msg[0] - buflen);
            msg[0] = msg[0] - buflen;  // store the new length
            delete[] (char*)msg[1];  // free the memory
            msg[1] = (INT32)buf;
            q_urgent (sid, msg);  // store data as first item in the queue
        } // if
        else delete[] (char*)msg[1];  // free the memory
        return SOkay;
    } // else
}

INT32 s_delete (INT32 sid, const ACSocketAddress *)
{
    char data[200];  // let's hope that is enough
    INT32 len = 200;
    while (s_receive (sid, STimeout, 0, data, &len) == SOkay)
    {
        len = 200;  // simply ignore the data
    }
    q_delete (sid);  // delete the (now empty) queue
    return SOkay;
}


#endif


