//
// sockinet.cpp
//
// implementation of the internet domain socket address
//
// note: in project SimUTC, if the client runs under Linux and uses
// CInternetAddress (host, port) as a ctor, the program crashes at some
// later time during memory allocation
// the crash can only be avoided if the call to gethostbyname() is commented
// out and the IP address is set directly (just setting the IP address and
// commenting out the call to memcpy() did not help, you really have to
// remove the call to gethostbyname())
// this is quite probably a bug somewhere else in the client code, since it
// never occured before; however, if it happens in another client as well,
// just find out the IP address of the server and use the ctor which takes
// the IP address, p.e., use CInternetAddress (0x808237A6, port) instead
// of CInternetAddress ("cantor", port)
//
// note: in pSOS, you can savely use the host name since gethostbyname() is
// supplied by this file
// note: the ctor has not been tested under DOS
//
// created: 06.02.97
//
// last change: 03.06.98
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/06/18 08:09:44 $
//


#ifdef PSOS_SIM
  #include <pna.h>
#else
  #include <netinet/in.h>
  #include <sys/socket.h>
  #include <netdb.h>
#endif
#include <string.h>

#include "error.hpp"
#include "sockinet.hpp"


#ifdef PSOS_SIM
// PSOS does not know the gethostbyname() function and its hostent
// structure, so we have to supply both

struct hostent
{
    void *h_addr;   // IP address
    INT8 h_length;  // length of the address (4 bytes)
};

static struct hostent *gethostbyname (const char *host);
// only the fields 'h_addr' and 'h_length' are set
// TAKE CARE: the function is *not* reentrant! you should use the
// h_addr and h_length fields before calling the function again!

#endif


//**************************************************************************
// class CInternetAddress: internet domain address
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CInternetAddress::CInternetAddress ()
{
    addr_ = new sockaddr_in;  // make a new structure
    addr_->sin_family = AF_INET;  // family is Internet
    addr_->sin_addr.s_addr = INADDR_ANY;  // receive from all addresses
    addr_->sin_port = 0;  // let the port be assigned
}

CInternetAddress::CInternetAddress (int port)
{
    addr_ = new sockaddr_in;  // make a new structure
    addr_->sin_family = AF_INET;  // family is Internet
    addr_->sin_addr.s_addr = INADDR_ANY;  // receive from all addresses
    addr_->sin_port = htons (port);  // set the service port
}

CInternetAddress::CInternetAddress (char *host, int port)
{
    Assert (host != NULL);

    struct hostent *hp;  // for obtaining the host address
    addr_ = new sockaddr_in;  // make a new structure

    hp = gethostbyname (host);
    Assert (hp != NULL);

    memcpy (&addr_->sin_addr, hp->h_addr, hp->h_length);  // copy the addr

    addr_->sin_family = AF_INET;  // family is Internet
    addr_->sin_port = htons (port);
}

CInternetAddress::CInternetAddress (unsigned long ipaddress, int port)
{
    addr_ = new sockaddr_in;  // make a new structure
    addr_->sin_addr.s_addr = htonl (ipaddress);  // store the IP address
    addr_->sin_family = AF_INET;  // family is Internet
    addr_->sin_port = htons (port);
}

CInternetAddress::CInternetAddress (const CInternetAddress& inet)
{
    addr_ = new sockaddr_in;
    memcpy (addr_, inet.addr_, inet.Size());
    // copy the whole structure
}

CInternetAddress::~CInternetAddress ()
{
    delete addr_;
}


//--------------------------------------------------------------------------
//--- necessary functions
//--------------------------------------------------------------------------

// returns the size of the address structure
int CInternetAddress::Size () const
{
    return sizeof (sockaddr_in);
}

// returns the address family
int CInternetAddress::Family () const
{
    return addr_->sin_family;
}

// returns the address structure itself
sockaddr* CInternetAddress::Addr() const
{
    return (sockaddr*)addr_;
}


#ifdef PSOS_SIM
//**************************************************************************
// local functions
//**************************************************************************

// the 'hosts' and 'haddress' tables must correspond; changes should always
// affect both tables alike

static const int NumberOfHosts = 50;
// maximum number of hosts stored in the tables


static char *hosts[NumberOfHosts];
// hosts: all known host names

static INT32 haddress[NumberOfHosts];
// haddress: the addresses of all known hosts


struct SHostent : public hostent
{
    SHostent() {h_addr = NULL; h_length = 4; }
    SHostent (INT32 *address) { h_addr = address; h_length = 4; }

    void Set (INT32 *address) { h_addr = address; }
};


struct SHostManager
{
public:
    SHostManager()  // init. the 'hosts' and 'haddress' tables
    {
        index_ = 0;  // start with first entry

        // set all known hosts and their addresses
        Set ("boole.auto.tuwien.ac.at", 0x80823722);
        Set ("boole", 0x80823722);
        Set ("cantor.auto.tuwien.ac.at", 0x808237A6);
        Set ("cantor", 0x808237A6);
        Set ("escher.auto.tuwien.ac.at", 0x80823721);
        Set ("escher", 0x80823721);
        //Set ("gauss.hannovermesse.de", 0xC3F3143C);
        //Set ("gauss", 0xC3F3143C);  // gauss, hannovermesse.de
        Set ("gauss.auto.tuwien.ac.at", 0x808237A3);
        Set ("gauss", 0x808237A3);
        Set ("maxwell.auto.tuwien.ac.at", 0x808237A2);
        Set ("maxwell", 0x808237A2);

        // this statement must be executed after the Set() calls
        if (index_ < NumberOfHosts) hosts[index_] = NULL;
    }

    void Set (const char *hostname, INT32 address)  // store name and address
    {
        Assert (index_ < NumberOfHosts);  // the array must be big enough
        Assert (hostname != NULL);

        hosts[index_] = new char[strlen(hostname)+1];  // alloc memory
        strcpy (hosts[index_], hostname);  // copy the hostname into it
        haddress[index_] = address;  // store the address
        index_++;  // increment array index
    }

    ~SHostManager()  // delete the hosts entries
    {
        for (int i=0; i<index_; i++) delete hosts[i];
    }


private:
    int index_;  // first unused array index
};

static SHostManager hostManager_;
// unfortunately, pSOS does not initialize the host and haddress tables
// (it sets them to zero instead), so the initialization has to be done
// manually; the hostManager does the initialization in its ctor


// only the fields 'h_addr' and 'h_length' are set
// TAKE CARE: the function is *not* reentrant! you should use the
// h_addr and h_length fields before calling the function again!
static struct hostent *gethostbyname (const char *host)
{
    static SHostent hoststruct;
    // this is static, and is returned if the host is found; therefore,
    // the function is not reentrant!

    Assert (host != NULL);

    for (int i=0; (i<NumberOfHosts) && (hosts[i]!=NULL); i++)
    {
        if (strcmp (hosts[i], host) == 0)  // found it!
        {
            hoststruct.Set (haddress+i);
            return &hoststruct;  // return the hostent structure
        } // if
    } // for

    // check if the 'host' is an internet address
    

    return NULL;  // unknown host
}

#endif

