//
// sockstre.cpp
//
// implementation of the basic stream socket interface class, and of the
// access classes for client and server stream sockets
//
// created: 07.02.97
//
// last change: 29.07.98
//
// RCS version control information:
//     $Revision: 1.8 $$State: Stable $$Date: 1998/07/29 16:23:44 $
//


#include "constant.hpp"
#include "error.hpp"
#include "socket.hpp"
#include "sockstre.hpp"
#include "structs.hpp"


static const INT16 ByteOrderTest = 0x1234;
// is sent to check if the byte order on the machine of the sender is the
// same as that on the machine of the receiver

// for converting the byte order of the 'size' parameter if necessary
static INT32 Convert (INT32 src)
{
     return ((src >> 24) & 0x000000FF) +
            ((src >>  8) & 0x0000FF00) +
            ((src <<  8) & 0x00FF0000) +
            ((src << 24) & 0xFF000000);
}


//**************************************************************************
// class CStreamInterface: used as an interface to a stream socket;
//      uses the communication protocol specified for the IFC between
//      supervisors and the evaluation
//      can send and receive messages (which consist of a message header, a
//      message ID, a size parameter and the message body itself), manages
//      the creation/deletion of the socket, keeps track of message IDs
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

// assume that a stream socket with the ID 'sid' exists and is open
CStreamInterface::CStreamInterface (ACSocketAddress* addr, INT32 sid)
    : CSocket (addr, sid, True)
{
    byteOrderReversed_ = False;
    Init();  // initialize the members
}


//--------------------------------------------------------------------------
//--- message ID management
//--------------------------------------------------------------------------

// returns the ID of the last message that has been received or
// NoAccount, if no message has ever been received
SINT32 CStreamInterface::GetMessageID () const
{
    return receiveMsgID_;
}


//--------------------------------------------------------------------------
//--- information about the byte order
//--------------------------------------------------------------------------

// returns True if the byte order of the communcation partners is not
// the same (use Send/ReceiveByteOrder() to find out about that), False
// if the order is the same or if it was never checked
EBool CStreamInterface::ByteOrderReversed () const
{
    return byteOrderReversed_;
}


//--------------------------------------------------------------------------
//--- message transmission
//--------------------------------------------------------------------------

// transmits the type, ID and structure over the socket (the ID is
// inserted automatically), returns the message ID or NoAccount (-1)
// if the send operation failed
// assert: data != NULL
INT32 CStreamInterface::SendMessage (EMessageType type, SBase *data)
{
    Assert (data != NULL);
    EBool result;  // result of the send operations

    result = SendMessageHeader (type);  // send the message
#ifdef LINUX_SIM
    operator &= (result, SendMessageID ());
    operator &= (result, SendMessage ((char*)data+sizeof(data->size),
                                      data->size - sizeof(data->size)));
    // note: g++ does not use the operator |= if it is called directly;
    // but if is is called explicitly, then the operator is used
#else
    result &= SendMessageID ();
    result &= SendMessage ((char*)data+sizeof(data->size),
                           data->size - sizeof(data->size));
#endif

    if (result) return sendMsgID_-1;  // return the message ID
    else return NoAccount;  // or NoAccount, if the operation has failed
}


//--------------------------------------------------------------------------
//--- message reception
//--------------------------------------------------------------------------

// get the first datum of a message (default: non-blocking); the call
// blocks according to 'timeout'; return True if successful;
// assert: type != NULL
EBool CStreamInterface::ReceiveMessageHeader (EMessageType *type,
                                             double timeout)
{
    Assert (type != NULL);

    INT8 header;
    INT32 len = sizeof(INT8);

    // get the header
    if (!Receive (&header, &len, timeout)) return False;  // call failed

    *type = (EMessageType)header;
    return True;
}

// the function assumes that the message ID is the first item in the
// socket, followed by a structure derived from SBase and that the
// first member in SBase is 'INT32 size', denoting the size of the
// structure (without the 'size' member); it allocates the appropriate
// amount of memory, reads the structure and returns a pointer to the
// memory in 'data'
// note: the memory *must* be returned using ReturnMessageBuffer()!
// assert: data != NULL
EBool CStreamInterface::ReceiveMessage (SBase **data)
{
    Assert (data != NULL);

    INT32 msgsize;
    INT32 len = sizeof(INT32);
    INT32 oldMsgID = receiveMsgID_;  // if it has to be restored

    // first get the message ID
    if (!ReceiveMessageID (GetTimeout())) return False;

    // now read the size (which must be at least the size of SBase if
    // one adds the size of the 'size' variable)
    if (!Receive (&msgsize, &len) || (len < sizeof(INT32)) ||
        (msgsize+sizeof(INT32) < sizeof(SBase)))
    {
        receiveMsgID_ = oldMsgID;  // restore the message ID
        return False;
    }

    if (ByteOrderReversed()) msgsize = Convert(msgsize);
    // the 'msgsize' has to be reverted

    *data = (SBase *)GetMessageBuffer (msgsize+sizeof(INT32));// get buffer
    Assert (*data != NULL);

    (*data)->size = msgsize+sizeof(INT32); // store the size of the struct

    // read the data (but leave out the first 4 bytes, which contain size)
    // the data may be split in several packets
    len = msgsize;  // length of the expected message
    INT32 index = sizeof(INT32);  // starting index
    EBool okay;  // the result of the call to Receive()

    while ((msgsize > 0) &&  // we still expect some data
           (okay = Receive ((char*)(*data)+index, &len)) &&  // recv new data
           (len < msgsize))  // the message is still not complete
    {
        msgsize = msgsize-len;  // reduce the length of the expected message
        index = index+len;   // set the starting index for the next portion
        len = msgsize;       // set the length for Receive()
    } // while

    if (!okay)  // we encountered an error during receive
    {
        receiveMsgID_ = oldMsgID;  // restore the message ID
        ReturnMessageBuffer (*data);  // return the buffer
        return False;
    }

    return True;
}


//--------------------------------------------------------------------------
//--- message reply
//--------------------------------------------------------------------------

// send a MsgReply to the last received message over the socket,
// containing the result (data of type SBase) of a message;
// returns True if successful, else False
// assert: data != NULL
// assert: a message has been received
EBool CStreamInterface::ReplyToMessage (SBase *data)
{
    Assert (data != NULL);
    Assert (GetMessageID() != NoAccount);
    // must have received a message to reply to

    EBool result;  // result of the send operations

    result = SendMessageHeader (MsgReply);
#ifdef LINUX_SIM
    operator &= (result, SendMessageID (GetMessageID()));
    operator &= (result, SendMessage ((char*)data+sizeof(data->size),
                                      data->size-sizeof(data->size)));
    // note: g++ does not use the operator |= if it is called directly;
    // but if is is called explicitly, then the operator is used
#else
    result &= SendMessageID (GetMessageID());  // reply to the last message
    result &= SendMessage ((char*)data+sizeof(data->size),
                           data->size-sizeof(data->size));
#endif
    return result;
}


//--------------------------------------------------------------------------
//--- special messages
//--------------------------------------------------------------------------

// send the resync byte to the evaluation
EBool CStreamInterface::SendSynchronizationByte()
{
    return SendMessageHeader (MsgSynchronizationByte);
}

// ask the evaluation to resynchronize
EBool CStreamInterface::Resynchronize()
{
    return SendMessageHeader (MsgResynchronize);
}

// send the module type (cast to INT8); returns True if the message
// was sent successfully, else False
// use ReceiveModule() to get the module of the communication partner
EBool CStreamInterface::SendModule (EModule module)
{
    EBool result;
    INT8 castedModule = (INT8)module;

    result = SendMessageHeader (MsgModule);  // send the header
#ifdef LINUX_SIM
    operator &= (result, Send ((void*)&castedModule, sizeof(INT8)));
    // note: g++ does not use the operator |= if it is called directly;
    // but if is is called explicitly, then the operator is used
#else
    result &= Send ((void*)&castedModule, sizeof(INT8));  // and module
#endif
    return result;
}

// receive the module; reception of the message header blocks according
// to 'timeout', reception of the INT8 uses the internal timeout;
// if the message was received successfully, 'module' will be modified
// to contain the module type that was received and True is returned;
// else False is returned and 'module' is not changed
EBool CStreamInterface::ReceiveModule (EModule& module, double timeout)
{
    EBool result;
    EMessageType header;
    INT8 castedModule = 0;
    INT32 len = sizeof(INT8);

    result = ReceiveMessageHeader (&header, timeout);  // read the header
    if (!result || (header != MsgModule)) return False;

    result = Receive (&castedModule, &len);  // then the INT8
    if (!result || (len != sizeof(INT8))) return False;

    module = (EModule)castedModule;  // set the module value

    return result;
}

// send the byte order test
EBool CStreamInterface::SendByteOrder()
{
    EBool result;

    result = SendMessageHeader (MsgByteOrder);  // send the header
#ifdef LINUX_SIM    
    operator &= (result, Send ((void*)&ByteOrderTest, sizeof(INT16)));
    // note: g++ does not use the operator |= if it is called directly;
    // but if is is called explicitly, then the operator is used
#else
    result &= Send ((void*)&ByteOrderTest, sizeof(INT16));  // and INT16
#endif
    return result;
}

// receive the byte order test; message header reception blocks
// according to 'timeout' (in seconds), the reception of the INT16
// blocks according to the internal timeout; if the order is reversed,
// byteOrderReversed is set to True; returns True if the message was
// received successfully, else False
EBool CStreamInterface::ReceiveByteOrder (double timeout)
{
    EBool result;
    EMessageType header;
    INT16 byteOrder = 0;
    INT32 len = sizeof(INT16);

    result = ReceiveMessageHeader (&header, timeout);  // read the header
    if (!result || (header != MsgByteOrder)) return False;

    result = Receive (&byteOrder, &len);  // then the INT16
    if (!result || (len != sizeof(INT16))) return False;

    if (byteOrder == ByteOrderTest) byteOrderReversed_ = False;
    else  byteOrderReversed_ = True;

    return result;
}


//--------------------------------------------------------------------------
//--- message buffer handling
//--------------------------------------------------------------------------

// allocates 'size' bytes and returns a pointer to the memory
void *CStreamInterface::GetMessageBuffer (INT32 size)
{
    Assert (size <= MaxMessageSize);
    return new INT8[size];
}

// deallocates the buffer; assumes that it has been allocated using
// GetMessageBuffer()
// note: buffers obtained by ReceiveMessage() *must* be returned using
// this function
void CStreamInterface::ReturnMessageBuffer (void *buffer)
{
    delete[] ((INT8*)buffer);
}


//--------------------------------------------------------------------------
//--- internal constructor
//--------------------------------------------------------------------------

// store the address of the socket
CStreamInterface::CStreamInterface (ACSocketAddress* addr)
    : CSocket (addr, True)
{
    byteOrderReversed_ = False;
    Init();  // initialize the members
}


//--------------------------------------------------------------------------
//--- internal member initialization
//--------------------------------------------------------------------------

void CStreamInterface::Init ()
{
    sendMsgID_ = 0;  // message IDs start with 0
    receiveMsgID_ = NoAccount;  // no message was received
    CSocket::Init();  // initialize members of the base class
}


//--------------------------------------------------------------------------
//--- internal message transmission
//--------------------------------------------------------------------------

// send the first byte of the message (ie. the type)
EBool CStreamInterface::SendMessageHeader (EMessageType type)
{
    INT8 msg = type;
    return Send (&msg, 1);
}

// send the (default) message ID
EBool CStreamInterface::SendMessageID ()
{
    if (!SendMessageID (sendMsgID_)) return False;

    sendMsgID_++;  // go to the next message ID
    return True;
}

// send a special message ID
EBool CStreamInterface::SendMessageID (INT32 id)
{
    INT32 convid;  // considers byte order

    if (ByteOrderReversed()) convid = Convert (id);
    else convid = id;

    return Send (&convid, sizeof (convid));
}

// send the size and data
EBool CStreamInterface::SendMessage (void *data, INT32 size)
{
    Assert (data != NULL);

    EBool result;  // result of the send operations
    INT32 convsize;  // converted size of the message

    if (ByteOrderReversed()) convsize = Convert (size);
    else convsize = size;

    result = Send (&convsize, sizeof(INT32));  // send the length
#ifdef LINUX_SIM
    operator &= (result, Send (data, size));
    // note: g++ does not use the operator |= if it is called directly;
    // but if is is called explicitly, then the operator is used
#else
    result &= Send (data, size);  // send the data
#endif

    return result;
}


//--------------------------------------------------------------------------
//--- internal message ID reception
//--------------------------------------------------------------------------

// receive the message ID (default: non-blocking); the call blocks
// according to 'timeout'; returns True if successful
EBool CStreamInterface::ReceiveMessageID (double timeout)
{
    INT32 len = sizeof(receiveMsgID_);  // expected length
    INT32 id;  // may have to be converted

    if (!Receive (&id, &len, timeout)) return False;
    // get the data from the socket

    if (ByteOrderReversed()) receiveMsgID_ = Convert (id);
    else receiveMsgID_ = id;
    // set the receiveMsgID, convert it if necessary

    return True;
}



//**************************************************************************
// class CStreamServer: creates a stream socket, listens to it, can accept
//      connections
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CStreamServer::CStreamServer (ACSocketAddress* addr)
    : CStreamInterface (addr)
{
    INT32 sid;
    Assert (s_create (addr, SStream, &sid) == SOkay);
    // create the socket
    SetSocketID (sid);  // store the ID
}

// close the socket (unlink it if necessary)
CStreamServer::~CStreamServer ()
{
    s_delete (GetSocketID(), GetSocketAddress());
    // delete the socket, maybe unlink it
}


//--------------------------------------------------------------------------
//--- accept connections
//--------------------------------------------------------------------------

// accepts a connection; returns NULL if the accept has failed, calls
// ReceiveByteOrder() if the accept has succeeded
CStreamInterface *CStreamServer::Accept()
{
    INT32 sid;

    if (s_accept (GetSocketID(), &sid) != SOkay) return NULL;
    // accept a new connection

    CStreamInterface *s = new CStreamInterface (GetSocketAddress(), sid);
    // make a new socket

    Assert(s->ReceiveByteOrder (SocketTimeout)); // get client byte order
    // note: the first statement causes the client to revert its received
    // msg size; the second statement causes the local ReceiveMessage() to
    // revert the msg size if necessary;

    return s;  // return the new socket
}



//**************************************************************************
// class CStreamClient: identifies a stream socket, connects to it
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor
//--------------------------------------------------------------------------

CStreamClient::CStreamClient (ACSocketAddress* addr)
    : CStreamInterface (addr)
{
    Assert (Connect (addr, MaxTrials));
    // establishing the connection is a must
}

CStreamClient::CStreamClient (ACSocketAddress* addr, int numberOfTrials)
    : CStreamInterface (addr)
{
    Assert (Connect (addr, numberOfTrials));
    // establishing the connection is a must
}


//--------------------------------------------------------------------------
//--- socket connection
//--------------------------------------------------------------------------

// tries to identify and to connect to the socket; gives up after
// 'numberOfTrials' attempts; if the connection could be established,
// SendByteOrder() is called
EBool CStreamClient::Connect (ACSocketAddress *addr, int numberOfTrials)
{
    INT32 sid;
    int i = 0;

    // wait until connection is accepted or till timeout
    while ((s_ident (addr, SStream, &sid) != SOkay) &&
           (++i < numberOfTrials)) ;

    if (i >= numberOfTrials) return False;
    // could not establish the connection

    SetSocketID (sid);  // store the ID

    Assert(SendByteOrder());  // send the local byte order
    // note: the first statement causes the local ReceiveMessage() to
    // revert the msg size if necessary; the second statement causes the
    // server to revert its received msg size
 
    return True;
}


