//
// supervis.cpp
//
// implementation of the synchronization supervisor
//
// created: 06.12.96
//
// last change: 11.03.98
//
// RCS version control information:
//     $Revision: 1.2 $$State: Stable $$Date: 1998/07/03 12:45:47 $
//


#include "clockifc.hpp"
#include "clocksyn.hpp"
#include "error.hpp"
#include "event.hpp"
#include "general.hpp"
#include "globals.hpp"
#include "lance.hpp"
#include "supervis.hpp"


//**************************************************************************
// class CSupervisor: this class controls the initialization and execution
//      of the clock synchronization; it creates all necessary moduls and
//      is their link to the evaluation
//**************************************************************************

//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CSupervisor::CSupervisor (char qname[4], ACSocketAddress *addr, SINT32 nodeID,
    int area, double clockFrequency) :
    CControllingObject (qname, addr, nodeID)
{
    nodeID_ = nodeID;
    area_ = area;
    snapshot_ = False;

    clock_ = new CClockInterface (this, clockFrequency, area_);
    cs_ = new CClockSync (this, area_);
    lance_ = new CLance (this, area_);

    Assert (clock_ != NULL);
    Assert (cs_ != NULL);
    Assert (lance_ != NULL);

    lance_->SetClockSync (cs_);  // connect the objects
    lance_->SetClock (clock_);
    cs_->SetClock (clock_);
    cs_->SetLance (lance_);

    listOfSupervisors.Insert (this, nodeID_, area_);
    // store the supervisor itself in the list of supervisors

    SendModule (Supervisor);  // send the module type to EvalSys
}


CSupervisor::~CSupervisor ()
{
    listOfSupervisors.Remove (nodeID_, area_);
    // remove the supervisor itself from the list of supervisors

    delete cs_;  // ClockSync must be deleted before Clock
    delete lance_;  // Lance must be deleted before Clock
    delete clock_;
}


//--------------------------------------------------------------------------
//--- reset of the application
//--------------------------------------------------------------------------

// set all parameters in all modules to their default values
void CSupervisor::Reset()
{
    snapshot_ = False;

    INT32 oldmode = DisableInterrupts();  // disable application interrupts
/**/// note: this should become obsolete; the Reset() functions should disable
    // interrupts if necessary

    clock_->Reset ();  // reset all modules
    cs_->Reset ();
    lance_->Reset ();

    EnableInterrupts (oldmode);  // enable application interrupts

    CControllingObject::Reset();  // reset the event queue as well
}


//--------------------------------------------------------------------------
//--- get the nodeID/area identifiers
//--------------------------------------------------------------------------

// copy the internal nodeID/area into the parameters
void CSupervisor::GetIdentification (SINT32& nodeID, int& area) const
{
    nodeID = nodeID_;
    area = area_;
}


//--------------------------------------------------------------------------
//--- suspend/resume application
//--------------------------------------------------------------------------

// suspend ClockSync and Lance
void CSupervisor::SuspendOperation ()
{
    cs_->Suspend ();  // suspend the clock sync. (disable duty interrupts)
    lance_->Suspend ();  // suspend the lance
    cs_->Suspend ();  // suspend the clock sync. (now really suspend)
    // note: the order of the three statements is important!
/**/// alternative: put lance_->Suspend() into CS::Suspend(), between
    // the disabling of the interrupt and the rest of the suspend
    // action; disadvantage: CS has to know when and how to suspend the
    // Lance
}

// resume operation
void CSupervisor::ResumeOperation ()
{
    cs_->Resume();  // let CS generate the next interrupt
}


//--------------------------------------------------------------------------
//--- return associated objects
//--------------------------------------------------------------------------

CClockInterface *CSupervisor::GetClock ()
{
    return clock_;
}

CClockSync *CSupervisor::GetClockSync ()
{
    return cs_;
}

CLance *CSupervisor::GetLance ()
{
    return lance_;
}


//--------------------------------------------------------------------------
//--- configuration of system parameters
//--------------------------------------------------------------------------

// sets the given parameter to its new value (or passes it on to one
// of the submodules); if the parameter does not exist or if the value
// is not valid, False is returned
EBool CSupervisor::SetParameter (const SParameter& param)
{
    // see if one of the objects can use the parameter
    switch (param.module)
    {
        case Supervisor:
            return CControllingObject::SetParameter (param);
            break;
        case ClockSync:
/**/        // check the type, maybe disable interrupts
            return cs_->SetParameter (param);
            break;
        case Clock:
            return clock_->SetParameter (param);
            break;
        case Lance:
            return lance_->SetParameter (param);
            break;

        default:
            return False;  // unknown parameter
    } // switch
}


// returns the value of the given parameter in the appropriate field of
// SParameter; if the parameter does not exist, False is returned (the
// parameter-fields may nevertheless be changed!)
EBool CSupervisor::GetParameter (SParameter& param) const
{
    // again, see if one of the objects can use the parameter
    switch (param.module)
    {
        case Supervisor:
            return CControllingObject::GetParameter (param);
            break;
        case ClockSync:
/**/        // check the type, maybe disable interrupts
            return cs_->GetParameter (param);
            break;
        case Clock:
            return clock_->GetParameter (param);
            break;
        case Lance:
            return lance_->GetParameter (param);
            break;

        default:
            return False;  // unknown parameter
    } // switch
}


//--------------------------------------------------------------------------
//--- execution of system commands
//--------------------------------------------------------------------------

// executes the command (or passes it on to one of the submodules);
// returns True if successfull, otherwise False
// own commands: EnableSnapshot, DisableSnapshot
EBool CSupervisor::ExecuteCommand (const SCommand& command)
{
    // see if one of the objects can use the command
    switch (command.module)
    {
        case Supervisor:
            if (command.type == EnableSnapshot)
            {
                snapshot_ = True;
                return True;
            }
            else if (command.type == DisableSnapshot)
            {
                snapshot_ = False;
                return True;
            }
            else return CControllingObject::ExecuteCommand (command);
            break;
        case ClockSync:
/**/        // check the type, maybe disable interrupts
            return cs_->ExecuteCommand (command);
            break;
        case Clock:
            return clock_->ExecuteCommand (command);
            break;
        case Lance:
            return lance_->ExecuteCommand (command);
            break;

        default:
            return False;  // unknown module
    } // switch
}


//--------------------------------------------------------------------------
//--- event logging
//--------------------------------------------------------------------------

// stores the event until it can be processed
void CSupervisor::NoteEvent (SEvent *event)
{
    Assert (event != NULL);
    Assert (event->size >= sizeof(SApplEvent));
    // events from the application modules must have that type
/**/// that's not very secure - think of something better!

    SApplEvent *appevent = (SApplEvent *)event;

    clock_->GetTimestamp (appevent->clockTime);
    // get the clock timestamp
    appevent->nodeID = nodeID_;
    appevent->area = area_;
    CControllingObject::NoteEvent (event);
}


//--------------------------------------------------------------------------
//--- periodic activities of derived classes
//--------------------------------------------------------------------------

// is called once every period; initiates a snapshot if the triggering of
// periodic snapshots is enabled
void CSupervisor::DoPeriodicActivities()
{
    if (snapshot_) MakeSystemSnapshot();
}


//--------------------------------------------------------------------------
//--- error reports
//--------------------------------------------------------------------------

int CSupervisor::ReportError (EReportType type, char *file, int line,
                              char *expression)
{
    return ::ReportError (type, file, line, expression, Supervisor,
                          nodeID_, area_);
    // call the global error report with module and nodeID/area
}

