//
// tester.cpp
//
// control process for testing the synchronization
//
// created: 19.10.96
//
// last change: 01.10.98
//
// RCS version control information:
//     $Revision: 1.4 $$State: Exp $$Date: 1998/10/01 10:19:38 $
//


#include <stdlib.h>

#ifndef RESOURCE_H_
#  include <Common/Resource.h>
#endif

#include "constant.hpp"
#include "cont_sw.hpp"
#include "error.hpp"
#include "errtask.hpp"
#include "globals.hpp"
#include "sim_glob.hpp"
#include "sockinet.hpp"
#include "sockunix.hpp"
#include "tester.hpp"


#define SIMUTC_DEBUG
// uncomment it to use the simple error handler (recommended if there
// are problems with process scheduling)


//**************************************************************************
// globals
//**************************************************************************

EBool g_finished = False;
// the 'tester' process is notified of the end of the simulation (as
// requested from EvalSys with the command KillSimulation) through
// this variable

extern double g_delayTime;
// the time that the sleeper process sleeps per second (thus delaying the
// whole simulation);
// it is set by the main() function (user definable via an option), is given
// in seconds (granularity is microseconds), and is needed to delay the
// simulation, because on a fast processor the events are so close together
// that they clog up the evaluation (XWindow-Server does not update screen
// anymore)

extern double g_sleeperPeriod;
// the period of the sleeper process
// this value specifies how often the process calls usleep() to delay
// the simulation; a big value causes the sleeper to run without delay
// for this time, and it is delayed for rather long when the sleeper
// sleeps; a small value lets the sleeper delay more often for shorter
// times - use this if you experience problems with sockets, if the
// evaluation is too slow, or if snapshot events are processed out of
// order


//**************************************************************************
// locals
//**************************************************************************

static const double WaitTime = 200.0;
// period of checking for end of simulation

static const INT32 PidTester  = 100;
static const INT32 PidSleeper = 200;
// process IDs of the controlling processes


//**************************************************************************
// class CSleeper: delays the simulation for g_delayTime seconds per
//	simulated second (only in LINUX_SIM)
//**************************************************************************

class CSleeper : public ACProcess
{
public:

    CSleeper::CSleeper () : ACProcess (PidSleeper) {}
    void Body();
};

void CSleeper::Body()
{
    for (;;)
    {
#ifdef LINUX_SIM
        if (g_finished) tm_wkafter (500*TICKS2SEC);
        // do not delay the deletion of the simulation

        if (g_delayTime > 0)
            usleep ((INT32)(g_delayTime * g_sleeperPeriod * 1e6));
        // sleep for g_delayTime*g_sleeperPeriod seconds, thus delaying the
        // whole program
        // note: usleep() takes microseconds as argument, but delays for
        //       at least 1/CLOCKS_PER_SEC seconds (constant from time.h)
        // note: this code even works if the program is started on a machine
        //       with more than one processor, since scheduling is done by
        //       C++SIM

        if (g_sleeperPeriod*TICKS2SEC < 1) tm_wkafter (1);
        else tm_wkafter ((INT32)(g_sleeperPeriod*TICKS2SEC));
        // wait g_sleeperPeriod seconds, but at least one pSOS tick
#else
#ifdef DOS_SIM
        // in DOS, there is no option to set a delay

        tm_wkafter (500*TICKS2SEC);  // wait very long
#else
        // in pSOS, this file and code should not be used!
        Assert (0);
#endif
#endif
    } // for
}


//**************************************************************************
// class CTester: sets up the simulation and starts it
//**************************************************************************

// the tester process gets a high dummy process ID
CTester::CTester () : ACProcess (PidTester)
{
}


// control the test process
void CTester::Body ()
{
    CErrorTask *errorHandler;  // error handler
#ifdef SIMUTC_DEBUG
    CErrorHandler *handler = new CErrorHandler;
    // use a simple handler in case the error task does not work
#endif
    CSoftwareController *controller;  // simulation controller
    CUnixAddress addr ("/tmp/SimUTC");
    //CInternetAddress addr ("escher.auto.tuwien.ac.at", EvalsysPort);
    //CInternetAddress addr (0x80823721, EvalsysPort);


    //----------------------------------------------------------------------
    // create and install the error handling process

    errorHandler = new CErrorTask ("QERR");  // create the error handler
    InstallErrorHandler (errorHandler);  // install it

#ifndef NO_RESOURCE
    Resource::ref (errorHandler);  // ref the process
#endif
    errorHandler->Activate();  // activate the process

#ifdef SIMUTC_DEBUG
    InstallErrorHandler (handler);
    // install the simple handler (that replaces the error task as handler)
#endif


    //----------------------------------------------------------------------
    // call the global initialization functions

    CheckAssumptions();
    InstallAlgorithms();
    InstallNetworks();


    //----------------------------------------------------------------------
    // create the simulation controller

    controller = new CSoftwareController ("CTRL", &addr);
#ifndef NO_RESOURCE
    Resource::ref (controller);  // ref the process
#endif
    controller->Activate();  // activate the process
    // that should be all one has to do; everything else goes over the
    // socket


    //----------------------------------------------------------------------
    // create the sleeper process

#ifdef LINUX_SIM
    // the sleeper does not delay in DOS anyway

    CSleeper *sleeper = new CSleeper ();
#ifndef NO_RESOURCE
    Resource::ref (sleeper);  // ref the process
#endif
    sleeper->Activate();  // activate the process

#endif  // LINUX_SIM


    //----------------------------------------------------------------------
    // start the simulation

    Scheduler::scheduler().Resume();

    Report (Tester process holding);    

    while (!g_finished)  // wait until the simulation has ended
    {
        Hold(WaitTime);
    } // while

    Report (Simulation terminating);    


    //----------------------------------------------------------------------
    // delete everything

    DeinstallErrorHandler();  // not needed, just to be on the safe side

    controller->Cancel();  // deschedule the processes
    errorHandler->Cancel();

    controller->terminate();  // terminate the controller
    errorHandler->terminate();  // terminate the error handler

    errorHandler->Flush();  // empty the error queue

#ifndef NO_RESOURCE
    Resource::unref (controller);  // unref the processes
    Resource::unref (errorHandler);
#else
    delete controller;  // or delete them
    delete errorHandler;
#endif


#ifdef LINUX_SIM
    sleeper->Cancel();  // deschedule the processes
    sleeper->terminate();  // terminate the controller
#ifndef NO_RESOURCE
    Resource::unref (sleeper);  // unref the processes
#else
    delete sleeper;  // or delete them
#endif
#endif  // LINUX_SIM


#ifdef SIMUTC_DEBUG
    delete handler;  // delete the simple error handler
#endif


    //----------------------------------------------------------------------
    // remove the scheduler

    Scheduler::scheduler().Suspend();
    Scheduler::scheduler().terminate();


#ifdef CPPSIM_V1_5 
    Thread_Type::mainResume();
#else
    Thread::mainResume();
#endif
}

void CTester::Await ()
{
    Resume();
    Thread::Self()->Suspend();
}

void CTester::Exit ()
{
#ifdef CPPSIM_V1_5
    Suspend();
    Thread_Type::Exit();
#else
    Thread::Exit();
#endif
}

