//
// time.cpp
//
// all structures, classes, functions etc. that are necessary for the
// time management
//
// created: 11.02.97
//
// last change: 05.09.97
//
// RCS version control information:
//     $Revision: 1.1 $$State: Stable $$Date: 1998/06/16 10:49:31 $
//


#include <iostream.h>

#include "time.hpp"


//**************************************************************************
// conversion functions NTP <--> seconds
//**************************************************************************


static const double TwoPow8 = 256;
// the value of bit 8 of the NTP time in seconds (bit 8 = 2^8 sec=^=4.27 min)
static const double TwoPow24 = 16777216;
// 1 second in timestamp format (timestamp: lsb is bit -24, 2^-24=^=59.6ns)
static const double TwoPow56 = 4294967296.*16777216.;
// 1/(lsb of the mircostamp)


// convert NTP time into seconds
double NtpToSeconds (INT32 ms, INT32 ts)
{
    return ((double)ms*TwoPow8 + (double)ts/TwoPow24);
}

// convert seconds into NTP
void SecondsToNtp (double sec, INT32& ms, INT32& ts)
{
    ms = (INT32)(sec/TwoPow8);
    ts = (INT32)((sec - (double)ms*TwoPow8)*TwoPow24 + 0.5);
}


// convert extended NTP time into seconds
double NtpToSeconds (INT32 ms, INT32 ts, INT32 us)
{
    return ((double)ms*TwoPow8 + (double)ts/TwoPow24 + (double)us/TwoPow56);
}

// convert seconds into extended NTP
void SecondsToNtp (double sec, INT32& ms, INT32& ts, INT32& us)
{
    ms = (INT32)(sec/TwoPow8);
    ts = (INT32)((sec - (double)ms*TwoPow8)*TwoPow24);
    us = (INT32)((sec-(double)ms*TwoPow8-(double)ts/TwoPow24)*TwoPow56 + 0.5);
}




//**************************************************************************
// class CNtpTime: stores the macrostamp, timestamp and microstamp; allows
//      arithmetic operations on the data
//**************************************************************************

static const INT32 Ceiling = 0xFFFFFFFF;


//--------------------------------------------------------------------------
//--- constructor, destructor
//--------------------------------------------------------------------------

CNtpTime::CNtpTime ()
{
    ms = 0;
    ts = 0;
    us = 0;
}

CNtpTime::CNtpTime (double seconds)
{
    SecondsToNtp (seconds, ms, ts, us);
}

// store the time
CNtpTime::CNtpTime (INT32 macro, INT32 time, INT32 micro)
{
    ms = macro;
    ts = time;
    us = micro;
}

CNtpTime::CNtpTime (const CNtpTime& time)
{
    ms = time.ms;
    ts = time.ts;
    us = time.us;
}


//--------------------------------------------------------------------------
//--- operators
//--------------------------------------------------------------------------

CNtpTime CNtpTime::operator= (const CNtpTime& time)
{
    ms = time.ms;
    ts = time.ts;
    us = time.us;
    // can't simply say *this=time, that would call operator= again
    return *this;
}

CNtpTime CNtpTime::operator+ (const CNtpTime& time) const
{
    CNtpTime result;
    INT8 carry = 0;

    if (us + time.us < us)  // overflow in the microstamp
/**/// this assumes that overflow means wrapping around; is this always
    // the case (on every machine)???
    {
        result.us = (us - (Ceiling - time.us + 1));
        carry = 1;
    }
    else
    {
        result.us = us + time.us;
        carry = 0;
    }

    if (ts + time.ts + carry < ts)  // overflow in the timestamp
    {
        result.ts = (ts - (Ceiling - time.ts + 1) + carry);
        carry = 1;
    }
    else
    {
        result.ts = ts + time.ts + carry;
        carry = 0;
    }

    result.ms = ms + time.ms + carry;

    return result;
}

CNtpTime CNtpTime::operator- (const CNtpTime& time) const
{
    CNtpTime result;
    INT8 carry = 0;

    if ((ms < time.ms) || ((ms == time.ms) && (ts < time.ts)) ||
        ((ms == time.ms) && (ts == time.ts) && (us < time.us)))
        return result;  // underflow --> return (0,0,0)


    if (us < time.us)  // take carry from ts
    {
        result.us = (Ceiling - time.us + us + 1);
        carry = 1;
    }
    else
    {
        result.us = us - time.us;
        carry = 0;
    }

    if (ts < time.ts + carry)  // take carry from ms
    {
        result.ts = (Ceiling - time.ts - carry + ts + 1);
        carry = 1;
    }
    else
    {
        result.ts = ts - time.ts - carry;
        carry = 0;
    }

    result.ms = ms - time.ms - carry;

    return result;
}

CNtpTime CNtpTime::operator+= (const CNtpTime& time)
{
    *this = *this + time;
    return *this;
}

CNtpTime CNtpTime::operator-= (const CNtpTime& time)
{
    *this = *this - time;
    return *this;
}


//**************************************************************************
// output functions
//**************************************************************************

ostream& operator<< (ostream& o, const STimestamp& time)
{
    o << "acc = " << time.acc << ", ms = " << time.ms;
    return o << ", ts = " << time.ts;
}

ostream& operator<< (ostream& o, const SExtendedTimestamp& time)
{
   return o << (STimestamp)time << ", us = " << time.us;
}

ostream& operator<< (ostream& o, const CNtpTime& time)
{
    return o << "ms = "<<time.ms<<", ts = "<<time.ts<<", us = "<<time.us;
}


